#
# syn_procs.tcl
#
# Kurt Kermes
# 2000/03/27
#
# library for Synergy JTAG functions
#
#
#
#       (C) COPYRIGHT SILICON GRAPHICS, INC.
#       UNPUBLISHED PROPRIETARY INFORMATION.
#       ALL RIGHTS RESERVED.
#



package provide syn_procs 0.0

################################################################################
# proc clk_ctrl
#
#	clkstp  : sysad | core | both | none"
#  sysadstp: force | <sync>
#  action  : nothing | gmoc | abist
#  pat(abist) : 1.5x 2x 3x 6x
#  pat(gmoc)  : 1s1x 2s1x 2s1.5x 2s2x 2s2.5x 2s3x 3s1x  
#  scanMode   : bits 7-4: chainSel
#               bit    2: LPRA access = 1 normal latch chain = 0
#               bit    1: invert chain
#               bit    0: circulate chain
################################################################################
proc clk_ctrl { clkstp sysadstp action pat {scanMode 0} } {

global modid brdidx mod brd chip chnidx

INST SC_CTL
GOINST

set scCtlTdr 0
# set clkstp parameter
if { $clkstp == "sysad" } { 
	set scCtlTdr [expr $scCtlTdr | 0x40000]
} elseif  { $clkstp == "core" } {
	set scCtlTdr [expr $scCtlTdr | 0x80000]
} elseif  { $clkstp == "both" } {
	set scCtlTdr [expr $scCtlTdr | 0xc0000]
} elseif  { $clkstp == "none" } {
	set scCtlTdr [expr $scCtlTdr | 0x00000]
} else {
	error "Invalid clock stop argument! Valid: sysad | core | both | none"
}

# set sysadstp parameter (valid only if stopping sysad)
if { $sysadstp == "force" } {
	set scCtlTdr [ expr $scCtlTdr | 0x20000]
}

set gmocSeq 0
# set action parameter
if { $action == "gmoc" } {
	if { $pat == "1s1x" } {
		set gmocSeq 0x5ff
	} elseif { $pat == "2s1x" } {
		set gmocSeq 0x55f
	} elseif { $pat == "2s1.5x" } {
		set gmocSeq 0x5af
	} elseif { $pat == "2s2x" } {
		set gmocSeq 0xacc
	} elseif { $pat == "2s2.5x" } {
		set gmocSeq 0x5f5
	} elseif { $pat == "3s1x" } {
		set gmocSeq 0x555
	} elseif { $pat == "2s3x" } {
		set gmocSeq 0x5fa
	} else { 
		error "Invalid gmoc pat! Valid: 1s1x, 2s1x, 2s1.5x, 2s2x, 2s2.5x"
	}

} elseif { $action == "abist" } {
	set scCtlTdr [expr $scCtlTdr | 0x1000]
	if { $pat == "1.5x" } {
		set gmocSeq 0x6db
	} elseif { $pat == "2x" } {
		set gmocSeq 0x333
	} elseif { $pat == "3x" } {
		set gmocSeq 0x1c7
	} elseif { $pat == "6x" } { 
		set gmocSeq 0x03f
	} else {
		error "Invalid abist pattern! Valid: 1.5x 2x 3x 6x"
	}
}

set scCtlTdr [ expr $scCtlTdr | $gmocSeq | ($scanMode << 24) ]
puts "scCtlTdr is [ format "0x%x"  $scCtlTdr]"

# scan it in
DATA $scCtlTdr
GODATA

if { [ string compare $action "nothing" ] != 0  } {
	# now do it
	INST GMOC
	GOINST
}

}

################################################################################
# set_trig - 
# sets the trigger_mask and triger_data 
# mask : if you don't care about certain bits coming out of
#        the debug port data, set the corresponding bit to 0
# data : port will trigger when data is matched -- 
#        Note: Make sure that if any MASK bit is zero, that 
#        the trigger data is also 0, otherwise it won't match
################################################################################
proc set_trig { mask data } {

global modid brdidx mod brd chip chnidx

INST TRIG_MASK
GOINST

DATA $mask
GODATA

INST TRIG_DATA
GOINST

set data [ expr $data & $mask ]
DATA $data
GODATA

}


################################################################################
# clk_count_ctrl - control the clock counter
# mode: 0: reserved; 
#       1: decrement every cycle
#       2: decrement on debug trigger
#       3: watchdog - reload when debug trigger is met
# enab: 1: enable clk counter 1
#       3: enable both clk counters
################################################################################
proc clk_count_ctrl { mode enab ncycles1hi ncycles1lo ncycles2hi ncycles2lo} {

global modid brdidx mod brd chip chnidx

set tmp 0

INST CK1_PHI
GOINST
DATA $ncycles1hi
GODATA

INST CK1_PLO
GOINST
DATA $ncycles1lo
GODATA

if { $enab != 0x1 && $enab != 0x3 } {
	error "Illegal enable setting --  only 1 or 3 is valid!!!"
}
if { $enab == 0x3 } {
	INST CK2_PHI
	GOINST
	DATA $ncycles2hi
	GODATA

INST CK2_PLO
GOINST
DATA $ncycles2lo
GODATA
}

INST SC_CTL
GOINST

if { $mode == 0 } {
	error "clk_count_ctrl: mode cannot be 0"
}

#set the Clock counter mode and clk counter enable 2
# also disable the counter 
set mode [ expr $mode & 3 ]
set tmp [ expr (($mode << 20) | ($enab & 2) << 21) ]
set tmp [ expr $tmp & 0xff3fffff] 
DATA $tmp
GODATA

# enable the counter -- start the countdown
set tmp [ expr ($tmp | (($enab & 2) << 21) | (($enab & 1) << 23) ) ] 
DATA $tmp
GODATA

}

################################################################################
# clk_count_stat - get status of clock counter
# Returns :
#    in ck1hi  - top 8 bits of value of counter 1
#    in ck1lo  - bottom 32 bits of counter
#    in status - bit 0 set is clk counter 2 is in use (is decrementing)
#    in status - bit 1 set if clk is stopped
################################################################################
proc clk_count_stat { ck1hi ck1lo status } {

global modid brdidx mod brd chip chnidx

upvar ck1hi  lck1hi
upvar ck1lo  lck1lo
upvar status lstatus

INST CK1_PHI
GOINST
GODATA
set lck1hi [ INFO CHIP $chip DATA CK1_PHI.Bits_31_downto_0 OUT ]

INST CK1_PLO
GOINST
GODATA
set lck1lo [ INFO CHIP $chip DATA CK1_PLO.Bits_31_downto_0 OUT ]

puts "clk_count_stat: hi: [format "0x%x" $lck1hi]  lo: [format "0x%x" $lck1lo] "

set lstatus [ expr ($lck1hi >> 30) & 3 ]

}


################################################################################
# pio_read - do a Local block read
# given addr and number of scans to determine timeout, 
# data is returned in datahi and datalo
# the status value is 1 on successful completion,
#                     the PIO_CTL return value if timeout occurred
################################################################################
proc pio_read { addr datahi datalo {timeout 2} } {

global modid brdidx mod brd chip chnidx
upvar datahi ldatahi 
upvar datalo ldatalo

set addr [ expr $addr & 0xffff ]

INST PIO_CTL
GOINST
DATA $addr
GODATA
#puts "pio_read: PIO_CTL done "

INST PIO
GOINST

# wait for response
INST PIO_CTL
GOINST
for { set loop 0} { $loop != $timeout } { incr loop 1 } {
	GODATA
	set status [ INFO CHIP $chip DATA PIO_CTL.Bits_31_downto_0 OUT ]
	#puts "pio_read: status is $status"
	if { [ expr $status & 0x110000 ] == 0x110000 } {
		break
	}
}
if { $loop == $timeout } {
	error "pio_read: timeout!"              
	return $status
}

INST PIO_DHI
GOINST
GODATA
set ldatahi [ INFO CHIP $chip DATA PIO_DHI.Bits_31_downto_0 OUT ]

INST PIO_DLO
GOINST
GODATA
set ldatalo [ INFO CHIP $chip DATA PIO_DLO.Bits_31_downto_0 OUT ]

#puts "pio_read: addr: [ format "0x%x" $addr ]  ldatahi = [format "0x%x" $ldatahi ]  ldatalo: [ format "0x%x" $ldatalo ]"

return 1

}


################################################################################
# pio_write - do a Local block write
# Given address, timeout, datahi and datalo, 
# writes data into given address
# errs out if PIO access didn't complete in the number of scans
#    defined in the timeout parameter
################################################################################
proc pio_write { addr datahi datalo {timeout 2} } {

global modid brdidx mod brd chip chnidx

#puts "pio_write: addr: [ format "0x%x" $addr ]  datahi: [ format "0x%x" $datahi ]  datalo: [ format "0x%x" $datalo ]"

INST PIO_CTL
GOINST
set addr [ expr $addr | 0x1000 ]		;# set the write bit
DATA $addr
GODATA

INST PIO_DHI
GOINST
DATA $datahi
GODATA

INST PIO_DLO
GOINST
DATA $datalo
GODATA

INST PIO
GOINST

# wait for response
INST PIO_CTL
GOINST
for { set loop 0 } { $loop != $timeout } { incr loop 1 } {
	GODATA
	set status [ INFO CHIP $chip DATA PIO_CTL.Bits_31_downto_0 OUT ]
	if { [ expr $status & 0x100000 ]  == 0x100000 } {
		break
	}
}
if { $loop == $timeout } {
	error "pio_write: timeout!"              
}

#puts "pio_write: complete"

}
