# 
#/**************************************************************************
# *                                                                        *
# *               Copyright (C) 1989-1999 Silicon Graphics, Inc.           *
# *                                                                        *
# *  These coded instructions, statements, and computer programs  contain  *
# *  unpublished  proprietary  information of Silicon Graphics, Inc., and  *
# *  are protected by Federal copyright law.  They  may  not be disclosed  *
# *  to  third  parties  or copied or duplicated in any form, in whole or  *
# *  in part, without the prior written consent of Silicon Graphics, Inc.  *
# *                                                                        *
# **************************************************************************/
# 
# @(#)$Id: cmdutil.pm,v 1.7 2005/02/22 20:31:08 beighley Exp $
#-------------------------------------------------------------------------------
#
#  FILE: cmdutil.pm
#
#  DESCRIPTION:  This file contains PERL subroutines to create a command utility
#     that allows the user to read and write MMR or memory from a GUI.
#
#  MODIFICATION HISTORY:
#
#     WHEN      WHO             WHAT
#     --------  --------------  -----------------------------------------
#     11/20/02  Karen Beighley  Initial source submission for release.
#     11/25/03  Karen Beighley  Added "cfg" command to list chips and features.
#                               Added support for TIO MMR address range.
#     12/04/03  Karen Beighley  Added "pio" read/write commands that allow
#                                  the user to specify the destination IDs.
#     12/11/03  Karen Beighley  Modified to call "pktgen_setup()" once per
#                                  target and store IDs for future use.
#                                  (assumes 2 address groups for "mmr" req's).
#     12/18/03  Karen Beighley  Added "piou" and "piol" commands.
#     02/10/04  Karen Beighley  Major changes to support multi-chain modules.
#
#-------------------------------------------------------------------------------


#-------------------------------------------------------------------------------
#  CONSTANTS - These cannot change!!!
#-------------------------------------------------------------------------------

$CWD_TARGET_MOD  =  0;  # Command window data: target module/brick
$CWD_TARGET_BRD  =  1;  # Command window data: target board in module
$CWD_TARGET_CHIP =  2;  # Command window data: target chip on board
$CWD_TARGET_CMP  =  3;  # Command window data: target chip component definition
$CWD_NODE_ID     =  4;  # Command window data: node ID of target
$CWD_WINDOW      =  5;  # Command window data: top-level window
$CWD_TEXT_REF    =  6;  # Command window data: reference to command text
$CWD_WID_CMD     =  7;  # Command window data: widget handle for commands
$CWD_WID_RES     =  8;  # Command window data: widget handle for results
$CWD_MODE_MMR    =  9;  # Command window data: mode of operation for MMRs
$CWD_MODE_MEM    = 10;  # Command window data: mode of operation for memory
$CWD_SETUP_FUN   = 11;  # Command window data: reference to setup function for new target

$MODE_MMR_NORM    = 0;
$MODE_MMR_ALL_PG  = 1;

@mode_mmr_names = ( "Use packet generator for II MMRs only",
                    "Use packet generator for all MMRs");

$MODE_MEM_PKTGEN  = 0;
#$MODE_MEM_MBIST   = 1;

@mode_mem_names = ( "Use packet generator for memory");
#                    "Use MBIST engine for memory");


#
#  Global variables for packet generator
#
my $INVALID_ID = -1;
my $tar_src_nid = $INVALID_ID;
my $tar_src_cid = $INVALID_ID;
my $tar_dst_nid = $INVALID_ID;
my $tar_dst_cid = $INVALID_ID;


################################################################################

sub cmdutil_create_window
{
   print "\ncmdutil_create_window( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $setup_fun = 0;
   if ( $#_ >= 0 )
   {  $setup_fun = $_[0];
   }
   my $curr_cmd = "";
   my $bwidth   = 12;
   my ($text, $mu_win, $menubar, $casmenu, $casmenu2, $help_btn, $frame, $cmd_entry, 
       $cmd_btn, $label, $res_wid, $btn);
   my @cw_data  = ( "", "", "", 0, 0, 0, 0, 0, 0, $MODE_MMR_NORM, $MODE_MEM_PKTGEN, 0 );

   #---------------------------------------
   #  Create the command window
   #---------------------------------------
   $text = "Interactive Command Window";
   $mu_win = MainWindow->new(-title => $text);

   #---------------------------------------
   #  Create the menubar for the window
   #---------------------------------------
   $menubar  = $mu_win->Menu(-type => 'menubar');
   $mu_win->configure(-menu => $menubar);
   $casmenu  = $menubar->cascade(-label => '~File', -tearoff => 0);
   $casmenu->command(-label => 'Close', -command => [ $mu_win => 'destroy' ]);
   $casmenu  = $menubar->cascade(-label => '~Settings', -tearoff => 0);
   $casmenu2 = $casmenu->cascade(-label => 'MMR access method', -tearoff => 0);
   for ($i = 0; $i <= $#mode_mmr_names; $i++)
   {  $casmenu2->radiobutton(-label => $mode_mmr_names[$i],
         -value => $i, -variable => \$cw_data[$CWD_MODE_MMR]);
   }
   $casmenu2 = $casmenu->cascade(-label => 'Memory access method', -tearoff => 0);
   for ($i = 0; $i <= $#mode_mem_names; $i++)
   {  $casmenu2->radiobutton(-label => $mode_mem_names[$i],
         -value => $i, -variable => \$cw_data[$CWD_MODE_MEM]);
   }
   $casmenu  = $menubar->cascade(-label => '~Help', -tearoff => 0);
   $help_btn = $casmenu->command(-label => 'List of Commands');

   #---------------------------------------
   #  Create the frame and widgets for the command specification
   #---------------------------------------
   $frame = $mu_win->Frame(-borderwidth => 10);
   $cmd_entry = $frame->BrowseEntry(-label => "Command:", -width => 80, -variable => \$curr_cmd);
   $cmd_entry->pack(-side => 'left', -fill => 'both', -expand => 'yes');
   $cmd_btn = $frame->Button(-text => 'Execute', -width => $bwidth);
   $cmd_btn->pack(-side => 'left');
   $frame->pack(-side => 'top', -fill => 'both');

   #---------------------------------------
   #  Create the frame and widgets for the command results display
   #---------------------------------------
   $frame = $mu_win->Frame(-borderwidth => 10);
   $label = $frame->Label(-anchor => 'w', -text => 'Command Results:');
   $label->pack(-side => 'top', -expand => 'no', -fill => 'both');
   $res_wid = $frame->Scrolled('ROText', -scrollbars => 'se');
   $res_wid->pack(-side => 'top', -expand => 'yes', -fill => 'both');
   $frame->pack(-side => 'top', -expand => 'yes', -fill => 'both');

   $help_btn->configure(-command => [\&cmdutil_display_help, $res_wid]);

   #---------------------------------------
   #  Create the frame and widgets for the action buttons
   #---------------------------------------
   $frame = $mu_win->Frame(-borderwidth => 10);
   $btn   = $frame->Button(-text => 'Clear Results', -width => $bwidth,
               -command => [\&cmdutil_cbf_result_btn, \@cw_data, 'clear'] );
   $btn->pack(-side => 'left', -padx => 5);
   $btn   = $frame->Button(-text => 'Save Results', -width => $bwidth,
               -command => [\&cmdutil_cbf_result_btn, \@cw_data, 'save'] );
   $btn->pack(-side => 'left', -padx => 5);
   $btn   = $frame->Button(-text => 'Close', -width => $bwidth,
               -command => [ $mu_win => 'destroy' ]);
   $btn->pack(-side => 'right', -padx => 5);
   $frame->pack(-side => 'top', -fill => 'both');

   #---------------------------------------
   #  Register the callback function for the command execution
   #---------------------------------------
   $cmd_entry->bind( '<Return>' => [\&cmdutil_cbf_execute_cmd, \@cw_data]);
#   $cmd_entry->bind( '<KeyPress>' => [\&cmdutil_cbf_keypress_cmd, \@cw_data]);
   $cmd_btn->configure(-command => [\&cmdutil_cbf_execute_cmd, 0, \@cw_data]);

   #---------------------------------------
   #  Store window widgets and variables
   #---------------------------------------
   $cw_data[$CWD_TARGET_MOD]  = "";
   $cw_data[$CWD_TARGET_BRD]  = "";
   $cw_data[$CWD_TARGET_CHIP] = "";
   $cw_data[$CWD_TARGET_CMP]  = 0;
   $cw_data[$CWD_NODE_ID]     = 0;
   $cw_data[$CWD_WINDOW]      = $mu_win;
   $cw_data[$CWD_TEXT_REF]    = \$curr_cmd;
   $cw_data[$CWD_WID_CMD]     = $cmd_entry;
   $cw_data[$CWD_WID_RES]     = $res_wid;
}


#-------------------------------------------------------------------------------

sub cmdutil_cbf_execute_cmd
{
   print "\ncmdutil_cbf_execute_cmd( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   return if ( $#_ != 1 );
   my $href     = $_[0];  # Hash pointer w/ Tk specifics
   my $cwd_ref  = $_[1];  # Reference to array with widgets and variables
   my $win_h    = $$cwd_ref[$CWD_WINDOW];   # Main window for dialog boxes
   my $cmd_ref  = $$cwd_ref[$CWD_TEXT_REF]; # Reference to current command string
   my ($text, @cmd_list, $i, $found, $addr, $num, @data, $label);
   if ( $debug_flag & $DEBUG_GENERAL )
   {  printf "CWD_REF = $cwd_ref = ( @$cwd_ref )\n";
   }

   #---------------------------------------
   #  Verify the command is valid and remove extra whitespace
   #---------------------------------------
   return if ( !defined($$cmd_ref) );
   $$cmd_ref =~ s/^\s+//g;  # Remove all leading whitespace
   $$cmd_ref =~ s/\s+$//g;  # Remove all trailing whitespace
#   $$cmd_ref =~ s/\s+/ /g;  # Remove duplicate spaces within command
   return if ( $$cmd_ref eq "" );
   if ( $debug_flag & $DEBUG_GENERAL )
   {  printf "\nSTART CMD>>>$$cmd_ref<<<END CMD\n";
   }

   #---------------------------------------
   #  Break up commands: multiple specified with "\n" or ";"
   #---------------------------------------
   my @top_cmds = split("\n", $$cmd_ref);
   my @cmds     = ();
   my @tmp_cmds = ();
   foreach $curr_cmd (@top_cmds)
   {  if ( $curr_cmd =~ /^#/ )
      {  push(@cmds, $curr_cmd);
      }
      else
      {  @tmp_cmds = split(";", $curr_cmd);
         push(@cmds, @tmp_cmds);
      }
   }

   #---------------------------------------
   #  Process each command specified
   #---------------------------------------
   $flag = $TRUE;
   foreach $curr_cmd ( @cmds )
   {
      $curr_cmd =~ s/^\s+//g;  # Remove all leading whitespace
      $curr_cmd =~ s/\s+$//g;  # Remove all trailing whitespace
      $curr_cmd =~ s/\s+/ /g;  # Remove duplicate spaces within command

      if ( ($curr_cmd =~ /^help/i) || ($curr_cmd eq "h") )
      {  cmdutil_display_help($$cwd_ref[$CWD_WID_RES]);
      }
      elsif ( ($curr_cmd =~ /^cfg/i) || ($curr_cmd =~ /^config/i) )
      {  cmdutil_display_config($$cwd_ref[$CWD_WID_RES]);
      }
      elsif ( ($curr_cmd =~ /^mmr /i) || ($curr_cmd =~ /^mem /i) || ($curr_cmd =~ /^pio/i) )
      {  if ( $$cwd_ref[$CWD_TARGET_MOD] eq "" )
         {  $text = "command> " . $curr_cmd . "\n";
            $$cwd_ref[$CWD_WID_RES]->insert('end', $text);
            $text = "ERROR - No target for operation; use \"target\" command to define.\n";
            $$cwd_ref[$CWD_WID_RES]->insert('end', $text);
            $$cwd_ref[$CWD_WID_RES]->see('end');
            $flag = $FALSE;
         }
         else
         {  $rv = cmdutil_process_access_cmd($cwd_ref, $curr_cmd);
            $flag = (( $rv == $FALSE ) ? $FALSE : $flag);
         }
      }
      elsif ( $curr_cmd =~ /^target/i )
      {  $rv = cmdutil_process_target_cmd($cwd_ref, $curr_cmd, $$cwd_ref[$CWD_SETUP_FUN]);
         $flag = (( $rv == $FALSE ) ? $FALSE : $flag);
      }
      elsif ( $curr_cmd =~ /^#/ )
      {  $text = "command> " . $curr_cmd . "\n";
         $$cwd_ref[$CWD_WID_RES]->insert('end', $text);
      }
      elsif ( $curr_cmd ne "" )
      {  $text = "command> " . $curr_cmd . "\n";
         $$cwd_ref[$CWD_WID_RES]->insert('end', $text);
         $text = "ERROR - Unknown command: \"$curr_cmd\"\n";
         $$cwd_ref[$CWD_WID_RES]->insert('end', $text);
         $$cwd_ref[$CWD_WID_RES]->see('end');
         $flag = $FALSE;
      }
   }

   #---------------------------------------
   #  If new command and no error: 
   #  1) add to the "command" history 
   #  2) clear/reset the command string
   #---------------------------------------
   if ( $flag )
   {  @cmd_list = $$cwd_ref[$CWD_WID_CMD]->get(0, 'end');
      for ($i = 0, $found = $FALSE; (($i <= $#cmd_list) && !$found); $i++)
      {  if ( $$cmd_ref eq $cmd_list[$i] )
         {  $found = $TRUE;
         }
      }
      if ( !$found )
      {  $$cwd_ref[$CWD_WID_CMD]->insert('end', $$cmd_ref);
      }

      $$cmd_ref = ""; # Clear/reset the command string
   }

   return($flag);
}

#-------------------------------------------------------------------------------

sub cmdutil_cbf_keypress_cmd
{
   print "\ncmdutil_cbf_keypress_cmd( @_ )\n" ;# if ($debug_flag & $DEBUG_FUNCTIONS);

   return if ( $#_ != 1 );
   my $wid_ref  = $_[0];  # Hash pointer w/ widget Tk specifics
   my $cwd_ref  = $_[1];  # Reference to array with widgets and variables

   my $e = $wid_ref->XEvent; # Get event object
   my $keysym_text = $e->K;
   my $keysym_val  = $e->N;

   printf "keysysm=\"$keysym_text\", numeric=$keysym_val\n";
}

#-------------------------------------------------------------------------------

sub cmdutil_process_access_cmd
{
   printf "\ncmdutil_process_access_cmd( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $cwd_ref   = $_[0];  # Reference to array with widgets and variables
   my $cmd_text  = $_[1];  # Text of command string
   if ( $debug_flag & $DEBUG_GENERAL )
   {  printf "CWD_REF  = $cwd_ref = ( @$cwd_ref )\n";
      printf "CMD_TEXT = ( $cmd_text )\n";
   }
   my $win_h = $$cwd_ref[$CWD_WINDOW];   # Main window for dialog boxes
   my @parms = split(" ", $cmd_text);
   my ($text, $base, $addr, $i, $addr_text, $data_text, $num, @data, $rv, $label,
       $chip_ptr, $flag, $name, $pnum, $src_node_id, $dst_node_id, $dst_chiplet, $len);
  
   #---------------------------------------
   #  Verify the command is valid
   #---------------------------------------
   $parms[0] = uc($parms[0]);
   if ( ($parms[0] ne "MMR") && ($parms[0] ne "MEM")  && ($parms[0] ne "PIO") &&
        ($parms[0] ne "PIOU") && ($parms[0] ne "PIOL") )
   {  return($FALSE); # programming error: shound not get here!!!
   }
   $name = $$cwd_ref[$CWD_TARGET_MOD] . ":" . $$cwd_ref[$CWD_TARGET_BRD] . ":" . $$cwd_ref[$CWD_TARGET_CHIP];

   #---------------------------------------
   #  Verify the target supports the operation
   #---------------------------------------
   $chip_ptr = $$cwd_ref[$CWD_TARGET_CMP];
   $flag     = $$chip_ptr[$CMP_IND_FLAG];
   if ( ($parms[0] eq "MMR") && (($flag & $CMP_FLAG_MMR) == 0) )
   {  $text = "command> " . $cmd_text . "\n";
      $$cwd_ref[$CWD_WID_RES]->insert('end', $text);
      $text = "ERROR - Target does not support JTAG initiated MMR operations.\n";
      $$cwd_ref[$CWD_WID_RES]->insert('end', $text);
      $$cwd_ref[$CWD_WID_RES]->see('end');
      return($FALSE);
   }
   elsif ( ($parms[0] eq "MEM") && (($flag & $CMP_FLAG_PTKGEN) == 0) )
   {  $text = "command> " . $cmd_text . "\n";
      $$cwd_ref[$CWD_WID_RES]->insert('end', $text);
      $text = "ERROR - Target does not support packet generated MEM operations.\n";
      $$cwd_ref[$CWD_WID_RES]->insert('end', $text);
      $$cwd_ref[$CWD_WID_RES]->see('end');
      return($FALSE);
   }
   elsif ( ($parms[0] =~ /^PIO/) && (($flag & $CMP_FLAG_PTKGEN) == 0) )
   {  $text = "command> " . $cmd_text . "\n";
      $$cwd_ref[$CWD_WID_RES]->insert('end', $text);
      $text = "ERROR - Target does not support packet generated PIO operations.\n";
      $$cwd_ref[$CWD_WID_RES]->insert('end', $text);
      $$cwd_ref[$CWD_WID_RES]->see('end');
      return($FALSE);
   }

   #---------------------------------------
   #  Target the chip for MMR/MEM operations
   #---------------------------------------
   mmr_target($SELECT, $$cwd_ref[$CWD_TARGET_MOD], $$cwd_ref[$CWD_TARGET_BRD], $$cwd_ref[$CWD_TARGET_CHIP]);

# NOTE: Should optimize to use block operations: make single read/write request
   
   #---------------------------------------
   #  <mem|mmr> <r|read> <addr> [<#words>]
   #---------------------------------------
   $parms[1] = uc($parms[1]);
   if ( ($parms[1] eq "R") || ($parms[1] eq "READ") ) 
   {
      $pnum = ( ($parms[0] =~ /^PIO/) ? 4 : 2 );
      if ( ($#parms != $pnum) && ($#parms != ($pnum + 1)) )
      {  $text = "command> " . $cmd_text . "\n";
         $$cwd_ref[$CWD_WID_RES]->insert('end', $text);
         $text = "ERROR - Incorrect parameters for read command: \"$curr_cmd\"\n";
         $$cwd_ref[$CWD_WID_RES]->insert('end', $text);
         $$cwd_ref[$CWD_WID_RES]->see('end');
         return($FALSE);
      }
      $label = sprintf("%-4s READ  ", $parms[0]);

      #  Convert address and number of locations to read
      $addr = convert_text2num($parms[$pnum]);
      $base = $addr;
      $num  = ( ($#parms == ($pnum + 1)) ? convert_text2num($parms[($pnum + 1)]) : 1 );

      #  Confirm operation of large number of locations requested
      if ( $num > 64 )
      {  $text = sprintf("Do you really want to read %u %s locations starting at address 0x%016llX?",
                    $num, ($parms[0] eq "MEM" ? "memory" : "MMR"), $addr);
         $rv = $win_h->messageBox(-icon => 'question', -type => 'YesNo', -message => $text,
                 -title => 'Confirm Read Request');
         if ( $rv eq 'No' )
         {  $text = "command> " . $cmd_text . "\n";
            $$cwd_ref[$CWD_WID_RES]->insert('end', $text);
            $text = "ERROR - Operation rejected due to the number of locations.\n";
            $$cwd_ref[$CWD_WID_RES]->insert('end', $text);
            $$cwd_ref[$CWD_WID_RES]->see('end');
            return($FALSE);
         }
      }
      elsif ( $num == 0 )
      {  $text = "command> " . $cmd_text . "\n";
         $$cwd_ref[$CWD_WID_RES]->insert('end', $text);
         $text = "ERROR - The number of locations specified is zero; no read issued.\n";
         $$cwd_ref[$CWD_WID_RES]->insert('end', $text);
         $$cwd_ref[$CWD_WID_RES]->see('end');
         return($FALSE);
      }

      #  Perform the read operation(s)
      @data = ();
      if ( $parms[0] eq "MEM" )  # Memory read operation
      {  for ($i = 0; $i < $num; $i++)
         {  $data[$i] = ( ($jtag_access == $JTAG_ACCESS_NONE) ? mem_read($addr) : pktgen_mem_read($addr) );
            return($FALSE) if ( $ecode < 0 );
            $addr += $MEM_ADDR_INCR;
         }
      }
      elsif ( $parms[0] =~ /^PIO/ )  # Packet generated PIO/MMR read operation
      {  
         if ( $tar_src_nid == $INVALID_ID ) # First time using packet generator on target
         {  ($tar_src_nid, $tar_src_cid, $tar_dst_nid, $tar_dst_cid) = pktgen_setup('mmr', 0);
         }
         $src_node_id = $tar_src_nid;
         $dst_node_id = convert_text2num($parms[2]);
         $dst_chiplet = convert_text2num($parms[3]);
         $text  = sprintf("DST[ID=0x%03X CHIPLET=0x%X] ", $dst_node_id, $dst_chiplet);
         $label .= $text;

         if ( $parms[0] eq "PIOL" )
         {  $suppl = 0x0F;
         }
         elsif ( $parms[0] eq "PIOU" )
         {  $suppl = 0xF0;
         }
         else
         {  $suppl = 0xFF;
         }
         for ($i = 0; $i < $num; $i++)
         {  $data[$i] = pktgen_mmr_read($addr, $src_node_id, $dst_node_id, $dst_chiplet, $suppl);
            return($FALSE) if ( $ecode < 0 );
            $addr += $MEM_ADDR_INCR;
         }
      }
      else # MMR read operation
      {  for ($i = 0; $i < $num; $i++)
         {  if ( ($jtag_access != $JTAG_ACCESS_NONE) &&
                 ( (($$chip_ptr[$CMP_IND_DESC] =~ /^shub/i) && (($addr & 0x100000000) == 0)) ||
                   (($$chip_ptr[$CMP_IND_DESC] =~ /^tio/i)  && ($addr < 0x80000000)) ) )
            { 
               if ( $tar_src_nid == $INVALID_ID ) # First time using packet generator on target
               {  ($tar_src_nid, $tar_src_cid, $tar_dst_nid, $tar_dst_cid) = pktgen_setup('mmr', 0);
               }
               $data[$i] = pktgen_mmr_read($addr, $tar_src_nid, $tar_dst_nid, $tar_dst_cid);
            }
            else # MMR operation using normal MMR subroutines
            {  $data[$i] = mmr_read($addr);
            }
            if ( $ecode < 0 )
            {  $text = "command> " . $cmd_text . "\n";
               $$cwd_ref[$CWD_WID_RES]->insert('end', $text);
               $text = ( ($num == 1) ?
                          "ERROR - Unable to complete command; verify MMR address is valid.\n" :
                          "ERROR - Unable to complete entire command; verify all MMR addresses are valid.\n" );
               $$cwd_ref[$CWD_WID_RES]->insert('end', $text);
               $$cwd_ref[$CWD_WID_RES]->see('end');
               return($FALSE) 
            }
            $addr += $MEM_ADDR_INCR;
         }
      }
   }
   #---------------------------------------
   #  <mem|mmr> <w|write> <addr> <data> [<data>]
   #---------------------------------------
   elsif ( ($parms[1] eq "W") || ($parms[1] eq "WRITE") ) 
   {
      $pnum = ( ($parms[0] =~ /^PIO/) ? 5 : 3 );
      if ( $#parms < $pnum )
      {  $text = "command> " . $cmd_text . "\n";
         $$cwd_ref[$CWD_WID_RES]->insert('end', $text);
         $text = "ERROR - Incorrect parameters for write command: \"$curr_cmd\"\n";
         $$cwd_ref[$CWD_WID_RES]->insert('end', $text);
         $$cwd_ref[$CWD_WID_RES]->see('end');
         return($FALSE);
      }
      $label = sprintf("%-4s WRITE ", $parms[0]);

      #  Convert address and number of locations to read
      $addr = convert_text2num($parms[($pnum-1)]);
      $base = $addr;
      @data = ();
      for ($num = 0; $num <= ($#parms - $pnum); $num++)
      {  $data[$num] = convert_text2num($parms[$num+$pnum]);
      }

      #  Perform the write operation(s)
      if ( $parms[0] eq "MEM" ) # Memory write operation
      {  
         for ($i = 0; $i < $num; $i++)
         {  if ( $jtag_access == $JTAG_ACCESS_NONE )
            {  mem_write($addr, $data[$i]);
            }
            else
            {  pktgen_mem_write($addr, $data[$i]);
            }
            return($FALSE) if ( $ecode < 0 );
            $addr += $MEM_ADDR_INCR;
         }
      }
      elsif ( $parms[0] =~ /^PIO/ ) # Packet generated PIO/MMR write operation
      {  
         if ( $tar_src_nid == $INVALID_ID ) # First time using packet generator on target
         {  ($tar_src_nid, $tar_src_cid, $tar_dst_nid, $tar_dst_cid) = pktgen_setup('mmr', 0);
         }
         $src_node_id = $tar_src_nid;
         $dst_node_id = convert_text2num($parms[2]);
         $dst_chiplet = convert_text2num($parms[3]);
         $text  = sprintf("DST[ID=0x%03X CHIPLET=0x%X] ", $dst_node_id, $dst_chiplet);
         $label .= $text;

         if ( $parms[0] eq "PIOL" )
         {  $suppl = 0x0F;
         }
         elsif ( $parms[0] eq "PIOU" )
         {  $suppl = 0xF0;
         }
         else
         {  $suppl = 0xFF;
         }
         for ($i = 0; $i < $num; $i++)
         {  pktgen_mmr_write($addr, $data[$i], $src_node_id, $dst_node_id, $dst_chiplet, $suppl);
            return($FALSE) if ( $ecode < 0 );
            $addr += $MEM_ADDR_INCR;
         }
      }
      else # MMR write operation
      {
         for ($i = 0; $i < $num; $i++)
         {  if ( ($jtag_access != $JTAG_ACCESS_NONE) &&
                 ( (($$chip_ptr[$CMP_IND_DESC] =~ /^shub/i) && (($addr & 0x100000000) == 0)) ||
                   (($$chip_ptr[$CMP_IND_DESC] =~ /^tio/i)  && ($addr < 0x80000000)) ) )
            {  if ( $tar_src_nid == $INVALID_ID ) # First time using packet generator on target
               {  ($tar_src_nid, $tar_src_cid, $tar_dst_nid, $tar_dst_cid) = pktgen_setup('mmr', 0);
               }
               pktgen_mmr_write($addr, $data[$i], $tar_src_nid, $tar_dst_nid, $tar_dst_cid);
            }
            else # MMR operation using normal MMR subroutines
            {  mmr_write($addr, $data[$i]);
            }
            return($FALSE) if ( $ecode < 0 );
            $addr += $MEM_ADDR_INCR;
         }
      }
   }
   else
   {  $text = "command> " . $cmd_text . "\n";
      $$cwd_ref[$CWD_WID_RES]->insert('end', $text);
      $text = "ERROR - Unknown command: \"$curr_cmd\"\n";
      $$cwd_ref[$CWD_WID_RES]->insert('end', $text);
      $$cwd_ref[$CWD_WID_RES]->see('end');
      return($FALSE);
   }

   #---------------------------------------
   #  Adjust data if 32-bit PIO operation(s)
   #---------------------------------------
   if ( $parms[0] eq "PIOL" )
   {  for ($i = 0; $i < $num; $i++)
      {  $data[$i] &= 0x00000000FFFFFFFF;
      }
      $len = 32;
   }
   elsif ( $parms[0] eq "PIOU" )
   {  for ($i = 0; $i < $num; $i++)
      {  $data[$i] = ($data[$i] >> 32) & 0x00000000FFFFFFFF;
      }
      $len = 32;
   }
   else
   {  $len = 64;
   }

   #---------------------------------------
   #  Add the command and results to the "results" window
   #---------------------------------------
   $text = "command> " . $cmd_text . "\n";
   $$cwd_ref[$CWD_WID_RES]->insert('end', $text);
   $addr = $base;
   for ($i = 0; $i < $num; $i++)
   {  $addr_text = convert_num2text($addr, 64, 'hex');
      $data_text = convert_num2text($data[$i], $len, $num_base);
      $addr += $MEM_ADDR_INCR;
      $text = $name . " " . $label . $addr_text . " : " . $data_text . "\n";
      $$cwd_ref[$CWD_WID_RES]->insert('end', $text);
   }
   $$cwd_ref[$CWD_WID_RES]->see('end');

   return($TRUE);
}


#-------------------------------------------------------------------------------

sub cmdutil_cbf_result_btn
{
   print "\ncmdutil_cbf_result_btn( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   return if ( $#_ != 1 );
   my $cwd_ref    = $_[0];  # Reference to array with widgets and variables
   my $mode       = $_[1];  # Mode: 'save' or 'clear'
   if ( $debug_flag & $DEBUG_GENERAL )
   {  printf "CWD_REF  = $cwd_ref = ( @$cwd_ref )\n";
      printf "MODE     = $mode\n";
   }
   my $win_h      = $$cwd_ref[$CWD_WINDOW];  # Main window for dialog boxes
   my $res_widget = $$cwd_ref[$CWD_WID_RES];  # Reference to command results widget

   #---------------------------------------
   #  Clear the command history buffer
   #---------------------------------------
   if ( $mode eq 'clear' )
   {  $res_widget->delete('1.0', 'end');
      return;
   }

   #---------------------------------------
   #  Continue: save the command history buffer
   #---------------------------------------
   my ($out_file, $file_id, $text);

   #---------------------------------------
   #  Get the contents of the command history
   #---------------------------------------
   $text = $res_widget->get('1.0', 'end');

   #---------------------------------------
   #  Get the output file name from user
   #---------------------------------------
   $out_file = $win_h->getSaveFile();
   return if ( !defined($out_file) );

   #---------------------------------------
   #  Open the output file and write the text
   #---------------------------------------
   $file_id = sprintf("hist%08lX", time());
   if ( !open($file_id, ">$out_file") )
   {  $text = "Unable to open output file for writing.\nFile: $out_file\nCommand history not saved.";
      $win_h->messageBox(-icon => 'error', -type => 'OK', -message => $text,
         -title => 'ICC Monitor: open file error');
      return;
   }
   printf $file_id $text;
   close($file_id);
   
   print "\nSaved command results: $out_file\n";
}

#-------------------------------------------------------------------------------

sub cmdutil_display_help
{
   my $res_widget = $_[0];  # Reference to command results widget

   @help_msg = 
   (  
      "\nSuperScope Interactive Command Window - Memory and MMR Access\n",
      "\nThe following general commands are supported:\n",
      "target <brick> <board> <chip>      = target chip for MMR and memory operations\n",
      "target                             = display current target chip for MMR and memory operations\n",
      "cfg|config                         = display configuration: chip name and features\n",
      "h|help                             = echo supported commands and parameters\n",
      "The following memory commands are supported:\n",
      "mem r|read  <addr>                 = read the location specified by <addr>\n",
      "mem r|read  <addr> <num>           = read <number> locations starting at <addr>\n",
      "mem w|write <addr> <data>          = write the location specified by <addr> with <data>\n",
      "mem w|write <addr> <data> [<data>] = write multiple locations starting at <addr>\n",
      "NOTE: Memory write operations are non-coherent (SN2NET_NCWRD = non-coherent dword write).\n",
      "The following MMR commands are supported:\n",
      "mmr r|read  <addr>                 = read the location specified by <addr>\n",
      "mmr r|read  <addr> <num>           = read <number> locations starting at <addr>\n",
      "mmr w|write <addr> <data>          = write the location specified by <addr> with <data>\n",
      "mmr w|write <addr> <data> [<data>] = write multiple locations starting at <addr>\n",
      "The following PIO commands are supported: PIO destination node and chiplet ID required\n",
      "  The default byte enable mask for a PIO operation is 0xFF (in the request packets SUPPL field)\n",
      "pio r|read  <node> <chiplet> <addr>                 = read the location specified by <addr>\n",
      "pio r|read  <node> <chiplet> <addr> <num>           = read <number> locations starting at <addr>\n",
      "pio w|write <node> <chiplet> <addr> <data>          = write the location specified by <addr> with <data>\n",
      "pio w|write <node> <chiplet> <addr> <data> [<data>] = write multiple locations starting at <addr>\n",
      "All the PIO commands can be modified by appending 'l' or 'u' to the command name:\n",
      "  If 'l' is appended, the a byte mask of 0x0F is used in the request packet's SUPPL field\n",
      "  If 'u' is appended, the a byte mask of 0xF0 is used in the request packet's SUPPL field\n",
      "piou r|read  <node> <chiplet> <addr>                = read the location specified by <addr>\n",
      "piol r|read  <node> <chiplet> <addr>                = read the location specified by <addr>\n",
      "NOTE: Memory, PIO and some MMR operations (SHUB II) use the packet generator.\n",
      "Multiple commands can be entered if seperated by a semicolon (';') or newline ('\\n')\n",
      "\n",
   );

   #---------------------------------------
   #  Add the command and results to the "results" window
   #---------------------------------------
   $res_widget->insert('end', "command> help\n");
   for ($i = 0; $i <= $#help_msg; $i++)
   {  $res_widget->insert('end', $help_msg[$i]);
   }
   $res_widget->see('end');
}

#-------------------------------------------------------------------------------

sub cmdutil_display_config
{
   my $res_widget = $_[0];  # Reference to command results widget
   my ($i, $mname, $mref, $chip_list, $chip_ref, $name, $text, $flag, $str);
   my ($chn_list, $chn_ptr, $c);

   $res_widget->insert('end', "command> config\n");
   $res_widget->insert('end', "\n");
   $res_widget->insert('end', "BRICK    BOARD    CHIP     SUPPORTED FEATURES\n");
   $res_widget->insert('end', "-------- -------- -------- --------------------------------------\n");

   foreach $mname (@module_names)
   {
      $mref     = scan_module_find($mname);  # Module definition structure
      $chn_list = $$mref{CHN_REFS}; # List of chain definition pointers
      for ($c = 0; $c <= $#$chn_list; $c++)
      {  $chn_ptr   = $$chn_list[$c];
         $chip_list = $$chn_ptr[$CHN_IND_CREFS];
         for ($i = 0; $i <= $#$chip_list; $i++)
         {  $chip_ref = $$chip_list[$i];
            $name = sprintf("%-8s %-8s %-8s ", $mname, $$chip_ref[$CMP_IND_BOARD], $$chip_ref[$CMP_IND_LOG]);
            $text = "(JTAG";
            $flag = $$chip_ref[$CMP_IND_FLAG];
            if ( $flag & $CMP_FLAG_MMR )
            {  $text = $text . ", MMR";
            }
            if ( $flag & $CMP_FLAG_DEBUG )
            {  $text = $text . ", DEBUG_PORT";
            }
            if ( $flag & $CMP_FLAG_ISCAN )
            {  $text = $text . ", INTERNAL_SCAN";
            }
            if ( $flag & $CMP_FLAG_PTKGEN )
            {  $text = $text . ", PACKET_GENERATOR";
            }
            $text = $text . ")";

            $str = $name . $text . "\n";
            $res_widget->insert('end', $str);
         }
      }
   }
   $res_widget->insert('end', "\n");
   $res_widget->see('end');
}


#-------------------------------------------------------------------------------

sub cmdutil_process_target_cmd
{
   printf "\ncmdutil_process_target_cmd( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $cwd_ref   = $_[0];  # Reference to array with widgets and variables
   my $cmd_text  = $_[1];  # Text of command string
   my $setup_fun = $_[2];  # Reference to setup function for new target
   if ( $debug_flag & $DEBUG_GENERAL )
   {  printf "CWD_REF  = $cwd_ref = ( @$cwd_ref )\n";
      printf "CMD_TEXT = ( $cmd_text )\n";
   }
   my $win_h = $$cwd_ref[$CWD_WINDOW];   # Main window for dialog boxes
   my @parms = split(" ", $cmd_text);
   my ($text, $mname, $bname, $cname, $mref, $chip_ptr);

   #---------------------------------------
   #  Invalidate the packet generator IDs
   #---------------------------------------
   $tar_src_nid = $INVALID_ID;
   $tar_src_cid = $INVALID_ID;
   $tar_dst_nid = $INVALID_ID;
   $tar_dst_cid = $INVALID_ID;

   #---------------------------------------
   #  Process the "target" command
   #---------------------------------------
   if ( $#parms == 0 )
   {  $text = "command> " . $cmd_text . "\n";
      $$cwd_ref[$CWD_WID_RES]->insert('end', $text);
      if ( $$cwd_ref[$CWD_TARGET_MOD] eq "" )
      {  $text = "GET TARGET = <undefined>\n";
      }
      else
      {  $chip_ptr = $$cwd_ref[$CWD_TARGET_CMP];
         if ( $debug_flag & $DEBUG_GENERAL )
         {  printf("GET TARGET = %s %s %s (@$chip_ptr)\n", $$cwd_ref[$CWD_TARGET_MOD],
               $$cwd_ref[$CWD_TARGET_BRD], $$cwd_ref[$CWD_TARGET_CHIP]);
         }
         $text = sprintf("GET TARGET = %s %s %s\n", $$cwd_ref[$CWD_TARGET_MOD], $$cwd_ref[$CWD_TARGET_BRD],
             $$cwd_ref[$CWD_TARGET_CHIP]);
      }
      $$cwd_ref[$CWD_WID_RES]->insert('end', $text);
      $$cwd_ref[$CWD_WID_RES]->see('end');
   }
   elsif ( $#parms == 3 )
   {  
      $mname = $parms[1];  # Name of target module/brick
      $bname = $parms[2];  # Name of board within target module
      $cname = $parms[3];  # Name of chip/ASIC on board

      #---------------------------------------
      #  Determine if the module name is valid for the configuration
      #---------------------------------------
      $mref = scan_module_find($mname);
      if ( !defined($mref) )
      {  $text = "command> " . $cmd_text . "\n";
         $$cwd_ref[$CWD_WID_RES]->insert('end', $text);
         $text = "ERROR - Unknown module specified for TARGET operation.";
         $text .= "\nModule: $mname";
         $text .= "\nModule not found in current configuration.\n";
         $$cwd_ref[$CWD_WID_RES]->insert('end', $text);
         $$cwd_ref[$CWD_WID_RES]->see('end');
         $$cwd_ref[$CWD_TARGET_MOD]  = "";
         $$cwd_ref[$CWD_TARGET_BRD]  = "";
         $$cwd_ref[$CWD_TARGET_CHIP] = "";
         $$cwd_ref[$CWD_TARGET_CMP]  = 0;
         return($FALSE);
      }

      #---------------------------------------
      #  Extract the scan module details for the specific board/chip
      #---------------------------------------
      $chip_ptr = scan_module_chip_ptr($mref, $bname, $cname, 'logical');
      if ( !defined($chip_ptr) )
      {  $text = "command> " . $cmd_text . "\n";
         $$cwd_ref[$CWD_WID_RES]->insert('end', $text);
         $text = "ERROR - Unknown target board/chip specified for TARGET operation.";
         $text .= "\nModule: $mname";
         $text .= "\nBoard:  $bname";
         $text .= "\nChip:   $cname";
         $text .= "\nLocation not found in current configuration.\n";
         $$cwd_ref[$CWD_WID_RES]->insert('end', $text);
         $$cwd_ref[$CWD_WID_RES]->see('end');
         $$cwd_ref[$CWD_TARGET_MOD]  = "";
         $$cwd_ref[$CWD_TARGET_BRD]  = "";
         $$cwd_ref[$CWD_TARGET_CHIP] = "";
         $$cwd_ref[$CWD_TARGET_CMP]  = 0;
         return($FALSE);
      }
      $$cwd_ref[$CWD_TARGET_MOD]  = $$mref{NAME};
      $$cwd_ref[$CWD_TARGET_BRD]  = $$chip_ptr[$CMP_IND_BOARD];
      $$cwd_ref[$CWD_TARGET_CHIP] = $$chip_ptr[$CMP_IND_LOG];
      $$cwd_ref[$CWD_TARGET_CMP]  = $chip_ptr;

      $text = "command> " . $cmd_text . "\n";
      $$cwd_ref[$CWD_WID_RES]->insert('end', $text);
      if ( $debug_flag & $DEBUG_GENERAL )
      {  printf("SET TARGET = %s %s %s (@$chip_ptr)\n", $$cwd_ref[$CWD_TARGET_MOD],
            $$cwd_ref[$CWD_TARGET_BRD], $$cwd_ref[$CWD_TARGET_CHIP]);
      }
      $text = sprintf("SET TARGET = %s %s %s\n", $$cwd_ref[$CWD_TARGET_MOD], $$cwd_ref[$CWD_TARGET_BRD], 
                 $$cwd_ref[$CWD_TARGET_CHIP]);
      $$cwd_ref[$CWD_WID_RES]->insert('end', $text);
      $$cwd_ref[$CWD_WID_RES]->see('end');

      if ( $setup_fun != 0 )
      {  mmr_target($SELECT, $$cwd_ref[$CWD_TARGET_MOD], $$cwd_ref[$CWD_TARGET_BRD], $$cwd_ref[$CWD_TARGET_CHIP]);
         &$setup_fun($mref, $chip_ptr);
      }
   }
   else
   {  $text = "command> " . $cmd_text . "\n";
      $$cwd_ref[$CWD_WID_RES]->insert('end', $text);
      $text = "ERROR - Incorrect parameters for target command: \"$curr_cmd\"\n";
      $$cwd_ref[$CWD_WID_RES]->insert('end', $text);
      $$cwd_ref[$CWD_WID_RES]->see('end');
      return($FALSE);
   }

   return($TRUE);
}


################################################################################

1;
