# 
#/**************************************************************************
# *                                                                        *
# *               Copyright (C) 1989-1999 Silicon Graphics, Inc.           *
# *                                                                        *
# *  These coded instructions, statements, and computer programs  contain  *
# *  unpublished  proprietary  information of Silicon Graphics, Inc., and  *
# *  are protected by Federal copyright law.  They  may  not be disclosed  *
# *  to  third  parties  or copied or duplicated in any form, in whole or  *
# *  in part, without the prior written consent of Silicon Graphics, Inc.  *
# *                                                                        *
# **************************************************************************/
# 
#-------------------------------------------------------------------------------
#
#  FILE: dbgaccess.pm
#
#  DESCRIPTION:  This file contains PERL for the debug port access windows.
#
#
#  MODIFICATION HISTORY:
#
#     WHEN      WHO             WHAT
#     --------  --------------  -----------------------------------------
#     10/18/02  Karen Beighley  Added modification history to source file.
#     10/18/02  Karen Beighley  Changed to show the current selection when
#                                  the group configuration window is opened.
#     12/20/02  Karen Beighley  Apply mask to 64-bit values from DebugSel.xs
#     10/30/03  Karen Beighley  Cleanup and support of testpoint override.
#     01/08/04  Karen Beighley  Added ability to configure consecutive debug
#                                  port groups (to consecutive selections).
#     01/15/04  Karen Beighley  Changed to use "Toplevel" for new window(s).
#     02/11/04  Karen Beighley  Added module scan chain index for resets.
#     07/14/05  Karen Beighley  Added break if MMR operation fails while
#                                  reading/writing control MMRs.
# 
#-------------------------------------------------------------------------------
#
# @(#)$Id: dbgaccess.pm,v 1.17 2005/07/14 17:10:31 beighley Exp $

################################################################################
#  GLOBAL VARIABLES
################################################################################

$FALSE = 0;
$TRUE  = 1;

my $fld_font = '-*-helvetica-medium-r-normal--*-80-*-*-*-*-*-*';
my $utext    = "<undefined>";


################################################################################
# The debug port access window data: 
# <data HEX label text> <data MMR entry> <ctrl HEX label text> <ctrl MMR entry>
#
#  The debug port access lookup  structure is a HASH with one entry per port:
#    key   = full name of the chip (<module>:<board>:<chip>:<debug port>)
#    {key} = reference to debug port access (DWD) structure of the ASIC
#  The port has an entry in the lookup only if the access window is "open".
################################################################################

$DWD_MAIN_WINDOW   = 0;  # Main window reference for debug port data
$DWD_DATA_TEXT_HEX = 1;  # Text entry for debug port data : in hex
$DWD_DATA_TEXT_MMR = 2;  # Text entry for debug port data : formatted
$DWD_CTRL_TEXT_HEX = 3;  # Text entry for debug port control : in hex
$DWD_CTRL_TEXT_MMR = 4;  # Text entry for debug port control : formatted
$DWD_TP_TEXT_LIST  = 5;  # Text widget with list of selected testpoints
$DWD_MAIN_HIST_WID = 6;  # Text widget with list of selected testpoints
$DWD_REF_CB_LIST   = 7;  # 

%dwd_lookup = ();


################################################################################
# The group selection window data: 
# <main window> <group menubutton> <group list> <TP text> [selection vals]
################################################################################

$GWD_MAIN_WINDOW     = 0;  # Reference to the main window
$GWD_WID_GRP_MENU    = 1;  # Widget reference: menu with group options
$GWD_WID_GRP_OPTION  = 2;  # Widget reference: text of current group option
$GWD_WID_GRP_LIST    = 3;  # Widget reference: list box with selections per option
$GWD_WID_GRP_CURR    = 4;  # Widget reference: current selection in list box
$GWD_WID_TP_TEXT     = 5;  # Widget reference: testpoints per selection
$GWD_WID_CFG_BTN     = 6;  # Widget reference: testpoints per selection
$GWD_SEL_VALS        = 7;  # Array of values that correspond to selections
$GWD_CB_CFG_MULTI    = 8;  # Checkbutton variable: configure subsequent debug port groups


################################################################################
#  The debug port data (DPD) structure is an ARRAY containing the following:
#  ( <debug port name>, <dport struct ptr>, <debug port reg_ref>, <flag>, 
#    <ctrl reg_ref>, <ctrl mask>, <ctrl data>,
#    [ <group 0 name>, <group 1 name>, ... <group N name> ],
#    [ <group 0 ptr>,  <group 1 ptr>,  ... <group N ptr>  ]
#  )
################################################################################

$DPD_NAME      = 0;    # Name of debug port (i.e. "SH_LB_DEBUG_DATA")
$DPD_DPORT_PTR = 1;    # Pointer to "dport" structure (in C-library)
$DPD_FLAGS     = 2;    # Flags: information on debug data and control register
$DPD_DATA_REG  = 3;    # Register instance if debug port has MMR
$DPD_CTRL_REG  = 4;    # Register instance of control MMR (to configure)
$DPD_CTRL_MASK = 5;    # Mask of bits in control MMR that affect debug port
$DPD_CTRL_DATA = 6;    # Value of bits in control MMR that affect debug port
$DPD_GRP_NAMES = 7;    # Reference to array of group names
$DPD_GRP_PTRS  = 8;    # Reference to array of group control pointers (in C)
$DPD_GRP_FLAGS = 9;    # Reference to array of group flags

$DPD_FLAG_MMR       = 0x01; # Flag: debug port data available from MMR
$DPD_FLAG_NEW_CTRL  = 0x02; # Flag: control register value has changed

################################################################################
#  A debug port data (DPD) list is created for each ASIC accessed.  The list
#  is an array of references.  There is one entry per debug port on the ASIC:
#  ( <DPD ref> <DPD ref> <DPD ref> ... <DPD ref> )
#
#  The debug port lookup structure is a HASH with one entry per ASIC:
#    key   = full name of the chip (<module>:<board>:<chip>)
#    {key} = reference to debug port data (DPD) list of the ASIC
################################################################################

%dpd_lookup = ();


################################################################################

sub dpd_list_create
{
   printf "\ndpd_list_create( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   return if ( $#_ != 1 );
   my $chip_name = $_[0];   # Full name of ASIC: <brick>:<board>:<chip>
   my $chip_type = $_[1];   # ASIC type/description (i.e. "shub_01");

   #
   #  Determine if the structures have already been created
   #
   if ( exists $dpd_lookup{$chip_name} )
   {  return($dpd_lookup{$chip_name});
   }

   #
   #  Get the list of debug port names and C-structure pointers
   #
   my @dp_names = ();
   my @dp_ptrs  = ();
   my $dp_num   = pcf_dport_list_create($chip_type, \@dp_names, \@dp_ptrs);
   if ( $dp_num <= 0 )
   {  $emess = "Unable to extract debug port data from library.";
      $emess .= "\nTarget chip : $chip_name";
      $emess .= "\nChip type   : $chip_type";
      return;
   }

   if ($debug_flag & $DEBUG_GENERAL)
   {  print "\nDEBUG PORT RESULTS: $chip_name\n";
      print "dp_num   = $dp_num\n";
      print "dp_names = ( @dp_names )\n";
      print "dp_ptrs  = ( @dp_ptrs )\n";
   }

   #
   #  Build the debug port data (DPD) structures
   #
   my ($dpd_ref, $i);
   my @dpd_list = ();
   for ($i = 0; $i < $dp_num; $i++)
   {
      $dpd_ref = dpd_create($chip_name, $chip_type, $dp_names[$i], $dp_ptrs[$i]);
      if ( !defined($dpd_ref) )
      {  $emess = "Unable to extract debug port groups from database.";
         $emess .= "\nTarget chip : $chip_name";
         $emess .= "\nChip type   : $chip_type";
         $emess .= "\nDebug port  : $dp_names[$i]";
         return;
      }

      $dpd_list[$i] = $dpd_ref;  # Store the reference in the list

      if ($debug_flag & $DEBUG_DETAILS)
      {  printf "\ncreated: dpd_ref = $dpd_ref\n";
         dpd_dump($dpd_ref);
      }
   }

   #
   #  Store the DPD list in the global lookup
   # 
   $dpd_lookup{$chip_name} = \@dpd_list;
   return(\@dpd_list);
}


#-------------------------------------------------------------------------------

sub dpd_list_search
{
   printf "\ndpd_list_search( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $chip_name = $_[0];   # Full name of ASIC: <brick>:<board>:<chip>
   my $name      = $_[1];   # Name of debug port or control register
   my $type      = $_[2];   # Type: 'debug' or 'control'
   my ($dpd_list, $dpd_ref, $ctrl_reg);

   if ( !exists($dpd_lookup{$chip_name}) )
   {  $emess = "Unable to locate debug port data for chip.\nChip: $chip_name";
      return;
   }
   $dpd_list = $dpd_lookup{$chip_name};

   #
   #  Search through each entry in the array
   #
   if ( $type eq 'debug' )
   {  foreach $dpd_ref ( @$dpd_list )
      {  if ( $name eq $$dpd_ref[$DPD_NAME] )
         {  return( $dpd_ref );
         }
      }
   }
   else
   {  foreach $dpd_ref ( @$dpd_list )
      {  $ctrl_reg = $$dpd_ref[$DPD_CTRL_REG];
         if ( $name eq reg_name($ctrl_reg) )
         {  return( $dpd_ref );
         }
      }
   }
   return;
}

#-------------------------------------------------------------------------------

sub dpd_create
{  
   printf "\ndpd_create( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   return if ( $#_ != 3 );
   my $chip_name = $_[0];   # Full name of ASIC: <brick>:<board>:<chip>
   my $chip_type = $_[1];   # ASIC type/description (i.e. "shub_01");
   my $name      = $_[2];   # Name of debug port (i.e. "SH_LB_DEBUG_DATA")
   my $dport     = $_[3];   # Pointer to the debug port C-structure
   my ($tmp, $full_name, $reg_ref, $i);

   #  Extract the list of group names and group structure pointers
   my @grp_names = ();      # Array for group names
   my @grp_ptrs  = ();      # Array for group C-structure pointers
   my $num = pcf_dport_get_gctrl_list($dport, \@grp_names, \@grp_ptrs);
   return if ( $num <= 0 ); # No valid groups returned
   my @grp_flags = ();      # Array of group flags
   for ($i = 0; $i < $num; $i++)
   {  $grp_flags[$i] = 0;
   }

   #  Store the values in the DPD structure
   my @dpd_data  = ("", 0, 0, 0, 0, 0, 0, 0, 0, 0);
   $dpd_data[$DPD_NAME]       = $name;
   $dpd_data[$DPD_DPORT_PTR]  = $dport;
   $dpd_data[$DPD_FLAGS]      = 0;
   $dpd_data[$DPD_GRP_NAMES]  = \@grp_names;
   $dpd_data[$DPD_GRP_PTRS]   = \@grp_ptrs;
   $dpd_data[$DPD_GRP_FLAGS]  = \@grp_flags;

   #  Search for the debug port in the MMR list (does it have an MMR?)
   $full_name = $chip_name . ":" . $name;
   $reg_ref = reg_lookup_create_inst($full_name, $chip_type, $name);
   if ( reg_is_valid($reg_ref) )
   {  $dpd_data[$DPD_DATA_REG]  = $reg_ref; # register from %reg_lookup
      $dpd_data[$DPD_FLAGS] |= $DPD_FLAG_MMR;
   }
   else
   {  $dpd_data[$DPD_DATA_REG]  = 0; # No MMR for debug port data
   }

   #  Extract the CTRL specifics: name, mask, data
   $tmp = pcf_dport_get_pctrl_parm($dport, 'name');
   $full_name = $chip_name . ":" . $tmp;
   $reg_ref = reg_lookup_create_inst($full_name, $chip_type, $tmp);
   if ( reg_is_valid($reg_ref) )
   {  $dpd_data[$DPD_CTRL_REG]  = $reg_ref; # register from %reg_lookup
   }
   else
   {  $emess = "Unable to locate control register in MMR database.";
      $emess .= "\nRegister name: $tmp";
      $emess .= "\nFull name: $full_name";
      dsp_error_and_exit($emess);
      return ;
   }
   $dpd_data[$DPD_CTRL_MASK] = pcf_dport_get_pctrl_parm($dport, 'mask');
   $dpd_data[$DPD_CTRL_MASK] &= 0xFFFFFFFFFFFFFFFF;
   $dpd_data[$DPD_CTRL_DATA] = pcf_dport_get_pctrl_parm($dport, 'value');
   $dpd_data[$DPD_CTRL_DATA] &= 0xFFFFFFFFFFFFFFFF;

   #  Add in the override values
   group_override_update_flag($dpd_ref);

   if ( $debug_flag & $DEBUG_DETAILS )
   {  my $ref = \@dpd_data;
      printf "dpd_data  = $ref = ( @dpd_data )\n";
   }
   return( \@dpd_data );
}

#-------------------------------------------------------------------------------

sub group_override_update_flag
{  
   printf "\ngroup_override_update_flag( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $dpd_ref   = $_[0];  # Reference to debug port definition data
   my $grp_ptrs  = $$dpd_ref[$DPD_GRP_PTRS];
   my $grp_flags = $$dpd_ref[$DPD_GRP_FLAGS];

   #
   #  Process each of the groups
   #
   my ($i, $gptr, $or_ptr, $mask, $oval);
   for ($i = 0; $i <= $#$grp_ptrs; $i++)
   {  $gptr   = $$grp_ptrs[$i];
      $or_ptr = pcf_gctrl_get_parm($gptr, 'override');
      if ( $or_ptr != 0 )
      {  $mask = pcf_gctrl_get_parm($gptr, 'override mask') & 0xFFFFFFFFFFFFFFFF;
         $oval = pcf_gctrl_get_parm($gptr, 'override value') & 0xFFFFFFFFFFFFFFFF;
         $$grp_flags[$i] = ( (($$dpd_ref[$DPD_CTRL_DATA] & $mask) == $oval) ? 1 : 0 );
      }
      else
      {  $$grp_flags[$i] = 0;
      }
   }

   if ( $debug_flag & $DEBUG_DETAILS )
   {  printf "done: grp_flags = $grp_flags = ( @$grp_flags )\n";
   }
}


#-------------------------------------------------------------------------------

sub dpd_dump
{  return if ( $#_ != 0 );
   my $dpd_ref = $_[0];
   my $aref;

   printf "\nDEBUG PORT DATA STRUCTURE CONTENTS:\n";
   printf "name      = $$dpd_ref[$DPD_NAME]\n";
   printf "port ptr  = 0x%08llX\n", $$dpd_ref[$DPD_DPORT_PTR];
   printf "flags     = 0x%08llX\n", $$dpd_ref[$DPD_FLAGS];
   printf "port reg  = $$dpd_ref[$DPD_DATA_REG]\n";
   printf "ctrl reg  = $$dpd_ref[$DPD_CTRL_REG]\n";
   printf "ctrl mask = 0x%016llX\n", $$dpd_ref[$DPD_CTRL_MASK];
   printf "ctrl data = 0x%016llX\n", $$dpd_ref[$DPD_CTRL_DATA];
   printf "grp names = $$dpd_ref[$DPD_GRP_NAMES]\n";
   $aref = $$dpd_ref[$DPD_GRP_NAMES];
   printf "          = ( @$aref )\n";
   printf "grp ptrs  = $$dpd_ref[$DPD_GRP_PTRS]\n";
   $aref = $$dpd_ref[$DPD_GRP_PTRS];
   printf "          = ( @$aref )\n";
}


################################################################################

sub create_debug_access_window
{
   printf "\ncreate_debug_access_window( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   return if ( $#_ != 5 );

   my $top_win   = $_[0]; # SuperScope top-level window
   my $chip_name = $_[1]; # Full name of ASIC: <brick>:<board>:<chip>
   my $chip_type = $_[2]; # ASIC type/description (i.e. "shub_01");
   my $name      = $_[3]; # Name of debug port (i.e. "SH_LB_DEBUG_DATA")
   my $dpd_ref   = $_[4]; # pointer to debug port data
   my $debug_log = $_[5]; # reference to handle for main window history/log widget

   my $dport     = $$dpd_ref[$DPD_DPORT_PTR];
   my $text      = $chip_name . ":" . $name;
   my @tpw_list  = ();
   my @dwd_data  = ( 0, "", "", "", "", \@tpw_list, $debug_log );
   my $val       = 0;

   my ($mw_debug, $frame, $label, $menub, $hexval, $entry, $canvas, $can_frame, $bwidth,
       $grp_names, $grp_ptrs, $num_grps, $i, $gptr, $num_tp, $btn, $text_wid,
       $reg_ref, $desc, $char_height, $can_height, $twid_width, $grp_flags,
       $or_ptr, $tp_pos, $tp_name, $text);

   my @names = split(":", $chip_name);
   if ( !(exists $modules{ $names[0] })  )
   {  $emess  = "Specified module does not exist in the current configuration.";
      $emess .= "\nModule: $names[0]\n";
      superscope_error_handler($emess);
      return($FALSE);
   }
   my $mod_ref = $modules{ $names[0] };
   my ($chn_ind, $chip_ind) = scan_module_find_chip($mod_ref, $names[1], $names[2], 'logical');

   #
   #  Create the window for the debug port
   #
   my $pname = $chip_name . ":" . $name;
   $mw_debug = $top_win->Toplevel(-title => "Debug Port Access Window: $pname");
   $mw_debug->bind('<Destroy>' => [\&destroy_debug_access_window, $pname, $mw_debug] );
   $dwd_data[$DWD_MAIN_WINDOW] = $mw_debug;

   #
   #  Create the menubar for the main debug port window
   #
   $menubar  = $mw_debug->Menu(-type => 'menubar');
   $mw_debug->configure(-menu => $menubar);
   $casmenu = $menubar->cascade(-label => '~File', -tearoff => 0);

   $casmenu->command(-label => 'Save Window Contents (to text file)...',
      -command => [ \&debug_access_save_data, $win_main, 
                    $chip_name, $chip_type, $dpd_ref, \@dwd_data]);
   $casmenu->separator;

   $casmenu->command(-label => 'Download Debug Port to Logic Analyzer...',
      -command => [ \&debug_access_download_to_la, $win_main, 
                    $chip_name, $chip_type, $dpd_ref]);
   $casmenu->separator;

   $casmenu->command(-label => 'Close',
       -command => [$mw_debug => 'destroy'] );

   $casmenu = $menubar->cascade(-label => '~Reset', -tearoff => 0);
   $casmenu->command(-label => 'Reset chip TAP(s) via TMS',
      -command => [\&jar_cbf_reset, $mod_ref, $chn_ind, 'soft']);
   $casmenu->command(-label => 'Reset chip TAP(s) via TRST',
      -command => [\&jar_cbf_reset, $mod_ref, $chn_ind, 'hard']);
   $casmenu->command(-label => 'Reset SIC and configure chain(s)',
      -command => [\&jar_cbf_reset, $mod_ref, $chn_ind, 'sic' ]);

   #
   #  Create the window header widgets
   #
   $frame = $mw_debug->Frame(-borderwidth => 10);
   $label = $frame->Label(-text => $chip_name, -anchor => 'w');
   $label->pack(-side => 'left', -fill => 'both');
   $frame->pack(-side => 'top', -fill => 'both');

   #
   #  Create the debug port data widgets
   #
   $frame = $mw_debug->Frame(-borderwidth => 10);
   $menub = $frame->Menubutton(-text => $name, -width => 40, -tearoff => 0,
                -relief => 'raised', -direction => 'below', -anchor => 'w');
   if ( $$dpd_ref[$DPD_FLAGS] & $DPD_FLAG_MMR )
   {  $reg_ref = $$dpd_ref[$DPD_DATA_REG];
      $val     = reg_get_value($reg_ref);
      $menub->command(-label => reg_note($reg_ref));
      $text = sprintf("Address: 0x%09lX", reg_addr($reg_ref));
      $menub->command(-label => $text);
   }
   else
   {  $val = $$dpd_ref[$DPD_DATA_REG];
      $menub->command(-label => "No corresponding MMR for debug data");
   }
   $menub->pack(-side => 'left');
   $dwd_data[$DWD_DATA_TEXT_HEX] = convert_num2text($val, 64, 'hex');
   $dwd_data[$DWD_DATA_TEXT_MMR] = convert_num2text($val, 64, 'hex');
   $hexval = $frame->Label(-textvariable => \$dwd_data[$DWD_DATA_TEXT_HEX], 
                ### -font => 'Fixed 10',
                -width => 24, -anchor => 'w');
   $hexval->pack(-side => 'left');
   $entry = $frame->Entry(-width => 26, ###-font => 'Fixed 10', 
               -textvariable => \$dwd_data[$DWD_DATA_TEXT_MMR]);
   $entry->bind( '<Return>' => [\&debug_access_entry_data, $chip_name, $dpd_ref, \@dwd_data ] );
   $entry->bind( '<Tab>' => [\&debug_access_entry_data, $chip_name, $dpd_ref, \@dwd_data ] );
   $entry->bind( '<Leave>' => [\&debug_access_entry_data, $chip_name, $dpd_ref, \@dwd_data ] );
   $entry->pack(-side => 'left');
   $frame->pack(-side => 'top', -fill => 'both');

   #
   #  Create the debug port control widgets
   #
   $reg_ref = $$dpd_ref[$DPD_CTRL_REG];
   if ( !reg_is_valid($reg_ref) )
   {  $emess = "Data corruption or programming error: no register!";
      $emess .= "\nControl register for $name is invalid.";
      return($FALSE); 
   }
   $frame = $mw_debug->Frame(-borderwidth => 10);
   $menub = $frame->Menubutton(-text => reg_name($reg_ref), -width => 40, 
                -tearoff => 0, -relief => 'raised', -direction => 'below', 
                -anchor => 'w');
   $menub->command(-label => reg_note($reg_ref));
   $text = sprintf("Address: 0x%09lX", reg_addr($reg_ref));
   $menub->command(-label => $text);
   $menub->pack(-side => 'left');
   $val = reg_get_value($reg_ref);
   $dwd_data[$DWD_CTRL_TEXT_HEX] = convert_num2text($val, 64, 'hex');
   $dwd_data[$DWD_CTRL_TEXT_MMR] = convert_num2text($val, 64, 'hex');
   $hexval = $frame->Label(-textvariable => \$dwd_data[$DWD_CTRL_TEXT_HEX],
                ###, -font => 'Fixed 10',
                -width => 24, -anchor => 'w');
   $hexval->pack(-side => 'left');
   $entry = $frame->Entry(-width => 26, ###-font => 'Fixed 10', 
               -textvariable => \$dwd_data[$DWD_CTRL_TEXT_MMR]);
   $entry->bind( '<Return>' => [\&debug_access_entry_ctrl, $chip_name, $dpd_ref, \@dwd_data] );
   $entry->bind( '<Tab>' => [\&debug_access_entry_ctrl, $chip_name, $dpd_ref, \@dwd_data] );
   $entry->bind( '<Leave>' => [\&debug_access_entry_ctrl, $chip_name, $dpd_ref, \@dwd_data] );
   $entry->pack(-side => 'left');
   $frame->pack(-side => 'top', -fill => 'both');

   #
   #  Create the canvas for the group descriptions
   # note: determine height by counting the number of registers and fields
   #
   $canvas = $mw_debug->Scrolled('Canvas', -scrollbars => 'se',
                -scrollregion => ['0', '0', '200', '600'],
                -borderwidth => 2, -relief => 'sunken');
   $can_frame = $canvas->Frame();

   $char_height = 12;   # Should get this from font parameters
   $can_height  = 10;
   $twid_width  = 160;

   #
   #  Create the widgets for each group
   #
   $grp_names  = $$dpd_ref[$DPD_GRP_NAMES];
   $grp_ptrs   = $$dpd_ref[$DPD_GRP_PTRS];
   $grp_flags  = $$dpd_ref[$DPD_GRP_FLAGS];
   $num_grps   = $#$grp_ptrs + 1;
   for ($i = 0; $i < $num_grps; $i++)
   {
      $gptr   = $$grp_ptrs[$i];
      $num_tp = pcf_gctrl_get_parm($gptr, 'num');

      #height      = (button height   ) + (ROText height                     ) + (border)
      $can_height += ($char_height * 2) + (($char_height + 1) * ($num_tp + 2)) + 4;

      #
      #  Create the widgets for the label and configure button
      #
      $frame = $can_frame->Frame();
      $label = $frame->Label(-width => 5);
      $label->pack(-side => 'left', -fill => 'both');
      $btn   = $frame->Button(-text => $$grp_names[$i], 
                  -width => 15, -anchor => 'w');
      $btn->pack(-side => 'left');
      $btn   = $frame->Button(-text => 'Configure', -width => 10);
      $btn->pack(-side => 'left');

      #
      #  Determine if group has override
      #
      $or_ptr = pcf_gctrl_get_parm($gptr, 'override');
      if ( $or_ptr != 0 )
      {  $label = $frame->Label(-width => 3);
         $label->pack(-side => 'left');
         $tp_pos  = pcf_gctrl_get_parm($gptr, 'override pos');
         $tp_name = pcf_gctrl_get_parm($gptr, 'override name');
         $text    = "Enable override of bit $tp_pos with \"$tp_name\"";
         $chk_btn = $frame->Checkbutton(-variable => \$$grp_flags[$i], -text => $text,
                       -command => [\&debug_access_override_cb, $chip_name, $dpd_ref, \@dwd_data, $i]);
         $chk_btn->pack(-side => 'left');
      }
      $frame->pack(-side => 'top', -fill => 'both');

      #
      #  Create the widgets for the group testpoints
      #
      $frame = $can_frame->Frame();
      $label = $frame->Label(-width => 5);
      $label->pack(-side => 'left', -fill => 'both');
      $text_wid = $frame->ROText(-height => ($num_tp + 1), -width => $twid_width);
      $text_wid->pack(-side => 'left', -fill => 'both');
      $frame->pack(-side => 'top', -fill => 'both');

      $btn->configure(-command => [\&create_configure_group_window, $chip_name, $name, $dpd_ref, $i]);

      #
      #  Update the group with the current testpoints
      #
      $val = convert_text2num($dwd_data[$DWD_DATA_TEXT_HEX]);
      debug_access_update_group_tps($text_wid, $dport, $$grp_ptrs[$i], $val);

      #  Store the text widget in the window display structure
      $tpw_list[$i] = $text_wid;
   }
   $can_frame->pack(-side => 'top', -fill => 'both', -expand => 'yes');
   $canvas->configure(-scrollregion => [0, 0, (80 + (7 * $twid_width)), $can_height]);
   $canvas->create('window', 4, 4, -anchor => 'nw', -window => $can_frame);
   $canvas->pack(-side => 'top', -fill => 'both', -expand => 'yes');

   #
   #  Create the action button widgets
   #
   $bwidth = 18;
   $frame = $mw_debug->Frame(-borderwidth => 10);
   $btn = $frame->Button(-text => 'Read Configuration', -width => $bwidth,
             -command => [\&debug_access_read_ctrl, $chip_name, $dpd_ref, \@dwd_data] );
   $btn->pack(-side => 'left', -padx => 5);
   $btn = $frame->Button(-text => 'Write Configuration', -width => $bwidth,
             -command => [\&debug_access_write_ctrl, $chip_name, $name, $FALSE, $debug_log] );
   $btn->pack(-side => 'left', -padx => 5);
   $btn = $frame->Button(-text => 'Read Debug Data', -width => $bwidth,
             -command => [\&debug_access_read_data, $chip_name, $dpd_ref, \@dwd_data] );
   $btn->pack(-side => 'left', -padx => 5);
   if ( ($$dpd_ref[$DPD_FLAGS] & $DPD_FLAG_MMR) == 0 )
   {  $btn->configure(-state => 'disable');
   }
   $label = $frame->Label(-anchor => 'w', -width => 8);
   $label->pack(-side => 'left', -expand => 'no');
   $btn = $frame->Button(-text => 'Close Window', -width => $bwidth,
             -command => [$mw_debug => 'destroy'] );
   $btn->pack(-side => 'left', -padx => 5);
   $frame->pack(-fill => 'both');


   #
   #  Add window specifics to global lookup for chip (window is open)
   #
   $desc = $chip_name . ":" . $name;
   $dwd_lookup{$desc} = \@dwd_data;

   if ($debug_flag & $DEBUG_GENERAL)
   {  print "\nSTORED WINDOW DATA FOR: $desc = ( @dwd_data )\n";
   }

   return($TRUE);
}

#-------------------------------------------------------------------------------

sub destroy_debug_access_window
{
#   printf "\ndestroy_debug_access_window( @_ )\n";

   return if ( $#_ != 2 );
   my $curr_wid   = $_[0];  # Current widget/window handle from Tk
   my $dport_name = $_[1];  # Text widget for debug group display
   my $win_ref    = $_[2];  # Reference/handle to the debug port window

   return if ( $curr_wid != $win_ref ); # Return if not main window

   # Remove entry from "dwd_lookup{}" for debug port
   delete $dwd_lookup{$dport_name};

   if ($debug_flag & $DEBUG_GENERAL)
   {  print "\nDELETED WINDOW DATA FOR: $dport_name\n";
   }
}


#-------------------------------------------------------------------------------

sub debug_access_save_data
{
   printf "\ndebug_access_save_data( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);
   return if ( $#_ != 4 );

   my $win_top   = $_[0];  # Top-level window for file dialog
   my $chip_name = $_[1];  # Full name of ASIC: <brick>:<board>:<chip>
   my $chip_type = $_[2];  # ASIC type/decription (i.e. "shub_01")
   my $dpd_ref   = $_[3];  # Reference to debug port definition data
   my $dwd_ref   = $_[4];  # Reference to debug port window data

   #
   #  Prompt the user for the file name
   #
   my @types = ( ['SuperScope Data Files (Text)', '.txt']  );
   my $out_file = $win_top->getSaveFile(
#                  -title => 'Save Window Contents'
#                  -filetypes => \@types,
#                  -defaultextension => '.txt'
                   );
   return if ( !defined($out_file) );

   #
   #  Open the output file for writing
   #
   if ( !open(DATA_OUT_FILE, ">$out_file") )
   {  $emess = "Unable to open output file for writing.";
      $emess .= "\nFile: $out_file";
      superscope_error_handler($emess);
      return;
   }
   my ($val, $reg_ptr, $dport, $grp_names, $grp_ptrs, $g, $gptr, $path, $index, 
       $num_tp, @tps, $i, $mask, $name, $bit, $text);

   #
   #  Write the file header
   #
   printf DATA_OUT_FILE "#\n";
   printf DATA_OUT_FILE "# The file contains debug port access contents from\n";
   printf DATA_OUT_FILE "#    the SuperScope debug environemnt.\n";
   printf DATA_OUT_FILE "#    Window title : Debug Port Access Window\n";
   printf DATA_OUT_FILE "#    Target chip  : $chip_name\n";
   printf DATA_OUT_FILE "#    Chip type    : $chip_type\n";
   printf DATA_OUT_FILE "#    Debug port   : $$dpd_ref[$DPD_NAME]\n";
   printf DATA_OUT_FILE "#\n";

   #
   #  Process the debug port and control register values
   #
   $val = ( ($$dpd_ref[$DPD_FLAGS] & $DPD_FLAG_MMR) ? 
             reg_get_value($$dpd_ref[$DPD_DATA_REG]) :
             $$dpd_ref[$DPD_DATA_REG] );
   $reg_ptr = $$dpd_ref[$DPD_CTRL_REG];       # MMR register for control

   printf DATA_OUT_FILE "\nDEBUG PORT :  %-32s = 0x%016llX\n", 
      $$dpd_ref[$DPD_NAME], $val;
   printf DATA_OUT_FILE "CONTROL    :  %-32s = 0x%016llX\n", 
      reg_name($reg_ptr), reg_get_value($reg_ptr);

   #
   #  Move thru each group of the debug port
   #
   $dport     = $$dpd_ref[$DPD_DPORT_PTR];
   $grp_names = $$dpd_ref[$DPD_GRP_NAMES];
   $grp_ptrs  = $$dpd_ref[$DPD_GRP_PTRS];
   for ($g = 0; $g <= $#$grp_ptrs; $g++)
   {  
      $gptr   = $$grp_ptrs[$g];
      $path   = pcf_gctrl_get_parm($gptr, 'path');
      $index  = pcf_gctrl_get_parm($gptr, 'index');
      $num_tp = pcf_gctrl_get_parm($gptr, 'num');
      @tps    = ();
      if ( pcf_dport_get_tp_names($dport, $index, $num_tp, \@tps) != $num_tp )
      {  close(DATA_OUT_FILE);
         $emess = "Unable to extract group testpoints from database.";
         $emess = "\nOutput file is incomplete.";
         $emess .= "\nFile: $out_file";
         superscope_error_handler($emess);
         return;
      }

      printf DATA_OUT_FILE "\nTESTPOINTS FOR GROUP : $$grp_names[$g]\n";
      printf DATA_OUT_FILE "PATH=$path\n";

      for ($i = 0, $mask = (1 << $index); $i <= $#tps; $i++, $mask<<=1)
      {  $name = ( $tps[$i] eq "" ) ? $utext : $tps[$i];
         $bit  = ( ($mask & $val) ? 1 : 0 );
         $text = sprintf(" %2u)  %-60s  %u", ($i + $index), $name, $bit);
         printf DATA_OUT_FILE "$text\n";
      }
   }

   reg_write_contents($reg_ptr, DATA_OUT_FILE, $num_base);

   close(DATA_OUT_FILE);

   print "\nSaved window contents: $out_file\n";
}


#-------------------------------------------------------------------------------

sub debug_access_download_to_la
{
   printf "\ndebug_access_download_to_la( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);
   return if ( $#_ != 3 );

   my $win_top   = $_[0];  # Top-level window for file dialog
   my $chip_name = $_[1];  # Full name of ASIC: <brick>:<board>:<chip>
   my $chip_type = $_[2];  # ASIC type/decription (i.e. "shub_01")
   my $dpd_ref   = $_[3];  # Reference to debug port definition data

   #
   #  Extract the testpoints for the debug port
   #
   my $dport  = $$dpd_ref[$DPD_DPORT_PTR];
   my $num_tp = pcf_dport_get_parm($dport, 'ntp');
   if ( $num_tp == -1 )
   {  $emess = "Unable to extract debug port parameter from database.";
      superscope_error_handler($emess);
      return;
   }
   my @tps    = ();
   if ( pcf_dport_get_tp_names($dport, 0, $num_tp, \@tps) != $num_tp )
   {  $emess = "Unable to extract debug port testpoints from database.";
      superscope_error_handler($emess);
      return;
   }

   if ( $debug_flag & $DEBUG_DETAILS )
   {  printf "num_tp = $num_tp\n";
      printf "tps    = ( @tps )\n";
   }

   my $debug_name = $chip_name . ":" . $$dpd_ref[$DPD_DPORT_NAME];
   create_la_download_window($win_top, $debug_name, \@tps, \@la_chnmap, \@la_parms);
}


#-------------------------------------------------------------------------------

sub debug_access_update_group_tps
{  
   printf "\ndebug_access_update_group_tps( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);
   return if ( $#_ != 3 );

   my $text_wid   = $_[0];  # Text widget for debug group display
   my $dport      = $_[1];  # Debug port pointer (to C-structure)
   my $gptr       = $_[2];  # Group control pointer (to C-structure)
   my $value      = $_[3];  # Actual value of debug port data
   my (@tps, $i, $name, $text, $bit, $mask, $rv);

   #
   #  Get the group parameters and testpoints
   #
   my $path   = pcf_gctrl_get_parm($gptr, 'path');
   my $index  = pcf_gctrl_get_parm($gptr, 'index');
   my $num_tp = pcf_gctrl_get_parm($gptr, 'num');
   printf "\nfrom gctrl: path=\"$path\", index=$index, num_tp=$num_tp\n" if ($debug_flag & $DEBUG_DETAILS);
   @tps = ();
   $rv = pcf_dport_get_tp_names($dport, $index, $num_tp, \@tps);
#print "rv = $rv\n";
   if ( $rv != $num_tp )
   {  return(-1);
   }
#printf "tps = ( @tps )\n";

   #
   #  Clear out existing information and add new values
   #
   $text_wid->delete('1.0', 'end');
   $text = $path . "\n";
   $text_wid->insert('end', $text);
   for ($i = 0, $mask = (1 << $index); $i <= $#tps; $i++, $mask<<=1)
   {  $name = ( $tps[$i] eq "" ) ? $utext : $tps[$i];
      $bit  = ( ($mask & $value) ? 1 : 0 );
      $text = sprintf(" %2u)  %-60s  %u\n", ($i + $index), $name, $bit);
      $text_wid->insert('end', $text);
   }
   return($i-1);
}


#-------------------------------------------------------------------------------

sub debug_access_entry_data
{
   print "\ndebug_access_entry_data( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   return if ( $#_ != 3 );
   my $entry_ref = $_[0];  # Tk callback Entry reference
   my $chip_name = $_[1];  # Full name of ASIC: <brick>:<board>:<chip>
   my $dpd_ref   = $_[2];  # Reference to debug port definition data
   my $dwd_ref   = $_[3];  # Reference to debug port window data

   #
   #  Get the new value from the entry widget and update
   #
   my $val = convert_text2num($$dwd_ref[$DWD_DATA_TEXT_MMR]);
   debug_access_update_data_value($chip_name, $dpd_ref, $dwd_ref, $val, $TRUE);
}


#-------------------------------------------------------------------------------

sub debug_access_entry_ctrl
{
   print "\ndebug_access_entry_ctrl( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   return if ( $#_ != 3 );
   my $entry_ref = $_[0];  # Tk callback Entry reference
   my $chip_name = $_[1];  # Full name of ASIC: <brick>:<board>:<chip>
   my $dpd_ref   = $_[2];  # Reference to debug port definition data
   my $dwd_ref   = $_[3];  # Reference to debug port window data

   #
   #  Update the control register value (only bits valid for this port)
   #
   my $reg_ref = $$dpd_ref[$DPD_CTRL_REG];  # Register reference for control
   my $val = convert_text2num($$dwd_ref[$DWD_CTRL_TEXT_MMR]);
   $val = reg_set_value($reg_ref, $val);
   debug_access_update_ctrl_value($chip_name, $reg_ref, $TRUE);

   #
   #  Update all the debug port(s) on the chip
   #
   debug_access_update_chip($chip_name);
}

#-------------------------------------------------------------------------------

sub debug_access_override_cb
{
   print "\ndebug_access_override_cb( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   return if ( $#_ != 3 );
   my $chip_name = $_[0];  # Full name of ASIC: <brick>:<board>:<chip>
   my $dpd_ref   = $_[1];  # Reference to debug port definition data
   my $dwd_ref   = $_[2];  # Reference to debug port window data
   my $grp_index = $_[3];  # Index to specific group in debug control
   my $grp_ptrs  = $$dpd_ref[$DPD_GRP_PTRS];
   my $grp_flags = $$dpd_ref[$DPD_GRP_FLAGS];
   my ($gptr, $mask, $reg_ref, $val);

   #
   #  Update the control register value (only bits valid for this port)
   #
   $reg_ref = $$dpd_ref[$DPD_CTRL_REG];  # Register reference for control
   $val  = convert_text2num($$dwd_ref[$DWD_CTRL_TEXT_MMR]);
   if ( pcf_gctrl_get_parm($gptr, 'override') != 0 )
   {  $gptr = $$grp_ptrs[$grp_index];
      $mask = pcf_gctrl_get_parm($gptr, 'override mask');
      $val  &= ~($mask);
      if ( $$grp_flags[$grp_index] )
      {  $val |= pcf_gctrl_get_parm($gptr, 'override value');
      }
   }
   $val = reg_set_value($reg_ref, $val);
   debug_access_update_ctrl_value($chip_name, $reg_ref, $TRUE);

   #
   #  Update all the debug port(s) on the chip
   #
   debug_access_update_chip($chip_name);
}


#-------------------------------------------------------------------------------

sub debug_access_write_ctrl
{
   printf "\ndebug_access_write_ctrl( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $chip_name = $_[0]; # Full name of ASIC: <brick>:<board>:<chip>
   my $port_name = $_[1]; # Name of debug port data
   my $all_flag  = $_[2]; # Flag: TRUE=write all control MMRs
   my $debug_log = $_[3]; # Reference to handle for main window history/log widget
   my ($dpd_list, $i, $dpd_ref, $dport, $reg_ref, $text);

   if ( defined($debug_log) )
   {  $text = ( $all_flag ? "\nWrite configuration for all debug ports:\n" 
                      : "\nWrite configuration for $port_name:\n" );
      $debug_log->insert('end', $text);
      $debug_log->see('end');
   }

   #
   #  Target the ASIC for MMR operations
   #
   my @parms = split(":", $chip_name);
   mmr_target($SELECT, $parms[0], $parms[1], $parms[2]);

   #
   #  Move thru each debug port on the ASIC
   #
   $dpd_list = $dpd_lookup{$chip_name};  # Extract the chip's list of DPDs
   for ($i = 0; $i <= $#$dpd_list; $i++)
   {
      $dpd_ref = $$dpd_list[$i];                  # this port's DPD reference
      $dport   = $$dpd_ref[$DPD_DPORT_PTR];       # pointer to C-structure

#      printf "i=%02u dpd_ref=$dpd_ref -> flag=0x%08llX\n", $i, $$dpd_ref[$DPD_FLAGS];

      #  Determine if the control value has changed (needs to be written)
      if ( $all_flag || ($$dpd_ref[$DPD_FLAGS] & $DPD_FLAG_NEW_CTRL) ||
           ($$dpd_ref[$DPD_NAME] eq $port_name) )
      {
         $reg_ref = $$dpd_ref[$DPD_CTRL_REG];  # Register reference for control
         reg_mmr_write($reg_ref);

         if ( $ecode < 0 )
         {  $text = "MMR write operation(s) not completed due to error.\n";
            if ( defined($debug_log) )
            {  $debug_log->insert('end', $text);
               $debug_log->see('end');
            }
print $text;
            return;
         }

         $$dpd_ref[$DPD_FLAGS] &= ~($DPD_FLAG_NEW_CTRL); # Clear change flag

         #  Display MMR operation details
         $text = sprintf("%-24s mmr_write(0x%016llX, 0x%016llX) %s\n",
            $chip_name, reg_addr($reg_ref), reg_get_value($reg_ref),
            reg_name($reg_ref));
         if ( defined($debug_log) )
         {  $debug_log->insert('end', $text);
            $debug_log->see('end');
         }
print $text;
      }
   }
}


#-------------------------------------------------------------------------------

sub debug_access_read_ctrl
{
   printf "\ndebug_access_read_ctrl( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $chip_name = $_[0]; # Full name of ASIC: <brick>:<board>:<chip>
   my $dpd_ref   = $_[1]; # Reference to debug port definition
   my $dwd_ref   = $_[2]; # Reference to debug port window data
   my $debug_log = $$dwd_ref[$DWD_MAIN_HIST_WID]; # Main window history/log widget

   #
   #  Extract the register representation
   #
   my $reg_ref = $$dpd_ref[$DPD_CTRL_REG];

   if ( defined($debug_log) )
   {  $text = sprintf("\nRead configuration for : %s\n", reg_name($reg_ref));
      $debug_log->insert('end', $text);
      $debug_log->see('end');
   }

   #
   #  Target the correct ASIC for MMR operations
   #
   my @parms = split(":", $chip_name);
   mmr_target($SELECT, $parms[0], $parms[1], $parms[2]);

   #
   #  Read the debug control MMR
   #
   $val = reg_mmr_read($reg_ref);

   #
   #  Display MMR operation details
   #
   my $text = sprintf("%-24s mmr_read(0x%016llX) %s -> 0x%016llX\n",
            $chip_name, reg_addr($reg_ref), reg_name($reg_ref), $val);
   if ( defined($debug_log) )
   {  $debug_log->insert('end', $text);
      $debug_log->see('end');
   }
print $text;

   #
   #  Update the debug port structures with the new value
   #
   debug_access_update_ctrl_value($chip_name, $reg_ref, $TRUE);

   #
   #  Update all the debug port(s) on the chip
   #
   debug_access_update_chip($chip_name);

   #
   #  Clear the change flag: just read from HW...no need to write changes
   #
   $$dpd_ref[$DPD_FLAGS] &= ~($DPD_FLAG_NEW_CTRL);
}

#-------------------------------------------------------------------------------

sub debug_access_read_all_ctrl
{
   printf "\ndebug_access_read_all_ctrl( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $chip_name = $_[0]; # Full name of ASIC: <brick>:<board>:<chip>
   my $debug_log = $_[1]; # Reference to handle for main window history/log widget
   my ($val, $dpd_list, $i, $dpd_ref, $reg_reg, $name, $dwd_ref, $text);

   if ( defined($debug_log) )
   {  $debug_log->insert('end', "\nReading control register MMRs for all debug ports:\n");
      $debug_log->see('end');
   }

   #
   #  Target the specific ASIC for MMR operations
   #
   my @parms = split(":", $chip_name);
   mmr_target($SELECT, $parms[0], $parms[1], $parms[2]);

   #
   #  Move thru each debug port on the ASIC: read control MMR
   #
   $dpd_list = $dpd_lookup{$chip_name};  # Extract the chip's list of DPDs
   for ($i = 0; $i <= $#$dpd_list; $i++)
   {
      $dpd_ref  = $$dpd_list[$i];                 # this port's DPD reference
      $reg_ref  = $$dpd_ref[$DPD_CTRL_REG];       # MMR register for control

      #
      #  Read the debug control MMR
      #
      $val = reg_mmr_read($reg_ref);
      if ( $ecode < 0 )
      {  $text = "MMR read operation(s) not completed due to error.\n";
         if ( defined($debug_log) )
         {  $debug_log->insert('end', $text);
            $debug_log->see('end');
         }
print $text;
         return;
      }

      #
      #  Display MMR operation details
      #
      $text = sprintf("%-24s mmr_read(0x%016llX) %s -> 0x%016llX\n",
               $chip_name, reg_addr($reg_ref), reg_name($reg_ref), $val);
      if ( defined($debug_log) )
      {  $debug_log->insert('end', $text);
         $debug_log->see('end');
      }
print $text;

      #
      #  Update the debug port structures with the new value
      #
      debug_access_update_ctrl_value($chip_name, $reg_ref, $TRUE);
   } 

   #
   #  Update all the debug port(s) on the chip
   #
   debug_access_update_chip($chip_name);

   #
   #  Clear the change flag: just read from HW...no need to write changes
   #
   for ($i = 0; $i <= $#$dpd_list; $i++)
   {  $dpd_ref  = $$dpd_list[$i];                 # this port's DPD reference
      $$dpd_ref[$DPD_FLAGS] &= ~($DPD_FLAG_NEW_CTRL);
   }
}


#-------------------------------------------------------------------------------

sub debug_access_read_data
{
   printf "\ndebug_access_read_data( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $chip_name = $_[0]; # Full name of ASIC: <brick>:<board>:<chip>
   my $dpd_ref   = $_[1]; # Reference to debug port definition
   my $dwd_ref   = $_[2]; # Reference to debug port window data
   my $debug_log = $$dwd_ref[$DWD_MAIN_HIST_WID]; # Main window history/log widget

   return if ( ($$dpd_ref[$DPD_FLAGS] & $DPD_FLAG_MMR) == 0 );
   my ($reg_ref, $val);

   #
   #  Extract the register definition
   #
   $reg_ref = $$dpd_ref[$DPD_DATA_REG];

   if ( defined($debug_log) )
   {  $text = sprintf("\nRead debug data for : %s\n", reg_name($reg_ref));
      $debug_log->insert('end', $text);
      $debug_log->see('end');
   }

   #
   #  Target the correct ASIC for MMR operations
   #
   my @parms = split(":", $chip_name);
   mmr_target($SELECT, $parms[0], $parms[1], $parms[2]);

   #
   #  Read the debug data MMR
   #
   $val = reg_mmr_read($reg_ref);

   #
   #  Update the debug port structures with the new value
   #
   debug_access_update_data_value($chip_name, $dpd_ref, $dwd_ref, $val, $TRUE);

   #
   #  Display MMR operation details
   #
   my $text = sprintf("%-24s mmr_read(0x%016llX) %s -> 0x%016llX\n",
            $chip_name, reg_addr($reg_ref), reg_name($reg_ref), $val);
   if ( defined($debug_log) )
   {  $debug_log->insert('end', $text);
      $debug_log->see('end');
   }
print $text;
}


#-------------------------------------------------------------------------------

sub create_configure_group_window
{
   printf "\ncreate_configure_group_window( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   return if ( $#_ != 3 ); 

   my $chip_name = $_[0]; # Full name of ASIC: <brick>:<board>:<chip>
   my $name      = $_[1]; # Name of debug port (i.e. "SH_LB_DEBUG_DATA")
   my $dpd_ref   = $_[2]; # Reference to debug port definition
   my $grp_num   = $_[3]; # Index of the group within the debug port

   my ($dport, $grp_names, $grp_ptrs, $gptr);
   $dport     = $$dpd_ref[$DPD_DPORT_PTR]; # Debug port pointer (to C-structure)
   $grp_names = $$dpd_ref[$DPD_GRP_NAMES];
   $grp_ptrs  = $$dpd_ref[$DPD_GRP_PTRS];
   $gptr      = $$grp_ptrs[$grp_num];      # Group pointer

   my @gwd_data = (0, 0, 0, 0, 0, 0, 0, 0, 0);  # Initialize window info array
   my @sel_vals = ();

   #  Create the main window 
   my $mw_config  = $win_main->Toplevel(-title => "Debug Port Group Selection Window");
   $mw_config->bind('<Destroy>' => [\&destroy_configure_group_window, $chip_name, $name, $mw_config] );

   #  Create the main frames in the window
   my $main_frame = $mw_config->Frame(-borderwidth => 5);
   my $frame      = $main_frame->Frame();

   my ($label, $text, @grps, $num, $curr_ind, $curr_sel, $gmenu, $glist, $tp_text, 
       $bwidth, $cfg_btn, $btn);

   #  Create the header information
   $label = $frame->Label();
   $label->pack(-side => 'top', -fill => 'both');
   $text  = "Target chip : $chip_name";
   $label = $frame->Label(-anchor => 'w', -text => $text);
   $label->pack(-side => 'top', -fill => 'both');
   $text  = "Debug port  : $$dpd_ref[$DPD_NAME]";
   $label = $frame->Label(-anchor => 'w', -text => $text);
   $label->pack(-side => 'top', -fill => 'both');
   $text  = "Debug group : $$grp_names[$grp_num]";
   $label = $frame->Label(-anchor => 'w', -text => $text);
   $label->pack(-side => 'top', -fill => 'both');
   $label = $frame->Label();
   $label->pack(-side => 'top', -fill => 'both');

   #  Create the selection options widgets for the group options
   $label  = $frame->Label(-anchor => 'w', -text => "Group Selection Options:");
   $label->pack(-side => 'top', -fill => 'both');
   @grps   = ();
   $num    = pcf_gctrl_get_grp_list($gptr, \@grps);
   if ( $num < 1 )
   {  printf "\nERROR - No valid group selection options!\n";
      return;
   }

   #  Get group options and determine the current selection
   ($curr_ind, $curr_sel) = configure_get_current_selection($gptr, $num, \@grps);

   #  Create the group option menu
   $gmenu = $frame->Optionmenu(-options => \@grps, -width => 30, -anchor => 'w',
               -textvariable => \$gwd_data[$GWD_WID_GRP_OPTION]);
   $gmenu->pack(-side => 'top', -fill => 'both');

   #  Create the list of selection options within a group option
   $glist = $frame->Scrolled('Listbox', -scrollbars => 'se', -height => 8,
               -width => 50);
   $glist->bind('<Double-1>' => [\&configure_cbf_select, $chip_name, $dpd_ref, $grp_num, \@gwd_data]);
   $glist->bind('<Button>' => [\&configure_cbf_sel_option, \@gwd_data, $gptr]);
   $glist->pack(-side => 'top', -fill => 'both');

   #  Create the list of testpoints for the selection option
   $label   = $frame->Label(-anchor => 'w', -text => "Selected Testpoints:");
   $tp_text = $frame->Scrolled('ROText', -scrollbars => 'se',
                -height => 6, -width => 50);
   $label->pack(-side => 'top', -fill => 'both');
   $tp_text->pack(-side => 'top', -fill => 'both');

   $label   = $frame->Label(-text => "");
   $label->pack(-side => 'top', -fill => 'both');
   $chk_btn = $frame->Checkbutton(
                 -text => "Configure all subsequent port groups to consecutive selection options.",
                 -variable => \$gwd_data[$GWD_CB_CFG_MULTI],
                 );
   $chk_btn->pack(-side => 'top', -fill => 'both');

   if ( $grp_num >= $#$grp_ptrs )
   {  $chk_btn->configure(-state => 'disable');
   }

   $frame->pack(-side => 'top', -fill => 'both');
   $main_frame->pack(-side => 'top');

   $bwidth = 10;
   $frame = $mw_config->Frame(-borderwidth => 10);
   $btn = $frame->Button(-text => 'Select', -width => $bwidth,
                  -command => [\&configure_cbf_select, 0, $chip_name, $dpd_ref, $grp_num, \@gwd_data]);
   $btn->pack(-side => 'left', -padx => 5);
   $btn->pack(-side => 'left', -padx => 5);
   $cfg_btn = $frame->Button(-text => 'Configure', -width => $bwidth, -state => 'disabled');
   $cfg_btn->pack(-side => 'left', -padx => 5);
   $btn = $frame->Button(-text => 'Cancel', -width => $bwidth,
      -command => [ $mw_config => 'destroy' ]);
   $btn->pack(-side => 'left', -padx => 5);
   $frame->pack(-side => 'top', -fill => 'both');

   $cfg_btn->configure(-command => [\&configure_cbf_selopt_configure, $chip_name, $dpd_ref, 
      $grp_num, \@gwd_data, $cfg_btn]);

   $mw_config->bind( '<Enter>' => [\&configure_group_window_enter, \@gwd_data, $gptr ] );

   #
   #  Store values into the "group window data" structure
   #
   $gwd_data[$GWD_MAIN_WINDOW]    = $mw_config;
   $gwd_data[$GWD_WID_GRP_MENU]   = $gmenu;
   $gwd_data[$GWD_WID_GRP_OPTION] = $grps[$curr_ind];
   $gwd_data[$GWD_WID_GRP_LIST]   = $glist;
   $gwd_data[$GWD_WID_GRP_CURR]   = $curr_sel;
   $gwd_data[$GWD_WID_TP_TEXT]    = $tp_text;
   $gwd_data[$GWD_WID_CFG_BTN]    = $cfg_btn;
   $gwd_data[$GWD_SEL_VALS]       = \@sel_vals;
   $gwd_data[$GWD_CB_CFG_MULTI]   = $FALSE;

   #  Initialize for the "current" selected group
   $gmenu->configure(-command => [\&configure_cbf_sel_group, \@gwd_data, $gptr]);
   configure_cbf_sel_group(\@gwd_data, $gptr, $grps[$curr_ind]);
}


#-------------------------------------------------------------------------------

sub configure_group_window_enter
{
#   print "\nconfigure_group_window_enter( @_ )\n";

   return if ( $#_ != 2 );
   my $wid_ref  = $_[0]; # Widget/window handle from Tk
   my $gwd_ref  = $_[1]; # Group window data reference
   my $gptr     = $_[2]; # Group control pointer (to C-structure)

#print "window_enter: current =  $$gwd_ref[$GWD_WID_GRP_CURR]\n";

   #
   #  Extract the parameters from the group window data
   #
   $glist  = $$gwd_ref[$GWD_WID_GRP_LIST];

   #
   #  Return if list is still active
   #
   $cur = $glist->curselection();
   return if ( defined($cur) );

   #
   #  Re-select group in the listbox (previously saved)
   #
   $glist->selection('set', $$gwd_ref[$GWD_WID_GRP_CURR]);
   configure_cbf_sel_option(0, $gwd_ref, $gptr);
}


#-------------------------------------------------------------------------------

sub destroy_configure_group_window
{
#   print "destroy_configure_group_window( @_ )\n";

   my $obj_ref   = $_[0];  # Widget or window object being destroyed
   my $chip_name = $_[1];  # Full name of ASIC: <brick>:<board>:<chip>
   my $name      = $_[2];  # Name of debug port (i.e. "SH_LB_DEBUG_DATA")
   my $win_ref   = $_[3];  # Reference for the main window

   #  Delete lookup entry and free memory ???
}


#-------------------------------------------------------------------------------

sub configure_cbf_sel_group
{
   print "\nconfigure_cbf_sel_group( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   return if ( $#_ != 2 );
   my $gwd_ref  = $_[0]; # Group window data reference
   my $gptr     = $_[1]; # Group control pointer (to C-structure)
   my $grp_name = $_[2]; # Current group option name
   my ($aref, @sels, $num, $i, $glist, $curr_sel);
   if ( $debug_flag & $DEBUG_DETAILS )
   {  printf "gwd_ref  = $gwd_ref = ( @$gwd_ref )\n";
      printf "gptr     = $gptr\n";
      printf "grp_name = $grp_name\n";
   }

   #  Extract variables from the group window data
   $glist = $$gwd_ref[$GWD_WID_GRP_LIST];
   $aref  = $$gwd_ref[$GWD_SEL_VALS];

   #
   #  Extract the selections for this group option
   #
   @sels = ();
   @$aref = ();
   $num = pcf_gctrl_get_opt_list($gptr, $grp_name, \@sels, $aref);

   if ( $debug_flag & $DEBUG_DETAILS )
   {  printf "\ngroup options list: num=$num\n";
      printf "sels = ( @sels )\n";
      printf "vals = $aref = ( @$aref )\n";
   }

   #
   #  Update the listbox with the selections for this group option
   #
   $glist->delete(0, 'end');

   for ($i = 0; $i < $num; $i++)
   {  $glist->insert('end', $sels[$i]);
   }

   $curr_sel = ($$gwd_ref[$GWD_WID_GRP_CURR] > $num) ? $num : $$gwd_ref[$GWD_WID_GRP_CURR];
   $glist->selection('set', $curr_sel);
   $$gwd_ref[$GWD_WID_GRP_CURR] = $curr_sel;

   configure_cbf_sel_option(0, $gwd_ref, $gptr);
}


#-------------------------------------------------------------------------------

sub configure_cbf_sel_option
{
   print "\nconfigure_cbf_sel_option( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   return if ( $#_ != 2 );
   my $rsvd_ref = $_[0]; # Tk HASH parameter
   my $gwd_ref  = $_[1]; # Group window data reference
   my $gptr     = $_[2]; # Group control pointer (to C-structure)

   #  Extract variables from the group window data
   my $glist   = $$gwd_ref[$GWD_WID_GRP_LIST];
   my $tp_text = $$gwd_ref[$GWD_WID_TP_TEXT];
   my $cfg_btn = $$gwd_ref[$GWD_WID_CFG_BTN];
   my $aref    = $$gwd_ref[$GWD_SEL_VALS];

   #
   #  Get the current selection from the listbox
   #
   my $cur = $glist->curselection();
   return if ( !defined($cur) );
   $$gwd_ref[$GWD_WID_GRP_CURR] = $cur;

   #
   #  Delete the existing TP names, retrieve and display new TP names
   #
   $tp_text->delete('1.0', 'end');
   my @names = (); 
   my $flag  = pcf_gctrl_get_tp_names($gptr, $$aref[$cur], \@names);
   my $num   = $#names + 1;
   return if ( ($flag < 0) || ($num <= 0) );
   my ($i, $text);
   for ($i = 0; $i < $num; $i++)
   {  $text = (($names[$i] eq "") ? $utext : $names[$i]) . "\n";
      $tp_text->insert('end', $text);
   }

   my $name = ( $flag ? 'disable' : 'normal' );
   $cfg_btn->configure(-state => $name);
}


#-------------------------------------------------------------------------------

sub configure_cbf_selopt_configure
{
   print "\nconfigure_cbf_selopt_configure( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   return if ( $#_ != 4 );
   my $chip_name = $_[0];  # Full name of ASIC: <brick>:<board>:<chip>
   my $dpd_ref   = $_[1];  # Reference to debug port definition
   my $grp_num   = $_[2];  # Index of the group within the debug port
   my $gwd_ref   = $_[3];  # Group window data reference
   my $cfg_btn   = $_[4];  # The configure button on the group window

   #  Extract data from the data structures
   my ($grp_ptrs, $gptr, $glist, $aref, $cur);
   $grp_ptrs = $$dpd_ref[$DPD_GRP_PTRS];
   $gptr     = $$grp_ptrs[$grp_num];      # Name of this specific port
   $glist    = $$gwd_ref[$GWD_WID_GRP_LIST];
   $aref     = $$gwd_ref[$GWD_SEL_VALS];
   $cur      = $$gwd_ref[$GWD_WID_GRP_CURR];
   return if ( !defined($cur) );

   #
   #  Get the selection option data for the group in the port
   #
   my @data = ();
   my $rv = pcf_gctrl_get_selopt_dport($gptr, $$aref[$cur], \@data);

   #
   #  Update the list of testpoints
   #
   my $sub_dport    = $data[0];
   my $sub_grp_num  = $data[1];
   my $sub_name     = $$gwd_ref[$GWD_WID_GRP_OPTION];
   my $dpd_list     = $dpd_lookup{ $chip_name };
   my $sub_dpd_ref  = 0;
   my $i;
   for ($i=0; (($i <= $#$dpd_list) && ($sub_dpd_ref == 0)); $i++) 
   {
      $dpd_ref = $$dpd_list[$i];
      $sub_dpd_ref = $dpd_ref  if ( $sub_name eq $$dpd_ref[$DPD_NAME] );
   }
   return if ( $sub_dpd_ref == 0 );

   #
   #  Create another debug group configuration window
   #
   create_configure_group_window($chip_name, $sub_name, $sub_dpd_ref, $sub_grp_num);
}


#-------------------------------------------------------------------------------

sub configure_cbf_select
{
   printf "\nconfigure_cbf_select( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   return if ( $#_ != 4 );

   my $rsvd      = $_[0];  # Tk parameter if called from button action
   my $chip_name = $_[1];  # Full name of ASIC: <brick>:<board>:<chip>
   my $dpd_ref   = $_[2];  # Reference to debug port definition
   my $grp_num   = $_[3];  # Index of the group within the debug port
   my $gwd_ref   = $_[4];  # Group window data reference

   #  Extract data from the data structure
   my $dport    = $$dpd_ref[$DPD_DPORT_PTR]; # Debug port pointer (to C-structure)
   my $grp_ptrs = $$dpd_ref[$DPD_GRP_PTRS];
   my $grp_flags= $$dpd_ref[$DPD_GRP_FLAGS];
   my $gptr     = $$grp_ptrs[$grp_num];      # Name of this specific port
   my $cfg_win  = $$gwd_ref[$GWD_MAIN_WINDOW];
   my $glist    = $$gwd_ref[$GWD_WID_GRP_LIST];
   my $aref     = $$gwd_ref[$GWD_SEL_VALS];
   my $cfg_flag = $$gwd_ref[$GWD_CB_CFG_MULTI];
   my ($val, $reg_ref, $or_ptr, $mask);

   #
   #  Get the current selection from the listbox
   #
   my $cur = $glist->curselection();
   return if ( !defined($cur) );

   #
   #  Destroy the configuration/selection window
   #
   $cfg_win->destroy();

   #
   #  Determine if group supports testpoint override
   #
   $val = $$aref[$cur]; # Current selection value for group
   $or_ptr = pcf_gctrl_get_parm($gptr, 'override');
   if ( $or_ptr != 0 )
   {  $mask = pcf_octrl_get_parm($or_ptr, 'mask');
      $val  &= ~($mask);
      if ( $$grp_flags[$grp_num] )
      {  $val |= pcf_octrl_get_parm($or_ptr, 'value');
      }
   }

   #
   #  Update the debug port structure with the current selection
   #
   my $flag = pcf_dport_update_group($dport, $grp_num, $val, $cfg_flag);
printf "\ndport_update_group() : group=$grp_num, val=0x%08lX, flag=0x%016llX\n", $val, $flag if ($debug_flag & $DEBUG_GENERAL);
   if ( $flag == -1 ) # Error encountered
   {  $emess = "Invalid or corrupt parameters passed to library function.";
      printf "\nERROR - $emess\n";
      return;
   }

   #
   #  Update the control MMR (other debug ports may use this control register)
   #
   $reg_ref = $$dpd_ref[$DPD_CTRL_REG];  # Register reference for control
   $val = pcf_dport_get_pctrl_parm($dport, 'value') & 0xFFFFFFFFFFFFFFFF;
   $val = (reg_get_value($reg_ref) & ~($$dpd_ref[$DPD_CTRL_MASK])) | $val;
   reg_set_value($reg_ref, $val);
   debug_access_update_ctrl_value($chip_name, $reg_ref, $TRUE);

   #
   #  Update display for this and all debug ports and clear flags
   #
   debug_access_update_chip($chip_name);
}


#-------------------------------------------------------------------------------

sub dbgaccess_update_reg
{
   print "\ndbgaccess_update_reg( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   return if ( $#_ != 1 );
   my $chip_name = $_[0]; # Full ASIC name: <brick>:<board>:<chip>
   my $reg_ref   = $_[1]; # register pointer from reg_create_mmr_inst()
   my ($dpd_list, $i, $dpd_ref, $data_ref, $dwd_ref, $val, $flag);

   #
   #  Determine if ASIC has debug port data structure
   #
   return if ( $chip_name ne $debug_chip_name );
   return if ( !exists($dpd_lookup{$chip_name}) );
   $dpd_list = $dpd_lookup{$chip_name};
   $val = reg_get_value($reg_ref);

   #
   #  Is this register used for debug port data?
   #
   for ($i = 0; $i <= $#$dpd_list; $i++)
   {  $dpd_ref  = $$dpd_list[$i];                 # This port's DPD reference
      if ( $$dpd_ref[$DPD_FLAGS] & $DPD_FLAG_MMR )
      {  $data_ref = $$dpd_ref[$DPD_DATA_REG];
         if ( $data_ref == $reg_ref )             # Register matches this port
         {  $name = $chip_name . ":" . $$dpd_ref[$DPD_NAME];
            if ( exists $dwd_lookup{$name} )      # Window currently displayed?
            {  $dwd_ref   = $dwd_lookup{$name};
               debug_access_update_data_value($chip_name, $dpd_ref, $dwd_ref, $val, $FALSE);
            }
         }
      }
   }

   #
   #  Is this register used for debug port control?
   #
   $flag = debug_access_update_ctrl_value($chip_name, $reg_ref, $FALSE);
   if ( $flag )
   {  debug_access_update_chip($chip_name);
   }
}


#-------------------------------------------------------------------------------

sub debug_access_update_ctrl_value
{
   print "\ndebug_access_update_ctrl_value( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $chip_name = $_[0];  # Full name of ASIC: <brick>:<board>:<chip>
   my $reg_ref   = $_[1];  # Reference to register (MMR) structure
   my $upd_flag  = $_[2];  # Flag: TRUE=update other access window(s)
   my ($val, $dpd_list, $i, $dpd_ref, $ctrl_reg, $name, $dwd_ref, $found);

   $val = reg_get_value($reg_ref);    #  Extract the value from the register
#printf "MMR: reg_ref = $reg_ref -> %s = 0x%016llX\n", reg_name($reg_ref), $val;

   #
   #  Move thru each debug port on the ASIC: does MMR control it?
   #
   $dpd_list = $dpd_lookup{$chip_name};  # Extract the chip's list of DPDs
   $found    = $FALSE;
   for ($i = 0; $i <= $#$dpd_list; $i++)
   {
      $dpd_ref  = $$dpd_list[$i];                 # this port's DPD reference
      $ctrl_reg = $$dpd_ref[$DPD_CTRL_REG];       # MMR register for control
      if ( $ctrl_reg == $reg_ref )                # Control MMR matches this one
      {  
printf "\ndport_update() : val=0x%016llX\n", $val if ($debug_flag & $DEBUG_GENERAL);
         pcf_dport_update($$dpd_ref[$DPD_DPORT_PTR], $val);

         if ( $val != $$dpd_ref[$DPD_CTRL_DATA] )  # The value is different
         {  $$dpd_ref[$DPD_CTRL_DATA] = $val;
            $$dpd_ref[$DPD_FLAGS]     |= $DPD_FLAG_NEW_CTRL;
            group_override_update_flag($dpd_ref);
printf "Need to update the control MMR: 0x%016llX -> 0x%016llX\n", $$dpd_ref[$DPD_CTRL_DATA], $val if ($debug_flag & $DEBUG_GENERAL);
            $name = $chip_name . ":" . $$dpd_ref[$DPD_NAME];
            if ( exists $dwd_lookup{$name} )         # Window currently displayed?
            {  $dwd_ref   = $dwd_lookup{$name};
               $$dwd_ref[$DWD_CTRL_TEXT_HEX] = convert_num2text($val, 64, 'hex');
               $$dwd_ref[$DWD_CTRL_TEXT_MMR] = convert_num2text($val, 64, 'hex');
            }
         }

         $found = $TRUE;
      }
   } 

   #  Update the register access structure(s) for this register
   if ( $upd_flag && $found )
   {  regaccess_update_reg($chip_name, $reg_ref);
   }

   return($found);
}


#-------------------------------------------------------------------------------

sub debug_access_update_data_value
{
   print "\ndebug_access_update_data_value( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $chip_name = $_[0];  # Full name of ASIC: <brick>:<board>:<chip>
   my $dpd_ref   = $_[1];  # Reference to debug port definition data
   my $dwd_ref   = $_[2];  # Reference to debug port window data
   my $val       = $_[3];  # Value for debug port data
   my $upd_flag  = $_[4];  # Flag: TRUE=update other access window(s)
   my ($reg_ref, $dport, $grp_ptrs, $tpw_list, $g);

   #
   #  Update the text label and entry widget
   #
   $$dwd_ref[$DWD_DATA_TEXT_MMR] = convert_num2text($val, 64, 'hex');
   $$dwd_ref[$DWD_DATA_TEXT_HEX] = convert_num2text($val, 64, 'hex');
   if ( $$dpd_ref[$DPD_FLAGS] & $DPD_FLAG_MMR )
   {  $reg_ref = $$dpd_ref[$DPD_DATA_REG];
      reg_set_value($reg_ref, $val);

      #  Update the register access structure(s) for this register
      if ( $upd_flag )
      {  regaccess_update_reg($chip_name, $reg_ref);
      }
   }
   else
   {  $$dpd_ref[$DPD_DATA_REG] = $val;
   }

   #
   #  Move thru each group of the debug port
   #
   $dport    = $$dpd_ref[$DPD_DPORT_PTR];
   $grp_ptrs = $$dpd_ref[$DPD_GRP_PTRS];
   $tpw_list = $$dwd_ref[$DWD_TP_TEXT_LIST];
   for ($g = 0; $g <= $#$grp_ptrs; $g++)
   {  debug_access_update_group_tps($$tpw_list[$g], $dport, $$grp_ptrs[$g], $val);
   }
}


#-------------------------------------------------------------------------------

sub debug_access_update_chip
{
   print "\ndebug_access_update_chip( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   return if ( $#_ != 0 );
   my $chip_name = $_[0]; # Full name of ASIC: <brick>:<board>:<chip>
   return if ( !exists $dpd_lookup{$chip_name} );

   my $dpd_list = $dpd_lookup{$chip_name};  # Extract the chip's list of DPDs
   my ($i, $dport, $dpd_ref, $val, $reg_ref, $dwd_ref, $tpw_list,
       $grp_names, $grp_ptrs, $g, $mask, $data, $num, $text, $debug_log);

   #
   #  Determine if the main debug window is open for the chip
   #
   if ( exists( $mdwd_lookup{$chip_name} ) )
   {  my $mdw_ref = $mdwd_lookup{$chip_name};
      $debug_log  = $$mdw_ref[$MDWD_WID_HIST];
   }

   #
   #  Read the update flags for each debug port (and clear them)
   #
   my @flags = ();
   $num = pcf_dport_list_rdclr_upd($chip_name, \@flags);
print "\ndport_list_rdclr_upd() : number of ports that changed = $num\n" if ($debug_flag & $DEBUG_GENERAL);
   if ( $num == -1 )
   {  $emess = "Unable to allocate memory required for data structure(s).";
      dsp_error_and_exit($emess);
   }
   if ( $num == 0 )  # Nothing changed: do not update
   {  return;
   }

   if ( defined($debug_log) )
   {  $text = "\nThe debug port configuration action caused changes for:\n";
      $debug_log->insert('end', $text);
      $debug_log->see('end');
   }

   #
   #  Process each debug port: does it need to be updated?
   #
printf "\nMoving thru the debug ports: $chip_name\n" if ($debug_flag & $DEBUG_GENERAL);
   for ($i = 0; $i <= $#flags; $i++)
   {
      $dpd_ref = $$dpd_list[$i];                  # this port's DPD reference
      $dport   = $$dpd_ref[$DPD_DPORT_PTR];       # pointer to C-structure

printf "i=%02u dpd_ref=$dpd_ref -> flag=0x%08llX\n", $i, $flags[$i] if ($debug_flag & $DEBUG_GENERAL);

      #
      #  Update the debug port if necessary (bits set in flag)
      #
      if ( $flags[$i] )
      {  
         #
         #  Update the debug port history and testpoint display (if port window is open)
         #
         $grp_names = $$dpd_ref[$DPD_GRP_NAMES];
         $grp_ptrs  = $$dpd_ref[$DPD_GRP_PTRS];
         $name      = $chip_name . ":" . $$dpd_ref[$DPD_NAME];
         if ( exists $dwd_lookup{$name} )  # Testpoint display window open for port
         {  
printf "Need to update the debug port display: $name\n" if ($debug_flag & $DEBUG_DETAILS);
            $dwd_ref   = $dwd_lookup{$name};
            $tpw_list  = $$dwd_ref[$DWD_TP_TEXT_LIST];
            $data      = convert_text2num($$dwd_ref[$DWD_DATA_TEXT_HEX]);
            for ($g = 0, $mask = 1; $g <= $#$grp_ptrs; $g++, $mask <<= 1)
            {
               if ( $flags[$i] & $mask )
               {  debug_access_update_group_tps($$tpw_list[$g], $dport, $$grp_ptrs[$g], $data);
                  if ( defined($debug_log) )
                  {  $text = sprintf("%-24s %s\n", $$dpd_ref[$DPD_NAME], $$grp_names[$g]);
                     $debug_log->insert('end', $text);
                     $debug_log->see('end');
                  }
               }
            }
         }
         else # No testpoint display window for port: update history only
         {  
printf "Debug port not current displayed: $name\n" if ($debug_flag & $DEBUG_DETAILS);
            for ($g = 0, $mask = 1; $g <= $#$grp_ptrs; $g++, $mask <<= 1)
            {  if ( $flags[$i] & $mask )
               {  if ( defined($debug_log) )
                  {  $text = sprintf("%-24s %s\n", $$dpd_ref[$DPD_NAME], $$grp_names[$g]);
                     $debug_log->insert('end', $text);
                     $debug_log->see('end');
                  }
               }
            }
         }

         #
         #  Update the "debug port download" window, if it exists
         #
         update_la_download_window($name, $dport);
 
      } # end: if statement
   } # end: for loop

}


#-------------------------------------------------------------------------------
#  The following function is a bit of a kludge: it attempts to determine
#    the current selection indices for the "Debug Port Group Selection Window".
#    It expects the group control path to be in a specific format!
#-------------------------------------------------------------------------------

sub configure_get_current_selection
{
   printf "\nconfigure_get_current_selection( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $gptr = $_[0];  # Pointer to group control (C-structure)
   my $num  = $_[1];  # Number of group selection names
   my $gref = $_[2];  # Reference to array of group selection names
   my ($path, $grp_name, @sels, @data, $i, $found, $curr_ind, $curr_sel);

   #  Get the full-path and extract selected group name
   $path = pcf_gctrl_get_parm($gptr, 'path');
   $path =~ s/\/.*//;  # Remove lower levels
   $path =~ s/\s+$//g; # Remove all trailing whitespace
   $grp_name = $path;      # Copy path
   $grp_name =~ s/ .*//;   # Remove selection option name

   #  Find the current group in the list of possible
   for ($i = 0, $found=$FALSE; (($i <= $#$gref) && !$found); $i++)
   {  $found = ( $grp_name eq $$gref[$i] );
   }
   $curr_ind = ( $found ? ($i - 1) : 0 );

   #  Search for the specific option within current group
   @sels = ();
   @data = ();
   $num = pcf_gctrl_get_opt_list($gptr, $grp_name, \@sels, \@data);
   for ($i = 0, $found=$FALSE; (($i <= $#sels) && !$found); $i++)
   {  $found = ( $path eq $sels[$i] );
   }
   $curr_sel = ( $found ? ($i - 1) : 0 );

   return($curr_ind, $curr_sel);
}


#-------------------------------------------------------------------------------


1;

