# 
#/**************************************************************************
# *                                                                        *
# *               Copyright (C) 1989-1999 Silicon Graphics, Inc.           *
# *                                                                        *
# *  These coded instructions, statements, and computer programs  contain  *
# *  unpublished  proprietary  information of Silicon Graphics, Inc., and  *
# *  are protected by Federal copyright law.  They  may  not be disclosed  *
# *  to  third  parties  or copied or duplicated in any form, in whole or  *
# *  in part, without the prior written consent of Silicon Graphics, Inc.  *
# *                                                                        *
# **************************************************************************/
# 
# @(#)$Id: dbgmain.pm,v 1.5 2004/03/29 17:46:57 beighley Exp $
#
#-------------------------------------------------------------------------------
#
#  FILE: dbgmain.pm
#
#  DESCRIPTION:  This file contains PERL for the main debug port window.
#
# NOTE: The debug access modules currently support only one ASIC at a time.
#       There is not support for debug access window for multiple chips.
#
#  MODIFICATION HISTORY:
#
#     WHEN      WHO             WHAT
#     --------  --------------  -----------------------------------------
#     01/15/04  Karen Beighley  Added modification history to source file.
#                               Store "main" window handle, and changed
#                                  how additional windows are created.
#
#-------------------------------------------------------------------------------

use DebugSel;
use dbgaccess;
use dbgsearch;

#-------------------------------------------------------------------------------
#  CONSTANTS - These should not change
#-------------------------------------------------------------------------------

$MDWD_WINDOW      = 0;  # Main debug window: window reference
$MDWD_MAIN_WINDOW = 1;  # Main debug window: SuperScope main window
$MDWD_CHIP_NAME   = 2;  # Main debug window: full name "<brick>:<board>:<chip>"
$MDWD_CHIP_TYPE   = 3;  # Main debug window: chip type/description (i.e. "shub_01")
$MDWD_DPD_LIST    = 4;  # Main debug window: reference to array of debug port structures
$MDWD_WID_HIST    = 5;  # Main debug window: handle of history/log text widget
$MDWD_WID_LIST    = 6;  # Main debug window: handle of listbox widget
$MDWD_REF_TSWIN   = 7;  # Main debug window: reference to TP search window (if open)


#-------------------------------------------------------------------------------
#  GLOBAL VARIABLES
#-------------------------------------------------------------------------------

$debug_win = 0;  # Reference to main debug access window
$debug_chip_name = "";  # Main debug window created for this ASIC

%mdwd_lookup = ();

#-------------------------------------------------------------------------------

sub create_main_debug_window
{
   printf "\ncreate_main_debug_window( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   return if ( $#_ != 4 );
   my $main_win  = $_[0]; # Handle to main SuperScope window
   my $chip_name = $_[1]; # Full name of ASIC: <brick>:<board>:<chip>
   my $chip_type = $_[2]; # ASIC type/description (i.e. "shub_01")
   my $top_menu  = $_[3]; # Reference to top-level cascase menu
   my $read_flag = $_[4]; # Flag: TRUE=read control values from MMRs
   my ($dpd_list, $dpd_ref, $dport, $text, $frame, $label, $btn, $rd_btn, $wr_btn,
       $list, $desc, $bwidth, $menubar, $casmenu, $reg_ref);
   my @mdw_data = (0, 0, "", "", 0, 0, 0, 0);
   my $debug_log;

   #
   #  Create the debug port data structure(s)
   #
   $dpd_list = dpd_list_create($chip_name, $chip_type);
   if ( !defined($dpd_list) )
   {  dsp_error_and_exit($emess);
   }

   if ( $debug_flag & $DEBUG_GENERAL )
   {  print "\nMAIN DEBUG PORT RESULTS: $chip_name\n";
      print "dpd_list = ( @$dpd_list )\n";
      my (@dp_names, @dp_ptrs, $i);
      @dp_names = ();
      @dp_ptrs  = ();
      for ($i = 0; $i <= $#$dpd_list; $i++)
      {  $dpd_ref = $$dpd_list[$i];
         print "dpd_ref = $dpd_ref = ( @$dpd_ref )\n";
         $dp_names[$i] = $$dpd_ref[$DPD_NAME];
         $dp_ptrs[$i]  = $$dpd_ref[$DPD_DPORT_PTR];
      }
   }

   $debug_chip_name = $chip_name;   # Current ASIC that debug access window created for

   #
   #  Create the main window
   #
   $text = "Main Debug Port Access Window: $chip_name";
   $debug_win = $main_win->Toplevel(-title => $text);

   #
   #  Create the menubar for the main debug port window
   #
   $menubar  = $debug_win->Menu(-type => 'menubar');
   $debug_win->configure(-menu => $menubar);
   $casmenu = $menubar->cascade(-label => '~File', -tearoff => 0);
   $casmenu->command(-label => 'Save Debug Port Configuration...',
#TEMP - the window should be $debug_win insteaf of $main_win!!!
      -command => [ \&window_cfg_save, $main_win, 'DEBUG_ACCESS', 
                    'Main Debug Port Access Window',
                    $chip_name, $chip_type, $dpd_list]);
   $casmenu->separator;
   $casmenu->command(-label => 'Close All Debug Windows', 
      -command => [ $debug_win => 'destroy'] );

   $casmenu = $menubar->cascade(-label => '~Search', -tearoff => 0);
   my $cmd = $casmenu->command(-label => 'Search for specific testpoint(s) ...');
   $cmd->configure(-command => [ \&create_tp_search_window, $main_win, \@mdw_data, \$cmd ]);

   #
   #  Define the callback function if the main window is destroyed
   #
   $debug_win->bind('<Destroy>' => [\&main_debug_cbf_close, \@mdw_data, $top_menu, \$cmd]);

   #
   #  Create the debug port listbox selection
   #
   $frame = $debug_win->Frame(-borderwidth => 10);
   $text  = "Available Debug Ports on $chip_name";
   $label = $frame->Label(-text => $text, -anchor => 'w');
   $label->pack(-side => 'top', -fill => 'both');
   $list  = $frame->Scrolled('Listbox', -scrollbars => 'se', -height => 10,
                -font => 'Clean 12 bold', -selectmode => 'extended' );
   #  Add each debug port to the listbox
   for ($i = 0; $i <= $#$dpd_list; $i++)
   {  $dpd_ref = $$dpd_list[$i];                  # this port's DPD reference
      $dport   = $$dpd_ref[$DPD_DPORT_PTR];
      $desc    = pcf_dport_get_parm($dport, 'desc');
      $text    = sprintf("%-36s %s", $$dpd_ref[$DPD_NAME], $desc);
      $list->insert('end', $text);
   }
   $list->pack(-side => 'top', -fill => 'both', -expand => 'yes');
   $frame->pack(-side => 'top', -fill => 'both', -expand => 'yes');

   $list->bind('<Button-2>' => [ \&main_debug_cbf_create, \@mdw_data ]);
   $list->bind('<Button-3>' => [ \&main_debug_cbf_create, \@mdw_data ]);
   $list->bind('<Double-1>' => [ \&main_debug_cbf_create, \@mdw_data ]);

   $bwidth = 12;

   #
   #  Create a frame with some action buttons
   #
   $frame = $debug_win->Frame(-borderwidth => 10);
   $btn   = $frame->Button(-text => 'Create Window', -width => $bwidth,
               -command => [ \&main_debug_cbf_create, 0, \@mdw_data ]);
   $btn->pack(-side => 'left', -padx => 5);
   $rd_btn   = $frame->Button(-text => 'Read All Configuration', -width => $bwidth+10);
   $rd_btn->pack(-side => 'right', -padx => 5);

   $wr_btn   = $frame->Button(-text => 'Write All Configuration', -width => $bwidth+10);
   $wr_btn->pack(-side => 'right', -padx => 5);
   $frame->pack(-side => 'top', -fill => 'both');

   #
   #  Create a frame with the history/log widgets
   #
   $frame = $debug_win->Frame(-borderwidth => 10);
   $label = $frame->Label(-anchor => 'w', -text => "Command History:");
   $label->pack(-side => 'top', -expand => 'no', -fill => 'both');
   $debug_log = $frame->Scrolled('ROText', -scrollbars => 'se', -height => 8);
   $debug_log->pack(-side => 'top', -expand => 'yes', -fill => 'both');
   $frame->pack(-side => 'top', -fill => 'both', -expand => 'yes');

   $rd_btn->configure(-command => [\&debug_access_read_all_ctrl, $chip_name, $debug_log]);
   $wr_btn->configure(-command => [\&debug_access_write_ctrl, $chip_name, " ", $TRUE, $debug_log]);

   #
   #  Create a frame with some action buttons
   #
   $frame = $debug_win->Frame(-borderwidth => 10);
   $btn   = $frame->Button(-text => 'Clear History', -width => $bwidth,
                -command => [\&main_debug_cbf_hist_btn, \@mdw_data, 'clear'] );
   $btn->pack(-side => 'left', -padx => 5);
   $btn   = $frame->Button(-text => 'Save History', -width => $bwidth,
                -command => [\&main_debug_cbf_hist_btn, \@mdw_data, 'save'] );
   $btn->pack(-side => 'left', -padx => 5);
   $btn   = $frame->Button(-text => 'Close All Debug Windows', -width => ($bwidth+10),
                -command => [ $debug_win => 'destroy'] );
   $btn->pack(-side => 'right', -padx => 5);
   $frame->pack(-side => 'top', -fill => 'both');

   #
   #  Read the current state of the debug port control registers
   #
   if ( $read_flag )
   {
      debug_access_read_all_ctrl($chip_name, $debug_log);
   }
   #
   #  Force current control register value as "new"
   #
   else
   {  foreach $dpd_ref ( @$dpd_list )
      {  $reg_ref  = $$dpd_ref[$DPD_CTRL_REG]; # control register instance
         debug_access_update_ctrl_value($chip_name, $reg_ref, $TRUE);
         $$dpd_ref[$DPD_FLAGS] |= $DPD_FLAG_NEW_CTRL; # force "new" value flag
      }
      debug_access_update_chip($chip_name);
   }

   #
   #  Store the data for the main debug window
   #
   $mdw_data[$MDWD_WINDOW]      = $debug_win;
   $mdw_data[$MDWD_MAIN_WINDOW] = $main_win;
   $mdw_data[$MDWD_CHIP_NAME]   = $chip_name;
   $mdw_data[$MDWD_CHIP_TYPE]   = $chip_type;
   $mdw_data[$MDWD_DPD_LIST]    = $dpd_list;
   $mdw_data[$MDWD_WID_HIST]    = $debug_log;
   $mdw_data[$MDWD_WID_LIST]    = $list;
   $mdw_data[$MDWD_REF_TSWIN]   = 0;

   printf "mdw_data = ( @mdw_data )\n" if ($debug_flag & $DEBUG_DETAILS);

   #
   #  Add the information to the global lookup table
   #
   $mdwd_lookup{$chip_name}    = \@mdw_data;
}


#-------------------------------------------------------------------------------

sub main_debug_cbf_hist_btn
{
   printf "\nmain_debug_cbf_hist_btn( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   return if ( $#_ != 1 );
   my $mdw_ref  = $_[0];  # Reference to window data
   my $mode     = $_[1];  # Mode: 'save' or 'clear'
   if ( $debug_flag & $DEBUG_GENERAL )
   {  printf "MDWD_REF = $mdw_ref = ( @$mdw_ref )\n";
      printf "MODE     = $mode\n";
   }
   my $win_h      = $$mdw_ref[$MDWD_WINDOW];    # Main window for dialog boxes
   my $res_widget = $$mdw_ref[$MDWD_WID_HIST];  # Reference to command results widget

   #---------------------------------------
   #  Clear the command history buffer
   #---------------------------------------
   if ( $mode eq 'clear' )
   {  $res_widget->delete('1.0', 'end');
      return;
   }

   #---------------------------------------
   #  Continue: save the command history buffer
   #---------------------------------------
   my ($out_file, $file_id, $text);

   #---------------------------------------
   #  Get the contents of the command history
   #---------------------------------------
   $text = $res_widget->get('1.0', 'end');

   #---------------------------------------
   #  Get the output file name from user
   #---------------------------------------
   my @types = ( ['History or Log file (Text)', '.txt'],
                 ['All Files',                     , '*'   ]  );
   $out_file = $win_h->getSaveFile(
                  -title => 'Save Debug Port History Data',
                  -filetypes => \@types,
                  -defaultextension => '.txt'
                   );
   return if ( !defined($out_file) );

   #---------------------------------------
   #  Open the output file and write the text
   #---------------------------------------
   $file_id = sprintf("hist%08lX", time());
   if ( !open($file_id, ">$out_file") )
   {  $text = "Unable to open output file for writing.\nFile: $out_file\nDebug port history not saved.";
      $win_h->messageBox(-icon => 'error', -type => 'OK', -message => $text,
         -title => 'SuperScope: open file error');
      return;
   }
   printf $file_id $text;
   close($file_id);

   print "\nSaved debug port history: $out_file\n";
}


#-------------------------------------------------------------------------------

sub main_debug_cbf_create
{
   print "\nmain_debug_cbf_create( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $rsvd_ref  = $_[0];  # Widget reference when invoked as callback
   my $mdw_ref   = $_[1];  # Reference to window data
   my $main_win  = $$mdw_ref[$MDWD_MAIN_WINDOW]; # SuperScope main window
   my $chip_name = $$mdw_ref[$MDWD_CHIP_NAME];   # Full chip name
   my $chip_type = $$mdw_ref[$MDWD_CHIP_NAME];   # Chip type/description
   my $dpd_list  = $$mdw_ref[$MDWD_DPD_LIST];    # List with debug port information for chip
   my $list_wid  = $$mdw_ref[$MDWD_WID_LIST];    # Reference to debug port listbox widget
   my (@cur, $cur, $dpd_ref, $name, $text, $mw_debug);

   #
   #  Get the current selection from the listbox
   #
   @cur = $list_wid->curselection();  # get current selection indices
   return unless @cur;                # return if no line selected

   #
   #  Process each entry in the selection list
   #
   foreach $cur ( @cur )
   {  $dpd_ref = $$dpd_list[$cur]; # Get the definition of the current debug port
      $name    = $chip_name . ":" . $$dpd_ref[$DPD_NAME];
      if ( exists $dwd_lookup{$name} )
      {  $text = "A window has already been created for debug port $$dpd_ref[$DPD_NAME].";
         $$mdw_ref[$MDWD_WINDOW]->messageBox(-icon => 'info', -type => 'OK',
               -message => $text, -title => 'Debug Port Access Window Message');
      }
      else # Create the window for the debug port
      {  create_debug_access_window($main_win, $chip_name, $chip_type, 
             $$dpd_ref[$DPD_NAME], $dpd_ref, $$mdw_ref[$MDWD_WID_HIST]);
      }
   }
}

#-------------------------------------------------------------------------------

sub main_debug_cbf_close
{
   printf "\nmain_debug_cbf_close( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);
   
   my $rsvd_ref  = $_[0];  # Widget reference when invoked as callback
   my $mdw_ref   = $_[1];  # Reference to parameters
   my $top_menu  = $_[2];  # Reference to top-level menu command : to enable
   my $sub_cmd   = $_[3];  # Reference to search command : to clear
   my $win_h     = $$mdw_ref[$MDWD_WINDOW];    # Extract handle to top-level window

   return if ( $rsvd_ref != $win_h );

   if ( $debug_flag & $DEBUG_DETAILS )
   {  printf "mdw_ref = $mdw_ref = ( @$mdw_ref )\n";
   }

   #
   #  Close each open debug port access window(s)
   #
   my $chip_name = $$mdw_ref[$MDWD_CHIP_NAME]; # Full chip name
   my $dpd_list  = $$mdw_ref[$MDWD_DPD_LIST];  # List with debug port information for chip
   my ($i, $dpd_ref, $full_name, $dwd_ref, $dpd_win);
   for ($i = 0; $i <= $#$dpd_list; $i++)
   {  $dpd_ref   = $$dpd_list[$i];
      $full_name = $chip_name . ":" . $$dpd_ref[$DPD_NAME];
      if ( exists($dwd_lookup{$full_name}) )
      {  $dwd_ref = $dwd_lookup{$full_name};
         $dpd_win = $$dwd_ref[$DWD_MAIN_WINDOW];
         $dpd_win->destroy();
      }
   }

   #
   #  Destroy the TP search window window, if open
   #
   if ( $$mdw_ref[$MDWD_REF_TSWIN] != 0 )
   {  $$sub_cmd = 0;
      $$mdw_ref[$MDWD_REF_TSWIN]->destroy();
   }

   $debug_chip_name = "";

   #
   #  Remove entry from "mdwd_lookup{}" for main debug windows
   #
   delete $mdwd_lookup{$chip_name};

   #
   #  Enable the top-level menu for the debug access window(s)
   #
   $top_menu->configure(-state => 'normal');
}

#-------------------------------------------------------------------------------

1;
