# 
#/**************************************************************************
# *                                                                        *
# *               Copyright (C) 1989-1999 Silicon Graphics, Inc.           *
# *                                                                        *
# *  These coded instructions, statements, and computer programs  contain  *
# *  unpublished  proprietary  information of Silicon Graphics, Inc., and  *
# *  are protected by Federal copyright law.  They  may  not be disclosed  *
# *  to  third  parties  or copied or duplicated in any form, in whole or  *
# *  in part, without the prior written consent of Silicon Graphics, Inc.  *
# *                                                                        *
# **************************************************************************/
# 
# @(#)$Id: dimm.pm,v 1.12 2006/11/28 20:20:35 meyer Exp $

#-------------------------------------------------------------------------------
#
#  FILE: dimm.dm
#
#  DESCRIPTION:  This file contains supporting PERL subroutines for the error
#     log decode utility for SHUB.
#
#  MODIFICATION HISTORY:
#
#     WHEN      WHO               WHAT
#     --------  ----------------  -----------------------------------------
#     01/22/03  Nelson Crumbaker  Original source code for submission.
#     08/21/03  Nelson Crumbaker  Moved the fsb/board/memory bit mappings
#                                 outside of this file and added more 
#                                 fields to identify chip location.
#                                 There are now 2 different bit mappings:
#                                 IP47 and IP57 (OPUS).
#     11/20/03  Gary Meyer        Removed "use lib" statement from module file.
#     03/18/04  Nelson Crumbaker  Check that the address not equal to zero 
#                                 before decoding dimm location to avoid 
#                                 decoding bogus errors.
#     07/13/04  Nelson Crumbaker  Added support for dram callout of x4 devices.
#                                 Decoding dimm info for directory errors.
#     12/17/04  Nelson Crumbaker  Added a function to compute FSB bit based
#                                 on Dir. Bit, Dir. ecc group, and L/R Bus.
#     06/08/05  Gary Meyer        Updated for Shub2.
#     02/07/06  Gary Meyer        To recognize IP73 blade type.
#     03/03/06  Gary Meyer        To recognize IP71 blade type.
#     04/14/06  Gary Meyer        Updated to recognize IP75 blade type.
#     10/28/06  Gary Meyer        Updated to recognize IP79 & IP81 blade types.
#
#-------------------------------------------------------------------------------

use ip41_memory_bit_map;
use ip57_memory_bit_map;
use ip63_memory_bit_map;
use ecc;

@x4_DRAM_TO_U_MAP = (
    2,   # Bank 0, D0
   13,   # Bank 0, D1
    3,   # Bank 0, D2
   14,   # Bank 0, D3
    4,   # Bank 0, D4
   15,   # Bank 0, D5
    5,   # Bank 0, D6
   16,   # Bank 0, D7
    8,   # Bank 0, D8
   18,   # Bank 0, D9
   19,   # Bank 0, D10
    9,   # Bank 0, D11
   20,   # Bank 0, D12
   10,   # Bank 0, D13
   21,   # Bank 0, D14
   11,   # Bank 0, D15
    6,   # Bank 0, D16
    7,   # Bank 0, D17
   31,   # Bank 1, D0
   39,   # Bank 1, D1
   30,   # Bank 1, D2
   38,   # Bank 1, D3
   29,   # Bank 1, D4
   37,   # Bank 1, D5
   28,   # Bank 1, D6
   36,   # Bank 1, D7
   25,   # Bank 1, D8
   35,   # Bank 1, D9
   34,   # Bank 1, D10
   24,   # Bank 1, D11
   33,   # Bank 1, D12
   23,   # Bank 1, D13
   32,   # Bank 1, D14
   22,   # Bank 1, D15
   27,   # Bank 1, D16
   26    # Bank 1, D17
);


############################################################################
#                   XBUS  XBUS  XBUS  XBUS  YBUS  YBUS  YBUS  YBUS         #
#    SHUB   LorR    DIMM  DIMM   x4    x8   DIMM  DIMM   x4    x8    FSB   #
#     BIT   DIMM     PIN   BIT  DRAM  DRAM   PIN   BIT  DRAM  DRAM   BIT   #
############################################################################
#   [  24, "LEFT",   33,   24,    3,    6,   33,   24,    3,    6  ], # 0

sub find_dimm
{
   local $address = $_[0];
   local $bit;
   local $XorY;
   local $LorR;
   local $DG;
   local $dimm;
   local $dw;
   local $bank;
   local $chip;
   local $board_type = $_[2]; 
   local $print_detailed_fru = $_[3]; 
   local $x4_chip = $_[4];

   if ( $board_type =~ /IP57/ )
   {
      @DoubleWordData = @IP57_MEMORY_BIT_MAP;
   }
   elsif ($board_type =~ /IP[678][1359]/)		#  IP63, IP71, IP73, IP75, IP79, IP81
   {
      @DoubleWordData = @IP63_MEMORY_BIT_MAP;
   }
   else
   {
      @DoubleWordData = @IP41_MEMORY_BIT_MAP;
   }

   if ( $address != 0 )
   {
      $DG = ($address >> 34) & 0x3;
      $XorY = ($address >> 7) & 0x1;
      $dw = ($address >> 3) & 0x1;
      $bank = ($address >> 12) & 0x1;

      $bit = get_bit_from_syndrome($_[1]);

      if ( $bit != -1 )
      {
         if ($bit < 64)
         {
            if ( $dw == 1 )
            {
               $bit += 64;
            }
            $LorR = fsb_bit_to_LorR($bit,$board_type);
         }
         else
         {
            $bit += 64;
            if ( $dw == 1 )
            {
               $bit += 8;
            }
            $LorR = fsb_bit_to_LorR($bit,$board_type);
         }

         $dimm = "DG" . $DG . "_" . (($XorY) ? "Y" : "X") . 
                 "_BUS_" . (($LorR) ?  "R" : "L" );

         if ( $print_detailed_fru )
         {
            $dimm .= ", PIN:" . $DoubleWordData[$bit][(2+(($XorY)*4))];
            $dimm .= ", BIT:" . $DoubleWordData[$bit][(3+(($XorY)*4))];

            if ( $x4_chip == 1 ) {
               $chip = $DoubleWordData[$bit][(5+(($XorY)*4))];
               hex($chip);
               $chip += ($bank * 18); 
               $chip = $x4_DRAM_TO_U_MAP[$chip];
               $chip = sprintf(", CHIP:U%d", $chip);
               $dimm .= $chip;
            } else {
               $chip = $DoubleWordData[$bit][(4+(($XorY)*4))];
               hex($chip);
               $chip += ($bank * 9); 
               $chip = sprintf(", CHIP:D%d", $chip);
               $dimm .= $chip;
            }

         }
      }
      else
      {
          $dimm = "DG" . $DG . "_" . (($XorY) ? "Y" : "X") .
                  "_BUS_" . "L" . "," .
                  "DG" . $DG . "_" . (($XorY) ? "Y" : "X") .
                  "_BUS_" . "R";
      }
   }
   else
   {
      $dimm = "UNKNOWN (INVALID ADDRESS)";
   }

   return($dimm);

}

sub fsb_bit_to_LorR
{
   my $fsb_bit = $_[0];
   my $board_type = $_[1];
   my $LorR;

   if ( $board_type =~ /IP57/ )
   {
      @DoubleWordData = @IP57_MEMORY_BIT_MAP;
   }
   elsif ($board_type =~ /IP[678][1359]/)		#  IP63, IP71, IP73, IP75, IP79, IP81
   {
      @DoubleWordData = @IP63_MEMORY_BIT_MAP;
   }
   else
   {
      @DoubleWordData = @IP41_MEMORY_BIT_MAP;
   }
   
   $LorR = ( $DoubleWordData[$fsb_bit][1] );
   
   if ( $LorR =~ /LEFT/ )
   {
      return(0);
   }  
   else
   {
      return(1);
   }  
}  

# This subroutine is for DIRECTORY correctable errors.
sub getDimmInfoFromMDBit
{
   my $mdbit      = $_[0];
   my $address    = $_[1];
   my $board_type = $_[2]; 
   my $x4_chip    = $_[3];
   my $LorR; 
   my $XorY; 
   my $bank;
   my $chip;
   my $dimmInfo;

   $LorR = ( $address >> 11 ) & 0x1;
   $XorY = ( $address >> 7 ) & 0x1;
   $bank = ($address >> 12) & 0x1;

   if ( $LorR == 0 )
   {
      $LorR = "LEFT";
   }  
   else
   {
      $LorR = "RIGHT";
   }  

   if ( $board_type =~ /IP57/ )
   {
      @DoubleWordData = @IP57_MEMORY_BIT_MAP;
   }
   elsif ($board_type =~ /IP[678][1359]/)		#  IP63, IP71, IP73, IP75, IP79, IP81
   {
      @DoubleWordData = @IP63_MEMORY_BIT_MAP;
   }
   else
   {
      @DoubleWordData = @IP41_MEMORY_BIT_MAP;
   }

   for($fsb_bit=0; $fsb_bit<144; $fsb_bit++)
   {
      if ( ( $DoubleWordData[$fsb_bit][0] == $mdbit ) &&
           ( $DoubleWordData[$fsb_bit][1] =~ $LorR ) )
      {
         $dimmInfo .= ", PIN:"   . $DoubleWordData[$fsb_bit][(2+(($XorY)*4))];
         $dimmInfo .= ", BIT:" . $DoubleWordData[$fsb_bit][(3+(($XorY)*4))];

         if ( $x4_chip == 1 ) {
            $chip = $DoubleWordData[$fsb_bit][(5+(($XorY)*4))];
            hex($chip);
            $chip += ($bank * 18);
            $chip = $x4_DRAM_TO_U_MAP[$chip];
            $chip = sprintf(", CHIP:U%d", $chip);
            $dimmInfo .= $chip;
         } else {
            $chip = $DoubleWordData[$fsb_bit][(4+(($XorY)*4))];
            hex($chip);
            $chip += ($bank * 9);
            $chip = sprintf(", CHIP:D%d", $chip);
            $dimmInfo .= $chip;
         }
      }
   }
   return($dimmInfo);
}

sub fsb_bit_from_dir_bit
{
   my $bit     = $_[0]; 
   my $eccgrp  = $_[1]; 
   my $LorR    = $_[2]; 
   my $mdbit;

   if ( $LorR == 0 )
   {
      $LorR = "LEFT";
   }  
   else
   {
      $LorR = "RIGHT";
   }  
   
   if ( $board_type =~ /IP57/ )
   {
      @DoubleWordData = @IP57_MEMORY_BIT_MAP;
   }  
   elsif ($board_type =~ /IP[678][1359]/)		#  IP63, IP71, IP73, IP75, IP79, IP81
   {
      @DoubleWordData = @IP63_MEMORY_BIT_MAP;
   }
   else
   {
      @DoubleWordData = @IP41_MEMORY_BIT_MAP;
   }  

   $mdbit = (($bit>>2) * 8) + ($bit%4);

   if ( $eccgrp == 0 ) 
   {
      $mdbit += 4;
   }

   for($fsb_bit=0; $fsb_bit<144; $fsb_bit++)
   {
      if ( ( $DoubleWordData[$fsb_bit][0] == $mdbit ) &&
           ( $DoubleWordData[$fsb_bit][1] =~ $LorR ) )
      {
         return($fsb_bit);
      }
   }

   return(-1);
}

1;
