#
#/**************************************************************************
# *                                                                        *
# *               Copyright (C) 1989-1999 Silicon Graphics, Inc.           *
# *                                                                        *
# *  These coded instructions, statements, and computer programs  contain  *
# *  unpublished  proprietary  information of Silicon Graphics, Inc., and  *
# *  are protected by Federal copyright law.  They  may  not be disclosed  *
# *  to  third  parties  or copied or duplicated in any form, in whole or  *
# *  in part, without the prior written consent of Silicon Graphics, Inc.  *
# *                                                                        *
# **************************************************************************/
#
# @(#)$Id: dvah_subs.pm,v 1.2 2002/03/06 16:25:23 beighley Exp $

#-------------------------------------------------------------------------------
#
#  FILE: dvah_subs.pm
#
#  DESCRIPTION:  This file contains PERL subroutines to handle arrays and
#     hash structures representing "memory" addresses and their corresponding
#     data values.  If the addresses and values are contained in two
#     seperate arrays, then they are called "data values arrays" and the
#     subroutines start with "dva_*".  If a hash structure is used, then
#     the address is used as the key, and the subroutines start with "dvh_*".
#
#     When an entire hash table is processed, the entries are sorted by
#     the address (key).  This order may be different from that found
#     when moving through the hash directly.
#
#  NOTES: The "common_defs.pm" file is required when this file is used.
#
#  SUBROUTINES:  The following subroutines are contained in this file:
#     (expected parameter usage displayed; some have optional parameters)
#     @dif   = dva_cmp(\@act, \@exp, \@msk)
#     @dif   = dva_xor(\@act, \@exp)
#     @data  = dva_and(\@act, \@msk)
#     @data  = dva_or(\@act, \@msk)
#
#     $num   = dva_file_read($filename, \@addr, \@data);
#     dva_file_write_block($filename, $addr, \@data);
#     dva_file_write_group($filename, \@addr, \@data);
#     %dhash = dvh_file_read($filename)
#     dvh_file_write($filename, \%dhash);
#
#     dva_display_block($title, $addr, \@data);
#     dva_display_block($title, $addr, \@data, $DATA_FMT_*);
#     dva_display_block($title, $addr, \@data, $DATA_FMT_*, \@labels);
#     dva_display_group($title, \@addr, \@data);
#     dva_display_group($title, \@addr, \@data, $DATA_FMT_*);
#     dva_display_group($title, \@addr, \@data, $DATA_FMT_*, \@labels);
#     dvh_display($title, \%dhash);
#     dvh_display($title, \%dhash, $DATA_FMT_*);
#     dvh_display($title, \%dhash, $DATA_FMT_*, \@labels);
#
#     @addrs = dvh_get_addrs(\%hash)
#     @data  = dvh_get_data(\%hash)
#     %dhash = dva_build_hash(\@addrs, \@data);
#     $str   = data_value_convert($val, $DATA_FMT_*)
#
#  MODIFICATION HISTORY:
#
#     WHEN      WHO             WHAT                                     
#     --------  --------------  -----------------------------------------
#     10/25/01  Karen Beighley  Original source code.
#     03/06/02  Karen Beighley  Changed "print" statements to display 64-bits.
# 
#-------------------------------------------------------------------------------


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: dva_cmp()
#
#    USAGE: @dif = dva_cmp(\@act, \@exp, \@msk)
#
#    DESCRIPTION: The subroutine compares the elements of two arrays and 
#       applies a mask: (act[i] ^ exp[i]) & msk[i]
#
#    RETURN VALUE: Array containing results of compare.
#
#-------------------------------------------------------------------------------

sub dva_cmp
{  
   my $num = $#_ + 1;
   print "\ndva_cmp(@_) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 3 )
   {  $emess = "Invalid parameters passed to subroutine dva_cmp().";
      dsp_error_and_exit($emess);
      return;
   }
   my($aref, $eref, $mref) = @_;
   if ( (ref($aref) ne 'ARRAY') || (ref($eref) ne 'ARRAY') ||
        (ref($mref) ne 'ARRAY') )
   {  $emess = "Invalid parameter passed to subroutine dva_cmp().";
      $emess .= "\nAll three parameters must be ARRAY references.";
      dsp_error_and_exit($emess);
   }
   my @act = @$aref;
   my @exp = @$eref;
   my @msk = @$mref;
   if ( ($#act != $#exp) || ($#act != $#msk) )
   {  $emess = "Invalid parameters passed to subroutine dva_cmp().";
      $emess .= "\nThe number of elements in each array must be equal.";
      dsp_error_and_exit($emess);
   }

   if ( $debug_flag & $DEBUG_GENERAL )
   {  print "actual data = ( @act )\n";
      print "expect data = ( @exp )\n";
      print "mask   data = ( @msk )\n";
   }

   my(@dif);
   $#dif = $#act;
   for ($i = 0; $i <= $#act; $i++)
   {  $dif[$i] = ($act[$i] ^ $exp[$i]) & $msk[$i];
   }

   return( @dif );
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: dva_xor()
#
#    USAGE: @dif = dva_xor(\@act, \@exp)
#
#    DESCRIPTION: The subroutine performs an exclusive-or (XOR) on the 
#       elements of two arrays: (act[i] ^ exp[i])
#
#    RETURN VALUE: Array containing results of XOR.
#
#-------------------------------------------------------------------------------

sub dva_xor
{  
   my $num = $#_ + 1;
   print "\ndva_xor(@_) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 2 )
   {  $emess = "Invalid parameters passed to subroutine dva_xor().";
      dsp_error_and_exit($emess);
      return;
   }
   my($aref, $eref) = @_;
   if ( (ref($aref) ne 'ARRAY') || (ref($eref) ne 'ARRAY') )
   {  $emess = "Invalid parameter passed to subroutine dva_xor().";
      $emess .= "\nBoth parameters must be ARRAY references.";
      dsp_error_and_exit($emess);
   }
   my @act = @$aref;
   my @exp = @$eref;
   if ( $#act != $#exp )
   {  $emess = "Invalid parameters passed to subroutine dva_xor().";
      $emess .= "\nThe number of elements in each array must be equal.";
      dsp_error_and_exit($emess);
   }

   if ( $debug_flag & $DEBUG_GENERAL )
   {  print "actual data = ( @act )\n";
      print "expect data = ( @exp )\n";
   }

   my(@dif);
   $#dif = $#act;
   for ($i = 0; $i <= $#act; $i++)
   {  $dif[$i] = $act[$i] ^ $exp[$i];
   }

   return( @dif );
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: dva_and()
#
#    USAGE: @data = dva_and(\@act, \@msk)
#
#    DESCRIPTION: The subroutine performs an bitwise-and (AND) on the 
#       elements of two arrays: (act[i] & msk[i])
#
#    RETURN VALUE: Array containing results of AND.
#
#-------------------------------------------------------------------------------

sub dva_and
{  
   my $num = $#_ + 1;
   print "\ndva_and(@_) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 2 )
   {  $emess = "Invalid parameters passed to subroutine dva_and().";
      dsp_error_and_exit($emess);
      return;
   }
   my($aref, $mref) = @_;
   if ( (ref($aref) ne 'ARRAY') || (ref($mref) ne 'ARRAY') )
   {  $emess = "Invalid parameter passed to subroutine dva_and().";
      $emess .= "\nBoth parameters must be ARRAY references.";
      dsp_error_and_exit($emess);
   }
   my @act = @$aref;
   my @msk = @$mref;
   if ( $#act != $#msk )
   {  $emess = "Invalid parameters passed to subroutine dva_and().";
      $emess .= "\nThe number of elements in each array must be equal.";
      dsp_error_and_exit($emess);
   }

   if ( $debug_flag & $DEBUG_GENERAL )
   {  print "actual data = ( @act )\n";
      print "mask   data = ( @msk )\n";
   }

   my(@data);
   $#data = $#act;
   for ($i = 0; $i <= $#act; $i++)
   {  $data[$i] = $act[$i] & $msk[$i];
   }
   return( @data );
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: dva_or()
#
#    USAGE: @data = dva_or(\@act, \@msk)
#
#    DESCRIPTION: The subroutine performs an bitwise-or (OR) on the 
#       elements of two arrays: (act[i] | msk[i])
#
#    RETURN VALUE: Array containing results of OR.
#
#-------------------------------------------------------------------------------

sub dva_or
{  
   my $num = $#_ + 1;
   print "\ndva_or(@_) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 2 )
   {  $emess = "Invalid parameters passed to subroutine dva_or().";
      dsp_error_and_exit($emess);
      return;
   }
   my($aref, $mref) = @_;
   if ( (ref($aref) ne 'ARRAY') || (ref($mref) ne 'ARRAY') )
   {  $emess = "Invalid parameter passed to subroutine dva_or().";
      $emess .= "\nBoth parameters must be ARRAY references.";
      dsp_error_and_exit($emess);
   }
   my @act = @$aref;
   my @msk = @$mref;
   if ( $#act != $#msk )
   {  $emess = "Invalid parameters passed to subroutine dva_or().";
      $emess .= "\nThe number of elements in each array must be equal.";
      dsp_error_and_exit($emess);
   }

   if ( $debug_flag & $DEBUG_GENERAL )
   {  print "actual data = ( @act )\n";
      print "mask   data = ( @msk )\n";
   }

   my(@data);
   $#data = $#act;
   for ($i = 0; $i <= $#act; $i++)
   {  $data[$i] = $act[$i] | $msk[$i];
   }
   return( @data );
}



#-------------------------------------------------------------------------------
#
#    SUBROUTINE: dvh_get_addrs()
#
#    USAGE: @addrs = dvh_get_addrs(\%hash)
#
#    DESCRIPTION: The subroutine extracts the address values (keys) from
#       the hash structure.  The values in the array are sorted based 
#       on the address value.
#
#    RETURN VALUE: Array of sorted address values.
#
#-------------------------------------------------------------------------------

sub dvh_get_addrs
{
   my $num = $#_ + 1;
   print "\ndvh_get_addrs(@_) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 1 )
   {  $emess = "Invalid parameters passed to subroutine dvh_get_addrs().";
      dsp_error_and_exit($emess);
      return;
   }
   my $hash_ref = $_[0];
   if ( ref($hash_ref) ne 'HASH' )
   {  $emess = "Invalid parameter passed to subroutine dvh_get_addrs().";
      $emess .= "\nThe parameter must be a HASH reference.";
      dsp_error_and_exit($emess);
   }
   my %hash = %$hash_ref;

   my @addrs = keys %hash;
   @addrs    = sort { $a <=> $b } @addrs;
   return(@addrs);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: dvh_get_data()
#
#    USAGE: @data = dvh_get_data(\%hash)
#
#    DESCRIPTION: The subroutine extracts the data values (values) from
#       the hash structure.  The values in the array are sorted based 
#       on the address value.
#
#    RETURN VALUE: Array of data values (in order based on address value).
#
#-------------------------------------------------------------------------------

sub dvh_get_data
{
   my $num = $#_ + 1;
   print "\ndvh_get_data(@_) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 1 )
   {  $emess = "Invalid parameters passed to subroutine dvh_get_data().";
      dsp_error_and_exit($emess);
      return;
   }
   my $hash_ref = $_[0];
   if ( ref($hash_ref) ne 'HASH' )
   {  $emess = "Invalid parameter passed to subroutine dvh_get_data().";
      $emess .= "\nThe parameter must be a HASH reference.";
      dsp_error_and_exit($emess);
   }
   my %hash   = %$hash_ref;
   my @addrs  = keys %hash;
   @addrs     = sort { $a <=> $b } @addrs;
   my @values = ();
   my $ind    = 0;
   foreach $a (@addrs)
   {  $values[$ind] = $hash{$a};
      $ind++;
   }
   return(@values);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: dva_build_hash()
#
#    USAGE: %dhash = dva_build_hash(\@addrs, \@data);
#
#    DESCRIPTION: The subroutine will build a hash structure using the
#       elements of the "addrs" array as the key and the corresponding
#       elements of the "data" array as the value.
#
#    RETURN VALUE: Hash structure created from array values.
#
#-------------------------------------------------------------------------------

sub dva_build_hash
{
   my $num = $#_ + 1;
   print "\ndva_build_hash(@_) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 2 )
   {  $emess = "Invalid number of parameters for subroutine dva_build_hash().";
      dsp_error_and_exit($emess);
   }

   my $addr_ptr = $_[0];
   if ( ref($addr_ptr) ne 'ARRAY' )
   {  $emess = "Invalid parameter passed to subroutine dva_build_hash().";
      $emess .= "\nFirst parameter must be an ARRAY reference.";
      dsp_error_and_exit($emess);
   }
   my $data_ptr = $_[1];
   if ( ref($data_ptr) ne 'ARRAY' )
   {  $emess = "Invalid parameter passed to subroutine dva_build_hash().";
      $emess .= "\nSecond parameter must be an ARRAY reference.";
      dsp_error_and_exit($emess);
   }
   my @addr = @$addr_ptr;
   my @data = @$data_ptr;
   if ( $#addr != $#data )
   {  $emess = "Invalid parameters passed to subroutine dva_build_hash().";
      $emess .= "\nThe number of elements in each array must be equal.";
      dsp_error_and_exit($emess);
   }

   %newhash = ();
   my $ind = 0;
   foreach $aval (@addr)
   {  $dval = $data[$ind];
printf "addr=%016llX=%-18ld  data=0x%016llX=%ld\n", $aval, $aval, $dval, $dval;
      $newhash{$aval} = $dval;
      $ind++;
   }
    
   return(%newhash);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: dva_file_read()
#
#    USAGE: $num = dva_file_read($filename, \@addr, \@data);
#
#    DESCRIPTION: The subroutine reads an input file and stores the
#       contents into "data value array" structures.
#       The format of each line in the input file must be:
#       <address as 64-bit hexidecimal> = <value as 64-bit hexidecimal>
#
#    RETURN VALUE: number of entries read from file
#
#-------------------------------------------------------------------------------

sub dva_file_read
{
   my $num = $#_ + 1;
   print "\ndva_file_read(@_) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 3 )
   {  $emess = "Invalid parameters passed to subroutine dva_file_read().";
      dsp_error_and_exit($emess);
      return;
   }
   my $filename = $_[0];
   my $addr_ref = $_[1];
   my $data_ref = $_[2];
   if ( (ref($addr_ref) ne 'ARRAY') || (ref($data_ref) ne 'ARRAY') )
   {  $emess = "Invalid parameter passed to subroutine dva_file_read().";
      $emess .= "\nThe second and third parameters must be ARRAY references.";
      dsp_error_and_exit($emess);
   }
   @$addr_ref = ();
   @$data_ref = ();

   #
   #  Open the input file for reading
   #
   if ( !open(fileId, "$filename") )
   {  $emess = "Unable to open input file for reading.";
      $emess .= "\nFile: $filename";
      dsp_error_and_exit($emess);
   }

   #
   #  Process each line from the input file; ignore comment lines
   #
   my @elems = ();
   my $i = 0;
   while (<fileId>)
   {  
#     printf "%03u read: $_\n", $i;
      if ( !/^#/ )
      {  @elems = split();
         if ( $#elems != 2 )
         {  $emess = "Invalid data format encountered in input file.";
            $emess .= "\nFile: $filename";
            dsp_error_and_exit($emess);
         }
         if ( $elems[1] ne "=" )
         {  $emess = "Invalid data format encountered in input file.";
            $emess .= "\nFile: $filename";
            dsp_error_and_exit($emess);
         }
         @$addr_ref[$i] = hex($elems[0]);
         @$data_ref[$i] = hex($elems[2]);
         $i++;
      }
   }
   close(fileId);
   return( $i );
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: dva_file_write_block()
#
#    USAGE: dva_file_write_block($filename, $addr, \@data);
#
#    DESCRIPTION: The subroutine writes the contents of "data value
#       array" structures to a specified file (in hexidecimal).
#       The format of each line in the output file:
#       <address as 64-bit hexidecimal> = <value as 64-bit hexidecimal>
#       The addresses are based on the "addr" which is incremented by
#       the value defined in "$MEM_ADDR_INCR".
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub dva_file_write_block
{
   my $num = $#_ + 1;
   print "\ndva_file_write_block(@_) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 3 )
   {  $emess = "Invalid parameters passed to subroutine dva_file_write_block().";
      dsp_error_and_exit($emess);
      return;
   }
   my $filename = $_[0];
   my $addr     = $_[1];
   my $data_ref = $_[2];
   if ( ref($data_ref) ne 'ARRAY' )
   {  $emess = "Invalid parameter passed to subroutine dva_file_write_block().";
      $emess .= "\nLast parameter must be an ARRAY reference.";
      dsp_error_and_exit($emess);
   }
   my @data = @$data_ref;

   #
   #  Open the output file for writing
   #
   if ( !open(fileId, ">$filename") )
   {  $emess = "Unable to open output file for writing.";
      $emess .= "\nFile: $filename";
      dsp_error_and_exit($emess);
   }
   printf fileId "#\n";
   printf fileId "#  Memory data file: map of values for data addresses\n";
   printf fileId "#\n";
   printf fileId "# Format of each line in file:\n";
   printf fileId "# <address value>  = <data value>\n";

   #
   #  Write each address/value pair
   #
   foreach $val (@data)
   {  printf fileId "0x%016llX = 0x%016llX\n", $addr, $val;
      $addr += $MEM_ADDR_INCR;
   }
   close(fileId);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: dva_file_write_group()
#
#    USAGE: dva_file_write_group($filename, \@addr, \@data);
#
#    DESCRIPTION: The subroutine writes the contents of "data value
#       array" structures to a specified file (in hexidecimal).
#       The format of each line in the output file:
#       <address as 64-bit hexidecimal> = <value as 64-bit hexidecimal>
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub dva_file_write_group
{
   my $num = $#_ + 1;
   print "\ndva_file_write_group(@_) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 3 )
   {  $emess = "Invalid parameters passed to subroutine dva_file_write_group().";
      dsp_error_and_exit($emess);
      return;
   }
   my $filename = $_[0];
   my $addr_ref = $_[1];
   my $data_ref = $_[2];
   if ( (ref($addr_ref) ne 'ARRAY') || (ref($data_ref) ne 'ARRAY') )
   {  $emess = "Invalid parameter passed to subroutine dva_file_write_group().";
      $emess .= "\nThe second and third parameters must be ARRAY references.";
      dsp_error_and_exit($emess);
   }
   my @addr = @$addr_ref;
   my @data = @$data_ref;
   if ( $#data != $#addr )
   {  $emess = "Invalid parameters passed to subroutine dva_file_write_group().";
      $emess .= "\nThe number of elements in each array must be equal.";
      dsp_error_and_exit($emess);
   }

   #
   #  Open the output file for writing
   #
   if ( !open(fileId, ">$filename") )
   {  $emess = "Unable to open output file for writing.";
      $emess .= "\nFile: $filename";
      dsp_error_and_exit($emess);
   }
   printf fileId "#\n";
   printf fileId "#  Memory data file: map of values for data addresses\n";
   printf fileId "#\n";
   printf fileId "# Format of each line in file:\n";
   printf fileId "# <address value>  = <data value>\n";

   #
   #  Write each address/value pair
   #
   for ($i = 0; $i <= $#addr; $i++)
   {  printf fileId "0x%016llX = 0x%016llX\n", $addr[$i], $data[$i];
   }
   close(fileId);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: dvh_file_read()
#
#    USAGE: %dhash = dvh_file_read($filename)
#
#    DESCRIPTION: The subroutine reads an input file and stores the
#       contents into "data value hash" structure.
#       The format of each line in the input file must be:
#       <address as 64-bit hexidecimal> = <value as 64-bit hexidecimal>
#
#    RETURN VALUE: The hash structure containing addresses and data.
#
#-------------------------------------------------------------------------------

sub dvh_file_read
{
   my $num = $#_ + 1;
   print "\ndvh_file_read(@_) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 1 )
   {  $emess = "Invalid parameters passed to subroutine dvh_file_read().";
      dsp_error_and_exit($emess);
      return;
   }
   my $filename = $_[0];
   my %dhash    = ();

   #
   #  Open the input file for reading
   #
   if ( !open(fileId, "$filename") )
   {  $emess = "Unable to open input file for reading.";
      $emess .= "\nFile: $filename";
      dsp_error_and_exit($emess);
   }

   #
   #  Process each line from the input file; ignore comment lines
   #
   my @elems = ();
   my $i    = 0;
   my $addr = 0;
   while (<fileId>)
   {  
#     printf "%03u read: $_\n", $i;
      if ( !/^#/ )
      {  @elems = split();
         if ( $#elems != 2 )
         {  $emess = "Invalid data format encountered in input file.";
            $emess .= "\nFile: $filename";
            dsp_error_and_exit($emess);
         }
         if ( $elems[1] ne "=" )
         {  $emess = "Invalid data format encountered in input file.";
            $emess .= "\nFile: $filename";
            dsp_error_and_exit($emess);
         }
         $addr = hex($elems[0]);
         $dhash{ $addr } = hex($elems[2]);
         $i++;
      }
   }
   close(fileId);
   return(%dhash);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: dvh_file_write()
#
#    USAGE: dvh_file_write($filename, \%dhash);
#
#    DESCRIPTION: The subroutine writes the contents of a "data value
#       hash" structure to the specified output file.
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub dvh_file_write
{  
   my $num = $#_ + 1;
   print "\ndvh_file_write(@_) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 2 )
   {  $emess = "Invalid parameters passed to subroutine dvh_file_write().";
      dsp_error_and_exit($emess);
      return;
   }
   my $filename = $_[0];
   my $data_ref = $_[1];
   if ( ref($data_ref) ne 'HASH' )
   {  $emess = "Invalid parameter passed to subroutine dvh_file_write().";
      $emess .= "\nThe second parameter must be a HASH reference.";
      dsp_error_and_exit($emess);
   }
   my %data = %$data_ref;

   #
   #  Open the output file for writing
   #
   if ( !open(fileId, ">$filename") )
   {  $emess = "Unable to open output file for writing.";
      $emess .= "\nFile: $filename";
      dsp_error_and_exit($emess);
   }
   printf fileId "#\n";
   printf fileId "#  Memory data file: map of values for data addresses\n";
   printf fileId "#\n";
   printf fileId "# Format of each line in file:\n";
   printf fileId "# <address value>  = <data value>\n";

   #
   #  Sort the addresses and write each address/value pair
   #
   my @addrs  = keys %data;
   @addrs     = sort { $a <=> $b } @addrs;
   my $addr   = 0;
   foreach $addr (@addrs)
   {  printf fileId "0x%016llX = 0x%016llX\n", $addr, $data{$addr};
   }
   close(fileId);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: dva_display_block()
#
#    USAGE: dva_display_block($title, $addr, \@data);
#           dva_display_block($title, $addr, \@data, $DATA_FMT_*);
#           dva_display_block($title, $addr, \@data, $DATA_FMT_*, \@labels);
#
#    DESCRIPTION: The subroutine displays the contents of "data value
#       array" structures in the specified format (decimal, octal, hexidecimal).
#       The "title" parameter is displayed if a non-NULL string.  If
#       all parameters are valid, then the format of each line:
#       <address as 64-bit hexidecimal> = <value> "<label>"
#       The addresses are based on the "addr" which is incremented by
#       the value defined in "$MEM_ADDR_INCR".
#       The "labels" parameter is optional, and is displayed if provided.
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub dva_display_block
{  
   my $num = $#_ + 1;
   print "\ndva_display_block(@_) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   my $fmt       = $DATA_FMT_HEX64;
   my $addr_flag = $FALSE;
   my $lab_flag  = $FALSE;
   my $lab_ref   = 0;
   my @names     = ();

   if ( $num == 4 )
   {  $fmt = $_[3];
   }
   elsif ( $num == 5 )
   {  $fmt = $_[3];
      $lab_flag = $TRUE;
   }
   elsif ( $num != 3 )
   {  $emess = "Invalid parameters passed to subroutine dva_display_block().";
      dsp_error_and_exit($emess);
      return;
   }
   my $data_ref  = $_[2];
   if ( ref($data_ref) ne 'ARRAY' )
   {  $emess = "Invalid parameter passed to subroutine dva_display_block().";
      $emess .= "\nThe last parameter must be ARRAY references.";
      dsp_error_and_exit($emess);
   }
   my @data      = @$data_ref;
   my $addr      = $_[1];
   if ( $lab_flag )
   {  $lab_ref = $_[4];
      if ( ref($lab_ref) ne 'ARRAY' )
      {  $emess = "Invalid parameter passed to subroutine dva_display_block().";
         $emess .= "\nThe label parameter must be ARRAY references.";
         dsp_error_and_exit($emess);
      }
      @names  = @$lab_ref;
      if ( $#data != $#names )
      {  $emess = "Invalid parameters passed to subroutine dva_display_block().";
         $emess .= "\nThe number of elements in each array must be equal.";
         dsp_error_and_exit($emess);
      }
   }

   print "$_[0]\n"  if ( $_[0] ne "" );
   my $str = "";
   my $ind = 0;

# How much slower check data format every element instead once and having
# a format-specific loop?  Something to observe...

   if ( $lab_flag ) # display: address value label
   {  $ind = 0;
      foreach $val (@data)
      {  $str = data_value_convert($val, $fmt);
         printf "%016llX = %s   \"%s\"\n", $addr, $str, $names[$ind];
         $ind++;
         $addr += $MEM_ADDR_INCR;
      }
   }
   else # display: address value
   {  foreach $val (@data)
      {  $str = data_value_convert($val, $fmt);
         printf "%016llX = %s\n", $addr, $str;
         $addr += $MEM_ADDR_INCR;
      }
   }

}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: dva_display_group()
#
#    USAGE: dva_display_group($title, \@addr, \@data);
#           dva_display_group($title, \@addr, \@data, $DATA_FMT_*);
#           dva_display_group($title, \@addr, \@data, $DATA_FMT_*, \@labels);
#
#    DESCRIPTION: The subroutine displays the contents of "data value
#       array" structures in the specified format (decimal, octal, hexidecimal).
#       The "title" parameter is displayed if a non-NULL string.  If
#       all parameters are valid, then the format of each line:
#       <address as 64-bit hexidecimal> = <value> "<label>"
#       The "address" field is not displayed if the parameter is "0".
#       The "labels" parameter is optional, and is displayed if provided.
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub dva_display_group
{  
   my $num = $#_ + 1;
   print "\ndva_display_group(@_) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   my $fmt       = $DATA_FMT_HEX64;
   my $addr_flag = $FALSE;
   my $lab_flag  = $FALSE;
   my $lab_ref   = 0;
   my @names     = ();
   my @addrs     = ();

   if ( $num == 4 )
   {  $fmt = $_[3];
   }
   elsif ( $num == 5 )
   {  $fmt = $_[3];
      $lab_flag = $TRUE;
   }
   elsif ( $num != 3 )
   {  $emess = "Invalid parameters passed to subroutine dva_display_group().";
      dsp_error_and_exit($emess);
      return;
   }
   my $data_ref = $_[2];
   if ( ref($data_ref) ne 'ARRAY' )
   {  $emess = "Invalid parameter passed to subroutine dva_display_group().";
      $emess .= "\nThe third parameter must be ARRAY references.";
      dsp_error_and_exit($emess);
   }
   my @data     = @$data_ref;
   my $addr_ref = $_[1];
   if ( $addr_ref != 0 )
   {  if ( ref($addr_ref) ne 'ARRAY' )
      {  $emess = "Invalid parameter passed to subroutine dva_display_group().";
         $emess .= "\nThe second parameter must be ARRAY references.";
         dsp_error_and_exit($emess);
      }
      @addrs = @$addr_ref;
      $addr_flag = $TRUE;
      if ( $#data != $#addrs )
      {  $emess = "Invalid parameters passed to subroutine dva_display_group().";
         $emess .= "\nThe number of elements in each array must be equal.";
         dsp_error_and_exit($emess);
      }
   }
   if ( $lab_flag )
   {  $lab_ref = $_[4];
      if ( ref($lab_ref) ne 'ARRAY' )
      {  $emess = "Invalid parameter passed to subroutine dva_display_group().";
         $emess .= "\nThe label parameter must be ARRAY references.";
         dsp_error_and_exit($emess);
      }
      @names  = @$lab_ref;
      if ( $#data != $#names )
      {  $emess = "Invalid parameters passed to subroutine dva_display_group().";
         $emess .= "\nThe number of elements in each array must be equal.";
         dsp_error_and_exit($emess);
      }
   }

   print "$_[0]\n"  if ( $_[0] ne "" );
   my $str = "";
   my $ind = 0;

# How much slower check data format every element instead once and having
# a format-specific loop?  Something to observe...

   if ( $addr_flag && $lab_flag ) # display: address value label
   {  $ind = 0;
      foreach $val (@data)
      {  $str = data_value_convert($val, $fmt);
         printf "%016llX = %s   \"%s\"\n", $addrs[$ind], $str, $names[$ind];
         $ind++;
      }
   }
   elsif ( $addr_flag ) # display: address value
   {  $ind = 0;
      foreach $val (@data)
      {  $str = data_value_convert($val, $fmt);
         printf "%016llX = %s\n", $addrs[$ind], $str;
         $ind++;
      }
   }
   elsif ( $lab_flag ) # display: value label
   {  $ind = 0;
      foreach $val (@data)
      {  $str = data_value_convert($val, $fmt);
         printf "%s   \"%s\"\n", $str, $names[$ind];
         $ind++;
      }
   }
   else # display: value
   {
      foreach $val (@data)
      {  $str = data_value_convert($val, $fmt);
         printf "$str\n";
      }
   }
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: dvh_display()
#
#    USAGE: dvh_display($title, \%dhash);
#           dvh_display($title, \%dhash, $DATA_FMT_*);
#           dvh_display($title, \%dhash, $DATA_FMT_*, \@labels);
#
#    DESCRIPTION: The subroutine displays the contents of a "data value
#       hash" structure in the specified format (decimal, octal, hexidecimal).
#       The "title" parameter is displayed if a non-NULL string.  If the
#       "labels" parameter is not provided, then the format of each line:
#       <address as 64-bit hexidecimal> = <value>
#       If the "labels" parameter is provided, then the format of each line:
#       <address as 64-bit hexidecimal> = <value> "<label>"
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub dvh_display
{  
   my $num = $#_ + 1;
   print "\ndvh_display(@_) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   my $fmt       = $DATA_FMT_HEX64;
   my $lab_flag  = $FALSE;
   my $lab_ref   = 0;

   # check to make sure parameters are the correct type (reference)
   # confirm that the number of elements in each array is the same

   if ( $num == 3 )
   {  $fmt = $_[2];
   }
   elsif ( $num == 4 )
   {  $fmt = $_[2];
      $lab_flag = $TRUE;
   }
   elsif ( $num != 2 )
   {  $emess = "Invalid parameters passed to subroutine dvh_display().";
      dsp_error_and_exit($emess);
      return;
   }
   my $data_ref = $_[1];
   if ( ref($data_ref) ne 'HASH' )
   {  $emess = "Invalid parameter passed to subroutine dvh_display().";
      $emess .= "\nThe second parameter must be HASH references.";
      dsp_error_and_exit($emess);
   }
   my %data     = %$data_ref;
   if ( $lab_flag )
   {  $lab_ref = $_[3];
      if ( ref($lab_ref) ne 'ARRAY' )
      {  $emess = "Invalid parameter passed to subroutine dvh_display().";
         $emess .= "\nThe label parameter must be ARRAY references.";
         dsp_error_and_exit($emess);
      }
      @names  = @$lab_ref;
# How many entries in hash table?  Need to compare!
#      if ( $#data != $#names )
#      {  $emess = "Invalid parameters passed to subroutine dvh_display().";
#         $emess .= "\nThe number of elements in each array must be equal.";
#         dsp_error_and_exit($emess);
#      }
   }

   print "$_[0]\n"  if ( $_[0] ne "" );
   my $str   = "";
   my $val   = 0;
   my $ind   = 0;
   my @addrs  = keys %data;
   @addrs     = sort { $a <=> $b } @addrs;
   my $addr  = 0;

# How much slower check data format every element instead once and having
# a format-specific loop?  Something to observe...

   if ( $lab_flag ) # display: addr value label
   {  $ind = 0;
#      while (($addr, $val) = each %data)
      foreach $addr (@addrs)
      {  $val = $data{$addr};
         $str = data_value_convert($val, $fmt);
         printf "%016llX = %s   \"%s\"\n", $addr, $str, $names[$ind];
         $ind++;
      }
   }
   else # display: addr value 
   {  $ind = 0;
#      while (($addr, $val) = each %data)
      foreach $addr (@addrs)
      {  $val = $data{$addr};
         $str = data_value_convert($val, $fmt);
         printf "%016llX = %s\n", $addr, $str;
         $ind++;
      }
   }
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: data_value_convert()
#
#    USAGE: $str = data_value_convert($val, $DATA_FMT_*)
#
#    DESCRIPTION: The subroutine converts a value to a string of a specific
#       format for display (decimal, octal, hexidecimal).
#
#    RETURN VALUE: A string containing the formatted value.
#
#-------------------------------------------------------------------------------

sub data_value_convert
{
   my $num = $#_ + 1;
   my $str = "";

   if ( $num != 2 )
   {  return( $str );
   }
   my $val = $_[0];
   my $fmt = $_[1];

# need to display preceeding "0" for octal and "0x" for hexidecimal?
# need to support upper 32-bits for 64-bit support!

   if ( $fmt == $DATA_FMT_DEC )
   {  $str = sprintf("%-18ld", $val);
   }
   elsif ( $fmt == $DATA_FMT_OCT )
   {  $str = sprintf("%-18lo", $val);
   }
   elsif ( $fmt == $DATA_FMT_HEX )
   {  $str = sprintf("%-16lX", $val);
   }
   elsif ( $fmt == $DATA_FMT_HEX8 )
   {  $str = sprintf("00 00 00 00 %02X %02X %02X %02X", 
         (($val >> 24) & 0xFF), (($val >> 16) & 0xFF), 
         (($val >> 8) & 0xFF), ($val & 0xFF));
   }
   elsif ( $fmt == $DATA_FMT_HEX16 )
   {  $str = sprintf("0000 0000 %04X %04X", (($val >> 16) & 0xFFFF), 
         ($val & 0xFFFF));
   }
   elsif ( $fmt == $DATA_FMT_HEX32 )
   {  $str = sprintf("00000000 %08X", ($val & 0xFFFFFFFF));
   }
   else # ( $fmt == $DATA_FMT_HEX64 )
   {  $str = sprintf("%016llX", $val);
   }
  
   return( $str ); 
}

#-------------------------------------------------------------------------------

1;
