#**************************************************************************
# *                                                                        *
# *               Copyright (C) 2002-2003 Silicon Graphics, Inc.           *
# *                                                                        *
# *  These coded instructions, statements, and computer programs  contain  *
# *  unpublished  proprietary  information of Silicon Graphics, Inc., and  *
# *  are protected by Federal copyright law.  They  may  not be disclosed  *
# *  to  third  parties  or copied or duplicated in any form, in whole or  *
# *  in part, without the prior written consent of Silicon Graphics, Inc.  *
# *                                                                        *
# **************************************************************************/
#
# # @(#)$Id: elog_decode.pm,v 1.13 2004/12/17 23:07:59 ncrumbak Exp $
#
#-------------------------------------------------------------------------------
#
#  FILE: elog_decode.pm
#
#  DESCRIPTION:  This file contains PERL source for the error log decode
#     utility for SHUB.
#
#  MODIFICATION HISTORY:
#
#     WHEN      WHO               WHAT
#     --------  ----------------  -----------------------------------------
#     04/09/03  Nelson Crumbaker  Moved core of elog_decode to this perl module.
#     06/06/03  Nelson Crumbaker  Fix in mfg_summary for uncorrectable errors 
#                                 not substituting for both dimms in the failing
#                                 dimm pair.
#     06/09/03  Nelson Crumbaker  Added code to collect timestamp information
#                                 from the shub errors and inserted function
#                                 sort_time (from errdmp) to print out a 
#                                 chronological list of timestamps to assist 
#                                 error debug.
#     08/21/03  Nelson Crumbaker  Added support for detailed FRU output (-F)
#                                 and support for BOARD_TYPE for IP57 (OPUS). 
#     11/21/03  Nelson Crumbaker  Added code to support only printing out the
#                                 decoded summary information.  Also modified
#                                 the MFG output to use the new DIMM format 
#                                 that is valid for both ALTIX and OPUS.
#     02/17/04  Nelson Crumbaker  Added support for TIO errors.  Also relaxed
#                                 pattern matching so that extra spaces within
#                                 elog entries as well as any characters that 
#                                 prefix lines are ignored. 
#     02/24/04  Nelson Crumbaker  Added more detailed decoding for XNHW_NI and 
#                                 PIHW errors.
#     03/18/04  Nelson Crumbaker  Check for address of zero before decoding 
#                                 dimm locations and nasid to avoid printing
#                                 bogus FRU locations.
#     07/13/04  Nelson Crumbaker  Added code for x4 dram chip support.
#                                 Fixed FSB Bit callout for DWORD 1 bits.
#                                 Added Directory Correctable details.
#     08/02/04  Nelson Crumbaker  Added support for Vortix node locations 2 & 3.
#     12/17/04  Nelson Crumbaker  Fixed output on fru summary of error count,
#                                 if is greater than 99, and also adjusted 
#                                 string length for errtype.  For MD Directory
#                                 SBE's I changed the fru summary to indicate
#                                 the equivalent FSB bit so that it can easily
#                                 be compared with PI/XN SBE's. 
#-------------------------------------------------------------------------------

# elog_decode subs
use ecc;
use dimm;

# Not needed for checked in code.
# use lib "/stand/sysco/lib/sscope";

# superscope subs
use jtag_subs;
use common_defs;
use shub_01_mmrs; 
use tio_01_mmrs;
use reg_subs;   
use mmr_subs;   
use fmt_subs;
use shub_subs;

$VNODE='';
$NASID='';
$MODULE='';
$LOCATION='';
$ENTRIES='';

# These variables control what is printed
$PRINT_VNODE_SUMMARY=1;
$USE_MFG_SUMMARY=0;
$PRINT_CPU_FRU_SUMMARY=1;
$PRINT_DETAILED_FRU=0;
$PRINT_SORTED_TIME_STAMPS=1;
$PRINT_BOARD_TYPE=1;
$x4_DRAM_CHIPS=0;

$PRINT_DIR_ERROR_FSB_BIT_FORMAT=1;

# Compare addresses on CPU errors with non-CPU errors on the node 
# and remove matching addresses.
$FILTER_CPU_ERRORS=0;

$PRINT_UNKNOWN_DATA=1;
$PRINT_CPU_ERRORS=1;
$PRINT_RTR_ERRORS=1;
$PRINT_INLINE_DECODE=1;
$OVERRIDE_REG_DETAIL='';

# Print warning messages.
$DEBUG_MSG=0;

# Prefix for embedded decoded print statements.
$DECODE_PRINT_PREFIX='|D| ';

# Attempt to process error log without full elog entries.
$PROCESS_INCOMPLETE_ELOGS=1;

$REG_FIELD_SKIP='skip';
$REG_FIELD_NONE='none';
$REG_FIELD_VALID='valid';
$REG_FIELD_ALL='all';

@CPU_STRING = ( 'A', 'B', 'C', 'D' );

$BOARD_TYPE="IP41";

$t_stamps = 0;


#
# The second value is used to determine how much information to
# print for the register:
#
#   skip  - Don't print anything.
#   none  - Print out the full register value.
#   valid - Print out the full register value and then break out each 
#           non-zero field value.
#   all   - Print out the full register value and then break out all
#           field values.
#
@REGS = ( ['SH_EVENT_OCCURRED',			$REG_FIELD_VALID],
          ['SH_FIRST_ERROR',			$REG_FIELD_VALID],
          ['SH_EVENT_OVERFLOW',			$REG_FIELD_VALID],

          ['SH_PI_ERROR_SUMMARY',		$REG_FIELD_VALID],
          ['SH_PI_ERROR_OVERFLOW',		$REG_FIELD_VALID],
          ['SH_PI_FIRST_ERROR',			$REG_FIELD_VALID],

          ['SH_PI_CORRECTED_DETAIL_1',		$REG_FIELD_ALL],
          ['SH_PI_CORRECTED_DETAIL_2',		$REG_FIELD_ALL],
          ['SH_PI_CORRECTED_DETAIL_3',		$REG_FIELD_ALL],
          ['SH_PI_CORRECTED_DETAIL_4',		$REG_FIELD_ALL],
          ['SH_PI_COR_TIME_STAMP',		$REG_FIELD_ALL],

          ['SH_PI_UNCORRECTED_DETAIL_1',	$REG_FIELD_ALL],
          ['SH_PI_UNCORRECTED_DETAIL_2',	$REG_FIELD_ALL],
          ['SH_PI_UNCORRECTED_DETAIL_3',	$REG_FIELD_ALL],
          ['SH_PI_UNCORRECTED_DETAIL_4',	$REG_FIELD_ALL],
          ['SH_PI_UNCOR_TIME_STAMP',		$REG_FIELD_ALL],

          ['SH_PI_ERROR_DETAIL_1',		$REG_FIELD_ALL],
          ['SH_PI_ERROR_DETAIL_2',		$REG_FIELD_ALL],
          ['SH_PI_HW_TIME_STAMP',		$REG_FIELD_ALL],

          ['SH_MEM_ERROR_SUMMARY',		$REG_FIELD_VALID],
          ['SH_MEM_ERROR_OVERFLOW',		$REG_FIELD_VALID],

          ['SH_DIR_COR_ERR_HDR_LOWER',		$REG_FIELD_ALL],
          ['SH_DIR_COR_ERR_HDR_UPPER',		$REG_FIELD_ALL],
          ['SH_MD_DQLP_MMR_XCERR1',		$REG_FIELD_ALL],
          ['SH_MD_DQLP_MMR_XCERR2',		$REG_FIELD_ALL],
          ['SH_MD_DQLP_MMR_YCERR1',		$REG_FIELD_ALL],
          ['SH_MD_DQLP_MMR_YCERR2',		$REG_FIELD_ALL],
          ['SH_MD_DQRP_MMR_XCERR1',		$REG_FIELD_ALL],
          ['SH_MD_DQRP_MMR_XCERR2',		$REG_FIELD_ALL],
          ['SH_MD_DQRP_MMR_YCERR1',		$REG_FIELD_ALL],
          ['SH_MD_DQRP_MMR_YCERR2',		$REG_FIELD_ALL],
          ['SH_MD_COR_TIME_STAMP',		$REG_FIELD_ALL],

          ['SH_DIR_UC_ERR_HDR_LOWER',		$REG_FIELD_ALL],
          ['SH_DIR_UC_ERR_HDR_UPPER',		$REG_FIELD_ALL],
          ['SH_MD_DQLP_MMR_XUERR1',		$REG_FIELD_ALL],
          ['SH_MD_DQLP_MMR_XUERR2',		$REG_FIELD_ALL],
          ['SH_MD_DQLP_MMR_YUERR1',		$REG_FIELD_ALL],
          ['SH_MD_DQLP_MMR_YUERR2',		$REG_FIELD_ALL],
          ['SH_MD_DQRP_MMR_XUERR1',		$REG_FIELD_ALL],
          ['SH_MD_DQRP_MMR_XUERR2',		$REG_FIELD_ALL],
          ['SH_MD_DQRP_MMR_YUERR1',		$REG_FIELD_ALL],
          ['SH_MD_DQRP_MMR_YUERR2',		$REG_FIELD_ALL],
          ['SH_MD_UNCOR_TIME_STAMP',		$REG_FIELD_ALL],

          ['SH_MD_DQLS_MMR_XAMOPW_ERR',		$REG_FIELD_ALL],
          ['SH_MD_DQRS_MMR_YAMOPW_ERR',		$REG_FIELD_ALL],

          ['SH_MISC_ERR_HDR_LOWER',		$REG_FIELD_ALL],
          ['SH_MISC_ERR_HDR_LOWER',		$REG_FIELD_ALL],
          ['SH_MD_DQLP_MMR_XPERR',		$REG_FIELD_ALL],
          ['SH_MD_DQLP_MMR_YPERR',		$REG_FIELD_ALL],
          ['SH_MD_DQRP_MMR_XPERR',		$REG_FIELD_ALL],
          ['SH_MD_DQRP_MMR_YPERR',		$REG_FIELD_ALL],
          ['SH_MD_HW_TIME_STAMP',		$REG_FIELD_ALL],

          ['SH_XN_ERROR_SUMMARY',		$REG_FIELD_VALID],
          ['SH_XN_ERROR_OVERFLOW',		$REG_FIELD_VALID],
          ['SH_XN_FIRST_ERROR',			$REG_FIELD_VALID],

          ['SH_XN_HW_TIME_STAMP',             	$REG_FIELD_ALL],

          ['SH_XN_CORRECTED_DETAIL_1', 		$REG_FIELD_ALL],
          ['SH_XN_CORRECTED_DETAIL_2', 		$REG_FIELD_ALL],
          ['SH_XN_CORRECTED_DETAIL_3', 		$REG_FIELD_ALL],
          ['SH_XN_CORRECTED_DETAIL_4', 		$REG_FIELD_ALL],
          ['SH_XN_COR_TIME_STAMP',		$REG_FIELD_ALL],

          ['SH_XN_UNCORRECTED_DETAIL_1', 	$REG_FIELD_ALL],
          ['SH_XN_UNCORRECTED_DETAIL_2', 	$REG_FIELD_ALL],
          ['SH_XN_UNCORRECTED_DETAIL_3', 	$REG_FIELD_ALL],
          ['SH_XN_UNCORRECTED_DETAIL_4', 	$REG_FIELD_ALL],
          ['SH_XN_UNCOR_TIME_STAMP',		$REG_FIELD_ALL],

          ['SH_XNMD_ERROR_SUMMARY',		$REG_FIELD_VALID],
          ['SH_XNMD_ERROR_OVERFLOW',		$REG_FIELD_VALID],
          ['SH_XNMD_FIRST_ERROR',		$REG_FIELD_VALID],
          ['SH_XNPI_ERROR_SUMMARY',		$REG_FIELD_VALID],
          ['SH_XNPI_ERROR_OVERFLOW',		$REG_FIELD_VALID],
          ['SH_XNPI_FIRST_ERROR',		$REG_FIELD_VALID],

          ['SH_XNMD_ERROR_DETAIL',		$REG_FIELD_ALL],
          ['SH_XNPI_ERROR_DETAIL',		$REG_FIELD_ALL],

          ['SH_XNIILB_ERROR_SUMMARY',		$REG_FIELD_VALID],
          ['SH_XNIILB_ERROR_OVERFLOW',		$REG_FIELD_VALID],
          ['SH_XNIILB_FIRST_ERROR',		$REG_FIELD_VALID],

          ['SH_XNIILB_ERROR_DETAIL_1',		$REG_FIELD_ALL],
          ['SH_XNIILB_ERROR_DETAIL_2',		$REG_FIELD_ALL],
          ['SH_XNIILB_ERROR_DETAIL_3',		$REG_FIELD_ALL],

          ['SH_NI0_ERROR_SUMMARY_1',		$REG_FIELD_VALID],
          ['SH_NI0_ERROR_SUMMARY_2',		$REG_FIELD_VALID],
          ['SH_NI0_ERROR_OVERFLOW_1',		$REG_FIELD_VALID],
          ['SH_NI0_ERROR_OVERFLOW_2',		$REG_FIELD_VALID],
          ['SH_NI0_FIRST_ERROR_1',		$REG_FIELD_VALID],
          ['SH_NI0_FIRST_ERROR_2',		$REG_FIELD_VALID],

          ['SH_NI0_ERROR_DETAIL_1',		$REG_FIELD_ALL],
          ['SH_NI0_ERROR_DETAIL_2',		$REG_FIELD_ALL],
          ['SH_NI0_ERROR_DETAIL_3',		$REG_FIELD_ALL],

          ['SH_NI1_ERROR_SUMMARY_1',            $REG_FIELD_VALID],
          ['SH_NI1_ERROR_SUMMARY_2',            $REG_FIELD_VALID],
          ['SH_NI1_ERROR_OVERFLOW_1',           $REG_FIELD_VALID],
          ['SH_NI1_ERROR_OVERFLOW_2',           $REG_FIELD_VALID],
          ['SH_NI1_FIRST_ERROR_1',              $REG_FIELD_VALID],
          ['SH_NI1_FIRST_ERROR_2',              $REG_FIELD_VALID],
          
          ['SH_NI1_ERROR_DETAIL_1',             $REG_FIELD_ALL],
          ['SH_NI1_ERROR_DETAIL_2',             $REG_FIELD_ALL],
          ['SH_NI1_ERROR_DETAIL_3',             $REG_FIELD_ALL], 

          ['SH_LB_ERROR_SUMMARY', 		$REG_FIELD_VALID],
          ['SH_LB_ERROR_OVERFLOW',		$REG_FIELD_VALID],
          ['SH_LB_FIRST_ERROR',  		$REG_FIELD_VALID],
          ['SH_LB_ERROR_DETAIL_1',		$REG_FIELD_ALL],
          ['SH_LB_ERROR_DETAIL_2',		$REG_FIELD_ALL],
          ['SH_LB_ERROR_DETAIL_3',		$REG_FIELD_ALL],
          ['SH_LB_ERROR_DETAIL_4',		$REG_FIELD_ALL],
          ['SH_LB_ERROR_DETAIL_5',		$REG_FIELD_ALL],
          ['SH_JUNK_ERROR_STATUS',		$REG_FIELD_ALL],
          ['SH_LB_HW_TIME_STAMP', 		$REG_FIELD_ALL]

        );


@TIO_REGS = 
        ( ['TIO_EVENT_OCCURRED',		$REG_FIELD_VALID],
          ['TIO_FIRST_ERROR',			$REG_FIELD_VALID],
          ['TIO_EVENT_OVERFLOW',		$REG_FIELD_VALID],

          ['TIO_ICE_HW_TIME_STAMP',		$REG_FIELD_ALL],
          ['TIO_ICE_GMS_ERROR_SUMMARY',		$REG_FIELD_VALID],
          ['TIO_ICE_GMS_FIRST_ERROR',		$REG_FIELD_VALID],
          ['TIO_ICE_GMS_ERROR_OVERFLOW',	$REG_FIELD_VALID],

          ['TIO_ICE_NPA_ERROR_SUMMARY',		$REG_FIELD_VALID],
          ['TIO_ICE_NPA_FIRST_ERROR',		$REG_FIELD_VALID],
          ['TIO_ICE_NPA_XN_ERROR_DETAIL_1',	$REG_FIELD_ALL],
          ['TIO_ICE_NPA_XN_ERROR_DETAIL_2',	$REG_FIELD_ALL],

          ['TIO_ICE_WRC_ERROR_SUMMARY',		$REG_FIELD_VALID],
          ['TIO_ICE_WRC_FIRST_ERROR',		$REG_FIELD_VALID],
          ['TIO_ICE_WRC_XN_ERROR_DETAIL_1',	$REG_FIELD_ALL],
          ['TIO_ICE_WRC_XN_ERROR_DETAIL_2',	$REG_FIELD_ALL],
          ['TIO_ICE_WRC_CT_ERROR_DETAIL_1',	$REG_FIELD_ALL],
          ['TIO_ICE_WRC_CT_ERROR_DETAIL_2',	$REG_FIELD_ALL],

          ['TIO_ICE_RRC_ERROR_SUMMARY',		$REG_FIELD_VALID],
          ['TIO_ICE_RRC_FIRST_ERROR',		$REG_FIELD_VALID],
          ['TIO_ICE_RRC_XN_ERROR_DETAIL_1',	$REG_FIELD_ALL],
          ['TIO_ICE_RRC_XN_ERROR_DETAIL_2',	$REG_FIELD_ALL],
          ['TIO_ICE_RRC_CT_ERROR_DETAIL_1',	$REG_FIELD_ALL],
          ['TIO_ICE_RRC_CT_ERROR_DETAIL_2',	$REG_FIELD_ALL],

          ['TIO_ICE_FRZ_ERROR_SUMMARY',		$REG_FIELD_VALID],
          ['TIO_ICE_FRZ_FIRST_ERROR',		$REG_FIELD_VALID],
          ['TIO_ICE_FRZ_XN_ERROR_DETAIL_1A',	$REG_FIELD_ALL],
          ['TIO_ICE_FRZ_XN_ERROR_DETAIL_2A',	$REG_FIELD_ALL],
          ['TIO_ICE_FRZ_XN_ERROR_DETAIL_1B',	$REG_FIELD_ALL],
          ['TIO_ICE_FRZ_XN_ERROR_DETAIL_2B',	$REG_FIELD_ALL],
          ['TIO_ICE_FRZ_CT_ERROR_DETAIL_1',	$REG_FIELD_ALL],

          ['TIO_ICE_PMI_RX_ERROR_SUMMARY',	$REG_FIELD_VALID],
          ['TIO_ICE_PMI_RX_FIRST_ERROR',	$REG_FIELD_VALID],
          ['TIO_ICE_PMI_RX_CT_ERROR_DETAIL_1',	$REG_FIELD_ALL],
          ['TIO_ICE_PMI_RX_CT_ERROR_DETAIL_2',	$REG_FIELD_ALL],

          ['TIO_ICE_PMI_TX_ERROR_SUMMARY',	$REG_FIELD_VALID],
          ['TIO_ICE_PMI_TX_FIRST_ERROR',	$REG_FIELD_VALID],
          ['TIO_ICE_PMI_TX_ERROR_OVERFLOW',	$REG_FIELD_VALID],

          ['TIO_ICE_CMA_ERROR_SUMMARY',		$REG_FIELD_VALID],
          ['TIO_ICE_CMA_FIRST_ERROR',		$REG_FIELD_VALID],
          ['TIO_ICE_CMA_ERROR_OVERFLOW',	$REG_FIELD_VALID],

          ['TIO_XN_ERROR_SUMMARY',		$REG_FIELD_VALID],
          ['TIO_XN_FIRST_ERROR',		$REG_FIELD_VALID],
          ['TIO_XN_ERROR_OVERFLOW',		$REG_FIELD_VALID],
          ['TIO_XN_HW_TIME_STAMP',		$REG_FIELD_ALL],

          ['TIO_XNIILB_ERROR_SUMMARY',		$REG_FIELD_VALID],
          ['TIO_XNIILB_ERROR_OVERFLOW',		$REG_FIELD_VALID],
          ['TIO_XNIILB_FIRST_ERROR',		$REG_FIELD_VALID],
          ['TIO_XNIILB_ERROR_SUMMARY_1',	$REG_FIELD_VALID],
          ['TIO_XNIILB_ERROR_OVERFLOW_1',	$REG_FIELD_VALID],
          ['TIO_XNIILB_FIRST_ERROR_1',		$REG_FIELD_VALID],
          ['TIO_XNIILB_ERROR_DETAIL_1',		$REG_FIELD_ALL],
          ['TIO_XNIILB_ERROR_DETAIL_2',		$REG_FIELD_ALL],
          ['TIO_XNIILB_ERROR_DETAIL_3',		$REG_FIELD_ALL],

          ['TIO_NI0_ERROR_SUMMARY_1',		$REG_FIELD_VALID],
          ['TIO_NI0_ERROR_OVERFLOW_1',		$REG_FIELD_VALID],
          ['TIO_NI0_FIRST_ERROR_1',		$REG_FIELD_VALID],
          ['TIO_NI0_ERROR_SUMMARY_2',		$REG_FIELD_VALID],
          ['TIO_NI0_ERROR_OVERFLOW_2',		$REG_FIELD_VALID],
          ['TIO_NI0_FIRST_ERROR_2',		$REG_FIELD_VALID],
          ['TIO_NI0_ERROR_DETAIL_1',		$REG_FIELD_ALL],
          ['TIO_NI0_ERROR_DETAIL_2',		$REG_FIELD_ALL],
          ['TIO_NI0_ERROR_DETAIL_3',		$REG_FIELD_ALL],
          ['TIO_NI0_LLP_STAT',			$REG_FIELD_ALL],
          ['TIO_NI0_LLP_ERR',			$REG_FIELD_ALL],

          ['TIO_NI1_ERROR_SUMMARY_1',		$REG_FIELD_VALID],
          ['TIO_NI1_ERROR_OVERFLOW_1',		$REG_FIELD_VALID],
          ['TIO_NI1_FIRST_ERROR_1',		$REG_FIELD_VALID],
          ['TIO_NI1_ERROR_SUMMARY_2',		$REG_FIELD_VALID],
          ['TIO_NI1_ERROR_OVERFLOW_2',		$REG_FIELD_VALID],
          ['TIO_NI1_FIRST_ERROR_2',		$REG_FIELD_VALID],
          ['TIO_NI1_ERROR_DETAIL_1',		$REG_FIELD_ALL],
          ['TIO_NI1_ERROR_DETAIL_2',		$REG_FIELD_ALL],
          ['TIO_NI1_ERROR_DETAIL_3',		$REG_FIELD_ALL],
          ['TIO_NI1_LLP_STAT',			$REG_FIELD_ALL],
          ['TIO_NI1_LLP_ERR',			$REG_FIELD_ALL],

          ['TIO_LB_ERROR_SUMMARY',		$REG_FIELD_VALID],
          ['TIO_LB_ERROR_OVERFLOW',		$REG_FIELD_VALID],
          ['TIO_LB_FIRST_ERROR',		$REG_FIELD_VALID],
          ['TIO_LB_ERROR_DETAIL_1',		$REG_FIELD_ALL],
          ['TIO_LB_ERROR_DETAIL_2',		$REG_FIELD_ALL],
          ['TIO_LB_ERROR_DETAIL_3',		$REG_FIELD_ALL],
          ['TIO_LB_ERROR_DETAIL_4',		$REG_FIELD_ALL],
          ['TIO_LB_HW_TIME_STAMP',		$REG_FIELD_ALL]

        );

sub elog_decode
{
   Init_Regs();

   if ( $OVERRIDE_REG_DETAIL )
   {  
      $reg_detail = $OVERRIDE_REG_DETAIL;
   }

   while ( $line = shift @_ )
   {
      if ($line =~ /Hardware Error Log -/)
      {
         ($VNODE) =    $line =~ /vnode\s+(\w+)/;
         ($NASID) =    $line =~ /nasid\s+(\w+)/;
         ($MODULE) =   $line =~ /module\s+(\w+)/;
         ($LOCATION) = $line =~ /location\s+(\w+)/;
         ($ENTRIES) =  $line =~ /:\s+(\d+)\s+entries/;

         $SMDK_INFO{$VNODE}{NASID}    = $NASID;
         $SMDK_INFO{$VNODE}{MODULE}   = $MODULE;
         $SMDK_INFO{$VNODE}{LOCATION} = $LOCATION;

         $NASID_TO_VNODE{$NASID} = $VNODE;

         if ( $ENTRIES eq "0" ) 
         {
            if ( $PRINT_UNKNOWN_DATA ) { print "$line"; }
         }
         else
         {
            if ( $ENTRIES eq "" ) 
            {
               if ( $PRINT_UNKNOWN_DATA ) { print "$line"; }
            }
            elsif ( ! ( $reg_detail =~ /skip/ ) ) 
            {
               printf("-------------------------------------------------------------------------------\n"); 
               print "$line";
            }
         }
      } 
      elsif ( $line =~ /SHUB:\s*((?:\w+\s+)+)\s*count:\s*(\w+)/ )
      {
         ($ERROR_TYPE,$COUNT) = $line =~ /SHUB:\s*((?:\w+\s+)+)\s*count:\s*(\w+)/;
         $COUNT = int($COUNT);

         if ( ! ( $reg_detail =~ /skip/ ) ) 
         {
            printf("-------------------------------------------------------------------------------\n"); 
            print $line;
         }
      }
      elsif ( $line =~ /(SH_\w+)\s*:\s*(\w+)/ )
      {
         @EINFO = ();
         CLR_Regs();

         while ( ($reg,$value) = $line =~ /(SH_\w+)\s*:\s*(\w+)/ )
         {
            reg_set_value($SHUB_MMR{$reg}, hex($value) );

            if ( $reg =~ /TIME_STAMP/) {
              $time_list[$t_stamps]->{NASID} = hex($NASID);
              $time_list[$t_stamps]->{REG} = $reg;
              $time_list[$t_stamps]->{TIME} = hex($value) & 0x7FFFFFFFFFFFFFFF;
              $time_list[$t_stamps]->{LOCATION} = sprintf("%s, %s", $MODULE, $LOCATION);
              $t_stamps++;
            }

            if ( ! ( $OVERRIDE_REG_DETAIL ) )
            {
               $reg_detail = $REG_PRINT_DETAIL{$reg};
            }

            if ( ! ( $reg_detail =~ /skip/ ) )
            {
               reg_write_value($SHUB_MMR{$reg}, STDOUT, $reg_detail );
            }

            push(@EINFO,$line);

            while ( @_[0] =~ /^\s*\n/ )
            {
               $line = shift @_;
            }

            $line = @_[0];

            if ( $line =~ /(SH_\w+)\s*:\s*(\w+)/ )
            {
               $line = shift @_;
            }
         }

         if ( $ERROR_TYPE =~ /PICE/ )
         {
            $einfo = Decode_PI_CE(@EINFO);
            if ( $PRINT_INLINE_DECODE )
            {
               printf("\n");
               printf("*******************************************************************************\n"); 
               print $DECODE_PRINT_PREFIX . $einfo;
               printf("*******************************************************************************\n"); 
               printf("\n");
            }
         }
         elsif ( $ERROR_TYPE =~ /XNCE/ )
         {
            $einfo = Decode_XN_CE(@EINFO);
            if ( $PRINT_INLINE_DECODE )
            {
               printf("\n");
               printf("*******************************************************************************\n");
               print $DECODE_PRINT_PREFIX . $einfo;
               printf("*******************************************************************************\n");
               printf("\n");
            }
         }
         elsif ( $ERROR_TYPE =~ /PIUCE/ )
         {
            $einfo = Decode_PI_UCE(@EINFO);
            if ( $PRINT_INLINE_DECODE )
            {
               printf("\n");
               printf("*******************************************************************************\n");
               print $DECODE_PRINT_PREFIX . $einfo;
               printf("*******************************************************************************\n");
               printf("\n");
            }
         }
         elsif ( $ERROR_TYPE =~ /XNUCE/ )
         {
            $einfo = Decode_XN_UCE(@EINFO);
            if ( $PRINT_INLINE_DECODE )
            {
               printf("\n");
               printf("*******************************************************************************\n");
               print $DECODE_PRINT_PREFIX . $einfo;
               printf("*******************************************************************************\n");
               printf("\n");
            }
         }
         elsif ( $ERROR_TYPE =~ /MDCE/ )
         {
            $einfo = Decode_MD_CE(@EINFO);
            if ( $PRINT_INLINE_DECODE )
            {
               printf("\n");
               printf("*******************************************************************************\n");
               print $DECODE_PRINT_PREFIX . $einfo;
               printf("*******************************************************************************\n");
               printf("\n");
            }
         }
         elsif ( $ERROR_TYPE =~ /MDUCE/ )
         {
            $einfo = Decode_MD_UCE(@EINFO);
            if ( $PRINT_INLINE_DECODE )
            {
               printf("\n");
               printf("*******************************************************************************\n");
               print $DECODE_PRINT_PREFIX . $einfo;
               printf("*******************************************************************************\n");
               printf("\n");
            }
         }
         elsif ( ($NI) = $ERROR_TYPE =~ /XNHW\s*NI(\d)/ )
         {
            $TIO = 0;
            $einfo = Decode_XNHW_NI($NI, $TIO);
            if ( $PRINT_INLINE_DECODE )
            {
               printf("\n");
               printf("*******************************************************************************\n");
               print $einfo;
               printf("*******************************************************************************\n");
               printf("\n");
            }
         }
         elsif ( $ERROR_TYPE =~ /PIHW/ )
         {
            $einfo = Decode_PIHW();
            if ( $PRINT_INLINE_DECODE )
            {
               printf("\n");
               printf("*******************************************************************************");
               print $einfo;
               printf("*******************************************************************************\n");
               printf("\n");
            }
         }
         $ERROR_TYPE='';
         $COUNT='';
      }
      elsif ( $line =~ /TIO:\s*\d*\s*((?:\w+\s+)+)\s*count:\s*(\w+)/ )
      {
         ($ERROR_TYPE,$COUNT) = $line =~ /TIO:\s*\d*\s*((?:\w+\s+)+)\s*count:\s*(\w+)/; 
         $COUNT = int($COUNT);

         if ( ! ( $reg_detail =~ /skip/ ) )
         {
            printf("-------------------------------------------------------------------------------\n");
            print $line;
         }
      }
      elsif ( $line =~ /(TIO_\w+)\s*:\s*(\w+)/ )
      {
         CLR_Regs();

         while ( ($reg,$value) = $line =~ /(TIO_\w+)\s*:\s*(\w+)/ )
         {
            reg_set_value($TIO_MMR{$reg}, hex($value) );

            if ( $reg =~ /TIME_STAMP/) {
              $time_list[$t_stamps]->{NASID} = hex($NASID);
              $time_list[$t_stamps]->{REG} = $reg;
              $time_list[$t_stamps]->{TIME} = hex($value) & 0x7FFFFFFFFFFFFFFF;
              $time_list[$t_stamps]->{LOCATION} = sprintf("%s, %s", $MODULE, $LOCATION);
              $t_stamps++;
            }

            if ( ! ( $OVERRIDE_REG_DETAIL ) )
            {
               $reg_detail = $REG_PRINT_DETAIL{$reg};
            }

            if ( ! ( $reg_detail =~ /skip/ ) )
            {
               reg_write_value($TIO_MMR{$reg}, STDOUT, $reg_detail );
            }

            while ( @_[0] =~ /^\s*\n/ )
            {
               $line = shift @_;
            }

            $line = @_[0];

            if ( $line =~ /(TIO_\w+)\s*:\s*(\w+)/ )
            {
               $line = shift @_;
            }
         }

         if ( ($NI) = $ERROR_TYPE =~ /XNHW\s*NI(\d)/ )
         {
            $TIO = 1;
            $einfo = Decode_XNHW_NI($NI, $TIO);
            if ( $PRINT_INLINE_DECODE )
            {
               printf("\n");
               printf("*******************************************************************************\n");
               print $einfo;
               printf("*******************************************************************************\n");
               printf("\n");
            }
         }

         $ERROR_TYPE='';
         $COUNT='';
      }
      elsif ( $line =~ /CPU:\s*(\w+\s+\w+\s+\w*)\s+count:\s*(\w+)/ )
      {
         ($ERROR_TYPE,$COUNT) = $line =~ /CPU:\s*(\w+\s+\w+\s+\w*)\s+count:\s*(\w+)/;
         $ERROR_TYPE =~ s/ /_/g;

         if ( $PRINT_CPU_ERRORS )
         {
            printf("\n");
            printf("-------------------------------------------------------------------------------\n"); 
            print $line;
         }
      }
      elsif ( $line =~ / ((?:MOD|PROC)_\w+)\s*:\s*(\w+)/ )
      {
         $fru_string = '';
         $nasid='';
         $cpu='';
         $address='';
         
         while ( ($reg,$value) = $line =~ / ((?:MOD|PROC)_\w+)\s*:\s*(\w+)/ )
         {
            $value = hex($value);

            if ( $reg =~ /PROC_CR_LID/ )
            {
               $nasid = ($value >> 16) & 0xfff;
               $cpu   = ($value >> 28) & 0x3;
               chop($line);
               $line = sprintf("%s # NASID: 0x%x, CPU_%s\n", $line, $nasid, $CPU_STRING[$cpu]);
            }
            elsif ( $reg =~ /MOD_TARGET_IDENTIFIER/ )
            {
               $address = $value;
            }

            if ( $PRINT_CPU_ERRORS ) { print $line; }

            while ( @_[0] =~ /^\s*\n/ )
            {
               $line = shift @_;
            }

            $line = @_[0];
            if ( $line =~ / ((?:MOD|PROC)_\w+)\s*:\s*(\w+)/ ) { $line = shift @_; }
         }

         $nasid = sprintf("0x%x,", $nasid);

         if ( $COUNT > 99 )
         {
            $count = sprintf("99+,", int($COUNT));
         }
         else
         {
            $count = sprintf("%d,", int($COUNT));
         }

         $fru_string .= sprintf("%-13s ADDR:0x%013llx, SYN:0x??, BIT:??, CNT:%-4s NASID:%-6s FRU:CPU_%s\n", 
            $ERROR_TYPE . ",", $address, $count, $nasid, $CPU_STRING[$cpu] );

         $SMDK_INFO{$VNODE}{CPU_FRU_STRING} .= $fru_string;

         if ( $PRINT_CPU_ERRORS )
         {
            printf("-------------------------------------------------------------------------------\n"); 
            printf("\n");
         }

         if ( $PRINT_INLINE_DECODE )
         {
            printf("\n");
            printf("*******************************************************************************\n");
            print $DECODE_PRINT_PREFIX . $fru_string;
            printf("*******************************************************************************\n");
            printf("\n");
         }

         $COUNT='';
         $ERROR_TYPE='';

      }
      elsif ( $line =~ /RT\s+router\s+\d+\s+port\s+\d+/ )
      {
         if ( $PRINT_RTR_ERRORS )
         {
            printf("\n");
            printf("-------------------------------------------------------------------------------\n"); 

            while ( $line =~ /RT\s+router\s+\d+\s+port\s+\d+/ )
            {
               print $line;
               $line = @_[0];
               if ( $line =~ /RT\s+router\s+\d+\s+port\s+\d+/ ) { $line = shift @_; }
            }

            printf("-------------------------------------------------------------------------------\n"); 
            printf("\n");
         }
      }
      else 
      {   
         if ( $PRINT_UNKNOWN_DATA ) { print "$line"; }
      }
   }

   if ( $FILTER_CPU_ERRORS)
   {
      foreach $vnode (keys % SMDK_INFO)
      {
         $fru_new = '';
         @fru = split(/\n/, $SMDK_INFO{$vnode}{CPU_FRU_STRING});
         foreach $fru (@fru) 
         {
            ($address) = $fru =~ /ADDR:(\w*),/;
            $address = hex($address);
            $address >>= 8;
            $address = sprintf("0x%011llx", $address);
            if ( ! ( $SMDK_INFO{$vnode}{FRU_STRING} =~ /$address/ ) )
            {
               $fru_new .= sprintf("$fru\n");
            }
         }
         $SMDK_INFO{$vnode}{CPU_FRU_STRING} = $fru_new;
      }
   }

   if ( $USE_MFG_SUMMARY )
   {
      foreach $VN (keys % SMDK_INFO)
      {
         $nasid    = $SMDK_INFO{$VN}{NASID};
         $module   = $SMDK_INFO{$VN}{MODULE};
         $location = $SMDK_INFO{$VN}{LOCATION};

         foreach $vnode (keys % SMDK_INFO)
         {
            $SMDK_INFO{$vnode}{FRU_STRING}     =~ s/NASID:$nasid,\s*/MODULE:$module.$location, /g;
            $SMDK_INFO{$vnode}{CPU_FRU_STRING} =~ s/NASID:$nasid,\s*/MODULE:$module.$location, /g;
         }
      }

      foreach $vnode (keys % SMDK_INFO)
      {
         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.([0-3]), FRU:)DG0_X_BUS_([L|R])/$1DIMM0_N$2_$3_BUS_X/g;
         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.([0-3]), FRU:)DG0_Y_BUS_([L|R])/$1DIMM1_N$2_$3_BUS_Y/g;
         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.([0-3]), FRU:)DG1_X_BUS_([L|R])/$1DIMM2_N$2_$3_BUS_X/g;
         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.([0-3]), FRU:)DG1_Y_BUS_([L|R])/$1DIMM3_N$2_$3_BUS_Y/g;
         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.([0-3]), FRU:)DG2_X_BUS_([L|R])/$1DIMM4_N$2_$3_BUS_X/g;
         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.([0-3]), FRU:)DG2_Y_BUS_([L|R])/$1DIMM5_N$2_$3_BUS_Y/g;
         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.([0-3]), FRU:)DG3_X_BUS_([L|R])/$1DIMM6_N$2_$3_BUS_X/g;
         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.([0-3]), FRU:)DG3_Y_BUS_([L|R])/$1DIMM7_N$2_$3_BUS_Y/g;
      }


#      foreach $vnode (keys % SMDK_INFO)
#      {
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.3, FRU:)DG0_X_BUS_L/$1DIMM0_N3_L_BUS_X/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.3, FRU:)DG0_Y_BUS_L/$1DIMM1_N3_L_BUS_Y/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.3, FRU:)DG1_X_BUS_L/$1DIMM2_N3_L_BUS_X/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.3, FRU:)DG1_Y_BUS_L/$1DIMM3_N3_L_BUS_Y/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.3, FRU:)DG2_X_BUS_L/$1DIMM4_N3_L_BUS_X/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.3, FRU:)DG2_Y_BUS_L/$1DIMM5_N3_L_BUS_Y/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.3, FRU:)DG3_X_BUS_L/$1DIMM6_N3_L_BUS_X/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.3, FRU:)DG3_Y_BUS_L/$1DIMM7_N3_L_BUS_Y/g;
#         
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.3, FRU:)DG0_X_BUS_R/$1DIMM0_N3_R_BUS_X/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.3, FRU:)DG0_Y_BUS_R/$1DIMM1_N3_R_BUS_Y/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.3, FRU:)DG1_X_BUS_R/$1DIMM2_N3_R_BUS_X/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.3, FRU:)DG1_Y_BUS_R/$1DIMM3_N3_R_BUS_Y/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.3, FRU:)DG2_X_BUS_R/$1DIMM4_N3_R_BUS_X/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.3, FRU:)DG2_Y_BUS_R/$1DIMM5_N3_R_BUS_Y/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.3, FRU:)DG3_X_BUS_R/$1DIMM6_N3_R_BUS_X/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.3, FRU:)DG3_Y_BUS_R/$1DIMM7_N3_R_BUS_Y/g;
#         
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.2, FRU:)DG0_X_BUS_L/$1DIMM0_N2_L_BUS_X/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.2, FRU:)DG0_Y_BUS_L/$1DIMM1_N2_L_BUS_Y/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.2, FRU:)DG1_X_BUS_L/$1DIMM2_N2_L_BUS_X/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.2, FRU:)DG1_Y_BUS_L/$1DIMM3_N2_L_BUS_Y/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.2, FRU:)DG2_X_BUS_L/$1DIMM4_N2_L_BUS_X/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.2, FRU:)DG2_Y_BUS_L/$1DIMM5_N2_L_BUS_Y/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.2, FRU:)DG3_X_BUS_L/$1DIMM6_N2_L_BUS_X/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.2, FRU:)DG3_Y_BUS_L/$1DIMM7_N2_L_BUS_Y/g;
#         
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.2, FRU:)DG0_X_BUS_R/$1DIMM0_N2_R_BUS_X/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.2, FRU:)DG0_Y_BUS_R/$1DIMM1_N2_R_BUS_Y/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.2, FRU:)DG1_X_BUS_R/$1DIMM2_N2_R_BUS_X/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.2, FRU:)DG1_Y_BUS_R/$1DIMM3_N2_R_BUS_Y/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.2, FRU:)DG2_X_BUS_R/$1DIMM4_N2_R_BUS_X/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.2, FRU:)DG2_Y_BUS_R/$1DIMM5_N2_R_BUS_Y/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.2, FRU:)DG3_X_BUS_R/$1DIMM6_N2_R_BUS_X/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.2, FRU:)DG3_Y_BUS_R/$1DIMM7_N2_R_BUS_Y/g;
#
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.1, FRU:)DG0_X_BUS_L/$1DIMM0_N1_L_BUS_X/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.1, FRU:)DG0_Y_BUS_L/$1DIMM1_N1_L_BUS_Y/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.1, FRU:)DG1_X_BUS_L/$1DIMM2_N1_L_BUS_X/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.1, FRU:)DG1_Y_BUS_L/$1DIMM3_N1_L_BUS_Y/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.1, FRU:)DG2_X_BUS_L/$1DIMM4_N1_L_BUS_X/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.1, FRU:)DG2_Y_BUS_L/$1DIMM5_N1_L_BUS_Y/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.1, FRU:)DG3_X_BUS_L/$1DIMM6_N1_L_BUS_X/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.1, FRU:)DG3_Y_BUS_L/$1DIMM7_N1_L_BUS_Y/g;
#   
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.1, FRU:)DG0_X_BUS_R/$1DIMM0_N1_R_BUS_X/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.1, FRU:)DG0_Y_BUS_R/$1DIMM1_N1_R_BUS_Y/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.1, FRU:)DG1_X_BUS_R/$1DIMM2_N1_R_BUS_X/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.1, FRU:)DG1_Y_BUS_R/$1DIMM3_N1_R_BUS_Y/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.1, FRU:)DG2_X_BUS_R/$1DIMM4_N1_R_BUS_X/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.1, FRU:)DG2_Y_BUS_R/$1DIMM5_N1_R_BUS_Y/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.1, FRU:)DG3_X_BUS_R/$1DIMM6_N1_R_BUS_X/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.1, FRU:)DG3_Y_BUS_R/$1DIMM7_N1_R_BUS_Y/g;
#
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.0, FRU:)DG0_X_BUS_L/$1DIMM0_N0_L_BUS_X/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.0, FRU:)DG0_Y_BUS_L/$1DIMM1_N0_L_BUS_Y/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.0, FRU:)DG1_X_BUS_L/$1DIMM2_N0_L_BUS_X/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.0, FRU:)DG1_Y_BUS_L/$1DIMM3_N0_L_BUS_Y/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.0, FRU:)DG2_X_BUS_L/$1DIMM4_N0_L_BUS_X/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.0, FRU:)DG2_Y_BUS_L/$1DIMM5_N0_L_BUS_Y/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.0, FRU:)DG3_X_BUS_L/$1DIMM6_N0_L_BUS_X/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.0, FRU:)DG3_Y_BUS_L/$1DIMM7_N0_L_BUS_Y/g;
#
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.0, FRU:)DG0_X_BUS_R/$1DIMM0_N0_R_BUS_X/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.0, FRU:)DG0_Y_BUS_R/$1DIMM1_N0_R_BUS_Y/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.0, FRU:)DG1_X_BUS_R/$1DIMM2_N0_R_BUS_X/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.0, FRU:)DG1_Y_BUS_R/$1DIMM3_N0_R_BUS_Y/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.0, FRU:)DG2_X_BUS_R/$1DIMM4_N0_R_BUS_X/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.0, FRU:)DG2_Y_BUS_R/$1DIMM5_N0_R_BUS_Y/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.0, FRU:)DG3_X_BUS_R/$1DIMM6_N0_R_BUS_X/g;
#         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(MODULE:\w*.0, FRU:)DG3_Y_BUS_R/$1DIMM7_N0_R_BUS_Y/g;
#
#      }

      foreach $vnode (keys % SMDK_INFO)
      {
         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/(DIMM[0-7]_N[0,1,2,3]_)[L,R](_BUS_[X,Y]),DG\w*/$1L$2,$1R$2/g;
      }

      # Convert DIMMs without module location mapping to MFG Format.
      foreach $vnode (keys % SMDK_INFO)
      {
         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/DG0_X_BUS_L/DIMM0_N?_L_BUS_X/g;
         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/DG0_Y_BUS_L/DIMM1_N?_L_BUS_Y/g;
         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/DG1_X_BUS_L/DIMM2_N?_L_BUS_X/g;
         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/DG1_Y_BUS_L/DIMM3_N?_L_BUS_Y/g;
         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/DG2_X_BUS_L/DIMM4_N?_L_BUS_X/g;
         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/DG2_Y_BUS_L/DIMM5_N?_L_BUS_Y/g;
         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/DG3_X_BUS_L/DIMM6_N?_L_BUS_X/g;
         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/DG3_Y_BUS_L/DIMM7_N?_L_BUS_Y/g;
         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/DG0_X_BUS_R/DIMM0_N?_R_BUS_X/g;
         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/DG0_Y_BUS_R/DIMM1_N?_R_BUS_Y/g;
         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/DG1_X_BUS_R/DIMM2_N?_R_BUS_X/g;
         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/DG1_Y_BUS_R/DIMM3_N?_R_BUS_Y/g;
         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/DG2_X_BUS_R/DIMM4_N?_R_BUS_X/g;
         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/DG2_Y_BUS_R/DIMM5_N?_R_BUS_Y/g;
         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/DG3_X_BUS_R/DIMM6_N?_R_BUS_X/g;
         $SMDK_INFO{$vnode}{FRU_STRING} =~ s/DG3_Y_BUS_R/DIMM7_N?_R_BUS_Y/g;
      }

   }

   if ( $PRINT_SORTED_TIME_STAMPS )
   {
      printf("-------------------------------------------------------------------------------\n");
      sort_times();
      printf("\n\n-------------------------------------------------------------------------------\n\n");
   }

   if ( $PRINT_BOARD_TYPE )
   {
      printf("-------------------------------------------------------------------------------\n");
      if ( $BOARD_TYPE =~ /IP57/ )
      {
         printf(" BOARD_TYPE:IP57\n");
      }
      elsif ( $BOARD_TYPE =~ /IP41/ )
      {
         printf(" BOARD_TYPE:IP41\n");
      }
      else
      {
         printf(" BOARD_TYPE:$BOARD_TYPE is UNKNOWN\n");
         exit(1);
      }
   }

   if ( $PRINT_VNODE_SUMMARY )
   {
      foreach $vnode (keys % SMDK_INFO)
      {
         $nasid    = $SMDK_INFO{$vnode}{NASID};
         $module   = $SMDK_INFO{$vnode}{MODULE};
         $location = $SMDK_INFO{$vnode}{LOCATION};
         if ( $SMDK_INFO{$vnode}{FRU_STRING} || 
              ( $SMDK_INFO{$vnode}{CPU_FRU_STRING} && $PRINT_CPU_FRU_SUMMARY ) )
         {
            printf("-------------------------------------------------------------------------------\n");
            printf(" SUMMARY: VNODE:$vnode, NASID:$nasid, MODULE:$module.$location\n" );
            printf("-------------------------------------------------------------------------------\n");

#           $line = "$SMDK_INFO{$vnode}{FRU_STRING}";
#           print $line;

            foreach $line ( split(/\n/, $SMDK_INFO{$vnode}{FRU_STRING}) )
            {
               print "$module.$location, $line\n";
            }

#           if ( $PRINT_CPU_FRU_SUMMARY ) { print $SMDK_INFO{$vnode}{CPU_FRU_STRING}; } 

            foreach $line ( split(/\n/, $SMDK_INFO{$vnode}{CPU_FRU_STRING}) )
            {
               print "$module.$location, $line\n";
            }

            print "\n";
         }
      }
   }
}

sub Init_Regs
{
   local $array_length;
   local $array;
   local $reg;

   $array_length = 0;
   $array_length = @REGS;

   for ($array=0; $array<$array_length; $array++)
   {
      $reg = $REGS[$array][0];
      $REG_PRINT_DETAIL{$reg} = $REGS[$array][1];
      $SHUB_MMR{$reg} = reg_create_mmr_inst('shub_01',$reg);
   }

   $array_length = @TIO_REGS;

   for ($array=0; $array<$array_length; $array++)
   {
      $reg = $TIO_REGS[$array][0];
      $REG_PRINT_DETAIL{$reg} = $TIO_REGS[$array][1];
      $TIO_MMR{$reg} = reg_create_mmr_inst('tio_01',$reg);
   }

### TEMP FIX FOR THE SH_FIRST_ERROR REGISTER DEFINE ###
#   $SHUB_MMR{'SH_FIRST_ERROR'} = 
#      reg_create_mmr_inst('shub_01','SH_EVENT_OCCURRED');
}

sub CLR_Regs
{
   local $array_length;
   local $array;
   local $reg;

   $array_length = 0;
   $array_length = @REGS;

   for ($array=0; $array<$array_length; $array++)
   {
      $reg = $REGS[$array][0];
      reg_set_value($SHUB_MMR{$reg}, 0);
   }

   $array_length = @TIO_REGS;

   for ($array=0; $array<$array_length; $array++)
   {
      $reg = $TIO_REGS[$array][0];
      reg_set_value($TIO_MMR{$reg}, 0);
   }

}

sub Decode_PI_CE 
{
   $summary = 
      reg_get_field_value ( $SHUB_MMR{'SH_PI_ERROR_SUMMARY'}, 'FSB_SHUB_CE' );
   $fru_string = '';

   if ( $summary )
   {
      $address = reg_get_field_value ( $SHUB_MMR{'SH_PI_CORRECTED_DETAIL_3'},
                                       'ADDRESS' );
      $address = $address << 3;
  
      $syndrome = reg_get_field_value ( $SHUB_MMR{'SH_PI_CORRECTED_DETAIL_3'},
                                       'SYNDROME' );

      $bit = get_bit_from_syndrome($syndrome);

      if (((($address & 0x8) == 0x8) && ($bit < 64)) || ((($address & 0x8) == 0x0) && ($bit > 63))) {
         $bit +=64;
      }
      elsif ((($address & 0x8) == 0x8) && ($bit > 63))  {
         $bit +=72;
      }

      $nasid = sprintf("0x%x,", int($NASID) );

      if ( $COUNT > 99 )
      {
         $count = sprintf("99+,", int($COUNT));
      }
      else
      {
         $count = sprintf("%d,", int($COUNT));
      }

      $fru_string .= sprintf("%-13s ADDR:0x%013llx, SYN:0x%02x,  FSB_BIT:%d, CNT:%-4s NASID:%-6s FRU:%s\n", 
         "FSB_SHUB_CE,", $address, $syndrome, $bit, $count, $nasid, "CPU" );

      $SMDK_INFO{$VNODE}{FRU_STRING} .= $fru_string;

   }

   $summary =
      reg_get_field_value ( $SHUB_MMR{'SH_PI_ERROR_SUMMARY'}, 'SHUB_FSB_CE' );

   if ( $summary )
   {
      $address = reg_get_field_value ( $SHUB_MMR{'SH_PI_CORRECTED_DETAIL_1'},
                                       'ADDRESS' );
      $address = $address << 3;

      $nasid = ($address >> 38) & 0x7ff;
                                       
      $syndrome = reg_get_field_value ( $SHUB_MMR{'SH_PI_CORRECTED_DETAIL_1'},
                                       'SYNDROME' );
                                       
      $bit = get_bit_from_syndrome($syndrome);

      if (((($address & 0x8) == 0x8) && ($bit < 64)) || ((($address & 0x8) == 0x0) && ($bit > 63))) {
         $bit +=64;
      }
      elsif ((($address & 0x8) == 0x8) && ($bit > 63))  {
         $bit +=72;
      }

      if ( $address != 0 )
      {
         $nasid = sprintf("0x%x,",$nasid);
      }
      else
      {
         $nasid = sprintf("0x??");
      }

      if ( $COUNT > 99 )
      {
         $count = sprintf("99+,", int($COUNT));
      }
      else
      {
         $count = sprintf("%d,", int($COUNT));
      }

      $fru_string .= sprintf("%-13s ADDR:0x%013llx, SYN:0x%02x,  FSB_BIT:%d, CNT:%-4s NASID:%-6s FRU:%s\n", 
         "SHUB_FSB_CE,", $address, $syndrome, $bit, $count, $nasid, 
         find_dimm($address,$syndrome,$BOARD_TYPE,$PRINT_DETAILED_FRU,$x4_DRAM_CHIPS) );
   }

   $SMDK_INFO{$VNODE}{FRU_STRING} .= $fru_string;

   return $fru_string;
}

sub Decode_XN_CE
{
   $summary =
      reg_get_field_value ( $SHUB_MMR{'SH_XN_ERROR_SUMMARY'}, 'XNMD_SBE_ERROR' );
   $fru_string = '';

   if ( $summary )
   {
      $addr_44_3 = reg_get_field_value ( $SHUB_MMR{'SH_XN_CORRECTED_DETAIL_3'}, 'HEADER0' );
      $addr_44_3 = (($addr_44_3 & 0xffffffffffc00000) >> 19);
      $addr_49_45 = reg_get_field_value ( $SHUB_MMR{'SH_XN_CORRECTED_DETAIL_4'}, 'HEADER1' );
      $addr_49_45 = ($addr_49_45 & 0x1f) << 45;
      $err_group = reg_get_field_value ( $SHUB_MMR{'SH_XN_CORRECTED_DETAIL_4'}, 'ERR_GROUP' );
      $address = $addr_49_45 | $addr_44_3 | (($err_group & 0x1)<<3);

      $nasid = ($address >> 38) & 0x7ff;

      $syndrome = reg_get_value ( $SHUB_MMR{'SH_XN_CORRECTED_DETAIL_1'} );
      $syndrome = ($syndrome >> ($err_group*16)) & 0xFF;

      $bit = get_bit_from_syndrome($syndrome);

      if (((($address & 0x8) == 0x8) && ($bit < 64)) || ((($address & 0x8) == 0x0) && ($bit > 63))) {
         $bit +=64;
      }
      elsif ((($address & 0x8) == 0x8) && ($bit > 63))  {
         $bit +=72;
      }

      if ( $address != 0 )
      {
         $nasid = sprintf("0x%x,",$nasid);
      }
      else
      {
         $nasid = sprintf("0x??");
      }

      if ( $COUNT > 99 )
      {
         $count = sprintf("99+,", int($COUNT));
      }
      else
      {
         $count = sprintf("%d,", int($COUNT));
      }

      $fru_string .= sprintf("%-13s ADDR:0x%013llx, SYN:0x%02x,  FSB_BIT:%d, CNT:%-4s NASID:%-6s FRU:%s\n",
         "XNMD_SBE,", $address, $syndrome, $bit, $count, $nasid,
         find_dimm($address,$syndrome,$BOARD_TYPE,$PRINT_DETAILED_FRU,$x4_DRAM_CHIPS) );
   }

   $SMDK_INFO{$VNODE}{FRU_STRING} .= $fru_string;
   return $fru_string;

}

sub Decode_PI_UCE 
{
   $summary = 
      reg_get_field_value ( $SHUB_MMR{'SH_PI_ERROR_SUMMARY'}, 'FSB_SHUB_UCE' );
   $fru_string = '';

   if ( $summary )
   {
      $address = reg_get_field_value ( $SHUB_MMR{'SH_PI_UNCORRECTED_DETAIL_3'},
                                       'ADDRESS' );
      $address = $address << 3;
  
      $syndrome = reg_get_field_value ( $SHUB_MMR{'SH_PI_UNCORRECTED_DETAIL_3'},
                                       'SYNDROME' );

      $nasid = sprintf("0x%x,", int($NASID) );

      if ( $COUNT > 99 )
      {
         $count = sprintf("99+,", int($COUNT));
      }
      else
      {
         $count = sprintf("%d,", int($COUNT));
      }

      $fru_string .= sprintf("%-13s ADDR:0x%013llx, SYN:0x%02x, CNT:%-4s NASID:%-6s FRU:%s,%s\n",
         "FSB_SHUB_UCE,", $address, $syndrome, $count, $nasid, 
         find_dimm($address,$syndrome,$BOARD_TYPE,$PRINT_DETAILED_FRU,$x4_DRAM_CHIPS), "CPU");
   }

   $summary =
      reg_get_field_value ( $SHUB_MMR{'SH_PI_ERROR_SUMMARY'}, 'SHUB_FSB_UCE' );

   if ( $summary )
   {
      $address = reg_get_field_value ( $SHUB_MMR{'SH_PI_UNCORRECTED_DETAIL_1'},
                                       'ADDRESS' );
      $address = $address << 3;

      $nasid = ($address >> 38) & 0x7ff;
                                       
      $syndrome = reg_get_field_value ( $SHUB_MMR{'SH_PI_UNCORRECTED_DETAIL_1'},
                                       'SYNDROME' );
                                       
      if ( $address != 0 )
      {
         $nasid = sprintf("0x%x,",$nasid);
      }
      else
      {
         $nasid = sprintf("0x??");
      }

      if ( $COUNT > 99 )
      {
         $count = sprintf("99+,", int($COUNT));
      }
      else
      {
         $count = sprintf("%d,", int($COUNT));
      }

      $fru_string .= sprintf("%-13s ADDR:0x%013llx, SYN:0x%02x, CNT:%-4s NASID:%-6s FRU:%s\n",
         "SHUB_FSB_UCE,", $address, $syndrome, $count, $nasid,
         find_dimm($address,$syndrome,$BOARD_TYPE,$PRINT_DETAILED_FRU,$x4_DRAM_CHIPS) );
   }

   $SMDK_INFO{$VNODE}{FRU_STRING} .= $fru_string;
   return $fru_string;

}

sub Decode_XN_UCE
{
   $summary =
      reg_get_field_value ( $SHUB_MMR{'SH_XN_ERROR_SUMMARY'}, 'XNMD_UCE_ERROR' );
   $fru_string = '';

   if ( $summary )
   {
      $addr_44_3 = reg_get_field_value ( $SHUB_MMR{'SH_XN_UNCORRECTED_DETAIL_3'}, 'HEADER0' );
      $addr_44_3 = (($addr_44_3 & 0xffffffffffc00000) >> 19);
      $addr_49_45 = reg_get_field_value ( $SHUB_MMR{'SH_XN_UNCORRECTED_DETAIL_4'}, 'HEADER1' );
      $addr_49_45 = ($addr_49_45 & 0x1f) << 45;
      $err_group = reg_get_field_value ( $SHUB_MMR{'SH_XN_UNCORRECTED_DETAIL_4'}, 'ERR_GROUP' );
      $address = $addr_49_45 | $addr_44_3 | (($err_group & 0x1)<<3);

      $nasid = ($address >> 38) & 0x7ff;

      $syndrome = reg_get_value ( $SHUB_MMR{'SH_XN_UNCORRECTED_DETAIL_1'} );
      $syndrome = ($syndrome >> ($err_group*16)) & 0xFF;

      if ( $address != 0 )
      {
         $nasid = sprintf("0x%x,",$nasid);
      }
      else
      {
         $nasid = sprintf("0x??");
      }

      if ( $COUNT > 99 )
      {
         $count = sprintf("99+,", int($COUNT));
      }
      else
      {
         $count = sprintf("%d,", int($COUNT));
      }

      $fru_string .= sprintf("%-13s ADDR:0x%013llx, SYN:0x%02x, CNT:%-4s NASID:%-6s FRU:%s\n",
         "XNMD_UCE,", $address, $syndrome, $count, $nasid,
         find_dimm($address,$syndrome,$BOARD_TYPE,$PRINT_DETAILED_FRU,$x4_DRAM_CHIPS) );
   }

   $SMDK_INFO{$VNODE}{FRU_STRING} .= $fru_string;
   return $fru_string;
}

sub Decode_MD_CE
{
   my $fsb_bit;

   $summary =
      reg_get_field_value ( $SHUB_MMR{'SH_MEM_ERROR_SUMMARY'}, 'DQLP_DIR_COR' );
   $fru_string = '';
      
   if ( $summary )
   {
      $address = reg_get_field_value ( $SHUB_MMR{'SH_DIR_COR_ERR_HDR_LOWER'}, 'ADDR' );
      $address <<= 3;
      $bus = ( $address >> 7 ) & 0x1;
      $LorR = ( $address >> 11 ) & 0x1;
      $dimm_group = ( $address >> 34 ) & 0x3;

      if ( $address != 0 )
      {
         $fru1 = "DG" . $dimm_group . "_" . (($bus) ? "X" : "Y") .
                 "_BUS_" . (($LorR) ?  "R" : "L" );
      }
      else
      {
         $fru1 = "UNKNOWN (INVALID ADDRESS)";
      }

      $nasid = sprintf("0x%x,", int($NASID) );

      if ( $COUNT > 99 )
      {
         $count = sprintf("99+,", int($COUNT));
      }
      else
      {
         $count = sprintf("%d,", int($COUNT));
      }

      if ( $bus == 1 )
      {
         if ( reg_get_field_value ( $SHUB_MMR{'SH_MD_DQLP_MMR_XCERR1'}, VAL ) )
         {
            $eccgrp = 0;
            $grp = reg_get_field_value ( $SHUB_MMR{'SH_MD_DQLP_MMR_XCERR1'}, GRP1 );
         }
         elsif ( reg_get_field_value ( $SHUB_MMR{'SH_MD_DQLP_MMR_XCERR2'}, VAL) )
         {
            $eccgrp = 1;
            $grp = reg_get_field_value ( $SHUB_MMR{'SH_MD_DQLP_MMR_XCERR2'}, GRP2 );
         }
         else
         {
            $grp = 0;
         }
      }
      else
      {
         if ( reg_get_field_value ( $SHUB_MMR{'SH_MD_DQLP_MMR_YCERR1'}, VAL ) )
         {
            $eccgrp = 0;
            $grp = reg_get_field_value ( $SHUB_MMR{'SH_MD_DQLP_MMR_YCERR1'}, GRP1 );
         }  
         elsif ( reg_get_field_value ( $SHUB_MMR{'SH_MD_DQLP_MMR_YCERR2'}, VAL))
         {
            $eccgrp = 1;
            $grp = reg_get_field_value ( $SHUB_MMR{'SH_MD_DQLP_MMR_YCERR2'}, GRP2 );
         }  
         else
         {
            $grp = 0;
         }  
      }

      $syn = gen_dir_syndrome($grp);
      $bit = get_dir_bit_from_syndrome($syn);

      $mdbit = (($bit>>2) * 8) + ($bit%4);

      if ( $eccgrp == 0 )
      {
         $mdbit += 4;
      }

      $fsb_bit = fsb_bit_from_dir_bit($bit,$eccgrp,$LorR);

      if ( $PRINT_DETAILED_FRU )
      {
         $dimmInfo = getDimmInfoFromMDBit($mdbit, $address, $BOARD_TYPE, $x4_DRAM_CHIPS);
      }
      else
      {
         $dimmInfo = "";
      }

      if ( $PRINT_DIR_ERROR_FSB_BIT_FORMAT )
      {
         if ( $fsb_bit != -1 )
         {
            $fru_string .= sprintf("%-13s ADDR:0x%013llx, DSYN:0x%02x, FSB_BIT:%d, CNT:%-4s NASID:%-6s FRU:%s%s\n", 
               "DQLP_DIR_COR,", $address, $syn, $fsb_bit, $count, $nasid, $fru1, $dimmInfo );
         }
         else
         {
            $fru_string .= sprintf("%-13s ADDR:0x%013llx, DSYN:0x%02x, FSB_BIT:??, CNT:%-4s NASID:%-6s FRU:%s\n",
               "DQLP_DIR_COR,", $address, $syn, $count, $nasid, $fru1 );
         }
      }
      else
      {
         if ( $bit != -1 )
         {
            $fru_string .= sprintf("%-13s ADDR:0x%013llx, DSYN:0x%02x,DBIT:%02d, EGRP:%d, CNT:%-4s NASID:%-6s FRU:%s%s\n", 
               "DQLP_DIR_COR,", $address, $syn, $bit, $eccgrp+1, $count, $nasid, $fru1, $dimmInfo );
         }
         else
         {
            $fru_string .= sprintf("%-13s ADDR:0x%013llx, DSYN:0x%02x,DBIT:??, CNT:%-4s NASID:%-6s FRU:%s\n",
               "DQLP_DIR_COR,", $address, $syn, $count, $nasid, $fru1 );
         }
      }
   }

   $summary =
      reg_get_field_value ( $SHUB_MMR{'SH_MEM_ERROR_SUMMARY'}, 'DQRP_DIR_COR' );

   if ( $summary )
   {
      $address = reg_get_field_value ( $SHUB_MMR{'SH_DIR_COR_ERR_HDR_LOWER'}, 'ADDR' );
      $address <<= 3;
      $bus = ( $address >> 7 ) & 0x1;
      $LorR = ( $address >> 11 ) & 0x1;
      $dimm_group = ( $address >> 34 ) & 0x3;
      $eccgrp = ( $address >> 8 ) & 0x1;

      if ( $address != 0 )
      {
         $fru1 = "DG" . $dimm_group . "_" . (($bus) ? "X" : "Y") .
                 "_BUS_" . (($LorR) ?  "R" : "L" );
      }
      else
      {
         $fru1 = "UNKNOWN (INVALID ADDRESS)";
      }

      $nasid = sprintf("0x%x,", int($NASID) );

      if ( $COUNT > 99 )
      {
         $count = sprintf("99+,", int($COUNT));
      }
      else
      {
         $count = sprintf("%d,", int($COUNT));
      }

      if ( $bus == 1 )
      {
         if ( reg_get_field_value ( $SHUB_MMR{'SH_MD_DQRP_MMR_XCERR1'}, VAL ) )
         {
            $eccgrp = 0;
            $grp = reg_get_field_value ( $SHUB_MMR{'SH_MD_DQRP_MMR_XCERR1'}, GRP1 );
         }
         elsif ( reg_get_field_value ( $SHUB_MMR{'SH_MD_DQRP_MMR_XCERR2'}, VAL))
         {
            $eccgrp = 1;
            $grp = reg_get_field_value ( $SHUB_MMR{'SH_MD_DQRP_MMR_XCERR2'}, GRP2 );
         }
         else
         {
            $grp = 0;
         }
      }
      else
      {
         if ( reg_get_field_value ( $SHUB_MMR{'SH_MD_DQRP_MMR_YCERR1'}, VAL ) )
         {
            $eccgrp = 0;
            $grp = reg_get_field_value ( $SHUB_MMR{'SH_MD_DQRP_MMR_YCERR1'}, GRP1 );
         }
         elsif ( reg_get_field_value ( $SHUB_MMR{'SH_MD_DQRP_MMR_YCERR2'}, VAL))
         {
            $eccgrp = 1;
            $grp = reg_get_field_value ( $SHUB_MMR{'SH_MD_DQRP_MMR_YCERR2'}, GRP2 );
         }
         else
         {
            $grp = 0;
         }
      }

      $syn = gen_dir_syndrome($grp);
      $bit = get_dir_bit_from_syndrome($syn);

      $mdbit = (($bit>>2) * 8) + ($bit%4);

      if ( $eccgrp == 0 )
      {
         $mdbit += 4;
      }

      $fsb_bit = fsb_bit_from_dir_bit($bit,$eccgrp,$LorR);

      if ( $PRINT_DETAILED_FRU )
      {
         $dimmInfo = getDimmInfoFromMDBit($mdbit, $address, $BOARD_TYPE, $x4_DRAM_CHIPS);
      }
      else
      {
         $dimmInfo = "";
      }
      
      if ( $PRINT_DIR_ERROR_FSB_BIT_FORMAT )
      {
         if ( $fsb_bit != -1 )
         {  
            $fru_string .= sprintf("%-13s ADDR:0x%013llx, DSYN:0x%02x, FSB_BIT:%d, CNT:%-4s NASID:%-6s FRU:%s%s\n", 
               "DQRP_DIR_COR,", $address, $syn, $fsb_bit, $count, $nasid, $fru1, $dimmInfo );
         }
         else
         {  
            $fru_string .= sprintf("%-13s ADDR:0x%013llx, DSYN:0x%02x, FSB_BIT:??, CNT:%-4s NASID:%-6s FRU:%s\n",
               "DQRP_DIR_COR,", $address, $syn, $count, $nasid, $fru1 );
         }
      }
      else
      {
         if ( $bit != -1 )
         {  
            $fru_string .= sprintf("%-13s ADDR:0x%013llx, DSYN:0x%02x, DBIT:%02d, EGRP:%d, CNT:%-4s NASID:%-6s FRU:%s%s\n", 
               "DQRP_DIR_COR,", $address, $syn, $bit, $eccgrp+1, $count, $nasid, $fru1, $dimmInfo );
         }
         else
         {  
            $fru_string .= sprintf("%-13s ADDR:0x%013llx, DSYN:0x%02x, DBIT:??, CNT:%-4s NASID:%-6s FRU:%s\n",
               "DQRP_DIR_COR,", $address, $syn, $count, $nasid, $fru1 );
         }
      }

   }
   $SMDK_INFO{$VNODE}{FRU_STRING} .= $fru_string;
   return $fru_string;
}

sub Decode_MD_UCE
{
   $summary =
      reg_get_field_value ( $SHUB_MMR{'SH_MEM_ERROR_SUMMARY'}, 'DQLP_DIR_UC' );
   $fru_string = '';
      
   if ( $summary )
   {
      $address = reg_get_field_value ( $SHUB_MMR{'SH_DIR_UC_ERR_HDR_LOWER'}, 'ADDR' );
      $address <<= 3;
      $bus = ( $address >> 7 ) & 0x1;
      $LorR = ( $address >> 11 ) & 0x1;
      $dimm_group = ( $address >> 34 ) & 0x3;
 
      if ( $address != 0 )
      {
         $fru1 = "DG" . $dimm_group . "_" . (($bus) ? "X" : "Y") .
                 "_BUS_" . (($LorR) ?  "R" : "L" );
      }
      else
      {
         $fru1 = "UNKNOWN (INVALID ADDRESS)";
      }

      $nasid = sprintf("0x%x,", int($NASID) );

      if ( $COUNT > 99 )
      {
         $count = sprintf("99+,", int($COUNT));
      }
      else
      {
         $count = sprintf("%d,", int($COUNT));
      }

      if ( $bus == 1 )
      {
         if ( reg_get_field_value ( $SHUB_MMR{'SH_MD_DQLP_MMR_XCERR1'}, VAL ) )
         {
            $grp = reg_get_field_value ( $SHUB_MMR{'SH_MD_DQLP_MMR_XCERR1'}, GRP1 );
         }
         elsif ( reg_get_field_value ( $SHUB_MMR{'SH_MD_DQLP_MMR_XCERR2'}, VAL) )
         {
            $grp = reg_get_field_value ( $SHUB_MMR{'SH_MD_DQLP_MMR_XCERR2'}, GRP2 );
         }
         else
         {
            $grp = 0;
         }
      }
      else
      {
         if ( reg_get_field_value ( $SHUB_MMR{'SH_MD_DQLP_MMR_YCERR1'}, VAL ) )
         {
            $grp = reg_get_field_value ( $SHUB_MMR{'SH_MD_DQLP_MMR_YCERR1'}, GRP1 );
         }
         elsif ( reg_get_field_value ( $SHUB_MMR{'SH_MD_DQLP_MMR_YCERR2'}, VAL))
         {
            $grp = reg_get_field_value ( $SHUB_MMR{'SH_MD_DQLP_MMR_YCERR2'}, GRP2 );
         }
         else
         {
            $grp = 0;
         }
      }

      $syn = gen_dir_syndrome($grp);

      $fru_string .= sprintf("%-13s ADDR:0x%013llx, DIR_SYN:0x%02x, CNT:%-4s NASID:%-6s FRU:%s\n",
         "DQLP_DIR_UC,", $address, $syn, $count, $nasid, $fru1 );
   }

   $summary =
      reg_get_field_value ( $SHUB_MMR{'SH_MEM_ERROR_SUMMARY'}, 'DQRP_DIR_UC' );

   if ( $summary )
   {
      $address = reg_get_field_value ( $SHUB_MMR{'SH_DIR_UC_ERR_HDR_LOWER'}, 'ADDR' );
      $address <<= 3;
      $bus = ( $address >> 7 ) & 0x1;
      $LorR = ( $address >> 11 ) & 0x1;
      $dimm_group = ( $address >> 34 ) & 0x3;

      if ( $address != 0 )
      {
         $fru1 = "DG" . $dimm_group . "_" . (($bus) ? "X" : "Y") .
                 "_BUS_" . (($LorR) ?  "R" : "L" );
      }
      else
      {
         $fru1 = "UNKNOWN (INVALID ADDRESS)";
      }

      $nasid = sprintf("0x%x,", int($NASID) );

      if ( $COUNT > 99 )
      {
         $count = sprintf("99+,", int($COUNT));
      }
      else
      {
         $count = sprintf("%d,", int($COUNT));
      }

      if ( $bus == 1 )
      {
         if ( reg_get_field_value ( $SHUB_MMR{'SH_MD_DQRP_MMR_XCERR1'}, VAL ) )
         {
            $grp = reg_get_field_value ( $SHUB_MMR{'SH_MD_DQRP_MMR_XCERR1'}, GRP1 );
         }
         elsif ( reg_get_field_value ( $SHUB_MMR{'SH_MD_DQRP_MMR_XCERR2'}, VAL))
         {
            $grp = reg_get_field_value ( $SHUB_MMR{'SH_MD_DQRP_MMR_XCERR2'}, GRP2 );
         }
         else
         {
            $grp = 0;
         }
      }
      else
      {
         if ( reg_get_field_value ( $SHUB_MMR{'SH_MD_DQRP_MMR_YCERR1'}, VAL ) )
         {
            $grp = reg_get_field_value ( $SHUB_MMR{'SH_MD_DQRP_MMR_YCERR1'}, GRP1 );
         }
         elsif ( reg_get_field_value ( $SHUB_MMR{'SH_MD_DQRP_MMR_YCERR2'}, VAL))
         {
            $grp = reg_get_field_value ( $SHUB_MMR{'SH_MD_DQRP_MMR_YCERR2'}, GRP2 );
         }
         else
         {
            $grp = 0;
         }
      }

      $syn = gen_dir_syndrome($grp);

      $fru_string .= sprintf("%-13s ADDR:0x%013llx, DIR_SYN:0x%02x, CNT:%-4s NASID:%-6s FRU:%s\n",
         "DQRP_DIR_UC,", $address, $syn, $count, $nasid, $fru1 );
   }

   $SMDK_INFO{$VNODE}{FRU_STRING} .= $fru_string;
   return $fru1_string;
} 
sub Decode_XNHW_NI {
   my $ni =  $_[0];
   my $tio = $_[1];
   my $header_l;
   my $header_u;
   my $fru_string;

   if ( $tio == 0 ) {
      if ( $ni == 0 ) {
         $header_l = reg_get_field_value ( $SHUB_MMR{'SH_NI0_ERROR_DETAIL_1'}, 'HEADER' );
         $header_u = reg_get_field_value ( $SHUB_MMR{'SH_NI0_ERROR_DETAIL_2'}, 'HEADER' );
      } else {
         $header_l = reg_get_field_value ( $SHUB_MMR{'SH_NI1_ERROR_DETAIL_1'}, 'HEADER' );
         $header_u = reg_get_field_value ( $SHUB_MMR{'SH_NI1_ERROR_DETAIL_2'}, 'HEADER' );
      }
   } else {
      if ( $ni == 0 ) {
         $header_l = reg_get_field_value ( $TIO_MMR{'TIO_NI0_ERROR_DETAIL_1'}, 'HEADER' );
         $header_u = reg_get_field_value ( $TIO_MMR{'TIO_NI0_ERROR_DETAIL_2'}, 'HEADER' );
      } else {
         $header_l = reg_get_field_value ( $TIO_MMR{'TIO_NI1_ERROR_DETAIL_1'}, 'HEADER' );
         $header_u = reg_get_field_value ( $TIO_MMR{'TIO_NI1_ERROR_DETAIL_2'}, 'HEADER' );
      }
   }

   $fru_string = decode_sn2net_header($header_l, $header_u);

   return $fru_string;
}

sub Decode_PIHW {
   my $fru_string;
   my $pi_first_error;

   $pi_first_error = reg_get_value($SHUB_MMR{'SH_PI_FIRST_ERROR'});

   $fru_string = decode_pi_err_detail ( "SH_PI_ERROR_DETAIL_1", 
                                        reg_get_value($SHUB_MMR{'SH_PI_ERROR_DETAIL_1'}), 
                                        $pi_first_error ); 
   chop($fru_string);
   $fru_string .= decode_pi_err_detail ( "SH_PI_ERROR_DETAIL_2", 
                                        reg_get_value($SHUB_MMR{'SH_PI_ERROR_DETAIL_2'}), 
                                        $pi_first_error ); 
   return $fru_string;
}

################################################################################
#
#  Sort the @time_list array, using the {TIME} key as the sort object
#
#
################################################################################
sub sort_times {

  my $chron_stamps = 0;                 # count chronologically ordered time stamps
  my $low_index = 0;                    # array index having the current low time value
  my $low_time = 0;                     # low time value for the inner for loop
  my $first_one = 1;                    # indicates this is the first unused entry 
  my $acquired = 0;                     # set this to prevent duplicate entries with the same time value

  print "\n\nThe following is a chronologically sorted list of valid time stamps:\n\n";
  print "    NASID    LOCATION \t\tREGISTER NAME \t\t\tTIME VALUE\n";
  print "    -----    -------- \t\t------------- \t\t\t----------\n";

  while ($chron_stamps < $t_stamps )  {
    for ($chron_loop = 0; $chron_loop < $t_stamps; $chron_loop++)  {

        # Always save the first unused entry

      if ((! defined $time_list[$chron_loop]->{USED}) && ($first_one == 1))  {
        $low_time = $time_list[$chron_loop]->{TIME};
        $low_index = $chron_loop;
        $first_one = 0;
        $acquired = 1;
        next;
      }
        # Save the entry if it's less than the previous low time value

      if ((! defined $time_list[$chron_loop]->{USED}) && ($time_list[$chron_loop]->{TIME} < $low_time)) {
        $low_time = $time_list[$chron_loop]->{TIME};
        $low_index = $chron_loop;
        $acquired = 1;
        next;
      }

        # Save the entry if it's equal to the low value seen on a previous iteration

      if ((! defined $time_list[$chron_loop]->{USED}) && ($time_list[$chron_loop]->{TIME} == $low_time) && (acquired == 0)) {
        $low_time = $time_list[$chron_loop]->{TIME};
        $low_index = $chron_loop;
        $acquired = 1;
        next;
      }
    } 

    printf "    0x%03X    %-9s       %-25s      0x%016llX\n", $time_list[$low_index]->{NASID}, $time_list[$low_index]->{LOCATION},
        $time_list[$low_index]->{REG}, $time_list[$low_index]->{TIME};

    $time_list[$low_index]->{USED} = 1;
    $chron_stamps++;
    $first_one = 1;
    $acquired = 0;
  }

  for ($aa = 0; $aa < $t_stamps; $aa++)  {
    if (! defined $time_list[$aa]->{USED})  {
      printf "\nALERT!  In sort_times routine - all time stamps not displayed in sorted list!\n";
      printf "    0x%03X    %-9s       %-25s      0x%016llX\n", $time_list[$aa]->{NASID}, $time_list[$aa]->{LOCATION},
          $time_list[$aa]->{REG}, $time_list[$aa]->{TIME};
    }
  }

}
