# 
#/**************************************************************************
# *                                                                        *
# *               Copyright (C) 1989-1999 Silicon Graphics, Inc.           *
# *                                                                        *
# *  These coded instructions, statements, and computer programs  contain  *
# *  unpublished  proprietary  information of Silicon Graphics, Inc., and  *
# *  are protected by Federal copyright law.  They  may  not be disclosed  *
# *  to  third  parties  or copied or duplicated in any form, in whole or  *
# *  in part, without the prior written consent of Silicon Graphics, Inc.  *
# *                                                                        *
# **************************************************************************/
# 
# @(#)$Id: error_regs.pm,v 1.4 2006/04/27 20:40:36 beighley Exp $
#-------------------------------------------------------------------------------
#
#  FILE: error_regs.pm
#
#  DESCRIPTION:  This file contains PERL subroutines to handle 
#
#  MODIFICATION HISTORY:
#
#     WHEN      WHO             WHAT
#     --------  --------------  -----------------------------------------
#     12/13/01  Karen Beighley  Original source code.
#     12/17/01  Karen Beighley  Minor changes; added dependency test.
#     10/14/02  Karen Beighley  Added use of block MMR read/write operations.
#                               Changed error handling to use common function.
#     11/20/02  Karen Beighley  Initial source code submission for release.
#     04/29/03  Karen Beighley  Added initial support for TIO and ORBIT chips.
#     07/19/04  Karen Beighley  Added initial support for SHUB2 chip.
#
#-------------------------------------------------------------------------------

use shub_subs;
use tio_subs;
use shub2_subs;


#-------------------------------------------------------------------------------
#  CONSTANTS - These values cannot change (other modules depend on them)
#-------------------------------------------------------------------------------

$EREG_FIRST_ERROR    = 0;
$EREG_DETAIL_1       = 1;
$EREG_DETAIL_2       = 2;
$EREG_DETAIL_3       = 3;
$EREG_DETAIL_4       = 4;
$EREG_DETAIL_5       = 5;
$EREG_SUMMARY        = 6;
$EREG_OVERFLOW       = 7;
$EREG_MASK           = 8;
$EREG_SUMMARY_ALIAS  = 9;
$EREG_OVERFLOW_ALIAS = 10;


#-------------------------------------------------------------------------------
#  GLOBAL VARIABLES
#-------------------------------------------------------------------------------

$EREG_ERROR_MASK            = 0x007FFFFF;
$EREG_ERROR_MASK_RP_BAD_CMD = 0x00000002;

$EREG_MAX_TIMEOUT_COUNT = 1000;

$ereg_addr_ref = \@shub_ereg_addr;  # Array of error register addresses (chip specific)


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: eregs_setup()
#
#    USAGE: $ref = eregs_setup()
#
#    DESCRIPTION: The subroutine will verify that a chip is targeted, and
#       initialize the error register address pointers for the target chip type.
#
#    RETURN VALUE: reference to the target error register address array
#
#-------------------------------------------------------------------------------

sub eregs_setup
{
   printf "\neregs_setup( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   #---------------------------------------
   #  Check number of MMR targets
   #---------------------------------------
   my $ntarget = mmr_target_get_num();
   if ( $ntarget != 1 )
   {  $emess = "A single chip must targeted before using the error registers.";
      $emess .= "\nCurrent number of targets = $ntarget";
      $emess .= "\nUse the \"mmr_target()\" subroutine to select a specific chip.";
      $ecode = -1;
      &$comm_err_fun($emess);
      return;
   }
   my $atype = mmr_target_list_get_detail(0, $DETAIL_IND_TYPE);

   #---------------------------------------
   #  Set the array pointers
   #---------------------------------------
   if ( $atype =~ /^shub2/i )
   {  $ereg_addr_ref = \@shub2_ereg_addr;
   }
   elsif ( $atype =~ /^shub/i )
   {  $ereg_addr_ref = \@shub_ereg_addr;
   }
   elsif ( $atype =~ /^orbit/i )
   {  $ereg_addr_ref = \@orbit_ereg_addr;
   }
   elsif ( $atype =~ /^tio/i )
   {  $ereg_addr_ref = \@tio_ereg_addr;
   }
   else
   {  $emess = "Targeted chip does not support packet generator operation.";
      $emess .= "\nTargeted chip is of type: $atype.";
      $ecode = -1;
      &$comm_err_fun($emess);
      return;
   }
   return( $ereg_addr_ref );
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: eregs_initialize()
#
#    USAGE: eregs_initialize()
#           eregs_initialize($error_mask)
#
#    DESCRIPTION: The subroutine will initialize the error reporting registers.
#       First, all errors are disabled (write 1's to ERROR_MASK), and then
#       the error detail, first error, error overflow and error summary MMRs
#       are cleared.  If the user provides an error mask value, that is
#       written to the ERROR_MASK register at the end of the sequence.
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub eregs_initialize
{  my $num = $#_ + 1;
   print "\neregs_initialize(@_) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   #
   #  Check the command-line parameters
   #
   if ( ($num != 0) && ($num != 1) )
   {  $emess = "Invalid parameters passed to subroutine eregs_initialize().";
      $ecode = -1;
      &$comm_err_fun($emess);
      return;
   }

   #
   #  Variables to the error reporting register (disable all errors)
   #
   my @addr = ( $$ereg_addr_ref[$EREG_MASK], 
                $$ereg_addr_ref[$EREG_DETAIL_1],
                $$ereg_addr_ref[$EREG_DETAIL_2],
                $$ereg_addr_ref[$EREG_DETAIL_3],
                $$ereg_addr_ref[$EREG_DETAIL_4],
                $$ereg_addr_ref[$EREG_DETAIL_5],
                $$ereg_addr_ref[$EREG_FIRST_ERROR],
                $$ereg_addr_ref[$EREG_OVERFLOW],
                $$ereg_addr_ref[$EREG_SUMMARY] );
   my @data = ( $EREG_ERROR_MASK, 0, 0, 0, 0, 0, 0, 0, 0 );

   #
   #  Add error mask enable if a value was provided to subroutine
   #
   if ( $num == 1 )
   {  push( @addr, $$ereg_addr_ref[$EREG_MASK] );
      push( @data, $_[0]                       );
#     push( @data, 0       ); # HACK - instead of value specified!!!!
   }

   #
   #  Initialize the error reporting register (disable all errors)
   #
   mmr_write_group(\@addr, \@data);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: eregs_read()
#
#    USAGE: eregs_read()
#
#    DESCRIPTION: The subroutine will read the MMRs associated with the
#       error reporting features.  The values read from the MMRs are returned
#       in an array; the order of the array is defined by the order of
#       the addresses (see below).
#
#    RETURN VALUE: Array containing values from error reporting MMRs.
#
#-------------------------------------------------------------------------------

sub eregs_read
{  my $num = $#_ + 1;
   print "\neregs_read(@_) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my @addr = ( $$ereg_addr_ref[$EREG_FIRST_ERROR],
                $$ereg_addr_ref[$EREG_DETAIL_1],
                $$ereg_addr_ref[$EREG_DETAIL_2],
                $$ereg_addr_ref[$EREG_DETAIL_3],
                $$ereg_addr_ref[$EREG_DETAIL_4],
                $$ereg_addr_ref[$EREG_DETAIL_5],
                $$ereg_addr_ref[$EREG_SUMMARY],
                $$ereg_addr_ref[$EREG_OVERFLOW],
                $$ereg_addr_ref[$EREG_MASK] );

   return( mmr_read_group(\@addr) );
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: eregs_clear_errors()
#
#    USAGE: eregs_clear_errors()
#           eregs_clear_errors($timeout_counter)
#
#    DESCRIPTION: The subroutine will clear the MMRs associated with the
#       error reporting features.  The error summary register is read, and
#       if it is non-zero, then the error MMRs are cleared.  After clearing
#       the error MMRs, the error summary register is re-read.  The sequence
#       is repeated until the error summary register is clear, or a timeout
#       condition is encountered.  An optional parameter "timeout_counter"
#       defines the number of times the read/clear loop is executed; the
#       default timeout value is defined by $EREG_MAX_TIMEOUT_COUNT.
#
#    RETURN VALUE: $TRUE if error registers were cleared; $FALSE if timeout.
#
#-------------------------------------------------------------------------------

sub eregs_clear_errors
{  my $num = $#_ + 1;
   print "\neregs_clear_errors(@_) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $max_count = $EREG_MAX_TIMEOUT_COUNT;

   #
   #  Check the command-line parameters
   #
   if ( $num == 1 )
   {  $max_count = $_[0];
   }
   elsif ( $num != 0 )
   {  $emess = "Invalid parameters passed to subroutine eregs_clear_errors().";
      $ecode = -1;
      &$comm_err_fun($emess);
      return;
   }
   if ($debug_flag & $DEBUG_GENERAL)
   {  print "max_count = $max_count\n";
   }

   my($summary, $oflow);
   my $count = 0;

   my @addr = ( $$ereg_addr_ref[$EREG_DETAIL_1],
                $$ereg_addr_ref[$EREG_DETAIL_2],
                $$ereg_addr_ref[$EREG_DETAIL_3],
                $$ereg_addr_ref[$EREG_DETAIL_4],
                $$ereg_addr_ref[$EREG_DETAIL_5],
                $$ereg_addr_ref[$EREG_FIRST_ERROR] );
   my @data = ( 0, 0, 0, 0, 0, 0 );

   my @sub_addr = ( $$ereg_addr_ref[$EREG_OVERFLOW_ALIAS], $$ereg_addr_ref[$EREG_SUMMARY_ALIAS] );
   my @sub_data = ( 0, 0);

   #
   #  Read the ERROR_SUMMARY register and attempt to clear all errors
   #
   $summary = mmr_read($$ereg_addr_ref[$EREG_SUMMARY]);
   return if ( $ecode < 0 ); # Return: short-circuit if error detected and returned
   while ( ($summary & $EREG_ERROR_MASK) != 0 )
   {  
      if ($debug_flag & $DEBUG_GENERAL)
      {  printf "clear errors: count=%-4u -> summary = 0x%016llX\n", $count, $summary;
      }

      #  NOTE: display timeout error and exit???
      return($FALSE) if ( $count > $max_count ) ;

      mmr_write_group(\@addr, \@data);
      return if ( $ecode < 0 );  # Return: short-circuit if error detected and returned

      $oflow = mmr_read( $$ereg_addr_ref[$EREG_OVERFLOW] );
      return if ( $ecode < 0 );  # Return: short-circuit if error detected and returned

      $sub_data[0] = $oflow;
      $sub_data[1] = $summary;
      mmr_write_group(\@sub_addr, \@sub_data);
      return if ( $ecode < 0 );  # Return: short-circuit if error detected and returned

      $summary = mmr_read($$ereg_addr_ref[$EREG_SUMMARY]);
      return if ( $ecode < 0 );  # Return: short-circuit if error detected and returned
      $count++;
   }

   if ($debug_flag & $DEBUG_GENERAL)
   {  printf "clear errors: count=%-4u -> summary = 0x%016llX\n", $count, $summary;
   }

   return($TRUE);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: eregs_wait_for_first_error()
#
#    USAGE: eregs_wait_for_first_error()
#           eregs_wait_for_first_error($timeout_counter)
#
#    DESCRIPTION: The subroutine will read the first error register until
#       an error is detected, or a timeout condition is encountered.
#       An optional parameter "timeout_counter"  defines the number of times 
#       the read/check loop is executed; the default timeout value is defined 
#       by $EREG_MAX_TIMEOUT_COUNT.
#
#    RETURN VALUE: $TRUE if error detected; $FALSE if timeout.
#
#-------------------------------------------------------------------------------

sub eregs_wait_for_first_error
{  my $num = $#_ + 1;
   print "\neregs_wait_for_first_error(@_) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $max_count = $EREG_MAX_TIMEOUT_COUNT;

   #
   #  Check the command-line parameters
   #
   if ( $num == 1 )
   {  $max_count = $_[0];
   }
   elsif ( $num != 0 )
   {  $emess = "Invalid parameters passed to subroutine eregs_wait_for_first_error().";
      $ecode = -1;
      &$comm_err_fun($emess);
      return;
   }
   if ( $debug_flag & $DEBUG_GENERAL )
   {  print "max_count = $max_count\n";
   }

   #
   #  Read the LB_FIRST_ERROR register: wait for error
   #
   my $count = 0;
   my $addr  = $$ereg_addr_ref[$EREG_FIRST_ERROR];
   my $data  = mmr_read($addr);
   return if ( $ecode < 0 ); # Return: short-circuit if error detected and returned
   while ( ($data & $EREG_ERROR_MASK) == 0 )
   {  
      if ( $debug_flag & $DEBUG_GENERAL )
      {  printf "wait for first error: count=%-4u -> data = 0x%016llX\n", $count, $data;
      }

      #  NOTE: display timeout error and exit???
      return($FALSE) if ( $count > $max_count ) ;

      $data  = mmr_read($addr);
      return if ( $ecode < 0 ); # Return: short-circuit if error detected and returned
      $count++;
   }
   if ( $debug_flag & $DEBUG_GENERAL )
   {  printf "wait for first error: count=%-4u -> data = 0x%016llX\n", $count, $data;
   }

   return($TRUE);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: eregs_wait_read_clear()
#
#    USAGE: eregs_wait_read_clear()
#           eregs_wait_read_clear($timeout_counter)
#
#    DESCRIPTION: The subroutine will wait for an error to be detected,
#       in the FIRST_ERROR register, and then read the error reporting registers
#       and clear the error registers.
#       The following subroutines are used: eregs_wait_for_first_error(),
#       eregs_read() and eregs_clear_errors().
#       An optional parameter "timeout_counter" is passed to the wait and
#       clear subroutines.  The default timeout_counter is defined by
#       $EREG_MAX_TIMEOUT_COUNT.
#
#    RETURN VALUE: An array containing the register values if no timeout.
#                  An empty array is returned if a timeout condition occurs.
#
#-------------------------------------------------------------------------------

sub eregs_wait_read_clear
{  my $num = $#_ + 1;
   print "\neregs_wait_read_clear(@_) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $max_count = $EREG_MAX_TIMEOUT_COUNT;

   #
   #  Check the command-line parameters
   #
   if ( $num == 1 )
   {  $max_count = $_[0];
   }
   elsif ( $num != 0 )
   {  $emess = "Invalid parameters passed to subroutine eregs_wait_read_clear().";
      $ecode = -1;
      &$comm_err_fun($emess);
      return;
   }
   if ($debug_flag & $DEBUG_GENERAL)
   {  print "max_count = $max_count\n";
   }

   my $rv;
   my @data = ();

   #
   #  Wait for the "first" error to be detected
   #
   $rv = eregs_wait_for_first_error($max_count);
   return if ( $ecode < 0 ); # Return: short-circuit if error detected and returned
   return( @data )  if ( $rv == $FALSE );

   #
   #  Read the error registers
   #
   @data = eregs_read();
   return if ( $ecode < 0 ); # Return: short-circuit if error detected and returned

   #
   #  Clear the error registers
   #
   $rv = eregs_clear_errors($max_count);
   return if ( $ecode < 0 ); # Return: short-circuit if error detected and returned
   @data = ()  if ( $rv == $FALSE );
   
   return( @data );
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: eregs_wait_for_summary_error()
#
#    USAGE: eregs_wait_for_summary_error($error_mask)
#           eregs_wait_for_summary_error($error_mask, $timeout_counter)
#
#    DESCRIPTION: The subroutine will read the error summary register until
#       a matching error is detected, or a timeout condition is encountered.
#       An optional parameter "timeout_counter"  defines the number of times 
#       the read/check loop is executed; the default timeout value is defined 
#       by $EREG_MAX_TIMEOUT_COUNT.
#
#    RETURN VALUE: $TRUE if error detected; $FALSE if timeout.
#
#-------------------------------------------------------------------------------

sub eregs_wait_for_summary_error
{  my $num = $#_ + 1;
   print "\neregs_wait_for_summary_error(@_) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $error_mask = 0;
   my $max_count  = $EREG_MAX_TIMEOUT_COUNT;

   #
   #  Check the command-line parameters
   #
   if ( $num == 2 )
   {  $error_mask = $_[0];
      $max_count  = $_[1];
   }
   elsif ( $num == 1 )
   {  $error_mask = $_[0];
   }
   elsif ( $num != 0 )
   {  $emess = "Invalid parameters passed to subroutine eregs_wait_for_summary_error().";
      $ecode = -1;
      &$comm_err_fun($emess);
      return;
   }
   if ( $debug_flag & $DEBUG_GENERAL )
   {  printf "error_mask = 0x%016lX\n", $error_mask;
      printf "max_count  = $max_count\n";
   }

   #
   #  Read the ERROR_SUMMARY register and wait for specific errors
   #
   my $count = 0;
   my $addr  = $$ereg_addr_ref[$EREG_SUMMARY];
   my $data  = mmr_read($addr);
   return if ( $ecode < 0 ); # Return: short-circuit if error detected and returned
   while ( ($data & $error_mask) == 0 )
   {  
      if ( $debug_flag & $DEBUG_GENERAL )
      {  printf "wait for summary error: count=%-4u -> data = 0x%016llX\n", $count, $data;
      }

      #  NOTE: display timeout error and exit???
      return($FALSE) if ( $count > $max_count ) ;

      $data  = mmr_read($addr);
      return if ( $ecode < 0 ); # Return: short-circuit if error detected and returned
      $count++;
   }
   if ( $debug_flag & $DEBUG_GENERAL )
   {  printf "wait for summary error: count=%-4u -> data = 0x%016llX\n", $count, $data;
   }

   return($TRUE);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: eregs_wait_summary_read_clear()
#
#    USAGE: eregs_wait_summary_read_clear($error_mask)
#           eregs_wait_summary_read_clear($error_mask, $timeout_counter)
#
#    DESCRIPTION: The subroutine will wait for an error to be detected,
#       in the ERROR_SUMMARY register, and then read the error reporting 
#       registers and clear the error registers.
#       The following subroutines are used: eregs_wait_for_summary_error(),
#       eregs_read() and eregs_clear_errors().
#       An optional parameter "timeout_counter" is passed to the wait and
#       clear subroutines.  The default timeout_counter is defined by
#       $EREG_MAX_TIMEOUT_COUNT.
#
#    RETURN VALUE: An array containing the register values if no timeout.
#                  An empty array is returned if a timeout condition occurs.
#
#-------------------------------------------------------------------------------

sub eregs_wait_summary_read_clear
{  my $num = $#_ + 1;
   print "\neregs_wait_summary_read_clear(@_) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $error_mask = 0;
   my $max_count  = $EREG_MAX_TIMEOUT_COUNT;

   #
   #  Check the command-line parameters
   #
   if ( $num == 2 )
   {  $error_mask = $_[0];
      $max_count  = $_[1];
   }
   elsif ( $num == 1 )
   {  $error_mask = $_[0];
   }
   elsif ( $num != 0 )
   {  $emess = "Invalid parameters passed to subroutine eregs_wait_summary_read_clear().";
      $ecode = -1;
      &$comm_err_fun($emess);
      return;
   }
   if ( $debug_flag & $DEBUG_GENERAL )
   {  printf "error_mask = 0x%016lX\n", $error_mask;
      printf "max_count  = $max_count\n";
   }

   my $rv;
   my @data = ();

   #
   #  Wait for the error to be detected in the summary register
   #
   $rv = eregs_wait_for_summary_error($error_mask, $max_count);
   return if ( $ecode < 0 ); # Return: short-circuit if error detected and returned
   return( @data )  if ( $rv == $FALSE );

   #
   #  Read the error registers
   #
   @data = eregs_read();
   return if ( $ecode < 0 ); # Return: short-circuit if error detected and returned

   #
   #  Clear the error registers
   #
   $rv = eregs_clear_errors($max_count);
   return if ( $ecode < 0 ); # Return: short-circuit if error detected and returned
   @data = ()  if ( $rv == $FALSE );
   
   return( @data );
}


#-------------------------------------------------------------------------------

1;
