#
#/**************************************************************************
# *                                                                        *
# *               Copyright (C) 2002-2005 Silicon Graphics, Inc.           *
# *                                                                        *
# *  These coded instructions, statements, and computer programs  contain  *
# *  unpublished  proprietary  information of Silicon Graphics, Inc., and  *
# *  are protected by Federal copyright law.  They  may  not be disclosed  *
# *  to  third  parties  or copied or duplicated in any form, in whole or  *
# *  in part, without the prior written consent of Silicon Graphics, Inc.  *
# *                                                                        *
# **************************************************************************/
#
# @(#)$Id: fru_funcs.pm,v 1.28 2007/02/07 18:39:33 meyer Exp $
#
#  FILE: fru_funcs.pm
#
#  DESCRIPTION:  This file contains PERL routines to do fru specific functions.
#
#  MODIFICATION HISTORY:
#
#     WHEN      WHO             WHAT
#     --------  --------------  -----------------------------------------
#     05/19/06  Gary Meyer      Original source code submitted for release.
#     05/19/06  Gary Meyer      Fixed bug with parsing NL4R config info.
#     06/26/06  Gary Meyer      Fleshed out routines within print_l1_info() to parse L1 status,
#				config info, and errors from the fru dump file.
#     06/28/06  Gary Meyer	Updates for NL4R error detail registers in fru file.
#				Fix for TIO & NL4R error summary problem in file mode.
#     06/30/06  Gary Meyer	Updated to also check for "Directory" in fru reason provided
#			  	from error detail info.
#     07/03/06  Gary Meyer	Added provisions to check for Speculative Nasid assignments.
#     07/19/06  Gary Meyer	Fix for possible endless loop case in display_fru_list if a 
#				non-matching SPEC_NASID value is encountered.
#     08/10/06  Gary Meyer	Fix for problem with LED error indication output.
#     08/25/06  Gary Meyer	Added toggle check for leds output found in the fru file.
#     08/28/06  Gary Meyer	Fix for a case when the led "not toggling" indication was on the
#				line after the cpu d.  Now listing which CPU isn't toggling.
#     08/29/06  Gary Meyer	Now capturing power data from the fru file.
#     09/14/06  Gary Meyer	Fixed formatting issue with LED output when status errors were found.
#     09/19/06  Gary Meyer	Update to display_fru_list() to prevent a corrected error that happens 
#                               very early from throwing off the fru score for a first-occurring special
#                               case HW error.
#     10/18/06  Gary Meyer	Update to generate Shub2 test point output and Shub2 PI CRB output from
#				fru dump file.
#     10/26/06  Gary Meyer	Fix for problem with incorrect PI CRB values from fru file. 
#     10/29/06  Gary Meyer	Change to limit to 7 the number of consecutive FRU entries implicating the
#                               same HW component to eliminate fru summaries with excessive redundant entries.
#     11/01/06  Gary Meyer	Fix and simplification for calculation of $total_same_dev.
#     12/10/06  Gary Meyer	Report instances of OPERATION_ERROR.
#     12/12/06  Gary Meyer	Formatting changes for correct alignment of TIO fru info in the summary.
#     12/19/06  Gary Meyer	Changes to ensure correct fru scoring for Tollhouse IRU routers. 
#				Corrected problem printing fru capture error info.
#     12/20/06  Gary Meyer	Fixed scoring issue seen when multiple LLP retry errors are seen.
#     12/23/06  Gary Meyer	Added save_fru function.
#     01-22-07  Gary Meyer      Ignore FSB testpoint errors if no processors are present on the node.
#     02-06-07  Gary Meyer      Updated redundant fru entry checks to prevent identical consecutive entries.
#     02-07-07  Gary Meyer      Fixed hole in FRU scoring for a first-occurring LLP retry timeout.
#
#-------------------------------------------------------------------------------




#--------------------------------------------------------------------------------
#
#  The main fru_file function - this is the first function jumped to from errdmp.
#
#--------------------------------------------------------------------------------

sub parse_fru_file {

  if ($show_clock == 1)  {
    chomp ($done_time = `date +'%T'`);
    print "\n errdmp status: fru file parsing begins at $done_time.\n";
  }
  get_cfg();
  if ($show_clock == 1)  {
    chomp ($done_time = `date +'%T'`);
    print "\n errdmp status: fru file get_cfg complete at $done_time.\n";
  }
  gather_cpu_cfg();
  if ($show_clock == 1)  {
    chomp ($done_time = `date +'%T'`);
    print "\n errdmp status: fru file gather_cpu_cfg complete at $done_time.\n";
  }
  parse_mmr_data();		
  if ($show_clock == 1)  {
    chomp ($done_time = `date +'%T'`);
    print "\n errdmp status: fru file parse_mmr_data complete at $done_time.\n";
  }
  output_summary();     	#  Print error summary for all nodes
  if ($show_clock == 1)  {
    chomp ($done_time = `date +'%T'`);
    print "\n errdmp status: output_summary complete at $done_time.\n";
  }
  display_sh2_mem_cfg();
  if ($show_clock == 1)  {
    chomp ($done_time = `date +'%T'`);
    print "\n errdmp status: display_sh2_mem_cfg complete at $done_time.\n";
  }
  display_prom_cfg();
  if ($show_clock == 1)  {
    chomp ($done_time = `date +'%T'`);
    print "\n errdmp status: display_prom_cfg complete at $done_time.\n";
  }
  print_l1_info();	
  if ($show_clock == 1)  {
    chomp ($done_time = `date +'%T'`);
    print "\n errdmp status: print_l1_info complete at $done_time.\n";
  }
  execution_done($global_err_val);

}


#--------------------------------------------------------------------------------
#
#  Basic fru_file parsing to get the number of Shub2, TIO, and NL4R chips,
#  and their locations.
#
#--------------------------------------------------------------------------------

sub get_cfg {

  @th_chassis_list = ();
  $th_iru_count = 0;
  my @rack_list = ();
  my $rack_count = 0;
  $fru_op_error = 0;
  $fru_op_error_msg = "   Errors found in the FRU file for following IRU's: \n";

  open (FRU_STAT, "$fru_file");
  while (<FRU_STAT>) {
    if (/ID: (\d\d\d)\.(\d\d) (\w+)/) {
      my $rack_val = $1;
      my $slot_val = $2;
      my $type_val = "$3";
      my $type_des = ($type_val =~ /19/) ? 'c' : 'r';		# should cover 1932, 1955, 191010
      $th_chassis_list[$th_iru_count]->{LOC} = "$rack_val"."$type_des"."$slot_val" ;
      $th_chassis_list[$th_iru_count]->{TYPE} = $type_val;
      $th_chassis_list[$th_iru_count]->{RACK} = $rack_val;
      $th_chassis_list[$th_iru_count]->{SLOT} = $slot_val;
      $th_iru_count++;
      if ($rack_count == 0) {
        push(@rack_list, $rack_val);
        $rack_count++;
      }
      else {
        my $rack_exists = 0;
        foreach $rack_entry (@rack_list) {
          if ($rack_entry == $rack_val) {
            $rack_exists = 1;
            last;
          }
        }
        if ($rack_exists == 0) {
          push(@rack_list, $rack_val);
          $rack_count++;
        }
      }
    }
	# OPERATION_ERROR::001.01:no response:IRouter:timeout waiting for packet

    if (/OPERATION_ERROR::(\d\d\d)\.(\d\d):(\w+)/) {
      my $rack_val = $1;
      my $slot_val = $2;
      my $err_dtl = $_;
      $fru_op_error_msg .= sprintf "\n  ALERT!:  OPERATION_ERROR detected on IRU at rack %d, slot %d : \n%s\n", 
	$rack_val, $slot_val, $err_dtl;
      $fru_op_error = 1;
    }

        # ERROR::004.01:11:1:error during capture # 191010 Chassis RTRR DUAL_ROUTER NL4R 1

    if (/ ERROR::(\d\d\d)\.(\d\d):(\w+)/) {
      my $rack_val = $1;
      my $slot_val = $2;
      my $err_dtl = $_;
      $fru_op_error_msg .= sprintf "\n  ALERT!:  ERROR detected on IRU at rack %d, slot %d : \n%s\n", 
	$rack_val, $slot_val, $err_dtl;
      $fru_op_error = 1;
    }

    if (/END_FRU_STATUS/) {
      last;
    }
  }
  close FRU_STAT;

	#  191010 :  
  	#  BEGIN_NL4R_NODE::001.01:10:0 # RTRL DUAL_ROUTER NL4R 0
  	#  BEGIN_NL4R_NODE::001.01:10:1 # RTRL DUAL_ROUTER NL4R 1
  	#  BEGIN_NL4R_NODE::001.01:11:0 # RTRR DUAL_ROUTER NL4R 0
  	#  BEGIN_NL4R_NODE::001.01:11:1 # RTRR DUAL_ROUTER NL4R 1
	#
	#  1955 :
	#  BEGIN_NL4R_NODE::001.01:5:0
	#  BEGIN_NL4R_NODE::001.01:6:0
	#
	#  QDR :
	#  BEGIN_NL4R_NODE::001.01:0:0
	#  BEGIN_NL4R_NODE::001.01:1:0
	#  BEGIN_NL4R_NODE::001.01:2:0
	#  BEGIN_NL4R_NODE::001.01:3:0


  open (NL4R_CHIPS, "grep BEGIN_NL4R_NODE $fru_file |");

  while (<NL4R_CHIPS>) {
    if (/BEGIN_NL4R_NODE::(\d\d\d)\.(\d\d):(\d\d):(\d)/) {
      $nl4r_target_list[$nl4r_node_count]->{MODULE} = "$1"."c"."$2";
      my $chip_val = "$4";
      my $side_val = ($3 =~ /10/) ? 'L' : 'R';
      $nl4r_target_list[$nl4r_node_count]->{NODE} = "RTR_"."$side_val";
      $nl4r_target_list[$nl4r_node_count]->{CHIP} = "$chip_val";
      $nl4r_target_list[$nl4r_node_count]->{CHASSIS} = 2;		# 191010
      printf("ADD NL4R: brick=%s  board=%s chip=%s  type=%d\n",
	$nl4r_target_list[$nl4r_node_count]->{MODULE}, $nl4r_target_list[$nl4r_node_count]->{NODE},
	$nl4r_target_list[$nl4r_node_count]->{CHIP}, $nl4r_target_list[$nl4r_node_count]->{CHASSIS}) if ($verbose_mode == 1);
      $nl4r_node_count++;
    }
    elsif (/BEGIN_NL4R_NODE::(\d\d\d)\.(\d\d):([5|6]):(\d)/) {
      $nl4r_target_list[$nl4r_node_count]->{MODULE} = "$1"."c"."$2";
      my $chip_val = "$4";
      my $side_val = ($3 =~ /5/) ? 'L' : 'R';
      $nl4r_target_list[$nl4r_node_count]->{NODE} = "RTR_"."$side_val";
      $nl4r_target_list[$nl4r_node_count]->{CHIP} = "$chip_val";
      $nl4r_target_list[$nl4r_node_count]->{CHASSIS} = 1;		# 1955
      printf("ADD NL4R: brick=%s  board=%s chip=%s  type=%d\n",
	$nl4r_target_list[$nl4r_node_count]->{MODULE}, $nl4r_target_list[$nl4r_node_count]->{NODE},
	$nl4r_target_list[$nl4r_node_count]->{CHIP}, $nl4r_target_list[$nl4r_node_count]->{CHASSIS}) if ($verbose_mode == 1);
      $nl4r_node_count++;
    }
    elsif (/BEGIN_NL4R_NODE::(\d\d\d)\.(\d\d):([0|1|2|3]):(\d)/) {
      $nl4r_target_list[$nl4r_node_count]->{MODULE} = "$1"."r"."$2";
      $nl4r_target_list[$nl4r_node_count]->{NODE} = "RTR"."$3";
      $nl4r_target_list[$nl4r_node_count]->{CHIP} = "$4";
      $nl4r_target_list[$nl4r_node_count]->{CHASSIS} = 0;		# Quad Dense Router
      printf("ADD NL4R: brick=%s  board=%s chip=%s  type=%d\n",
	$nl4r_target_list[$nl4r_node_count]->{MODULE}, $nl4r_target_list[$nl4r_node_count]->{NODE},
	$nl4r_target_list[$nl4r_node_count]->{CHIP}, $nl4r_target_list[$nl4r_node_count]->{CHASSIS}) if ($verbose_mode == 1);
      $nl4r_node_count++;
    }
  }
  close NL4R_CHIPS;


  open (SH2_CHIPS, "grep BEGIN_SHUB2_NODE $fru_file |");
  while (<SH2_CHIPS>) {
    if (/BEGIN_SHUB2_NODE::(\d\d\d)\.(\d\d):(\d).*(IP\w*)/) {
      $sh2_target_list[$sh2_node_count]->{MODULE} = "$1"."c"."$2";
      $sh2_target_list[$sh2_node_count]->{NODE} = "BLADE"."$3";
      $sh2_target_list[$sh2_node_count]->{CHIP} = "SHUB0";
      $sh2_target_list[$sh2_node_count]->{BLADE_TYPE} = "$4";
      $sh2_target_list[$sh2_node_count]->{CHASSIS} = get_chassis_type($sh2_target_list[$sh2_node_count]->{MODULE});
      printf("ADD SH2 SHUB0: brick=%s  board=%s chip=%s  type=%s\n",
	$sh2_target_list[$sh2_node_count]->{MODULE}, $sh2_target_list[$sh2_node_count]->{NODE},
	$sh2_target_list[$sh2_node_count]->{CHIP}, $sh2_target_list[$sh2_node_count]->{BLADE_TYPE})  if ($verbose_mode == 1);
      $sh2_node_count++;
    }
  }
  close SH2_CHIPS;

  open (TIO_CHIPS, "grep BEGIN_TIO_NODE $fru_file |");
  while (<TIO_CHIPS>) {
    if (/BEGIN_TIO_NODE::(\d\d\d)\.(\d\d):(\d):(\d) # B\d (\w+) TIO/) {
      $tio_target_list[$tio_node_count]->{MODULE} = "$1"."c"."$2";
      $tio_target_list[$tio_node_count]->{NODE} = "BLADE"."$3";
      $tio_target_list[$tio_node_count]->{CHIP} = "TIO_"."$4";
      $tio_target_list[$tio_node_count]->{CHASSIS} = get_chassis_type($tio_target_list[$tio_node_count]->{MODULE});
      printf("ADD TIO: brick=%s  board=%s chip=%s\n",
	$tio_target_list[$tio_node_count]->{MODULE}, $tio_target_list[$tio_node_count]->{NODE},
	$tio_target_list[$tio_node_count]->{CHIP}) if ($verbose_mode == 1);
      $tio_node_count++;
    }
  }
  close TIO_CHIPS;

  $config_summ = sprintf "\n\n=========================================================\n";
  $config_summ .= sprintf "\nSystem configuration summary extracted from fru dump file:\n\n";

  if ($sh2_node_count > 0) {
    $config_summ .= sprintf "     Shub 2x chips: %4d\n", $sh2_node_count;
  }
  if ($tio_node_count > 0) {
    $config_summ .= sprintf "         TIO chips: %4d\n", $tio_node_count;
    $tio_present = 1;
  }
  if ($nl4r_node_count > 0) {
    $config_summ .= sprintf "        NL4R chips: %4d\n", $nl4r_node_count;
    $nl4r_present = 1;
  }
  if ($th_iru_count > 0) {
    $config_summ .= sprintf "    Tollhouse IRUs: %4d\n", $th_iru_count;
  }
  $config_summ .=   sprintf "      System racks: %4d\n", $rack_count;
  $config_summ .= sprintf "\n\n------------------------------------------------------------------\n\n";

  if ($fru_op_error == 1) {
    $config_summ .= "$fru_op_error_msg\n";
  }
  print "$config_summ";


}


#--------------------------------------------------------------------------------
#
#  This subroutine parses the fru file for the L1 "serial all" output. 
#  The processor specific info is used to generate the processor config.
#
#--------------------------------------------------------------------------------

sub gather_cpu_cfg {

  unless (open FRU_FILE, "$fru_file") {
    die "\n\n ERROR!  Unable to open $fru_file for reading!  \n$!\n\n";
  }

  my $cpu_serial_string = "\n  Begin CPU info extracted from $fru_file\n";
  my $save_line = 0;
  
  my $tmp_string = "";

  while (<FRU_FILE>) {
    if ($save_line == 1) {
      $tmp_string .= $_;
      if (/BLADE ID/) {
        $cpu_serial_string .= $tmp_string;
        $save_line = 2;
      }
    }
    elsif ($save_line == 2) {
      $cpu_serial_string .= $_;
    }

    if (/BEGIN_SERIAL_DATA::(\d\d\d)\.(\d\d)/) {
      $tmp_string = "\n$_\n";
      $tmp_string .= "\n$1"."c"."$2\n";
      $save_line = 1;
    }
    elsif (/END_SERIAL_DATA/) {
      $save_line = 0;
    }
    elsif ((/BEGIN_SHUB_MMR_DATA/) || (/BEGIN_NL4R_MMR_DATA/) || (/BEGIN_TIO_MMR_DATA/)) {
      $save_line = 0;
      last;
    }
    elsif (/JEDEC/) {
      $save_line = 0;
    }

  }

  close FRU_FILE;
  parse_serial_info($cpu_serial_string); 	# in sys_ctrl_funcs.pm

}


#--------------------------------------------------------------------------------
#
#  parse through fru dump file and print the l1 env data to the errdmp log file.
#
#  need to parse:
#  	link		(done)
#  	leds		(done)
#  	cfg p		(done)
#	env		(done)
#	serial all
#
#--------------------------------------------------------------------------------

sub print_l1_info {

  unless (open FRU_FILE, "$fru_file") {
    die "\n\n ERROR!  Unable to open $fru_file for reading!  \n$!\n\n";
  }
  
  my $capture_link_info = 0;
  my $capture_leds_info = 0;
  my $capture_cfgp_info = 0;
  my $capture_pwr_info = 0;
  my $capture_env_info = 0;
  my $capture_serial_info = 0;

  my $current_brick = 0;
  my $current_blade = "";
  my $prev_blade = "";
  my $current_cpu = "";
  my $prev_cpu = "";
  my $prev_led_stat = 0;
  my $link_error_info = "";
  my $leds_error_info = "";
  my $link_error_found = 0;
  my $leds_error_found = 0;
  my $link_err_cnt = 0;
  my $leds_err_cnt = 0;
  my $led_entry = 0;
  my $toggle_msg = "";

  my $leds_output_string = "\n leds status captured by the L1 controller: \n";
  my $leds_error_string = "\n";		# accumulates all leds errors detected
  my $cfgp_output_string = "\n Port Connection Data captured by the L1 controller  \( cfg p \): \n";
  my $pwr_output_string = "\n Power Status Data captured by the L1 controller: \n";
  my $env_output_string = "\n Environmental Status Data captured by the L1 controller: \n";
  my $serial_output_string = "\n System Serial Information captured by the L1 controller: \n";

  while (<FRU_FILE>) {

	####   Begin Link parsing section  ####

    if (/BEGIN_LINK_DATA::(\d\d\d)\.(\d\d)/) {
      my $link_iru_rack = $1;
      my $link_iru_slot = $2; 
      $capture_link_info = 1;
      $current_brick = get_brick_loc($link_iru_slot, $link_iru_rack);
      $link_error_info = "  The following Numa link errors were found on $current_brick :\n";
    }
    if ((/ERROR:/) && ($capture_link_info == 1)) {
      $link_error_info .= "\t$_";
      $link_error_found = 1;
    }
    if (/END_LINK_DATA/) {
      if ($link_error_found == 1) {
        $l1_env_err = 1;
        $l1_env_err_msg .= " $link_error_info \n";

        $fru_summary .= sprintf "\tAlert:  $link_error_info \n";
        $fru_report_ready++;
        $global_err_val |= $L1_LINK_ERR;
	$link_err_cnt++;
      }
      $capture_link_info = 0;
      $link_error_found = 0;
    }

	####   Begin LEDS parsing section  ####

    if (/BEGIN_LEDS_STATUS::(\d\d\d)\.(\d\d)/) {
      my $leds_iru_rack = $1;
      my $leds_iru_slot = $2; 
      $capture_leds_info = 1;
      $led_entry = 0;
      $current_brick = get_brick_loc($leds_iru_slot, $leds_iru_rack);
      if ($current_brick =~ /c/) {		# is a compute brick
        $leds_error_info = "  The following leds status errors were found on $current_brick :\n";
	$leds_output_string .= "\n$current_brick : \n";
      }
      next;
    }
    if ($capture_leds_info == 1) {
      if (/END_LEDS_DATA/) {
        if ($leds_error_found == 1) {
          $l1_env_err = 1;
          $leds_error_string .= "$leds_error_info \n";
          $leds_err_cnt++;
        }
        $capture_leds_info = 0;
        $leds_error_found = 0;
	if ($led_entry == 1) {
	  $no_toggle_count++;
	  $l1_env_err = 1;
	  $leds_output_string .= "     <<<<<<<  Not toggling!";
	  $toggle_msg .= "  \n Found CPU LEDS not toggling on $current_brick, $prev_blade\n";
	}
	$leds_output_string .= "\n";
      }
      else {
	chomp($_);
	my $new_leds_line = sprintf "%-45s", $_;
	my $leds_good = check_leds_line($new_leds_line);	#  check if this led value indicates an error
	if ($leds_good == 0) {
	  $leds_error_info .= "        $new_leds_line\n";
	  $new_leds_line .= "      <<<<<<<";
	  $leds_error_found = 1;
	  if ($led_entry == 1)  {		# previous led entry wasn't toggling
	    $leds_output_string .= "      <<<<<<<  Not toggling!\n";
	  }
	  if ($prev_led_stat > 0) {
	    $leds_output_string .= "\n";
	  }
	  $led_entry = 2;
	  $prev_led_stat = 1;
	}
	elsif ($new_leds_line =~/(B[0-9]) CPU  ([A-D])/)  {
	  $prev_led_stat = 0;
	  $current_blade = $1;
	  $current_cpu = $2;
	  $current_blade =~ s/B/BLADE/;

 	    ###### cleanup from the previous line ########	

	  if ($led_entry == 1) {
	    $no_toggle_count++;
	    $l1_env_err = 1;
	    $leds_output_string .= "      <<<<<<<  Not toggling!\n";
	    $toggle_msg .= "  \n Found CPU $prev_cpu LEDS not toggling on $current_brick, $prev_blade\n";
	  }
	  else {
 	    $leds_output_string .= "\n";
	  }
	  $prev_blade = $current_blade;
	  $prev_cpu = $current_cpu;

	    ###### begin processing the new line  ########

	  if (($new_leds_line =~/slave loop/) || ($new_leds_line =~/CPU not present/)) {
            $led_entry = 2;
          }
	  else {
            $led_entry = 1;
          }
        }
        elsif ($new_leds_line =~ /0x[\da-fA-F][\da-fA-F]:/) {
	  $prev_led_stat = 0;
	  $led_entry++;
 	  $leds_output_string .= "\n";
	}
	else {
	  $prev_led_stat = 2;
	  if ($led_entry == 1) {
	    $no_toggle_count++;
	    $l1_env_err = 1;
	    $leds_output_string .= "      <<<<<<<  Not toggling!\n";
	    $toggle_msg .= "  \n Found CPU $prev_cpu LEDS not toggling on $current_brick, $prev_blade\n";
	    $led_entry = 0;
	  }
	  else {
 	    $leds_output_string .= "\n";
	  }
	}
 	$leds_output_string .= "$new_leds_line";
      }
    }

	####   Begin cfg p capture section  ####

    if (/BEGIN_CFGPORT_DATA::(\d\d\d)\.(\d\d)/) {
      my $cfgp_iru_rack = $1;
      my $cfgp_iru_slot = $2; 
      $capture_cfgp_info = 1;
      $current_brick = get_brick_loc($cfgp_iru_slot, $cfgp_iru_rack);
      $cfgp_output_string .= "\n$current_brick : \n";
      next;
    }
    if ($capture_cfgp_info == 1) {
      if (/END_CFGPORT_DATA/) {
        $capture_cfgp_info = 0;
        $cfgp_output_string .= "\n";
      } 
      else {
	$cfgp_output_string .= "$_";
      } 
    }
	
        ####   Begin env capture section  ####

    if (/BEGIN_ENV_DATA::(\d\d\d)\.(\d\d)/) {
      my $env_iru_rack = $1;
      my $env_iru_slot = $2;
      $capture_env_info = 1;
      $current_brick = get_brick_loc($env_iru_slot, $env_iru_rack);
      $env_output_string .= "\n$current_brick : \n";
      next;
    }
    if ($capture_env_info == 1) {
      if (/END_ENV_DATA/) {
        $capture_env_info = 0;
        $env_output_string .= "\n";
      }
      else {
        $env_output_string .= "$_";
      }
    }
    

        ####   Begin pwr status capture section  ####

    if (/BEGIN_POWER_DATA::(\d\d\d)\.(\d\d)/) {
      my $pwr_iru_rack = $1;
      my $pwr_iru_slot = $2;
      $capture_pwr_info = 1;
      $current_brick = get_brick_loc($pwr_iru_slot, $pwr_iru_rack);
      $pwr_output_string .= "\n$current_brick : \n";
      next;
    }
    if ($capture_pwr_info == 1) {
      if (/END_POWER_DATA/) {
        $capture_pwr_info = 0;
        $pwr_output_string .= "\n";
      }
      else {
        $pwr_output_string .= "$_";
      }
    }


        ####   Begin serial info capture section  ####

    if (/BEGIN_SERIAL_DATA::(\d\d\d)\.(\d\d)/) {
      my $serial_iru_rack = $1;
      my $serial_iru_slot = $2;
      $capture_serial_info = 1;
      $current_brick = get_brick_loc($serial_iru_slot, $serial_iru_rack);
      $serial_output_string .= "\n$current_brick : \n";
      next;
    }
    if ($capture_serial_info == 1) {
      if (/END_SERIAL_DATA/) {
        $capture_serial_info = 0;
        $serial_output_string .= "\n";
      }
      else {
        $serial_output_string .= "$_";
      }
    }

  }

  close FRU_FILE;

  if ($leds_err_cnt > 0) {
    $l1_env_err_msg .= "$leds_error_string"; 
  }

  if ($no_toggle_count > 0)  {
    $l1_env_err_msg .= "  Warning!  Observed LED status on $no_toggle_count cpu(s) not toggling.\n\n";
    if ($no_toggle_count < 5) {
      $fru_summary .= sprintf "\tAlert:  Observed LED status on $no_toggle_count cpu(s) not toggling.\n\n";
      $fru_summary .= "$toggle_msg";
      $l1_env_err_msg .= "$toggle_msg";
      $fru_report_ready++;
    }
  }

  if ($l1_env_err == 1)  {
    print "\n\n===========================================================================\n";
    print "\nFinal L1 Error Summary\n";
    print "$l1_env_err_msg\n";
    print "\n---------------------------------------------------------------------------\n";
  }

  if ($link_err_cnt == 0) {
    print "\n\n  There were no Numa link errors detected by the L1.\n\n";
  }
  else {
    print "\n\n  Numa link errors detected by the L1 and saved to L1 Error Summary.\n\n";
  }

  if ($leds_err_cnt == 0) {
    print "  There were no LED Status errors detected.\n\n"; 
  }

  print "\n$leds_output_string";
  print "\n$cfgp_output_string";
  print "\n$pwr_output_string";
  print "\n$env_output_string";
  print "\n$serial_output_string";
  print "\n\n";

}


#--------------------------------------------------------------------------------
#
#  parse through fru dump file and print chip specific MMR data
#
#--------------------------------------------------------------------------------

sub parse_mmr_data {

  unless (open FRU_FILE, "$fru_file") {
    die "\n\n ERROR!  Unable to open $fru_file for reading!  \n$!\n\n";
  }

  my $save_sh2 = 0;
  my $shub_tio = 0;
  my $save_nl4r = 0;

  my @sh2_reg_list = ();
  my @tio_reg_list = ();
  my @nl4r_reg_list = ();
  my @tp00_list = ();
  my @tp01_list = ();
  my $r = 0;
  $fru_rrb_cnt = 0;
  $fru_wrb_cnt = 0;
  $fru_irb_cnt = 0;
  $fru_prb_cnt = 0;

  while (<FRU_FILE>) {
    if (/BEGIN_SHUB2_NODE::(\d\d\d)\.(\d\d):(\d)/) {
      $save_sh2 = 1;
      $r = 0;
      @sh2_reg_list = ();
      next;
    }
    if ($save_sh2 > 0) {
      if (/END_SHUB2_NODE/) {
	$sh2_cnt++;
        $save_sh2 = 0;
      }
      elsif ((/(SH2_\w+) (0x\w+)/) && ($save_sh2 == 1)) {
        my $reg_name = $1;
        my $reg_data = oct $2;
        my $reg_ptr = reg_create_mmr_inst("shub2_01", $reg_name);
        reg_set_value ($reg_ptr, $reg_data);
        if ( !reg_is_valid($reg_ptr) )
        {  $emess = "Unable to create SHUB2 register instance: $reg_name.\n";
          dsp_error_and_exit($emess);
        }
        $sh2_reg_list[$r] = $reg_ptr;
	# printf " Entry: %03d  Name: %s   Value: 0x%16llX\n", $r, $reg_name, $reg_data;
        $r++;
      }
      elsif (/END_MMR_DATA/) {
        parse_sh2_mmr_list(\@sh2_reg_list);		# in shub2_funcs.pm
	$save_sh2 = 2;
	@tp00_list = ();
	@tp01_list = ();
      }
      elsif ((/(SH2_PI_DEBUG_\w+) (0x\w+)/) && ($save_sh2 == 2)) {
        my $reg_data = oct $2;
        push(@tp00_list, $reg_data);
      }
      elsif ((/END_TP_DATA_00/) && ($#tp00_list > 0) && ($save_sh2 == 2))  {
        my $tp_chk_val = parse_sh2_tp00_list(\@tp00_list);		# in pi_tools.pm
        if (($tp_chk_val & 2) == 2) {
          $global_err_val |= $SHUB_HW_ERR;
          $fru_list[$fru_count]->{NASID} = $nasid_value;		# global variable assigned in parse_shub2_node_id()
          $fru_list[$fru_count]->{REASON} = "PI is Hung due to Sustained snoop conflict.";
          $fru_list[$fru_count]->{DEVICE}   = "SHUB2";
          $fru_list[$fru_count]->{LOCATION} = $current_location;        # global variable assigned in parse_sh2_mmr_list()
          printf "\n%s: Added SHUB 2x NASID 0x%04X to the FRU list, entry %d.\n", $prefix, $nasid_value, $fru_count;
          $fru_count++;
        }
	$save_sh2 = 3;
      }
      elsif ((/(SH2_PI_DEBUG_\w+) (0x\w+)/) && ($save_sh2 == 3)) {
        my $reg_data = oct $2;
        push(@tp01_list, $reg_data);
      }
      elsif ((/END_TP_DATA_01/) && ($#tp01_list > 0) && ($save_sh2 == 3))  {
        ($fsb_stalled, $pi_crb_skid) = parse_sh2_tp01_list(\@tp01_list);		# in pi_tools.pm
        if (($fsb_stalled == 1) && ($flush_cfg_err == 0) && ($processor_count > 0)) {		
		# $flush_cfg_err is assigned in parse_sh2_mmr_list(), $processor_count is assigned in get_proc_cfg()
          $global_err_val |= $SHUB_HW_ERR;
          $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "\n    Shub PI testpoints indicate BPRI is continually asserted on the Front-side Bus.\n";
          $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "    This will prevent the processors on this node from issuing any requests.\n";
          $any_sh2_error = 1;

          $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
          $fru_list[$fru_count]->{NASID} = $nasid_value;
          $fru_list[$fru_count]->{REASON} = "Shub PI is continually asserting BPRI on the FSB.";
          $fru_list[$fru_count]->{DEVICE}   = "SHUB2";
          $fru_list[$fru_count]->{LOCATION} = $current_location;
          printf "\n%s: Added SHUB 2x NASID 0x%04X to the FRU list, entry %d.\n", $prefix, $nasid_value, $fru_count;
          $fru_count++;
        }
	$save_sh2 = 4;
	print "\n";
      }
      elsif ((/BEGIN_PI_CRB/) && ($save_sh2 == 4)) {
	@fru_rrb          = ();
	@fru_wrb          = ();
	@fru_irb          = ();
	@fru_prb          = ();
  	$fru_rrb_cnt = 0;
  	$fru_wrb_cnt = 0;
  	$fru_irb_cnt = 0;
  	$fru_prb_cnt = 0;
	$save_sh2 = 5;
      }
      elsif ((/(SH2_PI_CAM_\w+) (0x\w+)/) && ($save_sh2 == 5)) {
        my $reg_name = $1;
        my $reg_data = oct $2;
        parse_fru_crb($reg_name, $reg_data);
      }
      elsif ((/END_PI_CRB/) && ($save_sh2 == 5))  {
	$save_sh2 = 6;
	if ($verbose_mode == 1) {
	  printf "\n NASID 0x%04X FRU CRB Info:  RRB: %d   WRB: %d   IRB: %d    PRB: %d\n", 
	    $nasid_value, $fru_rrb_cnt, $fru_wrb_cnt, $fru_irb_cnt, $fru_prb_cnt;  
	}
        my $cam_status = read_all_crb("shub2_01");
      }
    }

    if (/BEGIN_TIO_NODE::(\d\d\d)\.(\d\d):(\d)/) {
      $save_tio = 1;
      $r = 0;
      @tio_reg_list = ();
      next;
    }
    if ($save_tio == 1) {
      if ((/END_TIO_NODE/) && ($save_tio == 1)) {
        parse_tio_mmr_list(\@tio_reg_list);             # 
        $tio_cnt++;
        $save_tio = 0;
      }
      elsif ((/(TIO_\w+) (0x\w+)/) || (/(OC_\w+) (0x\w+)/))  {
        my $reg_name = $1;
        my $reg_data = oct $2;
	$reg_name =~ s/^OC_/TIO_/;
        my $reg_ptr = reg_create_mmr_inst("tio_01", $reg_name);
        reg_set_value ($reg_ptr, $reg_data);
        if ( !reg_is_valid($reg_ptr) )
        {  $emess = "Unable to create TIO register instance: $reg_name.\n";
          dsp_error_and_exit($emess);
        }
        $tio_reg_list[$r] = $reg_ptr;
        # printf " Entry: %03d  Name: %s   Value: 0x%16llX\n", $r, $reg_name, $reg_data;
        $r++;
      }
    }
    if (/BEGIN_NL4R_NODE::(\d\d\d)\.(\d\d):/) {
      $save_nl4r = 1;
      $r = 0;
      @nl4r_reg_list = ();
      next;
    }
    if ($save_nl4r == 1) {
      if ((/END_NL4R_NODE/) && ($save_nl4r == 1)) {
        parse_nl4r_mmr_list(\@nl4r_reg_list);             #
        $nl4r_cnt++;
        $save_nl4r = 0;
      }
      elsif (/(R4_\w+) (0x\w+)/) {
        my $reg_name = $1;
        my $reg_data = oct $2;
        if ($reg_name =~ /R4_PORT._ERROR_DETAIL/) {
	  $reg_name =~s/_\d\d$//;
        }
        my $reg_ptr = reg_create_mmr_inst("nl4rtr_01", $reg_name);
        reg_set_value ($reg_ptr, $reg_data);
        if ( !reg_is_valid($reg_ptr) )
        {  $emess = "Unable to create NL4R register instance: $reg_name.\n";
          dsp_error_and_exit($emess);
        }
        $nl4r_reg_list[$r] = $reg_ptr;
        # printf " Entry: %03d  Name: %s   Value: 0x%16llX\n", $r, $reg_name, $reg_data;
        $r++;
      }
    }
  }

  close FRU_FILE;

}

#--------------------------------------------------------------------------------
#
#  Function to sort the PI CRB entries captured by the L1 FRU into the
#  appropriate CRB array location.
#
#--------------------------------------------------------------------------------

sub parse_fru_crb {

  my $cam_reg = $_[0];
  my $cam_data = $_[1];

  if ($cam_reg =~ /(SH2_PI_CAM_\w+)_RRB_(\d\d)/) {
    my $cam_type = "$1";
    my $crb_num = $2;
    if ($verbose_mode == 1) {
      printf " The type is %s : the RRB number is %d : the value is 0x%016llx\n", $cam_type, $crb_num, $cam_data;
    }
    if ($cam_type =~ /CAM_ADDRESS/) {
      $fru_rrb[$crb_num]->{CAM_ADDR} = $cam_data;
      $fru_rrb_cnt++;
    }
    elsif ($cam_type =~ /CAM_STATE/) {
      $fru_rrb[$crb_num]->{CAM_STATE} = $cam_data;
    }
    elsif ($cam_type =~ /CAM_LPRA/) {
      $fru_rrb[$crb_num]->{CAM_LPRA} = $cam_data;
    }
  }
  elsif ($cam_reg =~ /(SH2_PI_CAM_\w+)_WRB_(\d\d)/) {
    my $cam_type = "$1";
    my $crb_num = $2;
    if ($verbose_mode == 1) {
      printf " The type is %s : the WRB number is %d : the value is 0x%016llx\n", $cam_type, $crb_num, $cam_data;
    }
    if ($cam_type =~ /CAM_ADDRESS/) {
      $fru_wrb[$crb_num]->{CAM_ADDR} = $cam_data;
      $fru_wrb_cnt++;
    }
    elsif ($cam_type =~ /CAM_STATE/) {
      $fru_wrb[$crb_num]->{CAM_STATE} = $cam_data;
    }
    elsif ($cam_type =~ /CAM_LPRA/) {
      $fru_wrb[$crb_num]->{CAM_LPRA} = $cam_data;
    }
  }
  elsif ($cam_reg =~ /(SH2_PI_CAM_\w+)_IRB_(\d\d)/) {
    my $cam_type = "$1";
    my $crb_num = $2;
    if ($verbose_mode == 1) {
      printf " The type is %s : the IRB number is %d : the value is 0x%016llx\n", $cam_type, $crb_num, $cam_data;
    }
    if ($cam_type =~ /CAM_ADDRESS/) {
      $fru_irb[$crb_num]->{CAM_ADDR} = $cam_data;
      $fru_irb_cnt++;
    }
    elsif ($cam_type =~ /CAM_STATE/) {
      $fru_irb[$crb_num]->{CAM_STATE} = $cam_data;
    }
    elsif ($cam_type =~ /CAM_LPRA/) {
      $fru_irb[$crb_num]->{I_CAM_LPRA} = $cam_data;
    }
  }
  elsif ($cam_reg =~ /(SH2_PI_CAM_\w+)_PRB_(\d\d)/) {
    my $cam_type = "$1";
    my $crb_num = $2;
    if ($verbose_mode == 1) {
      printf " The type is %s : the PRB number is %d : the value is 0x%016llx\n", $cam_type, $crb_num, $cam_data;
    }
    if ($cam_type =~ /CAM_ADDRESS/) {
      $fru_prb[$crb_num]->{CAM_ADDR} = $cam_data;
      $fru_prb_cnt++;
    }
    elsif ($cam_type =~ /CAM_STATE/) {
      $fru_prb[$crb_num]->{CAM_STATE} = $cam_data;
    }
    elsif ($cam_type =~ /CAM_LPRA/) {
      $fru_prb[$crb_num]->{CAM_LPRA} = $cam_data;
    }
  }

  if (($cam_reg =~ /CAM_LPRA/) && ($verbose_mode == 1)) {
          print "\n";
          my $cam_index = 0;
          for ($cam_index = 0; $cam_index < $fru_rrb_cnt; $cam_index++) {
            printf "\nAfter parse: RRB %d: addr: 0x%016llx    state: 0x%016llx   lpra: 0x%016llx",
                $cam_index, $fru_rrb[$cam_index]->{CAM_ADDR}, $fru_rrb[$cam_index]->{CAM_STATE},$fru_rrb[$cam_index]->{CAM_LPRA};
          }
          for ($cam_index = 0; $cam_index < $fru_wrb_cnt; $cam_index++) {
            printf "\nAfter parse: WRB %d: addr: 0x%016llx    state: 0x%016llx   lpra: 0x%016llx",
                $cam_index, $fru_wrb[$cam_index]->{CAM_ADDR}, $fru_wrb[$cam_index]->{CAM_STATE},$fru_wrb[$cam_index]->{CAM_LPRA};
          }
          for ($cam_index = 0; $cam_index < $fru_irb_cnt; $cam_index++) {
            printf "\nAfter parse: IRB %d: addr: 0x%016llx    state: 0x%016llx   lpra: 0x%016llx",
                $cam_index, $fru_irb[$cam_index]->{CAM_ADDR}, $fru_irb[$cam_index]->{CAM_STATE},$fru_irb[$cam_index]->{CAM_LPRA};
          }
          for ($cam_index = 0; $cam_index < $fru_prb_cnt; $cam_index++) {
            printf "\nAfter parse: PRB %d: addr: 0x%016llx    state: 0x%016llx   lpra: 0x%016llx",
                $cam_index, $fru_prb[$cam_index]->{CAM_ADDR}, $fru_prb[$cam_index]->{CAM_STATE},$fru_prb[$cam_index]->{CAM_LPRA};
          }
          print "\n";
  }

}


#--------------------------------------------------------------------------------
#
#  Function to sort the entries in the fru_list array, and list them in the
#  FRU Summary report based on probability (score) values assigned within 
#  this function.
#
#--------------------------------------------------------------------------------


sub display_fru_list
{
  print "\nBeginning display_fru_list function\n" if ($verbose_mode == 1);

  my $x_cnt = 0;
  my $n_cnt = 0;
  my $f_cnt = 0;
  my @failed_node_list = ();
  my $failed_node_count = 0;
  my $match_found = 0;
  my $strike_counter = 0;
  my $max_score = 0;
  my $same_dev_limit = 7;
  my $total_same_dev = 0;

  	#  First, fill in the locations of the fru entries that only have NASID or LOCATION values assigned

  while (defined $fru_list[$x_cnt]) {
    if (!(defined $fru_list[$x_cnt]->{LOCATION}))  {
      $n_cnt = 0;
      $match_found = 0;
      if ($sh1_node_count > 0)  {		# check all Shub1 nodes for a matching NASID value
        while (defined $target_list[$n_cnt]) {
	  if ($fru_list[$x_cnt]->{NASID} == $target_list[$n_cnt]->{NASID})  {
            $fru_list[$x_cnt]->{LOCATION} = sprintf("%s, %s", $target_list[$n_cnt]->{MODULE}, $target_list[$n_cnt]->{NODE});
            $match_found = 1;
	    last;
	  }
	  else  {
            $n_cnt++;
	  }
        }
      }
      if (($match_found == 0) && ($sh2_node_count > 0))  {	# check all Shub2 nodes for a matching NASID value
	$n_cnt = 0;
        while (defined $sh2_target_list[$n_cnt]) {
	  if ($fru_list[$x_cnt]->{NASID} == $sh2_target_list[$n_cnt]->{NASID})  {
            $fru_list[$x_cnt]->{LOCATION} = sprintf("%s, %s", $sh2_target_list[$n_cnt]->{MODULE}, $sh2_target_list[$n_cnt]->{NODE});
            $match_found = 1;
	  }
	  elsif (defined $sh2_target_list[$n_cnt]->{SPEC_NASID}) {
	    if ($fru_list[$x_cnt]->{NASID} == $sh2_target_list[$n_cnt]->{SPEC_NASID})  {
              $fru_list[$x_cnt]->{LOCATION} = sprintf("%s, %s", $sh2_target_list[$n_cnt]->{MODULE}, $sh2_target_list[$n_cnt]->{NODE});
              $match_found = 1;
	    }
	  }
	  $n_cnt++;
	  if ($match_found == 1) {
	    last;	
	  }
        }
      }
      if (($match_found == 0) && ($tio_present == 1))  {	# check all TIO nodes for a matching NASID value
	$n_cnt = 0;
        while (defined $tio_target_list[$n_cnt]) {
	  if ($fru_list[$x_cnt]->{NASID} == $tio_target_list[$n_cnt]->{NASID})  {
            $fru_list[$x_cnt]->{LOCATION} = sprintf("%s, %s", $tio_target_list[$n_cnt]->{MODULE}, $tio_target_list[$n_cnt]->{NODE});
            $match_found = 1;
	    last;
	  }
	  else  {
            $n_cnt++;
	  }
        }
      }
								 # check all NL4R nodes for a matching NASID value
      if (($match_found == 0) && ($nl4r_present == 1) && ($dont_dump_nl4r_regs == 0))  {
	$n_cnt = 0;
        while (defined $nl4r_target_list[$n_cnt]) {
	  if ($fru_list[$x_cnt]->{NASID} == $nl4r_target_list[$n_cnt]->{NASID})  {
            $fru_list[$x_cnt]->{LOCATION} = sprintf("%s, %s", $nl4r_target_list[$n_cnt]->{MODULE}, $nl4r_target_list[$n_cnt]->{NODE});
            $match_found = 1;
	    last;
	  }
	  else  {
            $n_cnt++;
	  }
        }
      }
      if (($match_found == 0) && ($parallel_mode == 1))  {
	  gen_global_nasid_list();	# function to create a single list from the various partitions dumped
	  $n_cnt = 0;
	  while (defined $global_nasid_list[$n_cnt]) {
	    if ($fru_list[$x_cnt]->{NASID} == $global_nasid_list[$n_cnt]->{NASID})  {
              $fru_list[$x_cnt]->{LOCATION} = sprintf("%s, %s", $global_nasid_list[$n_cnt]->{MODULE}, $global_nasid_list[$n_cnt]->{NODE});
              $match_found = 1;
	      last;
	    }
	    else  {
              $n_cnt++;
	    }
	  }
      }
      if ($match_found == 0) {
        $fru_list[$x_cnt]->{LOCATION} = "Unknown"; 
      }
    }

    if (!(defined $fru_list[$x_cnt]->{NASID}))  {
      $n_cnt = 0;
      $match_found = 0;
      while (defined $target_list[$n_cnt]) {
	if (($target_list[$n_cnt]->{MODULE} =~ /$fru_list[$x_cnt]->{SLOT}/) 
	 && ($target_list[$n_cnt]->{NODE} =~ /$fru_list[$x_cnt]->{NODE}/)) {
          $fru_list[$x_cnt]->{NASID} = $target_list[$n_cnt]->{NASID};
          $match_found = 1;
	  last;
	}
	else  {
          $n_cnt++;
	}
      }
      if (($match_found == 0) && ($nl4r_present == 1) && ($dont_dump_nl4r_regs == 0))  {
	$n_cnt = 0;
        while (defined $nl4r_target_list[$n_cnt]) {
	  if (($nl4r_target_list[$n_cnt]->{MODULE} =~ /$fru_list[$x_cnt]->{SLOT}/) 
	   && ($nl4r_target_list[$n_cnt]->{NODE} =~ /$fru_list[$x_cnt]->{NODE}/)) {
            $fru_list[$x_cnt]->{NASID} = $nl4r_target_list[$n_cnt]->{NASID};
            $match_found = 1;
	    last;
	  }
	  else  {
            $n_cnt++;
	  }
        }
      }
      if (($match_found == 0) && ($tio_present == 1))  {
	$n_cnt = 0;
        while (defined $tio_target_list[$n_cnt]) {
	  if (($tio_target_list[$n_cnt]->{MODULE} =~ /$fru_list[$x_cnt]->{SLOT}/) 
	   && ($tio_target_list[$n_cnt]->{NODE} =~ /$fru_list[$x_cnt]->{NODE}/)) {
            $fru_list[$x_cnt]->{NASID} = $tio_target_list[$n_cnt]->{NASID};
            $match_found = 1;
	    last;
	  }
	  else  {
            $n_cnt++;
	  }
        }
      }
      if (($match_found == 0) && ($parallel_mode == 1))  {
          gen_global_nasid_list();      # function to create a single list from the various partitions dumped
          $n_cnt = 0;
          while (defined $global_nasid_list[$n_cnt]) {
	    if (($global_nasid_list[$n_cnt]->{MODULE} =~ /$fru_list[$x_cnt]->{SLOT}/) 
	     && ($global_nasid_list[$n_cnt]->{NODE} =~ /$fru_list[$x_cnt]->{NODE}/)) {
              $fru_list[$x_cnt]->{NASID} = $global_nasid_list[$n_cnt]->{NASID};
                $match_found = 1;
                last;
            }
            else  {
              $n_cnt++;
            }
          }
      }

      if ($match_found == 0) {
        $fru_list[$x_cnt]->{NASID} = "Unknown";
      }

    }

    $x_cnt++;
  }

  print "\nIn display_fru_list function - completed location values - fru entries: $#fru_list\n" if ($verbose_mode == 1);
  #  return;		#  OK up to this point


        #  Generate a failed_node_list that tabulates the number of strikes against each failing node
	#
	#  A strike against a NASID is incurred whenever a valid HW Timestamp value is found on a node,
	#  or when a NASID appears in the Source field of a valid error detail reg.

  $x_cnt = 0;
  
  while ($x_cnt < $fru_count) {
    $f_cnt = 0;
    if ($x_cnt == 0)  {  				#  Catch first one - always add this to the failing_node_list 
      $failed_node_list[$f_cnt]->{NASID} = $fru_list[$x_cnt]->{NASID};
      $failed_node_list[$f_cnt]->{LOCATION} = $fru_list[$x_cnt]->{LOCATION};
      $failed_node_list[$f_cnt]->{STRIKE} = 1;
      $failed_node_count++;
      $strike_counter++;
    }
    else  {
      $match_found = 0;	
      while (defined $failed_node_list[$f_cnt])  {	#  First possibility - match found so increment the strike counter
	if ((($fru_list[$x_cnt]->{NASID} == $failed_node_list[$f_cnt]->{NASID}) || ($fru_list[$x_cnt]->{DEVICE} =~ /NL4R/))
	   && ($fru_list[$x_cnt]->{LOCATION} eq $failed_node_list[$f_cnt]->{LOCATION})) {
	  if ($fru_list[$x_cnt]->{REASON} =~ /Source router/) { #  differentiate a remote router in those cases where the same
	    $failed_node_list[$f_cnt]->{STRIKE} += 5; 		#  router is implicated by several devices
	  }
	  else {
	    $failed_node_list[$f_cnt]->{STRIKE}++; 
	  }
          $strike_counter++;
          $match_found = 1;
	  print "\n Strike count for $fru_list[$x_cnt]->{LOCATION} = $failed_node_list[$f_cnt]->{STRIKE} .\n" if ($verbose_mode == 1);
	}
	$f_cnt++;
      }
      if ($match_found == 0)  {				#  No match found, so add this fru entry to the failing node list
        $failed_node_list[$f_cnt]->{NASID} = $fru_list[$x_cnt]->{NASID};
        $failed_node_list[$f_cnt]->{LOCATION} = $fru_list[$x_cnt]->{LOCATION};
        $failed_node_list[$f_cnt]->{STRIKE} = 1;
        $failed_node_count++;
        $strike_counter++;
      }
    }
    $x_cnt++;
  }

  if ($strike_counter != $fru_count)  {
    print "\n\nALERT!!!  strike counter ($strike_counter) not equal to total fru entries ($fru_count)\n\n";
  }
  else  {
    print "\nCurrent fru_count:  $fru_count\n\n" if ($verbose_mode == 1);
  }

  printf "\nIn display_fru_list function - completed failed_node_list - fru entries: %d\n",
    ($#fru_list + 1)  if ($verbose_mode == 1);

  # return;		#  OK up to here

	#  Now assign each fru entry a index value according to the time stamp values found for the node
	#  The algorithm will go something like:
	#	First, see if the fru nasid is found in the valid time stamp list (@sorted_times).  
	#	  If it is found, assign the index value from the time stamp list.
	#           If the immediately proceeding fru value is derived from a error detail reg,
        #           indicated by "Source" or "attached IO" found in the REASON string, then this fru entry should 
        #           be assigned the index value of the NASID having the valid hw time stamp. 
	#           If, later on, a lower index value is found for the secondary NASID, the lower
	#           index value will be used.
        #       If the fru nasid isn't found in the valid time stamp list, and it isn't a fru nasid
	#       derived from a error detail reg, it needs to receive the largest possible index value.
	#

  $x_cnt = 0;
  my $prev_time = 0;
  my $prev_time_cnt = 0;
  my $first_hw_time_index = 0;	    #  This is used to prevent a corrected error that happens way earlier from
  my $first_hw_time_acquired = 0;   #  throwing off the fru score for a first-occurring special case HW error.

  while ($x_cnt < $fru_count) {
    $n_cnt = 0;
    while (defined $sorted_times[$n_cnt])  {
      if ($n_cnt == 0)  {
        $prev_time = $sorted_times[$n_cnt]->{TIME};
        $prev_time_cnt = 0;
      }
      elsif ($prev_time < $sorted_times[$n_cnt]->{TIME})  {
	  $prev_time_cnt = $n_cnt;
      }
      if (($first_hw_time_acquired == 0) && ($sorted_times[$n_cnt]->{HW} == 1)) {
	$first_hw_time_index = $n_cnt;
        $first_hw_time_acquired = 1;
      }
      if (($fru_list[$x_cnt]->{NASID} == $sorted_times[$n_cnt]->{NASID})    	#  NASID & location value is found in time stamp list
	 && ($fru_list[$x_cnt]->{LOCATION} eq $sorted_times[$n_cnt]->{LOCATION}))  {
	if (!(defined $fru_list[$x_cnt]->{T_INDEX}))  {				#  FRU entry hasn't been assigned an index yet
	  if (($n_cnt != 0) && ($sorted_times[$n_cnt]->{TIME} == $prev_time))  {	#  Handle the case of duplicate time stamps
            $fru_list[$x_cnt]->{T_INDEX} = $prev_time_cnt; 
	  }
	  else {
            $fru_list[$x_cnt]->{T_INDEX} = $n_cnt; 
	  }
          printf "\nIn display_fru_list() - index assignment case 0 - fru entries: %d   iteration: %d  index: %d\n", 
	    ($#fru_list + 1), $x_cnt, $n_cnt  if ($verbose_mode == 1);
	}
	else  {
	  if ($n_cnt < $fru_list[$x_cnt]->{T_INDEX}) {		#  assign the lower value if fru entry already has an assigned index
	    $fru_list[$x_cnt]->{T_INDEX} = $n_cnt;
            printf "\nIn display_fru_list() - index assignment case 1 - fru entries: %d   iteration: %d  index: %d\n",
	      ($#fru_list + 1), $x_cnt, $n_cnt  if ($verbose_mode == 1);
	  }
	}
							#  if the next fru entry is derived from a detail reg, assign the index
							#  Keep in mind this look-ahead can set a defined value to $fru_list[$x_cnt + 1]
							#  even if the array element wasn't previously defined.
	if (($x_cnt + 1) < $fru_count) {		#  Only do the look-ahead check we're still within the array
	  if ( ($fru_list[$x_cnt + 1]->{REASON} =~ /Source node/) 
	    || ($fru_list[$x_cnt + 1]->{REASON} =~ /Source router/)
	    || ($fru_list[$x_cnt + 1]->{REASON} =~ /Directory node/)
	    || ($fru_list[$x_cnt + 1]->{REASON} =~ /attached IO/) )  {
	    $fru_list[$x_cnt + 1]->{T_INDEX} = $n_cnt;
            printf "\nIn display_fru_list() - index assignment case 2 - fru entries: %d   iteration: %d  index: %d\n",
	      ($#fru_list + 1), $x_cnt, $n_cnt  if ($verbose_mode == 1);
	  }
	}
        last;			#  since we have a match, no need to continue the sorted times while loop
      }
      else  {
        $prev_time = $sorted_times[$n_cnt]->{TIME};
        $n_cnt++;		#  Nasid and/or location doesn't match, so get next time entry 
      }
    }
    
    if (!(defined $fru_list[$x_cnt]->{T_INDEX}))  {		# no matching time stamp found, so assign max value 
      $fru_list[$x_cnt]->{T_INDEX} = $n_cnt;
      printf "\nIn display_fru_list() - index assignment case 3 - fru entries: %d   iteration: %d  index: %d\n",
        ($#fru_list + 1), $x_cnt, $n_cnt  if ($verbose_mode == 1);
    }

	#  Now assign the total count of strikes recorded on each nasid to each fru entry
	#  Initially try :  ($t_stamps - T_INDEX) + STRIKES

    $f_cnt = 0;
    while (defined $failed_node_list[$f_cnt])  {

      if ( (($fru_list[$x_cnt]->{NASID} == $failed_node_list[$f_cnt]->{NASID}) || ($fru_list[$x_cnt]->{DEVICE} =~ /NL4R/)) 
	 && ($fru_list[$x_cnt]->{LOCATION} eq $failed_node_list[$f_cnt]->{LOCATION})) {

        $fru_list[$x_cnt]->{SCORE} = ($t_stamps - $fru_list[$x_cnt]->{T_INDEX}) + $failed_node_list[$f_cnt]->{STRIKE};  
	if ($fru_list[$x_cnt]->{SCORE} > $max_score)  {
	  $max_score = $fru_list[$x_cnt]->{SCORE};
	}

        printf "\n score assignment  - fru entries: %d   iteration: %d  t_stamps: %d   index: %d  strike_count: %d  score: %d  \n",
	  ($#fru_list + 1), $x_cnt, $t_stamps, $fru_list[$x_cnt]->{T_INDEX}, $failed_node_list[$f_cnt]->{STRIKE}, 
	  $fru_list[$x_cnt]->{SCORE}   if ($verbose_mode == 1);
      }
      $f_cnt++;
    }

    printf "\nIn display_fru_list() - score value while loop - fru entries: %d   iteration: %d\n",
     ($#fru_list + 1), $x_cnt   if ($verbose_mode == 1);
    $x_cnt++;				#  Check next fru entry
  }

  printf "\nIn display_fru_list function - completed score value generation - fru entries: %d\n", ($#fru_list + 1)  if ($verbose_mode == 1);
  # return;		#  Failed before it hit this return


	#  Now generate final fru list which takes into account the index placing according to
	#  time stamp values as well as the number of times a nasid has a strike against it.
	#  Higher value means the fru entry goes to the top of the list.


  my $fru_sum = sprintf "\nFRU Summary:\n";
  $fru_sum .= sprintf "\n    NASID    LOCATION                 Score   Reason for Failure Event\n";
  $fru_sum .= sprintf   "    -----    -------------            -----   ------------------------";

  $x_cnt = 0;
  my $high_index = 0;                 # array index having the current high score value
  my $high_score = 0;                 # high score value for the inner for loop
  my $first_one = 1;                  # indicates this is the first unused entry
  my $acquired = 0;                   # set this to prevent duplicate entries with the same time value
  my @fru_dev_list = ();
  my $prev_fru_score = 0;
  my $prev_fru_dev = "Initial";
  my $prev_fru_reason = "Initial";
  my $prev_fru_nasid = 0xbeef;

  while ($x_cnt < $fru_count) {			# outer while loop for sorting & printing the fru summary
    my $special_case = 0;
    for ($n_cnt = 0; $n_cnt < $fru_count; $n_cnt++)  {

        #  Special case #1:  If the PI_FSB_HW_ERR or XN Packet Flow error occurred first,
	#  		     force the Score to the max_score + 5 to ensure that multiple
	#                    MemTimeouts to the same node which are an aftereffect of the 
	#                    halted packet traffic don't mistakenly throw off the scoring 
	#                    and incorrectly implicate the destination node of the timed out
	#                    memory accesses.
	#		     In the LLP Retry Timeout, POWER_NOT_OK, and CORESENTBADCRC cases, 
	#                    will bump the score regardless of time stamp value since the NL4R timestamps 
	#		     aren't always consistent.
	#
	#
      if ( (! defined $fru_list[$n_cnt]->{USED}) 			#  must be unused
	&& (($fru_list[$n_cnt]->{REASON} =~ /LLP Retry Timeout/) 	#  LLP retry timeout at any time will get promoted
	|| ($fru_list[$n_cnt]->{REASON} =~ /Fatal unknown error/)       #  Shub may be internally hung
	|| ($fru_list[$n_cnt]->{REASON} =~ /Data Corruption/) 		#  LLP data corruption will always get promoted
	|| ($fru_list[$n_cnt]->{REASON} =~ /Hung NL4R in Test points/)  #  NL4R test points indicate router is internally hung
	|| ($fru_list[$n_cnt]->{REASON} =~ /No ports have initialized/) #  NL4R is probably internally hung
	|| ($fru_list[$n_cnt]->{REASON} =~ /Cable Problem/)   		#  LLP POWER_NOT_OK problem will always get promoted
	|| (($fru_list[$n_cnt]->{T_INDEX} == $first_hw_time_index) 
	&&  (($fru_list[$n_cnt]->{REASON} =~ /Packet Flow Error/)	#  or a first occurring LLP Packet Flow error Shub XN or NL4R
	|| ($fru_list[$n_cnt]->{REASON} =~ /Chiplet NoMatch Header/)	#  or a first occurring Chiplet NoMatch error
	|| ($fru_list[$n_cnt]->{REASON} =~ /Suspect either Processor/)	#  or a first occurring any processor FSB issue
	|| ($fru_list[$n_cnt]->{REASON} =~ /Hung Front-side Bus/)	#  or a first occurring Hung FSB error (Shub 2)
	|| ($fru_list[$n_cnt]->{REASON} =~ /CPU in SKT/)		#  or a first occurring any processor FSB issue (Shub 2)
	|| ($fru_list[$n_cnt]->{REASON} =~ /Suspect Processor/))) ) )	#  or a first occurring processor FSB issue
      {
	  if ($x_cnt == 0) {		#  only increase the score the first time through the outer while loop
	    if (($fru_list[$n_cnt]->{SCORE} == $max_score) && ($fru_list[$n_cnt]->{T_INDEX} == $first_hw_time_index))  {
	      $fru_list[$n_cnt]->{SCORE} += 16;	# differentiate a node that has a first occurring HW error
	    }
	    elsif (($fru_list[$n_cnt]->{SCORE} < $max_score) && ($fru_list[$n_cnt]->{T_INDEX} == $first_hw_time_index))  {
	      $fru_list[$n_cnt]->{SCORE} = $max_score + 14;	# differentiate a node that has a first occurring HW error
	    }
	    elsif (($fru_list[$n_cnt]->{SCORE} == $max_score) && ($fru_list[$n_cnt]->{T_INDEX} != $first_hw_time_index))  {
	      $fru_list[$n_cnt]->{SCORE} += 12;	# differentiate a node that already has high fru score
	    }
	    else  {
	      $fru_list[$n_cnt]->{SCORE} = $max_score + 8;
	    }
	  }
	  $special_case = 1;
      }

		#  assign a lower value to the PI is Hung case since that can be an after-effect
		#  of a downed link

      if ( (! defined $fru_list[$n_cnt]->{USED}) 			#  must be unused
	&& ($fru_list[$n_cnt]->{REASON} =~ /PI is Hung/))   		#  PI test point dump indicates PI ICC problem
      {
	  if ($x_cnt == 0) {		#  only increase the score the first time through the outer while loop
	    if ($fru_list[$n_cnt]->{SCORE} == $max_score)  {
	      $fru_list[$n_cnt]->{SCORE} += 6;	# differentiate a node that is already at the top of the list
	    }
	    else  {
	      $fru_list[$n_cnt]->{SCORE} = $max_score + 4;
	    }
	  }
          $special_case = 1;
      }
        # Always save the first unused entry

      if ((! defined $fru_list[$n_cnt]->{USED}) 
	&& ($first_one == 1))  
      {
        $high_score = $fru_list[$n_cnt]->{SCORE};
        $high_index = $n_cnt;
        $first_one = 0;
        $acquired = 1;
        print "\nIn display_fru_list - acquired fru entry $n_cnt (high score=$high_score) in condition 1.\n" if ($verbose_mode == 1);
        next;
      }
        # Save the entry if it's greater than the previous high score value

      if ((! defined $fru_list[$n_cnt]->{USED}) 
	&& ($fru_list[$n_cnt]->{SCORE} > $high_score)) 
      {
        $high_score = $fru_list[$n_cnt]->{SCORE};
        $high_index = $n_cnt;
        $acquired = 1;
        print "\nIn display_fru_list - acquired fru entry $n_cnt (high score=$high_score) in condition 2.\n" if ($verbose_mode == 1);
        next;
      }

        # Save the entry if it's equal to the high value seen on a previous iteration

      if ((! defined $fru_list[$n_cnt]->{USED}) 
	&& ($fru_list[$n_cnt]->{SCORE} == $high_score) 
	&& ($acquired == 0)) 
      {
        $high_score = $fru_list[$n_cnt]->{SCORE};
        $high_index = $n_cnt;
        $acquired = 1;
        print "\nIn display_fru_list - acquired fru entry $n_cnt (high score=$high_score) in condition 3.\n" if ($verbose_mode == 1);
        next;
      }
      printf "\nIn display_fru_list - non acquired fru entry %d   score value = %d  (high score=%d) in condition 4.\n",
	$n_cnt, $fru_list[$n_cnt]->{SCORE}, $high_score  if ($verbose_mode == 1);
    }

    if ( ($fru_list[$high_index]->{REASON} =~ /DIMM/) 
      || ($fru_list[$high_index]->{LOCATION} =~ /Unknown/) 
      || ($verbose_mode == 1) 
      || ($full_fru_report == 1) ) {
      $special_case = 1;
    }
    my $same_dev_count = 0;		#  means of filtering out excessive redundant fru entries from fru summary
    if ($special_case == 0)  {
      push (@fru_dev_list, $fru_list[$high_index]->{LOCATION});
      my $fru_dev = "";
      foreach $fru_dev (@fru_dev_list) {
	if ($fru_dev eq $fru_list[$high_index]->{LOCATION}) {
          $same_dev_count++;
        }
      }
    }
    if ( ($same_dev_count < $same_dev_limit) && 
         ( ($fru_list[$high_index]->{NASID} != $prev_fru_nasid)		#  If all 4 of these "prev_fru" conditions are not met
        || ($fru_list[$high_index]->{SCORE} != $prev_fru_score)		#  then we won't print the fru item.  This should prevent
        || ($fru_list[$high_index]->{LOCATION} !~ $prev_fru_dev)	#  muliple consecutive identical fru items.
        || ($fru_list[$high_index]->{REASON} !~ $prev_fru_reason) ))  {
      my $gap = ($fru_list[$high_index]->{LOCATION} =~ /TIO_\d/) ? "" : 
		($fru_list[$high_index]->{LOCATION} =~ /TIO\d/) ? " " : "        ";
      $fru_sum .= sprintf "\n    0x%04X   %-15s %s %3d     %s", 
	  $fru_list[$high_index]->{NASID}, 
	  $fru_list[$high_index]->{LOCATION}, 
	  $gap,
	  $fru_list[$high_index]->{SCORE}, 
	  $fru_list[$high_index]->{REASON};
      print "\n printed fru entry.  score value is $fru_list[$high_index]->{SCORE}   high index is $high_index .\n" if ($verbose_mode == 1);  	
      $prev_fru_score =  $fru_list[$high_index]->{SCORE};
      $prev_fru_dev =    $fru_list[$high_index]->{LOCATION};
      $prev_fru_reason = $fru_list[$high_index]->{REASON};
      $prev_fru_nasid =  $fru_list[$high_index]->{NASID};
    }
    else {
      $total_same_dev++;
    }
    $fru_list[$high_index]->{USED} = 1;
    $first_one = 1;
    $acquired = 0;
    $x_cnt++;
  }			# end of outer while loop for sorting & printing the fru summary 
  $fru_sum .= sprintf "\n\n";

  if ($total_same_dev > 0) {
    $fru_sum .= sprintf "\n There were a total of %d redundant fru entries that were not included in this summary.\n\n", $total_same_dev;
  }

  return($fru_sum);
}


#--------------------------------------------------------------------
#
#  Return the chassis type value found when scanning through the list
#  of 1955 and 191010 chassis found so far.
#
#--------------------------------------------------------------------

sub get_chassis_type {

  my $ch_val = "$_[0]";		# user supplied chassis location
  my $ch_type = 0;

  my $iru_loc = "0";

  foreach $iru (@th_chassis_list) {
    if ($ch_val eq $iru->{LOC}) {
      $ch_type = ($iru->{TYPE} =~ /1955/) ? 1 :
                 ($iru->{TYPE} =~ /191010/) ? 2 : 0;
      last;
    }
  }

  return ($ch_type);
}

#--------------------------------------------------------------------
#
#  Returns the brick name (ie 004c05) given the supplied slot and rack values.
#
#--------------------------------------------------------------------

sub get_brick_loc {

  my $slot_val = $_[0];
  my $rack_val = $_[1];

  my $brick_name = "?????";

  foreach $iru (@th_chassis_list) {
    if (($slot_val == $iru->{SLOT}) && ($rack_val == $iru->{RACK})) {
      $brick_name = $iru->{LOC};
      last;
    }
  }

  return($brick_name);

}


#--------------------------------------------------------------------
#
#  Returns the brick name (ie 004c05) given the supplied slot and rack values.
#
#--------------------------------------------------------------------

sub check_leds_line {

  my $leds_line = "$_[0]";
  my $line_ok = 1;

  if (     ($leds_line =~ /fail/)            	# 0x82, 0x83, 0xb0, 0xb5, 0xb6, 0xb8, 0xba, 0xbc, 0xc1
        || ($leds_line =~ /MCA/)             	# 0x38, 0x39, 0x3a, 0x3e, 0x56
        || ($leds_line =~ /o memory/)        	# 0x80, 0x84
        || ($leds_line =~ /timed out/)          # 0x81
        || ($leds_line =~ /Too many/)        	# 0x8a
        || ($leds_line =~ /initialize/)      	# 0x85
        || ($leds_line =~ /duplicate rack/)  	# 0x86
        || ($leds_line =~ /invalid/)         	# 0xb1
        || ($leds_line =~ /overflow/)        	# 0xb7, 0xbd
        || ($leds_line =~ /[Bb]ad/)          	# 0xb3, 0xb9, 0xbb
        || ($leds_line =~ /No PROM segment/) 	# 0xbe
        || ($leds_line =~ /Out of SAL/)      	# 0xc0
        #  || ($leds_line =~ /SAL_PMI/)         # 0x57
        || ($leds_line =~ /Error/)           	# 0xc2
        || ($leds_line =~ /Could not write/) 	# 0xc3
        || ($leds_line =~ /illegal/)   )     	# 0xd8
    {
      $line_ok = 0;
    }

  return($line_ok);
}

#------------------------------------------------------
#
#  Function to add HW components to the list of FRU suspects.
#  If a nasid value of 0xffffff is supplied, it is considered
#  undefined and not provided to the hash table.
#  If a location value of 'UNDEFINED' is supplied, the location
#  in the fru hash table is left undefined.
#  
#  A valid nasid OR a valid location must be provided for the fru
#  summary stuff in display_fru_list() to work properly.
#
#------------------------------------------------------

sub save_fru {

  my $dev = $_[0];
  my $nsd = $_[1];
  my $loc = $_[2];
  my $rsn = $_[3];
  my $tag = $_[4];
  my $ret_msg = "";
     
  $fru_list[$fru_count]->{DEVICE}   = $dev;

  if ($nsd != 0xffffff) {
    $fru_list[$fru_count]->{NASID}    = $nsd;
    $ret_msg = sprintf "\n%s: Added NASID 0x%04X to the FRU list, entry %d  (%s).\n", 
      $prefix, $nsd, $fru_count, $tag;
  }
  else {
    $ret_msg = sprintf "\n%s: Added location %s to the FRU list, entry %d  (%s).\n", 
      $prefix, %loc, $fru_count, $tag;
  }

  if ($loc !~ /UNDEFINED/) {
    $fru_list[$fru_count]->{LOCATION} = $loc;
  }
  $fru_list[$fru_count]->{REASON}   = $rsn;
  $fru_count++;

  if (($nsd == 0xffffff) && ($loc =~ /UNDEFINED/)) {
    ret_msg .= "\nWARNING:  Invalid location and NASID values supplied to save_fru().\n";
  }

  return($ret_msg);
}

#-------------------------------------------------------------------------------
1;
