# 
#/**************************************************************************
# *                                                                        *
# *               Copyright (C) 1989-1999 Silicon Graphics, Inc.           *
# *                                                                        *
# *  These coded instructions, statements, and computer programs  contain  *
# *  unpublished  proprietary  information of Silicon Graphics, Inc., and  *
# *  are protected by Federal copyright law.  They  may  not be disclosed  *
# *  to  third  parties  or copied or duplicated in any form, in whole or  *
# *  in part, without the prior written consent of Silicon Graphics, Inc.  *
# *                                                                        *
# **************************************************************************/
# 
# @(#)$Id: groupdef.pm,v 1.4 2004/06/23 20:16:50 beighley Exp $
#-------------------------------------------------------------------------------
#
#  FILE: groupdef.pm
#
#  DESCRIPTION:  This file contains PERL subroutines to create an GUI for a
#    user to download debug port testpoint groups to the external logic
#    analyzer (LA).  The name selection GUI (namesel) is used to allow
#    creation/modification of the testpoint groups.
#
#  NOTES: The following modules file is required when this file is used:
#    common_defs, TLAControl, la_subs, namesel
#
#  MODIFICATION HISTORY:
#
#     WHEN      WHO             WHAT
#     --------  --------------  -----------------------------------------
#     10/08/02  Karen Beighley  Original source code.
#     11/20/02  Karen Beighley  Initial source submission for release.
#     09/08/03  Karen Beighley  Modified parameters to create_namesel_window().
#     10/17/03  Karen Beighley  Modified parameters to create_namesel_window().
#     05/07/04  Karen Beighley  Modified group configuration window and order
#                                  of testpoints within a group.
#                               Check for commas in testpoint and channel names.
#
#-------------------------------------------------------------------------------

use common_defs;
use TLAControl;
use la_subs;
use namesel;

#-------------------------------------------------------------------------------
#  CONSTANTS - These values cannot change!
#-------------------------------------------------------------------------------

$GWD_WINDOW   = 0;  # Group window data: main window handle
$GWD_NAME_REF = 1;  # Group window data: reference to array of names
$GWD_DATA_REF = 2;  # Group window data: reference to array of data
$GWD_BTN_CRT  = 3;  # Group window data: create button handle
$GWD_BTN_MOD  = 4;  # Group window data: modify button handle
$GWD_LISTBOX  = 5;  # Group window data: listbox handle
$GWD_MODE     = 6;  # Group window data: listbox handle
$GWD_INDEX    = 7;  # Group window data: index of group to modify
$GWD_LAP_REF  = 8;  # Group window data: reference to LA parameters
$GWD_WIN_SEL  = 9;  # Group window data: reference to group selection window

$seltitle = "Debug Port Testpoints for LA Group";

#------------------------------------------------------------------------------

sub create_groupdef_window
{
   printf "\ncreate_groupdef_window( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $dp_name    = $_[0]; # Full name of debug port
   my $name_ref   = $_[1]; # Reference to array of names
   my $cmap_ref   = $_[2]; # Reference to array with channel names
   my $lap_ref    = $_[3]; # Reference to array with LA parameters
   my $cmd_ref    = $_[4]; # Reference to command button to disable/enable
   my @gw_data    = (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
   my ($i, $text, $win_h, $menubard, $casmenu, $name, $frame, $label, $entry, 
      $grp_lbox, $btn, $bwidth, $tmp, $ngrp, $gmap, $ind_ref, $gname);

   #---------------------------------------
   #  Build combined names: TP and channel
   #---------------------------------------
   my @new_names = ();
   for ($i = 0; $i <= $#$name_ref; $i++)
   {  $text = sprintf("%s=%s", $$name_ref[$i], $$cmap_ref[$i]);
      $new_names[$i] = $text;
   }

   #---------------------------------------
   #  Create main window for selection
   #---------------------------------------
   $text  = "Logic Analyzer: Download Debug Port Group(s)";
   $win_h = MainWindow->new(-title => $text);
   $win_h->bind('<Destroy>' => [\&destroy_groupdef_window, \@gw_data, $cmd_ref]);
   if ( $$cmd_ref != 0 )
   {  $$cmd_ref->configure(-state => 'disable');
   }

   #---------------------------------------
   #  Create the menu bar
   #---------------------------------------
   $menubar  = $win_h->Menu(-type => 'menubar');
   $win_h->configure(-menu => $menubar);
   $casmenu = $menubar->cascade(-label => '~File', -tearoff => 0);
   $casmenu->command(-label => 'Close', -command => [$win_h => 'destroy']);
   $casmenu = $menubar->cascade(-label => '~Group', -tearoff => 0);
   $crt_btn = $casmenu->command(-label => 'Create New Group',         -command => [\&groupdef_cbf_create, \@gw_data]);
   $mod_btn = $casmenu->command(-label => 'Modify Selected Group',    -command => [\&groupdef_cbf_modify, 0, \@gw_data]);
   $casmenu->command(-label => 'Remove Selected Group(s)', -command => [\&groupdef_cbf_remove, \@gw_data]);

   #---------------------------------------
   #  Create the main window label
   #---------------------------------------
   $name = "Download Group(s) to Logic Analyzer: $dp_name";
   $frame = $win_h->Frame(-borderwidth => 10);
   $label = $frame->Label(-anchor => 'w', -text => $name);
   $label->pack(-side => 'left');
   $frame->pack(-side => 'top', -expand => 'no', -fill => 'both');


   #---------------------------------------
   #  Create the widgets to specify the LA address
   #---------------------------------------
   $frame = $win_h->Frame();
   $label = $frame->Label(-anchor => 'w', -width => 2);
   $label->pack(-side => 'left');
   $label = $frame->Label(-anchor => 'w', -text => "Logic Analyzer IP Address:", -width => 27);
   $label->pack(-side => 'left');
   $entry = $frame->Entry(-width => 15, -textvariable => \$$lap_ref[$LA_PARM_HOST]);
   $entry->pack(-side => 'left');
   $label = $frame->Label(-anchor => 'w', -width => 2);
   $label->pack(-side => 'left');
   $label = $frame->Label(-anchor => 'w', -text => "Port:", -width => 5);
   $label->pack(-side => 'left');
   $entry = $frame->Entry(-width => 6, -textvariable => \$$lap_ref[$LA_PARM_PORT]);
   $entry->pack(-side => 'left');
   $frame->pack(-side => 'top', -expand => 'no', -fill => 'both');

   #---------------------------------------
   #  Create the widgets to specify the LA module
   #---------------------------------------
   $frame = $win_h->Frame();
   $label = $frame->Label(-anchor => 'w', -width => 2);
   $label->pack(-side => 'left');
   $label = $frame->Label(-anchor => 'w', -text => "Logic Analyzer Module Name:", -width => 27);
   $label->pack(-side => 'left');
   $entry = $frame->Entry(-width => 15, -textvariable => \$$lap_ref[$LA_PARM_MODULE]);
   $entry->pack(-side => 'left');
   $frame->pack(-side => 'top', -expand => 'no', -fill => 'both');

   #---------------------------------------
   #  Create the listbox for group names
   #---------------------------------------
   $frame = $win_h->Frame(-borderwidth => 10);
   $sub_frame = $frame->Frame();
   $label = $sub_frame->Label(-anchor => 'w', -text => "Available Debug Port Testpoint Groups:");
   $label->pack(-side => 'top', -fill => 'both');
   $grp_lbox = $sub_frame->Scrolled('Listbox', -scrollbars => 'se',
                -height => 12, -font => 'Clean 12 bold', -selectmode => 'extended');
   $grp_lbox->pack(-side => 'top', -fill => 'both', -expand => 'yes');
   $grp_lbox->bind('<Double-1>' => [\&groupdef_cbf_modify, \@gw_data]);
   $sub_frame->pack(-side => 'left', -fill => 'both', -expand => 'yes');
   $sub_frame = $frame->Frame();
   $bwidth = 12;
   $btn = $frame->Button(-text => 'Create Group', -width => $bwidth,
             -command => [\&groupdef_cbf_create, \@gw_data]);
   $btn->pack(-side => 'top', -padx => 5, -pady => 3);
   $btn = $frame->Button(-text => 'Modify Group', -width => $bwidth,
             -command => [\&groupdef_cbf_modify, 0, \@gw_data]);
   $btn->pack(-side => 'top', -padx => 5, -pady => 3);
   $btn = $frame->Button(-text => 'Remove Group', -width => $bwidth,
             -command => [\&groupdef_cbf_remove, \@gw_data]);
   $btn->pack(-side => 'top', -padx => 5, -pady => 3);
   $sub_frame->pack(-side => 'right', -fill => 'y');
   $frame->pack(-side => 'top', -fill => 'both', -expand => 'yes');

   #---------------------------------------
   #  Insert the current groups and build array
   #---------------------------------------
   my @data = ();    # Group definition data
   ($ngrp, $gmap) = names_calc_group_mapping($name_ref);
   for ($i = 1; $i <= $ngrp; $i++)
   {
      $ind_ref = names_get_group_indices($name_ref, $i, $gmap);
      $text    = names_build_group_string(\@new_names, $ind_ref);

      $gname = $new_names[ $$ind_ref[0] ];
      $gname =~ s/\[.*//;        # Leading string before '[';
      $text  = sprintf("%-16s : $text", $gname);

      $grp_lbox->insert('end', $text); 

      my @tmp = ( $gname, $ind_ref );
      push(@data, \@tmp);
   }

   #---------------------------------------
   #  Create a mode option and warning label
   #---------------------------------------
   $frame = $win_h->Frame(-borderwidth => 10);
   $btn   = $frame->Radiobutton(-variable => \$gw_data[$GWD_MODE], -value => 0, -text => 'Download groups only');
   $btn->pack(-side => 'left', -fill => 'both');
   $label = $frame->Label(-width => 1);
   $label->pack(-side => 'left');
   $btn   = $frame->Radiobutton(-variable => \$gw_data[$GWD_MODE], -value => 1, -text => 'Download groups and channel names');
   $btn->pack(-side => 'left', -fill => 'both');
   $label = $frame->Label(-width => 2);
   $label->pack(-side => 'left');
   $label = $frame->Label(-text => "NOTE: Server program must be running on LA to download.", -anchor => 'w');
   $label->pack(-side => 'left');
   $frame->pack(side => 'top', -fill => 'both', -expand => 'no');

   #---------------------------------------
   #  Create a frame to hold the action buttons
   #---------------------------------------
   $frame = $win_h->Frame(-borderwidth => 10);
   $bwidth = 12;
   $btn = $frame->Button(-text => 'Write All', -width => $bwidth,
             -command => [\&groupdef_cbf_write, $FALSE, \@gw_data]);
   $btn->pack(-side => 'left', -padx => 5);
   $btn = $frame->Button(-text => 'Write Selected', -width => $bwidth,
             -command => [\&groupdef_cbf_write, $TRUE, \@gw_data]);
   $btn->pack(-side => 'left', -padx => 5);
   $btn = $frame->Button(-text => 'Select All', -width => $bwidth,
             -command => [\&groupdef_cbf_select, $TRUE, \@gw_data]);
   $btn->pack(-side => 'left', -padx => 5);
   $btn = $frame->Button(-text => 'Select None', -width => $bwidth,
             -command => [\&groupdef_cbf_select, $FALSE, \@gw_data]);
   $btn->pack(-side => 'left', -padx => 5);
   $btn = $frame->Button(-text => 'Close Window', -width => $bwidth,
             -command => [$win_h => 'destroy']);
   $btn->pack(-side => 'right', -padx => 5);
   $frame->pack(side => 'top', -fill => 'both', -expand => 'no');

   #---------------------------------------
   #  Store widget handles and variables
   #---------------------------------------
   $gw_data[$GWD_WINDOW]   = $win_h;
   $gw_data[$GWD_NAME_REF] = \@new_names;
   $gw_data[$GWD_DATA_REF] = \@data;
   $gw_data[$GWD_BTN_CRT]  = $crt_btn;
   $gw_data[$GWD_BTN_MOD]  = $mod_btn;
   $gw_data[$GWD_LISTBOX]  = $grp_lbox;
   $gw_data[$GWD_MODE]     = $TRUE;
   $gw_data[$GWD_LAP_REF]  = $lap_ref;
   $gw_data[$GWD_WIN_SEL]  = 0;
}

#------------------------------------------------------------------------------

sub destroy_groupdef_window
{
   printf "\ndestroy_groupdef_window( @_ )\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   my $curr_wid = $_[0]; # Current widget/window handle from Tk
   my $gwd_ref  = $_[1];  # Reference to group window data
   my $cmd_ref  = $_[2]; # Command button reference
   my $win      = $$gwd_ref[$GWD_WINDOW];

   return if ( $curr_wid != $win ); # Return if not main window

   #---------------------------------------
   #  Clear out the GUI widget handles
   #---------------------------------------
   $$gwd_ref[$GWD_WINDOW]   = 0;
   $$gwd_ref[$GWD_BTN_CRT]  = 0;
   $$gwd_ref[$GWD_BTN_MOD]  = 0;
   $$gwd_ref[$GWD_LISTBOX]  = 0;

   #---------------------------------------
   #  Enable button from parent window
   #---------------------------------------
   if ( $$cmd_ref != 0 )
   {  $$cmd_ref->configure(-state => 'normal');
   }

   #---------------------------------------
   #  Destroy the group selection window (if open)
   #---------------------------------------
   if ( $$gwd_ref[$GWD_WIN_SEL] != 0 )
   {  $$gwd_ref[$GWD_WIN_SEL]->destroy();
      $$gwd_ref[$GWD_WIN_SEL] = 0;
   }

}

#------------------------------------------------------------------------------
#------------------------------------------------------------------------------

sub groupdef_namesel_create_cbf
{
   printf "\ngroupdef_namesel_create_cbf( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $gwd_ref  = $_[0];  # Reference to group window data
   my $name     = $_[1];  # Name or title for group of selected strings/names
   my $aref     = $_[2];  # Reference to array with indices of selected strings/names
   my $name_ref = $$gwd_ref[$GWD_NAME_REF]; # Reference to array of names
   my $grp_ref  = $$gwd_ref[$GWD_DATA_REF]; # Reference to group definition data
   my $grp_lbox = $$gwd_ref[$GWD_LISTBOX];  # Listbox containing available groups
   my $win      = $$gwd_ref[$GWD_WINDOW];
   my ($btn, $text, $ptr, $flag, $done, $i);

   if ($debug_flag & $DEBUG_GENERAL)
   {  printf "ref  : $gwd_ref = ( @$gwd_ref )\n";
      printf "name : $name\n";
      printf "list : ( @$aref )\n";
   }

   #---------------------------------------
   #  Enable the "Create" and "Modify" buttons
   #---------------------------------------
   $btn = $$gwd_ref[$GWD_BTN_CRT];  # button from group definition window
   $btn->configure(-state => 'normal');
   $btn = $$gwd_ref[$GWD_BTN_MOD];  # button from group definition window
   $btn->configure(-state => 'normal');

   #---------------------------------------
   #  Add as only element to the list
   #---------------------------------------
   if ( $#$grp_ref < 0 )
   {  $text = names_build_group_string($name_ref, $aref);
      $text = sprintf("%-16s : $text", $name);
      $grp_lbox->insert('end', $text); 
      my @data = ( $name, $aref );
      push(@$grp_ref, \@data);
      return;
   }

   #---------------------------------------
   #  Insert into correct location 
   #---------------------------------------
   $ptr = $$grp_ref[0];
   if ( $name le $$ptr[0] )
   {  $text = names_build_group_string($name_ref, $aref);
      $text = sprintf("%-16s : $text", $name);
      $grp_lbox->insert(0, $text); 
      my @data = ( $name, $aref );
      unshift(@$grp_ref, \@data);
      $flag = ( $name eq $$ptr[0] );
   }
   else
   {  for ($i = 1, $done = $FALSE; (($i <= $#$grp_ref) && !$done); $i++)
      {  $ptr = $$grp_ref[$i];
         if ( $name le $$ptr[0] )
         {  $text = names_build_group_string($name_ref, $aref);
            $text = sprintf("%-16s : $text", $name);
            $grp_lbox->insert($i, $text); 
            my @data = ( $name, $aref );
            for ($j = $#$grp_ref; $j >= $i; $j--)
            {  $$grp_ref[$j+1] = $$grp_ref[$j];
            }
            $$grp_ref[$i] = \@data;
            $flag = ( $name eq $$ptr[0] );
            $done = $TRUE;
         }
      }
      if ( !$done )
      {  $text = names_build_group_string($name_ref, $aref);
         $text = sprintf("%-16s : $text", $name);
         $grp_lbox->insert('end', $text); 
         my @data = ( $name, $aref );
         push(@$grp_ref, \@data);
         $flag = $FALSE;
      }
   }

   #---------------------------------------
   #  Warning if this is a duplicate name
   #---------------------------------------
   if ( $flag )
   {  $win->messageBox(-icon => 'info', -type => 'OK',
         -message => "Multiple groups defined with the name \"$name\".",
         -title => 'Group Download Warning');
   }
}

#------------------------------------------------------------------------------

sub groupdef_namesel_modify_cbf
{
   printf "\ngroupdef_namesel_modify_cbf( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $gwd_ref  = $_[0];  # Reference to group window data
   my $name     = $_[1];  # Name or title for group of selected strings/names
   my $aref     = $_[2];  # Reference to array with indices of selected strings/names
   my $name_ref = $$gwd_ref[$GWD_NAME_REF]; # Reference to array of names
   my $grp_ref  = $$gwd_ref[$GWD_DATA_REF]; # Reference to group definition data
   my $grp_lbox = $$gwd_ref[$GWD_LISTBOX];  # Listbox containing available groups
   my $ind      = $$gwd_ref[$GWD_INDEX];     # Index of group within listbox
   my ($btn, $text, $ptr);

   if ($debug_flag & $DEBUG_GENERAL)
   {  printf "ref  : $gwd_ref = ( @$gwd_ref )\n";
      printf "name : $name\n";
      printf "list : ( @$aref )\n";
   }

   #---------------------------------------
   #  Enable the "Create" and "Modify" buttons
   #---------------------------------------
   $btn = $$gwd_ref[$GWD_BTN_CRT];  # button from group definition window
   $btn->configure(-state => 'normal');
   $btn = $$gwd_ref[$GWD_BTN_MOD];  # button from group definition window
   $btn->configure(-state => 'normal');

   #---------------------------------------
   #  Delete existing and insert modified group
   #---------------------------------------
   $grp_lbox->delete($ind); 
   $ptr = $$grp_ref[$ind];
   if ( $$ptr[0] eq $name )  # Name has not changed
   {  $text = names_build_group_string($name_ref, $aref);
      $text = sprintf("%-16s : $text", $name);
      $grp_lbox->insert($ind, $text); 
      $$ptr[0] = $name;
      $$ptr[1] = $aref;
   }
   else
   {  splice(@$grp_ref, $ind, 1);
      groupdef_namesel_create_cbf($gwd_ref, $name, $aref);
   }
}

#------------------------------------------------------------------------------

sub groupdef_namesel_cancel_cbf
{
   printf "\ngroupdef_namesel_cancel_cbf( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $gwd_ref = $_[0];  # Reference to group window data
   my $win     = $$gwd_ref[$GWD_WINDOW];

   #---------------------------------------
   #  Return if main window already destroyed
   #---------------------------------------
   return if ( $win == 0 );

   #---------------------------------------
   #  Enable the "Create" and "Modify" buttons
   #---------------------------------------
   $btn = $$gwd_ref[$GWD_BTN_CRT];  # button from group definition window
   $btn->configure(-state => 'normal');
   $btn = $$gwd_ref[$GWD_BTN_MOD];  # button from group definition window
   $btn->configure(-state => 'normal');
}

#------------------------------------------------------------------------------
#------------------------------------------------------------------------------

sub groupdef_cbf_create
{
   printf "\ngroupdef_cbf_create( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $gwd_ref  = $_[0]; # Reference to group window data
   my $name_ref = $$gwd_ref[$GWD_NAME_REF]; # Reference to array of names
   my $grp_ref  = $$gwd_ref[$GWD_DATA_REF]; # Reference to group definition data
   my $grp_lbox = $$gwd_ref[$GWD_LISTBOX];  # Listbox containing available groups
   my $btn;

   #---------------------------------------
   #  Disable the "Create" and "Modify" buttons
   #---------------------------------------
   $btn = $$gwd_ref[$GWD_BTN_CRT];  # button from group definition window
   $btn->configure(-state => 'disable');
   $btn = $$gwd_ref[$GWD_BTN_MOD];  # button from group definition window
   $btn->configure(-state => 'disable');

   #---------------------------------------
   #  Create window: create new group
   #---------------------------------------
   my $mode = $NAMESEL_MODE_SHOW_NAME;
   create_namesel_window($seltitle, $mode, "Create", $name_ref, "", NULL,
      \&groupdef_namesel_create_cbf, \&groupdef_namesel_cancel_cbf, $gwd_ref,
      \$$gwd_ref[$GWD_WIN_SEL]);
}

#------------------------------------------------------------------------------

sub groupdef_cbf_modify
{
   printf "\ngroupdef_cbf_modify( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $rsvd     = $_[0]; # Handle if called from listbox double-click
   my $gwd_ref  = $_[1]; # Reference to group window data
   my $name_ref = $$gwd_ref[$GWD_NAME_REF]; # Reference to array of names
   my $grp_ref  = $$gwd_ref[$GWD_DATA_REF]; # Reference to group definition data
   my $grp_lbox = $$gwd_ref[$GWD_LISTBOX];  # Listbox containing available groups
   my $win      = $$gwd_ref[$GWD_WINDOW];
   my ($btn, @list, $cur, $ptr, $gname, $iref);

   #---------------------------------------
   #  Get current groups(s) from listbox
   #---------------------------------------
   @list = $grp_lbox->curselection(); 
   if ( $#list < 0 )
   {  $win->messageBox(-icon => 'error', -type => 'OK',
         -message => "No group selected from list for modification.",
         -title => 'Group Download Warning');
      return;
   }

   #---------------------------------------
   #  Disable the "Create" and "Modify" buttons
   #---------------------------------------
   $btn = $$gwd_ref[$GWD_BTN_CRT];  # button from group definition window
   $btn->configure(-state => 'disable');
   $btn = $$gwd_ref[$GWD_BTN_MOD];  # button from group definition window
   $btn->configure(-state => 'disable');

   #---------------------------------------
   #  Get information for first group
   #---------------------------------------
   $cur   = $list[0];
   $ptr   = $$grp_ref[$cur];
   $gname = $$ptr[0];
   $iref  = $$ptr[1];
   $$gwd_ref[$GWD_INDEX] = $cur;

   #---------------------------------------
   #  Create window: modify group
   #---------------------------------------
   my $mode = $NAMESEL_MODE_SHOW_NAME;
   create_namesel_window($seltitle, $mode, "Modify", $name_ref, $gname, $iref,
      \&groupdef_namesel_modify_cbf, \&groupdef_namesel_cancel_cbf, $gwd_ref,
      \$$gwd_ref[$GWD_WIN_SEL]);
}


#------------------------------------------------------------------------------

sub groupdef_cbf_remove
{
   printf "\ngroupdef_cbf_remove( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $gwd_ref  = $_[0]; # Reference to group window data
   my $name_ref = $$gwd_ref[$GWD_NAME_REF]; # Reference to array of names
   my $grp_ref  = $$gwd_ref[$GWD_DATA_REF]; # Reference to group definition data
   my $grp_lbox = $$gwd_ref[$GWD_LISTBOX];  # Listbox containing available groups
   my $win      = $$gwd_ref[$GWD_WINDOW];
   my (@list, $cur, $i);

   #---------------------------------------
   #  Get current groups(s) from listbox
   #---------------------------------------
   @list = $grp_lbox->curselection(); 
   if ( $#list < 0 )
   {  $win->messageBox(-icon => 'error', -type => 'OK',
         -message => "No group(s) selected from list for removal.",
         -title => 'Group Download Warning');
      return;
   }

   #---------------------------------------
   #  Remove the selected group definition(s)
   #---------------------------------------
   for ($i = $#list; $i >= 0; $i--)
   {  $cur = $list[$i];
      $grp_lbox->delete($cur);
      splice(@$grp_ref, $cur, 1);
   }
}


#------------------------------------------------------------------------------

sub groupdef_cbf_select
{
   printf "\ngroupdef_cbf_select( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $flag     = $_[0];  # Flag: $TRUE=select, $FALSE=deselect
   my $gwd_ref  = $_[1]; # Reference to group window data
   my $grp_lbox = $$gwd_ref[$GWD_LISTBOX];  # Listbox containing available groups

   if ( $flag )
   {  $grp_lbox->selection('set', 0, 'end');
   }
   else
   {  $grp_lbox->selection('clear', 0, 'end');
   }
}

#------------------------------------------------------------------------------

sub groupdef_cbf_write
{
   printf "\ngroupdef_cbf_write( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $sel_flag = $_[0];  # Flag: $TRUE=write selected, $FALSE=write all
   my $gwd_ref  = $_[1]; # Reference to group window data
   my $name_ref = $$gwd_ref[$GWD_NAME_REF]; # Reference to group definition data
   my $grp_ref  = $$gwd_ref[$GWD_DATA_REF]; # Reference to group definition data
   my $grp_lbox = $$gwd_ref[$GWD_LISTBOX];  # Listbox containing available groups
   my $lap_ref  = $$gwd_ref[$GWD_LAP_REF];  # Reference to LA parameters
   my $win      = $$gwd_ref[$GWD_WINDOW];
   my ($i, $cur, $ptr, $gname, @list, $iref, $text, @data, $lab_text, $chn_text);

   #---------------------------------------
   #  Get current groups(s) from listbox
   #---------------------------------------
   if ( $sel_flag )
   {  @list = $grp_lbox->curselection(); 
      if ( $#list < 0 )
      {  $win->messageBox(-icon => 'error', -type => 'OK',
            -message => "No group(s) selected from list to write.",
            -title => 'Group Download Warning');
         return;
      }
   }
   else
   {  for ($i = 0; $i <= $#$grp_ref; $i++)
      {  $list[$i] = $i;
      }
   }

   #---------------------------------------
   #  Connect to the external LA
   #---------------------------------------
   la_connect($$lap_ref[$LA_PARM_HOST], $$lap_ref[$LA_PARM_PORT]);
   return if ( $ecode < 0 );

   #---------------------------------------
   #  Process each group selection
   #---------------------------------------
   foreach $cur ( @list )
   {
      $ptr   = $$grp_ref[$cur];  # Pointer to group data
      $gname = $$ptr[0];         # Group name
      $iref  = $$ptr[1];         # Reference to array of indexes

      # Process first element of group
      $text = $$name_ref[$$iref[0]];
      @data = split('=', $text, 2);
      $lab_text = $data[0];
      $chn_text = $data[1];

      if ( ($data[0] =~ m/,/) || ($data[1] =~ m/,/) )
      {  la_disconnect();
         $win->messageBox(-icon => 'error', -type => 'OK',
            -message => "Testpoint and channel names cannot contain comma characters.",
            -title => 'Group Download Error');
         return;
      }

      # Process remaining elements of group
      for ($i = 1; $i <= $#$iref; $i++)
      {  $text = $$name_ref[$$iref[$i]];
         @data = split('=', $text, 2);
         $lab_text .= "," . $data[0]; 
         $chn_text .= "," . $data[1];
         if ( ($data[0] =~ m/,/) || ($data[1] =~ m/,/) )
         {  la_disconnect();
            $win->messageBox(-icon => 'error', -type => 'OK',
               -message => "Testpoint and channel names cannot contain comma characters.",
               -title => 'Group Download Error');
            return;
         }
      }

      $text = ( ($$gwd_ref[$GWD_MODE]) ? "group and channel names" : "group" );
      printf "\nDownloading $text to LA: module=%s\n", $$lap_ref[$LA_PARM_MODULE];
      printf "\tGroup name : \"$gname\"\n";
      printf "\tChannels   : \"$chn_text\"\n";
      printf "\tLabels     : \"$lab_text\"\n";

      #---------------------------------------
      #  Download channel names to LA (if selected)
      #---------------------------------------
      if ( $$gwd_ref[$GWD_MODE] )
      {  la_module_set_channel_name($$lap_ref[$LA_PARM_MODULE], $chn_text, $lab_text);
         if ( $ecode < 0 )
         {  la_disconnect();
            return;
         } 
      }

      #---------------------------------------
      #  Download group definition to LA
      #---------------------------------------
      la_module_set_channel_group($$lap_ref[$LA_PARM_MODULE], $gname, $lab_text);
      if ( $ecode < 0 )
      {  la_disconnect();
         return;
      } 
   }

   #---------------------------------------
   #  Close connection to external LA
   #---------------------------------------
   la_disconnect();
}


###############################################################################

sub names_build_group_string
{
   printf "\nnames_build_group_string( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $name_ref = $_[0]; # Reference to array of names
   my $ind_ref  = $_[1]; # Reference to array of group indices
   my @list     = ();
   my ($str, $i, $ind);

   #---------------------------------------
   #  Build array with elements of the group
   #---------------------------------------
   for ($i = 0; $i <= $#$ind_ref; $i++)
   {  $ind = $$ind_ref[$i];
      push(@list, $$name_ref[$ind]);
   }
   return("") if ( $#list < 0 );  # No elements in group

   #---------------------------------------
   #  Build a string of group elements
   #---------------------------------------
   $str = $list[0];
   for ($i = 1; $i <= $#list; $i++)
   {  $str .= "," . $list[$i];
   }
   return($str);
}

#------------------------------------------------------------------------------

sub names_get_group_indices
{
   printf "\nnames_get_group_indices( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $name_ref = $_[0]; # Reference to array of names
   my $grp      = $_[1]; # Index of group to process
   my $map_ref  = $_[2]; # Mapping of names to groups
   my $flag     = (($#_ > 2) ? $_[3] : $TRUE);
   my @list     = ();
   my @inds     = ();
   my $num      = $#$name_ref;
   my ($str, $i, $j, $k, $bit, $done);

   #---------------------------------------
   #  Sort the elements by bit position in group
   #---------------------------------------
   for ($i = 0, $done=$FALSE; (($i <= $num) && !$done); $i++)
   {  if ( $$map_ref[$i] == $grp )
      {  $prefix = $$name_ref[$i];   # Copy this testpoint name
         $prefix =~ s/\[.*//;        # Leading string before '[';
         $bit = $$name_ref[$i];      # Copy this testpoint name
         $bit  =~ s/\].*//;          # Remove trailing string after ']';
         $bit  =~ s/^$prefix\[//;    # Remove prefix string
         $list[0] = $bit;            # Store bit position in group
         $inds[0] = $i;              # Store index of testpoint name
         $done = $TRUE;
      }
   } 
   return("") if ( !$done );  # No elements in group
   for ( ; $i <= $num; $i++)
   {  if ( $$map_ref[$i] == $grp )
      {  $bit = $$name_ref[$i];
         $bit  =~ s/\].*//;        # Remove trailing string after ']';
         $bit  =~ s/^$prefix\[//;  # Remove prefix string
         for ($j=0, $done=$FALSE; (($j <= $#list) && !$done); $j++)
         {  
            if ( $bit < $list[$j] )
            {  for ($k = ($#list + 1); $k > $j; $k--)
               {  $list[$k] = $list[($k-1)];
                  $inds[$k] = $inds[($k-1)];
               }
               $list[$j] = $bit;
               $inds[$j] = $i;
               $done = $TRUE;
            }
         }
         if ( !$done )
         {  push(@list, $bit);
            push(@inds, $i);
         }
      }
   }
   if ( $flag ) # Put in descending order?
   {  @inds = reverse(@inds);
   }
   return(\@inds);
}


#------------------------------------------------------------------------------

sub names_calc_group_mapping
{  
   printf "\nnames_calc_group_mapping( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $name_ref = $_[0]; # Reference to array of names
   my @mapping  = ();    # Index of group for each name
   my $grp      = 0;     # Index of current group
   my @names    = ();    # Array containing group names
   my @count    = ();    # Number of members in each group
   my ($num, $pos, $rem, $i, $cnt, $str, $sub);

   #---------------------------------------
   #  Initialize all group indices to zero
   #---------------------------------------
   $num   = $#$name_ref + 1;
   for ($i = 0; $i < $num; $i++)
   {  $mapping[$i] = 0;
   }

   #---------------------------------------
   #  Initialize variables for loop
   #---------------------------------------
   $pos = 0;        # Initial target index
   $rem = $num;     # Number of targets remaining
   $grp = 1;        # Current group index
   $names[0] = "";  # Group 0 is unused
   $count[0] = 0;   # Group 0 is unused

   #---------------------------------------
   #  Process the list of names
   #---------------------------------------
   while ( $rem > 0 ) # continue until all targets in a "group"
   {  $cnt = 1;                # Number of targets in this group
      $str = $$name_ref[$pos]; # Extract current string name
      $str =~ s/\[.*//;        # Leading string before '[';
      if ( ($str ne "") && ($str ne $$name_ref[$pos]) )
      {
         #---------------------------------------
         #  Move thru remaining targets (matching base name?)
         #---------------------------------------
         for ($i=($pos+1); $i < $num; $i++)
         {
            next if ( $mapping[$i] );

            $sub = $$name_ref[$i];           # Extract current string name
            $sub =~ s/\[.*//;                # Leading string before '['
            if (  ($sub ne $$name_ref[$i]) && ($str eq $sub) )
            {  $mapping[$i] = $grp;
               $cnt++;
            }
         }
   
         #---------------------------------------
         #  Update information for valid group
         #---------------------------------------
         if ( $cnt > 1 )
         {  $mapping[$pos] = $grp;  # Store the group index for this target
            $names[$grp]   = $str;  # Store the base name of the group
            $count[$grp]   = $cnt;  # Store the number of entries in this group
            $grp++;                 # Increment group number
         }
      }
      $rem -= $cnt;  # Update number of targets remaining
      for ($i=($pos+1), $pos=0; (($i < $num) && ($pos == 0)); $i++)
      {  $pos = $i if ( !$mapping[$i] );
      }
   }

   if ( $debug_flag & $DEBUG_DETAILS)
   {  printf("\nSUMMARY:\n");
      printf("NUM     = %u\n", ($grp - 1));
      printf("NAMES   = ( @names )\n");
      printf("COUNT   = ( @count )\n");
      printf("\n");
      printf("SIGNALS = ( @$name_ref )\n");
      printf("GROUP   = ( @mapping )\n");
   }

   #---------------------------------------
   #  Return number of groups and mapping
   #---------------------------------------
   return(($grp-1), \@mapping);
}


#------------------------------------------------------------------------------

1;
