#
#/**************************************************************************
# *                                                                        *
# *               Copyright (C) 1989-1999 Silicon Graphics, Inc.           *
# *                                                                        *
# *  These coded instructions, statements, and computer programs  contain  *
# *  unpublished  proprietary  information of Silicon Graphics, Inc., and  *
# *  are protected by Federal copyright law.  They  may  not be disclosed  *
# *  to  third  parties  or copied or duplicated in any form, in whole or  *
# *  in part, without the prior written consent of Silicon Graphics, Inc.  *
# *                                                                        *
# **************************************************************************/
#
# @(#)$Id: jtag_subs.pm,v 1.13 2005/07/14 17:04:41 beighley Exp $

$FALSE  = 0;
$TRUE   = 1;


#-------------------------------------------------------------------------------
#
#  FILE: jtag_subs.pm
#
#  DESCRIPTION:  This file contains PERL subroutines to handle reading and
#     writing JTAG instruction registers (IRs) and test data registers (TDRs).
#
#  NOTES: The "common_defs.pm" file is required when this file is used.
#
#  SUBROUTINES:  The following subroutines are contained in this file:
#     (expected parameter usage displayed; some have optional parameters)
#
#
#
#  MODIFICATION HISTORY:
#
#     WHEN      WHO             WHAT
#     --------  --------------  -----------------------------------------
#     xx/xx/xx  Karen Beighley  Original source code.
#     12/12/01  Karen Beighley  Added function: jtag_lib_msg_level()
#                               Added "use" to include JTAGAccess library.
#     12/17/01  Karen Beighley  Added functions: jtag_simacc_time() and
#                                  jtag_simacc_dump().
#                               Added definition of $USE_JTAG_SUBS_PM.
#     02/25/02  Karen Beighley  Added use of common error function pointer.
#     06/18/02  Karen Beighley  Added function: jtag_set_psi()
#     09/09/03  Karen Beighley  Added functions for stop-clock and internal
#                                  scan chain dumping.
#     02/10/04  Karen Beighley  Modified to target multi-chain modules. 
#     07/14/05  Karen Beighley  Added new subroutine: jtag_dr_load()
#
#-------------------------------------------------------------------------------

#use JTAGAccess ':all';    # JTAGAccess library built from C source code
use JTAGAccess;    # JTAGAccess library built from C source code

$USE_JTAG_SUBS_PM = 1;    # Define to allow test for module inclusion

#-------------------------------------------------------------------------------
#  CONSTANTS
#-------------------------------------------------------------------------------

$JTAG_RESET_HARD = 0;
$JTAG_RESET_SOFT = 1;

$JTAG_STATE_TLR  = 0;
$JTAG_STATE_RTI  = 1;
$JTAG_STATE_PIR  = 2;
$JTAG_STATE_PDR  = 3;

$JTAG_ACCESS_SYSCO  = 0;
$JTAG_ACCESS_VERSIM = 1;
$JTAG_ACCESS_SN0SB  = 2;
$JTAG_ACCESS_NONE   = 3;

$JTAG_JUST_LEFT     = 0;
$JTAG_JUST_RIGHT    = 1;

$JTAG_SIMACC_VEC    = 0;
$JTAG_SIMACC_INT    = 1;
$JTAG_SIMACC_SCAL   = 2;
$JTAG_SIMACC_STR    = 3;
$JTAG_MAX_SIMACC    = 4;

$JTAG_SIMACC_VCD_DUMP_ON  = 0x20;
$JTAG_SIMACC_VCD_DUMP_OFF = 0x21;
$JTAG_SIMACC_VPD_DUMP_ON  = 0x22;
$JTAG_SIMACC_VPD_DUMP_OFF = 0x23;

$DEFAULT_JTAG_PORT  = 8000;
$SYSCO_COPY_INVALID = 0;


#-------------------------------------------------------------------------------
#  GLOBAL VARIABLES
#-------------------------------------------------------------------------------

$jtag_init = $FALSE;
$jtag_host = "localhost";
$jtag_port = $DEFAULT_JTAG_PORT;
$jtag_id   = 0;
$jtag_cnum = 0;
$jtag_mnum = 0;
$jtag_parm = $SYSCO_COPY_INVALID;

$jtag_access = $JTAG_ACCESS_NONE;
@jtag_access_names = ( "SYSCO", "VERSIM", "SN0SB", "NONE" );

@jtag_reset_names = ( "HARD", "SOFT" );
@jtag_state_names = ( "TEST-LOGIC-RESET", "RUN-TEST-IDLE", "PAUSE-IR", "PAUSE-DR" );

$jtag_just = $JTAG_JUST_LEFT;
@jtag_just_names = ( "LEFT-JUSTIFIED", "RIGHT-JUSTIFIED" );


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: jtag_target()
#
#    USAGE: jtag_target($ctlr_id)
#           jtag_target($ctlr_id, $mod_ind)
#           jtag_target($ctlr_id, $mod_ind, $chn_ind)
#           jtag_target($ctlr_id, $mod_ind, $chn_ind, $sel_flag)
#
#    DESCRIPTION: The subroutine will define the target of future JTAG
#       operations.  Each module has a unique TAP controller ID, and may have
#       multiple scan chains (SICs in parallel or multi-port L1).
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub jtag_target
{
   print "\njtag_target(@_)\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   $jtag_id   = $_[0];  # Save the controller ID
   $jtag_mnum = ( ($#_ > 0) ? $_[1] : 0 );
   $jtag_cnum = ( ($#_ > 1) ? $_[2] : 0 );
   my $flag   = ( ($#_ > 2) ? $_[3] : $FALSE );

   # Store controller ID within C-library for future operations
   pcf_jtag_target($jtag_id, $jtag_mnum, $jtag_cnum, $flag, $ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess);
      return;
   }
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: jtag_lib_msg_level()
#
#    USAGE: jtag_lib_msg_level($level)
#
#    DESCRIPTION: Set the debug message level within the JTAGAccess library.
#       The "level" parameter is a bitmask using a combination of:
#       $DEBUG_FUNCTIONS, $DEBUG_GENERAL, $DEBUG_DETAILS, $DEBUG_VERBOSE
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub jtag_lib_msg_level
{
   my $num = $#_ + 1;
   print "\njtag_lib_msg_level(@_) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 1 )
   {  $emess = "Invalid parameters passed to subroutine jtag_lib_msg_level().";
      &$comm_err_fun($emess);
      return;
   }

   pcf_jtag_msg_level( $_[0] );  # set the message level within the library
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: jtag_connect()
#
#    USAGE: jtag_connect($jtag_access)             
#           jtag_connect($jtag_access, $jtag_host)             
#           jtag_connect($jtag_access, $jtag_host, $jtag_port)
#           jtag_connect($jtag_access, $jtag_host, $jtag_port, $jtag_parm)
#
#    DESCRIPTION: Open a connection to the specified JTAG controller.
#       The "jtag_access" parameter defines the JTAG controller access
#       mechanism: the system controller ($JTAG_ACCESS_SYSCO), 
#       a VERILOG simultor ($JTAG_ACCESS_VERSIM), an SN0-style scanbox
#       ($JTAG_ACCESS_SN0SB) or a software emulator ($JTAG_ACCESS_NONE).
#       The "jtag_host" and "jtag_port" parameters are required when the
#       access is $JTAG_ACCESS_SYSCO and $JTAG_ACCESS_VERSIM, and they
#       are ignored for the other access methods.  The default for
#       the "jtag_host" is "localhost", and the default port is 8000.
#       The option "jtag_parm" is used for the system controller access,
#       and contains the value associated with "--copy".
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub jtag_connect
{
   my $num = $#_ + 1;
   print "\njtag_connect(@_) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   jtag_disconnect()  if ( $jtag_init == $TRUE );

   if ( $num == 4 )
   {  $jtag_host = $_[1];
      $jtag_port = defined($_[2]) ? $_[2] : $DEFAULT_JTAG_PORT;
      $jtag_parm = defined($_[3]) ? $_[3] : $SYSCO_COPY_INVALID;
   }
   elsif ( $num == 3 )
   {  $jtag_host = $_[1];
      $jtag_port = defined($_[2]) ? $_[2] : $DEFAULT_JTAG_PORT;
      $jtag_parm = $SYSCO_COPY_INVALID;
   }
   elsif ( $num == 2 )
   {  $jtag_host = $_[1];
      $jtag_port = $DEFAULT_JTAG_PORT;
      $jtag_parm = $SYSCO_COPY_INVALID;
   }
   elsif ( $num != 1 )
   {  $emess = "Invalid parameters passed to subroutine jtag_connect().";
      &$comm_err_fun($emess);
      return;
   }

   $jtag_access = $_[0];
   if ( ($jtag_access < 0) || ($jtag_access > $JTAG_ACCESS_NONE) )
   {  $emess = "Unknown access parameter passed to subroutine jtag_connect().";
      &$comm_err_fun($emess);
      return;
   }

   if ($debug_flag & $DEBUG_GENERAL)
   {  print "access method: $jtag_access = ", $jtag_access_names[$jtag_access], "\n";
      print "host=\"$jtag_host\", port=$jtag_port, parm=$jtag_parm\n";
   }

   if ( $jtag_access == $JTAG_ACCESS_VERSIM )
   {  $ENV{JTAG_HOST} = $jtag_host;
      $ENV{JTAG_PORT} = $jtag_port;
   }

   #  Call C-function that does the "real" work
   pcf_jtag_connect($jtag_access, $jtag_host, $jtag_port, $jtag_parm, $ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess);
      return;
   }

   $jtag_init = $TRUE;
   pcf_jtag_set_just($jtag_just);  # set the default data justification
   return;
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: jtag_disconnect()
#
#    USAGE: jtag_disconnect()
#
#    DESCRIPTION: Close a connection to the logic analyzer (pre-opened). 
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub jtag_disconnect
{
   print "\njtag_disconnect(@_)\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $jtag_init == $TRUE )
   {  
      pcf_jtag_disconnect();
      printf "\nClosing connection to JTAG controller: host=$jtag_host, port=$jtag_port\n";
   }
   $jtag_init = $FALSE;
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: jtag_set_data_just()
#
#    USAGE: jtag_set_data_just($JTAG_JUST_LEFT|$JTAG_JUST_RIGHT)
#
#    DESCRIPTION: The subroutine sets the JTAG data justification.
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub jtag_set_data_just
{
   my $num = $#_ + 1;
   print "\njtag_set_data_just(@_) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 1 )
   {  $emess = "Invalid parameters passed to subroutine jtag_set_data_just().";
      &$comm_err_fun($emess);
      return;
   }
   $jtag_just = $_[0];
   if ( ($jtag_just != $JTAG_JUST_LEFT) && ($jtag_just != $JTAG_JUST_RIGHT) )
   {  $emess = "Unknown justification passed to subroutine jtag_set_data_just().";
      &$comm_err_fun($emess);
      return;
   }

   pcf_jtag_set_just($jtag_just);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: jtag_get_data_just()
#
#    USAGE: $just = jtag_get_data_just()
#
#    DESCRIPTION: The subroutine will get the current JTAG data justification.
#
#    RETURN VALUE: Current justification: $JTAG_DATA_LEFT or $JTAG_DATA_RIGHT
#
#-------------------------------------------------------------------------------

sub jtag_get_data_just
{
   my $num = $#_ + 1;
   print "\njtag_get_data_just(@_) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 0 )
   {  $emess = "Invalid parameters passed to subroutine jtag_get_data_just().";
      &$comm_err_fun($emess);
      return;
   }

   $jtag_just = pcf_jtag_get_just();
   return($jtag_just);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: jtag_set_psi()
#
#    USAGE: jtag_set_psi($psi_value)
#
#    DESCRIPTION: The subroutine will define the value of the PSI signal for
#       future operations.  
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub jtag_set_psi
{
   print "\njtag_set_psi(@_)\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $#_ != 0 )
   {  $emess = "Invalid parameters passed to subroutine jtag_set_psi().";
      &$comm_err_fun($emess);
      return;
   }

   # Save the PSI value for future use
   pcf_jtag_set_psi($_[0], $ecode, $emess);
   &$comm_err_fun($emess)   if ( $ecode < 0 );
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: jtag_reset()
#
#    USAGE: jtag_reset()
#           jtag_reset($JTAG_RESET_HARD|$JTAG_RESET_SOFT)
#           jtag_reset($JTAG_RESET_HARD|$JTAG_RESET_SOFT, $end_state)
#
#    DESCRIPTION: The subroutine will reset the JTAG TAP controller.  A HARD
#      reset uses the TRSTN signal; a SOFT reset uses TMS and TCK.  The end
#      state can be TEST-LOGIC-RESET or RUN-TEST-IDLE.  The default reset
#      type is $JTAG_RESET_SOFT and the default end-state is $JTAG_STATE_TLR.
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub jtag_reset
{
   my $num = $#_ + 1;
   print "\njtag_reset(@_) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   my $rtype = $JTAG_RESET_SOFT;
   my $state = $JTAG_STATE_TLR;

   if ( $num == 2 )
   {  $rtype = $_[0];
      if ( ($rtype != $JTAG_RESET_HARD) && ($rtype != $JTAG_RESET_SOFT) )
      {  $emess = "Invalid reset type passed to subroutine jtag_reset().";
         &$comm_err_fun($emess);
         return;
      }
      $state = $_[1];
      if ( ($state != $JTAG_STATE_TLR) && ($state != $JTAG_STATE_RTI) )
      {  $emess = "Invalid end-state passed to subroutine jtag_reset().";
         &$comm_err_fun($emess);
         return;
      }
   }
   elsif ( $num == 1 )
   {  $rtype = $_[0];
      if ( ($rtype != $JTAG_RESET_HARD) && ($rtype != $JTAG_RESET_SOFT) )
      {  $emess = "Invalid reset type passed to subroutine jtag_reset().";
         &$comm_err_fun($emess);
         return;
      }
   }
   elsif ( $num != 0 )
   {  $emess = "Invalid parameters passed to subroutine jtag_reset().";
      &$comm_err_fun($emess);
      return;
   }

   if ( $debug_flag & $DEBUG_DETAILS )
   {  print "jtag reset parameters:\n";
      print "reset type : $rtype = $jtag_reset_names[$rtype]\n";
      print "end-state  : $state = $jtag_state_names[$state]\n";
   }

   pcf_jtag_reset($rtype, $state, $ecode, $emess);
   &$comm_err_fun($emess)   if ( $ecode < 0 );
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: jtag_state_to_rti()
#
#    USAGE: jtag_state_to_rti()
#           jtag_state_to_rti($num)
#
#    DESCRIPTION: The subroutine will move the TAP controller to RUN-TEST-IDLE.
#       If the "num" parameter is given, and is non-zero, than that number of
#       TCKs is issued after reaching RUN-TEST-IDLE.
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub jtag_state_to_rti
{
   my $num = $#_ + 1;
   print "\njtag_state_to_rti(@_) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   my $nclks = 0;
   if ( $num == 1 )
   {  $nclks = $_[0];
   }
   elsif ( $num != 0 )
   {  $emess = "Invalid parameters passed to subroutine jtag_state_to_rti().";
      &$comm_err_fun($emess);
      return;
   }

   print "jtag_state_to_rti: #tcks = $nclks\n" if ($debug_flag & $DEBUG_GENERAL);

   pcf_jtag_state_to_rti($nclks, $ecode, $emess);
   &$comm_err_fun($emess)   if ( $ecode < 0 );
}

#-------------------------------------------------------------------------------
#
#    SUBROUTINE: jtag_dr_load()
#
#    USAGE: jtag_dr_load($ilen, \@ir_tdi, $dlen, \@dr_tdi)
#           jtag_dr_load($ilen, \@ir_tdi, $dlen, \@dr_tdi, $flag)
#
#    DESCRIPTION: The subroutine will perform both IR and DR scan operations.
#      The number of bits shifted in/out of the instruction register is
#      specified by "ilen", and "dlen" contains the length of the corresponding
#      data registers.  The TDI data ("ir_tdi" and "dr_tdi") must be an array
#      of 32-bit values.  The TDI data can be left or right-justified, as
#      defined using "jtag_set_just()".  After the each shift is complete, the
#      TAP controller is moved thru UPDATE-IR and then to RUN-TEST-IDLE.
#      
#      If the "flag" parameter is provided and is $TRUE, then a read-recirculate
#      operation is performed on the data register(s).  Default: the given
#      DR TDI data is shifted in. The sampled DR TDO data is returned.
#      
#    RETURN VALUE: An array containing the TDO data (same format as TDI).
#
#-------------------------------------------------------------------------------

sub jtag_dr_load
{
   my $num = $#_ + 1;
   print "\njtag_dr_load(@_) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   my $flag   = ( ($num == 5) ? $_[4] : $FALSE );
   my $ilen   = $_[0];
   my $ir_tdi = $_[1];
   my $dlen   = $_[2];
   my $dr_tdi = $_[3];
   if ( ref($ir_tdi) ne 'ARRAY' )
   {  $emess = "Invalid parameter passed to subroutine jtag_dr_load().";
      $emess .= "\nThe 2nd parameter must be an ARRAY reference.";
      &$comm_err_fun($emess);
      return;
   }
   if ( ref($dr_tdi) ne 'ARRAY' )
   {  $emess = "Invalid parameter passed to subroutine jtag_dr_load().";
      $emess .= "\nThe 4th parameter must be an ARRAY reference.";
      &$comm_err_fun($emess);
      return;
   }

   if ( $debug_flag & $DEBUG_DETAILS )
   {  printf "ir_len     = %lu\n", $ilen;
      printf "ir_data[0] = 0x%08lX\n", $$ir_data[0];
      printf "dr_len     = %lu\n", $dlen;
      printf "dr_data[0] = 0x%08lX\n", $$dr_data[0];
   }

   my $rv = pcf_jtag_dr_load($flag, $ilen, $ir_tdi, $dlen, $dr_tdi, $ecode, $emess);
   &$comm_err_fun($emess)   if ( $ecode < 0 );
   return(@$rv);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: jtag_scan_ir()
#
#    USAGE: jtag_scan_ir($len, \@tdi_data)
#           jtag_scan_ir($len, \@tdi_data, $end_state)
#
#    DESCRIPTION: The subroutine will perform an IR scan operation.  The 
#      number of bits shifted in/out of the instruction register is
#      specified by "len", and the input data (TDI) must be an array of
#      32-bit values.  The TDI data can be left or right-justified, as
#      defined using "jtag_set_just()".  After the shift is complete, the
#      TAP controller is moved thru UPDATE-IR and then to RUN-TEST-IDLE.
#      The "end_state" parameter can be given to leave the TAP controller
#      in PAUSE-IR (no update), or to move the controller to TEST-LOGIC-RESET
#      after reaching RUN-TEST-IDLE.
#      
#      Valid values for "end_state" are:
#      $JTAG_STATE_TLR, $JTAG_STATE_RTI or $JTAG_STATE_PIR
#
#    RETURN VALUE: An array containing the TDO data (same format as TDI).
#
#-------------------------------------------------------------------------------

sub jtag_scan_ir
{
   my $num = $#_ + 1;
   print "\njtag_scan_ir(@_) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   my $state = $JTAG_STATE_RTI;
   if ( $num == 3 )
   {  $state = $_[2];
      if ( ($state != $JTAG_STATE_TLR) && ($state != $JTAG_STATE_RTI) &&
           ($state != $JTAG_STATE_PIR) )
      {  $emess = "Invalid end-state passed to subroutine jtag_scan_ir().";
         &$comm_err_fun($emess);
         return;
      }
   }
   elsif ( $num != 2 )
   {  $emess = "Invalid parameters passed to subroutine jtag_scan_ir().";
      &$comm_err_fun($emess);
      return;
   }
   my $blen     = $_[0];
   my $data_ptr = $_[1];
   if ( ref($data_ptr) ne 'ARRAY' )
   {  $emess = "Invalid parameter passed to subroutine jtag_scan_ir().";
      $emess .= "\nSecond parameter must be an ARRAY reference.";
      &$comm_err_fun($emess);
      return;
   }
   if ( $debug_flag & $DEBUG_DETAILS )
   {  printf "#bits     = %lu\n", $blen;
      printf "end-state = $state = $jtag_state_names[$state]\n";
      printf "data      = @$data_ptr\n";
   }

   my $rv = pcf_jtag_scan_ir($blen, $data_ptr, $state, $ecode, $emess);
   &$comm_err_fun($emess)   if ( $ecode < 0 );
   return(@$rv);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: jtag_scan_dr()
#
#    USAGE: jtag_scan_dr($len, \@tdi_data)
#           jtag_scan_dr($len, \@tdi_data, $end_state)
#
#    DESCRIPTION: The subroutine will perform a DR scan operation.  The 
#      number of bits shifted in/out of the test data register is
#      specified by "len", and the input data (TDI) must be an array of
#      32-bit values.  The TDI data can be left or right-justified, as
#      defined using "jtag_set_just()".  After the shift is complete, the
#      TAP controller is moved thru UPDATE-DR and then to RUN-TEST-IDLE.
#      The "end_state" parameter can be given to leave the TAP controller
#      in PAUSE-DR (no update), or to move the controller to TEST-LOGIC-RESET
#      after reaching RUN-TEST-IDLE.
#      
#      Valid values for "end_state" are:
#      $JTAG_STATE_TLR, $JTAG_STATE_RTI or $JTAG_STATE_PIR
#
#    RETURN VALUE: An array containing the TDO data (same format as TDI).
#
#-------------------------------------------------------------------------------

sub jtag_scan_dr
{
   my $num = $#_ + 1;
   print "\njtag_scan_dr(@_) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   my $state = $JTAG_STATE_RTI;
   if ( $num == 3 )
   {  $state = $_[2];
      if ( ($state != $JTAG_STATE_TLR) && ($state != $JTAG_STATE_RTI) &&
           ($state != $JTAG_STATE_PDR) )
      {  $emess = "Invalid end-state passed to subroutine jtag_scan_dr().";
         &$comm_err_fun($emess);
         return;
      }
   }
   elsif ( $num != 2 )
   {  $emess = "Invalid parameters passed to subroutine jtag_scan_dr().";
      &$comm_err_fun($emess);
      return;
   }

   my $blen     = $_[0];
   my $data_ptr = $_[1];
   if ( ref($data_ptr) ne 'ARRAY' )
   {  $emess = "Invalid parameter passed to subroutine jtag_scan_dr().";
      $emess .= "\nSecond parameter must be an ARRAY reference.";
      &$comm_err_fun($emess);
      return;
   }
   if ( $debug_flag & $DEBUG_DETAILS )
   {  printf "#bits     = %lu\n", $blen;
      printf "end-state = $state = $jtag_state_names[$state]\n";
      printf "data      = @$data_ptr\n";
   }

   my $rv = pcf_jtag_scan_dr($blen, $data_ptr, $state, $ecode, $emess);
   &$comm_err_fun($emess)   if ( $ecode < 0 );
   return(@$rv);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: jtag_data_display()
#
#    USAGE: jtag_data_display(\@data)
#           jtag_data_display(\@data, $desc)
#
#    DESCRIPTION: The subroutine will display an array of JTAG data in
#      hexidecimal notation.
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub jtag_data_display
{
   my $num = $#_ + 1;
   print "\njtag_data_display(@_) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   my $desc = "DATA";
   if ( $num == 2 )
   {  $desc = $_[1];
   }
   elsif ( $num != 1 )
   {  $emess = "Invalid parameters passed to subroutine jtag_data_display().";
      &$comm_err_fun($emess);
      return;
   }
   my $data_ptr = $_[0];
   if ( ref($data_ptr) ne 'ARRAY' )
   {  $emess = "Invalid parameter passed to subroutine jtag_data_display().";
      $emess .= "\nFirst parameter must be an ARRAY reference.";
      &$comm_err_fun($emess);
      return;
   }
   my @data = @$data_ptr;

   $num = $#data + 1;
   my $q   = $num >> 3;
   my $r   = $num & 0x7;
   my ($i, $b);

   printf "\nJTAG BUFFER (HEX) = $desc\n";
   for ($i = 0, $b = 0; $i < $q; $i++, $b += 8)
   {  printf "%08lX %08lX %08lX %08lX %08lX %08lX %08lX %08lX\n",
           $data[$b+0], $data[$b+1], $data[$b+2], $data[$b+3],
           $data[$b+4], $data[$b+5], $data[$b+6], $data[$b+7];
   }
   if ( $r > 0 )
   {  $b = $q * 8;
      for ($i = 0; $i < $r; $i++)
      {  printf "%08lX ", $data[$b+$i];
      }
      printf "\n";
   }
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: jtag_simacc_set()
#
#    USAGE: jtag_simacc_set($code, $entity_name, $data)
#
#    NOTE: This subroutine is valid when driving a VERILOG simulator (the
#      access of $JTAG_ACCESS_VERSIM).
#
#    DESCRIPTION: The subroutine will issue a command to VERILOG simulator
#      to set the value of a specified entity.  The "code" parameter defines
#      the format of the data: 
#      $JTAG_SIMACC_VEC  = set the value of a register or port
#      $JTAG_SIMACC_INT  = set the value of an integer entity
#      $JTAG_SIMACC_SCAL = set the value of a scalar entity
#      $JTAG_SIMACC_STR  = set the value of a string entity
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub jtag_simacc_set
{
   my $num = $#_ + 1;
   print "\njtag_simacc_set(@_) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 3 )
   {  $emess = "Invalid parameters passed to subroutine jtag_simacc_set().";
      &$comm_err_fun($emess);
      return;
   }
   my $code = $_[0];
   if ( ($code < 0) || ($code >= $JTAG_MAX_SIMACC) )
   {  $emess = "Invalid code parameter passed to subroutine jtag_simacc_set().";
      &$comm_err_fun($emess);
      return;
   }

   pcf_jtag_simacc_set($code, $_[1], $_[2], $ecode, $emess);
   &$comm_err_fun($emess)   if ( $ecode < 0 );
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: jtag_simacc_get()
#
#    USAGE: $data = jtag_simacc_get($entity_name)
#
#    NOTE: This subroutine is valid when driving a VERILOG simulator (the
#      access of $JTAG_ACCESS_VERSIM).
#
#    DESCRIPTION: The subroutine will issue a command to VERILOG simulator
#      to get the value of a specified entity.
#
#    RETURN VALUE: The value of the entity from the VERILOG simulator.
#
#-------------------------------------------------------------------------------

sub jtag_simacc_get
{
   my $num = $#_ + 1;
   print "\njtag_simacc_get(@_) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 1 )
   {  $emess = "Invalid parameters passed to subroutine jtag_simacc_get().";
      &$comm_err_fun($emess);
      return;
   }

   my $rv = pcf_jtag_simacc_get($_[0], $ecode, $emess);
   &$comm_err_fun($emess)   if ( $ecode < 0 );
   return($rv);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: jtag_simacc_time()
#
#    USAGE: $time_stamp = jtag_simacc_time()
#
#    NOTE: This subroutine is valid when driving a VERILOG simulator (the
#      access of $JTAG_ACCESS_VERSIM).
#
#    DESCRIPTION: The subroutine will issue a command to VERILOG simulator
#      to get the current simulation time.
#
#    RETURN VALUE: The simulation time value from the VERILOG simulator.
#
#-------------------------------------------------------------------------------

sub jtag_simacc_time
{
   my $num = $#_ + 1;
   print "\njtag_simacc_time(@_) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 0 )
   {  $emess = "Invalid parameters passed to subroutine jtag_simacc_time().";
      &$comm_err_fun($emess);
      return;
   }

   my $rv = pcf_jtag_simacc_time($ecode, $emess);
   &$comm_err_fun($emess)   if ( $ecode < 0 );
   return($rv);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: jtag_simacc_dump()
#
#    USAGE: jtag_simacc_dump($mode)
#
#           $mode can be one of the following: $JTAG_SIMACC_VCD_DUMP_ON,
#              $JTAG_SIMACC_VCD_DUMP_OFF, $JTAG_SIMACC_VPD_DUMP_ON or
#              $JTAG_SIMACC_VPD_DUMP_OFF
#
#    NOTE: This subroutine is valid when driving a VERILOG simulator (the
#      access of $JTAG_ACCESS_VERSIM).
#
#    DESCRIPTION: The subroutine will issue a command to VERILOG simulator
#      to turn on or off dumping.
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub jtag_simacc_dump
{
   my $num = $#_ + 1;
   print "\njtag_simacc_dump(@_) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 1 )
   {  $emess = "Invalid parameters passed to subroutine jtag_simacc_dump().";
      &$comm_err_fun($emess);
      return;
   }
   if ( ($_[0] != $JTAG_SIMACC_VCD_DUMP_ON) &&
        ($_[0] != $JTAG_SIMACC_VCD_DUMP_OFF) &&
        ($_[0] != $JTAG_SIMACC_VPD_DUMP_ON) &&
        ($_[0] != $JTAG_SIMACC_VPD_DUMP_OFF) )
   {  $emess = "Invalid mode parameter passed to subroutine jtag_simacc_dump().";
      &$comm_err_fun($emess);
      return;
   }

   my $rv = pcf_jtag_simacc_dump($_[0], $ecode, $emess);
   &$comm_err_fun($emess)   if ( $ecode < 0 );
   return($rv);
}


#-------------------------------------------------------------------------------

sub jtag_icd_get_descs
{
   printf "\njtag_icd_get_descs( @_ )\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   #---------------------------------------
   #  Read all supported chip descriptions/types from library
   #---------------------------------------
   my @all_descs = ();
   my $ndescs    = pcf_icd_get_descs(\@all_descs, $ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess);
      return;
   }
   return( @all_descs );
}


#-------------------------------------------------------------------------------

sub jtag_dump_file_open
{
   my $num = $#_ + 1;
   print "\njtag_dump_file_open(@_) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 1 )
   {  $emess = "Invalid parameters passed to subroutine jtag_dump_file_open().";
      &$comm_err_fun($emess);
      return;
   }

   my $time_str = pcf_jtag_dump_file_open($_[0], $ecode, $emess);
   &$comm_err_fun($emess)   if ( $ecode < 0 );
   return($time_str);
}

#-------------------------------------------------------------------------------

sub jtag_dump_file_add
{
   my $num = $#_ + 1;
   print "\njtag_dump_file_add(@_) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 1 )
   {  $emess = "Invalid parameters passed to subroutine jtag_dump_file_add().";
      &$comm_err_fun($emess);
      return;
   }

   pcf_jtag_dump_file_add($_[0]);
}


#-------------------------------------------------------------------------------

sub jtag_dump_file_close
{
   my $num = $#_ + 1;
   print "\njtag_dump_file_close(@_) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 0 )
   {  $emess = "Invalid parameters passed to subroutine jtag_dump_file_close().";
      &$comm_err_fun($emess);
      return;
   }

   my $time_str = pcf_jtag_dump_file_close($ecode, $emess);
   &$comm_err_fun($emess)   if ( $ecode < 0 );
   return($time_str);
}


#-------------------------------------------------------------------------------

sub jtag_stop_clocks
{
   my $num = $#_ + 1;
   print "\njtag_stop_clocks(@_) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 1 )
   {  $emess = "Invalid parameters passed to subroutine jtag_stop_clocks().";
      &$comm_err_fun($emess);
      return;
   }

   pcf_jtag_stop_clocks($_[0], $ecode, $emess);
   &$comm_err_fun($emess)   if ( $ecode < 0 );
}


#-------------------------------------------------------------------------------

sub jtag_dump_chip
{
   my $num = $#_ + 1;
   print "\njtag_dump_chip(@_) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 5 )
   {  $emess = "Invalid parameters passed to subroutine jtag_dump_chip().";
      &$comm_err_fun($emess);
      return;
   }

   pcf_jtag_dump_chip($_[0], $_[1], $_[2], $_[3], $_[4], $ecode, $emess);
   &$comm_err_fun($emess)   if ( $ecode < 0 );
}

#-------------------------------------------------------------------------------


1;
