#
#/**************************************************************************
# *                                                                        *
# *               Copyright (C) 1989-1999 Silicon Graphics, Inc.           *
# *                                                                        *
# *  These coded instructions, statements, and computer programs  contain  *
# *  unpublished  proprietary  information of Silicon Graphics, Inc., and  *
# *  are protected by Federal copyright law.  They  may  not be disclosed  *
# *  to  third  parties  or copied or duplicated in any form, in whole or  *
# *  in part, without the prior written consent of Silicon Graphics, Inc.  *
# *                                                                        *
# **************************************************************************/
#
# @(#)$Id: jtdr_subs.pm,v 1.11 2005/01/10 21:16:07 beighley Exp $

#-------------------------------------------------------------------------------
#
#  FILE: jtdr_subs.pm
#
#  DESCRIPTION:  This file contains PERL subroutines to access and manipulate
#    a JTAG test data register (TDR) representation.  A JTAG TDR instance is
#    an ARRAY with two elements: a reference to the JTAG TDR definition 
#    (common for all instances of this register) and reference to a value 
#    (unique per instance).  The TDR definitions must be included for each 
#    ASIC type, and a lookup is done using the ASIC type name and JTAG TDR 
#    name (i.e. "shub_01" "IDCODE").
#
#    A JTAG TDR definition structure contains the register name, register
#    note/description, instruction code value, instruction code name,
#    register length (bits), reset value, and field defintions.
#
#  SUBROUTINES:  The following subroutines are contained in this file:
#    jtag_tdr_create_mmr_inst($asic_type, $tdr_name) -- create JTDR structure
#    jtag_tdr_is_valid($jtag_tdr_ref)  -- confirm created JTDR is valid
#    jtag_tdr_name($jtag_tdr_ref)      -- get the JTDR name
#    jtag_tdr_note($jtag_tdr_ref)      -- get the JTDR note or description
#    jtag_tdr_opname($jtag_tdr_ref)    -- get the JTDR opcode name
#    jtag_tdr_opcode($jtag_tdr_ref)    -- get the JTDR opcode value
#    jtag_tdr_length($jtag_tdr_ref)    -- get the JTDR length (bits)
#    jtag_tdr_reset($jtag_tdr_ref)     -- set the JTDR to the "reset" value
#    jtag_tdr_get_value($jtag_tdr_ref)         -- get the current JTDR value
#    jtag_tdr_set_value($jtag_tdr_ref, $value) -- set the JTDR value (entire)
#    jtag_tdr_get_fields($jtag_tdr_ref)        -- get the field labels
#    jtag_tdr_get_fields($jtag_tdr_ref, ORDER) -- get the field labels in order
#    jtag_tdr_is_valid_field($jtag_tdr_ref, $field_label) -- check field label
#    jtag_tdr_get_num_fields($jtag_tdr_ref)    -- get the number of fields
#    jtag_tdr_get_field_value($jtag_tdr_ref, $field) -- get field value
#    jtag_tdr_set_field_value($jtag_tdr_ref, $field, $value) -- set field value
#    jtag_tdr_field_parm($jtag_tdr_ref, $field, $entry_index)
#    jtag_tdr_dump_contents($jtag_tdr_ref) -- display register and field values
#
#  MODIFICATION HISTORY:
#
#     WHEN      WHO             WHAT
#     --------  --------------  -----------------------------------------
#     xx/xx/xx  Karen Beighley  Original source code.
#     06/06/03  Karen Beighley  Added subroutine: jtag_tdr_dump_contents().
#     07/17/03  Karen Beighley  Added change to description for TDR definition.
#     12/12/03  Karen Beighley  Increased label string length in display.
#     12/14/03  Karen Beighley  Added subroutine: jtag_tdr_is_valid_field().
#     02/18/04  Karen Beighley  Added support for new chip type: NL4RTR
#     04/28/04  Karen Beighley  Initial support for SHUB2 (not complete!)
#     01/04/05  Karen Beighley  Initial support for LBI (not complete!)
#
#-------------------------------------------------------------------------------

$USE_JTDR_SUBS_PM = 1;    # Define to allow test for module inclusion

use pic_01_jtdrs;
use shub_01_jtdrs;
use tio_01_jtdrs;
use nl4rtr_01_jtdrs;
use shub2_01_jtdrs;
use lbi_01_jtdrs;
use misc_jtdrs;
use vec_subs;


#-------------------------------------------------------------------------------
#  CONSTANTS --> these values must not change!
#-------------------------------------------------------------------------------

$TDR_FLD_INDEX  = 0;
$TDR_FLD_START  = 1;
$TDR_FLD_LENGTH = 2;
$TDR_FLD_MODE   = 3;
$TDR_FLD_DESC   = 4;
$TDR_MAX_FLD    = 5;

$TDR_REF_DEF    = 0;
$TDR_REF_VAL    = 1;


################################################################################
#   COMMON BYPASS REGISTER FIELD DEFINITION
################################################################################

%common_bypass_flds =
#  NAME                     IND  POS  LEN  MODE    DESCRIPTION
#  -------------------      ---  ---  ---  ------  -----------------------
(  "RSVD"              => [   0,   0,   1, 0x0000, ""                       ],
);


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: jtag_tdr_create_inst()
#
#    USAGE: jtag_tdr_create_inst($asic_type, $tdr_name)
#
#    DESCRIPTION:  This subroutine will create an instance of a JTAG TDR
#       for a specific for a specific ASIC type.  The TDR definitions must 
#       be included for the ASIC type, and a lookup is done using the
#       ASIC type name and JTAG TDR name (i.e. "shub_01" "IDCODE").
#       The structure is an array that contains a reference to the JTAG TDR
#       definition (common for all instances of this register) and 
#       reference to a value (unique per instance).
#
#    RETURN VALUE: A reference to a JTAG TDR instance structure.
#
#-------------------------------------------------------------------------------

sub jtag_tdr_create_inst
{
#   print "\njtag_create_tdr_inst(@_)\n";

   my @tdr_inst = (0, 0);
   return(\@tdr_inst) if ( $#_ != 1 );

   my ($def_name, $tdr_name, $ptr, $i, @data);
   $def_name = $_[0];
   $tdr_name = $_[1];

   my $base_name = $def_name;
   if ( $def_name =~ /^shub2/i )
   {  $base_name = "shub2_01";
   }
   elsif ( $def_name =~ /^shub/i )
   {  $base_name = "shub_01";
   }
   elsif ( $def_name =~ /^pic/i )
   {  $base_name = "pic_01";
   }
   elsif ( $def_name =~ /^tio/i )
   {  $base_name = "tio_01";
   }
   elsif ( $def_name =~ /^nl4r/i )
   {  $base_name = "nl4rtr_01";
   }
   else
   {  $base_name = $def_name;
   }
   $def_name = $base_name . "_tdr_defined";
   return(\@tdr_inst) if ( !defined($$def_name) );

   #  Search for the specified register
   $def_name = $_[0] . "_tdrs";
   if ( exists $$def_name{$_[1]} )
   {  $ptr = $$def_name{$_[1]};           # pointer to TDR definition structure (HASH)
      $tdr_inst[$TDR_REF_DEF] = $ptr;     # save definition reference in the TDR instance
      $ptr = $$ptr{RESET};                # pointer to the reset value structure (ARRAY)
      for ($i = 0; $i <= $#$ptr; $i++)    # copy the reset value
      {  $data[$i] = $$ptr[$i];
      }
      $tdr_inst[$TDR_REF_VAL] = \@data;   # save the data reference in the TDR instance
   }
   return(\@tdr_inst);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: jtag_tdr_create_bypass_inst()
#
#    USAGE: jtag_tdr_create_bypass_inst($ilen)
#
#    DESCRIPTION:  This subroutine will create an instance of a BYPASS TDR
#       for any ASIC type.  The instruction register length is required, so
#       the appropriate OPCODE can be created (all 1's for BYPASS).
#       The structure is an array that contains a reference to the JTAG TDR
#       definition and reference to a value.
#
#    RETURN VALUE: A reference to a JTAG BYPASS TDR instance structure.
#
#-------------------------------------------------------------------------------

sub jtag_tdr_create_bypass_inst
{
   my $ilen = 32;
   $ilen = $_[0] if ( $#_ >= 0 );  # Actual IR length for chip

   #  Build an instruction OPCODE of all 1's (BYPASS)
   my $code  = 0;
   my $i     = 0;
   for ($i = 0; $i < $ilen; $i++)
   {  $code = ($code << 1) | 1;
   }

   #  Create the JTAG TDR definition (HASH)
   my %bypass_tdr =
   (  NAME   => "BYPASS",
      NOTE   => "BYPASS JTAG Register",
      OPNAME => "BYPASS",
      OPCODE => $code,
      LENGTH => 1,
      RESET  => [ 0x00000000 ],
      FIELDS => \%common_bypass_flds
   );
   my @data = ( 0 );

   #  Create the JTAG TDR instance
   my @tdr_inst = (\%bypass_tdr, \@data);
   return(\@tdr_inst);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: jtag_tdr_is_valid()
#
#    USAGE: jtag_tdr_is_valid($jtag_tdr_ref)
#
#    DESCRIPTION: The subroutine will check the contents of a JTAG TDR
#      structure to confirm it contains a valid TDR definition pointer.
#
#    RETURN VALUE: 0 if the register is not valid; 1 if the register is valid.
#
#-------------------------------------------------------------------------------

sub jtag_tdr_is_valid
{  return(0) if ( $#_ != 0 );
   my $jtag_tdr = $_[0];
   return(0) if ( ref($jtag_tdr) ne 'ARRAY' );
   my $def_ptr = $$jtag_tdr[$TDR_REF_DEF];
   return(0) if ( ref($def_ptr) ne 'HASH' );
   my $val_ptr = $$jtag_tdr[$TDR_REF_VAL];
   return(0) if ( ref($val_ptr) ne 'ARRAY' );
   return(1);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: jtag_tdr_name()
#
#    USAGE: jtag_tdr_name($jtag_tdr_ref)
#
#    DESCRIPTION: The subroutine will extract the JTAG TDR name from the
#      specified TDR structure.
#
#    RETURN VALUE: the name of the JTAG TDR (text string: "CLKCNTL")
#
#-------------------------------------------------------------------------------

sub jtag_tdr_name
{  return("") if ( $#_ != 0 );
   my $jtag_tdr = $_[0];
   return("") if ( ref($jtag_tdr) ne 'ARRAY' );
   my $def_ptr = $$jtag_tdr[$TDR_REF_DEF];
   return("") if ( ref($def_ptr) ne 'HASH' );
   return( $$def_ptr{NAME} );
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: jtag_tdr_note()
#
#    USAGE: jtag_tdr_note($jtag_tdr_ref)
#
#    DESCRIPTION: The subroutine will extract the any note (description) from
#      the specified TDR structure.
#
#    RETURN VALUE: the note for the JTAG TDR (text string)
#
#-------------------------------------------------------------------------------

sub jtag_tdr_note
{  return("") if ( $#_ != 0 );
   my $jtag_tdr = $_[0];
   return("") if ( ref($jtag_tdr) ne 'ARRAY' );
   my $def_ptr = $$jtag_tdr[$TDR_REF_DEF];
   return("") if ( ref($def_ptr) ne 'HASH' );
   return( $$def_ptr{NOTE} );
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: jtag_tdr_opname()
#
#    USAGE: jtag_tdr_opname($jtag_tdr_ref)
#
#    DESCRIPTION: The subroutine will extract the instruction opcode name
#      from the specified TDR structure.
#
#    RETURN VALUE: the opcode name for the JTAG TDR (text string)
#
#-------------------------------------------------------------------------------

sub jtag_tdr_opname
{  return("") if ( $#_ != 0 );
   my $jtag_tdr = $_[0];
   return("") if ( ref($jtag_tdr) ne 'ARRAY' );
   my $def_ptr = $$jtag_tdr[$TDR_REF_DEF];
   return("") if ( ref($def_ptr) ne 'HASH' );
   return( $$def_ptr{OPNAME} );
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: jtag_tdr_opcode()
#
#    USAGE: jtag_tdr_opcode($jtag_tdr_ref)
#
#    DESCRIPTION: The subroutine will extract the instruction opcode value
#      from the specified TDR structure.
#
#    RETURN VALUE: the opcode value for the JTAG TDR
#
#-------------------------------------------------------------------------------

sub jtag_tdr_opcode
{  return("") if ( $#_ != 0 );
   my $jtag_tdr = $_[0];
   return("") if ( ref($jtag_tdr) ne 'ARRAY' );
   my $def_ptr = $$jtag_tdr[$TDR_REF_DEF];
   return("") if ( ref($def_ptr) ne 'HASH' );
   return( $$def_ptr{OPCODE} );
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: jtag_tdr_length()
#
#    USAGE: jtag_tdr_length($jtag_tdr_ref)
#
#    DESCRIPTION: The subroutine will extract the instruction register length
#      from the specified TDR structure.
#
#    RETURN VALUE: the instruction register length for the JTAG TDR
#
#-------------------------------------------------------------------------------

sub jtag_tdr_length
{  return("") if ( $#_ != 0 );
   my $jtag_tdr = $_[0];
   return("") if ( ref($jtag_tdr) ne 'ARRAY' );
   my $def_ptr = $$jtag_tdr[$TDR_REF_DEF];
   return("") if ( ref($def_ptr) ne 'HASH' );
   return( $$def_ptr{LENGTH} );
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: jtag_tdr_reset()
#
#    USAGE: jtag_tdr_reset($jtag_tdr_ref)
#
#    DESCRIPTION: The subroutine will return the reset value for the
#      specified TDR structure.
#
#    RETURN VALUE: an reference to an array containing the JTAG TDR reset value
#
#-------------------------------------------------------------------------------

sub jtag_tdr_reset
{  return("") if ( $#_ != 0 );
   my $jtag_tdr = $_[0];
   return("") if ( ref($jtag_tdr) ne 'ARRAY' );
   my $def_ptr = $$jtag_tdr[$TDR_REF_DEF]; # Reference to definition struct.
   my $val_ptr = $$jtag_tdr[$TDR_REF_VAL]; # Pointer to actual data
   return("") if ( ref($def_ptr) ne 'HASH' );
   my $aref = $$def_ptr{RESET};            # Pointer to reset data
   my $i    = 0;
   for ($i = 0; $i <= $#$aref; $i++)       # Copy reset value into data value
   {  $$val_ptr[$i] = $$aref[$i];
   }
   my @data = @$aref;
   return( \@data );                       # Return reset value (copy of array)
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: jtag_tdr_get_value()
#
#    USAGE: jtag_tdr_get_value($jtag_tdr_ref)
#
#    DESCRIPTION: The subroutine will extract the current JTAG TDR value to
#      from the specified structure.
#
#    RETURN VALUE: A reference to an array containing the JTAG TDR value.
#
#-------------------------------------------------------------------------------

sub jtag_tdr_get_value
{  return("") if ( $#_ != 0 );
   my $jtag_tdr = $_[0];
   return("") if ( ref($jtag_tdr) ne 'ARRAY' );
   my $val_ptr = $$jtag_tdr[$TDR_REF_VAL];
   my @data = (@$val_ptr);
   return(\@data);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: jtag_tdr_set_value()
#
#    USAGE: jtag_tdr_set_value($jtag_tdr_ref, $value_ref)
#
#    DESCRIPTION: The subroutine will change the current JTAG TDR value to
#      that specified by the "value_ref" parameter.  The "value_ref" pointer
#      should be a reference to an array with the actual values.
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub jtag_tdr_set_value
{  return if ( $#_ != 1 );
   my $jtag_tdr = $_[0];  # Reference to JTAG TDR structure
   my $data_ref = $_[1];  # Reference to array containing data
   return if ( ref($jtag_tdr) ne 'ARRAY' );
   return if ( ref($data_ref) ne 'ARRAY' );
   my $val_ptr = $$jtag_tdr[$TDR_REF_VAL]; # Pointer to actual data
# confirm length of input data array (same size?)
   vec_cpy_bits(0, $data_ref, 0, $val_ptr, jtag_tdr_length($jtag_tdr), 'right');
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: jtag_tdr_get_fields()
#
#    USAGE: jtag_tdr_get_fields($jtag_tdr_ref)
#           jtag_tdr_get_fields($jtag_tdr_ref, ORDER)
#
#    DESCRIPTION: The subroutine will extract the label for each field
#      defined within the JTAG TDR structure.  If the "ORDER" parameter is
#      specified, then the labels are ordered by starting bit position
#      (biggest to smallest starting bit position).
#
#    RETURN VALUE: an array of labels (each label is a text string)
#
#-------------------------------------------------------------------------------

sub jtag_tdr_get_fields
{  my $flag = 0;
   if ( $#_ == 1 )
   {  $flag = 1 if ( $_[1] eq ORDER );
   }
   elsif ( $#_ != 0 )
   {  return();
   }
   my $jtag_tdr_ptr = $_[0];                      # pointer to register instance
   return () if ( ref($jtag_tdr_ptr) ne 'ARRAY' );
   my $def_ptr = $$jtag_tdr_ptr[$TDR_REF_DEF];    # pointer to reg definition
   return () if ( ref($def_ptr) ne 'HASH' );
   my $fld_ptr = $$def_ptr{FIELDS};               # pointer to field hash table
   my @fields  = (keys %$fld_ptr);                # field labels/names (random)
   return( @fields ) if ( $flag == 0 );           # return labels in any order
   my ($fld, $parm_ptr, $i, @name);
   @name = ();                                    # initialize name array
   foreach $fld ( @fields )                       # process each field in hash
   {  $parm_ptr = $$fld_ptr{$fld};                # pointer to field parameters
      $i        = $$parm_ptr[$TDR_FLD_INDEX];     # index of field (bit ordered)
      $name[$i] = $fld;                           # save the label by index
   }
   return(@name);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: jtag_tdr_is_valid_field()
#
#    USAGE: jtag_tdr_is_valid_field($jtag_tdr_ref, $field_label)
#
#    DESCRIPTION: The subroutine will check to see if a field label is valid
#      for the given JTAG TDR structure.
#
#    RETURN VALUE: $TRUE if the field label is valid; $FALSE (0) if not valid
#
#-------------------------------------------------------------------------------

sub jtag_tdr_is_valid_field
{  return($FALSE) if ( $#_ != 1 );
#print "\njtag_tdr_is_valid_field( @_ )\n";
   my $jtag_tdr_ptr = $_[0];                      # pointer to register instance
   return if ( ref($jtag_tdr_ptr) ne 'ARRAY' );
   my $def_ptr = $$jtag_tdr_ptr[$TDR_REF_DEF];    # pointer to reg definition
   return if ( ref($def_ptr) ne 'HASH' );
   my $fld_ptr = $$def_ptr{FIELDS};               # pointer to field hash table
   return ( (exists $$fld_ptr{$_[1]}) ? $TRUE : $FALSE );
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: jtag_tdr_get_num_fields()
#
#    USAGE: jtag_tdr_get_num_fields($reg_ref)
#
#    DESCRIPTION: The subroutine will determine number of fields
#      defined within the JTAG TDR structure.
#
#    RETURN VALUE: number of fields in the JTAG TDR
#
#-------------------------------------------------------------------------------

sub jtag_tdr_get_num_fields
{  return if ( $#_ != 0 );
   my $jtag_tdr_ptr = $_[0];                      # pointer to register instance
   return () if ( ref($jtag_tdr_ptr) ne 'ARRAY' );
   my $def_ptr = $$jtag_tdr_ptr[$TDR_REF_DEF];    # pointer to reg definition
   return () if ( ref($def_ptr) ne 'HASH' );
   my $fld_ptr = $$def_ptr{FIELDS};               # pointer to field hash table
   my @flds  = (keys %$fld_ptr);                  # field labels/names (random)
   return( ($#flds + 1) );
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: jtag_tdr_set_field_value()
#
#    USAGE: jtag_tdr_set_field_value($reg_ref, $field_label, $value_ref)
#
#    DESCRIPTION: The subroutine will set the value of the specified
#      field in the given JTAG TDR structure.  The "value_ref" pointer
#      should be a reference to an array with the actual values.
#
#    RETURN VALUE: A reference to an array containing the JTAG TDR value.
#
#-------------------------------------------------------------------------------

sub jtag_tdr_set_field_value
{  return if ( $#_ != 2 );
print "\njtag_tdr_set_field_value( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);
   my $jtag_tdr_ptr = $_[0];                      # pointer to register instance
   return if ( ref($jtag_tdr_ptr) ne 'ARRAY' );
   return if ( ref($_[2]) ne 'ARRAY' );
   my $def_ptr = $$jtag_tdr_ptr[$TDR_REF_DEF];    # pointer to reg definition
   return if ( ref($def_ptr) ne 'HASH' );
   my $fld_ptr = $$def_ptr{FIELDS};               # pointer to field hash table
   return if ( !(exists $$fld_ptr{$_[1]}) );
   $def_ptr = $$fld_ptr{$_[1]};                   # pointer to field details
   vec_cpy_bits( 0, $_[2], $$def_ptr[$TDR_FLD_START], 
      $$jtag_tdr_ptr[$TDR_REF_VAL], $$def_ptr[$TDR_FLD_LENGTH], 'right');
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: jtag_tdr_get_field_value()
#
#    USAGE: jtag_tdr_get_field_value($jtag_tdr_ref, $field_label)
#
#    DESCRIPTION: The subroutine will extract the value of the specified
#      field from the given JTAG TDR structure.
#
#    RETURN VALUE: A reference to an array containing the field value.
#
#-------------------------------------------------------------------------------

sub jtag_tdr_get_field_value
{  return if ( $#_ != 1 );
#print "\njtag_tdr_get_field_value( @_ )\n";
   my $jtag_tdr_ptr = $_[0];                      # pointer to register instance
   return if ( ref($jtag_tdr_ptr) ne 'ARRAY' );
   my $def_ptr = $$jtag_tdr_ptr[$TDR_REF_DEF];    # pointer to reg definition
   return if ( ref($def_ptr) ne 'HASH' );
   my $fld_ptr = $$def_ptr{FIELDS};               # pointer to field hash table
   return if ( !(exists $$fld_ptr{$_[1]}) );
   $def_ptr = $$fld_ptr{$_[1]};                   # pointer to field details
   my $new_data = vec_create($$def_ptr[$TDR_FLD_LENGTH]);
   vec_cpy_bits($$def_ptr[$TDR_FLD_START], $$jtag_tdr_ptr[$TDR_REF_VAL], 
      0, $new_data, $$def_ptr[$TDR_FLD_LENGTH], 'right');
   return($new_data);
}


#-------------------------------------------------------------------------------

sub jtag_tdr_field_parm
{  return("") if ( $#_ != 2 );
   my $jtag_tdr_ptr = $_[0];
   return("") if ( ref($jtag_tdr_ptr) ne 'ARRAY' );
   my $def_ptr = $$jtag_tdr_ptr[$TDR_REF_DEF];
   return("") if ( ref($def_ptr) ne 'HASH' );
   my $fld_ptr = $$def_ptr{FIELDS};
   return("") if ( !(exists $$fld_ptr{$_[1]}) );
   return("") if ( $_[2] > $TDR_MAX_FLD );
   $def_ptr = $$fld_ptr{$_[1]};
   return( $$def_ptr[ $_[2] ] );
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: jtag_tdr_build_vec()
#
#    USAGE: jtag_tdr_build_vec($num, $lref, $iref, $oref, $just)
#
#    DESCRIPTION: The subroutine will build a JTAG vector from a number of
#       data entries.
#
#    RETURN VALUE: The total length of the vector.
#
#-------------------------------------------------------------------------------

sub jtag_tdr_build_vec
{  return(0) if ( $#_ != 4 );
#   print "\njtag_tdr_build_vec( @_ )\n";
   my ($num, $lref, $iref, $oref, $just);
   $num  = $_[0]; # number of entries in the input arrays
   $lref = $_[1]; # reference to array: each entry is the length
   $iref = $_[2]; # reference to array: each entry is pointer to the data
   $oref = $_[3]; # reference to array for vector data
   $just = $_[4]; # data justification: 'left' or 'right'
   return(0) if ( ref($lref) ne 'ARRAY' );
   return(0) if ( ref($iref) ne 'ARRAY' );
   return(0) if ( ref($oref) ne 'ARRAY' );

   my ($i, $len);
   for ($i = 0, $len = 0; $i < $num; $i++)
   {  vec_cpy_bits(0, $iref[$i], $len, $oref, $lref[$i], $just);
      $len += $lref[$i];
   }
   return($len);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: jtag_tdr_dump()
#
#    USAGE: jtag_tdr_dump($jtag_tdr_ref)
#
#    DESCRIPTION: The subroutine will display the contents of the JTAG TDR
#       structure (for debugging purposes).
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub jtag_tdr_dump
{  return if ( $#_ != 0 );
#printf "\njtag_tdr_dump()\n";

   my $jtag_tdr = $_[0];  # Reference to the JTAG TDR structure
   my $def_ptr  = $$jtag_tdr[$TDR_REF_DEF]; # Pointer to the definition struct.
   my @labels   = jtag_tdr_get_fields($jtag_tdr, ORDER);

   printf "\nJTAG TDR CONTENTS:\n";
   printf "name   = \"%s\"\n",  jtag_tdr_name($jtag_tdr); 
   printf "note   = \"%s\"\n",  jtag_tdr_note($jtag_tdr);
   printf "opname = \"%s\"\n",  jtag_tdr_opname($jtag_tdr);
   printf "opcode = 0x%02lX\n", jtag_tdr_opcode($jtag_tdr);
   printf "length = %lu\n",     jtag_tdr_length($jtag_tdr);
   printf "reset  = $$def_ptr{RESET}\n";
   printf "fields = $$def_ptr{FIELDS}\n";
   printf "labels = ( @labels )\n";
   printf "value  = $$jtag_tdr[$TDR_REF_VAL]\n";

   vec_dump_hex("JTAG_TDR RESET  DATA", $$def_ptr{RESET});
   vec_dump_hex("JTAG_TDR ACTUAL DATA", $$jtag_tdr[$TDR_REF_VAL]);
}

#-------------------------------------------------------------------------------
#
#    SUBROUTINE: jtag_tdr_dump_contents()
#
#    USAGE: jtag_tdr_dump_contents($jtag_tdr_ref)
#
#    DESCRIPTION: The subroutine will display the contents of the JTAG TDR:
#       the name of the register and complete value are displayed, followed by
#       the name and value of each field.
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub jtag_tdr_dump_contents
{
   my $jtdr_ref = $_[0];  # Reference to the JTAG TDR structure
   my @labels   = jtag_tdr_get_fields($jtdr_ref, ORDER);
   my ($field, $data_ref, $data_str, $len, $start, $text);

   $data_ref = jtag_tdr_get_value($jtdr_ref);
   $data_str = vec_convert_num2str_hex(jtag_tdr_length($jtdr_ref), $data_ref, 'right');
   printf "\nREGISTER %-27s            = %s\n", jtag_tdr_name($jtdr_ref), $data_str;

   foreach $field ( @labels )
   {  $data_ref = jtag_tdr_get_field_value($jtdr_ref, $field);
      $len      = jtag_tdr_field_parm($jtdr_ref, $field, $TDR_FLD_LENGTH);
      $start    = jtag_tdr_field_parm($jtdr_ref, $field, $TDR_FLD_START);
      $data_str = vec_convert_num2str_hex($len, $data_ref, 'right');
      if ( $len == 1 )
      {  $text = sprintf("<%u>", $start);
      }
      else
      {  $text = sprintf("<%u:%u>", ($start + $len - 1), $start);
      }
      printf "   FIELD %-27s %-10s = %s\n", $field, $text, $data_str;
   }
}

#-------------------------------------------------------------------------------

1;
