# 
#/**************************************************************************
# *                                                                        *
# *               Copyright (C) 1989-1999 Silicon Graphics, Inc.           *
# *                                                                        *
# *  These coded instructions, statements, and computer programs  contain  *
# *  unpublished  proprietary  information of Silicon Graphics, Inc., and  *
# *  are protected by Federal copyright law.  They  may  not be disclosed  *
# *  to  third  parties  or copied or duplicated in any form, in whole or  *
# *  in part, without the prior written consent of Silicon Graphics, Inc.  *
# *                                                                        *
# **************************************************************************/
# 
# @(#)$Id: laaccess.pm,v 1.5 2004/05/07 18:44:14 beighley Exp $
#
#-------------------------------------------------------------------------------
#
#  FILE: laaccess.pm
#
#  DESCRIPTION:  This file contains PERL for the logic analyzer windows.
#
#
#  MODIFICATION HISTORY:
#
#     WHEN      WHO             WHAT
#     --------  --------------  -----------------------------------------
#     01/15/04  Karen Beighley  Added modification history to source file.
#                               Changed to use "Toplevel" for new window(s).
#     05/07/04  Karen Beighley  Added channel config buttons to download menu.
#                               Check for commas in testpoint and channel names.
#
#-------------------------------------------------------------------------------

use TLAControl;
use la_subs;
use groupdef;


################################################################################
#  GLOBAL VARIABLES
################################################################################

$CHAN_MAP_TP_LO = 0;
$CHAN_MAP_TP_HI = 1;
$CHAN_MAP_LA_CH = 2;
$CHAN_MAP_LA_LO = 3;
$CHAN_MAP_LA_HI = 4;

$LA_PARM_HOST   = 0;
$LA_PARM_PORT   = 1;
$LA_PARM_MODULE = 2;
$LA_PARM_TRUNC  = 3;


################################################################################
#  The LA download window lookup  structure is a HASH with one entry per port:
#    key   = full name of the chip (<module>:<board>:<chip>:<debug port>)
#    {key} = reference to debug port signal list (ARRAY)
#  The port has an entry in the lookup only if the download window is "open".
################################################################################

%ldw_lookup = ();

#-------------------------------------------------------------------------------

sub create_la_download_window
{  my $num = $#_ + 1;

   printf "\ncreate_la_download_window( @_ ) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 5 )
   {  printf "\nERROR - Invalid parameters passed to function!\n";
      return($FALSE);
   }
   my $win_main = $_[0]; # Main window for application
   my $dp_name  = $_[1]; # Full name of debug port to download
   my $dp_sigs  = $_[2]; # Reference to array of debug port signals
   my $chnmap   = $_[3]; # Reference to array with channel names
   my $la_parms = $_[4]; # Reference to array with LA parameters

   my $win_title = "Logical Analyzer: Download Debug Port";
   my $top_win   = $win_main->Toplevel(-title => $win_title);

   my $nchn = $#$chnmap + 1;
   my $ntp  = $#$dp_sigs + 1;
   my ($menubar, $casmenu, $casmenu2, $name, $frame, $label, $entry, 
       $canvas, $canvas_frame, $i, $bwidth, $btn, $cmd);

   #
   #  Create the menubar for the window
   #
   $menubar  = $top_win->Menu(-type => 'menubar');
   $top_win->configure(-menu => $menubar);
   $casmenu = $menubar->cascade(-label => '~File', -tearoff => 0);

   $casmenu->command(-label => 'Load LA Channel Configuration...',
      -command => [ \&window_cfg_load, $win_main, $chnmap]);
   $casmenu->command(-label => 'Save LA Channel Configuration...',
      -command => [ \&window_cfg_save, $win_main, 'LA_CHNCFG_WINDOW', $win_title, "", "", $chnmap]);
   $casmenu->command(-label => 'Edit LA Channel Configuration...',
                     -command => [ \&create_la_chncfg_window, $win_main, $nchn, $chnmap ]);
   $casmenu->command(-label => 'Reset LA Channel Configuration to Default',
      -command => [ \&la_chncfg_cbf_default, $chnmap]);

   $casmenu->separator;
   $cmd = $casmenu->command(-label => 'LA Group Configuration...');
   $cmd->configure(-command => [ \&create_groupdef_window, $dp_name, $dp_sigs, $chnmap, $la_parms, \$cmd ]);
   $casmenu->separator;
   $casmenu->command(-label => 'Close', -command => [ $top_win => 'destroy' ]);

   $casmenu  = $menubar->cascade(-label => '~Settings', -tearoff => 0);
   $casmenu2 = $casmenu->cascade(-label => 'Long Testpoint Names', -tearoff => 0);
   $casmenu2->radiobutton(-label => "Truncate Starting Characters", 
                          -value => 'pre', -variable => \$$la_parms[$LA_PARM_TRUNC]);
   $casmenu2->radiobutton(-label => "Truncate Trailing Characters", 
                          -value => 'end', -variable => \$$la_parms[$LA_PARM_TRUNC]);

   $top_win->bind('<Destroy>' => [\&destroy_la_download_window, $dp_name, $top_win, \$cmd] );

   #
   #  Create the main window label
   #
   $name = "Download Debug Port to Logic Analyzer: $dp_name";
   $frame = $top_win->Frame(-borderwidth => 10);
   $label = $frame->Label(-anchor => 'w', -text => $name);
   $label->pack(-side => 'left');
   $frame->pack(-side => 'top', -expand => 'no', -fill => 'both');

   #
   #  Create the widgets to specify the LA address
   #
   $frame = $top_win->Frame();
   $label = $frame->Label(-anchor => 'w', -width => 2);
   $label->pack(-side => 'left');
   $label = $frame->Label(-anchor => 'w', -text => "Logic Analyzer IP Address:", -width => 27);
   $label->pack(-side => 'left');
   $entry = $frame->Entry(-width => 15, -textvariable => \$$la_parms[$LA_PARM_HOST]);
   $entry->pack(-side => 'left');
   $label = $frame->Label(-anchor => 'w', -width => 2);
   $label->pack(-side => 'left');
   $label = $frame->Label(-anchor => 'w', -text => "Port:", -width => 5);
   $label->pack(-side => 'left');
   $entry = $frame->Entry(-width => 6, -textvariable => \$$la_parms[$LA_PARM_PORT]);
   $entry->pack(-side => 'left');
   $frame->pack(-side => 'top', -expand => 'no', -fill => 'both');

   #
   #  Create the widgets to specify the LA module
   #
   $frame = $top_win->Frame();
   $label = $frame->Label(-anchor => 'w', -width => 2);
   $label->pack(-side => 'left');
   $label = $frame->Label(-anchor => 'w', -text => "Logic Analyzer Module Name:", -width => 27);
   $label->pack(-side => 'left');
   $entry = $frame->Entry(-width => 15, -textvariable => \$$la_parms[$LA_PARM_MODULE]);
   $entry->pack(-side => 'left');
   $frame->pack(-side => 'top', -expand => 'no', -fill => 'both');

   #
   #  Create the field labels
   #
   $frame = $top_win->Frame(-borderwidth => 10);
   $label = $frame->Label(-text => "Bit", -width => 4, -anchor => 'w');
   $label->pack(-side => 'left');
   $label = $frame->Label(-text => "Testpoint Name", -width => 50, -anchor => 'w');
   $label->pack(-side => 'left');
   $label = $frame->Label(-text => "Channel Probe", -width => 20, -anchor => 'w');
   $label->pack(-side => 'left');
   $label = $frame->Label(-text => "Select", -width => 8, -anchor => 'w');
   $label->pack(-side => 'left');
   $frame->pack(-side => 'top', -fill => 'both', -expand => 'no');

   #
   #  Create the main canvas widget with scrollbars
   #
   $canvas = $top_win->Scrolled('Canvas', -scrollbars => 'se',
                -scrollregion => [0, 0, 100, (20 + ($ntp * 22))],
                -borderwidth => 2, -relief => 'sunken');
   $canvas_frame = $canvas->Frame(-borderwidth => 10);

   #
   #  Create the widgets for the register information
   #
   my @cb_state = ();
   for ($i = 0; $i < $ntp; $i++)
   {  $frame = $canvas_frame->Frame();
      $label = $frame->Label(-text => $i, -width => 4, -anchor => 'w');
      $label->pack(-side => 'left');

      $entry = $frame->Entry(-textvariable => \$$dp_sigs[$i], -width => 50);
      $entry->pack(-side => 'left');
      $entry = $frame->Entry(-textvariable => \$$chnmap[$i], -width => 20);
      $entry->pack(-side => 'left');

      $cb_state[$i] = 1;
      $check = $frame->Checkbutton(-variable => \$cb_state[$i]);
      $check->pack(-side => 'left');
      $frame->pack(-side => 'top', -fill => 'both', -expand => 'yes');
   }
   $canvas_frame->pack(-side => 'top', -fill => 'both', -expand => 'yes');
   $canvas->create('window', 4, 4, -anchor => 'nw', -window => $canvas_frame);
   $canvas->pack(-side => 'top', -fill => 'both', -expand => 'yes');


   $frame = $top_win->Frame(-borderwidth => 10);
   $label = $frame->Label(-text => "NOTE: Server program must be running on LA to download.", -anchor => 'w');
   $label->pack(-side => 'left');
   $frame->pack(side => 'top', -fill => 'both', -expand => 'no');

   #
   #  Create a frame to hold the action buttons
   #
   $frame = $top_win->Frame(-borderwidth => 10);
   $bwidth = 12;
   $btn = $frame->Button(-text => 'Write All', -width => $bwidth,
             -command => [\&la_download_cbf_write, $top_win, $FALSE, $dp_sigs, $chnmap,
                          \@cb_state, $la_parms]);
   $btn->pack(-side => 'left', -padx => 5);
   $btn = $frame->Button(-text => 'Write Selected', -width => $bwidth,
             -command => [\&la_download_cbf_write, $top_win, $TRUE, $dp_sigs, $chnmap, 
                         \@cb_state, $la_parms]);
   $btn->pack(-side => 'left', -padx => 5);
   $btn = $frame->Button(-text => 'Select All', -width => $bwidth,
             -command => [\&la_download_cbf_select, 1, \@cb_state]);
   $btn->pack(-side => 'left', -padx => 5);
   $btn = $frame->Button(-text => 'Select None', -width => $bwidth,
             -command => [\&la_download_cbf_select, 0, \@cb_state]);
   $btn->pack(-side => 'left', -padx => 5);
   $btn = $frame->Button(-text => 'Close Window', -width => $bwidth,
                    -command => [ $top_win => 'destroy' ]);
   $btn->pack(-side => 'right', -padx => 5);
   $frame->pack(side => 'top', -fill => 'both', -expand => 'no');

   #
   #  Add entry to lookup for LA download window(s)
   #
   $lwd_lookup{$dp_name} = $dp_sigs;
   if ($debug_flag & $DEBUG_GENERAL)
   {  print "\nSTORED WINDOW DATA FOR: $dp_name = $dp_sigs = ( @$dp_sigs )\n";
   }

   return($TRUE);
}


#-------------------------------------------------------------------------------

sub destroy_la_download_window
{  
#   printf "\ndestroy_la_download_window( @_ )\n";

   return if ( $#_ != 3 );
   my $curr_wid = $_[0]; # Current widget/window handle from Tk
   my $dp_name  = $_[1]; # Full name of debug port to download
   my $win_ref  = $_[2]; # Reference/handle to the debug port window
   my $cmd_ref  = $_[3]; # Reference/handle to the debug port window

   return if ( $curr_wid != $win_ref ); # Return if not main window

   # Remove entry from "dwd_lookup{}" for debug port
   delete $lwd_lookup{$dp_name};
   if ($debug_flag & $DEBUG_GENERAL)
   {  print "\nDELETED WINDOW DATA FOR: $dp_name\n";
   }
   $$cmd_ref = 0;
}


#-------------------------------------------------------------------------------

sub update_la_download_window
{
   printf "\nupdate_la_download_window( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   return if ( $#_ != 1 );
   my $dp_name  = $_[0]; # Full name of debug port to download
   my $dport    = $_[1]; # Pointer to debug port C-structure
   my ($num, @new_sigs, $dp_sigs, $i);

   return($FALSE) if ( !exists($lwd_lookup{$dp_name}) );

   #  Get the list of testpoints from the database
   $num = pcf_dport_get_parm($dport, 'ntp');
   @new_sigs = ();
   if ( pcf_dport_get_tp_names($dport, 0, $num, \@new_sigs) != $num )
   {  $emess = "Unable to extract debug port testpoints from database.";
      superscope_error_handler($emess);
      return($FALSE);;
   }

   #  Update the testpoint pointers
   $dp_sigs = $lwd_lookup{$dp_name};
   for ($i = 0; $i < $num; $i++)
   {  $$dp_sigs[$i] = $new_sigs[$i];
   }
   return($TRUE);
}


#-------------------------------------------------------------------------------

sub create_la_chncfg_window
{  my $num = $#_ + 1;

   printf "\ncreate_la_chncfg_window( @_ ) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 3 )
   {  printf "\nERROR - Invalid parameters passed to function!\n";
      return($FALSE);
   }
   my $win_main  = $_[0]; # Main window for application
   my $num_tp    = $_[1]; # Number of testpoints in debug port
   my $chnmap    = $_[2]; # Reference to array with channel names
   my ($menubar, $casmenu, $name, $canvas, $canvas_frame, $i, 
       $map_frame, $frame, $label, $entry, $btn);

   my $win_title = "Logic Analyzer: Channel Configuration";
   my $top_win   = $win_main->Toplevel(-title => $win_title);
   my $bwidth    = 12;

   #
   #  Create the menubar for the window
   #
   $menubar  = $top_win->Menu(-type => 'menubar');
   $top_win->configure(-menu => $menubar);
   $casmenu = $menubar->cascade(-label => '~File', -tearoff => 0);
   $casmenu->command(-label => 'Load LA Channel Configuration...',
      -command => [ \&window_cfg_load, $win_main, $chnmap]);
   $casmenu->command(-label => 'Save LA Channel Configuration...',
      -command => [ \&window_cfg_save, $win_main, 'LA_CHNCFG_WINDOW', $win_title, "", "", $chnmap]);
   $casmenu->command(-label => 'Reset LA Channel Configuration to Default',
      -command => [ \&la_chncfg_cbf_default, $chnmap]);
   $casmenu->separator;
   $casmenu->command(-label => 'Close', -command => [ $top_win => 'destroy' ]);

   #
   #  Create the header
   #
   $name = "Debug Port Mapping to Logic Analyzer Channels";
   $frame = $top_win->Frame(-borderwidth => 10);
   $label = $frame->Label(-anchor => 'w', -text => $name);
   $label->pack(-side => 'left');
   $frame->pack(-side => 'top', -expand => 'no', -fill => 'both');

   $frame = $top_win->Frame(-borderwidth => 10);
   $label = $frame->Label(-text => "Bit", -width => 4, -anchor => 'w');
   $label->pack(-side => 'left');
   $label = $frame->Label(-text => "Channel Probe", -width => 15, -anchor => 'w');
   $label->pack(-side => 'left');
   $frame->pack(-side => 'top', -fill => 'both', -expand => 'no');

   #
   #  Create the main canvas widget with scrollbars
   #
   # NOTE: Determine height by counting the number of registers and fields
   $canvas = $top_win->Scrolled('Canvas', -scrollbars => 'se',
                -scrollregion => [0, 0, 15, (20 + ($num_tp * 22))],
                -borderwidth => 2, -relief => 'sunken');
   $canvas_frame = $canvas->Frame(-borderwidth => 10);

   #
   #  Create the widgets for the register information
   #
   for ($i = 0; $i < $num_tp; $i++)
   {  
      $frame = $canvas_frame->Frame();
      $label = $frame->Label(-text => $i, -width => 4, -anchor => 'w');
      $label->pack(-side => 'left');
      $entry = $frame->Entry(-width => 10, -textvariable => \$$chnmap[$i]);
      $entry->pack(-side => 'left');
      $frame->pack(-side => 'top', -fill => 'both', -expand => 'yes');
   }
   $canvas_frame->pack(-side => 'top', -fill => 'both', -expand => 'yes');
   $canvas->create('window', 0, 0, -anchor => 'nw', -window => $canvas_frame);
   $canvas->pack(-side => 'top', -fill => 'both', -expand => 'yes');

   my @map_data = ( "", "", "", "", "" );

   #
   #  Create a frame the mapping widgets
   #
   $map_frame = $top_win->Frame(-borderwidth => 10);

   $frame = $map_frame->Frame();
   $label = $frame->Label(-text => "Define Multiple Bits Using Bit Range Mapping:", -anchor => 'w');
   $label->pack(-side => 'left');
   $frame->pack(-side => 'top', -fill => 'both', -expand => 'yes');

   $frame = $map_frame->Frame();
   $label = $frame->Label(-text => 'Debug port data:', -width => 15, -anchor => 'w');
   $label->pack(-side => 'left');
   $label = $frame->Label(-text => 'bits', -width => 5, -anchor => 'e');
   $label->pack(-side => 'left');
   $entry = $frame->Entry(-width => 4, -textvariable => \$map_data[$CHAN_MAP_TP_LO]);
   $entry->pack(-side => 'left');
   $label = $frame->Label(-text => 'to', -width => 2, -anchor => 'w');
   $label->pack(-side => 'left');
   $entry = $frame->Entry(-width => 4, -textvariable => \$map_data[$CHAN_MAP_TP_HI]);
   $entry->pack(-side => 'left');
   $frame->pack(side => 'top', -fill => 'both');

   $frame = $map_frame->Frame();
   $label = $frame->Label(-text => 'LA channel:', -width => 10, -anchor => 'w');
   $label->pack(-side => 'left');
   $entry = $frame->Entry(-width => 4, -textvariable => \$map_data[$CHAN_MAP_LA_CH]);
   $entry->pack(-side => 'left');
   $label = $frame->Label(-text => 'bits', -width => 5, -anchor => 'e');
   $label->pack(-side => 'left');
   $entry = $frame->Entry(-width => 4, -textvariable => \$map_data[$CHAN_MAP_LA_LO]);
   $entry->pack(-side => 'left');
   $label = $frame->Label(-text => 'to', -width => 2, -anchor => 'w');
   $label->pack(-side => 'left');
   $entry = $frame->Entry(-width => 4, -textvariable => \$map_data[$CHAN_MAP_LA_HI]);
   $entry->pack(-side => 'left');
   $frame->pack(side => 'top', -fill => 'both');

   $map_frame->pack(side => 'top', -fill => 'both');

   #
   #  Create a frame to hold the action buttons
   #
   $frame = $top_win->Frame(-borderwidth => 10);
   $btn = $frame->Button(-text => 'Map Bits', -width => $bwidth,
               -command => [\&la_chncfg_cbf_map, $top_win, $num_tp, $chnmap, \@map_data]);
   $btn->pack(-side => 'left', -padx => 5);
   $btn = $frame->Button(-text => 'Close Window', -width => $bwidth,
                    -command => [ $top_win => 'destroy' ]);
   $btn->pack(-side => 'right', -padx => 5);
   $frame->pack(side => 'top', -fill => 'both');

   return($TRUE);
}

#-------------------------------------------------------------------------------

sub la_chncfg_cbf_default
{
   print "\nla_chncfg_cbf_default( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);
   my $chnmap    = $_[0]; # Reference to array with channel names

   la_default_chnmap($chnmap);
}


#-------------------------------------------------------------------------------

sub la_chncfg_cbf_map
{
   print "\nla_chncfg_cbf_map( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   return if ( $#_ != 3 );
   my $top_win   = $_[0]; # Top-level window for LA configuration
   my $num_tp    = $_[1]; # Number of testpoints in debug port
   my $chnmap    = $_[2]; # Reference to array with channel names
   my $map_data  = $_[3]; # Reference to array with parameters
   my ($dp_num, $dp_pos, $dp_inc, $la_num, $la_pos, $la_inc, $text, $i);

   #
   #  Verify the user defined parameters 
   #
   if ( ($$map_data[$CHAN_MAP_TP_LO] eq "") || ($$map_data[$CHAN_MAP_TP_HI] eq "") ||
        ($$map_data[$CHAN_MAP_LA_LO] eq "") || ($$map_data[$CHAN_MAP_LA_HI] eq "") ||
        ($$map_data[$CHAN_MAP_LA_CH] eq "") )
   {  $text = "Required parameter(s) undefined: all mapping fields must be specified.";
      $win_main->messageBox(-icon => 'error', -type => 'OK', 
            -message => $text, -title => 'LA Channel Configuration Error');
      return;
   }
   if ( $$map_data[$CHAN_MAP_TP_HI] >= $num_tp )
   {  $text = "Invalid bit range for debug port: upper bit exceeds limit of $num_tp.";
      $win_main->messageBox(-icon => 'error', -type => 'OK', 
            -message => $text, -title => 'LA Channel Configuration Error');
      return;
   }

   if ( $$map_data[$CHAN_MAP_TP_LO] > $$map_data[$CHAN_MAP_TP_HI] )
   {  $dp_num = $$map_data[$CHAN_MAP_TP_LO] - $$map_data[$CHAN_MAP_TP_HI] + 1;
      $dp_pos = $$map_data[$CHAN_MAP_TP_LO];
      $dp_inc = -1;
   }
   else
   {  $dp_num = $$map_data[$CHAN_MAP_TP_HI] - $$map_data[$CHAN_MAP_TP_LO] + 1;
      $dp_pos = $$map_data[$CHAN_MAP_TP_LO];
      $dp_inc = 1;
   }

   if ( $$map_data[$CHAN_MAP_LA_LO] > $$map_data[$CHAN_MAP_LA_HI] )
   {  $la_num = $$map_data[$CHAN_MAP_LA_LO] - $$map_data[$CHAN_MAP_LA_HI] + 1;
      $la_pos = $$map_data[$CHAN_MAP_LA_LO];
      $la_inc = -1;
   }
   else
   {  $la_num = $$map_data[$CHAN_MAP_LA_HI] - $$map_data[$CHAN_MAP_LA_LO] + 1;
      $la_pos = $$map_data[$CHAN_MAP_LA_LO];
      $la_inc = 1;
   }

   if ($debug_flag & $DEBUG_DETAILS)
   {  printf("DP: num=%02u pos=%02u inc=%d\n", $dp_num, $dp_pos, $dp_inc);
      printf("LA: num=%02u pos=%02u inc=%d\n", $la_num, $la_pos, $la_inc);
   }

   if ( $dp_num != $la_num )
   {  $text = "Range mismatch: number of debug port bits not equal to number of channel bits.";
      $win_main->messageBox(-icon => 'error', -type => 'OK', 
            -message => $text, -title => 'LA Channel Configuration Error');
      return;
   }

   #
   #  Change the channel mapping name(s)
   #
   for ($i = 0; $i < $dp_num; $i++)
   {  $text = sprintf("%s(%u)", $$map_data[$CHAN_MAP_LA_CH], $la_pos);
      $$chnmap[$dp_pos] = $text;
      $dp_pos += $dp_inc;
      $la_pos += $la_inc;
   }
}


#-------------------------------------------------------------------------------

sub la_download_cbf_select
{  print "\nla_download_cbf_select(@_) : #parms=$#_\n" if ($debug_flag & $DEBUG_FUNCTIONS);
   
   my $value    = $_[0]; # Value for all selection checkboxes
   my $cb_state = $_[1]; # Reference to array containing checkbutton state
   my $i;

   #  Set or clear all checkboxes
   for ($i = 0; $i <= $#$cb_state; $i++)
   {  $$cb_state[$i] = $value;
   }
}


#-------------------------------------------------------------------------------

sub la_download_cbf_write
{  print "\nla_download_cbf_write(@_) : #parms=$#_\n" if ($debug_flag & $DEBUG_FUNCTIONS);
 
   return if ( $#_ != 5 );
   my $top_win  = $_[0]; # Top-level window for LA configuration
   my $sel_flag = $_[1]; # Flag: 1=write selected, 0=write all in window 
   my $dp_sigs  = $_[2]; # Reference to array of debug port signals
   my $chnmap   = $_[3]; # Reference to array with channel names
   my $cb_state = $_[4]; # Reference to array containing checkbutton state
   my $la_parms = $_[5]; # Reference to array with LA parameters
   my ($i, $text, $num, $chn_text, $lab_text, $tmp, $len);

   #
   #  Verify the parameters
   #
   if ( $$la_parms[$LA_PARM_HOST] eq "" )
   {  $text = "The Logic Analyzer's IP address must be specified.";
      $top_win->messageBox(-icon => 'error', -type => 'OK', 
            -message => $text, -title => 'Download to LA Error');
      return;
   }
   if ( $$la_parms[$LA_PARM_PORT] eq "" )
   {  $text = "The Logic Analyzer's port must be specified.";
      $top_win->messageBox(-icon => 'error', -type => 'OK', 
            -message => $text, -title => 'Download to LA Error');
      return;
   }
   if ( $$la_parms[$LA_PARM_MODULE] eq "" )
   {  $text = "The Logic Analyzer's module name must be specified.";
      $top_win->messageBox(-icon => 'error', -type => 'OK', 
            -message => $text, -title => 'Download to LA Error');
      return;
   }

   #
   #  Build the list of channels and labels to write to the LA
   #
   $num      = 0;
   $chn_text = "";
   $lab_text = "";
   for ($i = 0; $i <= $#$dp_sigs; $i++)
   {  if ( ($sel_flag == $FALSE) ||
           (($sel_flag == $TRUE) && $$cb_state[$i]) )
      {  
         $tmp = $$dp_sigs[$i];
         $len = length($tmp);
         if ( $len > $LA_CHAN_NAME_LEN )
         {  if ( $$la_parms[$LA_PARM_TRUNC] eq 'end' )
            {  $tmp = substr($tmp, 0, $LA_CHAN_NAME_LEN);
            }
            else
            {  $tmp = substr($tmp, ($len-$LA_CHAN_NAME_LEN), $LA_CHAN_NAME_LEN);
            }
         }

         if ( $tmp =~ m/,/ )
         {  $text = sprintf("Testpoint name for bit $i contains comma: \"%s\"", $tmp);
            $top_win->messageBox(-icon => 'error', -type => 'OK', 
                  -message => $text, -title => 'Download to LA Error');
            return;
         }
         if ( $$chnmap[$i] =~ m/,/ )
         {  $text = sprintf("Channel probe name for bit $i contains comma: \"%s\"", $$chnmap[$i]);
            $top_win->messageBox(-icon => 'error', -type => 'OK', 
                  -message => $text, -title => 'Download to LA Error');
            return;
         }

         if ( $num > 0 )
         {  $chn_text .= "," . $$chnmap[$i];
            $lab_text .= "," . $tmp;
         }
         else
         {  $chn_text .= $$chnmap[$i];
            $lab_text .= $tmp;
         }
         $num++;
      }
   }
   if ( $num == 0 )
   {  $text = "No debug port data bits selected to write to logic analyzer.";
      $top_win->messageBox(-icon => 'error', -type => 'OK', 
            -message => $text, -title => 'Download to LA Error');
      return;
   }

   #
   #  Connect to the LA and write the labels to the specified channel probe(s)
   #
   la_connect($$la_parms[$LA_PARM_HOST], $$la_parms[$LA_PARM_PORT]);
   return if ( $ecode < 0 );

   printf "\nDownloading debug port to LA: module=%s, num=%u\n", $$la_parms[$LA_PARM_MODULE], $num;
   printf "\tChannels   : \"$chn_text\"\n";
   printf "\tLabels     : \"$lab_text\"\n";

   la_module_set_channel_name($$la_parms[$LA_PARM_MODULE], $chn_text, $lab_text);
   return if ( $ecode < 0 );

   la_disconnect();
   return;
}


#-------------------------------------------------------------------------------

1;
