# 
#/**************************************************************************
# *                                                                        *
# *               Copyright (C) 1989-1999 Silicon Graphics, Inc.           *
# *                                                                        *
# *  These coded instructions, statements, and computer programs  contain  *
# *  unpublished  proprietary  information of Silicon Graphics, Inc., and  *
# *  are protected by Federal copyright law.  They  may  not be disclosed  *
# *  to  third  parties  or copied or duplicated in any form, in whole or  *
# *  in part, without the prior written consent of Silicon Graphics, Inc.  *
# *                                                                        *
# **************************************************************************/
# 
# @(#)$Id: md_dir_state.pm,v 1.28 2007/03/14 22:53:01 meyer Exp $
#-------------------------------------------------------------------------------
#
#  FILE: md_dir_state.pm
#
#  DESCRIPTION:  This file contains PERL subroutines to extract md directory state
#
#  MODIFICATION HISTORY:
#
#     WHEN      WHO             WHAT
#     --------  --------------  -----------------------------------------
#     01/08/03  Gary Meyer      Original source code submitted for release.
#     01/15/03  Karen Beighley  Minor change to source.
#     03/18/03  Gary Meyer      Fixed bug in get_mem_cfg that caused banks to always be 0.
#     04/24/03  Gary Meyer      Added access and priority info to output.
#     10/29/03  Gary Meyer      Now displaying memory size for each group.  Compressed output
#                               of directory state information.
#     03/10/04  Gary Meyer      Modified chiplet value for TIO nodes.
#     09/07/04  Gary Meyer	Added retries for superscope mmr accesses.  Added printing of SHRD vectors.
#     01/06/05  Gary Meyer     	Added DIMM Speed and Cas Latency values, and improved output formatting.
#     05/23/05  Gary Meyer	Moved functions from errdmp to md_dir_state.pm.
#     06/18/05  Gary Meyer	Moved save_address() from errdmp.  Fix to grab the correct target_list
#				based on which shub version is being dumped.  Directory state dumping now
#				working in Shub2.
#     07/16/05  Gary Meyer	New function for decoding directory state.  Added Shub2 dimm config decode.
#     05/04/06  Gary Meyer	Update to prevent bogus entries in dir_target_nodes array from being used.
#     06/01/06  Gary Meyer	Added source_type to "already in the list" message.  Seems to be some type of
#				corner-case endless loop that we can hit where this message gets repeated 
#				continually.
#     06/09/06  Gary Meyer	Added debug code to identify why we sometimes get UNDEFINED dir state.
#     06/14/06  Gary Meyer	Printing Unc & Cor values for each directory state.  Added acc & pri to SHRD info.
#     06/15/06  Gary Meyer	Added BSYUP (0x79) directory state.  This is a new state for Shub2 that was 
#				omitted from original Shub2 updates.
#     06/19/06  Gary Meyer	Update to prevent re-use of address entries with duplicate NASIDs.
#     06/23/06  Gary Meyer	Update to abort directory state capture if mmr_target() fails.
#     06/29/06  Gary Meyer	Update for Shub2 to use 14 bit node value for $dir_node.
#     07/17/06  Gary Meyer	Update to decode_dir_state to ensure unsupplied access and priority values
#				are indicated as such.
#     07/24/06  Gary Meyer	Update to ensure all NASID values use 4 characters.
#     10/16/06  Gary Meyer	Initialize $SSCOPE_ERR before doing any jtag accesses.
#     12/11/06  Gary Meyer	Updates to ensure directory state is correctly captured on non-booted Shub nodes.
#     03-14-07  Gary Meyer      Update for physical dimm group remapping.
#
#
#-------------------------------------------------------------------------------


#-------------------------------------------------------------------------------
#
#  This subroutine is called after all MMRs have been read, and only if addresses
#  need to have the directory state checked.
#  The nasid list is stepped through once, and the address list is looped through
#  once each iteration, checking for any addresses that appear on each nasid.
#
#-------------------------------------------------------------------------------

sub sort_and_display_dir_state {

  print "\nIn sort_and_display_dir_state function\n" if ($verbose_mode == 1);

  my $lcl_nasid;        #  local nasid used when looping through the nasid_list array
  my $a_cnt;            #  variable used to loop through the address_list array
  my $first_one;        #  used to determine when to update the mmr_target used by superscope

  sort_by_addr();       #  First numerically sort the addresses

  if ($parallel_mode == 1)  {
    gen_global_nasid_list();
    *dir_target_nodes = \@global_nasid_list;    # create a symbolic reference
  }
  else {
    if ($shub_version == 2)  {
      *dir_target_nodes = \@sh2_target_list;          # create a symbolic reference
    }
    else {
      *dir_target_nodes = \@target_list;          # create a symbolic reference
    }
  }

  #  The outer loop iterates through available NASID values
  #  First print the list of addresses to check...

  print "\n  There are $addr_cnt addresses to obtain directory state for.\n";
  print "\n The list of Addresses to check directory state on is as follows: \n\n";
  foreach my $lcl_list (@dir_target_nodes)  {
    if ($verbose_mode == 1)  {
        printf "\n Entry values: NASID: 0x%04X  NODE: %s  MOD: %s\n",
            $lcl_list->{NASID}, $lcl_list->{NODE}, $lcl_list->{MODULE}
    }
    #  The inner loop iterates through the address list to get directory states for
    for ($a_cnt = 0; $a_cnt < $addr_cnt; $a_cnt++) {
      if (( $final_addr_list[$a_cnt]->{NASID} == $lcl_list->{NASID} ) && (! defined $final_addr_list[$a_cnt]->{DISPLAYED})) {
        printf "\tAddress 0x%010llX on NASID 0x%04X         Occurrences: %d\n",
          $final_addr_list[$a_cnt]->{ADDRESS}, $final_addr_list[$a_cnt]->{NASID}, $final_addr_list[$a_cnt]->{COUNT};
        $final_addr_list[$a_cnt]->{DISPLAYED} = 1;
      }
    }   # end of "for" loop - continue to the next address value
  }     # end of "foreach" loop - continue on to the next NASID value

  print "\n";
  $a_cnt = 0;
  foreach (@final_addr_list)  {
    if (! defined $final_addr_list[$a_cnt]->{DISPLAYED})  {
         printf "ALERT!  Invalid NASID indicated:  Address 0x%010llX on NASID 0x%04X\n",
                $final_addr_list[$a_cnt]->{ADDRESS}, $final_addr_list[$a_cnt]->{NASID};
    }
    $a_cnt++;
  }

  my $dir_state_complete = 0;
  my $initial_err_cnt = $reported_err_cnt;
  while (($reported_err_cnt < ($initial_err_cnt + 10)) && ($dir_state_complete == 0)) {
    my $shub_index = 0;
    foreach $lcl_list (@dir_target_nodes)  {
      $first_one = 1;
      my $mask_updated = 0;		# set if the MD error mask was changed
      my $shub_loc = "";
      #  The inner loop iterates through the address list to get directory states for
      for ($a_cnt = 0; $a_cnt < $addr_cnt; $a_cnt++) {
        if (( $final_addr_list[$a_cnt]->{NASID} == $lcl_list->{NASID} ) 
	    && (! defined $final_addr_list[$a_cnt]->{DONE}) 
	    && (defined $lcl_list->{MODULE}) 
	    && (defined $lcl_list->{NODE}) 
	    && (defined $lcl_list->{CHIP})) {
          if ($first_one == 1) {
            $first_one = 0;
	    $shub_loc = sprintf "%s of %s", $lcl_list->{NODE}, $lcl_list->{MODULE};
            $SSCOPE_ERR = 0;
            mmr_target($SELECT, $lcl_list->{MODULE}, $lcl_list->{NODE}, $lcl_list->{CHIP});
	    if ($SSCOPE_ERR == 1)  {
	      printf "\n mmr_target\(\) failed.  Aborting dir state read for addr 0x%010llX on NASID 0x%04X  %s\n\n",
                  $final_addr_list[$a_cnt]->{ADDRESS}, $final_addr_list[$a_cnt]->{NASID}, $shub_loc;
              $SSCOPE_ERR = 0;
              $first_one = 1;
	      next;
            }

	    if ($shub_version == 2)  {
	      if (($sh2_md_mask & (1 << 39)) == 0) {
		$mask_data = ($sh2_md_mask | (1 << 39));
		update_md_err_mask($mask_data);
		$mask_updated = 1;
	      }
	    }
          }

            # since the $target_list mapping doesn't correspond to the parallel list
            # force this value so all sscope failures during mbist operations get
            # allocated to the first shub.

          if ($parallel_mode == 1)  {
            $shub_index = 0;
          }

          print "\n-------------------------------------------------------\n";
          printf "\n Getting directory state for address 0x%010llX on NASID 0x%04X    %s\n\n",
                  $final_addr_list[$a_cnt]->{ADDRESS}, $final_addr_list[$a_cnt]->{NASID}, $shub_loc;
          get_dir_state(  $final_addr_list[$a_cnt]->{ADDRESS}, $verbose_mode, $shub_index);

	  $final_addr_list[$a_cnt]->{DONE} = 1;		# to prevent re-use of this entry with duplicate NASIDs
        }
      } # end of "for" loop - continue to the next address value
      $shub_index++;
      if (($shub_version == 2) && ($mask_updated == 1))  {
	update_md_err_mask($sh2_md_mask);
      }
    }   # end of "foreach" loop - continue on to the next NASID value

    print "\n----------------------------------------------------------------------\n\n";
    $dir_state_complete = 1;
  }

  if ($reported_err_cnt >= ($initial_err_cnt + 10))  {
    print "\n\nAborting attempt to read Shub MD directory state due to errors received.\n\n";
  }

}

#-------------------------------------------------------------------------------
#
#  Function to display directory state on non-booted Shub2 nodes.
#
#-------------------------------------------------------------------------------

sub display_nb_dir_state {

  print "\n\n  There are $nb_addr_cnt addresses on non-booted Shub2 nodes to obtain directory state for.\n";
  print "\n The list of Addresses to check directory state on is as follows: \n\n";

  foreach $nb_addr (@nb_address_list) {
    printf "\tAddress 0x%010llX on NASID 0x%04X         Location: %s, %s\n",
      $nb_addr->{ADDRESS}, $nb_addr->{NASID}, $nb_addr->{MODULE}, $nb_addr->{NODE};
  }

  foreach $nb_addr (@nb_address_list) {
    my $mask_updated = 0;
    my $shub_loc = sprintf "%s of %s", $nb_addr->{NODE}, $nb_addr->{MODULE};
    $SSCOPE_ERR = 0;
    mmr_target($SELECT, $nb_addr->{MODULE}, $nb_addr->{NODE}, $nb_addr->{CHIP});
    if ($SSCOPE_ERR == 1)  {
      printf "\n mmr_target\(\) failed.  Aborting dir state read for addr 0x%010llX on NASID 0x%04X  %s\n\n",
        $nb_addr->{ADDRESS}, $nb_addr->{NASID}, $shub_loc;
      $SSCOPE_ERR = 0;
      next;
    }
    if (($sh2_md_mask & (1 << 39)) == 0) {
      $mask_data = ($sh2_md_mask | (1 << 39));
      update_md_err_mask($mask_data);
      $mask_updated = 1;
    }

    print "\n-------------------------------------------------------\n";
    printf "\n Getting directory state for address 0x%010llX on NASID 0x%04X    %s\n\n",
        $nb_addr->{ADDRESS}, $nb_addr->{NASID}, $shub_loc;
    get_dir_state($nb_addr->{ADDRESS}, $verbose_mode, 0);

    if ($mask_updated == 1)  {
      update_md_err_mask($sh2_md_mask);
    }
  } 

  print "\n----------------------------------------------------------------------\n\n";

}


#-------------------------------------------------------------------------------
#
#  subroutine:  get_dir_state
#
#  arguments should be received in this order:
#
#    $starting_addr - hex value, up to 36 bits for the address to check (mandatory)
#    $verbose	    - print extra debug info
#	
#
#-------------------------------------------------------------------------------

sub get_dir_state  {

	$FALSE           = 0;
	$TRUE            = 1;
	$node            = 0;
	$verbose         = 0;
	$starting_addr   = 0x0;
	$fail_check      = 0;

	if (! defined $_[0])  {
	  print " ERROR!  Address not provided to get_dir_state subroutine.\n";
	  print " Returning without reading MD directory state.\n\n";
	}
	else  {
	  $starting_addr = $_[0];
	}

	if (defined $_[1]) {
	  $verbose  = $_[1];
	}

	my $shub_index = $_[2];

	if ($verbose > 0)  {
	  print "\n Directory state parameters to be used: \n";
	  printf "\tAddr: 0x%010llX   \n\n", $starting_addr;
	}

	if ($shub_version == 1)  {
	    ############## MD BIST DIRECTORY ENTRY MMRS ###################################
	  $dqlp_piowd_dir = reg_create_mmr_inst("shub_01", "SH_MD_DQLP_MMR_PIOWD_DIR_ENTRY");
	  $dqlp_piowd_ecc = reg_create_mmr_inst("shub_01", "SH_MD_DQLP_MMR_PIOWD_DIR_ECC");
	  $dqrp_piowd_dir = reg_create_mmr_inst("shub_01", "SH_MD_DQRP_MMR_PIOWD_DIR_ENTRY");
	  $dqrp_piowd_ecc = reg_create_mmr_inst("shub_01", "SH_MD_DQRP_MMR_PIOWD_DIR_ECC");

	  $lp_xdir_ecc   = reg_create_mmr_inst("shub_01", "SH_MD_DQLP_MMR_XPIORD_XDIR_ECC");
	  $lp_xdir_entry = reg_create_mmr_inst("shub_01", "SH_MD_DQLP_MMR_XPIORD_XDIR_ENTRY");
	  $lp_ydir_ecc   = reg_create_mmr_inst("shub_01", "SH_MD_DQLP_MMR_YPIORD_YDIR_ECC");
	  $lp_ydir_entry = reg_create_mmr_inst("shub_01", "SH_MD_DQLP_MMR_YPIORD_YDIR_ENTRY");
	  $rp_xdir_ecc   = reg_create_mmr_inst("shub_01", "SH_MD_DQRP_MMR_XPIORD_XDIR_ECC");
	  $rp_xdir_entry = reg_create_mmr_inst("shub_01", "SH_MD_DQRP_MMR_XPIORD_XDIR_ENTRY");
	  $rp_ydir_ecc   = reg_create_mmr_inst("shub_01", "SH_MD_DQRP_MMR_YPIORD_YDIR_ECC");
	  $rp_ydir_entry = reg_create_mmr_inst("shub_01", "SH_MD_DQRP_MMR_YPIORD_YDIR_ENTRY");

	    ############## MD SUMMARY ERROR MMRS ###################################
	  $sh_mmrbist_err = reg_create_mmr_inst("shub_01", "SH_MMRBIST_ERR");
	  $sum_err        = reg_create_mmr_inst("shub_01", "SH_MEM_ERROR_SUMMARY");
	  $sum_err_al     = reg_create_mmr_inst("shub_01", "SH_MEM_ERROR_SUMMARY_ALIAS");
	  $ovfl_err       = reg_create_mmr_inst("shub_01", "SH_MEM_ERROR_OVERFLOW");
	  $ovfl_err_al    = reg_create_mmr_inst("shub_01", "SH_MEM_ERROR_OVERFLOW_ALIAS");

	    ################ MD BIST CONTROL MMRS ##########################
	  $mbist_base = reg_create_mmr_inst("shub_01", "SH_MMRBIST_BASE");
	  $mbist_ctl  = reg_create_mmr_inst("shub_01", "SH_MMRBIST_CTL");
	}
	elsif ($shub_version == 2)  {
	    ############## MD BIST DIRECTORY ENTRY MMRS ###################################
	  $dqlp_piowd_dir = reg_create_mmr_inst("shub2_01", "SH2_MD_DQLP_MMR_PIOWD_DIR_ENTRY");
	  $dqlp_piowd_ecc = reg_create_mmr_inst("shub2_01", "SH2_MD_DQLP_MMR_PIOWD_DIR_ECC");
	  $dqrp_piowd_dir = reg_create_mmr_inst("shub2_01", "SH2_MD_DQRP_MMR_PIOWD_DIR_ENTRY");
	  $dqrp_piowd_ecc = reg_create_mmr_inst("shub2_01", "SH2_MD_DQRP_MMR_PIOWD_DIR_ECC");

	  $lp_xdir_ecc   = reg_create_mmr_inst("shub2_01", "SH2_MD_DQLP_MMR_XPIORD_XDIR_ECC");
	  $lp_xdir_entry = reg_create_mmr_inst("shub2_01", "SH2_MD_DQLP_MMR_XPIORD_XDIR_ENTRY");
	  $lp_ydir_ecc   = reg_create_mmr_inst("shub2_01", "SH2_MD_DQLP_MMR_YPIORD_YDIR_ECC");
	  $lp_ydir_entry = reg_create_mmr_inst("shub2_01", "SH2_MD_DQLP_MMR_YPIORD_YDIR_ENTRY");
	  $rp_xdir_ecc   = reg_create_mmr_inst("shub2_01", "SH2_MD_DQRP_MMR_XPIORD_XDIR_ECC");
	  $rp_xdir_entry = reg_create_mmr_inst("shub2_01", "SH2_MD_DQRP_MMR_XPIORD_XDIR_ENTRY");
	  $rp_ydir_ecc   = reg_create_mmr_inst("shub2_01", "SH2_MD_DQRP_MMR_YPIORD_YDIR_ECC");
	  $rp_ydir_entry = reg_create_mmr_inst("shub2_01", "SH2_MD_DQRP_MMR_YPIORD_YDIR_ENTRY");

	    ############## MD SUMMARY ERROR MMRS ###################################
	  $sh_mmrbist_err = reg_create_mmr_inst("shub2_01", "SH2_MMRBIST_ERR");
	  $sum_err        = reg_create_mmr_inst("shub2_01", "SH2_MEM_ERROR_SUMMARY");
	  $sum_err_al     = reg_create_mmr_inst("shub2_01", "SH2_MEM_ERROR_SUMMARY_ALIAS");
	  $ovfl_err       = reg_create_mmr_inst("shub2_01", "SH2_MEM_ERROR_OVERFLOW");
	  $ovfl_err_al    = reg_create_mmr_inst("shub2_01", "SH2_MEM_ERROR_OVERFLOW_ALIAS");

	    ################ MD BIST CONTROL MMRS ##########################
	  $mbist_base = reg_create_mmr_inst("shub2_01", "SH2_MMRBIST_BASE");
	  $mbist_ctl  = reg_create_mmr_inst("shub2_01", "SH2_MMRBIST_CTL");
	}


	###################################################################################
	# BIST READ_DIR_PAIR
	###################################################################################
	#Set up the bist control mmr data
	$cmd = 0x50;            # READ_DIR_PAIR = 0x50
	$block_length = 0x0;    # READ_DIR_PAIR command ignores the block_length

	$X = 1;
	$Y = 0;

	$LEFT  = 0;
	$RIGHT = 1;

	$DIRA = 0;
	$DIRB = 1;
	$k = 0;
	$ret_val = 0;
	$mmr_data = 0;
	$mmr_name = "NONAME";
	$dir  = 0;
	$dira = 0;
	$dirb = 0;
	$ecca = 0;
	$eccb = 0;
	my $cor  = 0;
	my $unc  = 0;
	my $pri  = 0;
	my $acc  = 0;
   
	my $addr = ($shub_version == 2) ? ($starting_addr & 0x3fffffffff) : ($starting_addr & 0xfffffffff);

	if ($verbose > 0)  {
	  printf "\nAddress = 0x%010llX ****************************\n", $addr;
	}

	if ((start_bist_engine($cmd, $block_length, $addr, $shub_index)) != 1)  {
	  print "\nError in get_dir_state().  Unable to successfully start the bist engine.\n";
	  return(-1);
	}

	# For directory: A7 = 0 -> Y
	#                A7 = 1 -> X
	#
	#                A11 = 0 -> L
	#                A11 = 1 -> R
	#
	#                A8 = 0 -> DIRA
	#                A8 = 1 -> DIRB

	$bit_7  = ($addr >> 7)  & 0x1;
	$bit_8  = ($addr >> 8)  & 0x1;
	$bit_11 = ($addr >> 11) & 0x1;

	my $xfr_complete = 0;
  	my $xfr_f_cnt = 0;

	if ($bit_7 == $Y) {
	  if ($bit_11 == $LEFT) {
	    $mmr_name = "DQLP_YDIR_ENTRY";
	    while (($xfr_complete == 0) && ($xfr_f_cnt < 3) && ($target_list[$shub_index]->{XFR_ERR_CNT} < 6))  {
                $SSCOPE_ERR = 0;
	        $mmr_data = reg_mmr_read($lp_ydir_entry);
    		if ($SSCOPE_ERR == 1)  {
      		    $xfr_f_cnt++;
      		    $target_list[$shub_index]->{XFR_ERR_CNT}++;
      		    $SSCOPE_ERR = 0;
    	  	}
    		else {
      		    $xfr_complete = 1;
    		}
  	    }
	    $dira     = reg_get_field_value($lp_ydir_entry, "DIRA");
	    $dirb     = reg_get_field_value($lp_ydir_entry, "DIRB");
	    $pri      = reg_get_field_value($lp_ydir_entry, "PRI");
	    $acc      = reg_get_field_value($lp_ydir_entry, "ACC");
	    $cor      = reg_get_field_value($lp_ydir_entry, "COR");
	    $unc      = reg_get_field_value($lp_ydir_entry, "UNC");
	  } elsif ($bit_11 == $RIGHT) {
	    $mmr_name = "DQRP_YDIR_ENTRY";
            while (($xfr_complete == 0) && ($xfr_f_cnt < 3) && ($target_list[$shub_index]->{XFR_ERR_CNT} < 6))  {
                $SSCOPE_ERR = 0;
	        $mmr_data = reg_mmr_read($rp_ydir_entry);
                if ($SSCOPE_ERR == 1)  {
                    $xfr_f_cnt++;
                    $target_list[$shub_index]->{XFR_ERR_CNT}++;
                    $SSCOPE_ERR = 0;
                }
                else {
                    $xfr_complete = 1;
                }
            }
	    $dira     = reg_get_field_value($rp_ydir_entry, "DIRA");
	    $dirb     = reg_get_field_value($rp_ydir_entry, "DIRB");
	    $pri      = reg_get_field_value($rp_ydir_entry, "PRI");
	    $acc      = reg_get_field_value($rp_ydir_entry, "ACC");
	    $cor      = reg_get_field_value($rp_ydir_entry, "COR");
	    $unc      = reg_get_field_value($rp_ydir_entry, "UNC");
	  }

	} elsif ($bit_7 == $X) {
	  if ($bit_11 == $LEFT) {      # L Dimm
	    $mmr_name = "DQLP_XDIR_ENTRY";
            while (($xfr_complete == 0) && ($xfr_f_cnt < 3) && ($target_list[$shub_index]->{XFR_ERR_CNT} < 6))  {
                $SSCOPE_ERR = 0;
	        $mmr_data = reg_mmr_read($lp_xdir_entry);
                if ($SSCOPE_ERR == 1)  {
                    $xfr_f_cnt++;
                    $target_list[$shub_index]->{XFR_ERR_CNT}++;
                    $SSCOPE_ERR = 0;
                }
                else {
                    $xfr_complete = 1;
                }
            }
	    $dira     = reg_get_field_value($lp_xdir_entry, "DIRA");
	    $dirb     = reg_get_field_value($lp_xdir_entry, "DIRB");
	    $pri      = reg_get_field_value($lp_xdir_entry, "PRI");
	    $acc      = reg_get_field_value($lp_xdir_entry, "ACC");
	    $cor      = reg_get_field_value($lp_xdir_entry, "COR");
	    $unc      = reg_get_field_value($lp_xdir_entry, "UNC");
	  } 
	  elsif ($bit_11 == $RIGHT) { # R Dimm
	    $mmr_name = "DQRP_XDIR_ENTRY";
            while (($xfr_complete == 0) && ($xfr_f_cnt < 3) && ($target_list[$shub_index]->{XFR_ERR_CNT} < 6))  {
                $SSCOPE_ERR = 0;
	        $mmr_data = reg_mmr_read($rp_xdir_entry);
		if ($SSCOPE_ERR == 1)  {
                    $xfr_f_cnt++;
                    $target_list[$shub_index]->{XFR_ERR_CNT}++;
                    $SSCOPE_ERR = 0;
                }
                else {
                    $xfr_complete = 1;
                }
            }
	    $dira     = reg_get_field_value($rp_xdir_entry, "DIRA");
	    $dirb     = reg_get_field_value($rp_xdir_entry, "DIRB");
	    $pri      = reg_get_field_value($rp_xdir_entry, "PRI");
	    $acc      = reg_get_field_value($rp_xdir_entry, "ACC");
	    $cor      = reg_get_field_value($rp_xdir_entry, "COR");
	    $unc      = reg_get_field_value($rp_xdir_entry, "UNC");
	  }
	}

	if($bit_8 == $DIRA) {
	  $dir = $dira;
	} 
	elsif($bit_8 == $DIRB) {
	  $dir = $dirb;
	}

	my $dir_value = decode_dir_state($dir, $cor, $unc, $acc, $pri);
	printf "addr: 0x%010llX %s", $addr, $dir_value;

    # Eliminate this extra step, as we weren't acting on the information
    # generated anyway.
	
	# if(read_bist_dir_mmrs(0, $shub_index)) {     
	  # $fail_check++;
	# }

}
################################################################################
#End of BIST testing.
################################################################################

########################################################################
# Starts the bist engine with a given command, d-word block length, and address.
# Usage: start_bist_engine(cmd, block_length, addr);
########################################################################
sub start_bist_engine
{
  my $cmd          = $_[0];
  my $block_length = $_[1];
  my $base_addr    = $_[2];
  my $shub_index   = $_[3];

  #cmd = FILL_DIR       = 0x20
  #cmd = FILL_DATA      = 0x30
  #cmd = DIMM_BIST_READ = 0x10
  #cmd = READ_DIR_PAIR  = 0x50

  # Set up and write the bist base address mmr
  reg_set_value($mbist_base, $base_addr);

  my $wr_xfr_complete = 0;
  my $xfr_f_cnt = 0;
  while (($wr_xfr_complete == 0) && ($xfr_f_cnt < 3) && ($target_list[$shub_index]->{XFR_ERR_CNT} < 6))  {
    $SSCOPE_ERR = 0;
    reg_mmr_write($mbist_base);
    if ($SSCOPE_ERR == 1)  {
      $xfr_f_cnt++;
      $target_list[$shub_index]->{XFR_ERR_CNT}++;
      $SSCOPE_ERR = 0;
    }
    else {
      $wr_xfr_complete = 1;
    }
  }

  #Set up and write the bist control mmr
  # NOTE: The bist engine ignores the bottom 6 bits of the block_length field
  # for a directory fill.
  reg_set_field_value($mbist_ctl, "CMD", $cmd);
  reg_set_field_value($mbist_ctl, "BLOCK_LENGTH", $block_length);
  reg_set_field_value($mbist_ctl, "IN_PROGRESS", 1);

  my $ctl_xfr_complete = 0;
  $xfr_f_cnt = 0;
  while (($ctl_xfr_complete == 0) && ($xfr_f_cnt < 3) && ($target_list[$shub_index]->{XFR_ERR_CNT} < 6))  {
    $SSCOPE_ERR = 0;
    reg_mmr_write($mbist_ctl);
    if ($SSCOPE_ERR == 1)  {
      $xfr_f_cnt++;
      $target_list[$shub_index]->{XFR_ERR_CNT}++;
      $SSCOPE_ERR = 0;
    }
    else {
      $ctl_xfr_complete = 1;
    }
  }
  return($wr_xfr_complete & $ctl_xfr_complete);

}

########################################################################
# Reads the Directory Entry MMRs from the READ_DIR_PAIR bist command
# Usage: $x = read_bist_dir_mmrs(read_ecc);
# NOTE: if x > 0, errors occured
########################################################################
sub read_bist_dir_mmrs
{
  my $read_ecc    = $_[0];
  my $shub_index  = $_[1];

  my $k = 0;
  my $ret_val = 0;
  my $mmr_data = 0;
  my $mmr_name = "NONAME";
  my $dira = 0;
  my $dirb = 0;
  my $ecca = 0;
  my $eccb = 0;
  my $pri  = 0;
  my $acc  = 0;
  my $cor  = 0;
  my $unc  = 0;

  for($k = 0; $k < 4; $k++) {
    if($k == 0) {
      my $xfr_complete = 0;
      my $xfr_f_cnt = 0;
      while (($xfr_complete == 0) && ($xfr_f_cnt < 3) && ($target_list[$shub_index]->{XFR_ERR_CNT} < 6))  {
          $SSCOPE_ERR = 0;
          $mmr_data = reg_mmr_read($lp_xdir_entry);
          if ($SSCOPE_ERR == 1)  {
              $xfr_f_cnt++;
              $target_list[$shub_index]->{XFR_ERR_CNT}++;
              $SSCOPE_ERR = 0;
          }
          else {
              $xfr_complete = 1;
          }
      }
      $mmr_name = "DQLP_XDIR_ENTRY";
      $dira = reg_get_field_value($lp_xdir_entry, "DIRA");
      $dirb = reg_get_field_value($lp_xdir_entry, "DIRB");
      $pri  = reg_get_field_value($lp_xdir_entry, "PRI");
      $acc  = reg_get_field_value($lp_xdir_entry, "ACC");
      $cor  = reg_get_field_value($lp_xdir_entry, "COR");
      $unc  = reg_get_field_value($lp_xdir_entry, "UNC");
    } elsif($k == 1) {
      my $xfr_complete = 0;
      my $xfr_f_cnt = 0;
      while (($xfr_complete == 0) && ($xfr_f_cnt < 3) && ($target_list[$shub_index]->{XFR_ERR_CNT} < 6))  {
          $SSCOPE_ERR = 0;
          $mmr_data = reg_mmr_read($rp_xdir_entry);
          if ($SSCOPE_ERR == 1)  {
              $xfr_f_cnt++;
              $target_list[$shub_index]->{XFR_ERR_CNT}++;
              $SSCOPE_ERR = 0;
          }
          else {
              $xfr_complete = 1;
          }
      }
      $mmr_name = "DQRP_XDIR_ENTRY";
      $dira = reg_get_field_value($rp_xdir_entry, "DIRA");
      $dirb = reg_get_field_value($rp_xdir_entry, "DIRB");
      $pri  = reg_get_field_value($rp_xdir_entry, "PRI");
      $acc  = reg_get_field_value($rp_xdir_entry, "ACC");
      $cor  = reg_get_field_value($rp_xdir_entry, "COR");
      $unc  = reg_get_field_value($rp_xdir_entry, "UNC");
    } elsif($k == 2) {
      my $xfr_complete = 0;
      my $xfr_f_cnt = 0;
      while (($xfr_complete == 0) && ($xfr_f_cnt < 3) && ($target_list[$shub_index]->{XFR_ERR_CNT} < 6))  {
          $SSCOPE_ERR = 0;
          $mmr_data = reg_mmr_read($lp_ydir_entry);
          if ($SSCOPE_ERR == 1)  {
              $xfr_f_cnt++;
              $target_list[$shub_index]->{XFR_ERR_CNT}++;
              $SSCOPE_ERR = 0;
          }
          else {
              $xfr_complete = 1;
          }
      }
      $mmr_name = "DQLP_YDIR_ENTRY";
      $dira = reg_get_field_value($lp_ydir_entry, "DIRA");
      $dirb = reg_get_field_value($lp_ydir_entry, "DIRB");
      $pri  = reg_get_field_value($lp_ydir_entry, "PRI");
      $acc  = reg_get_field_value($lp_ydir_entry, "ACC");
      $cor  = reg_get_field_value($lp_ydir_entry, "COR");
      $unc  = reg_get_field_value($lp_ydir_entry, "UNC");
    } elsif($k == 3) {
      my $xfr_complete = 0;
      my $xfr_f_cnt = 0;
      while (($xfr_complete == 0) && ($xfr_f_cnt < 3) && ($target_list[$shub_index]->{XFR_ERR_CNT} < 6))  {
          $SSCOPE_ERR = 0;
          $mmr_data = reg_mmr_read($rp_ydir_entry);
          if ($SSCOPE_ERR == 1)  {
              $xfr_f_cnt++;
              $target_list[$shub_index]->{XFR_ERR_CNT}++;
              $SSCOPE_ERR = 0;
          }
          else {
              $xfr_complete = 1;
          }
      }
      $mmr_name = "DQRP_YDIR_ENTRY";
      $dira = reg_get_field_value($rp_ydir_entry, "DIRA");
      $dirb = reg_get_field_value($rp_ydir_entry, "DIRB");
      $pri  = reg_get_field_value($rp_ydir_entry, "PRI");
      $acc  = reg_get_field_value($rp_ydir_entry, "ACC");
      $cor  = reg_get_field_value($rp_ydir_entry, "COR");
      $unc  = reg_get_field_value($rp_ydir_entry, "UNC");
    }

    if($verbose) {
      printf"\t$mmr_name = ";
      printf"DIRA = 0x%07llX\n", $dira;
      printf"\t\t\t  DIRB = 0x%07llX\n", $dirb;
      printf"\t\t\t  PRI  = 0x%01llX\n", $pri;
      printf"\t\t\t  ACC  = 0x%01llX\n", $acc;
      printf"\t\t\t  COR  = 0x$cor\n";
      printf"\t\t\t  UNC  = 0x$unc\n";
    } else {
      #printf"\t$mmr_name = ";
      #printf"0x%016llX\n", $mmr_data;
    }

    #if($cor == 1) {
    #  print"\t\t\t  ERROR: Correctable Error in Directory Entry Data!!!  ($mmr_name)\n";
    #  $ret_val++;
    #}
    #if($unc == 1) {
    #  print"\t\t\t  ERROR: Uncorrectable Error in Directory Entry Data!!!  ($mmr_name)\n";
    #  $ret_val++;
    #}
  }

  if($read_ecc) {
    my $k = 0;
    for($k = 0; $k< 4; $k++) {
      my $xfr_complete = 0;
      my $xfr_f_cnt = 0;
      if($k == 0) {
         while (($xfr_complete == 0) && ($xfr_f_cnt < 3) && ($target_list[$shub_index]->{XFR_ERR_CNT} < 6))  {
            $SSCOPE_ERR = 0;
            $mmr_data = reg_mmr_read($lp_xdir_ecc);
            if ($SSCOPE_ERR == 1)  {
               $xfr_f_cnt++;
               $target_list[$shub_index]->{XFR_ERR_CNT}++;
               $SSCOPE_ERR = 0;
            }
            else {
               $xfr_complete = 1;
            }
        }
        $mmr_name = "DQLP_XDIR_ECC";
        $ecca = reg_get_field_value($lp_xdir_ecc, "ECCA");
        $eccb = reg_get_field_value($lp_xdir_ecc, "ECCB");
      } elsif($k == 1) {
        while (($xfr_complete == 0) && ($xfr_f_cnt < 3) && ($target_list[$shub_index]->{XFR_ERR_CNT} < 6))  {
            $SSCOPE_ERR = 0;
            $mmr_data = reg_mmr_read($rp_xdir_ecc);
            if ($SSCOPE_ERR == 1)  {
               $xfr_f_cnt++;
               $target_list[$shub_index]->{XFR_ERR_CNT}++;
               $SSCOPE_ERR = 0;
            }
            else {
               $xfr_complete = 1;
            }
        }
        $mmr_name = "DQRP_XDIR_ECC";
        $ecca = reg_get_field_value($rp_xdir_ecc, "ECCA");
        $eccb = reg_get_field_value($rp_xdir_ecc, "ECCB");
      } elsif($k == 2) {
        while (($xfr_complete == 0) && ($xfr_f_cnt < 3) && ($target_list[$shub_index]->{XFR_ERR_CNT} < 6))  {
            $SSCOPE_ERR = 0;
            $mmr_data = reg_mmr_read($lp_ydir_ecc);
            if ($SSCOPE_ERR == 1)  {
               $xfr_f_cnt++;
               $target_list[$shub_index]->{XFR_ERR_CNT}++;
               $SSCOPE_ERR = 0;
            }
            else {
               $xfr_complete = 1;
            }
        }
        $mmr_name = "DQLP_YDIR_ECC";
        $ecca = reg_get_field_value($lp_ydir_ecc, "ECCA");
        $eccb = reg_get_field_value($lp_ydir_ecc, "ECCB");
      } elsif($k == 3) {
        while (($xfr_complete == 0) && ($xfr_f_cnt < 3) && ($target_list[$shub_index]->{XFR_ERR_CNT} < 6))  {
            $SSCOPE_ERR = 0;
            $mmr_data = reg_mmr_read($rp_ydir_entry);
            if ($SSCOPE_ERR == 1)  {
               $xfr_f_cnt++;
               $target_list[$shub_index]->{XFR_ERR_CNT}++;
               $SSCOPE_ERR = 0;
            }
            else {
               $xfr_complete = 1;
            }
        }
        $mmr_name = "DQRP_YDIR_ECC";
        $ecca = reg_get_field_value($rp_ydir_ecc, "ECCA");
        $eccb = reg_get_field_value($rp_ydir_ecc, "ECCB");
      }

      printf"\t$mmr_name =\t";
      printf"ECCA = 0x%07llX\n", $ecca;
      printf"\t\tECCB = 0x%07llX\n", $eccb;
    }
  }


  return $ret_val;
}



#--------------------------------------------------------------------------------
#
#  subroutine:  get_mem_cfg
#
#    input arguments:  The 64 bit value read from the X|Y|JNR_DIMM_CFG register
#  
#    return value: 8 bit value which indicates the banks & groups:
#	[3:0]	2 bank value [0 or 1]
#	[7:4]	group value [1,2,3, or 4]
#
#    rules:  
#
#	If any of the groups have only 1 bank, the 2 bank value is cleared for all.
#
#--------------------------------------------------------------------------------

sub get_mem_cfg {
  
  my $cfg_value;
  my $cfg_mmr_value; 
  my $cfg_mmr_value = $_[0];
  my $dimm_cfg_mmr_val = $_[0];
  my $grp_val = 0;
  my $two_bank_val = 1;
  my $cfg_loop;
  my $cfg_string = "";
  my $size_val;

  for ( $cfg_loop = 0; $cfg_loop < 4; $cfg_loop++ ) {
    if ($cfg_loop > 0)  {
      $cfg_string .= "   ";
    }
    if (($cfg_mmr_value & 7) == 7 ) {
      $cfg_string .= "   --     --  ";
    }
    else {
      $two_bank_val &= ($cfg_mmr_value & 8) >> 3;
      if (($cfg_mmr_value & 7) == 0) 	{  $size_val = "  64 MB";  }
      elsif (($cfg_mmr_value & 7) == 1) {  $size_val = " 128 MB";  }
      elsif (($cfg_mmr_value & 7) == 2) {  $size_val = " 256 MB";  }
      elsif (($cfg_mmr_value & 7) == 3) {  $size_val = " 512 MB";  }
      elsif (($cfg_mmr_value & 7) == 4) {  $size_val = "1024 MB";  }
      elsif (($cfg_mmr_value & 7) == 5) {  $size_val = "2048 MB";  }
      elsif (($cfg_mmr_value & 7) == 6) {  $size_val = "UNDEFINED";  }

      $cfg_string .= sprintf ("%-9s  %d  ",  $size_val, ($two_bank_val + 1) );
      $grp_val++;
    }
    $cfg_mmr_value = $cfg_mmr_value >> 8;
  }

  $target_list[$t_cnt]->{MEM_CFG} = "$cfg_string";

  $dimm_speed = $cfg_mmr_value & 0xf;
  if ($dimm_speed == 0) { $target_list[$t_cnt]->{DIMM_SPD} = "100 MHz  2.0"; }
  elsif ($dimm_speed == 1) { $target_list[$t_cnt]->{DIMM_SPD} = "133 MHz  2.0"; }
  elsif ($dimm_speed == 2) { $target_list[$t_cnt]->{DIMM_SPD} = "133 MHz  2.5"; }
  elsif ($dimm_speed == 3) { $target_list[$t_cnt]->{DIMM_SPD} = "160 MHz  2.0"; }
  elsif ($dimm_speed == 4) { $target_list[$t_cnt]->{DIMM_SPD} = "160 MHz  2.5"; }
  elsif ($dimm_speed == 5) { $target_list[$t_cnt]->{DIMM_SPD} = "160 MHz  3.0"; }
  elsif ($dimm_speed == 6) { $target_list[$t_cnt]->{DIMM_SPD} = "200 MHz  2.0"; }
  elsif ($dimm_speed == 7) { $target_list[$t_cnt]->{DIMM_SPD} = "200 MHz  2.5"; }
  elsif ($dimm_speed == 8) { $target_list[$t_cnt]->{DIMM_SPD} = "200 MHz  3.0"; }
  else { $target_list[$t_cnt]->{DIMM_SPD} = "??? MHz  ???"; }

  $cfg_value = $two_bank_val | ($grp_val << 4);

  my $cfg_reg = reg_create_mmr_inst("shub_01", "SH_X_DIMM_CFG");
  reg_set_value($cfg_reg, $dimm_cfg_mmr_val);
  $dimm_group_0_cfg = reg_get_field_value($cfg_reg, 'DIMM0_CS');  # Logical value for each physical DIMM0s
  $dimm_group_1_cfg = reg_get_field_value($cfg_reg, 'DIMM1_CS');  # Logical value for each physical DIMM1s
  $dimm_group_2_cfg = reg_get_field_value($cfg_reg, 'DIMM2_CS');  # Logical value for each physical DIMM2s
  $dimm_group_3_cfg = reg_get_field_value($cfg_reg, 'DIMM3_CS');  # Logical value for each physical DIMM3s

  return $cfg_value;
}

###########################################################################


#--------------------------------------------------------------------------------
#
#  subroutine:  get_sh2_mem_cfg
#
#    input arguments:  The 64 bit value read from the X|Y|JNR_DIMM_CFG register
#  
#    return value: 8 bit value which indicates the banks & groups:
#	[3:0]	2 bank value [0 or 1]
#	[7:4]	group value [1,2,3, or 4]
#
#    rules:  
#
#	If any of the groups have only 1 bank, the 2 bank value is cleared for all.
#
#--------------------------------------------------------------------------------

sub get_sh2_mem_cfg {
  
  my $cfg_mmr_value = $_[0]; 
  my $dimm_cfg_mmr_val = $_[0]; 
  my $grp_val = 0;
  my $two_bank_val = 1;
  my $cfg_loop;
  my $cfg_string = "";
  my $size_val;
  my $dimm_cs = 0;
  my $dimm_rv = 0;

  for ( $cfg_loop = 0; $cfg_loop < 3; $cfg_loop++ ) {
    if ($cfg_loop > 0)  {
      $cfg_string .= "   ";
    }
    if (($cfg_mmr_value & 7) == 7 ) {
      $cfg_string .= "   --     -- -- --  ";
    }
    else {
      if (($cfg_mmr_value & 7) == 0) 	{  $size_val = " 512 MB";  }
      elsif (($cfg_mmr_value & 7) == 1) {  $size_val = "1024 MB";  }
      elsif (($cfg_mmr_value & 7) == 2) {  $size_val = "2048 MB";  }
      elsif (($cfg_mmr_value & 7) == 3) {  $size_val = "4096 MB";  }
      elsif (($cfg_mmr_value & 7) == 4) {  $size_val = "8192 MB";  }
      elsif (($cfg_mmr_value & 7) == 5) {  $size_val = "  16 GB";  }
      elsif (($cfg_mmr_value & 7) == 6) {  $size_val = "  32 GB";  }
      
      $two_bank_val &= ($cfg_mmr_value >> 3) & 1 ;
      $dimm_rv = ($cfg_mmr_value >> 4) & 1;   
      $dimm_cs = ($cfg_mmr_value >> 5) & 3;   
      $cfg_string .= sprintf "%-9s  %d  %d  %d  ",  $size_val, ($two_bank_val + 1), $dimm_rv, $dimm_cs;
      $grp_val++;
    }
    $cfg_mmr_value = $cfg_mmr_value >> 8;
  }

  $sh2_target_list[$sh2_cnt]->{MEM_CFG} = "$cfg_string";

  my $dimm_speed = $cfg_mmr_value & 0xf;
  if ($dimm_speed == 0) { $sh2_target_list[$sh2_cnt]->{DIMM_SPD} = "100 MHz  2.0"; }
  elsif ($dimm_speed == 1) { $sh2_target_list[$sh2_cnt]->{DIMM_SPD} = "133 MHz  2.0"; }
  elsif ($dimm_speed == 2) { $sh2_target_list[$sh2_cnt]->{DIMM_SPD} = "133 MHz  2.5"; }
  elsif ($dimm_speed == 3) { $sh2_target_list[$sh2_cnt]->{DIMM_SPD} = "160 MHz  2.0"; }
  elsif ($dimm_speed == 4) { $sh2_target_list[$sh2_cnt]->{DIMM_SPD} = "160 MHz  2.5"; }
  elsif ($dimm_speed == 5) { $sh2_target_list[$sh2_cnt]->{DIMM_SPD} = "160 MHz  3.0"; }
  elsif ($dimm_speed == 6) { $sh2_target_list[$sh2_cnt]->{DIMM_SPD} = "200 MHz  2.0"; }
  elsif ($dimm_speed == 7) { $sh2_target_list[$sh2_cnt]->{DIMM_SPD} = "200 MHz  2.5"; }
  elsif ($dimm_speed == 8) { $sh2_target_list[$sh2_cnt]->{DIMM_SPD} = "200 MHz  3.0"; }
  else { $sh2_target_list[$sh2_cnt]->{DIMM_SPD} = "??? MHz  ???"; }

  my $cfg_reg = reg_create_mmr_inst("shub2_01", "SH2_X_DIMM_CFG");
  reg_set_value($cfg_reg, $dimm_cfg_mmr_val);
  $dimm_group_0_cfg = reg_get_field_value($cfg_reg, 'DIMM0_CS');   # Logical value for each physical DIMM0s
  $dimm_group_1_cfg = reg_get_field_value($cfg_reg, 'DIMM1_CS');   # Logical value for each physical DIMM1s
  $dimm_group_2_cfg = reg_get_field_value($cfg_reg, 'DIMM2_CS');   # Logical value for each physical DIMM2s
  $dimm_group_3_cfg = reg_get_field_value($cfg_reg, 'DIMM3_CS');   # Logical value for each physical DIMM3s

}

###########################################################################


################################################################################
#
#  Sort the @address_list array, using {ADDRESS} as the sort element to key on,
#  and create the @final_addr_list using the sorted output.
#
#
################################################################################

sub sort_by_addr {

  my $sorted_addr_cnt = 0;              # count of ordered addresses
  my $low_index = 0;                    # array index having the current low address value
  my $low_addr = 0;                     # low address value for the inner for loop
  my $first_one = 1;                    # indicates this is the first unused entry
  my $addr_loop = 0;
  @final_addr_list = ();		# Initialize the array

  while ($sorted_addr_cnt < $addr_cnt )  {
    for ($addr_loop = 0; $addr_loop < $addr_cnt; $addr_loop++)  {

        # Always save the first unused entry

      if ((! defined $address_list[$addr_loop]->{USED}) && ($first_one == 1))  {
        $low_addr = $address_list[$addr_loop]->{ADDRESS};
        $low_index = $addr_loop;
        $first_one = 0;
        next;
      }
        # Save the entry if it's less than the previous low time value

      if ((! defined $address_list[$addr_loop]->{USED}) && ($address_list[$addr_loop]->{ADDRESS} < $low_addr)) {
        $low_addr = $address_list[$addr_loop]->{ADDRESS};
        $low_index = $addr_loop;
        next;
      }
    }           #  end of for loop  - a new low address value should be available

    $address_list[$low_index]->{USED} = 1;
    $final_addr_list[$sorted_addr_cnt]->{NASID} = $address_list[$low_index]->{NASID};
    $final_addr_list[$sorted_addr_cnt]->{ADDRESS} = $address_list[$low_index]->{ADDRESS};
    $final_addr_list[$sorted_addr_cnt]->{COUNT} = $address_list[$low_index]->{COUNT};
    $sorted_addr_cnt++;
    $first_one = 1;
  }

  for ($aa = 0; $aa < $addr_cnt; $aa++)  {
    if (! defined $address_list[$aa]->{USED})  {
      printf "\nALERT!  In sort_by_addr routine - all addresses not allocated into final list!\n";
      printf "Address:  0x%016llX     NASID:  0x%04X\n", $address_list[$aa]->{ADDRESS}, $address_list[$low_index]->{NASID};
    }
  }
}

#-------------------------------------------------------------------------------
#
#  save_address subroutine
#
#  Arguments passed:    $_[0]   prefix string containing NASID
#                       $_[1]   address to be checked & added
#                       $_[2]   nasid home of directory address to be checked & added
#                       $_[3]   how this address came to be flagged
#
#-------------------------------------------------------------------------------
sub save_address {

  my $num_args		= $#_ + 1; 
  my $new_addr          = 1;
  my $aa                = 0;
  my $nasid_prefix      = $_[0];
  my $addr_val          = $_[1];
  my $nasid_val         = $_[2];
  my $source_type       = $_[3];

  my $not_booted	= ($num_args == 5) ? $_[4] : 0;
    
  if (($not_booted == 0) || ($nasid_val != 0))  {
    for($aa=0; $aa<$addr_cnt; $aa++) {
      if( ($addr_val==$address_list[$aa]->{ADDRESS}) && ($nasid_val==$address_list[$aa]->{NASID})) {
        $new_addr = 0;
        printf "\n   %s: NASID 0x%04X address 0x%010llX (%s) is already in the list...\n", $nasid_prefix, $nasid_val, $addr_val, $source_type;
        $address_list[$aa]->{COUNT} += 1;
        last;
      }
    }
    if($new_addr) {
      $address_list[$addr_cnt]->{NASID}   = $nasid_val;
      $address_list[$addr_cnt]->{ADDRESS} = $addr_val;
      $address_list[$addr_cnt]->{COUNT} = 1;
      printf "\n%s: added NASID 0x%04X  Address: 0x%010llX to the list (%s).\n", $nasid_prefix, $nasid_val, $addr_val, $source_type;
      $addr_cnt++;
    }
  }
  else {
    $nb_address_list[$nb_addr_cnt]->{NASID}   = $nasid_val;
    $nb_address_list[$nb_addr_cnt]->{ADDRESS} = $addr_val;
    $nb_address_list[$nb_addr_cnt]->{MODULE}  = $sh2_target_list[$sh2_cnt]->{MODULE};	   # these are set in read_and_parse_shub2x()
    $nb_address_list[$nb_addr_cnt]->{NODE}    = $sh2_target_list[$sh2_cnt]->{NODE};	   #  "  "
    $nb_address_list[$nb_addr_cnt]->{CHIP}    = $sh2_target_list[$sh2_cnt]->{CHIP};        #  "  "
    printf "\n%s: added Non-booted NASID 0x%04X  Address: 0x%010llX to the list (%s).\n", $nasid_prefix, $nasid_val, $addr_val, $source_type;
    $nb_addr_cnt++;
  }

}


#-------------------------------------------------------------------------------
#
#  update_md_err_mask subroutine
#
#  Arguments passed:    $_[0]   value to write to SH2_MEM_ERROR_MASK
#
#
#-------------------------------------------------------------------------------

sub update_md_err_mask  {

  my $mmr_write_data = $_[0];
  
  $mask_reg = reg_create_mmr_inst("shub2_01", "SH2_MEM_ERROR_MASK");
  reg_set_value($mask_reg, $mmr_write_data);

  my $xfr_complete = 0;
  my $xfr_f_cnt = 0;
  while (($xfr_complete == 0) && ($xfr_f_cnt < 3) && ($sh2_target_list[$shub_index]->{XFR_ERR_CNT} < 6))  {
    $SSCOPE_ERR = 0;
    reg_mmr_write($mask_reg);
    if ($SSCOPE_ERR == 1)  {
      $xfr_f_cnt++;
      $sh2_target_list[$shub_index]->{XFR_ERR_CNT}++;
      $SSCOPE_ERR = 0;
      print "\n Error detected after reg_mmr_write() attempt in update_md_err_mask()\n";
    }
    else {
      $xfr_complete = 1;
    }
  }
}


#------------------------------------------------------------------------------
#
#  The following function decodes and returns a string containing the 
#  decoded directory value.
#
#  directory state defines found in //sn2/chips/shub2/md/dq/v/sh2_md_dq_dir_proto.v
#
#------------------------------------------------------------------------------

sub decode_dir_state {

  	my $dir = $_[0];
  	my $cor = $_[1];
  	my $unc = $_[2];
	my $access = 'Unknown';
	my $access_hint = "\n";
	if (defined $_[3]) {
	  $access = sprintf "%d", $_[3];
	  $access_hint = sprintf "\n                     \(Access value of %s corresponds to LOCVEC%s and PRIVEC%s\)\n     ",
		 $access, $access, $access;
  	}

	my $priority = 'Unknown';
	if (defined $_[4]) {
	  $priority = sprintf "%d", $_[4];
	}


	my $ret_val = "";

  	my $BUSYI = 0x60;
  	my $BUSY  = 0x61;
  	my $BSYIW = 0x62;
  	my $BSYEI = 0x63;
  	my $BSYUW = 0x64;
  	my $BSYUR = 0x65;

  	my $BSYF  = 0x66;
  	my $BSYFN = 0x67;
  	my $BSYFY = 0x68;
  	my $BSYG  = 0x69;
  	my $BSYGN = 0x6a;
  	my $BSYGY = 0x6b;

  	my $BSYX  = 0x6c;
  	my $BSYXN = 0x6d;
  	my $BSYXY = 0x6e;
  	my $BSYN  = 0x6f;
  	my $BSYNN = 0x70;
  	my $BSYNY = 0x71;

  	my $BSYEN = 0x72;
  	my $BSYEY = 0x73;
  	my $BSYSN = 0x74;
  	my $BSYSY = 0x75;
  	my $BSYUN = 0x76;
  	my $BSYUY = 0x77;
  	my $BSYE  = 0x78;
  	my $BSYUP = 0x79;
  	my $POISS = 0x7c;
  	my $POISW = 0x7d;
  	my $POIS  = 0x7e;

	my $dir_state = ($dir >> 24) & 0x3;

	my $final_state = "????";

	if(($dir_state == 1) || ($dir_state == 2) || ($dir_state == 3)) {
	  # print"Directory State is SHRD\n";
	  if ($dir_state == 1)  {  $final_state = "SHRD-A ";  }
	  if ($dir_state == 2)  {  $final_state = "SHRD-B ";  }
	  if ($dir_state == 3)  {  $final_state = "SHRD-AB";  }
	
	  my $shr16_l = ($dir >> 0)  & 0xff;
	  my $shr16_u = ($dir >> 12) & 0xff;
	  my $shr8_l  = ($dir >> 8)  & 0xf;
	  my $shr8_u  = ($dir >> 20) & 0xf;

	  my $shr8_final  = $shr8_l | ($shr8_u << 4);
	  my $shr16_final = $shr16_l | ($shr16_u << 8);

	  $ret_val = sprintf "  State: %s    Shr16 vector: 0x%04X    Shr8 vector: 0x%02X   Access:%s  Priority:%s  Unc:%d  Cor:%d  %s ", 
		$final_state, $shr16_final, $shr8_final, $access, $priority, $unc, $cor, $access_hint;

	} 

	elsif ($dir_state == 0) {
	  my $dir_sxro     = ($dir >> 23) & 0x1;
	  my $dir_hint     = ($dir >> 21) & 0x3;
	  my $dir_substate = ($dir >> 14) & 0x7f;
	  my $dir_chiplet  = ($shub_version == 1) ? ($dir >> 11) & 0x7 : 2;		# default to PI for shub2
	  my $dir_node     = ($shub_version == 1) ? $dir & 0x7ff : $dir & 0x3fff;
	  my $dir_substate_pois  = 0;
	  my $dir_substate_unown = ($dir_substate >> 6) & 0x1;
	  my $dir_substate_excl  = ($dir_substate >> 5) & 0x3;

	  if ($shub_version == 1)  {
	    $dir_substate_pois  = ($dir_substate >> 3) & 0xf;
	  }
	  else {
	    $dir_substate_pois  = ($dir_substate >> 2) & 0x1f;
	  }

	  if($dir_sxro == 1) {
	    $final_state = "SXRO";
	  } elsif($dir_sxro == 0) {
	    my $busy_state = "UNDEFINED";
	    if($dir_substate_unown == 0x0) {
		 $busy_state = "UNOWN";
	    } elsif($dir_substate_excl == 0x2) {
		 $busy_state = "EXCL";
	    } elsif(($dir_substate_pois == 0xf) && ($shub_version == 1)) {
		 $busy_state = "POIS";
	    } elsif(($dir_substate_pois == 0x1f) && ($shub_version == 2)) {
		if ($dir_substate == $POISS) {
		  $busy_state = "POISS";
		}
		elsif ($dir_substate == $POISW) {
		  $busy_state = "POISW";
		}
		elsif ($dir_substate == $POIS) {
		  $busy_state = "POIS";
		}
		else {
		  $busy_state = "POIST";
		}
	    } else {
		if($dir_substate == $BUSYI) {
		 $busy_state = "BUSYI";
	       } elsif($dir_substate == $BSYIW) {
		 $busy_state = "BSYIW";
	       } elsif($dir_substate == $BUSY) {
		 $busy_state = "BUSY";
	       } elsif($dir_substate == $BSYEI) {
		 $busy_state = "BSYEI";
	       } elsif($dir_substate == $BSYUW) {
		 $busy_state = "BSYUW";
	       } elsif($dir_substate == $BSYUR) {
		 $busy_state = "BSYUR";
	       } elsif($dir_substate == $BSYF) {
		 $busy_state = "BSYF";
	       } elsif($dir_substate == $BSYFN) {
		 $busy_state = "BSYFN";
	       } elsif($dir_substate == $BSYFY) {
		 $busy_state = "BSYFY";
	       } elsif($dir_substate == $BSYG) {
		 $busy_state = "BSYG";
	       } elsif($dir_substate == $BSYGN) {
		 $busy_state = "BSYGN";
	       } elsif($dir_substate == $BSYGY) {
		 $busy_state = "BSYGY";
	       } elsif($dir_substate == $BSYX) {
		 $busy_state = "BSYX";
	       } elsif($dir_substate == $BSYXN) {
		 $busy_state = "BSYXN";
	       } elsif($dir_substate == $BSYXY) {
		 $busy_state = "BSYXY";
	       } elsif($dir_substate == $BSYN) {
		 $busy_state = "BSYN";
	       } elsif($dir_substate == $BSYNN) {
		 $busy_state = "BSYNN";
	       } elsif($dir_substate == $BSYNY) {
		 $busy_state = "BSYNY";
	       } elsif($dir_substate == $BSYEN) {
		 $busy_state = "BSYEN";
	       } elsif($dir_substate == $BSYEY) {
		 $busy_state = "BSYEY";
	       } elsif($dir_substate == $BSYSN) {
		 $busy_state = "BSYSN";
	       } elsif($dir_substate == $BSYSY) {
		 $busy_state = "BSYSY";
	       } elsif($dir_substate == $BSYUN) {
		 $busy_state = "BSYUN";
	       } elsif($dir_substate == $BSYUY) {
		 $busy_state = "BSYUY";
	       } elsif(($dir_substate == $BSYE) && ($shub_version == 2)) {
		 $busy_state = "BSYE";
	       } elsif(($dir_substate == $BSYUP) && ($shub_version == 2)) {
		 $busy_state = "BSYUP";
	       } 
	    }
	    # printf "Directory State: $busy_state     ";
	    $final_state = "$busy_state";

	    if($dir_hint == 0) {
	      # printf"Hint Field: UNOWN    ";
	      $final_state .= "  Hint Field: UNOWN";
	    } elsif($dir_hint == 1) {
	      # printf"Hint Field: EXCL to Local PI    ";
	      $final_state .= "  Hint Field: EXCL to Local PI";
	    } elsif($dir_hint == 2) {
	      # printf"Hint Field: EXCL to II or SXB    ";
	      $final_state .= "  Hint Field: EXCL to II or TIO";
	    } elsif($dir_hint == 3) {
	      # printf"Hint Field:  All Others   ";
	      $final_state .= "  Hint Field: All Others";
	    }

	  } 
	  else {
	    $final_state = "ERROR: UNDEFINED SXRO VALUE!";
	  }

	  if (($dir_node & 1) == 0)  {			#  Shub node
	    if ($shub_version == 2)  {
	      $chiplet = $shub2_chiplet_list[$dir_chiplet];
	    }
	    else {
	      $chiplet = $sn2net_chiplet_list[$dir_chiplet];
	    }
	  }
	  else  {					# TIO node
	    $chiplet = $tio_chiplet_list[$dir_chiplet];
	  }

	  $ret_val = sprintf "  State: %s  Nasid: 0x%04X  Chiplet: %s   Access:%s  Priority:%s  Unc:%d  Cor:%d  %s ", 
		$final_state, $dir_node, $chiplet, $access, $priority, $unc, $cor, $access_hint;

			#  GEM - add this debug output to determine why fall-through occurred
	  if ($final_state =~ /UNDEFINED/) {	   
	    $ret_val .= sprintf "dir_sxro=0x%X  dir_hint=0x%X  dir_substate=0x%X  dir_substate_pois=0x%X   dir_substate_unown=0x%X   dir_substate_excl=0x%X\n ", $dir_sxro, $dir_hint, $dir_substate, $dir_substate_pois, $dir_substate_unown, $dir_substate_excl;
	  }

	} 
	else {
		$ret_val = sprintf "ERROR: UNDEFINED DIR_STATE VALUE!\n";
	}

	return($ret_val);
}



sub decode_mybit {
  my $raw_mybit = $_[0];
  my $ret_val = "";

  if (($raw_mybit & 1) == 1) {
    $ret_val .= "\nMyBit = SRCII : Source is TIO Node\n";
  }
  elsif ((($raw_mybit >> 1) & 1) == 1) {
    $ret_val .= "\nMyBit = PTRII : Source is Shub II\n";
  }
  elsif ((($raw_mybit >> 2) & 1) == 1) {
    $ret_val .= "\nMyBit = ANYTIM : timeeq or timelast or timespec is valid\n";
  }
  elsif ((($raw_mybit >> 3) & 1) == 1) {
    $ret_val .= "\nMyBit = ONLYME: incoming request directory vector equals current directory value\n";
  }
  elsif ((($raw_mybit >> 4) & 1) == 1) {
    $ret_val .= "\nMyBit = TIMSPEC: directory timer value equals 0x1ff\n";
  }
  elsif ((($raw_mybit >> 5) & 1) == 1) {
    $ret_val .= "\nMyBit = TIMLAST: directory timer value equals previous timer value\n";
  }
  elsif ((($raw_mybit >> 6) & 1) == 1) {
    $ret_val .= "\nMyBit = TIMEQ: Current timer value equals directory timer field\n";
  }
  elsif ((($raw_mybit >> 7) & 1) == 1) {
    $ret_val .= "\nMyBit = PTREQ: Source field equals directory pointer field\n";
  }
  elsif ((($raw_mybit >> 8) & 1) == 1) {
    $ret_val .= "\nMyBit = DFPART: Source and directory pointer are in different partitions\n";
  }

  return($ret_val);

}


1;
