# 
#/**************************************************************************
# *                                                                        *
# *               Copyright (C) 1989-1999 Silicon Graphics, Inc.           *
# *                                                                        *
# *  These coded instructions, statements, and computer programs  contain  *
# *  unpublished  proprietary  information of Silicon Graphics, Inc., and  *
# *  are protected by Federal copyright law.  They  may  not be disclosed  *
# *  to  third  parties  or copied or duplicated in any form, in whole or  *
# *  in part, without the prior written consent of Silicon Graphics, Inc.  *
# *                                                                        *
# **************************************************************************/
# 
# @(#)$Id: mem_err_funcs.pm,v 1.43 2007/03/14 22:53:01 meyer Exp $
#-------------------------------------------------------------------------------
#
#  FILE: mem_err_funcs.pm
#
#  DESCRIPTION:  This file contains PERL subroutines to extract and interpret
#	memory errors logged in the shub.
#
#
#  MODIFICATION HISTORY:
#
#     WHEN      WHO             WHAT
#     --------  --------------  -----------------------------------------
#     01/08/03  Gary Meyer      Original source code submitted for release.
#     02/26/03  Gary Meyer  	Added ability to provide DTR & DIMM locations.
#     04/03/03  Gary Meyer  	Corrected problems decoding DTR location with XN detected UCE's.
#     06/02/03  Gary Meyer      Added notes on the structure of the DoubleWordData array.
#     08/27/03  Gary Meyer      Added dimm pin to single bit error info. Added decode_amo_dimm_info().
#                               Restructured to use Nelson's DIMM table info.  Modified XN SBE info
#                               to indicate the proper bit based on the reporting flit value.
#     09/11/03  Gary Meyer      Add suspect DIMM location to fru "REASON".
#     09/15/03  Gary Meyer      Added specific pointer to get the fru entry correct.
#     11/12/03  Gary Meyer      Reformatted output for Opus brick cases.
#     05/20/04  Gary Meyer      Reformatted MD output for Opus brick cases.
#     06/21/04  Gary Meyer	Updated to display the dimm location in the standard format.
#     08/03/04  Gary Meyer	DIMM location string now considers node locations within Vortix bricks.
#     09/17/04  Gary Meyer	Fixed cut & paste error with $dir_LorR used instead of $data_LorR.
#     12/09/04  Gary Meyer	Now only printing the XN DIMM info indicated by ERR_GROUP in DETAIL_4.
#     03/07/05  Gary Meyer	Added get_mbe_chip_side().  Will now attempt to determine L or R side for MBE errors.
#     03/16/05  Gary Meyer	Fixed problem in get_mbe_chip_side() that would allow for divide by 0.
#     06/07/05  Gary Meyer	Changes for Shub2.
#     06/14/05  Gary Meyer	More changes for Shub2.
#     06/17/05  Gary Meyer	Previous changes wiped out @SYNDROME_MASKS in get_mbe_chip_side().
#     08/30/05  Gary Meyer	Shub2 changes for dimm layout and naming convention.
#     10/24/05  Gary Meyer	Updates for Shub2 DIMM labeling, and IP67 Node naming.
#     02/01/06  Gary Meyer	Found & fixed more typos where $dir_LorR was used instead of $data_LorR.
#				Corrected bug where the Left DIMM side was called out for PI, XN, and MD-AMO
#				reported double bit errors having a syndrome not 100% on either side.
#     02/07/06  Gary Meyer	Updated to recognize IP73 brick/blade type.
#     03/22/06  Gary Meyer	Corrected problems with UCE Left Side and Right Side regular expressions.
#     03/30/06  Gary Meyer	Corrected problem with Shub2 XN ECC DIMM side reporting. ($xn_err_grp wasn't defined)
#     03/30/06  Gary Meyer	Updates to recognize IP71 blade type.
#     04/14/06  Gary Meyer	Updates to recognize IP75 blade type.
#     05/02/06  Gary Meyer	Now printing the NASID value with the address in the detail of a FSB_SHUB ecc error.
#     06/14/06  Gary Meyer      Fix for case where multiple errors logged by the MD cause the suspect dimm location
#                               to be attached to the wrong fru entry.
#     11/06/06  Gary Meyer      Suppress DIMM output and FRU DIMM indicator for XN reported poisoned ECC errors.
#			 	Recognize FSB_SHUB_UCE poisoned syndrome.
#     11/06/06  Gary Meyer      Fixed bug introduced with the last update.  The DIMM_SUMM wasn't being updated correctly.
#     11/09/06  Gary Meyer      Cleaned up sh2_decode_amo_dimm_info().  Print all hex characters in upper case.
#     11/10/06  Gary Meyer      Improved formatting of output to fru reason for XN poisoned data signature.
#     11/22/06  Gary Meyer      Update to ensure the AMO DIMM is assigned to the correct FRU entry.
#     11/28/06  Gary Meyer      Update to differentiate between AMO read & write errors.  Recognize IP79 & IP81.
#     11/30/06  Gary Meyer      Corrected issue with not clearing $read_error with a AMO store error.
#     12/10/06  Gary Meyer      Updated with correct dimm_offset for IP81.
#     03-14-07  Gary Meyer      Update for physical dimm group remapping.
#
#
#-------------------------------------------------------------------------------

use ip41_memory_bit_map;
use ip57_memory_bit_map;
use ip63_memory_bit_map;
use ecc;
use dimm;

############################################################################
#                   XBUS  XBUS  XBUS  XBUS  YBUS  YBUS  YBUS  YBUS         #
#    SHUB   LorR    DIMM  DIMM  DRAM  DRAM  DIMM  DIMM  DRAM  DRAM   FSB   #
#     BIT   DIMM     PIN   BIT  CHIP   PIN   PIN   BIT  CHIP   PIN   BIT   #
############################################################################
#   [  24, "LEFT",   33,   24,    3,   ??,   33,   24,    3,   ??  ], # 0


  @SH1_CHIPKILL_CHIP_UC_SYNDROMES = (
    #--------------------------------------------------------
    # 0x3  0x5  0x6  0x7  0x9  0xa  0xb  0xc  0xd  0xe  0xf    # MASK OF BITS ON CHIP THAT ARE SET.
    #--------------------------------------------------------
    [0x30,0x50,0x60,0x79,0x90,0xa0,0xb9,0xc0,0xd9,0xe9,0xf0,], # CHIP:0R, Bits: 40(0x19), 41(0x29), 42(0x49), 43(0x89)
    [0x7b,0x8d,0xf6,0x79,0xcf,0xb4,0x3b,0x42,0xcd,0xb6,0x39,], # CHIP:1R, Bits: 56(0x8f), 57(0xf4), 65(0x02), 70(0x40)
    [0xed,0x1e,0xf3,0xec,0x9f,0x72,0x6d,0x81,0x9e,0x73,0x6c,], # CHIP:2R, Bits: 62(0x1f), 63(0xf2), 64(0x01), 71(0x80)
    [0x30,0x50,0x60,0x76,0x90,0xa0,0xb6,0xc0,0xd6,0xe6,0xf0,], # CHIP:3R, Bits: 44(0x16), 45(0x26), 46(0x46), 47(0x86)
    [0x30,0x50,0x60,0x7c,0x90,0xa0,0xbc,0xc0,0xdc,0xec,0xf0,], # CHIP:4R, Bits: 32(0x1c), 33(0x2c), 34(0x4c), 35(0x8c)
    [0xde,0xf9,0x27,0xd6,0xe1,0x3f,0xce,0x18,0xe9,0x37,0xc6,], # CHIP:5R, Bits: 30(0xf1), 31(0x2f), 67(0x08), 68(0x10)
    [0xb7,0xfc,0x4b,0xb3,0xd8,0x6f,0x97,0x24,0xdc,0x6b,0x93,], # CHIP:6R, Bits: 24(0xf8), 25(0x4f), 66(0x04), 69(0x20)
    [0x03,0x05,0x06,0x67,0x09,0x0a,0x6b,0x0c,0x6d,0x6e,0x0f,], # CHIP:7R, Bits: 12(0x61), 13(0x62), 14(0x64), 15(0x68)
    [0x30,0x50,0x60,0x7a,0x90,0xa0,0xba,0xc0,0xda,0xea,0xf0,], # CHIP:8R, Bits: 36(0x1a), 37(0x2a), 38(0x4a), 39(0x8a)
    [0x03,0x05,0x06,0x57,0x09,0x0a,0x5b,0x0c,0x5d,0x5e,0x0f,], # CHIP:0L, Bits: 16(0x51), 17(0x52), 18(0x54), 19(0x58)
    [0x03,0x05,0x06,0x37,0x09,0x0a,0x3b,0x0c,0x3d,0x3e,0x0f,], # CHIP:1L, Bits: 20(0x31), 21(0x32), 22(0x34), 23(0x38)
    [0xa0,0x7e,0xde,0xae,0x7b,0xdb,0xab,0x05,0x75,0xd5,0xa5,], # CHIP:2L, Bits: 26(0x70), 27(0xd0), 28(0x0e), 29(0x0b)
    [0x03,0x05,0x06,0xc7,0x09,0x0a,0xcb,0x0c,0xcd,0xce,0x0f,], # CHIP:3L, Bits:  0(0xc1),  1(0xc2),  2(0xc4),  3(0xc8)
    [0x0a,0xe7,0xed,0xea,0xb7,0xbd,0xba,0x50,0x57,0x5d,0x5a,], # CHIP:4L, Bits: 58(0x07), 59(0x0d), 60(0xe0), 61(0xb0)
    [0x03,0x05,0x06,0x97,0x09,0x0a,0x9b,0x0c,0x9d,0x9e,0x0f,], # CHIP:5L, Bits:  8(0x91),  9(0x92), 10(0x94), 11(0x98)
    [0x30,0x50,0x60,0x75,0x90,0xa0,0xb5,0xc0,0xd5,0xe5,0xf0,], # CHIP:6L, Bits: 48(0x15), 49(0x25), 50(0x45), 51(0x85)
    [0x30,0x50,0x60,0x73,0x90,0xa0,0xb3,0xc0,0xd3,0xe3,0xf0,], # CHIP:7L, Bits: 52(0x13), 53(0x23), 54(0x43), 55(0x83)
    [0x03,0x05,0x06,0xa7,0x09,0x0a,0xab,0x0c,0xad,0xae,0x0f,], # CHIP:8L, Bits:  4(0xa1),  5(0xa2),  6(0xa4),  7(0xa8)
  );

  @SH1_BITS_ON_CHIP = (
    [40, 41, 42, 43], # CHIP:0R
    [56, 57, 65, 70], # CHIP:1R
    [62, 63, 64, 71], # CHIP:2R
    [44, 45, 46, 47], # CHIP:3R
    [32, 33, 34, 35], # CHIP:4R
    [30, 31, 67, 68], # CHIP:5R
    [24, 25, 66, 69], # CHIP:6R
    [12, 13, 14, 15], # CHIP:7R
    [36, 37, 38, 39], # CHIP:8R
    [16, 17, 18, 19], # CHIP:0L
    [20, 21, 22, 23], # CHIP:1L
    [26, 27, 28, 29], # CHIP:2L
    [ 0,  1,  2,  3], # CHIP:3L
    [58, 59, 60, 61], # CHIP:4L
    [ 8,  9, 10, 11], # CHIP:5L
    [48, 49, 50, 51], # CHIP:6L
    [52, 53, 54, 55], # CHIP:7L
    [ 4,  5,  6,  7], # CHIP:8L
  );


  @SH2_CHIPKILL_CHIP_UC_SYNDROMES = (
    #--------------------------------------------------------
    # 0x3  0x5  0x6  0x7  0x9  0xa  0xb  0xc  0xd  0xe  0xf    # MASK OF BITS ON CHIP THAT ARE SET.
    #--------------------------------------------------------
    [0x30,0x50,0x60,0x7c,0x90,0xa0,0xbc,0xc0,0xdc,0xec,0xf0,], # CHIP:0R, Bits: 32(0x1c), 33(0x2c), 34(0x4c), 35(0x8c)
    [0x30,0x50,0x60,0x7a,0x90,0xa0,0xba,0xc0,0xda,0xea,0xf0,], # CHIP:1R, Bits: 36(0x1a), 37(0x2a), 38(0x4a), 39(0x8a)
    [0x30,0x50,0x60,0x79,0x90,0xa0,0xb9,0xc0,0xd9,0xe9,0xf0,], # CHIP:2R, Bits: 40(0x19), 41(0x29), 42(0x49), 43(0x89)
    [0x30,0x50,0x60,0x76,0x90,0xa0,0xb6,0xc0,0xd6,0xe6,0xf0,], # CHIP:3R, Bits: 44(0x16), 45(0x26), 46(0x46), 47(0x86)
    [0x30,0x50,0x60,0x75,0x90,0xa0,0xb5,0xc0,0xd5,0xe5,0xf0,], # CHIP:4R, Bits: 48(0x15), 49(0x25), 50(0x45), 51(0x85)
    [0x30,0x50,0x60,0x73,0x90,0xa0,0xb3,0xc0,0xd3,0xe3,0xf0,], # CHIP:5R, Bits: 52(0x13), 53(0x23), 54(0x43), 55(0x83)
    [0x7b,0x8d,0xf6,0x79,0xcf,0xb4,0x3b,0x42,0xcd,0xb6,0x39,], # CHIP:6R, Bits: 56(0x8f), 57(0xf4), 65(0x02), 70(0x40)
    [0x0a,0xe7,0xed,0xea,0xb7,0xbd,0xba,0x50,0x57,0x5d,0x5a,], # CHIP:7R, Bits: 58(0x07), 59(0x0d), 60(0xe0), 61(0xb0)
    [0xed,0x1e,0xf3,0xec,0x9f,0x72,0x6d,0x81,0x9e,0x73,0x6c,], # CHIP:8R, Bits: 62(0x1f), 63(0xf2), 64(0x01), 71(0x80)
    [0x03,0x05,0x06,0xc7,0x09,0x0a,0xcb,0x0c,0xcd,0xce,0x0f,], # CHIP:0L, Bits:  0(0xc1),  1(0xc2),  2(0xc4),  3(0xc8)
    [0x03,0x05,0x06,0xa7,0x09,0x0a,0xab,0x0c,0xad,0xae,0x0f,], # CHIP:1L, Bits:  4(0xa1),  5(0xa2),  6(0xa4),  7(0xa8)
    [0x03,0x05,0x06,0x97,0x09,0x0a,0x9b,0x0c,0x9d,0x9e,0x0f,], # CHIP:2L, Bits:  8(0x91),  9(0x92), 10(0x94), 11(0x98)
    [0x03,0x05,0x06,0x67,0x09,0x0a,0x6b,0x0c,0x6d,0x6e,0x0f,], # CHIP:3L, Bits: 12(0x61), 13(0x62), 14(0x64), 15(0x68)
    [0x03,0x05,0x06,0x57,0x09,0x0a,0x5b,0x0c,0x5d,0x5e,0x0f,], # CHIP:4L, Bits: 16(0x51), 17(0x52), 18(0x54), 19(0x58)
    [0x03,0x05,0x06,0x37,0x09,0x0a,0x3b,0x0c,0x3d,0x3e,0x0f,], # CHIP:5L, Bits: 20(0x31), 21(0x32), 22(0x34), 23(0x38)
    [0xb7,0xfc,0x4b,0xb3,0xd8,0x6f,0x97,0x24,0xdc,0x6b,0x93,], # CHIP:6L, Bits: 24(0xf8), 25(0x4f), 66(0x04), 69(0x20)
    [0xa0,0x7e,0xde,0xae,0x7b,0xdb,0xab,0x05,0x75,0xd5,0xa5,], # CHIP:7L, Bits: 26(0x70), 27(0xd0), 28(0x0e), 29(0x0b)
    [0xde,0xf9,0x27,0xd6,0xe1,0x3f,0xce,0x18,0xe9,0x37,0xc6,], # CHIP:8L, Bits: 30(0xf1), 31(0x2f), 67(0x08), 68(0x10)
  );

  @SH2_BITS_ON_CHIP = (
    [32, 33, 34, 35], # CHIP:0R
    [36, 37, 38, 39], # CHIP:1R
    [40, 41, 42, 43], # CHIP:2R
    [44, 45, 46, 47], # CHIP:3R
    [48, 49, 50, 51], # CHIP:4R
    [52, 53, 54, 55], # CHIP:5R
    [56, 57, 65, 70], # CHIP:6R
    [58, 59, 60, 61], # CHIP:7R
    [62, 63, 64, 71], # CHIP:8R
    [ 0,  1,  2,  3], # CHIP:0L
    [ 4,  5,  6,  7], # CHIP:1L
    [ 8,  9, 10, 11], # CHIP:2L
    [12, 13, 14, 15], # CHIP:3L
    [16, 17, 18, 19], # CHIP:4L
    [20, 21, 22, 23], # CHIP:5L
    [24, 25, 66, 69], # CHIP:6L
    [26, 27, 28, 29], # CHIP:7L
    [30, 31, 67, 68], # CHIP:8L
  );


#---------------------------------------------------------------------------
#
#  subroutine:  decode_proc_err
#
#    When appropriate, this subroutine returns a string value to be used
#    in determining the location of a suspect failing dimm.
#
#    The following cases should be covered in this routine:
#
#	Correctable error info in SH_PI_CORRECTED_DETAIL_3
#	Uncorrectable error info in SH_PI_UNCORRECTED_DETAIL_3
#
#
#---------------------------------------------------------------------------

sub decode_proc_err 
{
  print "\n In decode_proc_err subroutine\n" if ($verbose_mode == 1);;
  my $mmr_name = "$_[0]";
  my $mmr_value = $_[1];
  my $name_tag = "$_[2]";
  my $return_msg = "";
  my $ch_type = ($shub_version == 2) ? "shub2_01" : "shub_01";
  my $addr_cor_name = "SH_PI_CORRECTED_DETAIL_3";
  my $addr_ucor_name = "SH_PI_UNCORRECTED_DETAIL_3";

  $mmr_name =~ s/\s*//g;        #  Remove spaces from $mmr_name

  if ($shub_version == 2)  {
    $addr_cor_name = "SH2_PI_CORRECTED_DETAIL_4";
    $addr_ucor_name = "SH2_PI_UNCORRECTED_DETAIL_4";
  }
  my $err_reg_ptr = reg_create_mmr_inst($ch_type, $mmr_name);
  if ( ! reg_is_valid($err_reg_ptr) ) {  
    $emess = "\nUnable to create register instance: $mmr_name.\n";
    return($emess);
  }
  reg_set_value($err_reg_ptr, $mmr_value);

  if ($mmr_name =~ /$addr_cor_name/)  {
    my $dep      = reg_get_field_value($err_reg_ptr, 'DEP');
    my $syndrome = reg_get_field_value($err_reg_ptr, 'SYNDROME');
    my $address  = reg_get_field_value($err_reg_ptr, 'ADDRESS');
    $address  = $address << 3;
    my $nasid_fr_addr = ($shub_version == 2) ? ($address >> $nasid_index) & $nasid_mask : ($address >> 38) & 0x7ff;
    my $bit = get_bit_from_syndrome($syndrome);
    if (((($address & 0x8) == 0x8) && ($bit < 64)) || ((($address & 0x8) == 0x0) && ($bit > 63))) {
      $bit +=64;
    }
    elsif ((($address & 0x8) == 0x8) && ($bit > 63))  {
      $bit +=72;
    }

    $return_msg .= sprintf("\n%s: FSB_CE Info:  Failing Addr:0x%016llX \(NASID 0x%04X\), Failing Syndrome: 0x%02llX, Bit: %d\n", 
	$name_tag, $address, $nasid_fr_addr, $syndrome, $bit );
  }

  if ($mmr_name =~ /$addr_ucor_name/)  {
    my $dep      = reg_get_field_value($err_reg_ptr, 'DEP');
    my $syndrome = reg_get_field_value($err_reg_ptr, 'SYNDROME');
    my $address  = reg_get_field_value($err_reg_ptr, 'ADDRESS');
    $address  = $address << 3;
    my $nasid_fr_addr = ($shub_version == 2) ? ($address >> $nasid_index) & $nasid_mask : ($address >> 38) & 0x7ff;

    $return_msg .= sprintf("\n%s: FSB_UCE Info:  Failing Addr: 0x%016llX \(NASID 0x%04X\), Failing Syndrome: 0x%02llX\n", 
	$name_tag, $address, $nasid_fr_addr, $syndrome );
    if ($syndrome == 0x3) {
      $return_msg .= sprintf "    Syndrome of 0x3 indicates data has been poisoned in the processor's cache due to read data\n";
      $return_msg .= sprintf "    that was bad from the Shub on a previous load or was corrupted internal to the processor part.\n";
    }
  }

  return ($return_msg);

}

#---------------------------------------------------------------------------
#
#  subroutine:  decode_mem_err
#
#    When appropriate, this subroutine returns a string value to be used
#    in determining the location of a suspect failing dimm.
#
#    The following cases should be covered in this routine:
#
#	Correctable directory error info in SH_DIR_COR_ERR_HDR_LOWER
#	Uncorrectable directory error info in SH_DIR_UC_ERR_HDR_LOWER
#	Correctable error info in SH_PI_CORRECTED_DETAIL_1
#	Uncorrectable error info in SH_PI_UNCORRECTED_DETAIL_1
#	Correctable error info in SH_XN_CORRECTED_DETAIL_1, 3 & 4
#	Uncorrectable error info in SH_XN_UNCORRECTED_DETAIL_3 & 4
#
#
#---------------------------------------------------------------------------

sub decode_mem_err 
{
  print "\n In decode_mem_err subroutine\n" if ($verbose_mode == 1);;
  my $mmr_name = "$_[0]";
  my $mmr_value = $_[1];
  my $name_tag = "$_[2]";
  my $board_type = "$_[3]";
  my $return_msg = "";
  my $sbe_dimm_pin;
  my $ch_type = ($shub_version == 2) ? "shub2_01" : "shub_01";
  if ($shub_version == 2)  {
    $t_cnt = 0;
    $target_list[$t_cnt]->{DIMM_SUMM} = "";	#  This is to allow re-use of shub1 code w/o re-writing too much
  }

  my $dtr_offset = 0;
  my $phys_id = "";
  my $dimm_node = 0;

  my $dimm_offset = ($board_type =~ /IP81/) ? 2 : 3;
 
  my $phys_node = $target_list[$t_cnt]->{NODE};
  $phys_node =~ s/-J2//;		# strip out the "-J2" that comes with IP-67

  if ($phys_node =~ /1/) {
    $dtr_offset +=2;
    $dimm_node = 1;
  }
  elsif ($phys_node =~ /2/) {
    $dimm_node = 2;
  }
  elsif ($phys_node =~ /3/) {
    $dimm_node = 3;
  }

  my $failing_dimm_node = ($shub_version == 2) ? "" : sprintf "N%d_", $dimm_node;

  $mmr_name =~ s/\s*//g;        #  Remove spaces from $mmr_name

  my $err_reg_ptr = reg_create_mmr_inst($ch_type, $mmr_name);
  if ( ! reg_is_valid($err_reg_ptr) ) {  
    $emess = "\nUnable to create register instance: $mmr_name.\n";
    return($emess);
  }
  reg_set_value($err_reg_ptr, $mmr_value);

  if ($mmr_name =~ /_DIR_COR_.*_HDR_LOWER/)  {			#  Directory correctable
    $dir_addr = reg_get_field_value($err_reg_ptr, 'ADDR');
    $dir_addr <<= 3;
    $dir_bus = ($dir_addr >> 7) & 0x1;			#  0=Y    1=X
    $dir_LorR = ($dir_addr >> 11) & 0x1;		#  0=L    1=R
    $data_side = ($dir_LorR) ? "RIGHT" : "LEFT";
    my $failing_dimm_bus = ($shub_version == 2) ? "" : sprintf "_BUS_%s", (($dir_bus) ? "X" : "Y");

    my $raw_dimm_group = ($dir_addr >> $dimm_group_index) & 0x3;
    $failing_dimm_group = get_physical_group($raw_dimm_group);
    $failing_dtr = $dtr_offset + $dir_LorR;

    if ($shub_version == 2)  {
      $failing_dimm_slot = $failing_dimm_group + ( $dimm_offset * (1 - $dir_bus));
    }
    else  {
      $failing_dimm_slot = ($failing_dimm_group * 2) + (1 - $dir_bus);
    }

    $return_msg .= sprintf("\n%s: MD DIMM Info: Directory correctable error - Dir. Addr: 0x%013llX Dimm Group %d, %s BUS %s.\n",
        $name_tag, $dir_addr, $failing_dimm_group, (($dir_bus) ? "X" : "Y"), (($dir_LorR) ? "R" : "L") );
    $target_list[$t_cnt]->{DIMM_SUMM} .= sprintf("\n%s: MD DIMM Info: Directory correctable error - Dir. Addr: 0x%013llX Dimm Group %d, %s BUS %s.",
        $name_tag, $dir_addr, $failing_dimm_group, (($dir_bus) ? "X" : "Y"), (($dir_LorR) ? "R" : "L") );

    my $failing_dimm_loc = sprintf("DIMM%d_%s%s%s", $failing_dimm_slot, $failing_dimm_node, 
	(($dir_LorR) ? "R" : "L"), $failing_dimm_bus );
    $target_list[$t_cnt]->{DIMM_SUMM} .= sprintf("\n%s: MD reported DIMM Location:  %s\n", $name_tag, $failing_dimm_loc);

    $fru_list[$md_ce_fru_entry]->{REASON} .= "Suspect $failing_dimm_loc.  ";

  }

  if ($mmr_name =~ /_DIR_UC_.*_HDR_LOWER/)  {			#  Directory uncorrectable
    $dir_addr = reg_get_field_value($err_reg_ptr, 'ADDR');
    $dir_addr <<= 3;
    $dir_bus = ($dir_addr >> 7) & 0x1;			#  0=Y    1=X
    my $failing_dimm_bus = ($shub_version == 2) ? "" : sprintf "_BUS_%s", (($dir_bus) ? "X" : "Y");
    $dir_LorR = ($dir_addr >> 11) & 0x1;		#  0=L    1=R
    $data_side = ($dir_LorR) ? "RIGHT" : "LEFT";

    my $raw_dimm_group = ($dir_addr >> $dimm_group_index) & 0x3;
    $failing_dimm_group = get_physical_group($raw_dimm_group);
    $failing_dtr = $dtr_offset + $dir_LorR;
    if ($shub_version == 2)  {
      $failing_dimm_slot = $failing_dimm_group + ( $dimm_offset * (1 - $dir_bus));
    }
    else  {
      $failing_dimm_slot = ($failing_dimm_group * 2) + (1 - $dir_bus);
    }

    $return_msg .= sprintf("\n%s: MD DIMM Info: Directory uncorrectable error - Dir. Addr: 0x%013llX Dimm Group %d, %s BUS %s.\n",
         $name_tag, $dir_addr, $failing_dimm_group, (($dir_bus) ? "X" : "Y"), (($dir_LorR) ? "R" : "L") );
    $target_list[$t_cnt]->{DIMM_SUMM} .= sprintf("\n%s: MD DIMM Info: Directory uncorrectable error - Dir. Addr: 0x%013llX Dimm Group %d, %s BUS %s.",
         $name_tag, $dir_addr, $failing_dimm_group, (($dir_bus) ? "X" : "Y"), (($dir_LorR) ? "R" : "L") );

    my $failing_dimm_loc = sprintf("DIMM%d_%s%s%s", $failing_dimm_slot, $failing_dimm_node, 
	(($dir_LorR) ? "R" : "L"), $failing_dimm_bus );
    $target_list[$t_cnt]->{DIMM_SUMM} .= sprintf("\n%s: MD reported DIMM Location:  %s\n", $name_tag, $failing_dimm_loc);

    $fru_list[$md_uce_fru_entry]->{REASON} .= "Suspect $failing_dimm_loc.  ";

  }

  if ($mmr_name =~ /_PI_CORRECTED_DETAIL_1/)  {
    $dep      = reg_get_field_value($err_reg_ptr, 'DEP');
    $syndrome = reg_get_field_value($err_reg_ptr, 'SYNDROME');
    my $fsb_address  = reg_get_field_value($err_reg_ptr, 'ADDRESS');
    $fsb_address  = $fsb_address << 3;
    if ($shub_version == 2)  {
      my $pi_dimm_group_index = $nasid_index - 4;
      my $raw_dimm_group = ($fsb_address >> $pi_dimm_group_index) & 0x3;
      $failing_dimm_group = get_physical_group($raw_dimm_group);
      $nasid_fr_addr = ($fsb_address >> $nasid_index) & $nasid_mask;
      $address  = (($fsb_address >> $pi_dimm_group_index) & 3) << $dimm_group_index;
      $address += $fsb_address & ($offset_mask >> 2);
      $address += (($fsb_address >> $nasid_index) & $nasid_mask) << 38;
    }
    else {
      $address = $fsb_address;
      $nasid_fr_addr = ($address >> 38) & 0x7ff;
      my $raw_dimm_group = ($address >> $dimm_group_index) & 0x3;
      $failing_dimm_group = get_physical_group($raw_dimm_group);
    }    
    #  $data     = reg_get_field_value($sh_pi_corrected_detail_2, 'DATA');
    $failing_bus = ($address >> 7) & 0x1;		#  0=X     1=Y
    my $failing_dimm_bus = ($shub_version == 2) ? "" : sprintf "_BUS_%s", (($failing_bus) ? "Y" : "X");
    $bit = get_bit_from_syndrome($syndrome);

    if (((($address & 0x8) == 0x8) && ($bit < 64)) || ((($address & 0x8) == 0x0) && ($bit > 63))) {
      $bit +=64;
    }
    elsif ((($address & 0x8) == 0x8) && ($bit > 63))  {
      $bit +=72;
    }

    $data_LorR = fsb_bit_to_LorR($bit, $board_type); 	
    $data_side = ($data_LorR) ? "RIGHT" : "LEFT";

    $failing_dtr = $dtr_offset + $data_LorR;
    if ($shub_version == 2)  {
      $failing_dimm_slot = $failing_dimm_group + ( $dimm_offset * $failing_bus);
    }
    else  {
      $failing_dimm_slot = ($failing_dimm_group * 2) + $failing_bus;
    }

    $sbe_dimm_pin = fsb_bit_to_dimm_pin($bit, $board_type);

    $return_msg .= sprintf("\n%s: PI DIMM Info:  Failing SN2NET Addr:0x%016llX (NASID 0x%04X) Failing Syndrome: 0x%02llX ", 
	$name_tag, $address, $nasid_fr_addr, $syndrome );
    $target_list[$t_cnt]->{DIMM_SUMM} .= sprintf("\n%s: PI DIMM Info:  Failing SN2NET Addr:0x%016llX (NASID 0x%04X) Failing Syndrome: 0x%02llX ", 
	$name_tag, $address, $nasid_fr_addr, $syndrome );

    $return_msg .= sprintf("\n%s: PI DIMM Info:   Syndrome:0x%02llX, Bit %d, Dimm Group %d, %s BUS %s   (SBE) \n",
          $name_tag, $syndrome, $bit, $failing_dimm_group, ($failing_bus) ? "Y" : "X", $data_side );
    $target_list[$t_cnt]->{DIMM_SUMM} .= sprintf("\n%s: PI DIMM Info:   Syndrome:0x%02llX, Bit %d, Dimm Group %d, %s BUS %s   (SBE) ",
          $name_tag, $syndrome, $bit, $failing_dimm_group, ($failing_bus) ? "Y" : "X", $data_side );

    my $failing_dimm_loc = sprintf("DIMM%d_%s%s%s", $failing_dimm_slot, $failing_dimm_node, 
	(($data_LorR) ? "R" : "L"), $failing_dimm_bus );
    $target_list[$t_cnt]->{DIMM_SUMM} .= sprintf("\n%s: PI reported DIMM Location:  %s\n", $name_tag, $failing_dimm_loc);

    $fru_list[$pi_ce_fru_entry]->{REASON} .= "Suspect $failing_dimm_loc.  ";

  }

  if ($mmr_name =~ /_PI_UNCORRECTED_DETAIL_1/)  {
    $dep      = reg_get_field_value($err_reg_ptr, 'DEP');
    $syndrome = reg_get_field_value($err_reg_ptr, 'SYNDROME');
    $address  = reg_get_field_value($err_reg_ptr, 'ADDRESS');
    my $fsb_address  = reg_get_field_value($err_reg_ptr, 'ADDRESS');
    $fsb_address  = $fsb_address << 3;
    if ($shub_version == 2)  {
      my $pi_dimm_group_index = $nasid_index - 4;
      my $raw_dimm_group = ($fsb_address >> $pi_dimm_group_index) & 0x3;
      $failing_dimm_group = get_physical_group($raw_dimm_group);
      $nasid_fr_addr = ($fsb_address >> $nasid_index) & $nasid_mask;
      $address  = (($fsb_address >> $pi_dimm_group_index) & 3) << $dimm_group_index;
      $address += $fsb_address & ($offset_mask >> 2);
      $address += (($fsb_address >> $nasid_index) & $nasid_mask) << 38;
    }
    else {
      $address = $fsb_address;
      $nasid_fr_addr = ($address >> 38) & 0x7ff;
      my $raw_dimm_group = ($address >> $dimm_group_index) & 0x3;
      $failing_dimm_group = get_physical_group($raw_dimm_group);
    }

    $return_msg .= sprintf("\n%s: PI DIMM Info:  Failing SN2NET Addr: 0x%016llX (NASID 0x%04X) , Failing Syndrome: 0x%02llX\n", 
	$name_tag, $address, $nasid_fr_addr, $syndrome );
    $target_list[$t_cnt]->{DIMM_SUMM} .= sprintf("\n%s: PI DIMM Info:  Failing SN2NET Addr: 0x%016llX (NASID 0x%04X) , Failing Syndrome: 0x%02llX\n", 
	$name_tag, $address, $nasid_fr_addr, $syndrome );
    $failing_bus = ($address >> 7) & 0x1;			#  0=X     1=Y
    my $failing_dimm_bus = ($shub_version == 2) ? "" : sprintf "_BUS_%s", (($failing_bus) ? "Y" : "X");
    if ($shub_version == 2)  {
      $failing_dimm_slot = $failing_dimm_group + ( $dimm_offset * $failing_bus);
    }
    else  {
      $failing_dimm_slot = ($failing_dimm_group * 2) + $failing_bus;
    }

    my $mbe_dimm_msg = get_mbe_chip_side($syndrome);
    $data_side = ($mbe_dimm_msg =~ /Right Side/) ? "RIGHT" : ($mbe_dimm_msg =~ /Left Side/) ? "LEFT" : "??";
    $data_LorR = ($mbe_dimm_msg =~ /Right Side/) ? "R" : ($mbe_dimm_msg =~ /Left Side/) ? "L" : "??";

    $return_msg .= sprintf("%s: PI DIMM Info:   Syndrome:0x%02llX, Dimm Group %d, %s BUS %s   (MBE)\n",
         $name_tag, $syndrome, $failing_dimm_group, ($failing_bus) ? "Y" : "X", $data_side);
    $return_msg .= sprintf("%s\n", $mbe_dimm_msg);

    $target_list[$t_cnt]->{DIMM_SUMM} .= sprintf("%s: PI DIMM Info:   Syndrome:0x%02llX, Dimm Group %d, %s BUS %s   (MBE)",
         $name_tag, $syndrome, $failing_dimm_group, ($failing_bus) ? "Y" : "X", $data_side);

    my $failing_dimm_loc = sprintf("DIMM%d_%s%s%s", $failing_dimm_slot, $failing_dimm_node, $data_LorR, $failing_dimm_bus);

    $target_list[$t_cnt]->{DIMM_SUMM} .= sprintf("\n%s: PI reported DIMM Location:  %s\n", $name_tag, $failing_dimm_loc);

    $fru_list[$pi_uce_fru_entry]->{REASON} .= "Suspect $failing_dimm_loc.  ";

  }

  if ($mmr_name =~ /_XN_.*_DETAIL_1/)  {
    $syndrome[0] = reg_get_field_value($err_reg_ptr, 'ECC0_SYNDROME');
    $syndrome[1] = reg_get_field_value($err_reg_ptr, 'ECC1_SYNDROME');
    $syndrome[2] = reg_get_field_value($err_reg_ptr, 'ECC2_SYNDROME');
    $syndrome[3] = reg_get_field_value($err_reg_ptr, 'ECC3_SYNDROME');

    if ($shub_version == 2) {
      $xn_err_grp = reg_get_field_value($err_reg_ptr, 'ERR_GROUP');;			#  which 64 bit group contains failing dword
    }

    @xn_failing_dtr = ();
    @xn_failing_side = ();

    if ($mmr_name =~ /_XN_CORRECTED_DETAIL_1/)  {
      for ($flit_num=0; $flit_num < 4; $flit_num++)  {
        if ($syndrome[$flit_num] != 0)  {
          $bit = get_bit_from_syndrome($syndrome[$flit_num]);

          if (((($flit_num & 1) == 1) && ($bit < 64)) || ((($flit_num & 0x1) == 0x0) && ($bit > 63))) {
            $bit +=64;
          }
          elsif ((($flit_num & 1) == 1) && ($bit > 63))  {
            $bit +=72;
          }

          $data_LorR = fsb_bit_to_LorR($bit, $board_type); 	

    	  $data_side = ($data_LorR) ? "RIGHT" : "LEFT";

          $final_dtr = $dtr_offset + $data_LorR;
	  $sbe_dimm_pin = fsb_bit_to_dimm_pin($bit, $board_type);
	  if (($board_type =~ /IP57/) || ($board_type =~ /IP[678][1359]/)) {
              $xn_failing_dtr[$flit_num] = sprintf "%s side     DIMM pin: %d", $data_side, $sbe_dimm_pin;
	  }
	  else  {
              $xn_failing_dtr[$flit_num] = sprintf "DTR %d     DIMM pin: %d", $final_dtr, $sbe_dimm_pin;
	  }

	  $xn_failing_side[$flit_num] = ($data_LorR) ? "R" : "L";

          $return_msg .= sprintf("\n%s: XN DIMM Info: Failing Syndrome in DWord %d: 0x%02llX   Bit: %d  DIMM Pin: %d  BUS: %s ", 
	    $name_tag, $flit_num, $syndrome[$flit_num], $bit, $sbe_dimm_pin, $data_side );
          $target_list[$t_cnt]->{DIMM_SUMM} .= sprintf("\n%s: XN DIMM Info:  Failing Syndrome in DWord %d: 0x%02llX    Bit: %d  DIMM Pin: %d  BUS %s ", 
	    $name_tag, $flit_num, $syndrome[$flit_num], $bit, $sbe_dimm_pin, $data_side );
        }
        else {
	  $xn_failing_dtr[$flit_num] = "NONE";
	  $xn_failing_side[$flit_num] = "??";
        }
      }
    }
    else  {			#  Uncorrected XN error case
      for ($flit_num=0; $flit_num < 4; $flit_num++)  {
        if (($syndrome[$flit_num] > 0) && ($syndrome[$flit_num] < 0xff))  {
    	  my $mbe_dimm_msg = get_mbe_chip_side($syndrome[$flit_num]);
    	  $data_side = ($mbe_dimm_msg =~ /Right Side/) ? "RIGHT" : ($mbe_dimm_msg =~ /Left Side/) ? "LEFT" : "Either Left or Right side";
    	  $xn_failing_side[$flit_num] = ($mbe_dimm_msg =~ /Right Side/) ? "R" : ($mbe_dimm_msg =~ /Left Side/) ? "L" : "??";

	  if (($board_type =~ /IP57/) || ($board_type =~ /IP[678][1359]/)) {
              $xn_failing_dtr[$flit_num] = sprintf "%s", $data_side;
	  }
	  else  {
	      if ($mbe_dimm_msg =~ /Right Side/)  {
                  $xn_failing_dtr[$flit_num] = sprintf "DTR %d", $dtr_offset + 1;
	      }
	      elsif ($mbe_dimm_msg =~ /Left Side/) {
                  $xn_failing_dtr[$flit_num] = sprintf "DTR %d", $dtr_offset;
	      }
	      else  {
                  $xn_failing_dtr[$flit_num] = sprintf "either DTR %d or %d", $dtr_offset, ($dtr_offset + 1);
	      }
	  }

          $return_msg .= sprintf("\n%s: XN DIMM Info:  Failing Syndrome in DWord %d: 0x%02llX    Bit: Multiple Bit ", 
	    $name_tag, $flit_num, $syndrome[$flit_num] );
          $target_list[$t_cnt]->{DIMM_SUMM} .= sprintf("\n%s: XN DIMM Info:  Failing Syndrome in DWord %d: 0x%02llX   Bit: Multiple Bit", 
	    $name_tag, $flit_num, $syndrome[$flit_num]);
	  $target_list[$t_cnt]->{DIMM_SUMM} .= sprintf("%s", $mbe_dimm_msg);

        }
        elsif ($syndrome[$flit_num] == 0xff)  {
	  $xn_failing_dtr[$flit_num] = "NONE - Poisoned Data";
	  $xn_failing_side[$flit_num] = "NONE - Poisoned Data";
          $return_msg .= sprintf("\n%s: XN DIMM Info:  Failing Syndrome in DWord %d: 0x%02llX    Bit: Multiple Bit - Poisoned ", 
	    $name_tag, $flit_num, $syndrome[$flit_num] );
        }
        else {
	  $xn_failing_dtr[$flit_num] = "NONE";
	  $xn_failing_side[$flit_num] = "??";
        }
      }
    }

    $return_msg .= "\n";
    $target_list[$t_cnt]->{DIMM_SUMM} .= "\n";
  }

  if ($mmr_name =~ /_XN_.*_DETAIL_3/)  {
    if ($shub_version == 2)  {
      $xn_suppl = $mmr_value & 0xffff;
      $xn_src_node = ($mmr_value >> 16) & 0x3fff;
      $xn_src_chip = ($mmr_value >> 32) & 0x3;
      $xn_echo = ($mmr_value >> 36) & 0x1ff;
      $xn_dst_node = ($mmr_value >> 48) & 0x3fff;
    }
    else {
      $xn_addr_44_3 = ($mmr_value & 0xffffffffffc00000) >> 19;  	#  fm_addr[41:0] are in detail_3 bits [63:22]
      $xn_dst_node = ($mmr_value &            0x7ff00) >> 8;	#  fm_dest[10:0] are in detail_3 bits [18:8]
      $xn_dst_chip = ($mmr_value &           0x380000) >> 19;	#  fm_dest[13:11] are in detail_3 bits [21:19]
      $xn_cmd	  = $mmr_value & 0xff;				#  fm_cmd[7:0] are in detail_3 bits [7:0]
    }
  }
  if ($mmr_name =~ /_XN_.*_DETAIL_4/)  {
    if ($shub_version == 2)  {
      $xn_dst_chip = $mmr_value & 0x3;
      $xn_cmd = ($mmr_value >> 4) & 0xff;
      $xn_addr = ($mmr_value >> 15) & 0x1ffffffffffff;
      $final_address = $xn_addr << 3;
      $nasid_fr_addr = ($final_address >> 38) & 0x3fff;
      $failing_bus = ($final_address >> 7) & 0x1;			#  0=X     1=Y
      $failing_dimm_bus = "";
      my $raw_dimm_group = ($final_address >> $dimm_group_index) & 0x3;
      $failing_dimm_group = get_physical_group($raw_dimm_group);
      $failing_dimm_slot = $failing_dimm_group + ( $dimm_offset * $failing_bus);
    }
    else  {
      $xn_src_chip   = ($mmr_value & 0x38000000000) >> 39;	#  fm_src[13:11] are in detail_4 bits [41:39]
      $xn_src_node   = ($mmr_value & 0x07ff0000000) >> 28;	#  fm_src[10:0] are in detail_4 bits [38:28]
      $xn_echo	   = ($mmr_value & 0x0000ff80000) >> 19;	#  fm_echo[8:0] are in detail_4 bits [27:19]
      $xn_suppl	   = ($mmr_value & 0x0000007ffe0) >> 5;		#  fm_suppl[13:0] are in detail_4 bits [18:5]
      $xn_addr_49_45 =  $mmr_value & 0x0000000001f;		#  fm_addr[46:42] are in detail_4 bits [4:0]
      $xn_err_grp    = ($mmr_value >> 62) & 3;			#  which 64 bit group contains failing dword
      $final_address = $xn_addr_44_3 + ($xn_addr_49_45 << 45);
      $nasid_fr_addr = ($final_address >> 38) & 0x7ff;
      $failing_bus = ($final_address >> 7) & 0x1;			#  0=X     1=Y
      $failing_dimm_bus = sprintf "_BUS_%s", (($failing_bus) ? "Y" : "X");
      my $raw_dimm_group = ($final_address >> $dimm_group_index) & 0x3;
      $failing_dimm_group = get_physical_group($raw_dimm_group);
      $failing_dimm_slot = ($failing_dimm_group * 2) + $failing_bus;
    }

    my $return_dimm_msg = sprintf("\n%s: XN DIMM Info:  Failing Addr: 0x%013llX (NASID 0x%04X), DWord: %d, Dimm Group %d, %s BUS %s\n",
	$name_tag, $final_address, $nasid_fr_addr, $xn_err_grp, $failing_dimm_group,  ($failing_bus) ? "Y" : "X");

    my $added_to_phys = 0;
    my $legit_xn_uce = 0;
    for ($flit_num=0; $flit_num < 4; $flit_num++)  {
      if ($flit_num == $xn_err_grp)  {
        if ($xn_failing_side[$flit_num] !~ /Poison/) {
	  $legit_xn_uce = 1;
	  my $tmp_phys_id = "";
          my $failing_dimm_loc = sprintf("DIMM%d_%s%s%s", $failing_dimm_slot, $failing_dimm_node,
              $xn_failing_side[$flit_num], $failing_dimm_bus );
          $return_dimm_msg .= sprintf("\n%s: XN reported DIMM Location:  %s\n", $name_tag, $failing_dimm_loc);
          $tmp_phys_id = "$failing_dimm_loc";
  
	  if ($phys_id !~ /$tmp_phys_id/)  {
	    if ($added_to_phys == 1)  {
	      $phys_id .= ", ";			#  add comma seperator if there's already a DIMM in the string
	    }
	    $phys_id .= "$tmp_phys_id";
	    $added_to_phys = 1;
	  }
        }
      }
    }
    $return_dimm_msg .= "\n";
    
    if (($mmr_name !~ /UN/) || ($legit_xn_uce == 1))  {
      $return_msg .= $return_dimm_msg;
      $target_list[$t_cnt]->{DIMM_SUMM} .= $return_dimm_msg;
    }	

    $return_msg .= "\n$name_tag: XN Detected Memory Error Header Field Info:\n";
    $return_msg .= sprintf("\tSource Node:\t\t0x%04X\n", $xn_src_node);
    $return_msg .= sprintf("\tSource Chiplet:\t\t0x%X  (%s)\n", $xn_src_chip, $sn2net_chiplet_list[$xn_src_chip]);
    $return_msg .= sprintf("\tCommand:\t\t0x%02X   (%s)\n", $xn_cmd, $sn2net_command_list[$xn_cmd]);
    $return_msg .= sprintf("\tAddress:\t\t0x%013llX\n", $final_address);
    $return_msg .= sprintf("\tSupplemental:\t\t0x%04X\n", $xn_suppl);
    $return_msg .= sprintf("\tEcho:\t\t\t0x%02X\n", $xn_echo);
    $return_msg .= sprintf("\tDest Node:\t\t0x%04X\n", $xn_dst_node);
    $return_msg .= sprintf("\tDest Chiplet:\t\t0x%X  (%s)\n", $xn_dst_chip, $sn2net_chiplet_list[$xn_dst_chip]);

    if ($mmr_name =~ /UN/)  {
      if ($legit_xn_uce == 1) {
        $fru_list[$xn_uce_fru_entry]->{REASON} .= "Suspect $phys_id.  ";
      }
      else {
        $fru_list[$xn_uce_fru_entry]->{REASON} .= "\n\t\t\t\t\t\tXN read Poisoned Data that was previously written to this MD.  ";
	$any_dimm_error--;
      }
    }  else  {
      $fru_list[$xn_ce_fru_entry]->{REASON} .= "Suspect $phys_id.  ";
    }  
  }

  if ($shub_version == 2)  {
    $sh2_target_list[$sh2_cnt]->{DIMM_SUMM} .= $target_list[$t_cnt]->{DIMM_SUMM};
  }

  return ($return_msg);

}

#---------------------------------------------------------------------------
#
#  This large, fairly clumsy function is to generate DIMM location info from
#  MD reported AMO memory errors.  
#
#  Inputs:	$dq_type - string, either DQRS or DQLS
#               $name_tag - string indicating which nasid we're on
#               $board_type - string; either IP41 or IP57
#
#
#---------------------------------------------------------------------------

sub decode_amo_dimm_info
{
  print "\n In decode_amo_dimm_info subroutine\n" if ($verbose_mode == 1);;
  my $dq_type = "$_[0]";
  my $name_tag = "$_[1]";
  my $board_type = "$_[2]";
  my $sbe_dimm_pin;
  my $dtr_offset = 0;
  my $phys_id = "";
  my $dimm_node = 0;
  my $dimm_offset = ($board_type =~ /IP81/) ? 2 : 3;

  my $phys_node = $target_list[$t_cnt]->{NODE};
  $phys_node =~ s/-J2//;                # strip out the "-J2" that comes with IP-67

  if ($phys_node =~ /1/) {
    $dtr_offset +=2;
    $dimm_node = 1;
  }
  elsif ($phys_node =~ /2/) {
    $dimm_node = 2;
  }
  elsif ($phys_node =~ /3/) {
    $dimm_node = 3;
  }

  my $failing_dimm_node = ($shub_version == 2) ? "" : sprintf "N%d_", $dimm_node;

  if ($shub_version == 2)  {
    $t_cnt = 0;
    $target_list[$t_cnt]->{DIMM_SUMM} = "";
  }
 
  if ($dq_type eq 'DQLS')  {
    my $address = $target_list[$t_cnt]->{DQLS_AMO_ERR_ADDR};
    $failing_bus = 0;			 #  0=X     1=Y
    my $failing_dimm_bus = ($shub_version == 2) ? "" : sprintf "_BUS_%s", (($failing_bus) ? "Y" : "X");
    my $raw_dimm_group = ($address >> $dimm_group_index) & 0x3;
    $failing_dimm_group = get_physical_group($raw_dimm_group);


    if (defined $target_list[$t_cnt]->{DQLS_AMO_SCOR})   {
      my $syndrome = $target_list[$t_cnt]->{DQLS_AMO_SSYN}; 
      $bit = get_bit_from_syndrome($syndrome);
      if (((($address & 0x8) == 0x8) && ($bit < 64)) || ((($address & 0x8) == 0x0) && ($bit > 63))) {
        $bit +=64;
      }
      elsif ((($address & 0x8) == 0x8) && ($bit > 63))  {
        $bit +=72;
      }
 
      $data_LorR = fsb_bit_to_LorR($bit, $board_type);

      $data_side = ($data_LorR) ? "RIGHT" : "LEFT";
 
      $failing_dtr = $dtr_offset + $data_LorR;
      if ($shub_version == 2)  {
        $failing_dimm_slot = $failing_dimm_group + ( $dimm_offset * $failing_bus);
      }
      else  {
        $failing_dimm_slot = ($failing_dimm_group * 2) + $failing_bus;
      }

      $sbe_dimm_pin = fsb_bit_to_dimm_pin($bit, $board_type);
      $target_list[$t_cnt]->{DIMM_SUMM} .= sprintf("\n%s: MD AMO (Store) DIMM Info:  Failing Addr:0x%016llX  Failing Syndrome: 0x%02llX ",
        $name_tag, $address, $syndrome );
      $target_list[$t_cnt]->{DIMM_SUMM} .= sprintf("\n%s: MD AMO (Store) DIMM Info:  Bit %d, Dimm Group %d, %s BUS %s   (SBE) ",
          $name_tag, $bit, $failing_dimm_group, ($failing_bus) ? "Y" : "X", $data_side );

      my $failing_dimm_loc = sprintf("DIMM%d_%s%s%s", $failing_dimm_slot, $failing_dimm_node,
          (($data_LorR) ? "R" : "L"), $failing_dimm_bus );
      $target_list[$t_cnt]->{DIMM_SUMM} .= sprintf("\n%s: MD AMO (Store) DIMM Location:  %s   DIMM pin: %d\n", 
	  $name_tag, $failing_dimm_loc, $sbe_dimm_pin);
      $phys_id = sprintf("$failing_dimm_loc");

    }


    if (defined $target_list[$t_cnt]->{DQLS_AMO_RCOR})   {
      my $syndrome = $target_list[$t_cnt]->{DQLS_AMO_RSYN}; 
      $bit = get_bit_from_syndrome($syndrome);
      if (((($address & 0x8) == 0x8) && ($bit < 64)) || ((($address & 0x8) == 0x0) && ($bit > 63))) {
        $bit +=64;
      }
      elsif ((($address & 0x8) == 0x8) && ($bit > 63))  {
        $bit +=72;
      }

      $data_LorR = fsb_bit_to_LorR($bit, $board_type);

      $data_side = ($data_LorR) ? "RIGHT" : "LEFT";
 
      $failing_dtr = $dtr_offset + $data_LorR;
      if ($shub_version == 2)  {
        $failing_dimm_slot = $failing_dimm_group + ( $dimm_offset * $failing_bus);
      }
      else  {
        $failing_dimm_slot = ($failing_dimm_group * 2) + $failing_bus;
      }
      $sbe_dimm_pin = fsb_bit_to_dimm_pin($bit, $board_type);
      $target_list[$t_cnt]->{DIMM_SUMM} .= sprintf("\n%s: MD AMO (Read) DIMM Info:  Failing Addr:0x%016llX  Failing Syndrome: 0x%02llX ",
        $name_tag, $address, $syndrome );
      $target_list[$t_cnt]->{DIMM_SUMM} .= sprintf("\n%s: MD AMO (Read) DIMM Info:  Bit %d, Dimm Group %d, %s BUS %s   (SBE) ",
          $name_tag, $bit, $failing_dimm_group, ($failing_bus) ? "Y" : "X", $data_side );

      my $failing_dimm_loc = sprintf("DIMM%d_%s%s%s", $failing_dimm_slot, $failing_dimm_node,
          (($data_LorR) ? "R" : "L"), $failing_dimm_bus );
      $target_list[$t_cnt]->{DIMM_SUMM} .= sprintf("\n%s: MD AMO (Read) DIMM Location:  %s   DIMM pin: %d\n", 
	  $name_tag, $failing_dimm_loc, $sbe_dimm_pin);
      $phys_id = sprintf("$failing_dimm_loc");

    }

    if (defined $target_list[$t_cnt]->{DQLS_AMO_SUNC})   {
      my $syndrome = $target_list[$t_cnt]->{DQLS_AMO_SSYN}; 
 
      if ($shub_version == 2)  {
        $failing_dimm_slot = $failing_dimm_group + ( $dimm_offset * $failing_bus);
      }
      else  {
        $failing_dimm_slot = ($failing_dimm_group * 2) + $failing_bus;
      }

      my $mbe_dimm_msg = get_mbe_chip_side($syndrome);
      $data_side = ($mbe_dimm_msg =~ /Right Side/) ? "RIGHT" : ($mbe_dimm_msg =~ /Left Side/) ? "LEFT" : "either side";
      $data_LorR = ($mbe_dimm_msg =~ /Right Side/) ? "R" : ($mbe_dimm_msg =~ /Left Side/) ? "L" : "??";

      $target_list[$t_cnt]->{DIMM_SUMM} .= sprintf("\n%s: MD AMO (Store) DIMM Info:  Failing Addr:0x%016llX  Failing Syndrome: 0x%02llX ",
        $name_tag, $address, $syndrome );
      $target_list[$t_cnt]->{DIMM_SUMM} .= sprintf("\n%s: MD AMO (Store) DIMM Info:  Dimm Group %d, %s BUS %s  (MBE) ",
          $name_tag, $failing_dimm_group, $failing_bus, $data_side );

      my $failing_dimm_loc = sprintf("DIMM%d_%s%s%s", $failing_dimm_slot, $failing_dimm_node, $data_LorR, $failing_dimm_bus );
      $target_list[$t_cnt]->{DIMM_SUMM} .= sprintf("\n%s: MD AMO (Store) DIMM Location:  %s\n", $name_tag, $failing_dimm_loc);
      $target_list[$t_cnt]->{DIMM_SUMM} .= sprintf("\n%s\n", $mbe_dimm_msg);
      $phys_id = sprintf("$failing_dimm_loc");

    }

    if (defined $target_list[$t_cnt]->{DQLS_AMO_RUNC})   {
      my $syndrome = $target_list[$t_cnt]->{DQLS_AMO_RSYN}; 
      my $mbe_dimm_msg = get_mbe_chip_side($syndrome);
      $data_side = ($mbe_dimm_msg =~ /Right Side/) ? "RIGHT" : ($mbe_dimm_msg =~ /Left Side/) ? "LEFT" : "either side";
      $data_LorR = ($mbe_dimm_msg =~ /Right Side/) ? "R" : ($mbe_dimm_msg =~ /Left Side/) ? "L" : "??";
 
      if ($shub_version == 2)  {
        $failing_dimm_slot = $failing_dimm_group + ( $dimm_offset * $failing_bus);
      }
      else  {
        $failing_dimm_slot = ($failing_dimm_group * 2) + $failing_bus;
      }

      $target_list[$t_cnt]->{DIMM_SUMM} .= sprintf("\n%s: MD AMO (Read) DIMM Info:  Failing Addr:0x%016llX  Failing Syndrome: 0x%02llX ",
        $name_tag, $address, $syndrome );
      $target_list[$t_cnt]->{DIMM_SUMM} .= sprintf("\n%s: MD AMO (Read) DIMM Info:  Dimm Group %d, %s BUS %s   (MBE) ",
          $name_tag, $failing_dimm_group, $failing_bus, $data_side );

      my $failing_dimm_loc = sprintf("DIMM%d_%s%s%s", $failing_dimm_slot, $failing_dimm_node, $data_LorR, $failing_dimm_bus );
      $target_list[$t_cnt]->{DIMM_SUMM} .= sprintf("\n%s: MD AMO (Read) DIMM Location:  %s\n", $name_tag, $failing_dimm_loc);
      $phys_id = sprintf("$failing_dimm_loc");

    }
  }       

  elsif ($dq_type eq 'DQRS')  {
    my $address = $target_list[$t_cnt]->{DQRS_AMO_ERR_ADDR};
    $failing_bus = 1;			 #  0=X     1=Y
    my $failing_dimm_bus = ($shub_version == 2) ? "" : sprintf "_BUS_%s", (($failing_bus) ? "Y" : "X");
    my $raw_dimm_group = ($address >> $dimm_group_index) & 0x3;
    $failing_dimm_group = get_physical_group($raw_dimm_group);


    if (defined $target_list[$t_cnt]->{DQRS_AMO_SCOR})   {
      my $syndrome = $target_list[$t_cnt]->{DQRS_AMO_SSYN}; 
      $bit = get_bit_from_syndrome($syndrome);
 
      if (((($address & 0x8) == 0x8) && ($bit < 64)) || ((($address & 0x8) == 0x0) && ($bit > 63))) {
        $bit +=64;
      }
      elsif ((($address & 0x8) == 0x8) && ($bit > 63))  {
        $bit +=72;
      }

      $data_LorR = fsb_bit_to_LorR($bit, $board_type);

      $data_side = ($data_LorR) ? "RIGHT" : "LEFT";

      $failing_dtr = $dtr_offset + $data_LorR;

      if ($shub_version == 2)  {
        $failing_dimm_slot = $failing_dimm_group + ( $dimm_offset * $failing_bus);
      }
      else  {
        $failing_dimm_slot = ($failing_dimm_group * 2) + $failing_bus;
      }

      $sbe_dimm_pin = fsb_bit_to_dimm_pin($bit, $board_type);
      $target_list[$t_cnt]->{DIMM_SUMM} .= sprintf("\n%s: MD AMO (Store) DIMM Info:  Failing Addr:0x%016llX  Failing Syndrome: 0x%02llX ",
        $name_tag, $address, $syndrome );
      $target_list[$t_cnt]->{DIMM_SUMM} .= sprintf("\n%s: MD AMO (Store) DIMM Info:  Bit %d, Dimm Group %d, %s BUS %s   (SBE) ",
          $name_tag, $bit, $failing_dimm_group, ($failing_bus) ? "Y" : "X", $data_side );

      my $failing_dimm_loc = sprintf("DIMM%d_%s%s%s", $failing_dimm_slot, $failing_dimm_node,
          (($data_LorR) ? "R" : "L"), $failing_dimm_bus );
      $target_list[$t_cnt]->{DIMM_SUMM} .= sprintf("\n%s: MD AMO (Store) DIMM Location:  %s   DIMM pin: %d\n", 
	  $name_tag, $failing_dimm_loc, $sbe_dimm_pin);
      $phys_id = sprintf("$failing_dimm_loc");

    }


    if (defined $target_list[$t_cnt]->{DQRS_AMO_RCOR})   {
      my $syndrome = $target_list[$t_cnt]->{DQRS_AMO_RSYN}; 
      $bit = get_bit_from_syndrome($syndrome);
 
      if (((($address & 0x8) == 0x8) && ($bit < 64)) || ((($address & 0x8) == 0x0) && ($bit > 63))) {
        $bit +=64;
      }
      elsif ((($address & 0x8) == 0x8) && ($bit > 63))  {
        $bit +=72;
      }

      $data_LorR = fsb_bit_to_LorR($bit, $board_type);

      $data_side = ($data_LorR) ? "RIGHT" : "LEFT";

      $failing_dtr = $dtr_offset + $data_LorR;

      if ($shub_version == 2)  {
        $failing_dimm_slot = $failing_dimm_group + ( $dimm_offset * $failing_bus);
      }
      else  {
        $failing_dimm_slot = ($failing_dimm_group * 2) + $failing_bus;
      }

      $sbe_dimm_pin = fsb_bit_to_dimm_pin($bit, $board_type);
      $target_list[$t_cnt]->{DIMM_SUMM} .= sprintf("\n%s: MD AMO (Read) DIMM Info:  Failing Addr:0x%016llX  Failing Syndrome: 0x%02llX ",
        $name_tag, $address, $syndrome );
      $target_list[$t_cnt]->{DIMM_SUMM} .= sprintf("\n%s: MD AMO (Read) DIMM Info:  Bit %d, Dimm Group %d, %s BUS %s   (SBE) ",
          $name_tag, $bit, $failing_dimm_group, ($failing_bus) ? "Y" : "X", $data_side );

      my $failing_dimm_loc = sprintf("DIMM%d_%s%s%s", $failing_dimm_slot, $failing_dimm_node,
          (($data_LorR) ? "R" : "L"), $failing_dimm_bus );
      $target_list[$t_cnt]->{DIMM_SUMM} .= sprintf("\n%s: MD AMO (Read) DIMM Location:  %s   DIMM pin: %d\n", 
	  $name_tag, $failing_dimm_loc, $sbe_dimm_pin);
      $phys_id = sprintf("$failing_dimm_loc");

    }


    if (defined $target_list[$t_cnt]->{DQRS_AMO_SUNC})   {
      my $syndrome = $target_list[$t_cnt]->{DQRS_AMO_SSYN}; 
      my $mbe_dimm_msg = get_mbe_chip_side($syndrome);
      $data_side = ($mbe_dimm_msg =~ /Right Side/) ? "RIGHT" : ($mbe_dimm_msg =~ /Left Side/) ? "LEFT" : "either side";
      $data_LorR = ($mbe_dimm_msg =~ /Right Side/) ? "R" : ($mbe_dimm_msg =~ /Left Side/) ? "L" : "??";
 
      if ($shub_version == 2)  {
        $failing_dimm_slot = $failing_dimm_group + ( $dimm_offset * $failing_bus);
      }
      else  {
        $failing_dimm_slot = ($failing_dimm_group * 2) + $failing_bus;
      }

      $target_list[$t_cnt]->{DIMM_SUMM} .= sprintf("\n%s: MD AMO (Store) DIMM Info:  Failing Addr:0x%016llX  Failing Syndrome: 0x%02llX ",
        $name_tag, $address, $syndrome );
      $target_list[$t_cnt]->{DIMM_SUMM} .= sprintf("\n%s: MD AMO (Store) DIMM Info:  Dimm Group %d, %s BUS %s   (MBE) ",
          $name_tag, $failing_dimm_group, $failing_bus, $data_side );

      my $failing_dimm_loc = sprintf("DIMM%d_%s%s%s", $failing_dimm_slot, $failing_dimm_node, $data_LorR, $failing_dimm_bus );
      $target_list[$t_cnt]->{DIMM_SUMM} .= sprintf("\n%s: MD AMO (Store) DIMM Location:  %s\n", $name_tag, $failing_dimm_loc);
      $phys_id = sprintf("$failing_dimm_loc");

    }


    if (defined $target_list[$t_cnt]->{DQRS_AMO_RUNC})   {
      my $syndrome = $target_list[$t_cnt]->{DQRS_AMO_RSYN}; 
      my $mbe_dimm_msg = get_mbe_chip_side($syndrome);
      $data_side = ($mbe_dimm_msg =~ /Right Side/) ? "RIGHT" : ($mbe_dimm_msg =~ /Left Side/) ? "LEFT" : "either side";
      $data_LorR = ($mbe_dimm_msg =~ /Right Side/) ? "R" : ($mbe_dimm_msg =~ /Left Side/) ? "L" : "??";
 
      if ($shub_version == 2)  {
        $failing_dimm_slot = $failing_dimm_group + ( $dimm_offset * $failing_bus);
      }
      else  {
        $failing_dimm_slot = ($failing_dimm_group * 2) + $failing_bus;
      }

      $target_list[$t_cnt]->{DIMM_SUMM} .= sprintf("\n%s: MD AMO (Read) DIMM Info:  Failing Addr:0x%016llX  Failing Syndrome: 0x%02llX ",
        $name_tag, $address, $syndrome );
      $target_list[$t_cnt]->{DIMM_SUMM} .= sprintf("\n%s: MD AMO (Read) DIMM Info:  Dimm Group %d, %s BUS %s   (MBE) ",
          $name_tag, $failing_dimm_group, $failing_bus, $data_side );

      my $failing_dimm_loc = sprintf("DIMM%d_%s%s%s", $failing_dimm_slot, $failing_dimm_node, $data_LorR, $failing_dimm_bus );
      $target_list[$t_cnt]->{DIMM_SUMM} .= sprintf("\n%s: MD AMO (Read) DIMM Location:  %s\n", $name_tag, $failing_dimm_loc);
      $phys_id = sprintf("$failing_dimm_loc");

    }

  }

  $fru_list[$fru_count-1]->{REASON} .= "Suspect $phys_id.  ";

  if ($shub_version == 2)  {
    $sh2_target_list[$sh2_cnt]->{DIMM_SUMM} .= $target_list[$t_cnt]->{DIMM_SUMM};
  }

}


#---------------------------------------------------------------------------
#
#  This large, fairly clumsy function is to generate DIMM location info from
#  Shub 2 MD reported AMO memory errors.  
#
#  Inputs:	$dq_type - string, either DQRS or DQLS
#               $name_tag - string indicating which nasid we're on
#               $board_type - string; either IP63, IP73, or IP65
#
#
#---------------------------------------------------------------------------

sub sh2_decode_amo_dimm_info
{
  print "\n In sh2_decode_amo_dimm_info subroutine\n" if ($verbose_mode == 1);;
  my $dq_type = "$_[0]";
  my $name_tag = "$_[1]";
  my $board_type = "$_[2]";
  my $sbe_dimm_pin;
  my $dtr_offset = 0;
  my $phys_id = "Indeterminate DIMM location";		# this should get replaced if valid AMO info is found
  my $dimm_node = 0;
  my $amo_fru_entry = $fru_count - 1;			# default
  my $read_error = 1;
  my $amo_src = "the local PI";
  my $dimm_offset = ($board_type =~ /IP81/) ? 2 : 3;

  my $failing_dimm_node = "";

  if ($dq_type eq 'DQLS')  {
    $amo_fru_entry = $sh2_target_list[$sh2_cnt]->{DQLS_AMO_FRU};
    my $address = $sh2_target_list[$sh2_cnt]->{DQLS_AMO_ERR_ADDR};
    $failing_bus = 0;			 #  0=X     1=Y
    my $failing_dimm_bus = ($shub_version == 2) ? "" : sprintf "_BUS_%s", (($failing_bus) ? "Y" : "X");
    my $raw_dimm_group = ($address >> $dimm_group_index) & 0x3;
    $failing_dimm_group = get_physical_group($raw_dimm_group);

    if (defined $sh2_target_list[$sh2_cnt]->{DQLS_AMO_SCOR})   {
      $read_error = 0;
      my $syndrome = $sh2_target_list[$sh2_cnt]->{DQLS_AMO_SSYN}; 
      if ($syndrome == 0xff) {
        $amo_src = "a remote node via the XN";
      }
      else {
        my $prob_syndrome = $syndrome ^ 0xff;
        $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf "\n\tProbable syndrome for AMO Write data error : 0x%X\n", $prob_syndrome;
      }
    }


    if (defined $sh2_target_list[$sh2_cnt]->{DQLS_AMO_RCOR})   {
      my $syndrome = $sh2_target_list[$sh2_cnt]->{DQLS_AMO_RSYN}; 
      $bit = get_bit_from_syndrome($syndrome);
      if (((($address & 0x8) == 0x8) && ($bit < 64)) || ((($address & 0x8) == 0x0) && ($bit > 63))) {
        $bit +=64;
      }
      elsif ((($address & 0x8) == 0x8) && ($bit > 63))  {		#  check bit
        $bit +=72;
      }

      $data_LorR = fsb_bit_to_LorR($bit, $board_type);
      $data_side = ($data_LorR) ? "RIGHT" : "LEFT";
      $failing_dimm_slot = $failing_dimm_group + ( $dimm_offset * $failing_bus);

      $sbe_dimm_pin = fsb_bit_to_dimm_pin($bit, $board_type);
      $sh2_target_list[$sh2_cnt]->{DIMM_SUMM} .= sprintf("\n%s: MD AMO (Read) DIMM Info:  Failing Addr:0x%016llX  Failing Syndrome: 0x%02llX ",
        $name_tag, $address, $syndrome );
      $sh2_target_list[$sh2_cnt]->{DIMM_SUMM} .= sprintf("\n%s: MD AMO (Read) DIMM Info:  Bit %d, Dimm Group %d, %s BUS %s   (SBE) ",
          $name_tag, $bit, $failing_dimm_group, ($failing_bus) ? "Y" : "X", $data_side );

      my $failing_dimm_loc = sprintf("DIMM%d_%s%s%s", $failing_dimm_slot, $failing_dimm_node,
          (($data_LorR) ? "R" : "L"), $failing_dimm_bus );
      $sh2_target_list[$sh2_cnt]->{DIMM_SUMM} .= sprintf("\n%s: MD AMO (Read) DIMM Location:  %s   DIMM pin: %d\n", 
	  $name_tag, $failing_dimm_loc, $sbe_dimm_pin);
      $phys_id = sprintf("$failing_dimm_loc");
    }

    if (defined $sh2_target_list[$sh2_cnt]->{DQLS_AMO_SUNC})   {
      $read_error = 0;
      my $syndrome = $sh2_target_list[$sh2_cnt]->{DQLS_AMO_SSYN}; 
      if ($syndrome == 0xff) {
        $amo_src = "a remote node via the XN";
      }
      else {
        my $prob_syndrome = $syndrome ^ 0xff;
        $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf "\n\tProbable syndrome for AMO Write data error : 0x%X\n", $prob_syndrome;
      }
    }

    if (defined $sh2_target_list[$sh2_cnt]->{DQLS_AMO_RUNC})   {
      my $syndrome = $sh2_target_list[$sh2_cnt]->{DQLS_AMO_RSYN}; 
      my $mbe_dimm_msg = get_mbe_chip_side($syndrome);
      $data_side = ($mbe_dimm_msg =~ /Right Side/) ? "RIGHT" : ($mbe_dimm_msg =~ /Left Side/) ? "LEFT" : "either side";
      $data_LorR = ($mbe_dimm_msg =~ /Right Side/) ? "R" : ($mbe_dimm_msg =~ /Left Side/) ? "L" : "??";
 
      $failing_dimm_slot = $failing_dimm_group + ( $dimm_offset * $failing_bus);

      $sh2_target_list[$sh2_cnt]->{DIMM_SUMM} .= sprintf("\n%s: MD AMO (Read) DIMM Info:  Failing Addr:0x%016llX  Failing Syndrome: 0x%02llX ",
        $name_tag, $address, $syndrome );
      $sh2_target_list[$sh2_cnt]->{DIMM_SUMM} .= sprintf("\n%s: MD AMO (Read) DIMM Info:  Dimm Group %d, %s BUS %s   (MBE) ",
          $name_tag, $failing_dimm_group, $failing_bus, $data_side );

      my $failing_dimm_loc = sprintf("DIMM%d_%s%s%s", $failing_dimm_slot, $failing_dimm_node, $data_LorR, $failing_dimm_bus );
      $sh2_target_list[$sh2_cnt]->{DIMM_SUMM} .= sprintf("\n%s: MD AMO (Read) DIMM Location:  %s\n", $name_tag, $failing_dimm_loc);
      $phys_id = sprintf("$failing_dimm_loc");
    }
  }       

  elsif ($dq_type eq 'DQRS')  {
    $amo_fru_entry = $sh2_target_list[$sh2_cnt]->{DQRS_AMO_FRU};
    my $address = $sh2_target_list[$sh2_cnt]->{DQRS_AMO_ERR_ADDR};
    $failing_bus = 1;			 #  0=X     1=Y
    my $failing_dimm_bus = ($shub_version == 2) ? "" : sprintf "_BUS_%s", (($failing_bus) ? "Y" : "X");
    my $raw_dimm_group = ($address >> $dimm_group_index) & 0x3;
    $failing_dimm_group = get_physical_group($raw_dimm_group);


    if (defined $sh2_target_list[$sh2_cnt]->{DQRS_AMO_SCOR})   {
      $read_error = 0;
      my $syndrome = $sh2_target_list[$sh2_cnt]->{DQRS_AMO_SSYN}; 
      if ($syndrome == 0xff) {
        $amo_src = "a remote node via the XN";
      }
      else {
        my $prob_syndrome = $syndrome ^ 0xff;
        $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf "\n\tProbable syndrome for AMO Write data error : 0x%X\n", $prob_syndrome;
      }
    }


    if (defined $sh2_target_list[$sh2_cnt]->{DQRS_AMO_RCOR})   {
      my $syndrome = $sh2_target_list[$sh2_cnt]->{DQRS_AMO_RSYN}; 
      $bit = get_bit_from_syndrome($syndrome);
 
      if (((($address & 0x8) == 0x8) && ($bit < 64)) || ((($address & 0x8) == 0x0) && ($bit > 63))) {
        $bit +=64;
      }
      elsif ((($address & 0x8) == 0x8) && ($bit > 63))  {		#  check bit
        $bit +=72;
      }

      $data_LorR = fsb_bit_to_LorR($bit, $board_type);
      $data_side = ($data_LorR) ? "RIGHT" : "LEFT";
      $failing_dimm_slot = $failing_dimm_group + ( $dimm_offset * $failing_bus);

      $sbe_dimm_pin = fsb_bit_to_dimm_pin($bit, $board_type);
      $sh2_target_list[$sh2_cnt]->{DIMM_SUMM} .= sprintf("\n%s: MD AMO (Read) DIMM Info:  Failing Addr:0x%016llX  Failing Syndrome: 0x%02llX ",
        $name_tag, $address, $syndrome );
      $sh2_target_list[$sh2_cnt]->{DIMM_SUMM} .= sprintf("\n%s: MD AMO (Read) DIMM Info:  Bit %d, Dimm Group %d, %s BUS %s   (SBE) ",
          $name_tag, $bit, $failing_dimm_group, ($failing_bus) ? "Y" : "X", $data_side );

      my $failing_dimm_loc = sprintf("DIMM%d_%s%s%s", $failing_dimm_slot, $failing_dimm_node,
          (($data_LorR) ? "R" : "L"), $failing_dimm_bus );
      $sh2_target_list[$sh2_cnt]->{DIMM_SUMM} .= sprintf("\n%s: MD AMO (Read) DIMM Location:  %s   DIMM pin: %d\n", 
	  $name_tag, $failing_dimm_loc, $sbe_dimm_pin);
      $phys_id = sprintf("$failing_dimm_loc");
    }


    if (defined $sh2_target_list[$sh2_cnt]->{DQRS_AMO_SUNC})   {
      $read_error = 0;
      my $syndrome = $sh2_target_list[$sh2_cnt]->{DQRS_AMO_SSYN}; 
      if ($syndrome == 0xff) {
        $amo_src = "a remote node via the XN";
      }
      else {
        my $prob_syndrome = $syndrome ^ 0xff;
        $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf "\n\tProbable syndrome for AMO Write data error : 0x%X\n", $prob_syndrome;
      }
    }


    if (defined $sh2_target_list[$sh2_cnt]->{DQRS_AMO_RUNC})   {
      my $syndrome = $sh2_target_list[$sh2_cnt]->{DQRS_AMO_RSYN}; 
      my $mbe_dimm_msg = get_mbe_chip_side($syndrome);
      $data_side = ($mbe_dimm_msg =~ /Right Side/) ? "RIGHT" : ($mbe_dimm_msg =~ /Left Side/) ? "LEFT" : "either side";
      $data_LorR = ($mbe_dimm_msg =~ /Right Side/) ? "R" : ($mbe_dimm_msg =~ /Left Side/) ? "L" : "??";
      $failing_dimm_slot = $failing_dimm_group + ( $dimm_offset * $failing_bus);

      $sh2_target_list[$sh2_cnt]->{DIMM_SUMM} .= sprintf("\n%s: MD AMO (Read) DIMM Info:  Failing Addr:0x%016llX  Failing Syndrome: 0x%02llX ",
        $name_tag, $address, $syndrome );
      $sh2_target_list[$sh2_cnt]->{DIMM_SUMM} .= sprintf("\n%s: MD AMO (Read) DIMM Info:  Dimm Group %d, %s BUS %s   (MBE) ",
          $name_tag, $failing_dimm_group, $failing_bus, $data_side );

      my $failing_dimm_loc = sprintf("DIMM%d_%s%s%s", $failing_dimm_slot, $failing_dimm_node, $data_LorR, $failing_dimm_bus );
      $sh2_target_list[$sh2_cnt]->{DIMM_SUMM} .= sprintf("\n%s: MD AMO (Read) DIMM Location:  %s\n", $name_tag, $failing_dimm_loc);
      $phys_id = sprintf("$failing_dimm_loc");
    }
  }


  if ($read_error == 1)  {
    $fru_list[$amo_fru_entry]->{REASON} .= "Suspect $phys_id.  ";
  }
  else {
    $fru_list[$amo_fru_entry]->{REASON} .= sprintf "\n\t\t\t\t\t\tData error occurred on a AMO write from %s.  ", $amo_src;
  }

}



sub fsb_bit_to_dimm_pin
{
   my $fsb_bit = $_[0];
   my $board_type = $_[1];

   if ($board_type =~ /IP57/) {
      @DoubleWordData = @IP57_MEMORY_BIT_MAP;
   }
   elsif ($board_type =~ /IP[678][1359]/) {
      @DoubleWordData = @IP63_MEMORY_BIT_MAP;
   }
   else {
      @DoubleWordData = @IP41_MEMORY_BIT_MAP;
   }

   return ($DoubleWordData[$fsb_bit][2] );

}


sub Calculate_ECC
{
   my $ecc;
   my $data = $_[0];

   $ecc = 0;

   for ($bit=0; $bit<64; $bit++)
   {
      if ( (0x1<<$bit) & $data )
      {
         $ecc ^= $ECC[$bit];
      }
   }
   
   return($ecc);
}


#-------------------------------------------------------------------------
#
#  This subroutine looks up the supplied syndrome to determine the probability
#  of which DIMM side the bits for the failing syndrome would occur on.
#
#-------------------------------------------------------------------------

sub get_mbe_chip_side {

    my $syn_to_find = $_[0];

    my $hit_cnt = 0;
    my $bit_details = "\n";
    my $left_cnt = 0;
    my $right_cnt = 0;
    my $chip = 0;
    my $synIdx = 0;
    my @CHIPKILL_CHIP_UC_SYNDROMES = ();
    my @BITS_ON_CHIP = ();

    my @SYNDROME_MASKS = ( 0x3, 0x5, 0x6, 0x7, 0x9, 0xa, 0xb, 0xc, 0xd, 0xe, 0xf );

    if ($shub_version == 2)  {
        @CHIPKILL_CHIP_UC_SYNDROMES = @SH2_CHIPKILL_CHIP_UC_SYNDROMES;
        @BITS_ON_CHIP = @SH2_BITS_ON_CHIP;
    }
    else {
        @CHIPKILL_CHIP_UC_SYNDROMES = @SH1_CHIPKILL_CHIP_UC_SYNDROMES;
        @BITS_ON_CHIP = @SH1_BITS_ON_CHIP;
    }

    for ( $chip=0; $chip<18; $chip++ ) {
       for ( $synIdx=0; $synIdx < 11; $synIdx++ ) {
          my $syn_in_list = $CHIPKILL_CHIP_UC_SYNDROMES[$chip][$synIdx];
          if ($syn_in_list == $syn_to_find)  {
            $hit_cnt++;
            my $bitMask = $SYNDROME_MASKS[$synIdx];
            $bit_details .= sprintf(" * CHIP%d%s - BITS (SYN) : ", ($chip%9), ($chip > 8) ? "L" : "R");
            my $numbits=0;
            for($b=0;$b<4;$b++) {
               if ( (0x1 << $b) & $bitMask ) {
                  $bit_details .= sprintf("%2d", $BITS_ON_CHIP[$chip][$b]);
                  $bit_details .= sprintf(" (0x%02x), ", $ECC[$BITS_ON_CHIP[$chip][$b]]);
                  $numbits++;
               }
            }
            $bit_details .= sprintf("\n");
            if ( $chip > 8 ) {
              $left_cnt++;
            }
            else {
              $right_cnt++;
            }
          }
       }
    }

    my $return_msg = sprintf " Failing DIMM side not confirmed for MBE syn 0x%x.", $syn_to_find;

    if (($left_cnt == 0) && ($right_cnt > 0))  {
        $return_msg = sprintf " Suspect DIMM for MBE syn 0x%x is confirmed on the Right Side DIMM", $syn_to_find;
    }
    elsif (($left_cnt > 0) && ($right_cnt == 0))  {
        $return_msg = sprintf " Suspect DIMM for MBE syn 0x%x is confirmed on the Left Side DIMM", $syn_to_find;
    }
    elsif ($hit_cnt == 0)  {
        $return_msg = sprintf " Syndrome \(0x%X\) did not match any dram bit-aligned MBE syndromes.", $syn_to_find;
    }
    else {
        my $left_prob = ($left_cnt * 100) / $hit_cnt;
        my $right_prob = ($right_cnt * 100) / $hit_cnt;
        $return_msg .= sprintf "\n  Left side probability: %d\%   Right side probability %d\%\n", $left_prob, $right_prob;
    }

    $return_msg .= "\n$bit_details\n";
    return($return_msg);

}

#--------------------------------------------------------------
#
#  Function to determine the physical dimm group to use when 
#  generating a dimm location.  This is supposed to handle those
#  cases when dimm groups have been remapped in the DIMM_CFG mmr's.
#
#--------------------------------------------------------------

sub get_physical_group {

  my $logical_group_val = $_[0];
  my $physical_group_val = 9;		# default, error value

  if    ($logical_group_val == $dimm_group_0_cfg) {  $physical_group_val = 0;  }
  elsif ($logical_group_val == $dimm_group_1_cfg) {  $physical_group_val = 1;  }
  elsif ($logical_group_val == $dimm_group_2_cfg) {  $physical_group_val = 2;  }
  elsif ($logical_group_val == $dimm_group_3_cfg) {  $physical_group_val = 3;  }
  
  return($physical_group_val);
  
}


###########################################################################

1;
