#
#/**************************************************************************
# *                                                                        *
# *               Copyright (C) 1989-1999 Silicon Graphics, Inc.           *
# *                                                                        *
# *  These coded instructions, statements, and computer programs  contain  *
# *  unpublished  proprietary  information of Silicon Graphics, Inc., and  *
# *  are protected by Federal copyright law.  They  may  not be disclosed  *
# *  to  third  parties  or copied or duplicated in any form, in whole or  *
# *  in part, without the prior written consent of Silicon Graphics, Inc.  *
# *                                                                        *
# **************************************************************************/
#
# @(#)$Id: mem_subs.pm,v 1.1 2002/02/20 17:25:12 beighley Exp $


#-------------------------------------------------------------------------------
#
#  FILE: mem_subs.pm
#
#  DESCRIPTION:  This file contains PERL subroutines to handle reading and/or
#     writing system "memory".  The addresses and corresponding data values
#     are contained in two arrays or in a hash structure.
#
#     When an entire hash table is processed, the entries are sorted by
#     the address (key).  This order may be different from that found
#     when moving through the hash directly.
#
#  NOTES: The "common_defs.pm" file is required when this file is used.
#
#  SUBROUTINES:  The following subroutines are contained in this file:
#     (expected parameter usage displayed; some have optional parameters)
#     mem_access($MEM_ACCESS_SMDK|$MEM_ACCESS_JTAG|$MEM_ACCESS_PROM)
#     $val  = mem_read($addr)
#     @data = mem_read_block($num, $addr)
#     @data = mem_read_group(\@addrs)
#     mem_write($addr, $value)
#     mem_write_block($addr, \@data)
#     mem_write_group(\@addr, \@data)
#
#  MODIFICATION HISTORY:
#
#     WHEN      WHO             WHAT
#     --------  --------------  -----------------------------------------
#     10/25/01  Karen Beighley  Original source code.
#     12/13/01  Karen Beighley  Minor change to mmr_write for "mem_sim_mode".
#
#-------------------------------------------------------------------------------


#-------------------------------------------------------------------------------
#  GLOBAL VARIABLES
#-------------------------------------------------------------------------------

$mem_access = $MEM_ACCESS_NONE;
@mem_access_names = ( "SMDK", "JTAG", "PROM", "NONE" );

$sim_mode = $TRUE;
%sim_mem = ( 0 => 0 );


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: mem_access()
#
#    USAGE: mem_access($MEM_ACCESS_SMDK|$MEM_ACCESS_JTAG|$MEM_ACCESS_PROM)
#
#    DESCRIPTION: The subroutine sets the memory access method for
#       future memory references.
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub mem_access
{
   my $num = $#_ + 1;
   print "\nmem_access(@_) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 1 )
   {  $emess = "Invalid parameters passed to subroutine mem_access().";
      dsp_error_and_exit($emess);
      return;
   }
   $mem_access = $_[0];
   if ( ($mem_access < 0) || ($mem_access > $MEM_ACCESS_NONE) )
   {  $emess = "Unknown access parameter passed to subroutine mem_access().";
      dsp_error_and_exit($emess);
      return;
   }

   print "access method: $mem_access = ", $mem_access_names[$mem_access], "\n" if ($debug_flag & $DEBUG_GENERAL);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: mem_read()
#
#    USAGE: $val = mem_read($addr)
#
#    DESCRIPTION: The subroutine reads the single memory location specified
#       by the "addr" parameter.
#
#    RETURN VALUE: The data value from the specified address.
#
#-------------------------------------------------------------------------------

sub mem_read
{
   my $num = $#_ + 1;
   print "\nmem_read(@_) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 1 )
   {  $emess = "Invalid parameters passed to subroutine mem_read().";
      dsp_error_and_exit($emess);
      return;
   }
   my $mem_addr = $_[0];
   my $mem_val = 0;
   printf "address = 0x%016lX\n", $mem_addr if ($debug_flag & $DEBUG_GENERAL);

   #  Support simulation/emulation envirnoment (no hardware connection)
   if ( $sim_mode )
   {  if ( exists $sim_mem{$mem_addr} )    # Address previously accessed
      {  $mem_val = $sim_mem{$mem_addr};
      }
      else                                # First access to address
      {  $sim_mem{$mem_addr} = 0;
         $mem_val = 0;
      }
      dvh_display("MEMORY CONTENTS:", \%sim_mem) if ($debug_flag & $DEBUG_DETAILS);
   }

   return( $mem_val );
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: mem_read_block()
#
#    USAGE: @data = mem_read_block($num, $addr)
#
#    DESCRIPTION: The subroutine reads a block of contiguous memory locations
#       beginning at the address specified by the "addr" parameter.
#       The address is incremented by the $MEM_ADDR_INCR value.
#
#    RETURN VALUE: The array of data values from the specified addresses.
#
#-------------------------------------------------------------------------------

sub mem_read_block
{
   my $num = $#_ + 1;
   print "\nmem_read_block(@_) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 2 )
   {  $emess = "Invalid parameters passed to subroutine mem_read_block().";
      dsp_error_and_exit($emess);
      return;
   }
   my $mem_num  = $_[0];
   my $mem_addr = $_[1];
   my @mem_val = ();
   if ( $debug_flag & $DEBUG_GENERAL )
   {  printf "number  = %lu\n", $mem_num;
      printf "address = 0x%016lX\n", $mem_addr;
   }

   #  Support simulation/emulation envirnoment (no hardware connection)
   if ( $sim_mode )
   {  for ($i = 0; $i < $mem_num; $i++)
      {  $mem_val[$i] = mem_read($mem_addr);
         $mem_addr += $MEM_ADDR_INCR
      }
   }

   return( @mem_val );
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: mem_read_group()
#
#    USAGE: @data = mem_read_group(\@addrs)
#
#    DESCRIPTION: The subroutine reads a group of memory locations 
#       specified by the "addr" parameter.  The array is moved thru
#       in the order given.
#
#    RETURN VALUE: The array of data values from the specified addresses.
#
#-------------------------------------------------------------------------------

sub mem_read_group
{
   my $num = $#_ + 1;
   print "\nmem_read_group(@_) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 1 )
   {  $emess = "Invalid parameters passed to subroutine mem_read_group().";
      dsp_error_and_exit($emess);
      return;
   }
   my $addr_ref = $_[0];
   if ( ref($addr_ref) ne 'ARRAY' )
   {  $emess = "Invalid parameter passed to subroutine mem_read_block().";
      $emess .= "\nFirst parameter must be an ARRAY reference.";
      dsp_error_and_exit($emess);
   }
   my @addrs    = @$addr_ref;
   my @mem_val  = ();
   my $ind      = 0;
   printf "address = 0x%016lX\n", @addrs if ( $debug_flag & $DEBUG_GENERAL );

   #  Support simulation/emulation envirnoment (no hardware connection)
   if ( $sim_mode )
   {  foreach $a (@addrs)
      {  $mem_val[$ind] = mem_read($a);
         $ind++;
      }
   }

   return( @mem_val );
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: mem_write()
#
#    USAGE: mem_write($addr, $value)
#
#    DESCRIPTION: The subroutine writes the single memory location specified
#       by the "addr" parameter with "value".
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub mem_write
{
   my $num = $#_ + 1;
   print "\nmem_write(@_) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 2 )
   {  $emess = "Invalid parameters passed to subroutine mem_write().";
      dsp_error_and_exit($emess);
      return;
   }
   my $mem_addr = $_[0];
   my $mem_data = $_[1];
   if ( $debug_flag & $DEBUG_GENERAL )
   {  printf "address = 0x%016lX\n", $mem_addr;
      printf "data    = 0x%016lX\n", $mem_data;
   }

   #  Support simulation/emulation envirnoment (no hardware connection)
   if ( $sim_mode )
   {  
      $sim_mem{$mem_addr} = $mem_data;

#      if ( exists $sim_mem{$mem_addr} )    # Address previously accessed
#      {  $sim_mem{$mem_addr} .= $mem_data;
#      }
#      else                                # First access to address
#      {  $sim_mem{$mem_addr} = $mem_data;
#      }

      dvh_display("MEMORY CONTENTS:", \%sim_mem) if ($debug_flag & $DEBUG_DETAILS);
   }
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: mem_write_block()
#
#    USAGE: mem_write_block($addr, \@data)
#
#    DESCRIPTION: The subroutine writes a block of contiguous memory
#       locations beginning at the address specified by the "addr" parameter.
#       The address is incremented by the $MEM_ADDR_INCR value.
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub mem_write_block
{
   my $num = $#_ + 1;
   print "\nmem_write_block(@_) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 2 )
   {  $emess = "Invalid parameters passed to subroutine mem_write_block().";
      dsp_error_and_exit($emess);
      return;
   }
   my $mem_addr = $_[0];
   my $data_ptr = $_[1];
   if ( ref($data_ptr) ne 'ARRAY' )
   {  $emess = "Invalid parameter passed to subroutine mem_write_block().";
      $emess .= "\nSecond parameter must be an ARRAY reference.";
      dsp_error_and_exit($emess);
   }
   my @mem_data = @$data_ptr;
   if ( $debug_flag & $DEBUG_GENERAL )
   {  printf "address = 0x%016lX\n", $mem_addr;
      printf "data    = @mem_data\n";
   }

   #  Support simulation/emulation envirnoment (no hardware connection)
   if ( $sim_mode )
   {  foreach $val (@mem_data)
      {  mem_write($mem_addr, $val);
         $mem_addr += $MEM_ADDR_INCR;
      }
      dvh_display("MEMORY CONTENTS:", \%sim_mem) if ($debug_flag & $DEBUG_DETAILS);
   }
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: mem_write_group()
#
#    USAGE: mem_write_group(\@addr, \@data)
#
#    DESCRIPTION: The subroutine writes a group of memory locations 
#       specified by the "addr" parameter.  The array is moved thru
#       in the order given, and the data value at the corresponding
#       array index of "data" is used.
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub mem_write_group
{
   my $num = $#_ + 1;
   print "\nmem_write_group(@_) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 2 )
   {  $emess = "Invalid parameters passed to subroutine mem_write_group().";
      dsp_error_and_exit($emess);
      return;
   }
   my $addr_ptr  = $_[0];
   if ( ref($addr_ptr) ne 'ARRAY' )
   {  $emess = "Invalid parameter passed to subroutine mem_write_group().";
      $emess .= "\nFirst parameter must be an ARRAY reference.";
      dsp_error_and_exit($emess);
   }
   my @mem_addr = @$addr_ptr;
   my $data_ptr = $_[1];
   if ( ref($data_ptr) ne 'ARRAY' )
   {  $emess = "Invalid parameter passed to subroutine mem_write_group().";
      $emess .= "\nSecond parameter must be an ARRAY reference.";
      dsp_error_and_exit($emess);
   }
   my @mem_data = @$data_ptr;
   if ( $#mem_addr != $#mem_data )
   {  $emess = "Invalid parameters passed to subroutine mem_write_group().";
      $emess .= "\nThe number of elements in each array must be equal.";
      dsp_error_and_exit($emess);
   }
   if ( $debug_flag & $DEBUG_GENERAL )
   {  printf "address = @mem_addr\n";
      printf "data    = @mem_data\n";
   }
   my $ind = 0;

   #  Support simulation/emulation envirnoment (no hardware connection)
   if ( $sim_mode )
   {  foreach $val (@mem_data)
      {  mem_write($mem_addr[$ind], $val);
         $ind++;
      }
      dvh_display("MEMORY CONTENTS:", \%sim_mem) if ($debug_flag & $DEBUG_DETAILS);
   }
}


#-------------------------------------------------------------------------------

1;
