#
#/**************************************************************************
# *                                                                        *
# *               Copyright (C) 1989-1999 Silicon Graphics, Inc.           *
# *                                                                        *
# *  These coded instructions, statements, and computer programs  contain  *
# *  unpublished  proprietary  information of Silicon Graphics, Inc., and  *
# *  are protected by Federal copyright law.  They  may  not be disclosed  *
# *  to  third  parties  or copied or duplicated in any form, in whole or  *
# *  in part, without the prior written consent of Silicon Graphics, Inc.  *
# *                                                                        *
# **************************************************************************/
#
# @(#)$Id: mmr_subs.pm,v 1.18 2007/01/23 22:45:17 beighley Exp $

#-------------------------------------------------------------------------------
#
#  FILE: mmr_subs.pm
#
#  DESCRIPTION:  This file contains PERL subroutines to handle reading and/or
#     writing system memory mapped registers (MMRs).  The addresses and 
#     corresponding data values are contained in two arrays or in a hash 
#     structure.
#
#     When an entire hash table is processed, the entries are sorted by
#     the address (key).  This order may be different from that found
#     when moving through the hash directly.
#
#  NOTES: The "common_defs.pm" file is required when this file is used.
#
#  SUBROUTINES:  The following subroutines are contained in this file:
#     (expected parameter usage displayed; some have optional parameters)
#     mmr_access($MMR_ACCESS_SMDK|$MMR_ACCESS_JTAG|$MMR_ACCESS_PROM)
#     mmr_set_timeout($value)
#     $val  = mmr_get_timeout()
#     $val  = mmr_read($addr)
#     @data = mmr_read_block($num, $addr)
#     @data = mmr_read_group(\@addrs)
#     mmr_write($addr, $value)
#     mmr_write_block($addr, \@data)
#     mmr_write_group(\@addr, \@data)
#     The following subroutines support register data structures:
#     $val  = reg_mmr_read($reg_ref) -- read MMR using register structure
#     reg_mmr_write($reg_ref) -- write MMR using register structure
#     reg_list_mmr_read($reg_list) -- read list of MMR register structures
#     reg_list_mmr_write($reg_list) -- write list of MMR register structures
#     mmr_list_read_and_save() -- read MMRs and save/display values
#
#  MODIFICATION HISTORY:
#
#     WHEN      WHO             WHAT
#     --------  --------------  -----------------------------------------
#     11/01/01  Karen Beighley  Original source code.
#     12/06/01  Karen Beighley  Added support for MMR operations via JTAG.
#                               Added subroutines to set/get timeout value.
#     12/11/01  Karen Beighley  Added subroutine: mmr_target().
#     12/13/01  Karen Beighley  Minor change to mmr_write for "mmr_sim_mode".
#     02/25/02  Karen Beighley  Added use of common error function pointer.
#     03/19/02  Karen Beighley  Added function: mmr_list_read_and_save()
#     11/21/02  Karen Beighley  Fixed function: mmr_list_read_and_save()
#     12/19/02  Karen Beighley  Apply mask to mmr_read values to force
#                                  conversion of the scalar to 64-bit number.
#     10/16/03  Karen Beighley  Changed output function (format) used in
#                                  function: mmr_list_read_and_save().
#     12/10/03  Karen Beighley  Added support for register TYPE field  (check
#                                  and call pktgen_mmr subroutines when needed).
#     02/10/04  Karen Beighley  Modified MMR target to call JTAG target.
#     08/12/04  Karen Beighley  Minor change to debugging output.
#     11/15/06  Karen Beighley  Changed how MMR data is returned from C-library.
#     01/23/07  Karen Beighley  Fixed uninitialized variable (total) in mmr_read_group().
#
#-------------------------------------------------------------------------------

use reg_subs;
use pktgen;
use dvah_subs;

$MMR_TARGET_MAX      = 16;

$MMR_LIST_READ_MAX   = 256;
$MMR_LIST_WRITE_MAX  = 128;

#$MMR_LIST_READ_MAX   = 64;
#$MMR_LIST_WRITE_MAX  = 64;

#-------------------------------------------------------------------------------
#  GLOBAL VARIABLES
#-------------------------------------------------------------------------------

$mmr_access = $MMR_ACCESS_NONE;
@mmr_access_names = ( "SMDK", "JTAG", "PROM", "NONE" );

$mmr_sim_mode = $TRUE;
%mmr_sim_mem  = ( 0 => 0 );

$mmr_timeout = 10;

my $mmr_num_targets = 0;
my @mmr_targets = ();


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: reg_mmr_read()
#
#    USAGE: $val = reg_mmr_read($reg_ref)
#
#    DESCRIPTION: The subroutine reads the single MMR location specified
#       by the register reference (from reg_create_mmr_inst).
#
#    RETURN VALUE: The data value from the specified address.
#
#-------------------------------------------------------------------------------

sub reg_mmr_read
{  print "\nreg_mmr_read(@_)\n"  if ($debug_flag & $DEBUG_FUNCTIONS);
   my $reg_ref = $_[0];
   my $val = ( (reg_type($reg_ref) == $REG_TYPE_PKTGEN) ? 
                pktgen_mmr_read(reg_addr($reg_ref)) :
                mmr_read(reg_addr($reg_ref)) );
   reg_set_value($reg_ref, $val);
   return($val);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: reg_mmr_write()
#
#    USAGE: $val = reg_mmr_write($reg_ref)
#
#    DESCRIPTION: The subroutine writes the single MMR location specified
#       by the register reference (from reg_create_mmr_inst) with the
#       data values from the structure.
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub reg_mmr_write
{  print "\nreg_mmr_write(@_)\n"  if ($debug_flag & $DEBUG_FUNCTIONS);
   my $reg_ref = $_[0];

  if ( reg_type($reg_ref) == $REG_TYPE_PKTGEN )
  {   pktgen_mmr_write(reg_addr($reg_ref), reg_get_value($reg_ref));
  }
  else
  {   mmr_write(reg_addr($reg_ref), reg_get_value($reg_ref));
  }
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: reg_list_mmr_read()
#
#    USAGE: reg_list_mmr_read($reg_list)
#
#    DESCRIPTION: The subroutine reads the list of MMR locations specified
#       by the register reference array (from reg_create_mmr_list).
#
#    RETURN VALUE: Empty string if successful; message string if error.
#
#-------------------------------------------------------------------------------

sub reg_list_mmr_read
{
   print "\nreg_list_mmr_read( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   return("Invalid number of parameters given to function.") if ( $#_ != 0 );
   my $regs  = $_[0];  # Reference to array with register pointers
   my ($reg_ref, $i, $pg_flag);

   my $num  = $#$regs + 1;
   my @addr = ();
   my @data = ();
   my ($val, $src_id, $src_chiplet, $dst_node_id, $dst_chiplet);

   #
   #  Determine if any MMRs require packet generator
   #
   for ($i = 0, $pg_flag = $FALSE; (($i < $num) && !$pg_flag); $i++)
   {  if ( reg_type($$regs[$i]) == $REG_TYPE_PKTGEN )
      {  $pg_flag = $TRUE;
      }
   }

   #
   #  Need packet generator for some of the MMRs (determined by address range)
   #     Read MMRs : single MMR if packet generator needed OR blocks of "normal" MMRs
   #
   if ( $pg_flag )
   {  $i--;
      ($src_id, $src_chiplet, $dst_node_id, $dst_chiplet) = pktgen_setup('mmr', reg_addr($$regs[$i]));
      my @tmp  = ();
      my $n    = 0;
      for ($i = 0; $i < $num; $i++)
      {  $reg_ref  = $$regs[$i];
         if ( reg_type($reg_ref) == $REG_TYPE_PKTGEN )
         {  if ( $n > 0 )
            {  @tmp = mmr_read_group(\@addr);  # Read list of addresses
               push(@data, @tmp);              # Add to end of data array
               $n    = 0;                      # Clear number of entries in list
            }
            $val = pktgen_mmr_read(reg_addr($reg_ref), $src_id);
            push(@data, $val);                 # Add to end of data array
         }
         else
         {  $addr[$n++] = reg_addr($reg_ref);
         }
      }
      if ( $n > 0 )
      {  @tmp = mmr_read_group(\@addr);  # Read list of addresses
         push(@data, @tmp);              # Add to end of data array
      }
   }
   #
   #  Build the MMR address list and read as a single block
   #
   else
   {
      for ($i = 0; $i < $num; $i++)
      {  $addr[$i] = reg_addr($$regs[$i]);
      }
      @data = mmr_read_group(\@addr);
   }


   #
   #  Save the MMR value, and write the value and field(s) to the output "file"
   #
   for ($i = 0; $i < $num; $i++)
   {  reg_set_value($$regs[$i], $data[$i]);
   }
   return("");
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: reg_list_mmr_write()
#
#    USAGE: reg_list_mmr_write($reg_list)
#
#    DESCRIPTION: The subroutine writes the list of MMR locations specified
#       by the register reference array (from reg_create_mmr_list).
#
#    RETURN VALUE: Empty string if successful; message string if error.
#
#-------------------------------------------------------------------------------

sub reg_list_mmr_write
{
   print "\nreg_list_mmr_write( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   return("Invalid number of parameters given to function.") if ( $#_ != 0 );
   my $regs  = $_[0];  # Reference to array with register pointers
   my ($reg_ref, $i, $pg_flag);

   my $num  = $#$regs + 1;
   my @addr = ();
   my @data = ();
   my ($val, $src_id, $src_chiplet, $dst_node_id, $dst_chiplet);

   #
   #  Determine if any MMRs require packet generator
   #
   for ($i = 0, $pg_flag = $FALSE; (($i < $num) && !$pg_flag); $i++)
   {  if ( reg_type($$regs[$i]) == $REG_TYPE_PKTGEN )
      {  $pg_flag = $TRUE;
      }
   }

   #
   #  Need packet generator for some of the MMRs (determined by address range)
   #     Write MMRs : single MMR if packet generator needed OR blocks of "normal" MMRs
   #
   if ( $pg_flag )
   {  $i--;
      ($src_id, $src_chiplet, $dst_node_id, $dst_chiplet) = pktgen_setup('mmr', reg_addr($$regs[$i]));
      my $n   = 0;
      for ($i = 0; $i < $num; $i++)
      {  $reg_ref  = $$regs[$i];
         if ( reg_type($reg_ref) == $REG_TYPE_PKTGEN )
         {  if ( $n > 0 )
            {  mmr_write_group(\@addr, \@data); # Write list of addresses
               @addr = ();                      # Clear address list
               @data = ();                      # Clear data list
               $n    = 0;                       # Clear number of entries in list
            }
            pktgen_mmr_write(reg_addr($reg_ref), reg_get_value($reg_ref), $src_id);
         }
         else
         {  $addr[$n] = reg_addr($reg_ref);
            $data[$n] = reg_get_value($reg_ref);
            $n++;
         }
      }
      if ( $n > 0 )
      {  mmr_write_group(\@addr, \@data);  # Write list of addresses
      }
   }
   #
   #  Build the MMR address/data lists, and write as a single block
   #
   else
   {
      for ($i = 0; $i < $num; $i++)
      {  $reg_ref  = $$regs[$i];
         $addr[$i] = reg_addr($reg_ref);
         $data[$i] = reg_get_value($reg_ref);
      }
      mmr_write_group(\@addr, \@data);
   }

   #
   #  Build the MMR address and data lists
   #

   @addr = ();
   @data = ();
   return("");
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: mmr_access()
#
#    USAGE: mmr_access($access_mode)
#           mmr_access($access_mode, $flag)
#
#    DESCRIPTION: The subroutine sets the memory mapped register (MMR) access 
#       method for future MMR references. 
#
#       The "access_mode" parameter can be: $MMR_ACCESS_JTAG or $MMR_ACCESS_NONE
#       Unsupported modes are: $MMR_ACCESS_SMDK and $MMR_ACCESS_PROM
#
#       For the $MMR_ACCESS_JTAG access mode, the optional "flag" parameter
#       determines if low-level L1 MMR commands are available for use.  If the 
#       "flag" parameter is $TRUE (non-zero), then low-level L1 MMR commands 
#       can be used (if the system controller is used for JTAG).
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub mmr_access
{  my $num = $#_ + 1;
   print "\nmmr_access(@_) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   #
   #  Extract and check the subroutine parameters
   #
   my $flag = $TRUE;
   if ( ($num == 2) && defined($_[1]) )
   {  $flag = $_[1];     # User specified value for flag
   }
   elsif ( $num != 1 )
   {  $emess = "Invalid parameters passed to subroutine mmr_access().";
      &$comm_err_fun($emess);
      return;
   }
   $mmr_access = $_[0];  # User specified MMR access mechanism
   if ( ($mmr_access < 0) || ($mmr_access > $MMR_ACCESS_NONE) )
   {  $emess = "Unknown access mode passed to subroutine mmr_access().";
      &$comm_err_fun($emess);
      return;
   }

   #
   #  Initialize the specified MMR access mechanism
   #
   if ( $mmr_access == $MMR_ACCESS_JTAG )
   {  $mmr_sim_mode = $FALSE;
      pcf_jtag_set_ctlr_flag($flag); # Set the library controller flag
   }
   elsif ( $mmr_access == $MMR_ACCESS_NONE )
   {  $mmr_sim_mode = $TRUE;
   }
   else # if ( ($mmr_access == $MMR_ACCESS_SMDK) || ($mmr_access == $MMR_ACCESS_PROM) )
   {  $emess = "Unsupported access mode passed to subroutine mmr_access().";
      &$comm_err_fun($emess);
      return;
   }

   print "access method: $mmr_access = ", $mmr_access_names[$mmr_access], "\n" if ($debug_flag & $DEBUG_GENERAL);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: mmr_target()
#
#    USAGE: mmr_target($SELECT|$DESELECT, $brick, $board, $chip)
#           mmr_target($SELECT|$DESELECT, $brick, $board, $chip,
#                                         $brick, $board, $chip,
#                                         ...
#                                         $brick, $board, $chip)
#
#    DESCRIPTION: The subroutine will define the target of future MMR
#       operations.  A specific chip (on a board within a brick) can be
#       selected or deselected.
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub mmr_target
{  
   print "\nmmr_target(@_) : $#_\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $#_ < 3   )
   {  $emess = "Invalid parameters passed to subroutine mmr_target().";
      &$comm_err_fun($emess);
      return;
   }
   my ($op, $i, $ref);

   #
   #  Build the MMR target list
   #
   $op = shift(@_);
   if ( $op == $SELECT )
   {
      $mmr_targets = mmr_target_list_create( \@_ );
   }
   elsif ( $op == $DESELECT )
   {  $i = 0;
      while ( ($i+2) <= $#_ )
      {  $mname = $_[$i];    # Module/brick name ("001c01")
         $bname = $_[$i+1];  # Board name within module ("NODE0")
         $cname = $_[$i+2];  # Chip name on board ("SHUB0")
         mmr_target_list_delete_entry($mmr_targets, $mname, $bname, $cname);
         $i += 3;
      }
   }
   else
   {  $emess = "Invalid selection parameter passed to subroutine mmr_target().";
      &$comm_err_fun($emess);
      return;
   }

   return if ( $#$mmr_targets < 0 );

   if ( $debug_flag & $DEBUG_GENERAL )
   {  mmr_target_list_dump($mmr_targets);
   }

   #
   #  Pass the MMR target information to the appropriate library
   #
   if ( $mmr_access == $MMR_ACCESS_JTAG  )
   {
      $ref = $$mmr_targets[0];
      pcf_jtag_mmr_target($SELECT, $ref, $ecode, $emess);
      if ( $ecode < 0 )
      {  &$comm_err_fun($emess);
         return;
      }
      for ($i = 1; $i <= $#$mmr_targets; $i++)
      {  $ref = $$mmr_targets[$i];
         pcf_jtag_mmr_target($APPEND, $ref, $ecode, $emess);
         if ( $ecode < 0 )
         {  &$comm_err_fun($emess);
            return;
         }
         jtag_target($$ref[$DETAIL_IND_MLOC], $$ref[$DETAIL_IND_MNUM], $$ref[$DETAIL_IND_CNUM]);
      }
      $ref = $$mmr_targets[0];
      jtag_target($$ref[$DETAIL_IND_MLOC], $$ref[$DETAIL_IND_MNUM], $$ref[$DETAIL_IND_CNUM]);
   }

   $mmr_num_targets = $#$mmr_targets + 1;
   return( $mmr_num_targets );
}


#-------------------------------------------------------------------------------

sub mmr_target_get_num
{
   return( $#$mmr_targets + 1 );
}


#-------------------------------------------------------------------------------

sub mmr_target_list_get_detail
{
   print "\nmmr_target_list_get_detail( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $ind   = $_[0];  # Index of target in the list
   my $detail = $_[1];  # Index of detail for the specified target

   return if ( $ind > $#$mmr_targets );
   my $ref = $$mmr_targets[$ind];
   return( $$ref[$detail] );
}


#-------------------------------------------------------------------------------

sub mmr_target_list_create
{
   print "\nmmr_target_list_create( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   return if ( $#_ != 0 );
   my $nref = $_[0];  # Reference to array with target names
   my ($i, $n, $mname, $bname, $cname, $mref, $dref);
   my @ptrs = ();

   #
   #  Process the name array: extract module/brick, board and chip names
   #
   $i = 0;
   $n = 0;
   while ( ($i+2) <= $#$nref )
   {
      $mname = $$nref[$i];    # Module/brick name ("001c01")
      $bname = $$nref[$i+1];  # Board name within module ("NODE0")
      $cname = $$nref[$i+2];  # Chip name on board ("SHUB0")

      #
      #  Determine if the module name is valid for the configuration
      #
      if ( !(exists($modules{$mname})) )
      {  $emess = "Unknown module passed to subroutine mmr_target().";
         $emess .= "\nModule: $mname";
         $emess .= "\nModule not found in current configuration.";
         &$comm_err_fun($emess);
         return;
      }
      $mref = $modules{$mname};

      #
      #  Extract the scan module details for the specific board/chip
      #
      $dref = scan_module_chip_details($mref, $bname, $cname, 'logical');
      if ( !defined($dref) )
      {  $emess = "Unknown target board/chip passed to subroutine mmr_target().";
         $emess .= "\nModule: $mname";
         $emess .= "\nBoard:  $bname";
         $emess .= "\nChip:   $cname";
         $emess .= "\nLocation not found in current configuration.";
         &$comm_err_fun($emess);
         return;
      }
      $ptrs[$n] = $dref;   

#     printf("%03u : %-20s = $dref = ( @$dref )\n", $n, $$dref[$DETAIL_IND_NAME]);

      $i += 3;
      $n++;   
   }

   return(\@ptrs);
}


#-------------------------------------------------------------------------------

sub mmr_target_list_delete_entry
{
   print "\nmmr_target_list_delete_entry( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   return if ( $#_ != 3 );
   my $trefs = $_[0];  # Reference to array with target references
   my $mname = $_[1];  # Module/brick name ("001c01")
   my $bname = $_[2];  # Board name within module ("NODE0")
   my $cname = $_[3];  # Chip name on board ("SHUB0")
   my ($i, $found, $ptr);
   my $name = $mname . ":" . $bname . ":" . $cname;

   #
   #  Search list for specified target
   #
   for ($i = 0, $found = $FALSE; (($i <= $#$trefs) && !$found); $i++)
   {  $ptr = $$trefs[$i];
      if ( $$ptr[$DETAIL_IND_NAME] eq $name )
      {  $found = $TRUE;
      }
   }

   #
   #  Remove the target from the list
   #
   if ( $found )
   {  splice(@$trefs, ($i-1), 1);
   }
}


#-------------------------------------------------------------------------------

sub mmr_target_list_calc_group_index
{
   print "\nmmr_target_list_calc_group_index( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   return if ( $#_ != 0 );
   my $trefs = $_[0];  # Reference to array with target references
   my ($i, $n, $pos, $rem, $grp, $src, $dst, $cnt);
   my @group = ();

   #
   #  Initialize all group indices to zero (0)
   #
   $n = $#$trefs + 1;
   for ($i = 0; $i < $n; $i++)
   {  $group[$i] = 0;
   }

   #
   #  Process the target list: group targets with same details
   #
   $pos = 0;   # Initial target index
   $rem = $n;  # Number of targets remaining
   $grp = 1;   # Current group index
   while ( $rem > 0 )  # continue until all targets in a "group"
   {
      $src = $$trefs[$pos];  # Extract pointer for first target
      $group[$pos] = $grp;   # Store group index for first target
      $cnt = 1;              # Number of targets in this group

      #
      #  Move thru remaining targets looking for matching details
      #
      for ($i=($pos+1); $i < $n; $i++)
      {  
         next if ( $group[$i] );

         $dst = $$trefs[$i];
         if ( same_target_location($src, $dst) )
         {  $group[$i] = $grp;
            $cnt++;
         }
      }
      $rem -= $cnt;  # Update number of targets remaining
      for ($i=($pos+1), $pos=0; (($i < $n) && ($pos == 0)); $i++)
      {  $pos = $i if ( !$group[$i] );
      }
      $grp++;        # Increment group number
   }

   return(($grp-1), \@group);
}


#-------------------------------------------------------------------------------

sub mmr_target_list_dump
{
   return if ( $#_ != 0 );
   my $trefs = $_[0];  # Reference to array with target references
   my ($i, $num, $ptr);

   $num = $#$trefs + 1;
   printf("\nSelected MMR target(s): num=%u\n", $num);
   for ($i = 0; $i < $num; $i++)
   {  $ptr = $$trefs[$i];
      printf("%03u : %-20s = $ptr = ( @$ptr )\n", $i, $$ptr[$DETAIL_IND_NAME]);
   }
}


#-------------------------------------------------------------------------------

sub same_target_location
{
   print "\nsame_target_location( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   return($FALSE) if ( $#_ != 1 );
   my $src = $_[0];  # Reference to array with chip details in scan chain
   my $dst = $_[1];  # Reference to array with chip details in scan chain

   return($FALSE) if ( $$src[$DETAIL_IND_TYPE] ne $$dst[$DETAIL_IND_TYPE] );
   return($FALSE) if ( $$src[$DETAIL_IND_ILEN] != $$dst[$DETAIL_IND_ILEN] );
   return($FALSE) if ( $$src[$DETAIL_IND_IPOS] != $$dst[$DETAIL_IND_IPOS] );
   return($FALSE) if ( $$src[$DETAIL_IND_CLEN] != $$dst[$DETAIL_IND_CLEN] );
   return($FALSE) if ( $$src[$DETAIL_IND_CPOS] != $$dst[$DETAIL_IND_CPOS] );

   return($TRUE); 
}


#-------------------------------------------------------------------------------

sub extract_group_targets
{
   print "\nextract_group_targets( @_ )\n" ;# if ($debug_flag & $DEBUG_FUNCTIONS);

   return if ( $#_ != 1 );
   my $gnum  = $_[0];  # Group to extract targets for
   my $glist = $_[1];  # Reference to array with group indices
   my ($i, $n);
   my @list = ();

   #
   #  Move through the list and check for matching groups
   #
   for ($i = 0, $n = 0; $i <= $#$glist; $i++)
   {  if ( $$glist[$i] == $gnum )
      {  $list[$n] = $i;
         $n++;
      }
   }
   return(\@list);
}


#-------------------------------------------------------------------------------


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: mmr_set_timeout()
#
#    USAGE: mmr_set_timeout($value)
#
#    DESCRIPTION: The subroutine sets the timeout value for an MMR operation.
#       The MMR access method must be defined before calling this function.
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub mmr_set_timeout
{
   my $num = $#_ + 1;
   print "\nmmr_set_timeout(@_) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 1 )
   {  $emess = "Invalid parameters passed to subroutine mmr_set_timeout().";
      &$comm_err_fun($emess);
      return;
   }
   $mmr_timeout = $_[0];

   if ( $mmr_access == $MMR_ACCESS_JTAG )
   {  pcf_jtag_set_timeout($mmr_timeout);
   }

}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: mmr_get_timeout()
#
#    USAGE: $value = mmr_get_timeout()
#
#    DESCRIPTION: The subroutine gets the timeout value for an MMR operation.
#       The MMR access method must be defined before calling this function.
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub mmr_get_timeout
{
   my $num = $#_ + 1;
   print "\nmmr_get_timeout(@_) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 0 )
   {  $emess = "Invalid parameters passed to subroutine mmr_get_timeout().";
      &$comm_err_fun($emess);
      return;
   }

   if ( $mmr_access == $MMR_ACCESS_JTAG )
   {  $mmr_timeout = pcf_jtag_get_timeout();
   }

   return( $mmr_timeout );
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: mmr_read()
#
#    USAGE: $val = mmr_read($addr)
#           $ref = mmr_read($addr)
#
#    DESCRIPTION: The subroutine reads the single MMR location specified
#       by the "addr" parameter.
#
#    RETURN VALUE: The data value from the MMR if one target, or a reference
#       to an array of MMR values if multiple targets.
#
#-------------------------------------------------------------------------------

sub mmr_read
{
   my $num = $#_ + 1;
   print "\nmmr_read(@_) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 1 )
   {  $emess = "Invalid parameters passed to subroutine mmr_read().";
      &$comm_err_fun($emess);
      return;
   }
   my $mmr_addr = $_[0];
   my $mmr_val  = 0;
   my @data     = ();
   my $i;
   printf "address = 0x%016llX\n", $mmr_addr if ($debug_flag & $DEBUG_GENERAL);

   #  Support simulation/emulation envirnoment (no hardware connection)
   if ( $mmr_sim_mode || ($mmr_access == $MMR_ACCESS_NONE) )
   {  if ( exists $mmr_sim_mem{$mmr_addr} )    # Address previously accessed
      {  $mmr_val = $mmr_sim_mem{$mmr_addr};
      }
      else                                # First access to address
      {  $mmr_sim_mem{$mmr_addr} = 0;
         $mmr_val = 0;
      }
      dvh_display("MMR SIM CONTENTS:", \%mmr_sim_mem) if ($debug_flag & $DEBUG_DETAILS);
      for ($i = 0; $i < $mmr_num_targets; $i++)
      {  $data[$i] = $mmr_val;
      }
      $mmr_val = \@data;
   }
   elsif ( $mmr_access == $MMR_ACCESS_JTAG )
   {  pcf_jtag_mmr_read($mmr_addr, \@data, $ecode, $emess);
      if ( $ecode < 0 )
      {  &$comm_err_fun($emess);
         return;
      }
      # Force conversion of scalar to 64-bit value
      for ($i = 0; $i < $mmr_num_targets; $i++)
      {  $data[$i] &= 0xFFFFFFFFFFFFFFFF;
      }
      $mmr_val = \@data;
   }

   # 
   #  If one target: return a single MMR value
   #  If multiple targets: return an array of MMR values
   #
   return($$mmr_val[0]) if ( $mmr_num_targets == 1 );
   return($mmr_val);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: mmr_read_block()
#
#    USAGE: @data = mmr_read_block($num, $addr)
#           @ptrs = mmr_read_block($num, $addr)
#
#    DESCRIPTION: The subroutine reads a block of contiguous MMR locations
#       beginning at the address specified by the "addr" parameter.
#       The address is incremented by the $MEM_ADDR_INCR value.
#
#    RETURN VALUE: The array of data values from the specified addresses if one
#       target, or an array of references if multiple targets.  If multiple
#       tagets, then each element is a reference to an array with the data
#       for one target.
#
#-------------------------------------------------------------------------------

sub mmr_read_block
{
   my $num = $#_ + 1;
   print "\nmmr_read_block(@_) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 2 )
   {  $emess = "Invalid parameters passed to subroutine mmr_read_block().";
      &$comm_err_fun($emess);
      return;
   }
   my $mmr_num  = $_[0];
   my $mmr_addr = $_[1];
   if ( $debug_flag & $DEBUG_GENERAL )
   {  printf "number  = %lu\n", $mmr_num;
      printf "address = 0x%016llX\n", $mmr_addr;
   }

   #
   #  Build up a list of addresses
   #
   my @addr_list = ();
   for ($i = 0; $i < $mmr_num; $i++)
   {  $addr_list[$i] = $mmr_addr;
      $mmr_addr += $MEM_ADDR_INCR
   }

   #
   #  Read the MMRs using the list of addresses
   #
   return( mmr_read_group(\@addr_list) );
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: mmr_read_group()
#
#    USAGE: @data = mmr_read_group(\@addrs)
#           @ptrs = mmr_read_block($num, $addr)
#
#    DESCRIPTION: The subroutine reads a group of MMR locations 
#       specified by the "addr" parameter.  The array is moved thru
#       in the order given.
#
#    RETURN VALUE: The array of data values from the specified addresses if one
#       target, or an array of references if multiple targets.  If multiple
#       tagets, then each element is a reference to an array with the data
#       for one target.
#
#-------------------------------------------------------------------------------

sub mmr_read_group
{
   print "\nmmr_read_group(@_)\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $#_ != 0 )
   {  $emess = "Invalid parameters passed to subroutine mmr_read_group().";
      &$comm_err_fun($emess);
      return;
   }
   my $addr_ref = $_[0];
   if ( ref($addr_ref) ne 'ARRAY' )
   {  $emess = "Invalid parameter passed to subroutine mmr_read_group().";
      $emess .= "\nFirst parameter must be an ARRAY reference.";
      &$comm_err_fun($emess);
      return;
   }
   my ($ind, @addrs, $a, $i, $t, $rem, $num, $total);
   my @mmr_val  = ();
   my @ptrs     = ();

   #  Support simulation/emulation envirnoment (no hardware connection)
   if ( $mmr_sim_mode || ($mmr_access == $MMR_ACCESS_NONE) )
   {  @addrs    = @$addr_ref;
      $ind      = 0;
      printf "address = 0x%016llX\n", @addrs if ( $debug_flag & $DEBUG_GENERAL );
      foreach $a (@addrs) # process each address in the list
      {  $mmr_val[$ind] = mmr_read($a);
         $ind++;
      }
      if ( $debug_flag & $DEBUG_GENERAL )
      {   dva_display_group("MMR Read  Data", $addr_ref, \@mmr_val, $DATA_FMT_HEX64);
      }
      return( @mmr_val );
   }
   #  Access MMRs using hardware connection specified
   else
   {  pcf_jtag_mmr_read_group($addr_ref, \@mmr_val, $ecode, $emess);
      if ( $ecode < 0 )
      {  &$comm_err_fun($emess);
         return(@mmr_val);
      }
      $num   = $#$addr_ref + 1;
      $total = $mmr_num_targets * $num;
      # Force conversion of scalar to 64-bit value
      for ($i = 0; $i < $total; $i++)
      {  $mmr_val[$i] &= 0xFFFFFFFFFFFFFFFF;
      }
      if ( $mmr_num_targets == 1 )
      {  if ( $debug_flag & $DEBUG_GENERAL )
         {   dva_display_group("MMR Read  Data", $addr_ref, \@mmr_val, $DATA_FMT_HEX64);
         }
         return( @mmr_val );
      }
      for ($t = 0; $t < $mmr_num_targets; $t++)
      {  my @data  = splice(@mmr_val, 0, $num);
         $ptrs[$t] = \@data;
      }
      if ( $debug_flag & $DEBUG_GENERAL )
      {   dva_display_group("MMR Read  Data", $addr_ref, $ptrs[0], $DATA_FMT_HEX64);
      }
      @mmr_val = ();
      return(@ptrs);
   }
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: mmr_write()
#
#    USAGE: mmr_write($addr, $value)
#
#    DESCRIPTION: The subroutine writes the single MMR location specified
#       by the "addr" parameter with "value".
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub mmr_write
{
   my $num = $#_ + 1;
   print "\nmmr_write(@_) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 2 )
   {  $emess = "Invalid parameters passed to subroutine mmr_write().";
      &$comm_err_fun($emess);
      return;
   }
   my $mmr_addr = $_[0];
   my $mmr_data = $_[1];
   if ( $debug_flag & $DEBUG_GENERAL )
   {  printf "address = 0x%016llX = %lld\n", $mmr_addr, $mmr_addr;
      printf "data    = 0x%016llX = %lld\n", $mmr_data, $mmr_data;
   }

   #  Support simulation/emulation envirnoment (no hardware connection)
   if ( $mmr_sim_mode || ($mmr_access == $MMR_ACCESS_NONE) )
   {  $mmr_sim_mem{$mmr_addr} = $mmr_data;
      dvh_display("MMR SIM CONTENTS:", \%mmr_sim_mem) if ($debug_flag & $DEBUG_DETAILS);
   }
   elsif ( $mmr_access == $MMR_ACCESS_JTAG )
   {  pcf_jtag_mmr_write($mmr_addr, $mmr_data, $ecode, $emess);
      if ( $ecode < 0 )
      {  &$comm_err_fun($emess);
         return;
      }
   }
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: mmr_write_block()
#
#    USAGE: mmr_write_block($addr, \@data)
#
#    DESCRIPTION: The subroutine writes a block of contiguous MMR
#       locations beginning at the address specified by the "addr" parameter.
#       The address is incremented by the $MEM_ADDR_INCR value.
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub mmr_write_block
{
   my $num = $#_ + 1;
   print "\nmmr_write_block(@_) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 2 )
   {  $emess = "Invalid parameters passed to subroutine mmr_write_block().";
      &$comm_err_fun($emess);
      return;
   }
   my $mmr_addr = $_[0];
   my $data_ptr = $_[1];
   if ( ref($data_ptr) ne 'ARRAY' )
   {  $emess = "Invalid parameter passed to subroutine mmr_write_block().";
      $emess .= "\nSecond parameter must be an ARRAY reference.";
      &$comm_err_fun($emess);
      return;
   }
   my $mmr_num  = $#$data_ptr + 1;
   if ( $debug_flag & $DEBUG_GENERAL )
   {  printf "address = 0x%016llX\n", $mmr_addr;
      printf "number  = $mmr_num\n";
      printf "data    = ( @$data_ptr )\n";
   }

   #
   #  Build up a list of addresses
   #
   my @addr_list = ();
   for ($i = 0; $i < $mmr_num; $i++)
   {  $addr_list[$i] = $mmr_addr;
      $mmr_addr += $MEM_ADDR_INCR
   }

   #
   #  Write the list of MMRs
   #
   mmr_write_group(\@addr_list, $data_ptr);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: mmr_write_group()
#
#    USAGE: mmr_write_group(\@addr, \@data)
#
#    DESCRIPTION: The subroutine writes a group of MMR locations 
#       specified by the "addr" parameter.  The array is moved thru
#       in the order given, and the data value at the corresponding
#       array index of "data" is used.
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub mmr_write_group
{
   print "\nmmr_write_group(@_)\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $#_ != 1 )
   {  $emess = "Invalid parameters passed to subroutine mmr_write_group().";
      &$comm_err_fun($emess);
      return;
   }
   my $addr_ptr  = $_[0];
   if ( ref($addr_ptr) ne 'ARRAY' )
   {  $emess = "Invalid parameter passed to subroutine mmr_write_group().";
      $emess .= "\nFirst parameter must be an ARRAY reference.";
      &$comm_err_fun($emess);
      return;
   }
   my @mmr_addr = @$addr_ptr;
   my $data_ptr = $_[1];
   if ( ref($data_ptr) ne 'ARRAY' )
   {  $emess = "Invalid parameter passed to subroutine mmr_write_group().";
      $emess .= "\nSecond parameter must be an ARRAY reference.";
      &$comm_err_fun($emess);
      return;
   }
   my @mmr_data = @$data_ptr;
   if ( $#mmr_addr != $#mmr_data )
   {  $emess = "Invalid parameters passed to subroutine mmr_write_group().";
      $emess .= "\nThe number of elements in each array must be equal.";
      &$comm_err_fun($emess);
      return;
   }
   if ( $debug_flag & $DEBUG_GENERAL )
   {   dva_display_group("MMR Write Data", \@mmr_addr, \@mmr_data, $DATA_FMT_HEX64);
   }
   my ($ind, $val, $rem, @addr, @data, $num, $i);

   #  Support simulation/emulation envirnoment (no hardware connection)
   if ( $mmr_sim_mode || ($mmr_access == $MMR_ACCESS_NONE) )
   {  $ind = 0;
      foreach $val (@mmr_data)
      {  mmr_write($mmr_addr[$ind], $val);
         $ind++;
      }
   }
   #  Access MMRs using hardware connection specified
   elsif ( $#$addr_ptr < $MMR_LIST_WRITE_MAX )
   {  pcf_jtag_mmr_write_group($addr_ptr, $data_ptr, $ecode, $emess);
      if ( $ecode < 0 )
      {  &$comm_err_fun($emess);
         return;
      }
   }
   else
   {  $rem = $#$addr_ptr + 1;
      $ind = 0;
      while ( $rem > 0 )
      {  $num   = ( ($rem < $MMR_LIST_WRITE_MAX) ? $rem : $MMR_LIST_WRITE_MAX );
         @addrs = ();
         @data  = ();
         for ($i = 0; $i < $num; $i++)
         {  $addrs[$i] = $$addr_ptr[$ind+$i];
            $data[$i]  = $$data_ptr[$ind+$i];
         }
         pcf_jtag_mmr_write_group(\@addrs, \@data, $ecode, $emess);
         if ( $ecode < 0 )
         {  &$comm_err_fun($emess);
            return(@mmr_val);
         }
         $rem = $rem - $num;
         $ind = $ind + $num;
      }
      return( @mmr_val );
   }
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: mmr_list_read_and_save()
#
#    USAGE: mmr_list_read_and_save($brick, $board, $chip, $type, $nref)
#           mmr_list_read_and_save($brick, $board, $chip, $type, $nref, $file)
#
#    DESCRIPTION: The subroutine will target the specified chip, read each
#       MMR specified, and write the register and field values to a file or
#       standard out.  If "$file" is given, then it is opened for output.
#       The "$nref" parameter is a reference to an array of MMR names. 
#
#    RETURN VALUE: An empty string ("") if no error; a message string if error.
#
#-------------------------------------------------------------------------------

sub mmr_list_read_and_save
{
   print "\nmmr_list_read_and_save( @_ )\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   return("Invalid number of parameters given to function.") if ( ($#_ != 4) && ($#_ != 5) );
   my $brick = $_[0];  # Brick name: "001c01"
   my $board = $_[1];  # Board name within brick: "NODE0" or "NODE1"
   my $chip  = $_[2];  # Chip name on board: "SHUB0" or "SHUB1"
   my $type  = $_[3];  # Chip type: "shub_01", "pic_01", etc.
   my $nref  = $_[4];  # Reference to array with MMR names
   my $oname = (($#_ == 5) ? $_[5] : ""); # Output file name
   my ($ofp, $name, $reg_ptr, $reg_list);

   #
   #  Open an output file or use standard output (STDOUT)
   #
   if ( $oname ne "" )
   {  if ( !open(MMR_OUT_FILE, ">$oname") )
      {  $emess = "Unable to open output file for writing.";
         $emess .= "\nFile: $oname";
         return($emess);
      }
      $ofp = MMR_OUT_FILE;
   }
   else
   {  $ofp = STDOUT;
   }

   #
   #  Create a list of register structures
   #
   $reg_list = reg_create_mmr_list($type, $nref);
   if ( !defined($reg_list) )
   {  return ("Unable to create register list; unknown register in list.");
   }

   #
   #  Read the value of each register in the list
   #
   reg_list_mmr_read($reg_list);

   #
   #  Write the contents of each register to the "output" file
   #
   for ($i = 0; $i <= $#$reg_list; $i++)
   {  $reg_ptr = $$reg_list[$i];
      reg_write_value($reg_ptr, $ofp, 'all', 'hex');
   }

   if ( $oname ne "" )
   {  close(MMR_OUT_FILE);
   }

   return("");
}


#-------------------------------------------------------------------------------

1;
