# 
#/**************************************************************************
# *                                                                        *
# *               Copyright (C) 1989-1999 Silicon Graphics, Inc.           *
# *                                                                        *
# *  These coded instructions, statements, and computer programs  contain  *
# *  unpublished  proprietary  information of Silicon Graphics, Inc., and  *
# *  are protected by Federal copyright law.  They  may  not be disclosed  *
# *  to  third  parties  or copied or duplicated in any form, in whole or  *
# *  in part, without the prior written consent of Silicon Graphics, Inc.  *
# *                                                                        *
# **************************************************************************/
# 
# @(#)$Id: mmrsel.pm,v 1.1 2003/10/17 18:19:04 beighley Exp $
#-------------------------------------------------------------------------------
#
#  FILE: mmrsel.pm
#
#  DESCRIPTION:  This file contains PERL subroutines to create an GUI for a
#    user to select from a list of memory mapped registers (MMRs).  One or more 
#    MMRs can be selected, and the selection order can be specified (if important).
#    A summary name or title can be provided for the group of selected MMRs.
#
#  USAGE:  The "create_mmrsel_window" should be called with the parameters:
#            <main_win>   = handle to the main window widget
#            <chip_name>  = full-name of chip (i.e. "002c01:NODE1:SHUB1")
#            <chip_type>  = chip type/description (i.e. "shub" or "tio")
#            <desc>       = description of operation ("for MMR Access Window")
#            <mode>       = bitmap of modes that control selection behavior
#            <btn_text>   = text for the action button or "" for "Create"
#            <curr_title> = reference to current name of group (for entry)
#            <curr_ref>   = reference to array of selected (indices!)
#            <create_fun> = reference to function to call for "create"
#            <cancel_fun> = reference to function to call for "cancel"
#            <sub_parm>   = parameter passed to create and cancel functions
#            <win_ref>    = reference for storing window handle
#            [<lab_text>] = optional: override label or disable "name entry" widgets
#
#  NOTES: The "common_defs.pm" file is required when this file is used.
#
#  MODIFICATION HISTORY:
#
#     WHEN      WHO             WHAT
#     --------  --------------  -----------------------------------------
#     10/16/03  Karen Beighley  Original source code submitted for release.
#
#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
#  CONSTANTS - These values cannot change!
#-------------------------------------------------------------------------------

$FALSE = 0;
$TRUE  = 1;

$MSWD_CHIP_NAME  =  0;  # MMR selection window data: full chip name
$MSWD_CHIP_TYPE  =  1;  # MMR selection window data: chip type (i.e. "shub")
$MSWD_LB_ALL     =  2;  # MMR selection window data: listbox with all names
$MSWD_LB_SEL     =  3;  # MMR selection window data: listbox with selected names
$MSWD_BTN_SEL    =  4;  # MMR selection window data: select button
$MSWD_BTN_UP     =  5;  # MMR selection window data: move up button
$MSWD_BTN_DWN    =  6;  # MMR selection window data: move down button
$MSWD_BTN_DEL    =  7;  # MMR selection window data: delete/remove button
$MSWD_NAME_TEXT  =  8;  # MMR selection window data: text string from name entry
$MSWD_NAME_LABEL =  9;  # MMR selection window data: label widget text string
$MSWD_SEL_LREF   = 10;  # MMR selection window data: reference to selected array
$MSWD_WINDOW     = 11;  # MMR selection window data: main window handle
$MSWD_CAN_FLAG   = 12;  # MMR selection window data: flag: TRUE=call cancel subroutine
$MSWD_MODE       = 13;  # MMR selection window data: mode
$MSWD_EVAL_NAME  = 14;  # MMR selection window data: mode
$MSWD_EVAL_ADDR  = 15;  # MMR selection window data: mode

$MMRSEL_MODE_DEFAULT     = 0x00;  # Default mode of operation
$MMRSEL_MODE_EMPTY_LIST  = 0x01;  # Support "empty" selection list
$MMRSEL_MODE_SHOW_NAME   = 0x10;  # Show the "name entry" widgets
$MMRSEL_MODE_EMPTY_NAME  = 0x20;  # Support "empty" text for name entry
$MMRSEL_MODE_SHOW_BROWSE = 0x40;  # Add browse button with "name entry" widgets


#-------------------------------------------------------------------------------

sub create_mmrsel_window
{
   printf "\ncreate_mmrsel_window( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   return if ( ($#_ != 11) && ($#_ != 12) );
   my $main_win   = $_[0];  # Handle to the main window widget
   my $chip_name  = $_[1];  # Full ASIC name (i.e. "001c02:NODE1:SHUB1")
   my $chip_type  = $_[2];  # ASIC type (i.e. "shub_01")
   my $desc       = $_[3];  # Description of operation ("for MMR Access Window")
   my $mode       = $_[4];  # Bitmap of modes that control selection behavior
   my $btn_text   = $_[5];  # Label for button: "Create", "Modify"
   my $curr_title = $_[6];  # Reference to current name/title of group
   my $curr_ref   = $_[7];  # Reference to array of selected (indices!)
   my $sub_create = $_[8];  # Reference to subroutine for create
   my $sub_cancel = $_[9];  # Reference to subroutine for cancel
   my $sub_parm   = $_[10]; # Reference to parameter for subroutine(s)
   my $win_ref    = $_[11]; # Reference: store window handle here
   my $lab_text   = ( ($mode & $MMRSEL_MODE_SHOW_BROWSE) ? "File:" : "Window Title:" );
   if ( exists($_[12]) )
   {  $lab_text = $_[12];
   }
   $btn_text = "Create" if ( $btn_text eq "" );

   #---------------------------------------
   #  Create and initialize data structures
   #---------------------------------------
   my @sel_data = ();
   my @msw_data = (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
   $msw_data[$MSWD_CHIP_NAME]  = $chip_name;
   $msw_data[$MSWD_CHIP_TYPE]  = $chip_type;
   $msw_data[$MSWD_MODE]       = $mode;
   $msw_data[$MSWD_SEL_LREF]   = \@sel_data;
   $msw_data[$MSWD_NAME_TEXT]  = ((defined($curr_title) && ($curr_title ne "")) ? $curr_title : "");
   $msw_data[$MSWD_NAME_LABEL] = $lab_text;
   $msw_data[$MSWD_CAN_FLAG]   = $TRUE;
   $msw_data[$MSWD_EVAL_NAME]  = "";
   $msw_data[$MSWD_EVAL_ADDR]  = "";

   my ($win_h, $frame, $label, $entry, $bwidth, $btn_create, $btn_cancel, $sel_lbox, 
       $all_box, $btn, $btn_sel, $btn_move1, $btn_move2, $btn_del1, $i, $text, $menubar,
       $name, $asic_mmr_list, $asic_mmr_hash, $hptr);

   #---------------------------------------
   #  Create main window for selection
   #---------------------------------------
   $win_h = $main_win->Toplevel(-title => "MMR Selection Window: $chip_name");
   $win_h->bind('<Destroy>' => [\&destroy_mmrsel_window, \@msw_data, $sub_cancel, $sub_parm, $win_ref]);
   if ( defined($win_ref) )
   {  $$win_ref = $win_h;  # Store so parent has access to window handle (to destroy)
   }

   #---------------------------------------
   #  Create the menu bar
   #---------------------------------------
   $menubar  = $win_h->Menu(-type => 'menubar');
   $win_h->configure(-menu => $menubar);
   $casmenu = $menubar->cascade(-label => '~File', -tearoff => 0);
   $casmenu->command(-label => $btn_text, -command => [\&mmrsel_cbf_create, \@msw_data, $sub_create, $sub_parm]);
   $casmenu->command(-label => 'Cancel', -command => [$win_h => 'destroy']);

   #---------------------------------------
   #  Create the search widgets
   #---------------------------------------
   $frame = $win_h->Frame(-borderwidth => 10);
   $label = $frame->Label(-anchor => 'w', -text => "Select $desc for $chip_name");
   $label->pack(-side => 'top', -fill => 'both');

   $sub_frame = $frame->Frame();
   $label = $sub_frame->Label(-text => "", -width => 5);
   $label->pack(-side => 'left');
   $label = $sub_frame->Label(-text => 'Search for MMR name starting with:', -width => 31, -anchor => 'w');
   $label->pack(-side => 'left');
   $entry = $sub_frame->Entry(-width => 40, -textvariable => \$msw_data[$MSWD_EVAL_NAME]);
   $entry->bind( '<Return>' => [\&mmrsel_search_name, \@msw_data] );
   $entry->pack(-side => 'left');
   $label = $sub_frame->Label(-text => '');
   $label->pack(-side => 'left', -expand => 'yes', -fill => 'x');
   $sub_frame->pack(-side => 'top', -expand => 'yes', -fill => 'both');
 
   $sub_frame = $frame->Frame();
   $label = $sub_frame->Label(-text => "", -width => 5);
   $label->pack(-side => 'left');
   $label = $sub_frame->Label(-text => 'Search for a specific MMR address:', -width => 31, -anchor => 'w');
   $label->pack(-side => 'left');
   $entry = $sub_frame->Entry(-width => 40, -textvariable => \$msw_data[$MSWD_EVAL_ADDR]);
   $entry->bind( '<Return>' => [\&mmrsel_search_addr, \@msw_data] );
   $entry->pack(-side => 'left');
   $label = $sub_frame->Label(-text => '');
   $label->pack(-side => 'left', -expand => 'yes', -fill => 'x');
   $sub_frame->pack(-side => 'top', -expand => 'yes', -fill => 'both');

   $frame->pack(-side => 'top', -fill => 'both', -expand => 'yes');

   #---------------------------------------
   #  Create the listbox for signal names
   #---------------------------------------
   $frame = $win_h->Frame(-borderwidth => 10);
   $label = $frame->Label(-anchor => 'w', -text => "Available $desc:");
   $label->pack(-side => 'top', -fill => 'both');
   $all_lbox = $frame->Scrolled('Listbox', -scrollbars => 'se', -width => 90,
                -height => 12, -font => 'Clean 12 bold', -selectmode => 'extended');
   $all_lbox->pack(-side => 'top', -fill => 'both', -expand => 'yes');
   $frame->pack(-side => 'top', -fill => 'both', -expand => 'yes');

   $all_lbox->bind('<Button-1>' => [\&mmrsel_cbf_all_lbox_btn1, \@msw_data]);
   $all_lbox->bind('<Button-2>' => [\&mmrsel_cbf_select, \@msw_data]);
   $all_lbox->bind('<Button-3>' => [\&mmrsel_cbf_select_and_create, \@msw_data, $sub_create, $sub_parm]);
   $all_lbox->bind('<Double-1>' => [\&mmrsel_cbf_select, \@msw_data]);

   #---------------------------------------
   #  Insert the names into the listbox
   #---------------------------------------
   $name = $chip_type . "_sorted_mmr_name";  # Assume chip type is valid!
   $asic_mmr_list = \@$name;
   $name = $chip_type . "_mmr_by_name";      # Assume chip type is valid!
   $asic_mmr_hash = \%$name;
   for ($i = 0; $i <= $#$asic_mmr_list; $i++)
   {  $label = $$asic_mmr_list[$i];
      $hptr  = $$asic_mmr_hash{$label};
      $text  = sprintf("%-36s 0x%09llX %s", $label, $$hptr{ADDRESS}, $$hptr{NOTE});
      $all_lbox->insert('end', $text);
   }

   #---------------------------------------
   #  Create the signal action buttons
   #---------------------------------------
   $bwidth = 12;
   $frame = $win_h->Frame(-borderwidth => 10);
   $btn_sel = $frame->Button(-text => 'Select', -width => $bwidth,
                  -state => 'disabled', -command => [\&mmrsel_cbf_select, \@msw_data]);
   $btn_sel->pack(-side => 'left', -padx => 5);
   $label = $frame->Label(-width => $bwidth, -anchor => 'e');
   $label->pack(-side => 'left', -padx => 5, -expand => 'yes');
   $btn_move1 = $frame->Button(-text => 'Move Up', -width => $bwidth,
                   -state => 'disabled', -command => [\&mmrsel_cbf_move_up, \@msw_data]);
   $btn_move1->pack(-side => 'left', -padx => 5);
   $btn_move2 = $frame->Button(-text => 'Move Down', -width => $bwidth,
                   -state => 'disabled', -command => [\&mmrsel_cbf_move_down, \@msw_data]);
   $btn_move2->pack(-side => 'left', -padx => 5);
   $btn_del1 = $frame->Button(-text => 'Remove', -width => $bwidth,
                  -state => 'disabled', -command => [\&mmrsel_cbf_del_sel, \@msw_data]);
   $btn_del1->pack(-side => 'left', -padx => 5);
   $frame->pack(-side => 'top', -fill => 'both');

   #---------------------------------------
   #  Create the listbox for signal names
   #---------------------------------------
   $frame = $win_h->Frame(-borderwidth => 10);
   $label = $frame->Label(-anchor => 'w', -text => "Selected $desc:");
   $label->pack(-side => 'top', -fill => 'both');
   $sel_lbox = $frame->Scrolled('Listbox', -scrollbars => 'se',
                -height => 12, -font => 'Clean 12 bold', -selectmode => 'extended');
   $sel_lbox->pack(-side => 'top', -fill => 'both', -expand => 'yes');
   $frame->pack(-side => 'top', -fill => 'both', -expand => 'yes');

   $sel_lbox->bind('<Button>' => [\&mmrsel_cbf_sel_lbox_btn, \@msw_data]);

   #---------------------------------------
   #  Insert the names into the listbox
   #---------------------------------------
   if ( ref($curr_ref) eq 'ARRAY' )
   {  for ($i = 0, $j = 0; $i <= $#$curr_ref; $i++)
      {  $cur = $$curr_ref[$i];
         if ( $cur <= $#$asic_mmr_list )
         {  $label = $$asic_mmr_list[$cur];
            $hptr  = $$asic_mmr_hash{$label};
            $text  = sprintf("%-36s 0x%09llX %s", $label, $$hptr{ADDRESS}, $$hptr{NOTE});
            $sel_lbox->insert('end', $text);
            $sel_data[$j] = $cur;  # store selected index in array
            $j++;
         }
         else
         {  printf("\nWARNING - Selected element does not exist in list: $cur\n");
         }
      }
   }

   #---------------------------------------
   #  Create the "Name/Title" widgets
   #---------------------------------------
   if ( $mode & $MMRSEL_MODE_SHOW_NAME )
   {  $frame = $win_h->Frame(-borderwidth => 10);
      $label = $frame->Label(-anchor => 'w', -text => $lab_text);
      $label->pack(-side => 'left', -expand => 'no');
      $entry = $frame->Entry(-width => 40, -textvariable => \$msw_data[$MSWD_NAME_TEXT]);
      $entry->pack(-side => 'left', -expand => 'yes', -fill => 'x');
      if ( $mode & $MMRSEL_MODE_SHOW_BROWSE )
      {  $btn = $frame->Button(-text => 'Browse', -width => $bwidth,
                   -command => [\&mmrsel_cbf_browse, \@msw_data]);
         $btn->pack(-side => 'right', -padx => 5);
      }
      $frame->pack(-side => 'top', -fill => 'both');
   }

   #---------------------------------------
   #  Create the action buttons
   #---------------------------------------
   $frame = $win_h->Frame(-borderwidth => 10);
   $bwidth = 20;
   $btn_create = $frame->Button(-text => $btn_text,
                    -width => $bwidth,
                    -command => [\&mmrsel_cbf_create, \@msw_data, $sub_create, $sub_parm]);
   $btn_create->pack(-side => 'left', -padx => 5);
   $btn_cancel = $frame->Button(-text => 'Cancel', -width => $bwidth, -command => [$win_h => 'destroy']);
   $btn_cancel->pack(-side => 'left', -padx => 5);
   $frame->pack(-side => 'top', -fill => 'both');

   #---------------------------------------
   #  Store widget handles for future use
   #---------------------------------------
   $msw_data[$MSWD_WINDOW]    = $win_h;
   $msw_data[$MSWD_LB_ALL]    = $all_lbox;
   $msw_data[$MSWD_LB_SEL]    = $sel_lbox;
   $msw_data[$MSWD_BTN_SEL]   = $btn_sel;
   $msw_data[$MSWD_BTN_UP]    = $btn_move1;
   $msw_data[$MSWD_BTN_DWN]   = $btn_move2;
   $msw_data[$MSWD_BTN_DEL]   = $btn_del1;
}

#-------------------------------------------------------------------------------

sub destroy_mmrsel_window
{
   print "\ndestroy_mmrsel_window( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $item_h     = $_[0];  # Handle to Tk widget/window
   my $swd_ref    = $_[1];  # Reference to selection window data
   my $sub_cancel = $_[2];  # Reference to subroutine
   my $sub_parm   = $_[3];  # Parameter to pass to subroutine
   my $ref        = $_[4];  # Reference to handle for parent
   my $win        = $$swd_ref[$MSWD_WINDOW];    # Top-level window handle

   return if ( $item_h != $win );  # Return if not main window

   #---------------------------------------
   #  Call "cancel" function (if defined)
   #---------------------------------------
   if ( $$swd_ref[$MSWD_CAN_FLAG] && defined($sub_cancel) && (ref($sub_cancel) eq 'CODE') )
   {  &$sub_cancel($sub_parm);
   }
   $$ref = 0; # Clear out window handle for parent
}

#-------------------------------------------------------------------------------

sub mmrsel_cbf_browse
{
   printf "\nmmrsel_cbf_browse( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $swd_ref  = $_[0];  # Reference to selection window data
   my $win_h    = $$swd_ref[$MSWD_WINDOW]; # Top-level window handle

   my @types = ( ['MMR Output File (Text)', '.txt'],
                 ['All Files',            , '*'   ]  );
   my $out_file = $win_h->getSaveFile(
                  -title => 'Save MMR Data',
                  -filetypes => \@types,
                  -defaultextension => '.txt'
                   );
   return if ( !defined($out_file) );

   $$swd_ref[$MSWD_NAME_TEXT] = $out_file;
}


#-------------------------------------------------------------------------------

sub mmrsel_cbf_all_lbox_btn1
{
   print "\nmmrsel_cbf_all_lbox_btn1( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $lb_ref  = $_[0];  # Reference to Listbox handle
   my $swd_ref = $_[1];  # Reference to selection window data
   my $cur     = $$swd_ref[$MSWD_LB_ALL]->curselection();
   if ( defined($cur) )
   {  mmrsel_set_button_state($swd_ref, 'all', 'normal');
      mmrsel_set_button_state($swd_ref, 'sel', 'disabled');
   }
}


#-------------------------------------------------------------------------------

sub mmrsel_cbf_sel_lbox_btn
{
   print "\nmmrsel_cbf_sel_lbox_btn( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $lb_ref  = $_[0];  # Reference to Listbox handle
   my $swd_ref = $_[1];  # Reference to selection window data
   my $cur     = $$swd_ref[$MSWD_LB_SEL]->curselection();
   if ( defined($cur) )
   {  mmrsel_set_button_state($swd_ref, 'all', 'disabled');
      mmrsel_set_button_state($swd_ref, 'sel', 'normal');
   }
}


#-------------------------------------------------------------------------------

sub mmrsel_cbf_select_and_create
{
   print "\nmmrsel_cbf_select_and_create( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   mmrsel_cbf_select($_[$#_-2]);
   mmrsel_cbf_create($_[$#_-2], $_[$#_-1], $_[$#_-0]);
}

#-------------------------------------------------------------------------------

sub mmrsel_cbf_select
{
   print "\nmmrsel_cbf_select( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $swd_ref = $_[$#_];  # Reference to selection window data
   my $sel_ref = $$swd_ref[$MSWD_SEL_LREF];  # Reference to selection list
   my ($cur, $text, $num, $i, $found);

   #---------------------------------------
   #  Get current selection(s) from listbox
   #---------------------------------------
   my @list = $$swd_ref[$MSWD_LB_ALL]->curselection();
   return unless @list;                  # return if no line selected

   #---------------------------------------
   #  Insert entries into the "selected" list
   #---------------------------------------
   foreach $cur (@list)
   {
      for ($i = 0, $found = $FALSE; ($i <= $#$sel_ref) && !$found; $i++)
      {  $found = ( $$sel_ref[$i] == $cur );
      }
      if ( !$found )
      {  $text = $$swd_ref[$MSWD_LB_ALL]->get($cur);    # get current line text from listbox
         $$swd_ref[$MSWD_LB_SEL]->insert('end', $text); # insert into the "selected" listbox
         $num  = $#$sel_ref + 1;  # increment index for array
         $$sel_ref[$num] = $cur;  # store selected index in array
      }
   }

   printf "\nselected =  ( @$sel_ref )\n" if ($debug_flag & $DEBUG_DETAILS);
}

#-------------------------------------------------------------------------------

sub mmrsel_cbf_move_up
{
   print "\nmmrsel_cbf_move_up( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $swd_ref = $_[0];  # Reference to selection window data
   my $sel_ref = $$swd_ref[$MSWD_SEL_LREF];  # Reference to selection list
   my $lbox    = $$swd_ref[$MSWD_LB_SEL];
   my ($cur, $text, $tmp);

   #---------------------------------------
   #  Get current selection(s) from listbox
   #---------------------------------------
   my @list = $lbox->curselection();
   return unless @list;                  # return if no line selected
   return unless $list[0];

   #---------------------------------------
   #  Delete the selected entries and insert in new position
   #---------------------------------------
   foreach $cur (@list)
   {
      $text = $lbox->get($cur);
      $lbox->delete($cur);
      $lbox->insert(($cur-1), $text);

      #  Update the array containing the selected indices
      $tmp = $$sel_ref[$cur];
      $$sel_ref[$cur] = $$sel_ref[($cur-1)];
      $$sel_ref[($cur-1)] = $tmp;
      $lbox->selection('set', ($cur-1));
      $lbox->see(($cur-1));
   }

   printf "\nselected =  ( @$sel_ref )\n" if ($debug_flag & $DEBUG_DETAILS);
}

#-------------------------------------------------------------------------------

sub mmrsel_cbf_move_down
{
   print "\nmmrsel_cbf_move_down( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $swd_ref = $_[0];  # Reference to selection window data
   my $sel_ref = $$swd_ref[$MSWD_SEL_LREF];  # Reference to selection list
   my $lbox    = $$swd_ref[$MSWD_LB_SEL];
   my ($n, $i, $cur, $text, $tmp);

   #---------------------------------------
   #  Get current selection(s) from listbox
   #---------------------------------------
   my @list = $lbox->curselection();
   return unless @list;                  # return if no line selected
   $n = $#list;
   return if ( $list[$n] == $#$sel_ref );

   #---------------------------------------
   #  Delete the selected entries and insert in new position
   #---------------------------------------
   for ($i = $n; $i >= 0; $i--)
   {  $cur = $list[$i];
      $text = $lbox->get($cur);
      $lbox->delete($cur);
      if ( ($cur+1) == $#$sel_ref) 
      {  $lbox->insert('end', $text); 
      }
      else 
      {  $lbox->insert(($cur+1), $text); 
      }

      #  Update the array containing the selected indices
      $tmp = $$sel_ref[$cur];
      $$sel_ref[$cur] = $$sel_ref[($cur+1)];
      $$sel_ref[($cur+1)] = $tmp;
      $lbox->selection('set', ($cur+1));
      $lbox->see(($cur+1));
   }

   printf "\nselected =  ( @$sel_ref )\n" if ($debug_flag & $DEBUG_DETAILS);
}

#-------------------------------------------------------------------------------

sub mmrsel_cbf_del_sel
{
   print "\nmmrsel_cbf_del_sel( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $swd_ref = $_[0];  # Reference to selection window data
   my $sel_ref = $$swd_ref[$MSWD_SEL_LREF];  # Reference to selection list
   my $lbox    = $$swd_ref[$MSWD_LB_SEL];
   my ($n, $i, $cur);

   #---------------------------------------
   #  Get current selection(s) from listbox
   #---------------------------------------
   my @list = $lbox->curselection();
   return unless @list;                  # return if no line selected
   $n = $#list;

   #---------------------------------------
   #  Delete the selected entries and insert in new position
   #---------------------------------------
   for ($i = $n; $i >= 0; $i--)
   {  $cur = $list[$i];

      $lbox->delete($cur);
      splice(@$sel_ref, $cur, 1);

      #  Update the buttons (disable) or the selected entry
      if ( $#$sel_ref < 0 )
      {  mmrsel_set_button_state($swd_ref, 'sel', 'disabled');
      }
      else
      {  $cur = $#$sel_ref  if ( $cur > $#$sel_ref );
         $lbox->selection('set', $cur) if ( $#$sel_ref >= 0 );
      }
   }

   printf "\nselected =  ( @$sel_ref )\n" if ($debug_flag & $DEBUG_DETAILS);
}


#-------------------------------------------------------------------------------

sub mmrsel_cbf_create
{
   print "\nmmrsel_cbf_create( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $swd_ref    = $_[0];  # Reference to selection window data
   my $sub_create = $_[1];  # Reference to subroutine
   my $sub_parm   = $_[2];  # Parameter to pass to subroutine

   my $sel_ref  = $$swd_ref[$MSWD_SEL_LREF];  # Reference to selection list
   my $win      = $$swd_ref[$MSWD_WINDOW];    # Top-level window handle
   my $lab_text = $$swd_ref[$MSWD_NAME_LABEL];

   #---------------------------------------
   #  Verify the user-defined selection(s)
   #---------------------------------------
   if ( ($#$sel_ref < 0) && 
        (($$swd_ref[$MSWD_MODE] & $MMRSEL_MODE_EMPTY_LIST) == 0) )
   {  $win->messageBox(-icon => 'error', -type => 'OK', 
            -message => "No MMR selections made from available list.",
            -title => 'MMR Selection Error');
      return;
   }
   if ( ($$swd_ref[$MSWD_MODE] & $MMRSEL_MODE_SHOW_NAME) &&
        ($$swd_ref[$MSWD_NAME_TEXT] eq "") &&
        (($$swd_ref[$MSWD_MODE] & $MMRSEL_MODE_EMPTY_NAME) == 0) )
   {  $win->messageBox(-icon => 'error', -type => 'OK', 
            -message => "The \"$lab_text\" parameter must be specified.",
            -title => 'MMR Selection Error');
      return;
   }

   if ( $debug_flag & $DEBUG_DETAILS )
   {  printf("\n********** SELECTION COMPLETE **********\n");
      printf("\nName : $$swd_ref[$MSWD_NAME_TEXT]\n");
      printf("List : ( @$sel_ref )\n");
   }

   #---------------------------------------
   #  Destroy the name selection window
   #---------------------------------------
   $$swd_ref[$MSWD_CAN_FLAG] = $FALSE;
   $win->destroy();

   #---------------------------------------
   #  Call "create" function (if defined)
   #---------------------------------------
   if ( defined($sub_create) && (ref($sub_create) eq 'CODE') )
   {  &$sub_create($sub_parm, $$swd_ref[$MSWD_NAME_TEXT], $sel_ref);
   }
}

#-------------------------------------------------------------------------------

sub mmrsel_search_name
{  print "\nmmrsel_search_name(@_) : #parms=$#_\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $rsvd      = $_[0];  # Tk handle to ENTRY
   my $swd_ref   = $_[1];  # Reference to selection window data
   my $eval_name = $$swd_ref[$MSWD_EVAL_NAME]; # Current entry widget contents
   my $win       = $$swd_ref[$MSWD_WINDOW];    # Top-level window handle

   return if ( length($eval_name) <= 0 );
   $eval_name =~ s/\s+//g;         # Remove all whitespace
   return if ( length($eval_name) <= 0 );
   my ($ch, $name, $tmp_name, $i, $cur, $found, $ind, $asic_mmr_list);

   #
   #  Create reference for specific ASIC type
   #
   $name = $$swd_ref[$MSWD_CHIP_TYPE] . "_sorted_mmr_name";
   $asic_mmr_list = \@$name;

   #  Determine where to start searching from in list
   $ind = $$swd_ref[$MSWD_LB_ALL]->curselection();
   if ( defined($ind) )
   {  $ind++;
      $ind = 0  if ( $ind > $#$asic_mmr_list );
   }
   else
   {  $ind = 0;
   }

   #  Search for the specified string; check for wild-card ('*')
   $ch = substr($eval_name, 0, 1); # Extract the first character
   if ( $ch eq '*' )  # User string starts with wild-card
   {  $tmp_name = substr($eval_name, 1, (length($eval_name) - 1));
      for ($i = 0, $found = 0; 
           (($i <= $#$asic_mmr_list) && ($found == 0)); $i++)
      {  if ( $$asic_mmr_list[$ind] =~ m/$tmp_name/i )
         {  $found = 1;
         }
         else
         {  $ind = (($ind == $#$asic_mmr_list) ? 0 : ($ind + 1));
         }
      }
      if ( $found != 1 )
      {  $name = sprintf("No register found with name containing \"$eval_name\".");
         $win->messageBox(-icon => 'info', -type => 'OK', -message => $name,
            -title => 'Search by Name Error');
         return;
      }
   }
   else # User string starts with non-wild-card (not '*')
   {  for ($i = 0, $found = 0; 
           (($i <= $#$asic_mmr_list) && ($found == 0)); $i++)
      {  if ( $$asic_mmr_list[$ind] =~ /^$eval_name/i )
         {  $found = 1 
         }
         else
         {  $ind = (($ind == $#$asic_mmr_list) ? 0 : ($ind + 1));
         }
      }
      if ( $found != 1 )
      {  $name = sprintf("No register found with name starting with \"$eval_name\".");
         $win->messageBox(-icon => 'info', -type => 'OK', -message => $name,
            -title => 'Search by Name Error');
         return;
      }
   }
   printf "Matching selection: ind=$ind \"$$asic_mmr_list[$ind]\"\n" if ($debug_flag & $DEBUG_GENERAL);

   #  Update the listbox so the specified MMR is selected and visible
   $cur = $$swd_ref[$MSWD_LB_ALL]->curselection();  # current listbox selection
   if ( !defined($cur) )             # no current listbox selection
   {  $$swd_ref[$MSWD_LB_ALL]->selection('set', $ind);
      mmrsel_set_button_state($swd_ref, 'all', 'normal');
      mmrsel_set_button_state($swd_ref, 'sel', 'disabled');
   }
   elsif ( $cur != $ind )              # current selection is different
   {  $$swd_ref[$MSWD_LB_ALL]->selection('clear', $cur);
      $$swd_ref[$MSWD_LB_ALL]->selection('set', $ind);
   }
   $$swd_ref[$MSWD_LB_ALL]->see($ind); # make selected entry visible to user
}


#-------------------------------------------------------------------------------

sub mmrsel_search_addr
{  print "\nmmrsel_search_addr(@_) : #parms=$#_\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $rsvd      = $_[0]; # Tk handle to ENTRY
   my $swd_ref   = $_[1];  # Reference to selection window data
   my $eval_addr = $$swd_ref[$MSWD_EVAL_ADDR]; # Current entry widget contents
   my $win       = $$swd_ref[$MSWD_WINDOW];    # Top-level window handle
   my ($addr, $hptr, $name, $i, $cur, $found, $asic_mmr_list, $asic_mmr_hash);

   $name = $$swd_ref[$MSWD_CHIP_TYPE] . "_sorted_mmr_name";
   $asic_mmr_list = \@$name;
   $name = $$swd_ref[$MSWD_CHIP_TYPE] . "_mmr_by_addr";
   $asic_mmr_hash = \%$name;

   return if ( length($eval_addr) == 0 );  # return if no valid string given
   $addr = convert_text2num($eval_addr);   # convert string to number
   $eval_addr = convert_num2text($addr, 64, $num_base); # update string

   printf "Searching for address: $addr=0x%08lX=\"$eval_addr\"\n", $addr if ($debug_flag & $DEBUG_GENERAL);
   #  Determine if the specified address is valid (entry in lookup table)
   if ( exists $$asic_mmr_hash{$addr} )
   {  $hptr = $$asic_mmr_hash{$addr};
      $name  = $$hptr{NAME};
   }
   else # The address is not valid (no entry in lookup table)
   {  $name = sprintf("No register found with address = %s", $$swd_ref[$MSWD_EVAL_ADDR]);
      $win->messageBox(-icon => 'info', -type => 'OK', -message => $name,
         -title => 'Search by Address Error');
      return;
   }
   #  Search for this register name in the sorted list (need index)
   for ($i = 0, $found = 0; 
        (($i <= $#$asic_mmr_list) && (!$found)); $i++)
   {  $found = 1  if ( $name eq $$asic_mmr_list[$i] );
   }
   if ( !$found )
   {  $name = sprintf("No register found with address = %s", $$swd_ref[$MSWD_EVAL_ADDR]);
      $win->messageBox(-icon => 'info', -type => 'OK', -message => $name,
         -title => 'Search by Address Error');
      return;
   }
   $i--;
   printf "Matching selection: i=$i \"$$asic_mmr_list[$i]\"\n" if ($debug_flag & $DEBUG_GENERAL);

   #  Update the listbox so the specified MMR is selected and visible
   $cur = $$swd_ref[$MSWD_LB_ALL]->curselection();  # current listbox selection
   if ( !defined($cur) )             # no current listbox selection
   {  $$swd_ref[$MSWD_LB_ALL]->selection('set', $i);
      mmrsel_set_button_state($swd_ref, 'all', 'normal');
      mmrsel_set_button_state($swd_ref, 'sel', 'disabled');
   }
   elsif ( $cur != $i )              # current selection is different
   {  $$swd_ref[$MSWD_LB_ALL]->selection('clear', $cur);
      $$swd_ref[$MSWD_LB_ALL]->selection('set', $i);
   }
   $$swd_ref[$MSWD_LB_ALL]->see($i); # make selected entry visible to user
}


#-------------------------------------------------------------------------------

sub mmrsel_set_button_state()
{  
   printf "\nmmrsel_set_button_state( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   return if ( $#_ != 2 );
   my $swd_ref = $_[0];  # Reference to selection window data
   return if ( ($_[2] ne 'normal') && ($_[2] ne 'disabled') );

   if ( $_[1] eq 'all' )
   {  $$swd_ref[$MSWD_BTN_SEL]->configure(-state => $_[2]);
   }
   elsif ( $_[1] eq 'sel' )
   {  $$swd_ref[$MSWD_BTN_UP]->configure(-state => $_[2]);
      $$swd_ref[$MSWD_BTN_DWN]->configure(-state => $_[2]);
      $$swd_ref[$MSWD_BTN_DEL]->configure(-state => $_[2]);
   }
}

#-------------------------------------------------------------------------------

1;
