# 
#/**************************************************************************
# *                                                                        *
# *               Copyright (C) 1989-1999 Silicon Graphics, Inc.           *
# *                                                                        *
# *  These coded instructions, statements, and computer programs  contain  *
# *  unpublished  proprietary  information of Silicon Graphics, Inc., and  *
# *  are protected by Federal copyright law.  They  may  not be disclosed  *
# *  to  third  parties  or copied or duplicated in any form, in whole or  *
# *  in part, without the prior written consent of Silicon Graphics, Inc.  *
# *                                                                        *
# **************************************************************************/
# 
# @(#)$Id: namesel.pm,v 1.5 2003/11/26 17:00:50 beighley Exp $
#-------------------------------------------------------------------------------
#
#  FILE: namesel.pm
#
#  DESCRIPTION:  This file contains PERL subroutines to create an GUI for a
#    user to select from a list of "names" or "strings".  One or more names
#    can be selected, and the selection order can be specified (if important).
#    A summary name or title can be provided for the group of selected "names".
#
#  USAGE:  The "create_namesel_window" should be called with the parameters:
#            <desc>       = description of names/strings (i.e. "Testpoints")
#            <mode>       = bitmap of modes that control selection behavior
#            <btn_text>   = text for the action button or "" for "Create"
#            <name_ref>   = reference to array containing names/string
#            <curr_title> = reference to current name of group (for entry)
#            <curr_ref>   = reference to array of selected (indices!)
#            <create_fun> = reference to function to call for "create"
#            <cancel_fun> = reference to function to call for "cancel"
#            <sub_parm>   = parameter passed to create and cancel functions
#            <win_ref>    = reference for storing window handle
#            [<grp_name>] = optional: override label or disable "Group Name" widgets
#
#  NOTES: The "common_defs.pm" file is required when this file is used.
#
#  MODIFICATION HISTORY:
#
#     WHEN      WHO             WHAT
#     --------  --------------  -----------------------------------------
#     09/30/02  Karen Beighley  Original source code.
#     11/20/02  Karen Beighley  Initial source submission for release.
#     06/24/03  Karen Beighley  Added optional parameter to override the label
#                                  or disable the "Group Name" widgets.
#     09/08/03  Karen Beighley  Added "mode" parameter to control behavior.
#     10/17/03  Karen Beighley  Added additional modes to control behavior.
#
#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
#  CONSTANTS - These values cannot change!
#-------------------------------------------------------------------------------

$FALSE = 0;
$TRUE  = 1;

$SWD_LB_ALL    =  0;  # Selection window data: listbox with all names
$SWD_LB_SEL    =  1;  # Selection window data: listbox with selected names
$SWD_BTN_SEL   =  2;  # Selection window data: select button
$SWD_BTN_UP    =  3;  # Selection window data: move up button
$SWD_BTN_DWN   =  4;  # Selection window data: move down button
$SWD_BTN_DEL   =  5;  # Selection window data: delete/remove button
$SWD_NAME_TEXT =  6;  # Selection window data: text string from name entry
$SWD_SEL_LREF  =  7;  # Selection window data: reference to selected array
$SWD_WINDOW    =  8;  # Selection window data: main window handle
$SWD_CAN_FLAG  =  9;  # Selection window data: flag: TRUE=call cancel subroutine
$SWD_MODE      = 10;  # Selection window data: mode

$NAMESEL_MODE_DEFAULT    = 0x00;  # Default mode of operation
$NAMESEL_MODE_EMPTY_LIST = 0x01;  # Support "empty" selection list
$NAMESEL_MODE_SHOW_NAME  = 0x10;  # Show the "group name" widgets
$NAMESEL_MODE_EMPTY_NAME = 0x20;  # Support "empty" text for group name entry



#-------------------------------------------------------------------------------

sub create_namesel_window
{
   printf "\ncreate_namesel_window( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $desc       = $_[0]; # Description of names/signals being selected
   my $mode       = $_[1]; # Bitmap of modes that control selection behavior
   my $btn_text   = $_[2]; # Label for button: "Create", "Modify"
   my $name_ref   = $_[3]; # Reference to array of names
   my $curr_title = $_[4]; # Reference to current name/title of group
   my $curr_ref   = $_[5]; # Reference to array of selected (indices!)
   my $sub_create = $_[6]; # Reference to subroutine for create
   my $sub_cancel = $_[7]; # Reference to subroutine for cancel
   my $sub_parm   = $_[8]; # Reference to parameter for subroutine(s)
   my $win_ref    = $_[9]; # Reference: store window handle here
   my $grp_name   = "Group Name:"; # Text to override or disable widgets;
   if ( exists($_[10]) )
   {  $grp_name = $_[10];
   }
   $btn_text = "Create" if ( $btn_text eq "" );

   #---------------------------------------
   #  Create and initialize data structures
   #---------------------------------------
   my @sel_data = ();
   my @sw_data  = (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
   $sw_data[$SWD_MODE]      = $mode;
   $sw_data[$SWD_NAME_TEXT] = ((defined($curr_title) && ($curr_title ne "")) ? $curr_title : "");
   $sw_data[$SWD_SEL_LREF]  = \@sel_data;
   $sw_data[$SWD_CAN_FLAG]  = $TRUE;

   my ($win_h, $frame, $label, $entry, $bwidth, $btn_create, $btn_cancel, $sel_lbox, 
       $all_box, $btn_sel, $btn_move1, $btn_move2, $btn_del1, $i, $text, $menubar);

   #---------------------------------------
   #  Create main window for selection
   #---------------------------------------
   $win_h = MainWindow->new(-title => "Selection Window: $desc");
   $win_h->bind('<Destroy>' => [\&destroy_namesel_window, \@sw_data, $sub_cancel, $sub_parm, $win_ref]);
   if ( defined($win_ref) )
   {  $$win_ref = $win_h;  # Store so parent has access to window handle (to destroy)
   }

   #---------------------------------------
   #  Create the menu bar
   #---------------------------------------
   $menubar  = $win_h->Menu(-type => 'menubar');
   $win_h->configure(-menu => $menubar);
   $casmenu = $menubar->cascade(-label => '~File', -tearoff => 0);
   $casmenu->command(-label => $btn_text, -command => [\&namesel_cbf_create, \@sw_data, $sub_create, $sub_parm]);
   $casmenu->command(-label => 'Cancel', -command => [$win_h => 'destroy']);

   #---------------------------------------
   #  Create the listbox for signal names
   #---------------------------------------
   $frame = $win_h->Frame(-borderwidth => 10);
   $label = $frame->Label(-anchor => 'w', -text => "Available $desc:");
   $label->pack(-side => 'top', -fill => 'both');
   $all_lbox = $frame->Scrolled('Listbox', -scrollbars => 'se',
                -height => 12, 
                -font => 'Clean 12 bold', -selectmode => 'extended');
   $all_lbox->pack(-side => 'top', -fill => 'both', -expand => 'yes');
   $frame->pack(-side => 'top', -fill => 'both', -expand => 'yes');

   $all_lbox->bind('<Button-1>' => [\&namesel_cbf_all_lbox_btn1, \@sw_data]);
   $all_lbox->bind('<Button-2>' => [\&namesel_cbf_select, \@sw_data]);
   $all_lbox->bind('<Button-3>' => [\&namesel_cbf_select_and_create, \@sw_data, $sub_create, $sub_parm]);
   $all_lbox->bind('<Double-1>' => [\&namesel_cbf_select, \@sw_data]);

   #---------------------------------------
   #  Insert the names into the listbox
   #---------------------------------------
   for ($i = 0; $i <= $#$name_ref; $i++)
   {  $text = $$name_ref[$i];
      $all_lbox->insert('end', $text);
   }

   #---------------------------------------
   #  Create the signal action buttons
   #---------------------------------------
   $bwidth = 12;
   $frame = $win_h->Frame(-borderwidth => 10);
   $btn_sel = $frame->Button(-text => 'Select', -width => $bwidth,
                  -state => 'disabled', -command => [\&namesel_cbf_select, \@sw_data]);
   $btn_sel->pack(-side => 'left', -padx => 5);
   $label = $frame->Label(-width => $bwidth, -anchor => 'e');
   $label->pack(-side => 'left', -padx => 5, -expand => 'yes');
   $btn_move1 = $frame->Button(-text => 'Move Up', -width => $bwidth,
                   -state => 'disabled', -command => [\&namesel_cbf_move_up, \@sw_data]);
   $btn_move1->pack(-side => 'left', -padx => 5);
   $btn_move2 = $frame->Button(-text => 'Move Down', -width => $bwidth,
                   -state => 'disabled', -command => [\&namesel_cbf_move_down, \@sw_data]);
   $btn_move2->pack(-side => 'left', -padx => 5);
   $btn_del1 = $frame->Button(-text => 'Remove', -width => $bwidth,
                  -state => 'disabled', -command => [\&namesel_cbf_del_sel, \@sw_data]);
   $btn_del1->pack(-side => 'left', -padx => 5);
   $frame->pack(-side => 'top', -fill => 'both');

   #---------------------------------------
   #  Create the listbox for signal names
   #---------------------------------------
   $frame = $win_h->Frame(-borderwidth => 10);
   $label = $frame->Label(-anchor => 'w', -text => "Selected $desc:");
   $label->pack(-side => 'top', -fill => 'both');
   $sel_lbox = $frame->Scrolled('Listbox', -scrollbars => 'se',
                -height => 8, 
                -font => 'Clean 12 bold', -selectmode => 'extended');
   $sel_lbox->pack(-side => 'top', -fill => 'both', -expand => 'yes');
   $frame->pack(-side => 'top', -fill => 'both', -expand => 'yes');

   $sel_lbox->bind('<Button>' => [\&namesel_cbf_sel_lbox_btn, \@sw_data]);

   #---------------------------------------
   #  Insert the names into the listbox
   #---------------------------------------
   if ( ref($curr_ref) eq 'ARRAY' )
   {  for ($i = 0, $j = 0; $i <= $#$curr_ref; $i++)
      {  $cur = $$curr_ref[$i];
         if ( $cur <= $#$name_ref )
         {  $sel_lbox->insert('end', $$name_ref[$cur]);
            $sel_data[$j] = $cur;  # store selected index in array
            $j++;
         }
         else
         {  printf("\nWARNING - Selected element does not exist in list: $cur\n");
         }
      }
   }

   #---------------------------------------
   #  Create the "Group Name" widgets
   #---------------------------------------
   if ( $mode & $NAMESEL_MODE_SHOW_NAME )
   {  $frame = $win_h->Frame(-borderwidth => 10);
      $label = $frame->Label(-anchor => 'w', -text => $grp_name);
      $label->pack(-side => 'left', -expand => 'no');
      $entry = $frame->Entry(-width => 40, -textvariable => \$sw_data[$SWD_NAME_TEXT]);
      $entry->pack(-side => 'left', -expand => 'yes', -fill => 'both');
      $frame->pack(-side => 'top', -fill => 'both');
   }

   #---------------------------------------
   #  Create the action buttons
   #---------------------------------------
   $frame = $win_h->Frame(-borderwidth => 10);
   $bwidth = 12;
   $btn_create = $frame->Button(-text => $btn_text,
                    -width => $bwidth,
                    -command => [\&namesel_cbf_create, \@sw_data, $sub_create, $sub_parm]);
   $btn_create->pack(-side => 'left', -padx => 5);
   $btn_cancel = $frame->Button(-text => 'Cancel', -width => $bwidth, -command => [$win_h => 'destroy']);
   $btn_cancel->pack(-side => 'left', -padx => 5);
   $frame->pack(-side => 'top', -fill => 'both');

   #---------------------------------------
   #  Store widget handles for future use
   #---------------------------------------
   $sw_data[$SWD_WINDOW]   = $win_h;
   $sw_data[$SWD_LB_ALL]   = $all_lbox;
   $sw_data[$SWD_LB_SEL]   = $sel_lbox;
   $sw_data[$SWD_BTN_SEL]  = $btn_sel;
   $sw_data[$SWD_BTN_UP]   = $btn_move1;
   $sw_data[$SWD_BTN_DWN]  = $btn_move2;
   $sw_data[$SWD_BTN_DEL]  = $btn_del1;
}

#-------------------------------------------------------------------------------

sub destroy_namesel_window
{
   print "\ndestroy_namesel_window( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $item_h     = $_[0];  # Handle to Tk widget/window
   my $swd_ref    = $_[1];  # Reference to selection window data
   my $sub_cancel = $_[2];  # Reference to subroutine
   my $sub_parm   = $_[3];  # Parameter to pass to subroutine
   my $ref        = $_[4];  # Reference to handle for parent
   my $win        = $$swd_ref[$SWD_WINDOW];    # Top-level window handle

   return if ( $item_h != $win );  # Return if not main window

   #---------------------------------------
   #  Call "cancel" function (if defined)
   #---------------------------------------
   if ( $$swd_ref[$SWD_CAN_FLAG] && defined($sub_cancel) && (ref($sub_cancel) eq 'CODE') )
   {  &$sub_cancel($sub_parm);
   }
   $$ref = 0; # Clear out window handle for parent
}


#-------------------------------------------------------------------------------

sub namesel_cbf_all_lbox_btn1
{
   print "\nnamesel_cbf_all_lbox_btn1( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $lb_ref  = $_[0];  # Reference to Listbox handle
   my $swd_ref = $_[1];  # Reference to selection window data
   my $cur     = $$swd_ref[$SWD_LB_ALL]->curselection();
   if ( defined($cur) )
   {  namesel_set_button_state($swd_ref, 'all', 'normal');
      namesel_set_button_state($swd_ref, 'sel', 'disabled');
   }
}


#-------------------------------------------------------------------------------

sub namesel_cbf_sel_lbox_btn
{
   print "\nnamesel_cbf_sel_lbox_btn( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $lb_ref  = $_[0];  # Reference to Listbox handle
   my $swd_ref = $_[1];  # Reference to selection window data
   my $cur     = $$swd_ref[$SWD_LB_SEL]->curselection();
   if ( defined($cur) )
   {  namesel_set_button_state($swd_ref, 'all', 'disabled');
      namesel_set_button_state($swd_ref, 'sel', 'normal');
   }
}


#-------------------------------------------------------------------------------

sub namesel_cbf_select_and_create
{
   print "\nnamesel_cbf_select_and_create( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   namesel_cbf_select($_[$#_-2]);
   namesel_cbf_create($_[$#_-2], $_[$#_-1], $_[$#_-0]);
}

#-------------------------------------------------------------------------------

sub namesel_cbf_select
{
   print "\nnamesel_cbf_select( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $swd_ref = $_[$#_];  # Reference to selection window data
   my $sel_ref = $$swd_ref[$SWD_SEL_LREF];  # Reference to selection list
   my ($cur, $text, $num, $i, $found);

   #---------------------------------------
   #  Get current selection(s) from listbox
   #---------------------------------------
   my @list = $$swd_ref[$SWD_LB_ALL]->curselection();
   return unless @list;                  # return if no line selected

   #---------------------------------------
   #  Insert entries into the "selected" list
   #---------------------------------------
   foreach $cur (@list)
   {
      for ($i = 0, $found = $FALSE; ($i <= $#$sel_ref) && !$found; $i++)
      {  $found = ( $$sel_ref[$i] == $cur );
      }
      if ( !$found )
      {  $text = $$swd_ref[$SWD_LB_ALL]->get($cur);    # get current line text from listbox
         $$swd_ref[$SWD_LB_SEL]->insert('end', $text); # insert into the "selected" listbox
         $num  = $#$sel_ref + 1;  # increment index for array
         $$sel_ref[$num] = $cur;  # store selected index in array
      }
   }

   printf "\nselected =  ( @$sel_ref )\n" if ($debug_flag & $DEBUG_DETAILS);
}

#-------------------------------------------------------------------------------

sub namesel_cbf_move_up
{
   print "\nnamesel_cbf_move_up( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $swd_ref = $_[0];  # Reference to selection window data
   my $sel_ref = $$swd_ref[$SWD_SEL_LREF];  # Reference to selection list
   my $lbox    = $$swd_ref[$SWD_LB_SEL];
   my ($cur, $text, $tmp);

   #---------------------------------------
   #  Get current selection(s) from listbox
   #---------------------------------------
   my @list = $lbox->curselection();
   return unless @list;                  # return if no line selected
   return unless $list[0];

   #---------------------------------------
   #  Delete the selected entries and insert in new position
   #---------------------------------------
   foreach $cur (@list)
   {
      $text = $lbox->get($cur);
      $lbox->delete($cur);
      $lbox->insert(($cur-1), $text);

      #  Update the array containing the selected indices
      $tmp = $$sel_ref[$cur];
      $$sel_ref[$cur] = $$sel_ref[($cur-1)];
      $$sel_ref[($cur-1)] = $tmp;
      $lbox->selection('set', ($cur-1));
      $lbox->see(($cur-1));
   }

   printf "\nselected =  ( @$sel_ref )\n" if ($debug_flag & $DEBUG_DETAILS);
}

#-------------------------------------------------------------------------------

sub namesel_cbf_move_down
{
   print "\nnamesel_cbf_move_down( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $swd_ref = $_[0];  # Reference to selection window data
   my $sel_ref = $$swd_ref[$SWD_SEL_LREF];  # Reference to selection list
   my $lbox    = $$swd_ref[$SWD_LB_SEL];
   my ($n, $i, $cur, $text, $tmp);

   #---------------------------------------
   #  Get current selection(s) from listbox
   #---------------------------------------
   my @list = $lbox->curselection();
   return unless @list;                  # return if no line selected
   $n = $#list;
   return if ( $list[$n] == $#$sel_ref );

   #---------------------------------------
   #  Delete the selected entries and insert in new position
   #---------------------------------------
   for ($i = $n; $i >= 0; $i--)
   {  $cur = $list[$i];
      $text = $lbox->get($cur);
      $lbox->delete($cur);
      if ( ($cur+1) == $#$sel_ref) 
      {  $lbox->insert('end', $text); 
      }
      else 
      {  $lbox->insert(($cur+1), $text); 
      }

      #  Update the array containing the selected indices
      $tmp = $$sel_ref[$cur];
      $$sel_ref[$cur] = $$sel_ref[($cur+1)];
      $$sel_ref[($cur+1)] = $tmp;
      $lbox->selection('set', ($cur+1));
      $lbox->see(($cur+1));
   }

   printf "\nselected =  ( @$sel_ref )\n" if ($debug_flag & $DEBUG_DETAILS);
}

#-------------------------------------------------------------------------------

sub namesel_cbf_del_sel
{
   print "\nnamesel_cbf_del_sel( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $swd_ref = $_[0];  # Reference to selection window data
   my $sel_ref = $$swd_ref[$SWD_SEL_LREF];  # Reference to selection list
   my $lbox    = $$swd_ref[$SWD_LB_SEL];
   my ($n, $i, $cur);

   #---------------------------------------
   #  Get current selection(s) from listbox
   #---------------------------------------
   my @list = $lbox->curselection();
   return unless @list;                  # return if no line selected
   $n = $#list;

   #---------------------------------------
   #  Delete the selected entries and insert in new position
   #---------------------------------------
   for ($i = $n; $i >= 0; $i--)
   {  $cur = $list[$i];

      $lbox->delete($cur);
      splice(@$sel_ref, $cur, 1);

      #  Update the buttons (disable) or the selected entry
      if ( $#$sel_ref < 0 )
      {  namesel_set_button_state($swd_ref, 'sel', 'disabled');
      }
      else
      {  $cur = $#$sel_ref  if ( $cur > $#$sel_ref );
         $lbox->selection('set', $cur) if ( $#$sel_ref >= 0 );
      }
   }

   printf "\nselected =  ( @$sel_ref )\n" if ($debug_flag & $DEBUG_DETAILS);
}


#-------------------------------------------------------------------------------

sub namesel_set_button_state()
{  
   printf "\nnamesel_set_button_state( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   return if ( $#_ != 2 );
   my $swd_ref = $_[0];  # Reference to selection window data
   return if ( ($_[2] ne 'normal') && ($_[2] ne 'disabled') );

   if ( $_[1] eq 'all' )
   {  $$swd_ref[$SWD_BTN_SEL]->configure(-state => $_[2]);
   }
   elsif ( $_[1] eq 'sel' )
   {  $$swd_ref[$SWD_BTN_UP]->configure(-state => $_[2]);
      $$swd_ref[$SWD_BTN_DWN]->configure(-state => $_[2]);
      $$swd_ref[$SWD_BTN_DEL]->configure(-state => $_[2]);
   }
}

#-------------------------------------------------------------------------------

sub namesel_cbf_create
{
   print "\nnamesel_cbf_create( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $swd_ref    = $_[0];  # Reference to selection window data
   my $sub_create = $_[1];  # Reference to subroutine
   my $sub_parm   = $_[2];  # Parameter to pass to subroutine

   my $sel_ref = $$swd_ref[$SWD_SEL_LREF];  # Reference to selection list
   my $win     = $$swd_ref[$SWD_WINDOW];    # Top-level window handle

   #---------------------------------------
   #  Verify the user-defined selection(s)
   #---------------------------------------
   if ( ($#$sel_ref < 0) && 
        (($$swd_ref[$SWD_MODE] & $NAMESEL_MODE_EMPTY_LIST) == 0) )
   {  $win->messageBox(-icon => 'error', -type => 'OK', 
            -message => "No selections made from list.",
            -title => 'Selection Error');
      return;
   }
   if ( ($$swd_ref[$SWD_MODE] & $NAMESEL_MODE_SHOW_NAME) &&
        ($$swd_ref[$SWD_NAME_TEXT] eq "") &&
        (($$swd_ref[$SWD_MODE] & $NAMESEL_MODE_EMPTY_NAME) == 0) )
   {  $win->messageBox(-icon => 'error', -type => 'OK', 
            -message => "The \"Group Name\" parameter must be specified.",
            -title => 'Selection Error');
      return;
   }

   if ( $debug_flag & $DEBUG_DETAILS )
   {  printf("\n********** SELECTION COMPLETE **********\n");
      printf("\nName : $$swd_ref[$SWD_NAME_TEXT]\n");
      printf("List : ( @$sel_ref )\n");
   }

   #---------------------------------------
   #  Destroy the name selection window
   #---------------------------------------
   $$swd_ref[$SWD_CAN_FLAG] = $FALSE;
   $win->destroy();

   #---------------------------------------
   #  Call "create" function (if defined)
   #---------------------------------------
   if ( defined($sub_create) && (ref($sub_create) eq 'CODE') )
   {  &$sub_create($sub_parm, $$swd_ref[$SWD_NAME_TEXT], $sel_ref);
   }
}

#-------------------------------------------------------------------------------

1;
