# 
#/**************************************************************************
# *                                                                        *
# *               Copyright (C) 1989-1999 Silicon Graphics, Inc.           *
# *                                                                        *
# *  These coded instructions, statements, and computer programs  contain  *
# *  unpublished  proprietary  information of Silicon Graphics, Inc., and  *
# *  are protected by Federal copyright law.  They  may  not be disclosed  *
# *  to  third  parties  or copied or duplicated in any form, in whole or  *
# *  in part, without the prior written consent of Silicon Graphics, Inc.  *
# *                                                                        *
# **************************************************************************/
# 
# @(#)$Id: pi_tools.pm,v 1.49 2007/03/23 19:38:43 meyer Exp $
#-------------------------------------------------------------------------------
#
#  FILE:  pi_tools.pm  
#
#
#  DESCRIPTION: contains routines & utilities for PI specific testing
#		used in conjunction with superscope scripts.
#
#  MODIFICATION HISTORY:
#
#     WHEN      WHO             WHAT
#     --------  --------------  -----------------------------------------
#     01/08/03  Gary Meyer      Original source code submitted for release.
#     05/02/03  Gary Meyer      Added tnd capture and additional PI livelock case detection.
#     10/28/03  Gary Meyer      Added dump flag to node attribute if problem is detected.
#     11/17/03  Karen Beighley  Modified to use MMR group operations (read/write).
#     11/17/03  Gary Meyer      Now only doing 100 loops when reading debug port.
#     08/03/04  Gary Meyer	Small change to implement improved PI problem reporting.
#				Added pi_deadlock_chk().
#     09/07/04  Gary Meyer	Added retries for superscope mmr reads & writes.
#     11/23/04  Gary Meyer	Added check for excessive TND assertions.
#     04/11/05  Gary Meyer	Added check for ICC stuck < 8.
#     06/06/05  Gary Meyer	Moved PI specific functions from errdmp to pi_tools.pm.
#				Changes for Shub2.
#     06/14/05  Gary Meyer	More Shub2 changes.
#     06/17/05  Gary Meyer	More Shub2 changes.  Still need updates to capture_sh2_pi_dbg().
#     07/14/05  Gary Meyer	Changes for proper proc decode from PRB value.  
#     09/12/05  Gary Meyer	Fixed bug where incorrect NASID could be used when saving the
#                               CRB address for directory state capture.
#     09/16/05  Gary Meyer	Changes for dumping directory state of addresses in firmware 
#				space.
#     10/24/05  Gary Meyer	Reduced polling of shub2 test points from 256 to 64.
#     10/27/05  Gary Meyer	Noting RRB's in INVAL_CNT > 0.  Provide more info for noteable RRBs.
#     10/28/05  Gary Meyer	Updated pi_deadlock_chk() to check either request or reply credits.
#     02/05/06  Gary Meyer	Updates for decoding agent id bits with AID_DECODE setting.
#     03/28/06  Gary Meyer	Added check for Sustained snoop conflict on the FSB in capture_sh2_pi_dbg_00().
#     03/31/06  Gary Meyer	Change to promote FRU score for Sustained snoop conflict on the FSB.
#     04/10/06  Gary Meyer	Change to promote FRU score for PI's with failing CAM reads.
#     05/23/06  Gary Meyer	Changes to improve output format for testpoint data in capture_sh2_pi_dbg_01().
#     06/02/06  Gary Meyer	Increased $sc_threshold from 12 to 24 in capture_sh2_pi_dbg_00() to prevent 
#				bogus snoop conflict indications.
#     06/23/06  Gary Meyer	Added error handling for mmr_target() function calls.
#     07/24/06  Gary Meyer      Update to ensure all NASID values use 4 characters.
#     07/27/06  Gary Meyer      Update to arm_pi_credits() to first read, then write the VC_CONFIG registers,
#				as opposed to using hard-coded values for the fields not written.
#     08/02/06  Gary Meyer	Qualify the $stall_stuck indication with PI CRB overuse in capture_sh2_pi_dbg_01().
#     10/18/06  Gary Meyer	Added clear for SH2_PERFORMANCE_MONITOR_DESKEW when reading test points.
#				Added seperate parsing functions for Shub2 testpoint data so they could be
#				used with fru testpoint data.  Initialize $SSCOPE_ERR before any jtag accesses.
#				Modified read_cam() to allow it to be used with PI crb info found in a fru dump.
#     10/20/06  Gary Meyer	Update to ensure "valid" value returned from read_cam() is correct in fru file mode.
#     10/26/06  Gary Meyer	Fix for WRB repeating problem in file mode.  Added verbose output option for fru crb.
#     11/08/06  Gary Meyer	Added support for monitor mode - BW# 951089.
#     11/13/06  Gary Meyer	Added support for detection of address in header corruption signature.
#     11/14/06  Gary Meyer	Modified debug message for address in header corruption signature so columns align.
#     12/01/06  Gary Meyer      Added code to recognize signature for BW 945998.
#     12/07/06  Gary Meyer      Fixed typo that caused RRB DID decode to be always 0.
#     12/11/06  Gary Meyer      Changes for getting directory state on nodes that haven't been booted.
#     12/16/06  Gary Meyer      Use global variables for CRB address compares.
#     12/20/06  Gary Meyer      Added L3 rtrdmp command line message if address in header corruption is found.
#     12/21/06  Gary Meyer      Fix for address in header corruption error reporting.
#     12/23/06  Gary Meyer      Minor output change to align address values in header compare.
#     01/02/07  Gary Meyer      Fix for address in header corruption error reporting - even_odd cacheline problem.
#     01/24/07  Gary Meyer      Fixed typo in address in header corruption error reporting that caused even routes
#				to be dumped in all cases. 
#     03/23/07  Gary Meyer      Address in header corruption update - could be failing in either direction.
#
#
#  NOTE: capture_debug_info provided by Mike Malewicki
#
#-------------------------------------------------------------------------------


sub capture_debug_info {

  if ($verbose_mode == 1)  {
    printf "\n  in capture_debug_info( @_ )\n";
  }

   my $livelock_cnt 	= 0;
   my $cmd_s1x_cnt	= 0;
   my $pipe_s4x_cnt	= 0;
   my $replay_s4x_cnt	= 0;
   my $valid_s4x_cnt	= 0;
   my $tnd_cnt		= 0;
   my $read_loop_limit 	= 100;
   my $ll_threshold	= $read_loop_limit / 2;
   my $tp_problem	= 0;
   my $pi_msg		= "";
   my $icc_cnt_lt_8	= 0;

   #
   #  Setup the debug port to monitor specific testpoints
   #
   my @addr = ();
   my @data = ();
   my $i = 0;
   $addr[$i] = 0x0000000110063000; # SH_DEBUG_SELECT
   $data[$i] = 0x00005D65544D2450;
   $i++;
   $addr[$i] = 0x0000000120050f00; # SH_PI_TEST_POINT_SELECT
   $data[$i] = 0x7666564636201006;
   $i++;
   $addr[$i] = 0x0000000120050180; # SH_PI_AFI_TEST_POINT_SELECT
   $data[$i] = 0x0000000000002100;
   $i++;
   $addr[$i] = 0x0000000120050700; # SH_PI_CRBP_TEST_POINT_SELECT
   $data[$i] = 0x1404027340000057;
   $i++;
   my $xfr_complete = 0;
   my $xfr_f_cnt = 0;
   while (($xfr_complete == 0) && ($xfr_f_cnt < 3) && ($target_list[$t_cnt]->{XFR_ERR_CNT} < 6))  {
     $SSCOPE_ERR = 0;
     mmr_write_group(\@addr, \@data);
     if ($SSCOPE_ERR == 1)  {
       $xfr_f_cnt++;
       $target_list[$t_cnt]->{XFR_ERR_CNT}++;
       $SSCOPE_ERR = 0;
     }
     else {
       $xfr_complete = 1;
     }
   }

   #
   #  Sample the debug port data
   #
   $port_addr = 0x0000000110072000; # SH_DEBUG_PORT
   @addr = ();
   @data = ();
   for ($i = 0; $i < $read_loop_limit; $i++)
   {  $addr[$i] = $port_addr;
   }

   $xfr_complete = 0;
   $xfr_f_cnt = 0;
   while (($xfr_complete == 0) && ($xfr_f_cnt < 3) && ($target_list[$t_cnt]->{XFR_ERR_CNT} < 6))  {
     $SSCOPE_ERR = 0;
     @data = mmr_read_group(\@addr);
     if ($SSCOPE_ERR == 1)  {
       $xfr_f_cnt++;
       $target_list[$t_cnt]->{XFR_ERR_CNT}++;
       $SSCOPE_ERR = 0;
     }
     else {
       $xfr_complete = 1;
     }
   }

   #
   #  Process the debug port data and display
   #
   printf("icc ioq_af ioq_f hitm hit tnd snp_s2f pipe_s2f reply_s2f proc_s2f snp_s4x pipe_s4x fsb_s4x replay_s4x valid_s4x cmd_s1x\n");

   my $icc_cnt_prev = 0;

   for($i=0; $i<$read_loop_limit; $i++) {
      $debug_port_val = $data[$i];

      $icc             = ( $debug_port_val & 0xf );
      $ioq_almost_full = ( $debug_port_val >> 5 ) & 1;
      $ioq_full        = ( $debug_port_val >> 6 ) & 1;
      $hitm            = ( $debug_port_val >> 8 ) & 1;
      $hit             = ( $debug_port_val >> 9 ) & 1;
      $tnd             = ( $debug_port_val >> 11 ) & 1;
      $snp_s2f         = ( $debug_port_val >> 12 ) & 1;
      $pipe_s2f        = ( $debug_port_val >> 13 ) & 1;
      $reply_s2f       = ( $debug_port_val >> 14 ) & 1;
      $proc_s2f        = ( $debug_port_val >> 15 ) & 1;
      $snp_s4x         = ( $debug_port_val >> 16 ) & 1;
      $pipe_s4x        = ( $debug_port_val >> 17 ) & 1;
      $fsb_s4x         = ( $debug_port_val >> 18 ) & 1;
      $replay_s4x      = ( $debug_port_val >> 19 ) & 1;
      $valid_s4x       = ( $debug_port_val >> 23 ) & 1;
      $cmd_s1x         = ( $debug_port_val >> 24 ) & 0xff;
    
      printf("%3d %6d %5d %4d %3d %3d %7d %8d %9d %8d %7d %8d %7d %10d %9d %7x %s\n", 
	$icc,$ioq_almost_full,$ioq_full, $hitm,$hit,$tnd,$snp_s2f,$pipe_s2f,$reply_s2f,$proc_s2f,
	$snp_s4x,$pipe_s4x,$fsb_s4x,$replay_s4x,$valid_s4x,$cmd_s1x, ($cmd_s1x != 0) ? $sn2net_command_list[$cmd_s1x]: " ");
    
      if (($ioq_almost_full == 1) && ($icc < 6)) {
	$livelock_cnt++;
      }
      if ($valid_s4x == 1)  {
	$valid_s4x_cnt++;
      }
      if ($replay_s4x == 1)  {
	$replay_s4x_cnt++;
      }
      if ($cmd_s1x > 0)  {
	$cmd_s1x_cnt++;
      }
      if ($pipe_s4x == 1)  {
	$pipe_s4x_cnt++;
      }
      if ($tnd == 1)  {
	$tnd_cnt++;
      }
      if ($i == 0)  {		# first iteration compare won't be compared against previous
	if ($icc < 8) {
	  $icc_cnt_lt_8++;
	}
      }
      else  {
	if (($icc == $icc_cnt_prev) && ($icc < 8)) {
	  $icc_cnt_lt_8++;
	}
      }
      $icc_cnt_prev = $icc;
   }

   if ($livelock_cnt > $ll_threshold)  
   {
     $any_error = 1;		# global flag from sn2dv
     $tp_problem = 1;
     $target_list[$t_cnt]->{ERROR} = 1;
     $pi_msg .= sprintf "   %s: Possible Livelock in PI XB Pipe.\n", $prefix;
     $target_list[$t_cnt]->{DUMP_ME} = 1;
   }
   if ($tnd_cnt > $ll_threshold)  
   {
     $any_error = 1;		# global flag from sn2dv
     $tp_problem = 1;
     $target_list[$t_cnt]->{ERROR} = 1;
     $pi_msg .= sprintf "   %s: TND Assertions on the FSB indicate Possible PTC Hang on either Processor.\n", $prefix;
     $target_list[$t_cnt]->{DUMP_ME} = 1;
   }
   if ( ($valid_s4x_cnt > $ll_threshold)
      || ($replay_s4x_cnt > $ll_threshold)
      || ($cmd_s1x_cnt > $ll_threshold)
      || ($pipe_s4x_cnt > $ll_threshold)
      || ($icc_cnt_lt_8 == $read_loop_limit) )
   {
     $any_error = 1;		# global flag from sn2dv
     $tp_problem = 1;
     $target_list[$t_cnt]->{ERROR} = 1;
     $pi_msg .= sprintf "   %s: Possible ICC Underflow in PI XB Pipe.\n", $prefix;
     $target_list[$t_cnt]->{DUMP_ME} = 1;
   }

   if ($tp_problem) {
     $target_list[$t_cnt]->{PI_ERR_MSG} = "$pi_msg\n";
   }
  
   my $icc_problem = ($icc_cnt_lt_8 == $read_loop_limit) ? 1 : 0;
   return($icc_problem);
}



#-------------------------------------------------------------------
#
#  In Shub2, we can read the debug data from SH2_PI_DEBUG_DATA, so 
#  this should allow us to run errdmp w/o affecting any tools that
#  gather system info from the top level debug port.
#
#  May also want to consider using the on-board logic analyzer.
#
#-------------------------------------------------------------------

sub capture_sh2_pi_dbg_00 {

   printf "\n  in capture_sh2_pi_dbg_00( @_ )\n"  if ($verbose_mode == 1);

   my $tp_problem = 0;

   #
   #  Setup the debug port to monitor specific testpoints
   #
   my @addr = ();
   my @data = ();
   my $i = 0;
   $addr[$i] = 0x0000000020050f00; # SH2_PI_TEST_POINT_SELECT
   $data[$i] = 0xF466564636261606; # Nibble 7- CRB CRBP CRBP CRBP CRBP CRBP CRBP CRBP
   $i++;
   $addr[$i] = 0x0000000020050700; # SH2_PI_CRBP_TEST_POINT_SELECT
   $data[$i] = 0x0150604073741404;
   $i++;
   $addr[$i] = 0x0000000010050100; # SH2_LB_DEBUG_MUX0_SEL
   $data[$i] = 0x5300000000000000;
   $i++;
   $addr[$i] = 0x0000000010050080; # SH2_LB_DEBUG_LOCAL_SEL
   $data[$i] = 0x7000000000000000;
   $i++;
   $addr[$i] = 0x0000000010101000; # SH2_PERFORMANCE_MONITOR_DESKEW
   $data[$i] = 0x0000000000000000;
   $i++;
   $addr[$i] = 0x0000000010064000; # SH2_DEBUG_SEL
   $data[$i] = 0x1726252423222120;
   $i++;


   my $xfr_complete = 0;
   my $xfr_f_cnt = 0;
   while (($xfr_complete == 0) && ($xfr_f_cnt < 3) && ($sh2_target_list[$sh2_cnt]->{XFR_ERR_CNT} < 6))  {
     $SSCOPE_ERR = 0;
     mmr_write_group(\@addr, \@data);
     if ($SSCOPE_ERR == 1)  {
       $xfr_f_cnt++;
       $sh2_target_list[$sh2_cnt]->{XFR_ERR_CNT}++;
       $SSCOPE_ERR = 0;
     }
     else {
       $xfr_complete = 1;
     }
   }

   #
   #  Sample the debug port data
   #
   $port_addr = 0x10072000; # SH2_DEBUG_PORT

   @addr = ();
   @data = ();

   my $read_loop_limit 	= 60;

   for ($i = 0; $i < $read_loop_limit; $i++)
   {  $addr[$i] = $port_addr;
   }

   $xfr_complete = 0;
   $xfr_f_cnt = 0;
   while (($xfr_complete == 0) && ($xfr_f_cnt < 3) && ($sh2_target_list[$sh2_cnt]->{XFR_ERR_CNT} < 6))  {
     $SSCOPE_ERR = 0;
     @data = mmr_read_group(\@addr);
     if ($SSCOPE_ERR == 1)  {
       $xfr_f_cnt++;
       $sh2_target_list[$sh2_cnt]->{XFR_ERR_CNT}++;
       $SSCOPE_ERR = 0;
     }
     else {
       $xfr_complete = 1;
     }
   }

   $tp_problem = parse_sh2_tp00_list(\@data);

   return($tp_problem);

}


#----------------------------------------------------------------------
#
#  This function parses a list of tp data
#
#----------------------------------------------------------------------

sub parse_sh2_tp00_list {

   my $tp_list_ref = $_[0];
   my @tp_data = ();
   @tp_data = @$tp_list_ref;

   my $tp_data_limit 	= ($#tp_data) + 1;
   my $ll_threshold	= $tp_data_limit / 2;
   my $pi_msg		= "";
   my $sc_threshold     = ($tp_data_limit / 2) - 8;
   my $sc_cnt		= 0;
   my $tnd_cnt		= 0;
   my $tp_issue		= 0;


   #
   #  Process the debug port data and display
   #
  # 0)  cmnd_s1x[0]   
  # 1)  cmnd_s1x[1]   
  # 2)  cmnd_s1x[2]                                                 
  # 3)  cmnd_s1x[3]                                                 
  # 4)  cmnd_s1x[4]
  # 5)  cmnd_s1x[5]
  # 6)  cmnd_s1x[6]
  # 7)  cmnd_s1x[7]               
  # 8)  inval_rrb_clear            
  # 9)  inval_rrb_idx_val          
 # 10)  rd_xfer_conflict_s1x       
 # 11)  irb_alloc_fail_s1x          
 # 12)  snoop_conflict_s4x         
 # 13)  pipe_conflict_s4x          
 # 14)  fsb_conflict_s4x           
 # 15)  do_inmsg_replay_s4x          
 # 16)  snoop_conflict_s2f                
 # 17)  pipe_conflict_s2f                 
 # 18)  reply_conflict_s2f                
 # 19)  crb_afi_proc_conflict             
 # 20)  is_read_s1f                      
 # 21)  is_pio_s1f                       
 # 22)  is_req_s1f                       
 # 23)  stage_valid_s1f                  
 # 24)  afi_crb_mem_valid                
 # 25)  afi_crb_pio_valid                
 # 26)  afi_crb_snoop_valid              
 # 27)  afi_crb_proc_soft_error     (Req format or Addr Access error, valid only when ads_2nd is asserted in afi_core.bus_check)           
 # 31)  tnd

   printf("                                                   s4x conflicts  s2f conflicts                               \n");
   printf("inv_r_clr inv_r_val rd_xfr_cfl irb_alloc_fl replay snp pipe fsb   snp pipe reply ca_proc_cfl is_read is_pio is_req stg_vld_s1f ac_mem_val ac_pio_val ac_snp_val tnd cmd_s1x\n");

   my $i = 0;

   for($i=0; $i<$tp_data_limit; $i++) {
      $debug_port_val = $tp_data[$i];

      $cmd_s1x                	=  $debug_port_val & 0xff;		# CRBP
      $inval_rrb_clear 		= ( $debug_port_val >> 8 ) & 1;		# CRBP
      $inval_rrb_idx_val 	= ( $debug_port_val >> 9 ) & 1;		# CRBP
      $rd_xfer_conflict_s1x     = ( $debug_port_val >> 10) & 1;		# CRBP
      $irb_alloc_fail_s1x       = ( $debug_port_val >> 11) & 1;		# CRBP
      $snoop_conflict_s4x       = ( $debug_port_val >> 12) & 1;		# CRBP
      $pipe_conflict_s4x        = ( $debug_port_val >> 13) & 1;		# CRBP
      $fsb_conflict_s4x         = ( $debug_port_val >> 14) & 1;		# CRBP
      $do_inmsg_replay_s4x      = ( $debug_port_val >> 15) & 1;		# CRBP
      $snoop_conflict_s2f       = ( $debug_port_val >> 16) & 1;		# CRBP
      $pipe_conflict_s2f        = ( $debug_port_val >> 17) & 1;		# CRBP
      $reply_conflict_s2f       = ( $debug_port_val >> 18) & 1;		# CRBP
      $crb_afi_proc_conflict    = ( $debug_port_val >> 19) & 1;		# CRBP
      $is_read_s1f         	= ( $debug_port_val >> 20) & 1;		# CRBP
      $is_pio_s1f         	= ( $debug_port_val >> 21) & 1;		# CRBP
      $is_req_s1f         	= ( $debug_port_val >> 22) & 1;		# CRBP
      $stage_valid_s1f         	= ( $debug_port_val >> 23) & 1;		# CRBP
      $afi_crb_mem_valid        = ( $debug_port_val >> 24) & 1;		# CRBP
      $afi_crb_pio_valid        = ( $debug_port_val >> 25) & 1;		# CRBP
      $afi_crb_snoop_valid      = ( $debug_port_val >> 26) & 1;		# CRBP
      $afi_crb_proc_error       = ( $debug_port_val >> 27) & 1;		# CRBP
      $tnd                      = ( $debug_port_val >> 31) & 1;		# LB
    
      printf("%4d %9d %10d %12d %8d %4d %4d %3d    %3d %4d %4d %9d %7d %6d %6d %11d %10d %10d %10d %6d %5X %s\n", 
	$inval_rrb_clear, $inval_rrb_idx_val, $rd_xfer_conflict_s1x, $irb_alloc_fail_s1x, $do_inmsg_replay_s4x,
	$snoop_conflict_s4x, $pipe_conflict_s4x, $fsb_conflict_s4x, $snoop_conflict_s2f, $pipe_conflict_s2f,
	$reply_conflict_s2f, $crb_afi_proc_conflict, $is_read_s1f, $is_pio_s1f, $is_req_s1f, $stage_valid_s1f,
	$afi_crb_mem_valid, $afi_crb_pio_valid, $afi_crb_snoop_valid,
	$tnd, $cmd_s1x, ($cmd_s1x != 0) ? $sn2net_command_list[$cmd_s1x]: " ");
    
      if ($tnd == 1)  {
	$tnd_cnt++;
      }
      if (($snoop_conflict_s2f == 1) && ($crb_afi_proc_conflict == 1)) {
	$sc_cnt++;
      }
   }

   print "\n\n";

   if ($tnd_cnt > $ll_threshold)  
   {
     $any_sh2_error = 1;		# global flag from sn2dv
     $tp_issue += 1;
     $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
     $pi_msg .= sprintf "   %s: TND Assertions on the FSB indicate Possible PTC Hang on either Processor.\n", $prefix;
     $sh2_target_list[$sh2_cnt]->{DUMP_ME} = 1;
   }

   if ($sc_cnt > $sc_threshold)  
   {
     $any_sh2_error = 1;		# global flag from sn2dv
     $tp_issue += 2;
     $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
     $pi_msg .= sprintf "   %s:  Sustained snoop conflict detected on the FSB indicates possible FSB Hang.\n", $prefix;
   }

   if ($tp_issue > 0) {
     $sh2_target_list[$sh2_cnt]->{PI_ERR_MSG} = "$pi_msg\n";
   }
   
   return($tp_issue);
}


#-------------------------------------------------------------------
#
#  In Shub2, we can read the debug data from SH2_PI_DEBUG_DATA, so 
#  this should allow us to run errdmp w/o affecting any tools that
#  gather system info from the top level debug port.
#
#  May also want to consider using the on-board logic analyzer.
#
#-------------------------------------------------------------------

sub capture_sh2_pi_dbg_01 {

  if ($verbose_mode == 1)  {
    printf "\n  in capture_sh2_pi_dbg_01( @_ )\n";
  }
   my $read_loop_limit 	= 60;

   #
   #  Setup the debug port to monitor specific testpoints
   #
   my @addr = ();
   my @data = ();
   my $i = 0;
   $addr[$i] = 0x0000000020050f00; # SH2_PI_TEST_POINT_SELECT
   $data[$i] = 0x0000000000251505; # 
   $i++;
   $addr[$i] = 0x0000000020050400,; # SH2_PI_CRBC_TEST_POINT_SELECT
   $data[$i] = 0x0000000000701060;

   my $xfr_complete = 0;

   my $xfr_f_cnt = 0;
   while (($xfr_complete == 0) && ($xfr_f_cnt < 3) && ($sh2_target_list[$sh2_cnt]->{XFR_ERR_CNT} < 6))  {
     $SSCOPE_ERR = 0;
     mmr_write_group(\@addr, \@data);
     if ($SSCOPE_ERR == 1)  {
       $xfr_f_cnt++;
       $sh2_target_list[$sh2_cnt]->{XFR_ERR_CNT}++;
       $SSCOPE_ERR = 0;
     }
     else {
       $xfr_complete = 1;
     }
   }

   #
   #  Sample the debug port data
   #
   $port_addr = 0x10072280; 		# SH2_PI_DEBUG_DATA

   @addr = ();
   @data = ();
   for ($i = 0; $i < $read_loop_limit; $i++)
   {  $addr[$i] = $port_addr;
   }

   $xfr_complete = 0;
   $xfr_f_cnt = 0;
   while (($xfr_complete == 0) && ($xfr_f_cnt < 3) && ($sh2_target_list[$sh2_cnt]->{XFR_ERR_CNT} < 6))  {
     $SSCOPE_ERR = 0;
     @data = mmr_read_group(\@addr);
     if ($SSCOPE_ERR == 1)  {
       $xfr_f_cnt++;
       $sh2_target_list[$sh2_cnt]->{XFR_ERR_CNT}++;
       $SSCOPE_ERR = 0;
     }
     else {
       $xfr_complete = 1;
     }
   }

   my $bpri_on = 0;
   my $crb_low = 0;
   my $cmd_q_af = 0;
   ($bpri_on, $crb_low, $cmd_q_af) = parse_sh2_tp01_list(\@data);
   return($bpri_on, $crb_low, $cmd_q_af);
}

#----------------------------------------------------------------------
#
#  This function parses a list of tp data
#
#----------------------------------------------------------------------

sub parse_sh2_tp01_list {

   my $tp_list_ref = $_[0];
   my @tp_data = ();
   @tp_data = @$tp_list_ref;
   my $tp_output = "";

   my $tp_data_limit 	= ($#tp_data) + 1;
   my $pi_msg		= "";
   my $stall_cnt	= 0;
   my $wtl_q_af_cnt   = 0;
   my $rrb_skid_cnt	= 0;
   my $wrb_skid_cnt	= 0;
   my $prb0_skid_cnt	= 0;
   my $prb1_skid_cnt	= 0;
   my $prb2_skid_cnt	= 0;
   my $prb3_skid_cnt	= 0;

  #  0)  stall_fsb_for_ext_request_tp 
  #  1)  dpc_crbc_fsb_wtl_cmnd_q_af_tp
  #  2)  wrb_wr_xfer_cnt_at_max_tp  
  #  3)  crb_afi_busy              
  #  4)  fsb_pipe_rrb_rd_xfer_flip_idx_0[4]  
  #  5)  fsb_pipe_rrb_rd_xfer_flip_en_0     
  #  6)  too_few_rrbs_tp                   
  #  7)  too_few_wrbs_tp                  
  #  8)  too_few_prbs_thread00_tp        
  #  9)  too_few_prbs_thread01_tp       
  #  10)  too_few_prbs_thread10_tp      
  #  11)  too_few_prbs_thread11_tp     


   $tp_output = sprintf("stall_fsb_for  dpc_crbc_fsb_wtl  wrb_wr_xfer  crb_afi_busy  too_few_rrbs  too_few_wrbs  too_few_prbs \n");
   $tp_output .= sprintf("_ext_request   _cmnd_q_af        _at_max                                                thread00 thread01 thread10 thread11 \n");

   my $i = 0;
   for($i=0; $i<$tp_data_limit; $i++) {
     my $stall    = $tp_data[$i] & 0x1;
     my $fsb_wtl  = ( $tp_data[$i] >> 1 ) & 0x1;
     my $wr_xfr   = ( $tp_data[$i] >> 2 ) & 0x1;
     my $afi_bsy  = ( $tp_data[$i] >> 3 ) & 0x1;
     my $few_rrb  = ( $tp_data[$i] >> 6 ) & 0x1;
     my $few_wrb  = ( $tp_data[$i] >> 7 ) & 0x1; 
     my $few_prb0 = ( $tp_data[$i] >> 8 ) & 0x1;
     my $few_prb1 = ( $tp_data[$i] >> 9 ) & 0x1;
     my $few_prb2 = ( $tp_data[$i] >> 10 ) & 0x1;
     my $few_prb3 = ( $tp_data[$i] >> 11 ) & 0x1;
    
     $tp_output .= sprintf("   %d             %d                  %d            %d               %d             %d           %d        %d        %d       %d\n", 
	$stall, $fsb_wtl, $wr_xfr, $afi_bsy, $few_rrb, $few_wrb, $few_prb0, $few_prb1, $few_prb2, $few_prb3);

     $stall_cnt += $stall; 
     $wtl_q_af_cnt += $fsb_wtl;
     $rrb_skid_cnt += $few_rrb; 
     $wrb_skid_cnt += $few_wrb; 
     $prb0_skid_cnt += $few_prb0; 
     $prb1_skid_cnt += $few_prb1; 
     $prb2_skid_cnt += $few_prb2; 
     $prb3_skid_cnt += $few_prb3; 
   }

   my $stall_stuck = ($stall_cnt > ($tp_data_limit - 3)) ? 1 : 0;
   my $wtl_cmd_q_af = ($wtl_q_af_cnt > ($tp_data_limit - 3)) ? 1 : 0;
   my $crb_skid =  ($rrb_skid_cnt > ($tp_data_limit - 5)) ? 1 
		 : ($wrb_skid_cnt > ($tp_data_limit - 5)) ? 1
		 : ($prb0_skid_cnt > ($tp_data_limit - 5)) ? 1
		 : ($prb1_skid_cnt > ($tp_data_limit - 5)) ? 1
		 : ($prb2_skid_cnt > ($tp_data_limit - 5)) ? 1
		 : ($prb3_skid_cnt > ($tp_data_limit - 5)) ? 1 : 0;

   if (($monitor_mode == 0) || ($stall_stuck == 1) || ($wtl_cmd_q_af == 1)) {
     printf "\n%s\n\n", $tp_output;
   }
   return($stall_stuck, $crb_skid, $wtl_cmd_q_af);
}



##########################
#
#  This function reads the SH_XNPI_SIC_FLOW register 40 times, to check
#  if the PI is consistently starving the XN of request or reply credits.  A value
#  of 1 is returned if each read indicates request or reply credits are == 0.
#
##########################

sub pi_deadlock_chk  {
  if ($verbose_mode == 1)  {
    print "\n In pi_deadlock_chk().\n";
  }

  $chip_type = "$_[0]";
  my $cred_type = $_[1];	#  0 is request     1 is reply
  my $cred_stuck = 0;
  my $zero_count    = 0;
  my $read_count    = 40;
  my @pi_cred_reg_list = ();
  my $crd_reg_name = ($chip_type =~ /shub_01/) ? "SH_XNPI_SIC_FLOW" : "SH2_XNPI_SIC_FLOW";
  
  for (my $i = 0; $i < $read_count; $i++)  { 
    $reg_ptr = reg_create_mmr_inst($chip_type, $crd_reg_name);
    if ( !reg_is_valid($reg_ptr) )
    {  $emess = "\nWarning!  Unable to create register instance: $name.\n";
       print "$emess";
    }
    $pi_cred_reg_list[$i] = $reg_ptr;
  }

  my $xfr_complete = 0;
  my $xfr_f_cnt = 0;
  my $init_chip_err_cnt = ($chip_type =~ /shub_01/) ? $target_list[$t_cnt]->{XFR_ERR_CNT} 
					       : $sh2_target_list[$sh2_cnt]->{XFR_ERR_CNT};
  my $chip_err_cnt = $init_chip_err_cnt;

  while (($xfr_complete == 0) && ($xfr_f_cnt < 3) && ($chip_err_cnt < 6))  {
    $SSCOPE_ERR = 0;
    reg_list_mmr_read(\@pi_cred_reg_list);
    if ($SSCOPE_ERR == 1)  {
      $xfr_f_cnt++;
      $chip_err_cnt++;
      $SSCOPE_ERR = 0;
    }
    else {
      $xfr_complete = 1;
    }
  }

  if ($init_chip_err_cnt != $chip_err_cnt)  {       # update counters if errors occurred
    if ($chip_type =~ /shub_01/)  {
      $target_list[$t_cnt]->{XFR_ERR_CNT} = $chip_err_cnt;
    }
    else {
      $sh2_target_list[$sh2_cnt]->{XFR_ERR_CNT} = $chip_err_cnt;
    }
  }

  my $dyn_field = ($cred_type == 0) ? 'CREDIT_VC0_DYN' : 'CREDIT_VC2_DYN';
  my $tst_field = ($cred_type == 0) ? 'CREDIT_VC0_TEST' : 'CREDIT_VC2_TEST';

  if ($xfr_complete == 1)  {
    for (my $i = 0; $i < $read_count; $i++)  { 
      $reg_ptr = $pi_cred_reg_list[$i];
      if ( ((reg_get_field_value($reg_ptr, $dyn_field)) == 0)
         && ((reg_get_field_value($reg_ptr, $tst_field)) == 8) ) {
	  $zero_count++;
      }
    }
    if ($zero_count == $read_count)  {
      printf "\nRe-checked %s %s VC credit levels and they were at 0 %d times.\n\n",
	$crd_reg_name, (($cred_type == 0) ? 'Request' : 'Reply'), $zero_count;
      $cred_stuck = 1;
    }
  }
  else {
      print "Error in pi_deadlock_chk() reading $crd_reg_name.\n\n";
  }
  return ($cred_stuck);
}


#-------------------------------------------------------------------------------
#
#  Subroutine to extract the Intervention Credit count from the PI chiplet.
#  Based on the following code from Darin Oster:
#
#      REMOTE_HUB_S(nasid, SH_DEBUG_SELECT,              0x0000000000000450L);
#      REMOTE_HUB_S(nasid, SH_PI_TEST_POINT_SELECT,      0x0000000000001606L);
#      REMOTE_HUB_S(nasid, SH_PI_CRBP_TEST_POINT_SELECT, 0x0000000000006757L);
#      icc = (REMOTE_HUB_L(nasid, SH_DEBUG_PORT)) & 0xfL;
#
#  Returns the ICC value.
#
#-------------------------------------------------------------------------------

sub get_icc()
{
  print "\nIn get_icc function\n" if ($verbose_mode == 1);

  $SH_DEBUG_SELECT_ADDR =               0x10063000;
  $SH_DEBUG_PORT_ADDR =                 0x10072000;
  $SH_PI_TEST_POINT_SELECT_ADDR =       0x20050f00;
  $SH_PI_CRBP_TEST_POINT_SELECT_ADDR =  0x20050700;
  $SH_PI_AFI_TEST_POINT_SELECT_ADDR =   0x20050180;

  my @adds = ( $SH_DEBUG_SELECT_ADDR,
               $SH_PI_TEST_POINT_SELECT_ADDR,
               $SH_PI_CRBP_TEST_POINT_SELECT_ADDR,
               $SH_PI_AFI_TEST_POINT_SELECT_ADDR );
  my @vals  = ( 0x12450, 0x201606, 0x6757, 0x210000);

  #
  #  Write the group of MMRs
  #
  my $xfr_complete = 0;
  my $xfr_f_cnt = 0;
  while (($xfr_complete == 0) && ($xfr_f_cnt < 3) && ($target_list[$t_cnt]->{XFR_ERR_CNT} < 6))  {
    $SSCOPE_ERR = 0;
    mmr_write_group(\@adds, \@vals);                    #  return mask setting to original value
    if ($SSCOPE_ERR == 1)  {
      $xfr_f_cnt++;
      $target_list[$t_cnt]->{XFR_ERR_CNT}++;
      $SSCOPE_ERR = 0;
    }
    else {
      $xfr_complete = 1;
    }
  }

  #
  #  Read the SH_DEBUG_PORT MMR
  #

  $xfr_complete = 0;
  $xfr_f_cnt = 0;
  while (($xfr_complete == 0) && ($xfr_f_cnt < 3) && ($target_list[$t_cnt]->{XFR_ERR_CNT} < 6))  {
    $SSCOPE_ERR = 0;
    $debug_value = mmr_read($SH_DEBUG_PORT_ADDR);
    if ($SSCOPE_ERR == 1)  {
      $target_list[$t_cnt]->{XFR_ERR_CNT}++;
      $xfr_f_cnt++;
      $SSCOPE_ERR = 0;
    }
    else {
      $xfr_complete = 1;
    }
  }

  if ($verbose_mode == 1)  {
    printf ("\nValue read from SH_DEBUG_PORT:  0x%016llX\n", $debug_value);
  }
  $icc_value = $debug_value & 0xf;
  $ioq_almost_full = ($debug_value >> 9) & 1;
  $ioq_full = ($debug_value >> 10) & 1;

  return($icc_value);

}



#-------------------------------------------------------------------------------
#
#  read_all_crb()  subroutine
#
#  Type: RRB: 0, WRB: 1, PRB: 2, IRB: 3
#
#-------------------------------------------------------------------------------

sub read_all_crb {

  my $num = $#_ + 1;
  $chip_type = "$_[0]";
  print "\nIn read_all_crb function    (@_) : num=$num  \n" if ($verbose_mode == 1);
   
  my $crb_to_mask = 0;

  my $crb;
  my $crb_retry_count = 0;
  my $crb_same_count = 0;
  my $total_bad_crb = 0;
  my $bad_crb_thresh = ($chip_type =~ /shub_01/) ? 100 : 132;	   # Actual Shub1: 104  /  Shub2: 136
  my $crb_status = 0;
  my $previous_crb_val = 0;
  my $current_crb_val = 0;
  my $lpra_zero_cnt = 0;
  my $crb_limit = 32;
  my $irb_limit = 8;
  my $prb_limit = ($chip_type =~ /shub_01/) ? 32 : 64;
  $type_shift = ($chip_type =~ /shub_01/) ? 5 : 6;
  $init_chip_err_cnt = ($chip_type =~ /shub_01/) ? $target_list[$t_cnt]->{XFR_ERR_CNT} 
						 : $sh2_target_list[$sh2_cnt]->{XFR_ERR_CNT};
  $chip_err_cnt = $init_chip_err_cnt;
  my $address = 0;
  my $nodenum = 0;
  my $cam_node = 0;
  my $ia64_comp_mask = ($chip_type =~ /shub_01/) ? 0x3ffff00000000 : 0x3fffe00000000;	# indicates IA64 compatibility space
  my $firmware_comp_mask = 0xff000000;

  @cam_regs = ($chip_type =~ /shub_01/) ? 
			(  "SH_PI_CAM_ADDRESS_READ_DATA",
                           "SH_PI_CAM_LPRA_READ_DATA",
                           "SH_PI_CAM_STATE_READ_DATA",
                           "SH_PI_CAM_CONTROL"  )
			:
			(  "SH2_PI_CAM_ADDRESS_READ_DATA",
                           "SH2_PI_CAM_LPRA_READ_DATA",
                           "SH2_PI_CAM_STATE_READ_DATA",
                           "SH2_PI_CAM_CONTROL"  );

  $cam_address_inst = reg_create_mmr_inst($chip_type, $cam_regs[0]);
  $cam_lpra_inst = reg_create_mmr_inst($chip_type, $cam_regs[1]);
  $cam_state_inst = reg_create_mmr_inst($chip_type, $cam_regs[2]);
  $cam_control_inst = reg_create_mmr_inst($chip_type, $cam_regs[3]);


#  Read RRB states

  my $crb_id = "";
  my $orig_pi_crb_cnt = $pi_crb_cnt;
  my $addr_in_hdr_bit = 0;

  if ($show_rrb_state == 1)  {
    for($crb=0; $crb < $crb_limit; $crb++) {
      my $crb_valid = read_cam(0, $crb);
      $cam_addr     = (reg_get_field_value($cam_address_inst, 'CAM_ADDR')) << 3;
      $cam_state    = reg_get_field_value($cam_state_inst, 'CAM_STATE');
      $cam_to       = reg_get_field_value($cam_state_inst, 'CAM_TO');
      $cam_rd_pend  = reg_get_field_value($cam_state_inst, 'CAM_STATE_RD_PEND');
      $lpra_val     = reg_get_field_value($cam_lpra_inst, 'CAM_LPRA');          #  LPRA Bits [63:0]
      $lpra_upr_val = reg_get_field_value($cam_state_inst, 'CAM_LPRA');         #  LPRA Bits [81:64] (shub1)  [97:64] shub2

      if (($cam_addr == 0x5555555555550 ) && ($lpra_val == 0xAAAAAAAAAAAAAAAA)) {
        $total_bad_crb++;
      }
      my $any_cam_error = ($crb_valid == 1) && (($cam_state != 0) || ($cam_to != 0) || (($cam_addr & 0x4000000000000) != 0));

      my $lpra_upr_str = ($chip_type =~ /shub_01/) ? sprintf "LPRA(81:64)=0x%04llX", $lpra_upr_val
						   : sprintf "LPRA(97:64)=0x%09llX", $lpra_upr_val;

      my $rrb_state = decode_rrb_state($cam_state);
      printf "%s: RRB[%2d]  ADDR=0x%013llX  STATE=0x%llX (%s)  TO=0x%llX  RD_PEND=0x%llX  %s  LPRA(63:0)=0x%016llX  ",
        $prefix, $crb, $cam_addr, $cam_state, $rrb_state, $cam_to, $cam_rd_pend, $lpra_upr_str, $lpra_val;

      my ($rrb_note_list, $note_string) = decode_rrb($lpra_val, $lpra_upr_val);		# return value determines if CRB is noteable
      $crb_id = sprintf "RRB[%2d]", $crb;

      if (($crb_valid == 1) && ($rrb_note_list > 0) && ($cam_state != 0)) {
        $pi_crb_list[$pi_crb_cnt]->{NASID} = $prefix;
        $pi_crb_list[$pi_crb_cnt]->{INDEX} = $crb_id;
        $pi_crb_list[$pi_crb_cnt]->{ADDR} = $cam_addr;
        $pi_crb_list[$pi_crb_cnt]->{COMMENT} = $note_string;
        $pi_crb_cnt++;
        printf "\n%s: added PI RRB %d  Address: 0x%013llX to the pi_crb_list.\n", $prefix, $crb, $cam_addr;
      }

      if (($crb_valid == 1) && ($cam_state != 0) && ($pi_xb_proto == 1)) {
	my $diff_addr = $pi_dtl1_addr ^ $cam_addr;
	my $bit_count = 0;
	my $bit_loop = 0;
	for ($bit_loop == 0; $bit_loop < 52; $bit_loop++) {
	  if ((($diff_addr >> $bit_loop) & 1) == 1) {  
	    $bit_count++;  
	    $addr_in_hdr_bit = $bit_loop;
	  }
	}
        if ($bit_count == 1)  {
          capture_aihc_err($crb_id, $addr_in_hdr_bit);
	}
        printf "\n DEBUG_MSG:  %s : Found %02d bit\(s\) in difference for crb address: 0x%013llX", $crb_id, $bit_count, $cam_addr;
        printf "\n                                        PI XB Protocol error address: 0x%013llX", $pi_dtl1_addr;
        printf "\n                                                                diff: 0x%013llX \n\n", $diff_addr;
      }
 
      if ($any_cam_error) {
	if ($shub_version == 2)  {
	  ($address, $cam_node) = fsb_to_net($cam_addr & 0x3ffffffffff80);
	  $address = $address & 0x3fffffff80;
	}
	else {
          $address = $cam_addr & $offset_mask;
          $cam_node = ($cam_addr>>$nasid_index) & $nasid_mask;           # Extract entire NASID
	}
        $nodenum = (($cam_addr & $ia64_comp_mask) == 0) ? $nasid_value : $cam_node;

	if (($address & $firmware_comp_mask) == $firmware_comp_mask) {
	  if ($remap_enabled == 1 ) {
	    $address = ($address & 0xffffff) | ($remap_offset << 24);
            save_address($prefix, $address, $nodenum, 'RRB', $node_not_booted);
	  }
	  else {
	    print "\n  Address is in firmware space - therefore will not capture directory state for it.\n";
	  }
	}
	else {
          save_address($prefix, $address, $nodenum, 'RRB', $node_not_booted);
	}
      }

      if (($lpra_val == 0) && ($lpra_upr_val == 0)) {   $lpra_zero_cnt++;   }
      $current_crb_val = reg_get_value($cam_address_inst);
      # printf "\nCurrent value: 0x%X   Previous value: 0x%X   Same count:  %d\n", $current_crb_val, $previous_crb_val, $crb_same_count;
      if ($previous_crb_val == $current_crb_val)  {
        $crb_same_count++;
      }
      else  {
        $crb_same_count = 0;
      }
      $previous_crb_val = $current_crb_val;
    }
  }

  if (($crb_same_count > 8 ) || ($lpra_zero_cnt > 4))  {
    print "\n\nWARNING:  Data returned from RRB reads appears to be unreliable!!!!\n\n";
    if ( $orig_pi_crb_cnt != $pi_crb_cnt)  {
      for (my $pp = $orig_pi_crb_cnt; $pp < $pi_crb_cnt; $pp++)  {
        $pi_crb_list[$pp]->{IGNORE} = 1;
      }
    }
  }

# Read WRB states

  if ($show_wrb_state == 1)  {
    $lpra_zero_cnt = 0;
    for($crb=0; $crb < $crb_limit; $crb++) {
      my $good_cam_read = 0;
      my $crb_valid = 0;
      my $retry_limit   = 7;
      $retry_count = 0; 		#  v 2.09  added retries in case we hit PI bug that corrupts CRB read data

      while (( $good_cam_read == 0 ) && ( $retry_count < $retry_limit )) {
        $crb_valid = read_cam(1, $crb);
        $lpra_val    = reg_get_field_value($cam_lpra_inst, 'CAM_LPRA');
        $lpra_string = decode_wrb($lpra_val);
        if (( $lpra_string !~ /GOOFY/ ) || ($file_mode == 1))  {
          $good_cam_read = 1;
        }  else  {
          $retry_count++;
          $crb_retry_count++;
        }
      }

      $cam_addr    = (reg_get_field_value($cam_address_inst, 'CAM_ADDR')) << 3;
      $cam_state   = reg_get_field_value($cam_state_inst, 'CAM_STATE');
      $cam_to      = reg_get_field_value($cam_state_inst, 'CAM_TO');
      $cam_rd_pend = reg_get_field_value($cam_state_inst, 'CAM_STATE_RD_PEND');

      if (($cam_addr == 0x5555555555550 ) && ($lpra_val == 0xAAAAAAAAAAAAAAAA)) {
        $total_bad_crb++;
      }

      my $wrb_state = decode_wrb_state($cam_state);

      my $any_cam_error = ($crb_valid == 1) && (($cam_state != 0) || ($cam_to != 0) || (($cam_addr & 0x4000000000000) != 0));

      $crb_id = sprintf "WRB[%2d]", $crb;

      printf "%s: WRB[%2d]  ADDR=0x%013llX  STATE=0x%llX (%s)  TO=0x%llX  RD_PEND=0x%llX  LPRA(63:0)=0x%016llX  %s",
        $prefix, $crb, $cam_addr, $cam_state, $wrb_state, $cam_to, $cam_rd_pend, $lpra_val, $lpra_string;

      if ($any_cam_error) {
        if ($shub_version == 2)  {
          ($address, $cam_node) = fsb_to_net($cam_addr & 0x3ffffffffff80);
	  $address = $address & 0x3fffffff80;
        }
        else {
          $address = $cam_addr & $offset_mask;
          $cam_node = ($cam_addr>>$nasid_index) & $nasid_mask;           # Extract entire NASID
        }
        if (($chip_type =~ /shub2_01/) || ($cam_node  != ($nasid_value ^ 0x400)))    # WRB timed out in boot prom
        {
          $nodenum = (($cam_addr & $ia64_comp_mask) == 0) ? $nasid_value : $cam_node;
          if (($address & $firmware_comp_mask) == $firmware_comp_mask) {
            if ($remap_enabled == 1 ) {
              $address = ($address & 0xffffff) | ($remap_offset << 24);
              save_address($prefix, $address, $nodenum, 'WRB', $node_not_booted);
            }
            else {
              print "\n  Address is in firmware space - therefore will not capture directory state for it.\n";
            }
          }
          else {
            save_address($prefix, $address, $nodenum, 'WRB', $node_not_booted);
	  }
        }
        if ( ($chip_type =~ /shub_01/) && ($cam_node  == ($nasid_value ^ 0x400)) && ($cam_to == 1))  {
          $target_list[$t_cnt]->{WRB_DEALLOC} ++;
          $wrb_d_cnt++;
        }
      }

      if (($crb_valid == 1) && ($cam_state != 0) && ($pi_xb_proto == 1)) {
        my $diff_addr = $pi_dtl1_addr ^ $cam_addr;
        my $bit_count = 0;
        my $bit_loop = 0;
        for ($bit_loop == 0; $bit_loop < 52; $bit_loop++) {
          if ((($diff_addr >> $bit_loop) & 1) == 1) {
            $bit_count++;
            $addr_in_hdr_bit = $bit_loop;
          }
        }
        if ($bit_count == 1)  {
          capture_aihc_err($crb_id, $addr_in_hdr_bit);
        }
        printf "\n DEBUG_MSG:  %s : Found %02d bit\(s\) in difference for crb address: 0x%013llX", $crb_id, $bit_count, $cam_addr;
        printf "\n                                        PI XB Protocol error address: 0x%013llX", $pi_dtl1_addr;
        printf "\n                                                                diff: 0x%013llX \n\n", $diff_addr;
      }

      $current_crb_val = reg_get_value($cam_address_inst);
      # printf "\nCurrent value: 0x%X   Previous value: 0x%X   Same count:  %d\n", $current_crb_val, $previous_crb_val, $crb_same_count;

      if (($lpra_val == 0) && ($lpra_upr_val == 0)) {   $lpra_zero_cnt++;   }

      if ($previous_crb_val == $current_crb_val)  {
        $crb_same_count++;
      }
      else  {
        $crb_same_count = 0;
      }
      $previous_crb_val = $current_crb_val;

    }
  }

  if (($crb_same_count > 8 ) || ($lpra_zero_cnt > 4))  {
    print "\n\nWARNING:  Data returned from WRB reads appears to be unreliable!!!!\n\n";
  }


# Read IRB states

  if ($show_irb_state == 1)  {
    $lpra_zero_cnt = 0;
    for($crb=0; $crb < $irb_limit; $crb++) {
      my $crb_valid = read_cam(3, $crb);
      $cam_addr    = (reg_get_field_value($cam_address_inst, 'CAM_ADDR')) << 3;
      $cam_state   = reg_get_field_value($cam_state_inst, 'CAM_STATE');
      $cam_to      = reg_get_field_value($cam_state_inst, 'CAM_TO');
      $lpra_val    = reg_get_field_value($cam_lpra_inst, 'CAM_LPRA');
      $cam_rp      = reg_get_field_value($cam_state_inst, 'CAM_STATE_RD_PEND');

      if (($cam_addr == 0x5555555555550 ) && ($lpra_val == 0xAAAAAAAAAAAAAAAA)) {
        $total_bad_crb++;
      }
      $lpra_string = decode_irb($lpra_val);
      my $irb_state = decode_irb_state($cam_state);
      $crb_id = sprintf "IRB[%2d]", $crb;

      printf "%s: IRB[%2d]  ADDR=0x%013llX  STATE=0x%llX (%s)  TO=0x%llX  RD_PEND=0x%llX  LPRA(63:0)=0x%09llX %s\n", 
	$prefix, $crb, $cam_addr, $cam_state, $irb_state, $cam_to, $cam_rp, $lpra_val, $lpra_string;

      my $any_cam_error = ($crb_valid == 1) && (($cam_state != 0) || ($cam_to != 0) || (($cam_addr & 0x4000000000000) != 0));

      if ($any_cam_error) {
        if ($shub_version == 2)  {
          ($address, $cam_node) = fsb_to_net($cam_addr & 0x3ffffffffff80);
	  $address = $address & 0x3fffffff80;
        }
        else {
          $address = $cam_addr & $offset_mask;
          $cam_node = ($cam_addr>>$nasid_index) & $nasid_mask;           # Extract entire NASID
        }
        $nodenum = (($cam_addr & $ia64_comp_mask) == 0) ? $nasid_value : $cam_node;
        if (($address & $firmware_comp_mask) == $firmware_comp_mask) {
          if ($remap_enabled == 1 ) {
            $address = ($address & 0xffffff) | ($remap_offset << 24);
            save_address($prefix, $address, $nodenum, 'IRB', $node_not_booted);
          }
          else {
            print "\n  Address is in firmware space - therefore will not capture directory state for it.\n";
          }
        }
        else {
          save_address($prefix, $address, $nodenum, 'IRB', $node_not_booted);
	}
      }

      if (($crb_valid == 1) && ($cam_state != 0) && ($pi_xb_proto == 1)) {
        my $diff_addr = $pi_dtl1_addr ^ $cam_addr;
        my $bit_count = 0;
        my $bit_loop = 0;         for ($bit_loop == 0; $bit_loop < 52; $bit_loop++) {
          if ((($diff_addr >> $bit_loop) & 1) == 1) {
            $bit_count++;
            $addr_in_hdr_bit = $bit_loop;
          }
        }
        if ($bit_count == 1)  {
          capture_aihc_err($crb_id, $addr_in_hdr_bit);
        }
      }


      $current_crb_val = reg_get_value($cam_address_inst);
      # printf "\nCurrent value: 0x%X   Previous value: 0x%X   Same count:  %d\n", $current_crb_val, $previous_crb_val, $crb_same_count;

      if (($lpra_val == 0) && ($lpra_upr_val == 0)) {   $lpra_zero_cnt++;   }

      if ($previous_crb_val == $current_crb_val)  {
        $crb_same_count++;
      }
      else  {
        $crb_same_count = 0;
      }
      $previous_crb_val = $current_crb_val;

    }
  }

  if (($crb_same_count > 8 ) || ($lpra_zero_cnt > 4))  {
    print "\n\nWARNING:  Data returned from IRB reads appears to be unreliable!!!!\n\n";
  }

# Read PRB states

  if ($show_prb_state == 1)  {
    $lpra_zero_cnt = 0;
    for($crb=0; $crb < $prb_limit; $crb++) {
      my $crb_valid = read_cam(2, $crb);
      $cam_addr    = (reg_get_field_value($cam_address_inst, 'CAM_ADDR')) << 3;
      $cam_state   = reg_get_field_value($cam_state_inst, 'CAM_STATE');
      $cam_to      = reg_get_field_value($cam_state_inst, 'CAM_TO');
      $lpra_val    = reg_get_field_value($cam_lpra_inst,'CAM_LPRA');
      $cam_rp      = reg_get_field_value($cam_state_inst, 'CAM_STATE_RD_PEND');

      if (($cam_addr == 0x5555555555550 ) && ($lpra_val == 0xAAAAAAAAAAAAAAAA)) {
        $total_bad_crb++;
      }
      my $prb_state = decode_prb_state($cam_state);
      $lpra_string = decode_prb($lpra_val);
      $crb_id = sprintf "PRB[%2d]", $crb;

      printf "%s: PRB[%2d]  ADDR=0x%013llX  STATE=0x%llX (%s)  TO=0x%llX  RD_PEND=0x%llX  LPRA(63:0)=0x%07llX %s\n", 
	$prefix, $crb, $cam_addr, $cam_state, $prb_state, $cam_to, $cam_rp, $lpra_val, $lpra_string;

      $any_cam_error = ($crb_valid == 1) && (($cam_state != 0) || ($cam_to != 0) || (($cam_addr & 0x4000000000000) != 0));

      if ($any_cam_error) {
        $address = $cam_addr & $offset_mask;
        $nodenum = ($cam_addr>>$nasid_index) & $nasid_mask;           # Extract entire NASID
        save_pio_address($prefix, $address, $nodenum, $crb);
      }

      if (($crb_valid == 1) && ($cam_state != 0) && ($pi_xb_proto == 1)) {
        my $diff_addr = $pi_dtl1_addr ^ $cam_addr;
        my $bit_count = 0;
        my $bit_loop = 0;         
        for ($bit_loop == 0; $bit_loop < 52; $bit_loop++) {           
          if ((($diff_addr >> $bit_loop) & 1) == 1) {
            $bit_count++;
            $addr_in_hdr_bit = $bit_loop;
          }
        }
        if ($bit_count == 1)  {
          capture_aihc_err($crb_id, $addr_in_hdr_bit);
        }
      }

      $current_crb_val = reg_get_value($cam_address_inst);
      $current_crb_val = reg_get_value($cam_address_inst);
      # printf "\nCurrent value: 0x%X   Previous value: 0x%X   Same count:  %d\n", $current_crb_val, $previous_crb_val, $crb_same_count;

      if (($lpra_val == 0) && ($lpra_upr_val == 0)) {   $lpra_zero_cnt++;   }

      if (($crb < 32) && ($previous_crb_val == $current_crb_val))  {    # GEM   Temporary WAR for Hyper-thread disabled processors
      # if ($previous_crb_val == $current_crb_val)       # GEM   Original statement
        $crb_same_count++;
      }
      else  {
        $crb_same_count = 0;
      }
      $previous_crb_val = $current_crb_val;

    }
  }

  if (($crb_same_count > 8 ) || ($lpra_zero_cnt > 4))  {
    print "\n\nWARNING:  Data returned from PRB reads appears to be unreliable!!!!\n\n";
  }

  if (($verbose_mode == 1) && ($crb_retry_count > 0))  {
    print "\nThere were $crb_retry_count retries in read_all_crb function\n";
  }

  if ($init_chip_err_cnt != $chip_err_cnt)  {       # update counters if errors occurred
    if ($chip_type =~ /shub_01/)  {
      $target_list[$t_cnt]->{XFR_ERR_CNT} = $chip_err_cnt;
    }
    else {
      $sh2_target_list[$sh2_cnt]->{XFR_ERR_CNT} = $chip_err_cnt;
    }
  }

  if ($total_bad_crb > $bad_crb_thresh)  {
    $crb_status = 1;		#  1 is bad
    my $crb_err_msg = sprintf "   %s: Possible PI Hang - all PI Cam accesses are failing.\n", $prefix;
    if ($chip_type =~ /shub_01/)  {
      if (defined $target_list[$t_cnt]->{PI_ERR_MSG}) {
        $target_list[$t_cnt]->{PI_ERR_MSG} .= $crb_err_msg;
      }
      else {
        $target_list[$t_cnt]->{PI_ERR_MSG} = $crb_err_msg;
      }
    }
    else {
      if (defined $sh2_target_list[$sh2_cnt]->{PI_ERR_MSG}) {
        $sh2_target_list[$sh2_cnt]->{PI_ERR_MSG} .= $crb_err_msg;
      }
      else {
        $sh2_target_list[$sh2_cnt]->{PI_ERR_MSG} = $crb_err_msg;
      }
    }
  }

  return($crb_status);

}

#-------------------------------------------------------------------------------
#
#  subroutine: read_cam
#
#	type values:  0 = RRB     
#		      1 = WRB   
#		      2 = PRB
#		      3 = IRB
#
#  If this function is called when using a fru file ($file_mode == 1), the
#  values are looked up from the arrays that were loaded up in parse_fru_crb()
#  within fru_funcs.pm.
#  Otherwise, read the data values directly from the HW via jtag accesses.
#
#-------------------------------------------------------------------------------

sub read_cam {

  print "\nIn read_cam function - ( @_ )\n" if ($verbose_mode == 1);

  my $cam_type = $_[0];
  my $index = $_[1];
  my $cam_read_ok = 0;
  my $cam_err_val = 0xAAAAAAAAAAAAAAAA;

  if ($file_mode == 1) {			#  defined in errdmp
    my $cam_addr_val = $cam_err_val;
    my $cam_state_val = $cam_err_val;
    my $cam_lpra_val = $cam_err_val;

    if ($cam_type == 0) {
      $cam_addr_val = $fru_rrb[$index]->{CAM_ADDR};
      $cam_state_val = $fru_rrb[$index]->{CAM_STATE};
      $cam_lpra_val = $fru_rrb[$index]->{CAM_LPRA};
    }
    elsif ($cam_type == 1) {
      $cam_addr_val = $fru_wrb[$index]->{CAM_ADDR};
      $cam_state_val = $fru_wrb[$index]->{CAM_STATE};
      $cam_lpra_val = $fru_wrb[$index]->{CAM_LPRA};
    }
    elsif ($cam_type == 2) {
      $cam_addr_val = $fru_prb[$index]->{CAM_ADDR};
      $cam_state_val = $fru_prb[$index]->{CAM_STATE};
      $cam_lpra_val = $fru_prb[$index]->{CAM_LPRA};
    }
    elsif ($cam_type == 3) {
      $cam_addr_val = $fru_irb[$index]->{CAM_ADDR};
      $cam_state_val = $fru_irb[$index]->{CAM_STATE};
      $cam_lpra_val = $fru_irb[$index]->{CAM_LPRA};
    }

    reg_set_value($cam_address_inst, $cam_addr_val);
    reg_set_value($cam_lpra_inst, $cam_lpra_val);
    reg_set_value($cam_state_inst, $cam_state_val);
    $cam_read_ok = (($cam_addr_val == $cam_err_val)
                  ||($cam_lpra_val == $cam_err_val)
                  ||($cam_state_val == $cam_err_val)) ? 0 : 1;

    if ($verbose_mode == 1) {
     printf "\nThe value for cam_type %d index %d  CAM address: 0x%016llx", $cam_type, $index, reg_get_value($cam_address_inst);
     printf "\nThe value for cam_type %d index %d  CAM lpra   : 0x%016llx", $cam_type, $index, reg_get_value($cam_lpra_inst);
     printf "\nThe value for cam_type %d index %d  CAM state  : 0x%016llx\n", $cam_type, $index, reg_get_value($cam_state_inst);
    }
  }
  else {
    my $data = 0;
    my $cnt = 0;
  
    my @wr_adds = ();
    my @wr_vals = ();
  
    my @adds    = ();
    my @vals    = ();
  
    my $cam_index = ($index & 0x3f) | (($cam_type & 0x3) << $type_shift);

    reg_set_value($cam_address_inst, 0);
    reg_set_value($cam_lpra_inst, 0);
    reg_set_value($cam_state_inst, 0);

    reg_set_value($cam_control_inst, 0);
    reg_set_field_value($cam_control_inst, 'CAM_INDX', $cam_index);
    reg_set_field_value($cam_control_inst, 'START', 1);

    $wr_adds[0] = reg_addr($cam_address_inst);
    $wr_adds[1] = reg_addr($cam_state_inst);
    $wr_adds[2] = reg_addr($cam_control_inst);
    $wr_vals[0] = 0;
    $wr_vals[1] = 0;
    $wr_vals[2] = reg_get_value($cam_control_inst);

    my $xfr_complete = 0;
    my $xfr_f_cnt = 0;
    while (($xfr_complete == 0) && ($xfr_f_cnt < 3) && ($chip_err_cnt < 6))  {
      $SSCOPE_ERR = 0;
      mmr_write_group(\@wr_adds, \@wr_vals);                    #  return mask setting to original value
      if ($SSCOPE_ERR == 1)  {
        $xfr_f_cnt++;
        $chip_err_cnt++;
        $SSCOPE_ERR = 0;
      }
      else {
        $xfr_complete = 1;
      }
    }

    $adds[0] = reg_addr($cam_address_inst);
    $adds[1] = reg_addr($cam_state_inst);

    do {
      if($cnt++ > 23) {
        if((reg_get_field_value($cam_state_inst, 'CAM_RD_DATA_VAL')) == 0) {
          print "ERROR: Timed out waiting for CAM_RD_DATA_VAL to set - type = $cam_type  index = $index\n";
        }
  
        if((reg_get_field_value($cam_address_inst, 'CAM_ADDR_VAL')) == 0) {
          print "ERROR: Timed out waiting for CAM_ADDR_VAL to set - type = $cam_type  index = $index\n";
        }
      }
  
      if ($cnt > 2)  {
        my $xfr_complete = 0;
        my $xfr_f_cnt = 0;
        while (($xfr_complete == 0) && ($xfr_f_cnt < 3) && ($chip_err_cnt < 6))  {
          $SSCOPE_ERR = 0;
          reg_mmr_write($cam_control_inst);
          if ($SSCOPE_ERR == 1)  {
            $xfr_f_cnt++;
            $chip_err_cnt++;
            $SSCOPE_ERR = 0;
          }
          else {
            $xfr_complete = 1;
          }
        }
      }
      #  Read the two MMRs
      $xfr_complete = 0;
      $xfr_f_cnt = 0;
      while (($xfr_complete == 0) && ($xfr_f_cnt < 3) && ($chip_err_cnt < 6))  {
        $SSCOPE_ERR = 0;
        @vals = mmr_read_group(\@adds);
        if ($SSCOPE_ERR == 1)  {
          $xfr_f_cnt++;
          $chip_err_cnt++;
          $SSCOPE_ERR = 0;
        }
        else {
          $xfr_complete = 1;
        }
      }

      reg_set_value($cam_address_inst, $vals[0]);
      reg_set_value($cam_state_inst, $vals[1]);
  
    } while( ( $cnt < 24) && (((reg_get_field_value($cam_state_inst, 'CAM_RD_DATA_VAL')) == 0 ) ||
             ((reg_get_field_value($cam_address_inst, 'CAM_ADDR_VAL')) == 0 )) );
  
    print "\nThe value of \$cnt (number of read attempts) is $cnt.\n" if ($verbose_mode == 1);

    if ($cnt > 23)  {
      reg_set_value($cam_address_inst, $cam_err_val);
      reg_set_value($cam_lpra_inst, $cam_err_val);
      reg_set_value($cam_state_inst, $cam_err_val);
    }
    else {
      $cam_read_ok = 1;
      $xfr_complete = 0;
      $xfr_f_cnt = 0;
      while (($xfr_complete == 0) && ($xfr_f_cnt < 3) && ($chip_err_cnt < 6))  {
        $SSCOPE_ERR = 0;
        reg_mmr_read($cam_lpra_inst);
        if ($SSCOPE_ERR == 1)  {
          $xfr_f_cnt++;
          $chip_err_cnt++;
          $SSCOPE_ERR = 0;
        }
          else {
          $xfr_complete = 1;
        }
      }
    }
  }
 
  return($cam_read_ok);

}


#-------------------------------------------------------------------------------
#
#  subroutine: decode_wrb
#
#  This subroutine decodes the WRB lpra bit group read from the SH_PI_CAM_LPRA_READ_DATA
#  mmr (bits [63:0]).
#
#-------------------------------------------------------------------------------

sub decode_wrb {

    print "\nIn decode_wrb function - ( @_ )\n" if ($verbose_mode == 1);

    my $return_string = "";

    local($lpra) = $_[0];                       #  LPRA bits [63:0]

    if ($chip_type =~ /shub_01/)  {
      $lpra_type = $lpra & 0x1f;
      $lpra_b = ($lpra >> 5) & 0x3;
      $lpra_p = ($lpra >> 7) & 0x1f;
      $lpra_hs = ($lpra >> 12) & 0x3fff;
      $lpra_echo = ($lpra >> 26) & 0x1ff;
      $lpra_ha63 = ($lpra >> 35) & 0xf;
      $lpra_a63 = ($lpra >> 39) & 0xf;
      $lpra_len = ($lpra >> 43) & 0x7;
      $lpra_be = ($lpra >> 46) & 0xff;
      $lpra_mbz = ($lpra >> 54) & 0xf;
    }
    else {
      $lpra_type = $lpra & 0x1f;
      $lpra_b = ($lpra >> 5) & 0x3;
      $lpra_p = ($lpra >> 7) & 0x1f;
      $lpra_hs = ($lpra >> 12) & 0xffff;
      $lpra_echo = ($lpra >> 28) & 0x1ff;
      $lpra_ha63 = ($lpra >> 37) & 0xf;
      $lpra_a63 = ($lpra >> 41) & 0xf;
      $lpra_len = ($lpra >> 45) & 0x7;
      $lpra_be = ($lpra >> 48) & 0xff;
      $lpra_aid = ($lpra >> 56) & 0xf;
      $lpra_mbz = ($lpra >> 60) & 0xf;
      my $proc_agent = (($sh2_target_list[$sh2_cnt]->{AID_DECODE} & 1) == 1) ? (($lpra_aid >> 1) & 1) : (($lpra_aid >> 2) & 1);
      my $proc_core =  (($sh2_target_list[$sh2_cnt]->{AID_DECODE} & 1) == 1) ? (($lpra_aid >> 2) & 1) : 0;
      my $proc_thread = (($sh2_target_list[$sh2_cnt]->{AID_DECODE} & 1) == 1) ? (($lpra_aid >> 3) & 1) : 0;
      $agent_id = sprintf "PROC:%d  CORE:%d  THREAD:%d", $proc_agent, $proc_core, $proc_thread; 
    }

    $type = &decode_wrb_type($lpra_type);
    $p = &decode_wrb_type($lpra_p);

    # WRB busy ack received (B field)
    if ($lpra_b == 0) {$b = "NONE";}
    if ($lpra_b == 1) {$b = "EACK";}
    if ($lpra_b == 2) {$b = "IACK";}
    if ($lpra_b == 3) {$b = "HOME";}

    if ($chip_type =~ /shub_01/)  {
      $return_string = sprintf ("T(%s) B(%s) P(%s) HS(0x%04X) ECHO(0x%03X) HA63(0x%X) A63(0x%X) LEN(0x%X) BE(0x%02X) MBZ(0x%X)\n",
        $type, $b, $p, $lpra_hs, $lpra_echo, $lpra_ha63, $lpra_a63, $lpra_len, $lpra_be, $lpra_mbz);
    }
    else {
      $return_string = sprintf ("T(%s) B(%s) P(%s) HS(0x%04X) ECHO(0x%03X) HA63(0x%X) A63(0x%X) LEN(0x%X) BE(0x%02X) AGENT_ID(0x%X  %s) MBZ(0x%X)\n",
        $type, $b, $p, $lpra_hs, $lpra_echo, $lpra_ha63, $lpra_a63, $lpra_len, $lpra_be, $lpra_aid, $agent_id, $lpra_mbz);
    }

    return ($return_string);
}


#-------------------------------------------------------------------------------
#
#
#-------------------------------------------------------------------------------

sub decode_wrb_type {

    print "\nIn decode_wrb_type function - ( @_ )\n" if ($verbose_mode == 1);

    local($raw_type) = $_[0];
    local($type_str) = "GOOFY";

    # WRB request type (T and P fields)
    if ($raw_type == 0) {$type_str = "NONE";}
    if ($raw_type == 17) {$type_str = "BWL";}
    if ($raw_type == 18) {$type_str = "BWLR";}
    if ($raw_type == 19) {$type_str = "BRQSH";}
    if ($raw_type == 20) {$type_str = "BRQHR";}
    if ($raw_type == 21) {$type_str = "BWLN";}
    if ($raw_type == 22) {$type_str = "BWLRN";}
    if ($raw_type == 23) {$type_str = "BIWE";}
    if ($raw_type == 24) {$type_str = "BCCL";}
    if ($raw_type == 28) {$type_str = "BWPD";}
    if ($raw_type == 29) {$type_str = "BWPF";}
    if (($raw_type == 30) && ($chip_type =~ /shub_01/))  {$type_str = "BAMOW";}
    if (($raw_type == 25) && ($chip_type =~ /shub2_01/))  {$type_str = "BAMOW";}
    if (($raw_type == 30) && ($chip_type =~ /shub2_01/))  {$type_str = "BWWCD";}
    if (($raw_type == 31) && ($chip_type =~ /shub2_01/))  {$type_str = "BWWCF";}

    return($type_str);
}



#-------------------------------------------------------------------------------
#
#  subroutine: decode_rrb
#
#  This subroutine decodes the 2 RRB lpra bit groups read from the SH_PI_CAM_LPRA_READ_DATA
#  mmr (bits [63:0]) and SH_PI_CAM_STATE_READ_DATA (bits [81:64]).
#
#-------------------------------------------------------------------------------

sub decode_rrb {

    print "\nIn decode_rrb function - ( @_ )\n" if ($verbose_mode == 1);

    local($lpra) = $_[0];                       #  LPRA bits [63:0]
    local($lpra_u) = $_[1];                     #  LPRA bits [81:64] (shub1)  [97:64] (shub2)

    if ($chip_type =~ /shub_01/)  {
      $lpra_type = $lpra & 0x7;
      $lpra_shd = ($lpra >> 3) & 0x1;
      $lpra_inval_cnt = ($lpra >> 4) & 0x3ff;
      $lpra_doom = ($lpra >> 14) & 0x1;
      $lpra_kill = ($lpra >> 15) & 0x1;
      $lpra_e = ($lpra >> 16) & 0x3;
      $lpra_p = ($lpra >> 18) & 0x3;
      $lpra_h = ($lpra >> 20) & 0xf;
      $lpra_hs = ($lpra >> 24) & 0x3fff;
      $lpra_echo = ($lpra >> 38) & 0x1ff;
      $lpra_ha63 = ($lpra >> 47) & 0xf;
      $lpra_nack = ($lpra >> 51) & 0xfff;
      $lpra_did = (($lpra >> 63) & 0x1) + (($lpra_u << 1) & 0x3fe);
      $lpra_a63 = ($lpra_u >> 9) & 0xf;
      $lpra_len = ($lpra_u >> 13) & 0x7;
      $lpra_err = ($lpra_u >> 16) & 0x1;
      $lpra_mbz = ($lpra_u >> 17) & 0x1;
      $agent_id = sprintf "PROC:%d", (($lpra_did >> 6) & 1); 
    }
    else {				#  use shub2 field widths
      $lpra_type = $lpra & 0xf;
      $lpra_shd = ($lpra >> 4) & 0x1;
      $lpra_inval_cnt = ($lpra >> 5) & 0x3ff;
      $lpra_kill = ($lpra >> 16) & 0x1;
      $lpra_e = ($lpra >> 17) & 0x3;
      $lpra_p = ($lpra >> 19) & 0x3;
      $lpra_h = ($lpra >> 21) & 0xf;
      $lpra_hs = ($lpra >> 25) & 0xffff;
      $lpra_echo = ($lpra >> 41) & 0x1ff;
      $lpra_ha63 = ($lpra >> 50) & 0xf;
      $lpra_nack = (($lpra >> 54) & 0x3ff) + ((lpra_u & 0x1f) << 10);
      $lpra_did = ($lpra_u >> 5) & 0x3ff;
      $lpra_a63 = ($lpra_u >> 15) & 0xf;
      $lpra_len = ($lpra_u >> 19) & 0x7;
      $lpra_be = ($lpra_u >> 22) & 0xff;
      $lpra_hfail = ($lpra_u >> 30) & 0x1;
      $lpra_poison = ($lpra_u >> 31) & 0x1;
      $lpra_mbz = ($lpra_u >> 17) & 0x1;
      my $proc_agent = (($sh2_target_list[$sh2_cnt]->{AID_DECODE} & 1) == 1) ? (($lpra_did >> 6) & 1) : (($lpra_did >> 7) & 1);
      my $proc_core =  (($sh2_target_list[$sh2_cnt]->{AID_DECODE} & 1) == 1) ? (($lpra_did >> 7) & 1) : 0;
      my $proc_thread = (($sh2_target_list[$sh2_cnt]->{AID_DECODE} & 1) == 1) ? (($lpra_did >> 8) & 1) : 0;
      $agent_id = sprintf "PROC:%d  CORE:%d  THREAD:%d", $proc_agent, $proc_core, $proc_thread; 
    }

    # RRB request type actions
    if ($lpra_type == 0) {$type = "BRLC";}
    if ($lpra_type == 1) {$type = "BRLD";}
    if ($lpra_type == 2) {$type = "BRIL";}
    if ($lpra_type == 3) {$type = "BFCL";}
    if ($lpra_type == 4) {$type = "BFCLG";}
    if ($lpra_type == 5) {$type = "BRP";}
    if ($lpra_type == 6) {$type = "BGET";}              #
    if ($lpra_type == 7) {$type = "BPRM";}              #  Prom Read
    if ($lpra_type == 8) {$type = "BRPA";}              #  Prom Read
    if ($lpra_type == 9) {$type = "BRWCD";}              #  Prom Read
    if ($lpra_type == 10) {$type = "BRWCF";}              #  Prom Read

    # RRB spec data tracking (E field)
    if ($lpra_e == 0) {$e = "NONE";}
    if ($lpra_e == 1) {$e = "DATA";}
    if ($lpra_e == 2) {$e = "ACK ";}

    # RRB pending request type (P field)
    if ($lpra_p == 0) {$p = "NONE";}
    if ($lpra_p == 1) {$p = "IWWRB";}
    if ($lpra_p == 2) {$p = "IW";}
    if ($lpra_p == 3) {$p = "BFCL";}

    # RRB held intervention type actions
    if ($lpra_h == 0) {$h = "NONE";}
    if ($lpra_h == 1) {$h = "INVAL";}
    if ($lpra_h == 9) {$h = "INTER";}
    if ($lpra_h == 10) {$h = "INEXC";}
    if ($lpra_h == 11) {$h = "ININV";}
    if ($lpra_h == 12) {$h = "ININF";}
    if ($lpra_h == 13) {$h = "FLSH";}
    if ($lpra_h == 14) {$h = "ERASE";}

    my $rrb_comment = "";
    my $noteable = 0;

    if ($lpra_h != 0) {
      $noteable += 1;
      $rrb_comment .= sprintf "Held Intervention type is %s.  ", $h;
    }
    if ($lpra_inval_cnt > 0) { 
      $noteable += 2;
      $rrb_comment .= sprintf "Invalidation Acknowledgement Count is %d.  ", $lpra_inval_cnt;
    }

    if ($chip_type =~ /shub_01/)  {
      printf "T(%s) SHD(%X) INVAL_CNT(0x%03X) DOOM(%X) KILL(%X) E(%s) PEND(%s) HELD(%s) HS(0x%04X) ECHO(0x%03X) HA63(0x%X) NACK(0x%03X) DID(0x%03X %s) A63(0x%X) LEN(0x%X) ERR(%X) MBZ(%X)\n",
        $type, $lpra_shd, $lpra_inval_cnt, $lpra_doom, $lpra_kill, $e, $p, $h, $lpra_hs, $lpra_echo,
        $lpra_ha63, $lpra_nack, $lpra_did, $agent_id, $lpra_a63, $lpra_len, $lpra_err, $lpra_mbz;
    }
    else {
      printf "T(%s) SHD(%X) INVAL_CNT(0x%03X) KILL(%X) E(%s) PEND(%s) HELD(%s) HS(0x%04X) ECHO(0x%03X) HA63(0x%X) NACK(0x%04X) DID(0x%03X %s) A63(0x%X) LEN(0x%X) BE(0x%02X) HFAIL(%X) POISON(%X) MBZ(%X)\n",
        $type, $lpra_shd, $lpra_inval_cnt, $lpra_kill, $e, $p, $h, $lpra_hs, $lpra_echo,
        $lpra_ha63, $lpra_nack, $lpra_did, $agent_id, $lpra_a63, $lpra_len, $lpra_be, $lpra_hfail, $lpra_poison, $lpra_mbz;
    }

    return ($noteable, $rrb_comment);

}

#-------------------------------------------------------------------------------
#
#  subroutine: decode_irb
#
#  This subroutine decodes the IRB lpra bit group read from the SH_PI_CAM_LPRA_READ_DATA
#  mmr (bits [35:0]).
#
#-------------------------------------------------------------------------------

sub decode_irb {

    print "\nIn decode_irb function - ( @_ )\n" if ($verbose_mode == 1);

    my $return_string = "";

    local($lpra) = $_[0];                       #  LPRA bits [35:0]

    if ($chip_type =~ /shub_01/)  {
      $lpra_type = $lpra & 0x1f;
      $lpra_is = ($lpra >> 5) & 0x3fff;
      $lpra_echo = ($lpra >> 19) & 0x1ff;
      $lpra_a63 = ($lpra >> 28) & 0xf;
      $lpra_mbz = ($lpra >> 32) & 0xf;
    }
    else {
      $lpra_type = $lpra & 0x1f;
      $lpra_is = ($lpra >> 5) & 0xffff;
      $lpra_echo = ($lpra >> 21) & 0x1ff;
      $lpra_a63 = ($lpra >> 30) & 0xf;
      $lpra_mbz = ($lpra >> 32) & 0x3;
    }

    $type = decode_irb_type($lpra_type);

    $return_string = sprintf (" T(%s) IS(0x%04X) ECHO(0x%03X) A63(0x%X) MBZ(0x%X)",
        $type, $lpra_is, $lpra_echo, $lpra_a63, $lpra_mbz);

    return ($return_string);
}


#-------------------------------------------------------------------------------
#
#  subroutine: decode_prb
#
#  This subroutine decodes the PRB lpra bit group read from the SH_PI_CAM_LPRA_READ_DATA mmr.
#
#-------------------------------------------------------------------------------

sub decode_prb {

    print "\nIn decode_prb function - ( @_ )\n" if ($verbose_mode == 1);

    my $return_string = "";

    local($lpra) = $_[0];                       #  LPRA bits [35:0]

    if ($chip_type =~ /shub_01/)  {
      $lpra_mbz = $lpra & 0x1;
      $lpra_did = ($lpra >> 1) & 0x3ff;
      $lpra_len = ($lpra >> 11) & 0x3;
      $lpra_be  = ($lpra >> 13) & 0xff;
      $lpra_r   = ($lpra >> 21) & 0x3;
      $agent_id = sprintf "PROC:%d", (($lpra_did >> 6) & 1);
    }
    else {
      $lpra_did = $lpra & 0x3ff;
      $lpra_len = ($lpra >> 10) & 0x3;
      $lpra_be  = ($lpra >> 12) & 0xff;
      $lpra_r   = ($lpra >> 20) & 0x3;
      $lpra_hf  = ($lpra >> 22) & 0x1;
      $lpra_pois = ($lpra >> 23) & 0x1;
      $lpra_mbz = ($lpra >> 24) & 0x3;
      my $proc_agent = (($sh2_target_list[$sh2_cnt]->{AID_DECODE} & 1) == 1) ? (($lpra_did >> 6) & 1) : (($lpra_did >> 7) & 1);
      my $proc_core =  (($sh2_target_list[$sh2_cnt]->{AID_DECODE} & 1) == 1) ? (($lpra_did >> 7) & 1) : 0;
      my $proc_thread = (($sh2_target_list[$sh2_cnt]->{AID_DECODE} & 1) == 1) ? (($lpra_did >> 8) & 1) : 0;
      $agent_id = sprintf "PROC:%d  CORE:%d  THREAD:%d", $proc_agent, $proc_core, $proc_thread; 
    }

    $type = ($lpra_r == 0) ? "WRITE" 
	  : ($lpra_r == 1) ? "READ" : "????";

    if ($chip_type =~ /shub_01/)  {
      $return_string = sprintf (" R(%s) LEN(0x%X) BE(0x%02X) DID(0x%03X %s) MBZ(0x%X)",
        $type, $lpra_len, $lpra_be, $lpra_did, $agent_id, $lpra_mbz);
    }
    else {
      $return_string = sprintf (" R(%s) LEN(0x%X) BE(0x%02X) DID(0x%03X %s) HARD_FAIL(0x%X) POISON(0x%X) MBZ(0x%X)",
        $type, $lpra_len, $lpra_be, $lpra_did, $agent_id, $lpra_hf, $lpra_pois, $lpra_mbz);
    }

    return ($return_string);
}



#-------------------------------------------------------------------------------
#
# Combines original setting of the SH_CRB_MESSAGE_CONTROL register with settings to
# enable flush interval of 1 and flush duration of 1
#
#-------------------------------------------------------------------------------

sub enable_single_flush
{
  my $msg_ctl_val = ( $crb_msg_ctl & 0xffffffff) | (1 << 32) | (1 << 48);
  my $xfr_complete = 0;
  my $xfr_f_cnt = 0;
  while (($xfr_complete == 0) && ($xfr_f_cnt < 3) && ($target_list[$t_cnt]->{XFR_ERR_CNT} < 6))  {
    $SSCOPE_ERR = 0;
    mmr_write (0x20000000, $msg_ctl_val);
    if ($SSCOPE_ERR == 1)  {
      $xfr_f_cnt++;
      $target_list[$t_cnt]->{XFR_ERR_CNT}++;
      $SSCOPE_ERR = 0;
    }
    else {
      $xfr_complete = 1;
    }
  }
}


#-------------------------------------------------------------------------------
#
# Restores original setting of the SH_CRB_MESSAGE_CONTROL register
#
#-------------------------------------------------------------------------------

sub disable_single_flush
{
  my $xfr_complete = 0;
  my $xfr_f_cnt = 0;
  while (($xfr_complete == 0) && ($xfr_f_cnt < 3) && ($target_list[$t_cnt]->{XFR_ERR_CNT} < 6))  {
    $SSCOPE_ERR = 0;
    mmr_write (0x20000000, $crb_msg_ctl);
    if ($SSCOPE_ERR == 1)  {
      $xfr_f_cnt++;
      $target_list[$t_cnt]->{XFR_ERR_CNT}++;
      $SSCOPE_ERR = 0;
    }
    else {
      $xfr_complete = 1;
    }
  }

}


sub enable_all_flush
{

  $t_cnt = 0;

  print "\nEnabling system flush on all nodes in the configuration.\n";

  while (defined $target_list[$t_cnt]) {
    $SSCOPE_ERR = 0;
    mmr_target($SELECT, $target_list[$t_cnt]->{MODULE},
                      $target_list[$t_cnt]->{NODE},
                      $target_list[$t_cnt]->{CHIP});

    if ($SSCOPE_ERR == 1)  {
      printf "\n mmr_target\(\) failed in enable_all_flush\(\).  Aborting mmr read for %s of %s\n\n",
        $target_list[$t_cnt]->{NODE}, $target_list[$t_cnt]->{MODULE};
      $SSCOPE_ERR = 0;
      $t_cnt++;
      next;
    }

    if ($t_cnt == 0)  {
      my $xfr_complete = 0;
      my $xfr_f_cnt = 0;
      while (($xfr_complete == 0) && ($xfr_f_cnt < 3) && ($target_list[$t_cnt]->{XFR_ERR_CNT} < 6))  {
        $crb_msg_ctl = mmr_read(0x20000000);
        if ($SSCOPE_ERR == 1)  {
          $target_list[$t_cnt]->{XFR_ERR_CNT}++;
          $xfr_f_cnt++;
          $SSCOPE_ERR = 0;
        }
        else {
          $xfr_complete = 1;
        }
      }
    }
    enable_single_flush();
    $t_cnt++;
  }
}


sub disable_all_flush
{

  $t_cnt = 0;

  print "\nDisabling system flush on all nodes in the configuration.\n";

  while (defined $target_list[$t_cnt]) {
    $SSCOPE_ERR = 0;
    mmr_target($SELECT, $target_list[$t_cnt]->{MODULE},
                      $target_list[$t_cnt]->{NODE},
                      $target_list[$t_cnt]->{CHIP});

    if ($SSCOPE_ERR == 1)  {
      printf "\n mmr_target\(\) failed in disable_all_flush\(\).  Aborting mmr read for %s of %s\n\n",
        $target_list[$t_cnt]->{NODE}, $target_list[$t_cnt]->{MODULE};
      $SSCOPE_ERR = 0;
      $t_cnt++;
      next;
    }

    disable_single_flush();
    $t_cnt++;
  }
}


#-------------------------------------------------------------------------------
#
#  subroutine to read credits received by the PI from the XN & MD
#
#  The following are "normal" credit values:
#        pi_crbp/pipe_shell/axb_shell/axb_out/arb
#             md_req_crd_cnt_reg: 0x8
#             md_rpy_crd_cnt_reg: 0x4
#             xn_req_crd_cnt_reg: 0x6
#             xn_rpy_crd_cnt_reg: 0x6
#
#-------------------------------------------------------------------------------

sub arm_pi_credits
{

  my $arm_only = 1;      #  set to a 1 to only write the config MMRs and bypass status MMR reads
  $chip_type = "$_[0]";

  @pi_mmr_names = ($chip_type =~ /shub_01/) ? 
		(  "SH_PI_MD2PI_REQUEST_VC_CONFIG",
                   "SH_PI_MD2PI_REPLY_VC_CONFIG",
                   "SH_PI_XN2PI_REQUEST_VC_CONFIG",
                   "SH_PI_XN2PI_REPLY_VC_CONFIG"  )
		:
		(  "SH2_PI_MD2PI_REQUEST_VC_CONFIG",
                   "SH2_PI_MD2PI_REPLY_VC_CONFIG",
                   "SH2_PI_XN2PI_REQUEST_VC_CONFIG",
                   "SH2_PI_XN2PI_REPLY_VC_CONFIG"  )  ;

  my $md_rq_max = ($chip_type =~ /shub_01/) ? 8 : 0x10;
  my $md_rp_max = ($chip_type =~ /shub_01/) ? 8 : 4;

  @pi_reg_list = ();

#--------------------------------------------
#  Create the register structures for each write MMR
#--------------------------------------------
  for ($i = 0; $i <= $#pi_mmr_names; $i++) {
    $name    = $pi_mmr_names[$i];
    $reg_ptr = reg_create_mmr_inst($chip_type, $name);
    if ( !reg_is_valid($reg_ptr) )
    {  $emess = "\nWarning!  Unable to create register instance: $name.\n";
       print "$emess";
    }
    $pi_reg_list[$i] = $reg_ptr;
  }

  my $xfr_complete = 0;
  my $xfr_f_cnt = 0;
  my $init_chip_err_cnt = ($chip_type =~ /shub_01/) ? $target_list[$t_cnt]->{XFR_ERR_CNT} 
						    : $sh2_target_list[$sh2_cnt]->{XFR_ERR_CNT};
  my $chip_err_cnt = $init_chip_err_cnt;
 
  while (($xfr_complete == 0) && ($xfr_f_cnt < 3) && ($chip_err_cnt < 6))  {
    $SSCOPE_ERR = 0;
    reg_list_mmr_read(\@pi_reg_list);        #  Read the MMRs (use L1 list operation)
    if ($SSCOPE_ERR == 1)  {
      $xfr_f_cnt++;
      $chip_err_cnt++;
      $SSCOPE_ERR = 0;
    }
    else {
      $xfr_complete = 1;
    }
  }

  foreach $pi_reg (@pi_reg_list) {
    reg_set_field_value($pi_reg, 'CAPTURE_CREDIT_STATUS', 1);
  }

  $xfr_complete = 0;
  $xfr_f_cnt = 0;
 
  while (($xfr_complete == 0) && ($xfr_f_cnt < 3) && ($chip_err_cnt < 6))  {
    $SSCOPE_ERR = 0;
    reg_list_mmr_write(\@pi_reg_list);        #  Write the MMRs (use L1 list operation)
    if ($SSCOPE_ERR == 1)  {
      $xfr_f_cnt++;
      $chip_err_cnt++;
      $SSCOPE_ERR = 0;
    }
    else {
      $xfr_complete = 1;
    }
  }


#--------------------------------------------
#  Create the register structures for each read MMR
#--------------------------------------------

  if ($arm_only == 0 )  {
    @pi_read_mmr_names = ($chip_type =~ /shub_01/) ?
			( "SH_PI_PI2MD_REQUEST_VC_STATUS",
                          "SH_PI_PI2MD_REPLY_VC_STATUS",
                          "SH_PI_PI2XN_REQUEST_VC_STATUS",
                          "SH_PI_PI2XN_REPLY_VC_STATUS"  )
			:
			( "SH2_PI_PI2MD_REQUEST_VC_STATUS",
                          "SH2_PI_PI2MD_REPLY_VC_STATUS",
                          "SH2_PI_PI2XN_REQUEST_VC_STATUS",
                          "SH2_PI_PI2XN_REPLY_VC_STATUS"  );

    @pi_read_reg_list = ();

    for ($i = 0; $i <= $#pi_read_mmr_names; $i++) {
      $name    = $pi_read_mmr_names[$i];
      $reg_ptr = reg_create_mmr_inst($chip_type, $name);
      if ( !reg_is_valid($reg_ptr) )
      {  $emess = "\nWarning!  Unable to create register instance: $name.\n";
         print "$emess";
      }
      $pi_read_reg_list[$i] = $reg_ptr;
    }

    my $xfr_complete = 0;
    my $xfr_f_cnt = 0;
    while (($xfr_complete == 0) && ($xfr_f_cnt < 3) && ($chip_err_cnt < 6))  {
      $SSCOPE_ERR = 0;
      reg_list_mmr_read(\@pi_read_reg_list);         #  Read the MMRs (use L1 list operation)
      if ($SSCOPE_ERR == 1)  {
        $xfr_f_cnt++;
        $chip_err_cnt++;
        $SSCOPE_ERR = 0;
      }
      else {
        $xfr_complete = 1;
      }
    }

    print "\n";

  #--------------------------------------------
  #  Display the contents of each register that was read
  #--------------------------------------------
    for ($i = 0; $i <= $#pi_read_mmr_names; $i++)
    {
      my $read_data = reg_get_value($pi_read_reg_list[$i]);
      if ($read_data == 0)  {
        $notify = "  <<<<<<<";
        $any_error = 1;
    	if ($chip_type =~ /shub_01/)  {
          $target_list[$t_cnt]->{ERROR} = 1;
          $target_list[$t_cnt]->{ERR_SUMM} .= describe_mmr($pi_read_mmr_names[$i], $read_data, $prefix);
	}
	else {
          $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
          $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= describe_mmr($pi_read_mmr_names[$i], $read_data, $prefix);
	}
      }
      else  {
        $notify = "";
      }

      printf "\n%s: %-36s  Addr: 0x%08lX,  Data = 0x%016llX  %s", $prefix, $pi_read_mmr_names[$i], reg_addr($pi_read_reg_list[$i]), $read_data, $notify;
    }
    print "\n";
  }     		# end of arm_only if statement


  if ($init_chip_err_cnt != $chip_err_cnt)  {       # update counters if errors occurred
    if ($chip_type =~ /shub_01/)  {
      $target_list[$t_cnt]->{XFR_ERR_CNT} = $chip_err_cnt;
    }
    else {
      $sh2_target_list[$sh2_cnt]->{XFR_ERR_CNT} = $chip_err_cnt;
    }
  }

}

#------------------------------------------------------
#
#  This function converts the numeric RRB state value 
#  to a more descriptive definition value.
#
#------------------------------------------------------
sub decode_rrb_state {

  my $state_val = $_[0];
  my $state_def = "????";

  if ($state_val == 0)     { $state_def = "FREE"; }
  elsif ($state_val == 4)  { $state_def = "WAIT_FOR_WRB"; }
  elsif ($state_val == 5)  { $state_def = "WRESP"; }
  elsif ($state_val == 7)  { $state_def = "SNOOP"; }

  return($state_def);
}

#------------------------------------------------------
#
#  This function converts the numeric WRB state value 
#  to a more descriptive definition value.
#
#------------------------------------------------------
sub decode_wrb_state {

  my $state_val = $_[0];
  my $state_def = "????";

  if ($state_val == 0) 	   { $state_def = "FREE"; }
  elsif ($state_val == 8)  { $state_def = "WACK"; }
  elsif ($state_val == 12) { $state_def = "SNOOP"; }
  elsif ($state_val == 13) { $state_def = "SNPP"; }
  elsif ($state_val == 14) { $state_def = "SNPRY"; }
  elsif ($state_val == 15) { $state_def = "SNPRR"; }

  return($state_def);
}

#------------------------------------------------------
#
#  This function converts the numeric IRB state value 
#  to a more descriptive definition value.
#
#------------------------------------------------------
sub decode_irb_state {

  my $state_val = $_[0];
  my $state_def = "????";

  if ($state_val == 0)     { $state_def = "FREE"; }
  elsif ($state_val == 1)  { $state_def = "WAIT_FOR_WRB"; }

  return($state_def);
}

#------------------------------------------------------
#
#  This function converts the numeric PRB state value 
#  to a more descriptive definition value.
#
#------------------------------------------------------
sub decode_prb_state {

  my $state_val = $_[0];
  my $state_def = "????";

  if ($state_val == 0)     { $state_def = "FREE"; }
  elsif ($state_val == 1)  { $state_def = "SNOOP"; }
  elsif ($state_val == 2)  { $state_def = "WRESP"; }
  elsif ($state_val == 3)  { $state_def = "MISC"; }

  return($state_def);
}


#------------------------------------------------------
#
#  This function converts the numeric IRB state value
#  to a more descriptive definition value.
#
#------------------------------------------------------
sub decode_irb_type {

  my $type_val = $_[0];
  my $type_def = "????";

  if ($type_val == 0)     { $type_def = "NONE"; }
  elsif ($type_val == 1)  { $type_def = "INTER"; }
  elsif ($type_val == 2)  { $type_def = "INEXC"; }
  elsif ($type_val == 3)  { $type_def = "ININV"; }
  elsif ($type_val == 4)  { $type_def = "ININF"; }
  elsif ($type_val == 5)  { $type_def = "FLSH"; }
  elsif ($type_val == 6)  { $type_def = "INVAL"; }
  elsif ($type_val == 7)  { $type_def = "ERASE"; }
  elsif ($type_val == 8)  { $type_def = "INIE"; }
  elsif ($type_val == 9)  { $type_def = "INIEF"; }
  elsif ($type_val == 10) { $type_def = "INNE"; }
  elsif ($type_val == 11) { $type_def = "INNEF"; }
  elsif ($type_val == 12) { $type_def = "IEIE"; }
  elsif ($type_val == 13) { $type_def = "IEIEF"; }
  elsif ($type_val == 14) { $type_def = "IENE"; }
  elsif ($type_val == 15) { $type_def = "IENEF"; }
  elsif ($type_val == 16) { $type_def = "IFIE"; }
  elsif ($type_val == 17) { $type_def = "IFIEF"; }
  elsif ($type_val == 18) { $type_def = "IFNE"; }
  elsif ($type_val == 19) { $type_def = "IFNEF"; }
  elsif ($type_val == 20) { $type_def = "FLIE"; }
  elsif ($type_val == 21) { $type_def = "FLIEF"; }
  elsif ($type_val == 22) { $type_def = "FLNE"; }
  elsif ($type_val == 23) { $type_def = "FLNEF"; }
  elsif ($type_val == 24) { $type_def = "ERIE"; }
  elsif ($type_val == 25) { $type_def = "INTR"; }
  elsif ($type_val == 26) { $type_def = "PTC"; }

  return($type_def);
}


#-------------------------------------------------------------------------------
#
#  save_pio_address subroutine
#
#  Arguments passed:    $_[0]   prefix string containing NASID
#                       $_[1]   address to be checked & added
#                       $_[2]   nasid home of directory address to be checked & added
#                       $_[3]   prb number with the pending PIO request
#
#-------------------------------------------------------------------------------
sub save_pio_address {

  my $new_addr          = 1;
  my $aa                = 0;
  my $nasid_prefix      = $_[0];
  my $addr_val          = $_[1];
  my $nasid_val         = $_[2];
  my $prb_num           = $_[3];

  # for($aa=0; $aa<$pio_addr_cnt; $aa++) {
      # if( ($addr_val==$pio_address_list[$aa]->{ADDRESS}) && ($nasid_val==$pio_address_list[$aa]->{NASID})) {
          # $new_addr = 0;
          # printf "\n   %s: NASID 0x%04X PIO address 0x%09llX is already in the list...\n", $nasid_prefix, $nasid_val, $addr_val;
          # $pio_address_list[$aa]->{COUNT} += 1;
          # last;
      # }
  # }


  if($new_addr) {
      my $proc_string = "";
      if ($shub_version == 2)  {
	$proc_string = ($prb_num < 16) ? "Proc Core 0, Thread 0" 
	  : (($prb_num > 15) && ($prb_num < 32)) ? "Proc Core 1, Thread 0"
	  : (($prb_num > 31) && ($prb_num < 48)) ? "Proc Core 0, Thread 1" : "Proc Core 1, Thread 1";
      }
      else {
	$proc_string = ($prb_num > 15) ? "Processor 1" : "Processor 0";
      }
      $pio_address_list[$pio_addr_cnt]->{NASID}   = $nasid_val;
      $pio_address_list[$pio_addr_cnt]->{ADDRESS} = $addr_val;
      $pio_address_list[$pio_addr_cnt]->{COUNT} = 1;
      $pio_address_list[$pio_addr_cnt]->{REPORTING_NASID} = $nasid_prefix;
      $pio_address_list[$pio_addr_cnt]->{REPORTING_LOC} = $current_location;
      $pio_address_list[$pio_addr_cnt]->{REPORTING_PROC} = $proc_string;
      printf "\n%s: added NASID 0x%04X  PIO Address: 0x%09llX to the list (PRB).\n", $nasid_prefix, $nasid_val, $addr_val;
      $pio_addr_cnt++;
  }
}

sub capture_aihc_err {

  my $c_id = "$_[0]";
  my $bad_bit = $_[1];

  my $ev_od = "Even";			# set to default of even
  my $bool_ev_od = 1;			# set to default of even
  my $cl_ev_od = "-even_cache_lines";	# set to default of even

  if ((($pi_dtl1_addr >> 7) & 1) == 1)  {
    $ev_od = "Odd";
    $bool_ev_od = 0;
    $cl_ev_od = "-odd_cache_lines";
  }

  my $rtd_msg = sprintf "\n  smdk command:  run rtrdmp -t R -n N0x%X,N0x%X      \t\t#   %s cacheline failed",
      $nasid_value, $pi_dtl2_src_node, $ev_od;

  if ( $l2_host eq "" ) {
    $l2_host = "[L2 host ip address]";
  }
  $rtd_msg .= sprintf "\n  or\n    L3 command:  rtrdmp --l2 %s -src_nasid 0x%X -dst_nasid 0x%X %s  ",
     $l2_host, $nasid_value, $pi_dtl2_src_node, $cl_ev_od;
  $rtd_msg .= sprintf "\n  and\n    L3 command:  rtrdmp --l2 %s -src_nasid 0x%X -dst_nasid 0x%X %s  ",
     $l2_host, $pi_dtl2_src_node, $nasid_value, $cl_ev_od;

  $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "\n\n Possible Numa-Link address in header error corruption detected by the PI on this node.";
  $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf "\n In %s, it appears address bit %d was corrupted in the header of a message.\n", 
	$c_id, $bad_bit;
  $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "\n Run the following command lines to determine the routes used.";
  $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "    $rtd_msg\n\n";


  $fru_list[$pi_xb_proto_fru_entry]->{REASON} .= "\n\t\t\t\t\t\tSuspect Numa-Link address in header error corruption.";
  $fru_list[$pi_xb_proto_fru_entry]->{REASON} .= "\n\t\t\t\t\t\tRun specified rtrdmp commands to trace failing route.";

  $rtd_list[$rtd_cnt]->{RTD_MSG}   = $rtd_msg;
  $rtd_list[$rtd_cnt]->{SRC_NASID} = $nasid_value;
  $rtd_list[$rtd_cnt]->{DST_NASID} = $pi_dtl2_src_node;
  $rtd_list[$rtd_cnt]->{CHIPLET}   = "pi";
  $rtd_list[$rtd_cnt]->{EVEN_ODD}  = $bool_ev_od;
  $rtd_list[$rtd_cnt]->{REQUEST}   = 1;                   #  failing packet could have been a request
  $rtd_list[$rtd_cnt]->{SKIP_MSG}  = 1;
  $rtd_cnt++;

  $rtd_list[$rtd_cnt]->{RTD_MSG}   = $rtd_msg;
  $rtd_list[$rtd_cnt]->{SRC_NASID} = $pi_dtl2_src_node;
  $rtd_list[$rtd_cnt]->{DST_NASID} = $nasid_value;
  $rtd_list[$rtd_cnt]->{CHIPLET}   = $pi_dtl2_chip;
  $rtd_list[$rtd_cnt]->{EVEN_ODD}  = $bool_ev_od;
  $rtd_list[$rtd_cnt]->{REQUEST}   = 0;                   #  failing packet could have been a response
  $rtd_cnt++;

}

###########################################################################

1;

