# 
#/**************************************************************************
# *                                                                        *
# *               Copyright (C) 1989-1999 Silicon Graphics, Inc.           *
# *                                                                        *
# *  These coded instructions, statements, and computer programs  contain  *
# *  unpublished  proprietary  information of Silicon Graphics, Inc., and  *
# *  are protected by Federal copyright law.  They  may  not be disclosed  *
# *  to  third  parties  or copied or duplicated in any form, in whole or  *
# *  in part, without the prior written consent of Silicon Graphics, Inc.  *
# *                                                                        *
# **************************************************************************/
# 
# @(#)$Id: pktgen.pm,v 1.12 2006/07/05 14:35:34 meyer Exp $
#-------------------------------------------------------------------------------
#
#  FILE: pktgen.pm
#
#  DESCRIPTION:  This file contains PERL subroutines to function the LB
#     packet generator to read/write memory and MMRs.
#
#  MODIFICATION HISTORY:
#
#     WHEN      WHO             WHAT
#     --------  --------------  -----------------------------------------
#     12/13/01  Karen Beighley  Original source code.
#     12/17/01  Karen Beighley  Minor changes; added dependency test.
#     12/18/01  Karen Beighley  Fixed typo with dump title.
#     10/14/02  Karen Beighley  Added use of block MMR read/write operations.
#                               Changed error handling to use common function.
#     10/16/02  Karen Beighley  Added MMR and memory read/write functions.
#                               Minor changes to debug and verbose messages
#     11/20/02  Karen Beighley  Major changes to read/write functions.
#                               Initial source submission for release.
#     11/22/02  Karen Beighley  Added additional/more detailed error messages.
#     04/28/03  Karen Beighley  Added initial support for TIO and ORBIT chips.
#     12/11/03  Karen Beighley  Added support for software emulation mode in
#                                  pktgen_mmr_read() and pktgen_mmr_write().
#     12/18/03  Karen Beighley  Added optional SUPPL parameter to subroutines:
#                                  pktgen_mmr_read() and pktgen_mmr_write().
#     01/09/04  Karen Beighley  Modified to disable LB interrupt before using
#                                  packet generator on TIO.
#     03/02/04  Karen Beighley  Fixed MEM read bug; changes to error checking.
#     07/19/04  Karen Beighley  Added initial support for SHUB2 chip.
#     08/12/04  Karen Beighley  Read and restore TIO_LB_HW_TIME_STAMP MMR.
#     04/27/06  Karen Beighley  Clear LB_HW_INT in TIO_EVENT_OCCURRED_ALIAS MMR.
#     04/28/06  Gary Meyer      Now clearing bit 3 in TIO_EVENT_OCCURRED_ALIAS MMR.
#     04/28/06  Gary Meyer      Now clearing bit 3 in TIO_EVENT_OCCURRED_ALIAS MMR.
#     05/25/06  Gary Meyer      Updated @chn32hdr_fields for shub2 expanded address and node_id fields.
#     07/05/06  Gary Meyer      Expanded address compare masks to 51 bits for larger machine sizes.
#
#-------------------------------------------------------------------------------

use error_regs;
use shub_subs;
use tio_subs;
use shub2_subs;


#-------------------------------------------------------------------------------
#  Structures and variables to build a channel_header data structure
#-------------------------------------------------------------------------------

@chn32hdr_labels = 
( "SRC_ID",         #  Length of SRC_ID      = 11 bits
  "SRC_CHIPLET",    #  Length of SRC_CHIPLET = 3  bits
  "DST_ID",         #  Length of DST_ID      = 11 bits
  "DST_CHIPLET",    #  Length of DST_CHIPLET = 3  bits
  "CMD",            #  Length of CMD         = 8  bits
  "ADDR",           #  Length of ADDR        = 50 bits
  "ECHO",           #  Length of ECHO        = 9  bits
  "SUPPL",          #  Length of SUPPL       = 14 bits
  "DIR",            #  Length of DIR         = 4  bits
  "AGE",            #  Length of AGE         = 8  bits
  "VC"              #  Length of VC          = 1  bit
);

$CHN32HDR_FIELD_LABEL  = 0;
$CHN32HDR_FIELD_MASK   = 1;
$CHN32HDR_FIELD_INDEX  = 2;
$CHN32HDR_FIELD_ADJUST = 3;
$CHN32HDR_FIELD_START  = 4;

#  original for Shub1x
#  @chn32hdr_fields = #             HDR    ADJUST  START
#      LABEL          DATA MASK     INDEX  POS     POS      
#  ( [ "VC"         , 0x00000001,   0,      0,      0 ],  # Length = 1 bits
  #  [ "AGE"        , 0x000000FF,   0,      0,      1 ],  # Length = 8 bits
  #  [ "DIR"        , 0x0000000F,   0,      0,      9 ],  # Length = 4 bits
  #  [ "DST_ID"     , 0x000007FF,   0,      0,     13 ],  # Length = 11 bits
  #  [ "CMD"        , 0x000000FF,   0,      0,     24 ],  # Length = 8 bits
  #  [ "DST_CHIPLET", 0x00000007,   0,      0,     32 ],  # Length = 3 bits
  #  [ "ADDR"       , 0x07FFFFFF,   0,      6,     35 ],  # Length = 27+3+17 bits
  #  [ "SRC_ID"     , 0x000007FF,   1,      0,     16 ],  # Length = 11 bits
  #  [ "SRC_CHIPLET", 0x00000007,   1,      0,     27 ],  # Length = 3 bits
  #  [ "ECHO"       , 0x000001FF,   1,      0,     55 ],  # Length = 9 bits
  #  [ "SUPPL"      , 0x00003FFF,   1,      0,      0 ],  # Length = 14 bits
  #  [ "ADDR"       , 0x00000007,   1,      3,     35 ],  # Length = 27+3+17 bits
  #  [ "ADDR"       , 0x0001FFFF,   1,     33,     38 ]   # Length = 27+3+17 bits
#  );

#  updated for Shub2
@chn32hdr_fields = #             HDR    ADJUST  START
#   LABEL          DATA MASK     INDEX  POS     POS
( [ "VC"         , 0x00000001,   0,      0,      0 ],  # Length = 1 bits
  [ "AGE"        , 0x000000FF,   0,      0,      1 ],  # Length = 8 bits
  [ "DIR"        , 0x0000000F,   0,      0,      9 ],  # Length = 4 bits
  [ "DST_ID"     , 0x000007FF,   0,      0,     13 ],  # Length = 11+1+2 bits	:  dest node bits [10:0]
  [ "DST_ID"     , 0x00000001,   0,     11,     34 ],  # Length = 11+1+2 bits	:  dest node bits [11]
  [ "DST_ID"     , 0x00000003,   0,     12,     62 ],  # Length = 11+1+2 bits	:  dest node bits [13:12]
  [ "CMD"        , 0x000000FF,   0,      0,     24 ],  # Length = 8 bits
  [ "DST_CHIPLET", 0x00000003,   0,      0,     32 ],  # Length = 2 bits
  [ "ADDR"       , 0x07FFFFFF,   0,      6,     35 ],  # Length = 27+3+17+2 bits  :  addr bits [32:6]
  [ "SRC_ID"     , 0x000007FF,   1,      0,     16 ],  # Length = 11+3 bits  :  source node bits [10:0]
  [ "SRC_ID"     , 0x00000007,   1,     11,     29 ],  # Length = 11+3 bits  :  source node bits [13:11]
  [ "SRC_CHIPLET", 0x00000003,   1,      0,     27 ],  # Length = 2 bits
  [ "ECHO"       , 0x000001FF,   1,      0,     55 ],  # Length = 9 bits
  [ "SUPPL"      , 0x0000FFFF,   1,      0,      0 ],  # Length = 16 bits
  [ "ADDR"       , 0x00000007,   1,      3,     35 ],  # Length = 27+3+17+2 bits  :  addr bits [5:3]
  [ "ADDR"       , 0x0001FFFF,   1,     33,     38 ],  # Length = 27+3+17+2 bits  :  addr bits [49:33]
  [ "ADDR"       , 0x00000003,   1,     50,     32 ]   # Length = 27+3+17+2 bits  :  addr bits [51:50]
);


$DIAG_MSG_CNTRL_IEE   = 0x0000000000001000;
$DIAG_MSG_CNTRL_PORT  = 0x0000000000002000;
$DIAG_MSG_CNTRL_BUSY  = 0x8000000000000000;
$DIAG_MSG_CNTRL_START = 0x4000000000000000;

#define SN2NET_PUT              0x41  // Write invalidate
#define SN2NET_NCWRD            0x2c  // Non-coherent write, dword
#define SN2NET_NCWRF            0x4c  // Non-coherent write, cache line


#-------------------------------------------------------------------------------
#  Flags to control the operation of the packet generator
#-------------------------------------------------------------------------------

$PKTGEN_CNTRL_SET_START = 0x01;
$PKTGEN_CNTRL_CLR_START = 0x00;

$PKTGEN_CNTRL_RPLY_CHN  = 0x04;
$PKTGEN_CNTRL_RQST_CHN  = 0x00;

$PKTGEN_WAIT_BUSY       = 0x08;
$PKTGEN_NOWAIT_BUSY     = 0x00;

$PKTGEN_MAX_BUSY_POLL   = 10000;

$PKTGEN_MAX_DBL_WORDS  = 9;
$PKTGEN_MAX_DATA_WORDS = $PKTGEN_MAX_DBL_WORDS * 2;


#-------------------------------------------------------------------------------
#  CONSTANTS - The MMRs addresses must be stored according to these constants
#-------------------------------------------------------------------------------

$PKTGEN_IND_ID     = 0;
$PKTGEN_IND_CNTRL  = 1;
$PKTGEN_IND_HDR0   = 2;
$PKTGEN_IND_HDR1   = 3;
$PKTGEN_IND_DATA0L = 4;
$PKTGEN_IND_DATA0U = 5;


#-------------------------------------------------------------------------------
#  GLOBAL VARIABLES
#-------------------------------------------------------------------------------

$pktgen_verbose = $FALSE;

$pktgen_retry_timeout = 30;

$ptkg_tgt_type = "shub";
$pktg_addr_ref = \@shub_pktgen_addr;


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: pktgen_setup()
#
#    USAGE: @ids = pktgen_setup($op, $addr)
#           @ids = pktgen_setup($op, $addr, $src_id)
#           @ids = pktgen_setup($op, $addr, $src_id, $dst_id)
#           @ids = pktgen_setup($op, $addr, $src_id, $dst_id, $dst_chiplet)
#
#    DESCRIPTION: The subroutine will verify that a chip is targeted, and 
#       initialize the error register and packet generator address pointers
#       for the target chip type.  It will call the chip-specific function
#       to determine the source and destination IDs for the specified
#       packet generated operation ('mem' or 'mmr').
#
#    RETURN VALUE: an array with four entries: source node ID, source chiplet
#       ID, destination node ID and destination chiplet ID.
#
#-------------------------------------------------------------------------------

sub pktgen_setup
{
   print "\npktgen_setup(@_)\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   #---------------------------------------
   #  Setup the error register pointer
   #---------------------------------------
   eregs_setup();
   return if ( $ecode < 0 );

   #---------------------------------------
   #  Check number of MMR targets
   #---------------------------------------
   my $ntarget = mmr_target_get_num();
   if ( $ntarget != 1 )
   {  $emess = "A single chip must targeted before using the packet generator.";
      $emess .= "\nCurrent number of targets = $ntarget";
      $emess .= "\nUse the \"mmr_target()\" subroutine to select a specific chip.";
      $ecode = -1;
      &$comm_err_fun($emess);
      return;
   }
   my $atype = mmr_target_list_get_detail(0, $DETAIL_IND_TYPE);

   #---------------------------------------
   #  Set the array pointers 
   #---------------------------------------
   if ( $atype =~ /^shub2/i )
   {  $ptkg_tgt_type = "shub2";
      $pktg_addr_ref = \@shub2_pktgen_addr;
      $inte_addr_ref = \@shub2_int_enable_addr;
      return( shub2_pktgen_get_ids(@_) );
   }
   elsif ( $atype =~ /^shub/i )
   {  $ptkg_tgt_type = "shub";
      $pktg_addr_ref = \@shub_pktgen_addr;
      $inte_addr_ref = \@shub_int_enable_addr;
      return( shub_pktgen_get_ids(@_) );
   }
   elsif ( $atype =~ /^tio/i )
   {  $ptkg_tgt_type = "tio";
      $pktg_addr_ref = \@tio_pktgen_addr;
      $inte_addr_ref = \@tio_int_enable_addr;
      return( tio_pktgen_get_ids(@_) );
   }
   else
   {  $emess = "Targeted chip does not support packet generator operation.";
      $emess .= "\nTargeted chip is of type: $atype.";
      $ecode = -1;
      &$comm_err_fun($emess);
      return;
   }
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: pktgen_mem_read()
#
#    USAGE: $val = pktgen_mem_read($addr)
#                  pktgen_mem_read($addr, $src_id)
#                  pktgen_mem_read($addr, $src_id, $dst_id)
#                  pktgen_mem_read($addr, $src_id, $dst_id, $dst_chiplet)
#
#    DESCRIPTION: The subroutine reads the single memory location specified
#       by the "addr" parameter using the packet generator.  If no other
#       parameters are specified the NODE_ID field of the NODE_ID MMR
#       is read, and the NODE_ID is used for both the source and destination.
#       If only the source is provided, then that value is also used for the
#       destination.
#
#-------------------------------------------------------------------------------

sub pktgen_mem_read
{
   my $num = $#_ + 1;
   print "\npktgen_mem_read(@_) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   #--------------------------------------- 
   #  Verify the parameters
   #--------------------------------------- 
   if ( ($num < 1) || ($num > 4) )
   {  $emess = "Invalid parameters passed to subroutine pktgen_mem_read().";
      $ecode = -1;
      &$comm_err_fun($emess);
      return;
   }
   my $mem_addr = $_[0];  # Address of memory location to read
   my ($mem_data, $src_id, $src_chiplet, $dst_id, $dst_chiplet);

   #--------------------------------------- 
   #  Set the pointers and determine IDs
   #--------------------------------------- 
   ($src_id, $src_chiplet, $dst_id, $dst_chiplet) = pktgen_setup('mem', @_);

   if ( $pktgen_verbose || ($debug_flag & $DEBUG_GENERAL) )
   {  printf "\n";
      printf "---------------------------------------------------------------------\n";
      printf "-- PKTGEN MEM READ : ADDR=0x%016llX\n", $mem_addr;
      printf "--             SRC : NODE_ID=0x%05X CHIPLET_ID=0x%02X\n", $src_id, $src_chiplet;
      printf "--             DST : NODE_ID=0x%05X CHIPLET_ID=0x%02X\n", $dst_id, $dst_chiplet;
      printf "---------------------------------------------------------------------\n";
   }
   my ($final_addr, %rq_hdr, $mode, @edata, $ereg_err_detail_1, $ereg_err_detail_2, 
       $emask_addr, $emask_data, $einit_data);
   my $ts_data = 0;

   #--------------------------------------- 
   #  Determine the node ID and adjust address
   #--------------------------------------- 
   $final_addr = $mem_addr | ($dst_id << 38);

   #--------------------------------------- 
   #  Create and initialize the packet header
   #--------------------------------------- 
   %rq_hdr = chn32hdr_new();
   $rq_hdr{SRC_ID}         = $src_id;
   $rq_hdr{SRC_CHIPLET}    = $src_chiplet;
   $rq_hdr{DST_ID}         = $dst_id;
   $rq_hdr{DST_CHIPLET}    = $dst_chiplet;
   $rq_hdr{ADDR}           = $final_addr;
   $rq_hdr{ECHO}           = 0x123;
   $rq_hdr{CMD}            = 0x04; # Coherent shapshot, forced
   $rq_hdr{SUPPL}          = 0xff;

   #--------------------------------------- 
   #  Initialize the error registers
   #--------------------------------------- 
   $emask_addr = $$ereg_addr_ref[$EREG_MASK];
   $emask_data = mmr_read($emask_addr);
   $einit_data = ( ($ptkg_tgt_type eq 'shub') ? ($emask_data | $EREG_ERROR_MASK_RP_BAD_CMD) :
                                                ($EREG_ERROR_MASK & ~($EREG_ERROR_MASK_RP_BAD_CMD)) );
   eregs_initialize($einit_data);
   if ( $ecode < 0 )
   {  $emess = "Packet generated memory read operation unsuccessful.";
      $emess .= "\nUnable to initialize the error reporting registers.";
      $ecode = -1;
      &$comm_err_fun($emess);
      return;
   }

   #--------------------------------------- 
   #  Modify the interrupt enable registers 
   #--------------------------------------- 
   my (@inte_data, @inte_copy, $inte_mask, $i);
   if ( $ptkg_tgt_type eq 'tio' )
   {  @inte_data = mmr_read_group($inte_addr_ref);
      @inte_copy = ();
      $inte_mask = 0xFFFFFFFFFFFFFFF7;
      for ($i = 0; $i <= $#inte_data; $i++)
      {  $inte_copy[$i] = $inte_data[$i] & $inte_mask;
      }
      mmr_write_group($inte_addr_ref, \@inte_copy);
      $ts_data = mmr_read($TIO_LB_HW_TIME_STAMP_MMR);
   }

   #--------------------------------------- 
   #  Write MMRs to launch packet (header and data)
   #--------------------------------------- 
   $mode = $PKTGEN_CNTRL_SET_START | $PKTGEN_WAIT_BUSY;
   $PKTGEN_MAX_BUSY_POLL = 20;      # limit the number of loops in pktgen_wait_if_busy()
   pktgen_write($mode, \%rq_hdr);
   if ( $ecode < 0 ) 
   {  mmr_write($emask_addr, $emask_data);  # Restore mask setting
      if ( $ptkg_tgt_type eq 'tio' )        # Restore interrupt settings
      {  mmr_write_group($inte_addr_ref, \@inte_data);
         mmr_write($TIO_EVENT_OCCURRED_ALIAS_MMR, 0x0000000000000008); 
         mmr_write($TIO_LB_HW_TIME_STAMP_MMR, $ts_data);
      }
      return;                               # Return: short-circuit if error detected and returned
   }

   #--------------------------------------- 
   #  Monitor error registers for response packet
   #--------------------------------------- 
   @edata = eregs_wait_summary_read_clear($EREG_ERROR_MASK_RP_BAD_CMD, $pktgen_retry_timeout);
   if ( ($#edata <= 0) || ($ecode < 0) )
   {  mmr_write($emask_addr, $emask_data);  # Restore mask setting
      if ( $ptkg_tgt_type eq 'tio' )        # Restore interrupt settings
      {  mmr_write_group($inte_addr_ref, \@inte_data);
         mmr_write($TIO_EVENT_OCCURRED_ALIAS_MMR, 0x0000000000000008); 
         mmr_write($TIO_LB_HW_TIME_STAMP_MMR, $ts_data);
      }
      $emess = "Packet generated memory read operation unsuccessful.";
      $emess .= "\nUnable to detect error or clear error reporting registers.";
      $ecode = -1;
      &$comm_err_fun($emess);
      return;
   }
   mmr_write($emask_addr, $emask_data);  # Restore mask setting

   #--------------------------------------- 
   #  Restore the interrupt enable registers 
   #--------------------------------------- 
   if ( $ptkg_tgt_type eq 'tio' )
   {  mmr_write_group($inte_addr_ref, \@inte_data);
      mmr_write($TIO_EVENT_OCCURRED_ALIAS_MMR, 0x0000000000000008); 
      mmr_write($TIO_LB_HW_TIME_STAMP_MMR, $ts_data);
   }
 
   #--------------------------------------- 
   #  Examine the error registers 
   #--------------------------------------- 
   $ereg_err_detail_1 = $edata[1];
   $ereg_err_detail_2 = $edata[2];

   if ( $ereg_err_detail_1 & 0x0002000000000000 )
   {  $emess = "Packet generated memory read operation unsuccessful.";
      $emess .= "\nUnexpected response packet captured in error detail registers.";
      $emess .= "\nLB_ERROR_DETAIL_1 has DATA_ERR set (0x0002000000000000).";
      $emess .= "\nUnable to confirm completion of packet generated request.";
      $ecode = -1;
      printf "\nActual data from LB ERROR MMRs:\n";
      decode_lb_status(\@edata);
      &$comm_err_fun($emess);
      return;
   }
   if ( $ereg_err_detail_1 & 0x0001000000000000 )
   {  $emess = "Packet generated memory read operation unsuccessful.";
      $emess .= "\nUnexpected response packet captured in error detail registers.";
      $emess .= "\nLB_ERROR_DETAIL_1 has HDR_ERR set (0x0001000000000000).";
      $emess .= "\nUnable to confirm completion of packet generated request.";
      $ecode = -1;
      printf "\nActual data from LB ERROR MMRs:\n";
      decode_lb_status(\@edata);
      &$comm_err_fun($emess);
      return;
   }

   $tmp = $ereg_err_detail_1 & 0xFF;
   if ( ($tmp != 0xCA) && ($tmp != 0xCF) )
   {  $emess = "Packet generated memory read operation unsuccessful.";
      $emess .= "\nUnexpected response packet captured in error detail registers.";
      $text   = sprintf("LB_ERROR_DETAIL_1 does not have COMMAND of 0xCA or 0xCF (COMMAND=0x%02X).", $tmp);
      $emess .= "\n$text";
      $emess .= "\nUnable to confirm completion of packet generated request.";
      $ecode = -1;
      printf "\nActual data from LB ERROR MMRs:\n";
      decode_lb_status(\@edata);
      &$comm_err_fun($emess);
      return;
   }

   $tmp = ($ereg_err_detail_2 & 0x1FFFFFFFFFFFF) << 3;
   if ( $tmp != $final_addr )
   {  $emess = "Packet generated memory read operation unsuccessful.";
      $emess .= "\nUnexpected response packet captured in error detail registers.";
      $text  = sprintf("LB_ERROR_DETAIL_2 does not have ADDR of 0x%012llX (ADDR=0x%012llX).", $final_addr, $tmp);
      $emess .= "\n$text";
      $emess .= "\nUnable to confirm completion of packet generated request.";
      $ecode = -1;
      printf "\nActual data from LB ERROR MMRs:\n";
      decode_lb_status(\@edata);
      &$comm_err_fun($emess);
      return;
   }

   if ( $pktgen_verbose || ($debug_flag & $DEBUG_GENERAL) )
   {  printf "\nActual data from LB ERROR MMRs:\n";
      decode_lb_status(\@edata);
   }

   #--------------------------------------- 
   #  Data should  maintain quadword ordering
   #--------------------------------------- 
   $mem_data = ( $final_addr & 0x8 ) ? $edata[3] : $edata[4];
   return($mem_data);
}

#-------------------------------------------------------------------------------
#
#    SUBROUTINE: pktgen_mem_write()
#
#    USAGE: pktgen_mem_write($addr, $value)
#           pktgen_mem_write($addr, $value, $src_id)
#           pktgen_mem_write($addr, $value, $src_id, $dst_id)
#           pktgen_mem_write($addr, $value, $src_id, $dst_id, $dst_chiplet)
#
#    DESCRIPTION: The subroutine writes the single memory location specified
#       by the "addr" parameter with "value" using the packet generator.
#       If no other parameters are specified the NODE_ID field of the
#       NODE_ID MMR is read, and the NODE_ID is used for both the source and
#       destination. If only the source is provided, then that value is also
#       used for the destination.
#
#       WARNING - This is a non-coherent write operation!!!
#
#-------------------------------------------------------------------------------

sub pktgen_mem_write
{
   my $num = $#_ + 1;
   print "\npktgen_mem_write(@_) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   #--------------------------------------- 
   #  Verify the parameters
   #--------------------------------------- 
   if ( ($num < 2) || ($num > 5) )
   {  $emess = "Invalid parameters passed to subroutine pktgen_mem_write().";
      $ecode = -1;
      &$comm_err_fun($emess);
      return;
   }
   my $mem_addr = $_[0];  # Address of memory to write
   my $mem_data = $_[1];  # Data to write to memory location
   my ($src_id, $src_chiplet, $dst_id, $dst_chiplet);

   #--------------------------------------- 
   #  Set the pointers and determine IDs
   #--------------------------------------- 
   splice(@_, 1, 1);   # Remove the memory data entry from the array
   ($src_id, $src_chiplet, $dst_id, $dst_chiplet) = pktgen_setup('mem', @_);

   if ( $pktgen_verbose || ($debug_flag & $DEBUG_GENERAL) )
   {  printf "\n";
      printf "---------------------------------------------------------------------\n";
      printf "-- PKTGEN MEM WRITE : ADDR=0x%016llX DATA=0x%016llX\n", $mem_addr, $mem_data;
      printf "--              SRC : NODE_ID=0x%05X CHIPLET_ID=0x%02X\n", $src_id, $src_chiplet;
      printf "--              DST : NODE_ID=0x%05X CHIPLET_ID=0x%02X\n", $dst_id, $dst_chiplet;
      printf "---------------------------------------------------------------------\n";
   }
   my ($final_addr, %rq_hdr, $mode, @edata, $ereg_err_detail_1, $ereg_err_detail_2, 
       $emask_addr, $emask_data, $einit_data);
   my $ts_data = 0;

   #--------------------------------------- 
   #  Determine the node ID and adjust address
   #--------------------------------------- 
   $final_addr = $mem_addr | ($dst_id << 38);

   #--------------------------------------- 
   #  Create and initialize the packet header
   #--------------------------------------- 
   %rq_hdr = chn32hdr_new();
   $rq_hdr{SRC_ID}         = $src_id;
   $rq_hdr{SRC_CHIPLET}    = $src_chiplet;
   $rq_hdr{DST_ID}         = $dst_id;
   $rq_hdr{DST_CHIPLET}    = $dst_chiplet;
   $rq_hdr{ADDR}           = $final_addr;
   $rq_hdr{ECHO}           = 0x123;
   $rq_hdr{CMD}            = 0x2c;
   $rq_hdr{SUPPL}          = 0xFF;

   #--------------------------------------- 
   #  Data should  maintain quadword ordering
   #--------------------------------------- 
   if ( $final_addr & 0x8 )  
   {  @data = (0, $mem_data);
   }  
   else
   {  @data = ($mem_data, 0);
   }
 
   #--------------------------------------- 
   #  Initialize the error registers
   #--------------------------------------- 
   $emask_addr = $$ereg_addr_ref[$EREG_MASK];
   $emask_data = mmr_read($emask_addr);
   $einit_data = ( ($ptkg_tgt_type eq 'shub') ? ($emask_data | $EREG_ERROR_MASK_RP_BAD_CMD) :
                                                ($EREG_ERROR_MASK & ~($EREG_ERROR_MASK_RP_BAD_CMD)) );
   eregs_initialize($einit_data);
   if ( $ecode < 0 )
   {  $emess = "Packet generated memory write operation unsuccessful.";
      $emess .= "\nUnable to initialize the error reporting registers.";
      $ecode = -1;
      &$comm_err_fun($emess);
      return;
   }

   #--------------------------------------- 
   #  Modify the interrupt enable registers 
   #--------------------------------------- 
   my (@inte_data, @inte_copy, $inte_mask, $i);
   if ( $ptkg_tgt_type eq 'tio' )
   {  @inte_data = mmr_read_group($inte_addr_ref);
      @inte_copy = ();
      $inte_mask = 0xFFFFFFFFFFFFFFF7;
      for ($i = 0; $i <= $#inte_data; $i++)
      {  $inte_copy[$i] = $inte_data[$i] & $inte_mask;
      }
      mmr_write_group($inte_addr_ref, \@inte_copy);
      $ts_data = mmr_read($TIO_LB_HW_TIME_STAMP_MMR);
   }

   #--------------------------------------- 
   #  Write MMRs to launch packet (header and data)
   #--------------------------------------- 
   $mode = $PKTGEN_CNTRL_SET_START | $PKTGEN_WAIT_BUSY;
   $PKTGEN_MAX_BUSY_POLL = 20;      # limit the number of loops in pktgen_wait_if_busy()
   pktgen_write($mode, \%rq_hdr, \@data);
   if ( $ecode < 0 ) 
   {  mmr_write($emask_addr, $emask_data);  # Restore mask setting
      if ( $ptkg_tgt_type eq 'tio' )        # Restore interrupt settings
      {  mmr_write_group($inte_addr_ref, \@inte_data);
         mmr_write($TIO_EVENT_OCCURRED_ALIAS_MMR, 0x0000000000000008); 
         mmr_write($TIO_LB_HW_TIME_STAMP_MMR, $ts_data);
      }
      return;                               # Return: short-circuit if error detected and returned
   }
 
   #--------------------------------------- 
   #  Monitor error registers for response packet
   #--------------------------------------- 
   @edata = eregs_wait_summary_read_clear($EREG_ERROR_MASK_RP_BAD_CMD, $pktgen_retry_timeout);
   if ( ($#edata <= 0) || ($ecode < 0) )
   {  mmr_write($emask_addr, $emask_data);  # Restore mask setting
      if ( $ptkg_tgt_type eq 'tio' )        # Restore interrupt settings
      {  mmr_write_group($inte_addr_ref, \@inte_data);
         mmr_write($TIO_EVENT_OCCURRED_ALIAS_MMR, 0x0000000000000008); 
         mmr_write($TIO_LB_HW_TIME_STAMP_MMR, $ts_data);
      }
      $emess = "Packet generated memory write operation unsuccessful.";
      $emess .= "\nUnable to detect error or clear error reporting registers.";
      $ecode = -1;
      &$comm_err_fun($emess);
      return;
   }
   mmr_write($emask_addr, $emask_data);  # Restore mask setting

   #--------------------------------------- 
   #  Restore the interrupt enable registers 
   #--------------------------------------- 
   if ( $ptkg_tgt_type eq 'tio' )
   {  mmr_write_group($inte_addr_ref, \@inte_data);
      mmr_write($TIO_EVENT_OCCURRED_ALIAS_MMR, 0x0000000000000008); 
      mmr_write($TIO_LB_HW_TIME_STAMP_MMR, $ts_data);
   }
 
   #--------------------------------------- 
   #  Examine the error registers 
   #--------------------------------------- 
   $ereg_err_detail_1 = $edata[1];
   $ereg_err_detail_2 = $edata[2];

   if ( $ereg_err_detail_1 & 0x0002000000000000 )
   {  $emess = "Packet generated memory write operation unsuccessful.";
      $emess .= "\nUnexpected response packet captured in error detail registers.";
      $emess .= "\nLB_ERROR_DETAIL_1 has DATA_ERR set (0x0002000000000000).";
      $emess .= "\nUnable to confirm completion of packet generated request.";
      $ecode = -1;
      printf "\nActual data from LB ERROR MMRs:\n";
      decode_lb_status(\@edata);
      &$comm_err_fun($emess);
      return;
   }
   if ( $ereg_err_detail_1 & 0x0001000000000000 )
   {  $emess = "Packet generated memory write operation unsuccessful.";
      $emess .= "\nUnexpected response packet captured in error detail registers.";
      $emess .= "\nLB_ERROR_DETAIL_1 has HDR_ERR set (0x0001000000000000).";
      $emess .= "\nUnable to confirm completion of packet generated request.";
      $ecode = -1;
      printf "\nActual data from LB ERROR MMRs:\n";
      decode_lb_status(\@edata);
      &$comm_err_fun($emess);
      return;
   }

   $tmp = $ereg_err_detail_1 & 0xFF;
   if ( $tmp != 0xE0 )
#   if ( ($tmp != 0x98) && ($tmp != 0xE0) )  # 0x98 is for PUT (write/invalidate cache line)
   {  $emess = "Packet generated memory write operation unsuccessful.";
      $emess .= "\nUnexpected response packet captured in error detail registers.";
      $text   = sprintf("LB_ERROR_DETAIL_1 does not have COMMAND of 0xE0 (COMMAND=0x%02X).", $tmp);
      $emess .= "\n$text";
      $emess .= "\nUnable to confirm completion of packet generated request.";
      $ecode = -1;
      printf "\nActual data from LB ERROR MMRs:\n";
      decode_lb_status(\@edata);
      &$comm_err_fun($emess);
      return;
   }

   $tmp = ($ereg_err_detail_2 & 0x1FFFFFFFFFFFF) << 3;
   if ( $tmp != $final_addr )
   {  $emess = "Packet generated memory write operation unsuccessful.";
      $emess .= "\nUnexpected response packet captured in error detail registers.";
      $text  = sprintf("LB_ERROR_DETAIL_2 does not have ADDR of 0x%012llX (ADDR=0x%012llX).", $final_addr, $tmp);
      $emess .= "\n$text";
      $emess .= "\nUnable to confirm completion of packet generated request.";
      $ecode = -1;
      printf "\nActual data from LB ERROR MMRs:\n";
      decode_lb_status(\@edata);
      &$comm_err_fun($emess);
      return;
   }

   if ( $pktgen_verbose || ($debug_flag & $DEBUG_GENERAL) )
   {  printf "\nActual data from LB ERROR MMRs:\n";
      decode_lb_status(\@edata);
   }
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: pktgen_mem_write_cl()
#
#    USAGE: pktgen_mem_write_cl($addr, $data_ref)
#           pktgen_mem_write_cl($addr, $data_ref, $src_id)
#           pktgen_mem_write_cl($addr, $data_ref, $src_id, $dst_id)
#           pktgen_mem_write_cl($addr, $data_ref, $src_id, $dst_id, $dst_chiplet)
#
#    DESCRIPTION: The subroutine writes the cache line(s) starting at
#       the "addr" parameter with "data_ref" using the packet generator.
#       If no other parameters are specified the NODE_ID field of the
#       NODE_ID MMR is read, and the NODE_ID is used for both the source and
#       destination. If only the source is provided, then that value is also
#       used for the destination.
#
#       WARNING - This is a non-coherent write operation!!!
#
#-------------------------------------------------------------------------------

sub pktgen_mem_write_cl
{
   my $num = $#_ + 1;
   print "\npktgen_mem_write_cl(@_) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   #--------------------------------------- 
   #  Verify the parameters
   #--------------------------------------- 
   if ( ($num < 2) || ($num > 5) )
   {  $emess = "Invalid parameters passed to subroutine pktgen_mem_write_cl().";
      $ecode = -1;
      &$comm_err_fun($emess);
      return;
   }
   my $mem_addr = $_[0];  # Address of memory to write
   my $mem_dref = $_[1];  # Reference to array of 64-bit values

   #--------------------------------------- 
   #  Verify the data is cache-line aligned
   #--------------------------------------- 
   my $nline = $#$mem_dref + 1;
   if ( ($nline % 16) != 0 )
   {  $emess = "Invalid parameters passed to subroutine pktgen_mem_write_cl().";
      $ecode = -1;
      &$comm_err_fun($emess);
      return;
   }
   $nline = $nline / 16;


   #--------------------------------------- 
   #  Set the pointers and determine IDs
   #--------------------------------------- 
   splice(@_, 1, 1);   # Remove the memory data entry from the array
   my ($src_id, $src_chiplet, $dst_id, $dst_chiplet);
   ($src_id, $src_chiplet, $dst_id, $dst_chiplet) = pktgen_setup('mem', @_);

   if ( $pktgen_verbose || ($debug_flag & $DEBUG_GENERAL) )
   {  printf "\n";
      printf "---------------------------------------------------------------------\n";
      printf "-- PKTGEN MEM WRITE : ADDR=0x%016llX DATA=0x%016llX\n", $mem_addr, $$mem_dref[0];
      printf "--              SRC : NODE_ID=0x%05X CHIPLET_ID=0x%02X\n", $src_id, $src_chiplet;
      printf "--              DST : NODE_ID=0x%05X CHIPLET_ID=0x%02X\n", $dst_id, $dst_chiplet;
      printf "---------------------------------------------------------------------\n";
   }
   my ($final_addr, %rq_hdr, $mode, @edata, $ereg_err_detail_1, $ereg_err_detail_2, 
       $emask_addr, $emask_data, $einit_data);
   my $ts_data = 0;

   #--------------------------------------- 
   #  Determine the node ID and adjust address
   #--------------------------------------- 
   $final_addr = $mem_addr | ($dst_id << 38);

   #--------------------------------------- 
   #  Create and initialize the packet header
   #--------------------------------------- 
   %rq_hdr = chn32hdr_new();
   $rq_hdr{SRC_ID}         = $src_id;
   $rq_hdr{SRC_CHIPLET}    = $src_chiplet;
   $rq_hdr{DST_ID}         = $dst_id;
   $rq_hdr{DST_CHIPLET}    = $dst_chiplet;
   $rq_hdr{ADDR}           = $final_addr;
   $rq_hdr{ECHO}           = 0x123;
   $rq_hdr{CMD}            = 0x41; # Non-coherent write, cache line
   $rq_hdr{SUPPL}          = 0xFF;

   #--------------------------------------- 
   #  Data should  maintain quadword ordering
   #--------------------------------------- 
   my $i;
   my @data = ();
   for ($i = 0; $i < 16; $i++)
   {  $data[$i] = $$mem_dref[$i];
   }
 
   #--------------------------------------- 
   #  Initialize the error registers
   #--------------------------------------- 
   $emask_addr = $$ereg_addr_ref[$EREG_MASK];
   $emask_data = mmr_read($emask_addr);
   $einit_data = ( ($ptkg_tgt_type eq 'shub') ? ($emask_data | $EREG_ERROR_MASK_RP_BAD_CMD) :
                                                ($EREG_ERROR_MASK & ~($EREG_ERROR_MASK_RP_BAD_CMD)) );
   eregs_initialize($einit_data);
   if ( $ecode < 0 )
   {  $emess = "Packet generated memory write operation unsuccessful.";
      $emess .= "\nUnable to initialize the error reporting registers.";
      $ecode = -1;
      &$comm_err_fun($emess);
      return;
   }

   #--------------------------------------- 
   #  Modify the interrupt enable registers 
   #--------------------------------------- 
   my (@inte_data, @inte_copy, $inte_mask, $i);
   if ( $ptkg_tgt_type eq 'tio' )
   {  @inte_data = mmr_read_group($inte_addr_ref);
      @inte_copy = ();
      $inte_mask = 0xFFFFFFFFFFFFFFF7;
      for ($i = 0; $i <= $#inte_data; $i++)
      {  $inte_copy[$i] = $inte_data[$i] & $inte_mask;
      }
      mmr_write_group($inte_addr_ref, \@inte_copy);
      $ts_data = mmr_read($TIO_LB_HW_TIME_STAMP_MMR);
   }

   #--------------------------------------- 
   #  Write MMRs to launch packet (header and data)
   #--------------------------------------- 
   $mode = $PKTGEN_CNTRL_SET_START | $PKTGEN_WAIT_BUSY;
   $PKTGEN_MAX_BUSY_POLL = 20;      # limit the number of loops in pktgen_wait_if_busy()
   pktgen_write($mode, \%rq_hdr, \@data);
   if ( $ecode < 0 ) 
   {  mmr_write($emask_addr, $emask_data);  # Restore mask setting
      if ( $ptkg_tgt_type eq 'tio' )        # Restore interrupt settings
      {  mmr_write_group($inte_addr_ref, \@inte_data);
         mmr_write($TIO_EVENT_OCCURRED_ALIAS_MMR, 0x0000000000000008); 
         mmr_write($TIO_LB_HW_TIME_STAMP_MMR, $ts_data);
      }
      return;                               # Return: short-circuit if error detected and returned
   }
 
   #--------------------------------------- 
   #  Monitor error registers for response packet
   #--------------------------------------- 
   @edata = eregs_wait_summary_read_clear($EREG_ERROR_MASK_RP_BAD_CMD, $pktgen_retry_timeout);
   if ( ($#edata <= 0) || ($ecode < 0) )
   {  mmr_write($emask_addr, $emask_data);  # Restore mask setting
      if ( $ptkg_tgt_type eq 'tio' )        # Restore interrupt settings
      {  mmr_write_group($inte_addr_ref, \@inte_data);
         mmr_write($TIO_EVENT_OCCURRED_ALIAS_MMR, 0x0000000000000008); 
         mmr_write($TIO_LB_HW_TIME_STAMP_MMR, $ts_data);
      }
      $emess = "Packet generated memory write operation unsuccessful.";
      $emess .= "\nUnable to detect error or clear error reporting registers.";
      $ecode = -1;
      &$comm_err_fun($emess);
      return;
   }
   mmr_write($emask_addr, $emask_data);  # Restore mask setting

   #--------------------------------------- 
   #  Restore the interrupt enable registers 
   #--------------------------------------- 
   if ( $ptkg_tgt_type eq 'tio' )
   {  mmr_write_group($inte_addr_ref, \@inte_data);
      mmr_write($TIO_EVENT_OCCURRED_ALIAS_MMR, 0x0000000000000008); 
      mmr_write($TIO_LB_HW_TIME_STAMP_MMR, $ts_data);
   }
 
   #--------------------------------------- 
   #  Examine the error registers 
   #--------------------------------------- 
   $ereg_err_detail_1 = $edata[1];
   $ereg_err_detail_2 = $edata[2];

   if ( $ereg_err_detail_1 & 0x0002000000000000 )
   {  $emess = "Packet generated memory write operation unsuccessful.";
      $emess .= "\nUnexpected response packet captured in error detail registers.";
      $emess .= "\nLB_ERROR_DETAIL_1 has DATA_ERR set (0x0002000000000000).";
      $emess .= "\nUnable to confirm completion of packet generated request.";
      $ecode = -1;
      printf "\nActual data from LB ERROR MMRs:\n";
      decode_lb_status(\@edata);
      &$comm_err_fun($emess);
      return;
   }
   if ( $ereg_err_detail_1 & 0x0001000000000000 )
   {  $emess = "Packet generated memory write operation unsuccessful.";
      $emess .= "\nUnexpected response packet captured in error detail registers.";
      $emess .= "\nLB_ERROR_DETAIL_1 has HDR_ERR set (0x0001000000000000).";
      $emess .= "\nUnable to confirm completion of packet generated request.";
      $ecode = -1;
      printf "\nActual data from LB ERROR MMRs:\n";
      decode_lb_status(\@edata);
      &$comm_err_fun($emess);
      return;
   }

   $tmp = $ereg_err_detail_1 & 0xFF;
#   if ( $tmp != 0xE0 )
   if ( ($tmp != 0x98) && ($tmp != 0xE0) )  # 0x98 is for PUT (write/invalidate cache line)
   {  $emess = "Packet generated memory write operation unsuccessful.";
      $emess .= "\nUnexpected response packet captured in error detail registers.";
      $text   = sprintf("LB_ERROR_DETAIL_1 does not have COMMAND of 0xE0 (COMMAND=0x%02X).", $tmp);
      $emess .= "\n$text";
      $emess .= "\nUnable to confirm completion of packet generated request.";
      $ecode = -1;
      printf "\nActual data from LB ERROR MMRs:\n";
      decode_lb_status(\@edata);
      &$comm_err_fun($emess);
      return;
   }

   $tmp = ($ereg_err_detail_2 & 0x1FFFFFFFFFFFF) << 3;
   if ( $tmp != $final_addr )
   {  $emess = "Packet generated memory write operation unsuccessful.";
      $emess .= "\nUnexpected response packet captured in error detail registers.";
      $text  = sprintf("LB_ERROR_DETAIL_2 does not have ADDR of 0x%012llX (ADDR=0x%012llX).", $final_addr, $tmp);
      $emess .= "\n$text";
      $emess .= "\nUnable to confirm completion of packet generated request.";
      $ecode = -1;
      printf "\nActual data from LB ERROR MMRs:\n";
      decode_lb_status(\@edata);
      &$comm_err_fun($emess);
      return;
   }

   if ( $pktgen_verbose || ($debug_flag & $DEBUG_GENERAL) )
   {  printf "\nActual data from LB ERROR MMRs:\n";
      decode_lb_status(\@edata);
   }
}



#-------------------------------------------------------------------------------
#
#    SUBROUTINE: pktgen_mmr_read()
#
#    USAGE: $val = pktgen_mmr_read($addr)
#                  pktgen_mmr_read($addr, $src_id)
#                  pktgen_mmr_read($addr, $src_id, $dst_id)
#                  pktgen_mmr_read($addr, $src_id, $dst_id, $dst_chiplet)
#                  pktgen_mmr_read($addr, $src_id, $dst_id, $dst_chiplet, $suppl)
#
#    DESCRIPTION: The subroutine reads the single MMR location specified
#       by the "addr" parameter using the packet generator.  If no other
#       parameters are specified the NODE_ID field of the NODE_ID MMR
#       is read, and the NODE_ID is used for both the source and destination.
#       If only the source is provided, then that value is also used for the
#       destination.
#
#-------------------------------------------------------------------------------

sub pktgen_mmr_read
{
   my $num = $#_ + 1;
   print "\npktgen_mmr_read(@_) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   #--------------------------------------- 
   #  Verify the parameters
   #--------------------------------------- 
   if ( ($num < 1) || ($num > 5) )
   {  $emess = "Invalid parameters passed to subroutine pktgen_mmr_read().";
      $ecode = -1;
      &$comm_err_fun($emess);
      return;
   }
   my $mmr_addr = $_[0];  # Address of MMR to read
   my $suppl    = ( ($num == 5) ? pop(@_) : 0xFF );
   my ($mmr_data, $src_id, $src_chiplet, $dst_id, $dst_chiplet);

   #--------------------------------------- 
   #  Support software emulation mode
   #--------------------------------------- 
   if ($mmr_access == $MMR_ACCESS_NONE)
   {  return( mmr_read($mmr_addr) );
   }

   #--------------------------------------- 
   #  Set the pointers and determine IDs
   #--------------------------------------- 
   ($src_id, $src_chiplet, $dst_id, $dst_chiplet) = pktgen_setup('mmr', @_);

   if ( $pktgen_verbose || ($debug_flag & $DEBUG_GENERAL) )
   {  printf "\n";
      printf "---------------------------------------------------------------------\n";
      printf "-- PKTGEN MMR READ : ADDR=0x%016llX\n", $mmr_addr;
      printf "--             SRC : NODE_ID=0x%05X CHIPLET_ID=0x%02X\n", $src_id, $src_chiplet;
      printf "--             DST : NODE_ID=0x%05X CHIPLET_ID=0x%02X\n", $dst_id, $dst_chiplet;
      printf "---------------------------------------------------------------------\n";
   }
   my ($final_addr, %rq_hdr, $mode, @edata, $ereg_err_detail_1, $ereg_err_detail_2,
       $emask_addr, $emask_data, $einit_data);
   my $ts_data = 0;

   #--------------------------------------- 
   #  Determine the node ID and adjust address
   #--------------------------------------- 
   $final_addr = $mmr_addr | ($dst_id << 38);

   #--------------------------------------- 
   #  Create and initialize the packet header
   #--------------------------------------- 
   %rq_hdr = chn32hdr_new();
   $rq_hdr{SRC_ID}         = $src_id;
   $rq_hdr{SRC_CHIPLET}    = $src_chiplet;
   $rq_hdr{DST_ID}         = $dst_id;
   $rq_hdr{DST_CHIPLET}    = $dst_chiplet;
   $rq_hdr{ADDR}           = $final_addr;
   $rq_hdr{ECHO}           = 0x123;
   $rq_hdr{CMD}            = 0x11;			# PIO Read
   $rq_hdr{SUPPL}          = $suppl;                    # default = 0xff;

   #--------------------------------------- 
   #  Initialize the error registers
   #--------------------------------------- 
   $emask_addr = $$ereg_addr_ref[$EREG_MASK];
   $emask_data = mmr_read($emask_addr);
   $einit_data = ( ($ptkg_tgt_type eq 'shub') ? ($emask_data | $EREG_ERROR_MASK_RP_BAD_CMD) :
                                                ($EREG_ERROR_MASK & ~($EREG_ERROR_MASK_RP_BAD_CMD)) );
   eregs_initialize($einit_data);
   if ( $ecode < 0 )
   {  $emess = "Packet generated MMR read operation unsuccessful.";
      $emess .= "\nUnable to initialize the error reporting registers.";
      $ecode = -1;
      &$comm_err_fun($emess);
      return;
   }

   #--------------------------------------- 
   #  Modify the interrupt enable registers 
   #--------------------------------------- 
   my (@inte_data, @inte_copy, $inte_mask, $i);
   if ( $ptkg_tgt_type eq 'tio' )
   {  @inte_data = mmr_read_group($inte_addr_ref);
      @inte_copy = ();
      $inte_mask = 0xFFFFFFFFFFFFFFF7;
      for ($i = 0; $i <= $#inte_data; $i++)
      {  $inte_copy[$i] = $inte_data[$i] & $inte_mask;
      }
      mmr_write_group($inte_addr_ref, \@inte_copy);
      $ts_data = mmr_read($TIO_LB_HW_TIME_STAMP_MMR);
   }

   #--------------------------------------- 
   #  Write MMRs to launch packet (header and data)
   #--------------------------------------- 
   $mode = $PKTGEN_CNTRL_SET_START | $PKTGEN_WAIT_BUSY;
   $PKTGEN_MAX_BUSY_POLL = 20;      # limit the number of loops in pktgen_wait_if_busy()
   pktgen_write($mode, \%rq_hdr);
   if ( $ecode < 0 ) 
   {  mmr_write($emask_addr, $emask_data);  # Restore mask setting
      if ( $ptkg_tgt_type eq 'tio' )        # Restore interrupt settings
      {  mmr_write_group($inte_addr_ref, \@inte_data);
         mmr_write($TIO_EVENT_OCCURRED_ALIAS_MMR, 0x0000000000000008); 
         mmr_write($TIO_LB_HW_TIME_STAMP_MMR, $ts_data);
      }
      return;                               # Return: short-circuit if error detected and returned
   }

   #--------------------------------------- 
   #  Monitor error registers for response packet
   #--------------------------------------- 
   @edata = eregs_wait_summary_read_clear($EREG_ERROR_MASK_RP_BAD_CMD, $pktgen_retry_timeout);
   if ( ($#edata <= 0) || ($ecode < 0) )
   {  mmr_write($emask_addr, $emask_data);  # Restore mask setting
      if ( $ptkg_tgt_type eq 'tio' )        # Restore interrupt settings
      {  mmr_write_group($inte_addr_ref, \@inte_data);
         mmr_write($TIO_EVENT_OCCURRED_ALIAS_MMR, 0x0000000000000008); 
         mmr_write($TIO_LB_HW_TIME_STAMP_MMR, $ts_data);
      }
      $emess = "Packet generated MMR read operation unsuccessful.";
      $emess .= "\nUnable to detect error or clear error reporting registers.";
      $ecode = -1;
      &$comm_err_fun($emess);
      return;
   }
   mmr_write($emask_addr, $emask_data);  # Restore mask setting

   #--------------------------------------- 
   #  Restore the interrupt enable registers 
   #--------------------------------------- 
   if ( $ptkg_tgt_type eq 'tio' )
   {  mmr_write_group($inte_addr_ref, \@inte_data);
      mmr_write($TIO_EVENT_OCCURRED_ALIAS_MMR, 0x0000000000000008); 
      mmr_write($TIO_LB_HW_TIME_STAMP_MMR, $ts_data);
   }
 
   #--------------------------------------- 
   #  Examine the error registers 
   #--------------------------------------- 
   $ereg_err_detail_1 = $edata[1];
   $ereg_err_detail_2 = $edata[2];

   if ( $ereg_err_detail_1 & 0x0002000000000000 )
   {  $emess = "Packet generated MMR read operation unsuccessful.";
      $emess .= "\nUnexpected response packet captured in error detail registers.";
      $emess .= "\nLB_ERROR_DETAIL_1 has DATA_ERR set (0x0002000000000000).";
      $emess .= "\nUnable to confirm completion of packet generated request.";
      $ecode = -1;
      printf "\nActual data from LB ERROR MMRs:\n";
      decode_lb_status(\@edata);
      &$comm_err_fun($emess);
      return;
   }
   if ( $ereg_err_detail_1 & 0x0001000000000000 )
   {  $emess = "Packet generated MMR read operation unsuccessful.";
      $emess .= "\nUnexpected response packet captured in error detail registers.";
      $emess .= "\nLB_ERROR_DETAIL_1 has HDR_ERR set (0x0001000000000000).";
      $emess .= "\nUnable to confirm completion of packet generated request.";
      $ecode = -1;
      printf "\nActual data from LB ERROR MMRs:\n";
      decode_lb_status(\@edata);
      &$comm_err_fun($emess);
      return;
   }

   $tmp = $ereg_err_detail_1 & 0xFF;
   if ( $tmp != 0xB4 )
   {  $emess = "Packet generated MMR read operation unsuccessful.";
      $emess .= "\nUnexpected response packet captured in error detail registers.";
      $text   = sprintf("LB_ERROR_DETAIL_1 does not have COMMAND of 0xB4 (COMMAND=0x%02X).", $tmp);
      $emess .= "\n$text";
      $emess .= "\nUnable to confirm completion of packet generated request.";
      $ecode = -1;
      printf "\nActual data from LB ERROR MMRs:\n";
      decode_lb_status(\@edata);
      &$comm_err_fun($emess);
      return;
   }

   $tmp = ($ereg_err_detail_2 & 0x1FFFFFFFFFFFF) << 3;
   if ( $tmp != $final_addr )
   {  $emess = "Packet generated MMR read operation unsuccessful.";
      $emess .= "\nUnexpected response packet captured in error detail registers.";
      $text  = sprintf("LB_ERROR_DETAIL_2 does not have ADDR of 0x%012llX (ADDR=0x%012llX).", 
                  $final_addr, $tmp);
      $emess .= "\n$text";
      $emess .= "\nUnable to confirm completion of packet generated request.";
      $ecode = -1;
      printf "\nActual data from LB ERROR MMRs:\n";
      decode_lb_status(\@edata);
      &$comm_err_fun($emess);
      return;
   }

   if ( $pktgen_verbose || ($debug_flag & $DEBUG_GENERAL) )
   {  printf "\nActual data from LB ERROR MMRs:\n";
      decode_lb_status(\@edata);
   }

   #--------------------------------------- 
   #  Data should  maintain quadword ordering
   #--------------------------------------- 
   $mmr_data = ( $final_addr & 0x8 ) ? $edata[3] : $edata[4];
   return($mmr_data);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: pktgen_mmr_write()
#
#    USAGE: pktgen_mmr_write($addr, $value)
#           pktgen_mmr_write($addr, $value, $src_id)
#           pktgen_mmr_write($addr, $value, $src_id, $dst_id)
#           pktgen_mmr_write($addr, $value, $src_id, $dst_id, $dst_chiplet)
#           pktgen_mmr_write($addr, $value, $src_id, $dst_id, $dst_chiplet, $suppl)
#
#    DESCRIPTION: The subroutine writes the single MMR location specified
#       by the "addr" parameter with "value" using the packet generator. 
#       If no other parameters are specified the NODE_ID field of the 
#       NODE_ID MMR is read, and the NODE_ID is used for both the source and 
#       destination. If only the source is provided, then that value is also 
#       used for the destination.
#
#-------------------------------------------------------------------------------

sub pktgen_mmr_write
{
   my $num = $#_ + 1;
   print "\npktgen_mmr_write(@_) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   #--------------------------------------- 
   #  Verify the parameters
   #--------------------------------------- 
   if ( ($num < 2) || ($num > 6) )
   {  $emess = "Invalid parameters passed to subroutine pktgen_mmr_write().";
      $ecode = -1;
      &$comm_err_fun($emess);
      return;
   }
   my $mmr_addr = $_[0];  # Address of MMR to write
   my $mmr_data = $_[1];  # Data to write to MMR
   my $suppl    = ( ($num == 6) ? pop(@_) : 0xFF );
   my ($src_id, $src_chiplet, $dst_id, $dst_chiplet);

   #--------------------------------------- 
   #  Support software emulation mode
   #--------------------------------------- 
   if ($mmr_access == $MMR_ACCESS_NONE)
   {  mmr_write($mmr_addr, $mmr_data);
      return;
   }

   #--------------------------------------- 
   #  Set the pointers and determine IDs
   #--------------------------------------- 
   splice(@_, 1, 1);   # Remove the MMR data entry from the array
   ($src_id, $src_chiplet, $dst_id, $dst_chiplet) = pktgen_setup('mmr', @_);

   if ( $pktgen_verbose || ($debug_flag & $DEBUG_GENERAL) )
   {  printf "\n";
      printf "---------------------------------------------------------------------\n";
      printf "-- PKTGEN MMR WRITE : ADDR=0x%016llX DATA=0x%016llX\n", $mmr_addr, $mmr_data;
      printf "--              SRC : NODE_ID=0x%05X CHIPLET_ID=0x%02X\n", $src_id, $src_chiplet;
      printf "--              DST : NODE_ID=0x%05X CHIPLET_ID=0x%02X\n", $dst_id, $dst_chiplet;
      printf "---------------------------------------------------------------------\n";
   }
   my ($final_addr, %rq_hdr, $mode, @edata, $ereg_err_detail_1, $ereg_err_detail_2,
       $emask_addr, $emask_data, $einit_data);
   my $ts_data = 0;

   #--------------------------------------- 
   #  Determine the node ID and adjust address
   #--------------------------------------- 
   $final_addr = $mmr_addr | ($dst_id << 38);

   #--------------------------------------- 
   #  Create and initialize the packet header
   #--------------------------------------- 
   %rq_hdr = chn32hdr_new();
   $rq_hdr{SRC_ID}         = $src_id;
   $rq_hdr{SRC_CHIPLET}    = $src_chiplet;
   $rq_hdr{DST_ID}         = $dst_id;
   $rq_hdr{DST_CHIPLET}    = $dst_chiplet;
   $rq_hdr{ADDR}           = $final_addr;
   $rq_hdr{ECHO}           = 0x123;
   $rq_hdr{CMD}            = 0x31;			# PIO Write
   $rq_hdr{SUPPL}          = $suppl;                    # Default = 0xff

   #--------------------------------------- 
   #  Data should  maintain quadword ordering
   #--------------------------------------- 
   if ( $final_addr & 0x8 )  
   {  @data = (0, $mmr_data);
   }  
   else
   {  @data = ($mmr_data, 0);
   }
 
   #--------------------------------------- 
   #  Initialize the error registers
   #--------------------------------------- 
   $emask_addr = $$ereg_addr_ref[$EREG_MASK];
   $emask_data = mmr_read($emask_addr);
   $einit_data = ( ($ptkg_tgt_type eq 'shub') ? ($emask_data | $EREG_ERROR_MASK_RP_BAD_CMD) :
                                                ($EREG_ERROR_MASK & ~($EREG_ERROR_MASK_RP_BAD_CMD)) );
   eregs_initialize($einit_data);
   if ( $ecode < 0 )
   {  $emess = "Packet generated MMR write operation unsuccessful.";
      $emess .= "\nUnable to initialize the error reporting registers.";
      $ecode = -1;
      &$comm_err_fun($emess);
      return;
   }

   #--------------------------------------- 
   #  Modify the interrupt enable registers 
   #--------------------------------------- 
   my (@inte_data, @inte_copy, $inte_mask, $i);
   if ( $ptkg_tgt_type eq 'tio' )
   {  @inte_data = mmr_read_group($inte_addr_ref);
      @inte_copy = ();
      $inte_mask = 0xFFFFFFFFFFFFFFF7;
      for ($i = 0; $i <= $#inte_data; $i++)
      {  $inte_copy[$i] = $inte_data[$i] & $inte_mask;
      }
      mmr_write_group($inte_addr_ref, \@inte_copy);
      $ts_data = mmr_read($TIO_LB_HW_TIME_STAMP_MMR);
   }

   #--------------------------------------- 
   #  Write MMRs to launch packet (header and data)
   #--------------------------------------- 
   $mode = $PKTGEN_CNTRL_SET_START | $PKTGEN_WAIT_BUSY;
   $PKTGEN_MAX_BUSY_POLL = 20;      # limit the number of loops in pktgen_wait_if_busy()
   pktgen_write($mode, \%rq_hdr, \@data);
   if ( $ecode < 0 )
   {  mmr_write($emask_addr, $emask_data);  # Restore mask setting
      if ( $ptkg_tgt_type eq 'tio' )        # Restore interrupt settings
      {  mmr_write_group($inte_addr_ref, \@inte_data);
         mmr_write($TIO_EVENT_OCCURRED_ALIAS_MMR, 0x0000000000000008); 
         mmr_write($TIO_LB_HW_TIME_STAMP_MMR, $ts_data);
      }
      return;
   }
 
   #--------------------------------------- 
   #  Monitor error registers for response packet
   #--------------------------------------- 
   @edata = eregs_wait_summary_read_clear($EREG_ERROR_MASK_RP_BAD_CMD, $pktgen_retry_timeout);
   if ( ($#edata <= 0) || ($ecode < 0) )
   {  mmr_write($emask_addr, $emask_data);  # Restore mask setting
      if ( $ptkg_tgt_type eq 'tio' )        # Restore interrupt settings
      {  mmr_write_group($inte_addr_ref, \@inte_data);
         mmr_write($TIO_EVENT_OCCURRED_ALIAS_MMR, 0x0000000000000008); 
         mmr_write($TIO_LB_HW_TIME_STAMP_MMR, $ts_data);
      }
      $emess = "Packet generated MMR write operation unsuccessful.";
      $emess .= "\nUnable to detect error or clear error reporting registers.";
      $ecode = -1;
      &$comm_err_fun($emess);
      return;
   }
   mmr_write($emask_addr, $emask_data);  # Restore mask setting

   #--------------------------------------- 
   #  Restore the interrupt enable registers 
   #--------------------------------------- 
   if ( $ptkg_tgt_type eq 'tio' )
   {  mmr_write_group($inte_addr_ref, \@inte_data);
      mmr_write($TIO_EVENT_OCCURRED_ALIAS_MMR, 0x0000000000000008); 
      mmr_write($TIO_LB_HW_TIME_STAMP_MMR, $ts_data);
   }
 
   #--------------------------------------- 
   #  Examine the error registers 
   #--------------------------------------- 
   $ereg_err_detail_1 = $edata[1];
   $ereg_err_detail_2 = $edata[2];

   if ( $ereg_err_detail_1 & 0x0002000000000000 )
   {  $emess = "Packet generated MMR write operation unsuccessful.";
      $emess .= "\nUnexpected response packet captured in error detail registers.";
      $emess .= "\nLB_ERROR_DETAIL_1 has DATA_ERR set (0x0002000000000000).";
      $emess .= "\nUnable to confirm completion of packet generated request.";
      $ecode = -1;
      printf "\nActual data from LB ERROR MMRs:\n";
      decode_lb_status(\@edata);
      &$comm_err_fun($emess);
      return;
   }
   if ( $ereg_err_detail_1 & 0x0001000000000000 )
   {  $emess = "Packet generated MMR write operation unsuccessful.";
      $emess .= "\nUnexpected response packet captured in error detail registers.";
      $emess .= "\nLB_ERROR_DETAIL_1 has HDR_ERR set (0x0001000000000000).";
      $emess .= "\nUnable to confirm completion of packet generated request.";
      $ecode = -1;
      printf "\nActual data from LB ERROR MMRs:\n";
      decode_lb_status(\@edata);
      &$comm_err_fun($emess);
      return;
   }

   $tmp = $ereg_err_detail_1 & 0xFF;
   if ( $tmp != 0x92 )
   {  $emess = "Packet generated MMR write operation unsuccessful.";
      $emess .= "\nUnexpected response packet captured in error detail registers.";
      $text   = sprintf("LB_ERROR_DETAIL_1 does not have COMMAND of 0x92 (COMMAND=0x%02X).", $tmp);
      $emess .= "\n$text";
      $emess .= "\nUnable to confirm completion of packet generated request.";
      $ecode = -1;
      printf "\nActual data from LB ERROR MMRs:\n";
      decode_lb_status(\@edata);
      &$comm_err_fun($emess);
      return;
   }

   $tmp = ($ereg_err_detail_2 & 0x1FFFFFFFFFFFF) << 3;
   if ( $tmp != $final_addr )
   {  $emess = "Packet generated MMR write operation unsuccessful.";
      $emess .= "\nUnexpected response packet captured in error detail registers.";
      $text  = sprintf("LB_ERROR_DETAIL_2 does not have ADDR of 0x%012llX (ADDR=0x%012llX).", $final_addr, $tmp);
      $emess .= "\n$text";
      $emess .= "\nUnable to confirm completion of packet generated request.";
      $ecode = -1;
      printf "\nActual data from LB ERROR MMRs:\n";
      decode_lb_status(\@edata);
      &$comm_err_fun($emess);
      return;
   }

   if ( $pktgen_verbose || ($debug_flag & $DEBUG_GENERAL) )
   {  printf "\nActual data from LB ERROR MMRs:\n";
      decode_lb_status(\@edata);
   }
#decode_lb_status(\@edata); # HACK
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: pktgen_write()
#
#    USAGE: pktgen_write($mode, \%hdr)
#           pktgen_write($mode, \%hdr, \@data)
#           pktgen_write($mode, \%hdr, \@data, $epoint)
#
#    DESCRIPTION: This subroutine will write the MMRs associated with the
#       packet generator (diagnostic message generator). 
#
#-------------------------------------------------------------------------------

sub pktgen_write
{  my $num = $#_ + 1;
   print "\npktgen_write(@_) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $nwords = 0;
   my $aref   = 0;
   my @adata  = ();
   my $epoint = 0;
   my $i      = 0;

   #
   #  Check and extract the parameters
   #
   if ( ($num == 3) || ($num == 4) )
   {  $aref = $_[2]; 
      if ( ref($aref) ne 'ARRAY' )
      {  $emess = "Invalid parameter passed to subroutine pktgen_write().";
         $emess .= "\nThird parameter must be an ARRAY reference.";
         $ecode = -1;
         &$comm_err_fun($emess);
         return;
      }
      @adata  = @$aref;
      $nwords = $#adata + 1;
      if ( ($nwords < 0) || ($nwords > $PKTGEN_MAX_DATA_WORDS) )
      {  $emess = "Invalid data double words passed to subroutine pktgen_write().";
         $emess .= "\nNumber of data double words must be between 0 and $PKTGEN_MAX_DBL_WORDS.\n";
         $ecode = -1;
         &$comm_err_fun($emess);
         return;
      }
      if ( $num == 4 )
      {  $epoint = $_[3];
      }
   }
   elsif ( $num != 2 )
   {  $emess = "Invalid parameters passed to subroutine pktgen_write().";
      $ecode = -1;
      &$comm_err_fun($emess);
      return;
   }
   my $mode = $_[0];
   my $href = $_[1]; 
   if ( ref($href) ne 'HASH' )
   {  $emess = "Invalid parameter passed to subroutine pktgen_write().";
      $emess .= "\nSecond parameter must be an ARRAY reference.";
      $ecode = -1;
      &$comm_err_fun($emess);
      return;
   }

   my @hdr_dwords = chn32hdr_build($href);

   #
   #  Build the value for the DIAG_MSG_CNTRL register
   #
   my $value =( $nwords >> 1 ) & 0x3F;
   $value |= $DIAG_MSG_CNTRL_PORT   if ( $mode & $PKTGEN_CNTRL_RPLY_CHN );
   $value |= $DIAG_MSG_CNTRL_START  if ( $mode & $PKTGEN_CNTRL_SET_START );
   if ( $num == 4 )
   {  $value |= $DIAG_MSG_CNTRL_IEE;
      $value |= ($epoint << 6) & 0xFC0;
   }

   #
   #  Build arrays with the address and data values
   #
   my @addr = ( $$pktg_addr_ref[$PKTGEN_IND_HDR0], $$pktg_addr_ref[$PKTGEN_IND_HDR1] );
   my @data = chn32hdr_build($href);
   if ( $nwords > 0 )
   {  for ($i = 0; $i < $nwords; $i++)
      {  push(@addr, $$pktg_addr_ref[$PKTGEN_IND_DATA0L + $i]);
         push(@data, $adata[$i]);
      }
   }
   push(@addr, $$pktg_addr_ref[$PKTGEN_IND_CNTRL]);
   push(@data, $value);

   if ( $pktgen_verbose || ($debug_flag & $DEBUG_GENERAL) )
   {  printf "\nWriting header and data for packet generator:\n";
      chn32hdr_dump($href);
      printf "\nActual data for DIAG_MSG_HDR and DIAG_MSG_DATA MMRs:\n";
      for ($i = 0; $i <= $#addr; $i++)
      {  printf "MMR @ 0x%016llX = 0x%016llX\n", $addr[$i], $data[$i];
      }
   }

   #
   #  Write the HEADER MMRs and DATA MMRs (if given)
   #
   mmr_write_group(\@addr, \@data);
   return if ( $ecode < 0 ); # Return: short-circuit if error detected and returned

   #
   #  Check mode: wait for idle if desired
   #
   if ( $mode & $PKTGEN_WAIT_BUSY )
   {
      if ( $pktgen_verbose )
      {  print "\nChecking BUSY flag of DIAG_MSG_CNTRL; wait while busy...\n";
      }

      pktgen_wait_if_busy();
      return if ( $ecode < 0 ); # Return: short-circuit if error detected and returned
   }
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: pktgen_read()
#
#    USAGE: %hdr = pktgen_read()
#           %hdr = pktgen_read($mode)
#           %hdr = pktgen_read($mode, $num, \@data)
#
#    DESCRIPTION: This subroutine will read the MMRs associated with the
#       packet generator (diagnostic message generator).  The two header
#       MMRs are read and returned in a chn32hdr structure.  The "num"
#       parameter specifies the number of double words to read (starting
#       with DIAG_MSG_DATA0L) and store in "data".  There are two MMRs for
#       each double word; there will be two entries in the "data" array.
#       If "num" is not given, or is zero, then no data MMRs are read.
#
#       The "mode" parameter can be:
#          0 = read and return MMRs immediately (default if "mode" not given)
#          $PKTGEN_WAIT_BUSY = wait for packeg generator to be idle
#
#    RETURN VALUE: $TRUE when packet generator is busy; $FALSE if idle.
#
#-------------------------------------------------------------------------------

sub pktgen_read
{  my $num = $#_ + 1;
   print "\npktgen_read(@_) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $mode   = 0;
   my $nwords = 0;
   my $aref   = 0;
   my $i      = 0;

   #
   #  Check and extract the parameters
   #
   if ( $num == 3 )
   {  $mode   = $_[0];
      if ( ($_[1] < 0) || ($_[1] > $PKTGEN_MAX_DBL_WORDS) )
      {  $emess = "Invalid number of double words passed to subroutine pktgen_read().";
         $emess .= "\nNumber of data double words must be between 0 and $PKTGEN_MAX_DATA_WORDS.\n";
         $ecode = -1;
         &$comm_err_fun($emess);
         return;
      }
      $nwords = $_[1] * 2;
      $aref   = $_[2]; 
      if ( ref($aref) ne 'ARRAY' )
      {  $emess = "Invalid parameter passed to subroutine pktgen_read().";
         $emess .= "\nThird parameter must be an ARRAY reference.";
         $ecode = -1;
         &$comm_err_fun($emess);
         return;
      }
      @$aref = ();
   }
   elsif ( $num == 1 )
   {  $mode  = $_[0];
   }
   elsif ( $num != 0 )
   {  $emess = "Invalid parameters passed to subroutine pktgen_read().";
      $ecode = -1;
      &$comm_err_fun($emess);
      return;
   }
   if ( $debug_flag & $DEBUG_GENERAL )
   {  printf "read mode=0x%04X,  number of data MMRs=%lu\n", $mode, $nwords;
   }

   #
   #  Check mode: wait for idle if desired
   #
   if ( $mode & $PKTGEN_WAIT_BUSY )
   {
      if ( $pktgen_verbose )
      {  print "\nChecking BUSY flag of DIAG_MSG_CNTRL; wait while busy...\n";
      }

      pktgen_wait_if_busy();
      return if ( $ecode < 0 ); # Return: short-circuit if error detected and returned
   }

   #
   #  Read the header MMRs and extract the structure fields
   #
   my @addr = ( $$pktg_addr_ref[$PKTGEN_IND_HDR0], $$pktg_addr_ref[$PKTGEN_IND_HDR1] );
   my @data = mmr_read_group( \@addr );
   return if ( $ecode < 0 ); # Return: short-circuit if error detected and returned
   my %hdr = chn32hdr_parse(\@data);

   if ( $pktgen_verbose || ($debug_flag & $DEBUG_GENERAL) )
   {  printf "\nReading header and data for packet generator:\n";
      chn32hdr_dump(\%hdr);
      printf "\nActual data from DIAG_MSG_HDR MMRs:\n";
      printf "MMR read @ 0x%016llX = 0x%016llX\n", $addr[0], $data[0];
      printf "MMR read @ 0x%016llX = 0x%016llX\n", $addr[1], $data[1];
   }

   #
   #  Read the data MMRs and store the values
   #
   my $val = 0;
   if ( $nwords > 0 )
   {   @addr = ();
       for ($i = 0; $i < $nwords; $i++)
       {  $addr[$i] = $$pktg_addr_ref[$PKTGEN_IND_DATA0L + $i];
       }
       @data = mmr_read_group(\@addr);
       return if ( $ecode < 0 ); # Return: short-circuit if error detected and returned

      for ($i = 0; $i < $nwords; $i++)
      {  $val = $data[$i];
         @$aref[$i] = $val;
      }

      if ( $pktgen_verbose || ($debug_flag & $DEBUG_GENERAL) )
      {  printf "\nActual data from DIAG_MSG_HDR_DATA MMRs:\n";
         for ($i = 0; $i < $nwords; $i++)
         {  printf "MMR @ 0x%016llX = 0x%016llX\n", $addr[$i], $val;
         }
      }
   }

   return(%hdr);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: pktgen_is_busy()
#
#    USAGE: pktgen_is_busy()
#
#    DESCRIPTION: This subroutine will check the status of the packet
#       generator by reading the control register and examining the BUSY bit.
#
#    RETURN VALUE: $TRUE when packet generator is busy; $FALSE if idle.
#
#-------------------------------------------------------------------------------

sub pktgen_is_busy
{  my $num = $#_ + 1;
   print "\npktgen_is_busy(@_) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 0 )
   {  $emess = "Invalid parameters passed to subroutine pktgen_is_busy().";
      $ecode = -1;
      &$comm_err_fun($emess);
      return;
   }

   my $addr = $$pktg_addr_ref[$PKTGEN_IND_CNTRL];
   my $data = mmr_read($addr);
   return if ( $ecode < 0 ); # Return: short-circuit if error detected and returned

   if ( ($debug_flag & $DEBUG_GENERAL) || $pktgen_verbose )
   {  printf "LB_DIAG_MSG_CNTRL @ 0x%016llX = 0x%016llX\n", $addr, $data;
   }

   return( ($data & $DIAG_MSG_CNTRL_BUSY) ? $TRUE : $FALSE );
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: pktgen_wait_if_busy()
#
#    USAGE: pktgen_wait_if_busy()
#           pktgen_wait_if_busy($timeout)
#
#    DESCRIPTION: This subroutine will check the status of the packet
#       generator, and wait/loop until it becomes idle.  The maximum number
#       of times to read and check the control register is an optional
#       parameter (default is $PKTGEN_MAX_BUSY_POLL).
#
#    RETURN VALUE: $TRUE when packet generator is idle (not busy)
#
#-------------------------------------------------------------------------------

sub pktgen_wait_if_busy
{  my $num = $#_ + 1;
   print "\npktgen_wait_if_busy(@_) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $counter = $PKTGEN_MAX_BUSY_POLL;
   my $i = 0;

   if ( $num == 1 )
   {  $counter = $_[0];
   }
   elsif ( $num != 0 )
   {  $emess = "Invalid parameters passed to subroutine pktgen_wait_if_busy().";
      $ecode = -1;
      &$comm_err_fun($emess);
      return;
   }
   if ($debug_flag & $DEBUG_GENERAL )
   {  print "maximum poll counter = $counter\n";
   }

   #
   #  LOOP: check if busy (return $TRUE if not busy)
   #
   for ($i = 0; $i < $counter; $i++)
   {  return($TRUE)  if ( pktgen_is_busy() == $FALSE );
      return if ( $ecode < 0 ); # Return: short-circuit if error detected and returned
   }

   $emess = "Timeout waiting for BUSY flag of DIAG_MSG_CNTRL to clear.\nSubroutine pktgen_wait_if_busy() using maximum poll counter of $counter.";
   $ecode = -1;
   &$comm_err_fun($emess);
   return($FALSE);;
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: chn32hdr_build()
#
#    USAGE: @regdata = chn32hdr_build(\%hdr)
#
#    DESCRIPTION: This function will build the MMR double-words from the
#       specified header structure.
#
#    RETURN VALUE: Array containing two double-words (MMR HDR0 and HDR1)
#
#------------------------------------------------------------------------------

sub chn32hdr_build
{
   my $num = $#_ + 1;
   print "\nchn32hdr_build(@_) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   #
   #  Check and extract the parameters
   #
   if ( $num != 1 )
   {  $emess = "Invalid parameters passed to subroutine chn32hdr_build().";
      $ecode = -1;
      &$comm_err_fun($emess);
      return;
   }
   my $href = $_[0];
   if ( ref($href) ne 'HASH' )
   {  $emess = "Invalid parameter passed to subroutine chn32hdr_build().";
      $emess .= "\nFirst parameter must be a HASH reference.";
      $ecode = -1;
      &$comm_err_fun($emess);
      return;
   }
   my %hdata = %$href;

   #
   #  Build the packet header values from the hash structure
   #
   my @adata = ( 0x00000000, 0x00000000);
   my $label = "";
   my $val   = 0;
   my ($elem, @parms);
   foreach $elem (@chn32hdr_fields)
   {  @parms = @$elem;
#      printf "$elem = @parms\n";
      $label = $parms[$CHN32HDR_FIELD_LABEL];
      $val   = $hdata{$label} >> $parms[$CHN32HDR_FIELD_ADJUST];
      $val   = ($val & $parms[$CHN32HDR_FIELD_MASK]) << $parms[$CHN32HDR_FIELD_START];
      $adata[$parms[$CHN32HDR_FIELD_INDEX]] |= $val;
   }

   return(@adata);
}

#------------------------------------------------------------------------------
#
#    SUBROUTINE: chn32hdr_parse()
#
#    USAGE: %hdr = chn32hdr_parse(\@regdata)
#
#    DESCRIPTION: This function will parse the MMR header double-words
#       and place the values in the corresponding header entries.
#
#    RETURN VALUE:  A hash structure containing the header entries.
#
#------------------------------------------------------------------------------

sub chn32hdr_parse
{
   my $num = $#_ + 1;
   print "\nchn32hdr_parse(@_) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   #
   #  Check and extract the parameters
   #
   if ( $num != 1 )
   {  $emess = "Invalid parameters passed to subroutine chn32hdr_parse().";
      $ecode = -1;
      &$comm_err_fun($emess);
      return;
   }
   my $aref = $_[0];
   if ( ref($aref) ne 'ARRAY' )
   {  $emess = "Invalid parameter passed to subroutine chn32hdr_parse().";
      $emess .= "\nFirst parameter must be an ARRAY reference.";
      $ecode = -1;
      &$comm_err_fun($emess);
      return;
   }
   my @adata = @$aref;
   my %hdr   = chn32hdr_new();
   my $val   = 0;
   my $label = "";
   my ($elem, @parms);

   foreach $elem (@chn32hdr_fields)
   {  @parms = @$elem;
#      printf "$elem = @parms\n";
      $label = $parms[$CHN32HDR_FIELD_LABEL];
      $val   = $adata[$parms[$CHN32HDR_FIELD_INDEX]] >> $parms[$CHN32HDR_FIELD_START];
      $val   = ($val & $parms[$CHN32HDR_FIELD_MASK]) << $parms[$CHN32HDR_FIELD_ADJUST];
      $hdr{$label} = $val;
   }

   return(%hdr);
}


#------------------------------------------------------------------------------
#
#    SUBROUTINE: chn32hdr_new()
#
#    USAGE: %hdr = chn32hdr_new()
#
#    DESCRIPTION: This function will create and initialize a header hash
#       structure.
#
#    RETURN VALUE:  Initialize header hash structure.
#
#------------------------------------------------------------------------------

sub chn32hdr_new
{  my %chn32hdr = ();
   my $key;
   foreach $key (@chn32hdr_labels)
   {  $chn32hdr{$key} = 0;
   }
   return( %chn32hdr );
}


#------------------------------------------------------------------------------
#
#    SUBROUTINE: chn32hdr_dump()
#
#    USAGE: chn32hdr_dump(\@regdata)
#
#    DESCRIPTION: This function will display the contents of a header hash  
#       structure.
#
#    RETURN VALUE: none
#
#------------------------------------------------------------------------------

sub chn32hdr_dump
{  
   my $num = $#_ + 1;
   print "\nchn32hdr_dump(@_) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   #
   #  Check and extract the parameters
   #
   if ( $num != 1 )
   {  $emess = "Invalid parameters passed to subroutine chn32hdr_dump().";
      $ecode = -1;
      &$comm_err_fun($emess);
      return;
   }
   my $href = $_[0];
   if ( ref($href) ne 'HASH' )
   {  $emess = "Invalid parameter passed to subroutine chn32hdr_dump().";
      $emess .= "\nFirst parameter must be a HASH reference.";
      $ecode = -1;
      &$comm_err_fun($emess);
      return;
   }
   my %hdata = %$href;
   my $key;

   #
   #  Display the contents of the channel header structure
   #
   print "\nContents of CHAN_3_2_HEADER table:\n";
   foreach $key (@chn32hdr_labels)
   {  # need to apply mask to value before displaying!
      printf "%-12s = 0x%016llX = %u\n", $key, $hdata{$key}, $hdata{$key};
   }
}


#------------------------------------------------------------------------------

1;
