#
#/**************************************************************************
# *                                                                        *
# *               Copyright (C) 1989-1999 Silicon Graphics, Inc.           *
# *                                                                        *
# *  These coded instructions, statements, and computer programs  contain  *
# *  unpublished  proprietary  information of Silicon Graphics, Inc., and  *
# *  are protected by Federal copyright law.  They  may  not be disclosed  *
# *  to  third  parties  or copied or duplicated in any form, in whole or  *
# *  in part, without the prior written consent of Silicon Graphics, Inc.  *
# *                                                                        *
# **************************************************************************/
#
# @(#)$Id: reg_subs.pm,v 1.12 2005/06/15 22:01:23 beighley Exp $


#-------------------------------------------------------------------------------
#
#  FILE: reg_subs.pm
#
#  DESCRIPTION:  This file contains PERL subroutines to access and manipulate a
#    register representation.  A register instance contains the register value, 
#    and a pointer to the register definition structure.  The pointer, or 
#    reference, to the register instance is required for the subroutines.
#
#    A register definition structure contains the register name, register 
#    address, note/description, reset value and a group of field defintions.  
#    Each field definition contains a label, starting bit position, length, 
#    mask, access mode, reset value and description.  
#
#  SUBROUTINES:  The following subroutines are contained in this file:
#    reg_create_mmr_inst($asic_name, $mmr_name) -- create register structure
#    reg_create_mmr_inst($asic_name, $mmr_addr, MMR_ADDRESS) -- use MMR address
#    reg_create_mmr_list($asic_name, $mmr_names) -- create list of registers
#    reg_is_valid($reg_ref)  -- confirm created register is valid
#    reg_name($reg_ref)      -- get the register name
#    reg_note($reg_ref)      -- get the register note or description
#    reg_addr($reg_ref)      -- get the register address
#    reg_reset($reg_ref)     -- set the register to the "reset" value
#    reg_get_value($reg_ref) -- get the current register value
#    reg_set_value($reg_ref, $value) -- set the register value (entire)
#    reg_get_fields($reg_ref)        -- get the field labels
#    reg_get_fields($reg_ref, ORDER) -- get the field labels in order
#    reg_get_num_fields($reg_ref)    -- get the number of fields
#    reg_get_field_value($reg_ref, $field) -- get the value of a specific field
#    reg_set_field_value($reg_ref, $field, $value) -- set value of a field
#    reg_field_parm($reg_ref, $field, $entry_index) -- get field parameters
#    reg_write_contents($reg_ref, $file_desc, $num_base) -- write register values
#    reg_write_value($reg_ref, $file_desc, $mode, $numeric_base) -- write register values
#
#  MODIFICATION HISTORY:
#
#     WHEN      WHO             WHAT
#     --------  --------------  -----------------------------------------
#     01/03/02  Karen Beighley  Original source code.
#     01/16/02  Karen Beighley  Added additional functions: reg_is_valid()
#                                  and reg_create_mmr_inst().
#     02/04/02  Karen Beighley  Added additional functions: reg_dump()
#                                  and reg_get_num_fields().
#                               Added optional parameter to: reg_get_fields().
#     03/19/02  Karen Beighley  Added additional function: reg_write_contents().
#     04/03/02  Karen Beighley  Changed field bit position string.
#     05/21/02  Karen Beighley  Added additional function: reg_write_value().
#     07/09/02  Karen Beighley  Added additional function: reg_create_mmr_list().
#     01/09/03  Karen Beighley  Hardcoded any SHUB or PIC ASIC to use the same
#                                  MMR database (base version is valid).
#     10/16/03  Karen Beighley  Added additional function: reg_find_sorted_index().
#     12/09/03  Karen Beighley  Added support for register TYPE parameter.
#     02/18/04  Karen Beighley  Added support for new chip type: NL4RTR
#
#-------------------------------------------------------------------------------

$USE_REG_SUBS_PM = 1;    # Define to allow test for module inclusion


#-------------------------------------------------------------------------------
#  CONSTANTS --> these values must not change!
#-------------------------------------------------------------------------------

$REG_FLD_INDEX  = 0;
$REG_FLD_START  = 1;
$REG_FLD_LENGTH = 2;
$REG_FLD_MASK   = 3;
$REG_FLD_ACCESS = 4;
$REG_FLD_RFLAG  = 5;
$REG_FLD_RESET  = 6;
$REG_FLD_DESC   = 7;
$REG_MAX_FLD    = 8;

$REG_INDEX_PTR  = 0;
$REG_INDEX_DATA = 1;


$REG_TYPE_NORMAL = 0;
$REG_TYPE_PKTGEN = 1;


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: reg_create_mmr_inst()
#
#    USAGE: reg_create_mmr_inst($asic_name, $mmr_name)
#           reg_create_mmr_inst($asic_name, $mmr_name, MMR_NAME)
#           reg_create_mmr_inst($asic_name, $mmr_addr, MMR_ADDRESS)
#
#    DESCRIPTION: This subroutine will create a register structure for a
#       specific memory mapped register (MMR).  The ASIC name is required, and
#       must match that defined in the ASIC's MMR module (i.e. "shub_01").
#       The MMR can be specified using the name or address.
#
#    RETURN VALUE: A reference to a register structure.  The "reg_is_valid()"
#       subroutine can be used to confirm the register creation was successful.
#
#-------------------------------------------------------------------------------

sub reg_create_mmr_inst
{
   my $num = $#_ + 1;
#print "\nreg_create_mmr_inst(@_) : num=$num\n";

   my @rdata = (0, 0);
   my $ptr   = 0;
   my $nflag = 1;

   if ( $num == 3 )
   {  if ( $_[2] eq MMR_ADDRESS )
      {  $nflag = 0;
      }
      elsif ( $_[2] ne MMR_NAME )
      {  return(\@rdata);
      }
   }
   elsif ( $num != 2 )
   {  return(\@rdata);
   }
   my $def_name = $_[0];
   my $base_name = $def_name;
   if ( $def_name =~ /^shub2/i )
   {  $base_name = "shub2_01";
   }
   elsif ( $def_name =~ /^shub/i )
   {  $base_name = "shub_01";
   }
   elsif ( $def_name =~ /^pic/i )
   {  $base_name = "pic_01";
   }
   elsif ( $def_name =~ /^tio/i )
   {  $base_name = "tio_01";
   }
   elsif ( $def_name =~ /^nl4r/i )
   {  $base_name = "nl4rtr_01";
   }

   $def_name = $base_name . "_defined";
#print "check   : $def_name\n";
   return(\@rdata) if ( !defined($$def_name) );
#print "defined : $def_name\n";

   $def_name = $base_name;
   if ( $nflag )  # search by MMR name
   {  $def_name .= "_mmr_by_name";
#print "search by name: $def_name...\n";
   }
   else           # seach by MMR address
   {  $def_name .= "_mmr_by_addr";
#print "search by address: $def_addr...\n";
   }

   #  Search for the specified register
   if ( exists $$def_name{$_[1]} )
   {  $ptr = $$def_name{$_[1]};
      $rdata[$REG_INDEX_PTR] = $ptr;
      reg_reset(\@rdata);
   }

#printf "rdata = ( @rdata )\n";
   return(\@rdata);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: reg_create_mmr_list()
#
#    USAGE: reg_create_mmr_list($asic_name, $mmr_names)
#
#    DESCRIPTION: This subroutine will create an array of register structure 
#       for the specified memory mapped registers (MMRs).  The ASIC name is 
#       required, and must match that defined in the ASIC's MMR module 
#       (i.e. "shub_01").  The "mmr_names" parameter is a reference to an
#       array of MMRs names.
#
#    RETURN VALUE: A reference to an array or undefined if error.
#
#-------------------------------------------------------------------------------

sub reg_create_mmr_list
{
   print "\nreg_create_mmr_list( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   return if ( $#_ != 1 );
   my $type  = $_[0];  # Chip type: "shub_01", "pic_01", etc.
   my $nref  = $_[1];  # Reference to array with MMR names
   my ($name, $i, $reg_ref);

   #
   #  Create each register instance and add to the list
   #
   my @regs = ();
   for ($i = 0; $i <= $#$nref; $i++)
   {  $reg_ref = reg_create_mmr_inst($type, $$nref[$i]);
      return if ( !reg_is_valid($reg_ref) );
      $regs[$i] = $reg_ref;
   }
   return(\@regs);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: reg_is_valid()
#
#    USAGE: reg_is_valid($reg_ref)
#
#    DESCRIPTION: The subroutine will check the contents of a register
#      structure to confirm it contains a valid register definition pointer.
#
#    RETURN VALUE: 0 if the register is not valid; 1 if the register is valid.
#
#-------------------------------------------------------------------------------

sub reg_is_valid
{  return(0) if ( $#_ != 0 );
   my $reg_ptr = $_[0];
   return(0) if ( ref($reg_ptr) ne 'ARRAY' );
   my $def_ptr = $$reg_ptr[$REG_INDEX_PTR];
   return(0) if ( ref($def_ptr) ne 'HASH' );
   return(1);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: reg_find_sorted_index()
#
#    USAGE: reg_find_sorted_index($asic_name, $mmr_name)
#
#    DESCRIPTION: This subroutine will search for a specific name in the MMR
#       list for a specific chip type.
#
#    RETURN VALUE: A index into the MMR "sorted by name" array.
#
#-------------------------------------------------------------------------------

sub reg_find_sorted_index
{
   print "\nreg_find_sorted_index( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   return if ( $#_ != 1 );

   #
   #  Convert the chip type to format used in MMR data files
   #
   my $def_name = $_[0];
   my $mmr_name = uc($_[1]);
   my $base_name = $def_name;
   if ( $def_name =~ /^shub2/i )
   {  $base_name = "shub2_01";
   }
   elsif ( $def_name =~ /^shub/i )
   {  $base_name = "shub_01";
   }
   elsif ( $def_name =~ /^pic/i )
   {  $base_name = "pic_01";
   }
   elsif ( $def_name =~ /^tio/i )
   {  $base_name = "tio_01";
   }
   elsif ( $def_name =~ /^nl4r/i )
   {  $base_name = "nl4rtr_01";
   }
   $def_name = $base_name . "_defined";
   return if ( !defined($$def_name) );
   $def_name = $base_name . "_sorted_mmr_name";
#printf "\ndef_name = $def_name\n";

   #
   #  Attempt to find the specified MMR name
   #
   my $asic_mmr_list = \@$def_name;
   my ($i, $found);
  
   for ($i = 0, $found = 0;
        (($i <= $#$asic_mmr_list) && ($found == 0)); $i++)
   {  if ( $mmr_name eq $$asic_mmr_list[$i] )
      {  $found = 1;
      }
      elsif ( $mmr_name le $$asic_mmr_list[$i] )
      {   $found = 2
      }
   }
   return if ( $found != 1 );
   $i--;
#printf "i=$i -> $$asic_mmr_list[$i]\n";
   return($i); 
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: reg_name()
#
#    USAGE: reg_name($reg_ref)
#
#    DESCRIPTION: The subroutine will extract the register name from the
#      specified register structure.
#
#    RETURN VALUE: the name of the register (text string: "SH_II_INT0_CONFIG")
#
#-------------------------------------------------------------------------------

sub reg_name
{  return("") if ( $#_ != 0 );
   my $reg_ptr = $_[0];
   return("") if ( ref($reg_ptr) ne 'ARRAY' );
   my $def_ptr = $$reg_ptr[$REG_INDEX_PTR];
   return("") if ( ref($def_ptr) ne 'HASH' );
   return( $$def_ptr{NAME} );
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: reg_type()
#
#    USAGE: reg_type($reg_ref)
#
#    DESCRIPTION: The subroutine will extract the register type from the
#      specified register structure.
#
#    RETURN VALUE: the register type ($REG_TYPE_NORMAL or $REG_TYPE_PKTGEN)
#
#-------------------------------------------------------------------------------

sub reg_type
{  return("") if ( $#_ != 0 );
   my $reg_ptr = $_[0];
   return("") if ( ref($reg_ptr) ne 'ARRAY' );
   my $def_ptr = $$reg_ptr[$REG_INDEX_PTR];
   return("") if ( ref($def_ptr) ne 'HASH' );
   return( $$def_ptr{TYPE} ) if ( defined($$def_ptr{TYPE}) );
   return($REG_TYPE_NORMAL);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: reg_note()
#
#    USAGE: reg_note($reg_ref)
#
#    DESCRIPTION: The subroutine will extract any note (description) from the
#      specified register structure.
#
#    RETURN VALUE: the note for the register (text string)
#
#-------------------------------------------------------------------------------

sub reg_note
{  return("") if ( $#_ != 0 );
   my $reg_ptr = $_[0];
   return("") if ( ref($reg_ptr) ne 'ARRAY' );
   my $def_ptr = $$reg_ptr[$REG_INDEX_PTR];
   return("") if ( ref($def_ptr) ne 'HASH' );
   return( $$def_ptr{NOTE} );
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: reg_addr()
#
#    USAGE: reg_addr($reg_ref)
#
#    DESCRIPTION: The subroutine will extract the register address from the
#      specified register structure.
#
#    RETURN VALUE: the address for the register (64-bit value)
#
#-------------------------------------------------------------------------------

sub reg_addr
{  return(0xFFFFFFFFFFFFFFFF) if ( $#_ != 0 );
   my $reg_ptr = $_[0];
   return(0xFFFFFFFFFFFFFFFF) if ( ref($reg_ptr) ne 'ARRAY' );
   my $def_ptr = $$reg_ptr[$REG_INDEX_PTR];
   return(0xFFFFFFFFFFFFFFFF) if ( ref($def_ptr) ne 'HASH' );
   return( $$def_ptr{ADDRESS} );
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: reg_reset()
#
#    USAGE: reg_reset($reg_ref)
#
#    DESCRIPTION: The subroutine will change the current register value
#      to the value defined by "reset".
#
#    RETURN VALUE: the reset value of the register (64-bit value)
#
#-------------------------------------------------------------------------------

sub reg_reset
{  return(0xFFFFFFFFFFFFFFFF) if ( $#_ != 0 );
   my $reg_ptr = $_[0];
   return(0xFFFFFFFFFFFFFFFF) if ( ref($reg_ptr) ne 'ARRAY' );
   my $def_ptr = $$reg_ptr[$REG_INDEX_PTR];
   return(0xFFFFFFFFFFFFFFFF) if ( ref($def_ptr) ne 'HASH' );
   $$reg_ptr[$REG_INDEX_DATA] = $$def_ptr{RESET};
   return( $$reg_ptr[$REG_INDEX_DATA] );
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: reg_get_value()
#
#    USAGE: reg_get_value($reg_ref)
#
#    DESCRIPTION: The subroutine will extract the current register value from
#      the specified register structure.
#
#    RETURN VALUE: the value of the register (64-bit value)
#
#-------------------------------------------------------------------------------

sub reg_get_value
{  return(0xFFFFFFFFFFFFFFFF) if ( $#_ != 0 );
   my $reg_ptr = $_[0];
   return(0xFFFFFFFFFFFFFFFF) if ( ref($reg_ptr) ne 'ARRAY' );
   return( $$reg_ptr[$REG_INDEX_DATA] );
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: reg_set_value()
#
#    USAGE: reg_set_value($reg_ref, $value)
#
#    DESCRIPTION: The subroutine will change the current register value to
#      that specified by the "value" parameter.
#
#    RETURN VALUE: the value of the register (64-bit value)
#
#-------------------------------------------------------------------------------

sub reg_set_value
{  return if ( $#_ != 1 );
   my $reg_ptr = $_[0];
   return if ( ref($reg_ptr) ne 'ARRAY' );
   $$reg_ptr[$REG_INDEX_DATA] = $_[1];
   return( $$reg_ptr[$REG_INDEX_DATA] );
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: reg_get_fields()
#
#    USAGE: reg_get_fields($reg_ref)
#           reg_get_fields($reg_ref, ORDER)
#
#    DESCRIPTION: The subroutine will extract the label for each field 
#      defined within the register structure.  If the "ORDER" parameter is
#      specified, then the labels are ordered by starting bit position 
#      (biggest to smallest starting bit position).
#
#    RETURN VALUE: an array of labels (each label is a text string)
#
#-------------------------------------------------------------------------------

sub reg_get_fields
{  my $flag = 0;
   if ( $#_ == 1 )
   {  $flag = 1 if ( $_[1] eq ORDER );
   }
   elsif ( $#_ != 0 )
   {  return();
   }
   my $reg_ptr = $_[0];                           # pointer to register instance
   return () if ( ref($reg_ptr) ne 'ARRAY' );
   my $def_ptr = $$reg_ptr[$REG_INDEX_PTR];       # pointer to reg definition
   return () if ( ref($def_ptr) ne 'HASH' );
   my $fld_ptr = $$def_ptr{FIELDS};               # pointer to field hash table 
   my @fields  = (keys %$fld_ptr);                # field labels/names (random)
   return( @fields ) if ( $flag == 0 );           # return labels in any order
   my ($fld, $parm_ptr, $i, @name);
   @name = ();                                    # initialize name array
   foreach $fld ( @fields )                       # process each field in hash
   {  $parm_ptr = $$fld_ptr{$fld};                # pointer to field parameters
      $i        = $$parm_ptr[$REG_FLD_INDEX];     # index of field (bit ordered)
      $name[$i] = $fld;                           # save the label by index
   }
   return(@name);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: reg_get_num_fields()
#
#    USAGE: reg_get_num_fields($reg_ref)
#
#    DESCRIPTION: The subroutine will determine number of fields
#      defined within the register structure.
#
#    RETURN VALUE: number of fields in the register
#
#-------------------------------------------------------------------------------

sub reg_get_num_fields
{  return () if ( $#_ != 0 );
   my $reg_ptr = $_[0];
   return () if ( ref($reg_ptr) ne 'ARRAY' );
   my $def_ptr = $$reg_ptr[$REG_INDEX_PTR];
   return () if ( ref($def_ptr) ne 'HASH' );
   my $fld_ptr = $$def_ptr{FIELDS};
   my @flds = keys %$fld_ptr;
   return( ($#flds + 1) );
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: reg_get_field_value()
#
#    USAGE: reg_get_field_value($reg_ref, $field_label)
#
#    DESCRIPTION: The subroutine will extract the value of the specified
#      field from the given register structure.
#
#    RETURN VALUE: the value of the field (64-bit value)
#
#-------------------------------------------------------------------------------

sub reg_get_field_value
{  return("") if ( $#_ != 1 );
   my $reg_ptr = $_[0];
   return("") if ( ref($reg_ptr) ne 'ARRAY' );
   my $def_ptr = $$reg_ptr[$REG_INDEX_PTR];
   return("") if ( ref($def_ptr) ne 'HASH' );
   my $fld_ptr = $$def_ptr{FIELDS};
   return("") if ( !(exists $$fld_ptr{$_[1]}) );
   $def_ptr = $$fld_ptr{$_[1]};
   return ( ($$reg_ptr[$REG_INDEX_DATA] & $$def_ptr[$REG_FLD_MASK] ) 
            >> $$def_ptr[$REG_FLD_START] );
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: reg_set_field_value()
#
#    USAGE: reg_set_field_value($reg_ref, $field_label, $value)
#
#    DESCRIPTION: The subroutine will set the value of the specified
#      field in the given register structure.
#
#    RETURN VALUE: the value of the register (64-bit value)
#
#-------------------------------------------------------------------------------

sub reg_set_field_value
{  return("") if ( $#_ != 2 );
   my $reg_ptr = $_[0];
   return("") if ( ref($reg_ptr) ne 'ARRAY' );
   my $def_ptr = $$reg_ptr[$REG_INDEX_PTR];
   return("") if ( ref($def_ptr) ne 'HASH' );
   my $fld_ptr = $$def_ptr{FIELDS};
   return("") if ( !(exists $$fld_ptr{$_[1]}) );
   $def_ptr = $$fld_ptr{$_[1]};
   my $val = $$reg_ptr[$REG_INDEX_DATA] & ~($$def_ptr[$REG_FLD_MASK]);
   $val |= (($_[2] << $$def_ptr[$REG_FLD_START]) & $$def_ptr[$REG_FLD_MASK]);
   $$reg_ptr[$REG_INDEX_DATA] = $val;
   return( $val );
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: reg_field_parm()
#
#    USAGE: reg_field_parm($reg_ref, $field_label, $entry_index)
#
#    DESCRIPTION: The subroutine will extract field parameters for a specific
#      field in the given register structure.  The "entry_index" is a number
#      and can be one of: $REG_FLD_START, $REG_FLD_LENGTH, $REG_FLD_MASK,
#      $REG_FLD_ACCESS, $REG_FLD_RFLAG, $REG_FLD_RESET, $REG_FLD_DESC, or
#      $REG_MAX_FLD (note the '$' is required since this is a numeric index).
#
#    RETURN VALUE: the value of the field parameter
#
#-------------------------------------------------------------------------------

sub reg_field_parm
{  return("") if ( $#_ != 2 );
   my $reg_ptr = $_[0];
   return("") if ( ref($reg_ptr) ne 'ARRAY' );
   my $def_ptr = $$reg_ptr[$REG_INDEX_PTR];
   return("") if ( ref($def_ptr) ne 'HASH' );
   my $fld_ptr = $$def_ptr{FIELDS};
   return("") if ( !(exists $$fld_ptr{$_[1]}) );
   return("") if ( $_[2] > $REG_MAX_FLD );
   $def_ptr = $$fld_ptr{$_[1]};
   return( $$def_ptr[ $_[2] ] );
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: reg_write_contents()
#
#    USAGE: reg_write_contents($reg_ref, $file_desc)
#           reg_write_contents($reg_ref, $file_desc, $numeric_base)
#
#    DESCRIPTION: The subroutine will write register name, address, value
#       and field values to the specified file descriptor (open file or STDOUT).
#       The optional numeric base parameter can be: 'hex', 'oct' or 'dec'.
#       Hexidecimal is used as the default numeric base if not specified.
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub reg_write_contents
{
   print "\nreg_write_contents( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   return if ( $#_ < 1 );
   my $reg_ptr = $_[0];  # Register instance reference
   my $ofp     = $_[1];  # Open output file or standard out
   my $nbase = ($#_ > 1) ? $_[2] : 'hex';  # Optional numeric base

   return if ( !reg_is_valid($reg_ptr) );
   my ($dptr, $fptr, $aptr, @fields, $field, $val, $text, $nstr);

   $dptr   = $$reg_ptr[$REG_INDEX_PTR]; # pointer to definition structure
   $fptr   = $$dptr{FIELDS};            # pointer to field hash table
   @fields = reg_get_fields($reg_ptr, ORDER); # list of field names

   #
   #  Display the register name, address and contents
   #
   $val    = reg_get_value($reg_ptr);   # Value of register
   $nstr   = convert_num2text($val, 64, $nbase);
   printf $ofp "\nREGISTER %-32s @ 0x%09llX = %s\n", reg_name($reg_ptr),
      reg_addr($reg_ptr), $nstr;

   #
   #  Process each field within the register
   #
   foreach $field (@fields)
   {  $aptr = $$fptr{$field};           # pointer to field definition

      if ( $$aptr[$REG_FLD_LENGTH] == 1 )
      {  $text = sprintf("<%u>", $$aptr[$REG_FLD_START]);
      }
      else
      {  $text = sprintf("<%u:%u>", ($$aptr[$REG_FLD_START] + $$aptr[$REG_FLD_LENGTH] - 1),
            $$aptr[$REG_FLD_START]);
      }

      $val  = reg_get_field_value($reg_ptr, $field);
      $nstr = convert_num2text($val, $$aptr[$REG_FLD_LENGTH], $nbase);

      printf $ofp "         FIELD %-32s %-7s = %s\n", $field, $text, $nstr;
   }
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: reg_write_value()
#
#    USAGE: reg_write_value($reg_ref, $file_desc, $mode)
#           reg_write_value($reg_ref, $file_desc, $mode, $numeric_base)
#
#    DESCRIPTION: The subroutine will write register name, address, value
#       and field values to the specified file descriptor (open file or STDOUT).
#       The mode parameter determines which fields are displayed:
#          none  = no fields are displayed
#          all   = all fields are displayed
#          valid = fields with valid data (non-zero value)
#       The optional numeric base parameter can be: 'hex', 'oct' or 'dec'.
#       Hexidecimal is used as the default numeric base if not specified.
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub reg_write_value
{
   print "\nreg_write_value( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   return if ( $#_ < 2 );
   my $reg_ptr = $_[0];  # Register instance reference
   my $ofp     = $_[1];  # Open output file or standard out
   my $mode    = $_[2];  # Open output file or standard out
   my $nbase = ($#_ > 2) ? $_[3] : 'hex';  # Optional numeric base

   return if ( !reg_is_valid($reg_ptr) );
   my ($dptr, $fptr, $aptr, @fields, $field, $val, $text, $nstr, $name);

   $dptr   = $$reg_ptr[$REG_INDEX_PTR]; # pointer to definition structure
   $fptr   = $$dptr{FIELDS};            # pointer to field hash table
   @fields = reg_get_fields($reg_ptr, ORDER); # list of field names

   #
   #  Display the register name, address and contents
   #
   $val    = reg_get_value($reg_ptr);   # Value of register
   $nstr   = convert_num2text($val, 64, $nbase);
   printf $ofp "\n%-32s @ 0x%09llX = %s\n", reg_name($reg_ptr),
      reg_addr($reg_ptr), $nstr;

   #
   #  Process each field within the register (depending on mode)
   #
   foreach $field (@fields)
   {  $aptr = $$fptr{$field};           # pointer to field definition
      $val  = reg_get_field_value($reg_ptr, $field);
      if ( ($mode eq 'all') || (($mode eq 'valid') && ($val != 0)) )
      {  $nstr = convert_num2text($val, $$aptr[$REG_FLD_LENGTH], $nbase);
         $name = sprintf("%s:%s", reg_name($reg_ptr), $field);
         $text = ( ($$aptr[$REG_FLD_LENGTH] == 1) ? sprintf("<%u>", $$aptr[$REG_FLD_START]) :
                     sprintf("<%u:%u>", ($$aptr[$REG_FLD_START] + $$aptr[$REG_FLD_LENGTH] - 1),
                        $$aptr[$REG_FLD_START]) );
         printf $ofp "%-58s %-7s = %s\n", $name, $text, $nstr;
      }
   }
}

#-------------------------------------------------------------------------------

sub reg_dump
{  return if ( $#_ != 0 );
   my $reg_ptr = $_[0];
   return if ( !reg_is_valid($reg_ptr) );
   my ($name, $note, $addr, $data, @labels);

   print "\nREGISTER STRUCTURE CONTENTS:\n";
   $name = reg_name($reg_ptr);
   printf "name   = \"$name\"\n";
   $note = reg_note($reg_ptr);
   printf "note   = \"$note\"\n";
   $addr = reg_addr($reg_ptr);
   printf "addr   = 0x%016llX\n", $addr;
   $data = reg_get_value($reg_ptr);
   printf "value  = 0x%016llX\n", $data;
   @labels = reg_get_fields($reg_ptr, ORDER);
   print  "fields = ( @labels )\n";
}


#-------------------------------------------------------------------------------

1;

