#
#/**************************************************************************
# *                                                                        *
# *               Copyright (C) 1989-1999 Silicon Graphics, Inc.           *
# *                                                                        *
# *  These coded instructions, statements, and computer programs  contain  *
# *  unpublished  proprietary  information of Silicon Graphics, Inc., and  *
# *  are protected by Federal copyright law.  They  may  not be disclosed  *
# *  to  third  parties  or copied or duplicated in any form, in whole or  *
# *  in part, without the prior written consent of Silicon Graphics, Inc.  *
# *                                                                        *
# **************************************************************************/
#
# @(#)$Id: scancfg.pm,v 1.24 2006/08/11 19:56:01 beighley Exp $
#
#-------------------------------------------------------------------------------
#
#  FILE: scancfg.pm
#
#  DESCRIPTION:  This file contains PERL subroutines to handle load and
#     represent the system-level scan infrastructure.
#
#  MODIFICATION HISTORY:
#
#     WHEN      WHO             WHAT
#     --------  --------------  ------------------------------------------------
#     xx/xx/xx  Karen Beighley  Original source code.
#     11/26/03  Karen Beighley  Added modificiation history to file.
#                               Added RACK, SLOT and DESC entries to module
#                                  data structure (hash table).
#     12/14/03  Karen Beighley  Added subroutines: scan_module_write_ir() and
#                                  scan_module_write_tdr(). 
#     12/31/03  Karen Beighley  Disabled debug messages in new subroutines.
#     02/10/04  Karen Beighley  Major changes to support multi-chain modules.
#     02/11/04  Karen Beighley  Added chain index to: scancfg_hardware_init().
#     02/18/04  Karen Beighley  Added support for new chip type: NL4RTR
#     03/05/04  Karen Beighley  Fixed bug of missing module data (RACK/SLOT).
#     03/23/04  Karen Beighley  Minor change to local variables (added "my").
#     04/28/04  Karen Beighley  Initial support for SHUB2 (not complete!)
#     06/01/04  Karen Beighley  Fixed bug to reset TDR: scan_module_write_tdr().
#     01/04/05  Karen Beighley  Initial support for LBI (not complete!)
#     ...
#     08/11/06  Karen Beighley  Added subroutine: scancfg_ce_connected_to_tio()
#
#-------------------------------------------------------------------------------

use jtdr_subs;
use jtag_subs;

$USE_SCANCFG_PM = 1;    # Define to allow test for module inclusion


#-------------------------------------------------------------------------------
#  CONSTANTS - These values cannot change!
#-------------------------------------------------------------------------------

$CHN_IND_MODE   = 0;  # Chain array index: chain mode
$CHN_IND_FLAG   = 1;  # Chain array index: chain flag(s)
$CHN_IND_LOC    = 2;  # Chain array index: location/port on JTAG controller
$CHN_IND_IRLEN  = 3;  # Chain array index: total IR chain length
$CHN_IND_DRLEN  = 4;  # Chain array index: total DR chain length (per TDR)
$CHN_IND_NCHIP  = 5;  # Chain array index: number of chips in scan chain
$CHN_IND_CREFS  = 6;  # Chain array index: array of references (one per chip)
$MAX_CHN_IND    = 7;  # Number of elements in chain array

$CMP_IND_BOARD  = 0;   # Component array index: board name
$CMP_IND_PHY    = 1;   # Component array index: physical chip name
$CMP_IND_LOG    = 2;   # Component array index: logical chip name
$CMP_IND_DESC   = 3;   # Component array index: chip type/description
$CMP_IND_ILEN   = 4;   # Component array index: IR length of chip
$CMP_IND_IPOS   = 5;   # Component array index: IR start position of chip
$CMP_IND_FLAG   = 6;   # Component array index: flag of supported op's
$CMP_IND_CURR   = 7;   # Component array index: current IR/TDR index
$CMP_IND_LIST   = 8;   # Component array index: ref to array of IR/TDR pointers
$MAX_CMP_IND    = 9;   # Number of elements in component array

$CMP_FLAG_JTAG   = 0x0001;               # Flag: supports JTAG
$CMP_FLAG_TDR    = 0x0002;               # Flag: JTAG TDR definitions
$CMP_FLAG_MMR    = 0x0004;               # Flag: supports MMR access via JTAG
$CMP_FLAG_DEBUG  = 0x0008;               # Flag: supports DEBUG port select
$CMP_FLAG_ISCAN  = 0x0010;               # Flag: supports internal/full scan
$CMP_FLAG_PTKGEN = 0x0020;               # Flag: supports packet generator operations

$DETAIL_IND_TYPE  = 0;  # Chip type ("shub_01" or "pic_01")
$DETAIL_IND_ILEN  = 1;  # Total IR length for scan chain
$DETAIL_IND_IPOS  = 2;  # Chip's IR starting position
$DETAIL_IND_CLEN  = 3;  # Number of chips in scan chain
$DETAIL_IND_CPOS  = 4;  # Chip index in scan chain
$DETAIL_IND_MLOC  = 5;  # Module/brick location ID
$DETAIL_IND_NAME  = 6;  # Target name (module:board:chip)
$DETAIL_IND_MNUM  = 7;  # Number number/index in configuration
$DETAIL_IND_CNUM  = 8;  # Chain number/index on module


$JTDR_BGND_USE_CURRENT  = 0;  # Use the current value in the chip's TDR structure
$JTDR_BGND_USE_DEFAULT  = 1;  # Use the "default" setting for the JTAG TDR
$JTDR_BGND_USE_HARDWARE = 2;  # Read the actual TDR value from the hardware

$FLAG_CE_CHIP_PRESENT   = 0x0001;

$FALSE = 0;
$TRUE  = 1;


#-------------------------------------------------------------------------------
#  GLOBAL VARIABLES
#-------------------------------------------------------------------------------

%modules      = ();  # Module definitions by name: $modules{$name} = $mod_ref
@module_names = ();  # Array of module names (same order as @module_ptrs)
@module_ptrs  = ();  # Array of module pointers (same order as @module_names)


#-------------------------------------------------------------------------------
#  Default hard-coded configuration if no file specified
#-------------------------------------------------------------------------------

@module_00000_chips =
(
#               PHYSICAL   LOGICAL                   IR   IR            CURRENT  LIST
#    BOARD      NAME       NAME       DESCRIPTION    LEN  POS  FLAG     IR/TDR   IR/TDR
#    ---------  ---------  ---------  -------------  ---  ---  -------  -------  ------
   [ "NODE1" ,  "K4H6B" ,  "PROC2" ,  "mck_b0"    ,   8,   0,  0x0001,   0,      0    ],
   [ "NODE1" ,  "P9H6B" ,  "PROC3" ,  "mck_b0"    ,   8,   8,  0x0001,   0,      0    ],
   [ "NODE1" ,  "M7H6"  ,  "SHUB1" ,  "shub_01"   ,   7,  16,  0x000F,   0,      0    ],
   [ "NODE0" ,  "C0H6B" ,  "PROC0" ,  "mck_b0"    ,   8,  23,  0x0001,   0,      0    ],
   [ "NODE0" ,  "G5H6B" ,  "PROC1" ,  "mck_b0"    ,   8,  31,  0x0001,   0,      0    ],
   [ "NODE0" ,  "E2H6"  ,  "SHUB0" ,  "shub_01"   ,   7,  39,  0x000F,   0,      0    ]
);

@module_00000_chains =
(
#                     IR    DR   NUM    
#    MODE  FLAG  LOC  LEN   LEN  CHIPS  ARRAY OF CHIP REFS
#    ----  ----  ---  ----  ---  -----  ---------------------
   [   0,    0,   0,   46,   6,     6,  \@module_00000_chips ],
);

%module_00000 =
( NAME  => "001c01",     # Module name or brick name
  INDEX => 0,            # Index of module in configuration
  LOC   => 65,           # JTAG controller location/ID
  MODE  => 0,            # CFG mode: multiple SICs in parallel?
  FLAG  => 0,            # Flag: application-specific parameter
  RACK  => 1,            # Rack location from name (if formatted)
  SLOT  => 1,            # Slot location from name (if formatted)
  TYPE  => 'c',          # Type of brick from name (if formatted)
  DESC  => "",           # Brick description - NOT YET IMPLEMENTED
  NCHN  => 1,            # Number of chains on the module
  CHN_REFS => \@module_00000_chains, # Array of pointers to chain data
  SELCHN => 0,           # Currently selected scan chain
);


#-------------------------------------------------------------------------------

sub scancfg_force
{
   my $mod_ref = \%module_00000;
   $modules{ $$mod_ref{NAME} } = $mod_ref;
   push(@module_names, $$mod_ref{NAME});
   push(@module_ptrs,  $mod_ref);
}


#-------------------------------------------------------------------------------

sub scancfg_ce_connected_to_tio
{  
   print "\nscancfg_ce_connected_to_tio( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);
   return if ( $#_ != 3 );
   my $mref  = $_[0];  # Reference to module structure
   my $bname = $_[1];  # Name of board in the module
   my $cname = $_[2];  # Name of chip on the board
   my $ntype = ( $_[3] eq 'physical' ? $CMP_IND_PHY : $CMP_IND_LOG );
   return(-1) if ( ref($mref) ne 'HASH' );
   my ($chip_list, $chip_ptr, $i, $uc_bname, $uc_cname);

   #
   #  Convert names to upper case for comparison
   #
   $uc_bname = uc($bname);
   $uc_cname = uc($cname);

   #
   #  Search each chip on the module: match board and chip
   #
   $chn_list = $$mref{CHN_REFS}; # List of chain definition pointers
   for ($c = 0; $c <= $#$chn_list; $c++)
   {  $chn_ptr   = $$chn_list[$c];
      $chip_list = $$chn_ptr[$CHN_IND_CREFS];
      for ($i = 0; $i <= $#$chip_list; $i++)
      {  $chip_ptr = $$chip_list[$i];  # Pointer to chip definition
         if ( ($uc_bname eq uc($$chip_ptr[$CMP_IND_BOARD])) && ($uc_cname eq uc($$chip_ptr[$ntype])) )
         {  if ( $$chn_ptr[$CHN_IND_FLAG] & $FLAG_CE_CHIP_PRESENT )
            {  return($TRUE);
            }
            else
            {  return($FALSE);
            } 
         }
      }
   }
   return($FALSE);
}


#-------------------------------------------------------------------------------

sub scancfg_load
{
   print "\nscancfg_load( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $brk_types = "";
   my $part_str  = "";

   if ( $#_ == 3 )
   {  $brk_types = $_[2]; # List of valid brick types for discovery
      $part_str  = $_[3]; # List of valid partitions for discovery
   }
   elsif ( $#_ == 2 )
   {  $brk_types = $_[2]; # List of valid brick types for discovery
   }
   elsif ( $#_ != 1 )
   {  $emess = "Invalid parameters passed to subroutine scancfg_load_file().";
      &$comm_err_fun($emess);
      return;
   }
   my $cfg_name = $_[0];  # Name of configuration file
   my $tdr_flag = $_[1];  # Flag: TRUE=build JTAG TDR definition structures
   my ($list_ref, $mod_ind, $mod_ref);

   #
   #  Load the configuration file
   #
   $list_ref = pcf_scancfg_load($cfg_name, $brk_types, $part_str, $ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess);
      return;
   }
   print "list_ref = $list_ref = ( @$list_ref )\n" if ($debug_flag & $DEBUG_DETAILS);

   #
   #  Process each module in the configuration
   #
   foreach $mod_ind ( @$list_ref )
   {
      #  Build the module structure with chip definitions
      $mod_ref = scancfg_build_module($mod_ind);

      #  Build the module's JTAG TDR definitions
      if ( $tdr_flag )
      {  $emess = scan_module_init_jtdrs($mod_ref);
         if ( $emess ne "" )
         {  $emess = "Encountered error during build of JTAG TDRs.\n" . $emess;
            &$comm_err_fun($emess);
            return;
         }
      }

      #  Store the module information in the global structures
      $modules{ $$mod_ref{NAME} } = $mod_ref;
      push(@module_names, $$mod_ref{NAME});
      push(@module_ptrs,  $mod_ref);

      scan_module_dump($mod_ref, $FALSE) if ($debug_flag & $DEBUG_DETAILS);
   }


   return( ($#$list_ref + 1) );
}

#-------------------------------------------------------------------------------

sub scancfg_dump
{
   print "\nscancfg_dump( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $flag = (( $#_ >= 0 ) ? $_[0] : $FALSE );
   my ($name, $mref);

   print "module_names = ( @module_names )\n" if ($debug_flag & $DEBUG_DETAILS);
   print "module_ptrs  = ( @module_ptrs )\n" if ($debug_flag & $DEBUG_DETAILS);

   #
   #  Dump the contents of each module structure
   #
   foreach $name (@module_names)
   {
      $mref = $modules{$name};
      scan_module_dump($mref, $flag);
   }

}


#-------------------------------------------------------------------------------

sub scancfg_check_for_chip_type
{
   print "\nscancfg_check_for_chip_type( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   return($FALSE) if ( $#_ < 0 );
   my $type = $_[0];
   my ($mod_ptr, $c, $i, $chn_list, $chn_ptr, $chip_list, $chip_ptr, $curr_type);

   #
   #  Process each module in the configuration
   # 
   for $mod_ptr ( @module_ptrs )
   {  $chn_list = $$mod_ptr{CHN_REFS}; # List of chain definition pointers
      for ($c = 0; $c <= $#$chn_list; $c++)
      {  $chn_ptr = $$chn_list[$c];
         $chip_list = $$chn_ptr[$CHN_IND_CREFS];
         for ($i = 0; $i <= $#$chip_list; $i++)
         {  $chip_ptr  = $$chip_list[$i];  # Pointer to chip definition
            $curr_type = $$chip_ptr[$CMP_IND_DESC];
            if ( $curr_type =~ /^$type/i )
            {   return($TRUE);
            }
         }
      }

   }
   return($FALSE);
}


#-------------------------------------------------------------------------------

sub scancfg_build_module
{  
   print "\nscancfg_build_module( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $#_ != 0 )
   {  $emess = "Invalid parameters passed to subroutine scancfg_build_module().";
      &$comm_err_fun($emess);
      return;
   }
   my $mod  = $_[0];  # Module index in configuration
   my %mod_data = (); # Module definition structure (HASH)
   my @data = ();     # Initialize data array
   my ($chn, $chn_ref, $chip, $chip_ref, $tmp, $val);

   #
   #  Extract the module details
   #
   $aref = pcf_scancfg_get_module($mod, $ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess);
      return;
   }

   print "module: aref = $aref = ( @$aref )\n" if ($debug_flag & $DEBUG_GENERAL);

   #
   #  Initialize the module definition structure
   #
   $mod_data{NAME}   = $$aref[0];
   $mod_data{INDEX}  = $mod;
   $mod_data{LOC}    = $$aref[1];
   $mod_data{MODE}   = $$aref[2];
   $mod_data{FLAG}   = 0;
   $mod_data{NCHN}   = $$aref[3];
   $mod_data{SELCHN} = 0;

   #
   #  Extract additional parameters if valid brick name format
   #
   if ( length($mod_data{NAME}) == 6 )
   {  $tmp = substr($mod_data{NAME}, 0, 3);
      $val = $tmp + 0;
      $val = ((($val < 1) || ($val > 999)) ? 0 : $val);
      $mod_data{RACK} = $val;
      $tmp = substr($mod_data{NAME}, 4, 2);
      $val = $tmp + 0;
      $val = ((($val < 1) || ($val > 99)) ? 0 : $val);
      $mod_data{SLOT} = $val;
      $mod_data{TYPE} = substr($mod_data{NAME}, 3, 1);
      $mod_data{DESC} = "";
   }
   else
   {  $mod_data{RACK} = 0;
      $mod_data{SLOT} = 0;
      $mod_data{TYPE} = ' ';
      $mod_data{DESC} = "";
   }

   #
   #  Build the chain definition structure
   #
   my @chn_refs = ();
   for ($chn = 0; $chn < $mod_data{NCHN}; $chn++)
   {
      $chn_ref = pcf_scancfg_get_chain($mod, $chn, $ecode, $emess);
      if ( $ecode < 0 )
      {  &$comm_err_fun($emess);
         return;
      }
      $chn_refs[$chn] = $chn_ref;

      #
      #  Build the chip definition structures
      #
      my @chip_refs = ();
      for ($chip = 0; $chip < $$chn_ref[$CHN_IND_NCHIP]; $chip++)
      {
         $chip_ref = pcf_scancfg_get_chip($mod, $chn, $chip, $ecode, $emess);
         if ( $ecode < 0 )
         {  &$comm_err_fun($emess);
            return;
         }
         $chip_refs[$chip] = $chip_ref;
         print "chip_ref = $chip_ref = ( @$chip_ref )\n" if ($debug_flag & $DEBUG_DETAILS);
      }
      $$chn_ref[$CHN_IND_CREFS] = \@chip_refs;

      print "chain_ref = $chn_ref = ( @$chn_ref )\n" if ($debug_flag & $DEBUG_DETAILS);
   }
   $mod_data{CHN_REFS} = \@chn_refs;

   return( \%mod_data ); 
}


#-------------------------------------------------------------------------------

sub scancfg_hardware_init
{
   my $mod_index = 0xFFFFFFFF;
   my $chn_ind   = 0;
   if ( $#_ >= 0 )
   {  if ( !exists $modules{$_[0]} )
      {  $emess = "Invalid module passed to subroutine scancfg_hardware_init.";
         $emess .= "\nModule: $_[0]";
         &$comm_err_fun($emess);
         return;
      }
      $mref = $modules{$_[0]};
      $mod_index = $$mref{INDEX};
      if ( $#_ > 0 )
      {  $$mref{SELCHN} = $_[1];
      }
      $chn_ind   = $$mref{SELCHN};
   }

   #  Call the C-function that does the "real" work
   pcf_scancfg_hardware_init($mod_index, $chn_ind, $ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess);
      return;
   }

# TEMP - Need to set the jtag target here ???

}


#-------------------------------------------------------------------------------

sub scan_module_is_valid
{  return(0) if ( $#_ != 0 );
   my $mod_ptr = $_[0];
   return(0) if ( ref($mod_ptr) ne 'HASH' );
   return(0) if ( !(exists $$mod_ptr{NAME}) );
   return(0) if ( !(exists $$mod_ptr{NCHN}) );
   my $chn_list = $$mod_ptr{CHN_REFS};  # List of chip definition pointers
   return(0) if ( ref($chn_list) ne 'ARRAY' );
   return(1);
}


#-------------------------------------------------------------------------------

sub scan_module_find
{  return if ( $#_ != 0 );

   my $mname = lc($_[0]);
   foreach $name ( keys %modules )
   {  if ( $mname eq lc($name) )
      {  return( $modules{$name} );
      }
   }
   return;
}

#-------------------------------------------------------------------------------

sub scan_module_chips
{  
   print "\nscan_module_chips( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);
   my @data  = ();
   if ( $#_ < 2 )
   {  return(@data);
   }
   my $mod_ptr = $_[0];  # Pointer to module definition structure
   my $mask    = $_[1];  # Mask of flags that must match
   my $ntype   = ( $_[2] eq 'physical' ? $CMP_IND_PHY : $CMP_IND_LOG );

   return(@data) if ( !(exists $$mod_ptr{CHN_REFS}) );
   my ($chn_list, $chn_ptr, $chip_list, $chip_ptr, $name, $c, $i);

   #
   #  Move through each chip on each chain for module
   #
   $chn_list = $$mod_ptr{CHN_REFS}; # List of chain definition pointers
   for ($c = 0; $c <= $#$chn_list; $c++)
   {  $chn_ptr   = $$chn_list[$c];
      $chip_list = $$chn_ptr[$CHN_IND_CREFS];
      for ($i = 0; $i <= $#$chip_list; $i++)
      {  $chip_ptr = $$chip_list[$i];  # Pointer to chip definition
         if ( ($mask == 0) || 
              ($mask && ($$chip_ptr[$CMP_IND_FLAG] & $mask)) )
         {  $name = sprintf("%s:%s:%s", $$mod_ptr{NAME}, 
                       $$chip_ptr[$CMP_IND_BOARD], $$chip_ptr[$ntype]);
            push (@data, $name);       # Add chip name to list (ARRAY)
         }
      }
   }
   return(@data);
}


#-------------------------------------------------------------------------------

sub scan_module_find_chip
{  
   print "\nscan_module_find_chip( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   return(-1) if ( $#_ != 3 );
   my $mref  = $_[0];  # Reference to module structure
   my $bname = $_[1];  # Name of board in the module
   my $cname = $_[2];  # Name of chip on the board
   my $ntype = ( $_[3] eq 'physical' ? $CMP_IND_PHY : $CMP_IND_LOG );
   return(-1) if ( ref($mref) ne 'HASH' );
   my ($chip_list, $chip_ptr, $c, $i, $uc_bname, $uc_cname);

   #
   #  Convert names to upper case for comparison
   #
   $uc_bname = uc($bname);
   $uc_cname = uc($cname);

   #
   #  Check each chip on the module
   #
   $chn_list = $$mref{CHN_REFS}; # List of chain definition pointers
   for ($c = 0; $c <= $#$chn_list; $c++)
   {  $chn_ptr   = $$chn_list[$c];
      $chip_list = $$chn_ptr[$CHN_IND_CREFS];
      for ($i = 0; $i <= $#$chip_list; $i++)
      {  $chip_ptr = $$chip_list[$i];  # Pointer to chip definition
         if ( ($uc_bname eq uc($$chip_ptr[$CMP_IND_BOARD])) && ($uc_cname eq uc($$chip_ptr[$ntype])) )
         {  return($c, $i);
         }
      }
   }

   return(-1, -1);
}


#-------------------------------------------------------------------------------

sub scan_module_chip_ptr
{  
   print "\nscan_module_chip_ptr( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   return if ( $#_ != 3 );
   my $mref  = $_[0];  # Reference to module structure
   my $bname = $_[1];  # Name of board in the module
   my $cname = $_[2];  # Name of chip on the board
   my $ntype = ( $_[3] eq 'physical' ? $CMP_IND_PHY : $CMP_IND_LOG );
   return if ( ref($mref) ne 'HASH' );
   my ($chip_list, $chip_ptr, $i, $uc_bname, $uc_cname);

   #
   #  Convert names to upper case for comparison
   #
   $uc_bname = uc($bname);
   $uc_cname = uc($cname);

   #
   #  Search each chip on the module: match board and chip
   #
   $chn_list = $$mref{CHN_REFS}; # List of chain definition pointers
   for ($c = 0; $c <= $#$chn_list; $c++)
   {  $chn_ptr   = $$chn_list[$c];
      $chip_list = $$chn_ptr[$CHN_IND_CREFS];
      for ($i = 0; $i <= $#$chip_list; $i++)
      {  $chip_ptr = $$chip_list[$i];  # Pointer to chip definition
         return($chip_ptr) if ( ($uc_bname eq uc($$chip_ptr[$CMP_IND_BOARD])) &&
                         ($uc_cname eq uc($$chip_ptr[$ntype])) );
      }
   }
   return;
}


#-------------------------------------------------------------------------------

sub scan_module_chip_details
{
   print "\nscan_module_chip_details( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   return if ( $#_ != 3 );
   my $mref  = $_[0];  # Reference to module structure
   my $bname = $_[1];  # Name of board in the module
   my $cname = $_[2];  # Name of chip on the board
   my $ntype = $_[3];  # Name type: logical or physical
   my ($chip_list, $chip_ref);

   #
   #  Find the chip location on the module/board
   #
   ($chn_ind, $chip_ind) = scan_module_find_chip($mref, $bname, $cname, $ntype);
   if ( $chn_ind < 0 )
   {  $emess = "Unable to find specified chip in module configuration.";
      $emess .= "\nTarget: $$mref{NAME}:$bname:$cname";
      return;
   }
 
   #
   #  Extract the chain and chip pointers
   # 
   $chn_list  = $$mref{CHN_REFS}; # List of chain definition pointers
   $chn_ptr   = $$chn_list[$chn_ind];
   $chip_list = $$chn_ptr[$CHN_IND_CREFS];
   $chip_ptr  = $$chip_list[$chip_ind];

   #
   #  Build up array with details of chip location in scan chain
   #
   my @chip_details = ();
   $chip_details[$DETAIL_IND_TYPE] = $$chip_ptr[$CMP_IND_DESC]; # Chip type: "shub_01", "pic_01"
   $chip_details[$DETAIL_IND_IPOS] = $$chip_ptr[$CMP_IND_IPOS]; # Chip's IR starting position
   $chip_details[$DETAIL_IND_ILEN] = $$chn_ptr[$CHN_IND_IRLEN]; # Total IR length for scan chain
   $chip_details[$DETAIL_IND_CLEN] = $$chn_ptr[$CHN_IND_NCHIP]; # Number of chips in scan chain
   $chip_details[$DETAIL_IND_CPOS] = $chip_ind;                 # Chip position in scan chain
   $chip_details[$DETAIL_IND_CNUM] = $chn_ind;                  # Chain number/index on module
   $chip_details[$DETAIL_IND_MNUM] = $$mref{INDEX};             # Module number/module in configuration
   $chip_details[$DETAIL_IND_MLOC] = $$mref{LOC};               # Module/brick location ID
   $chip_details[$DETAIL_IND_NAME] = $$mref{NAME} . ":" . $bname . ":" . $cname; # Target name

   printf("chip details: ( @chip_details )\n") if ($debug_flag & $DEBUG_DETAILS);

   return(\@chip_details);
}


#-------------------------------------------------------------------------------

sub scan_module_dump
{  my $flag = 0;
   if ( $#_ == 1 )
   {  $flag = $_[1];
   }
   elsif ( $#_ != 0 )
   {  return;
   }
   my $mref = $_[0];
   return if ( ref($mref) ne 'HASH' );
   my $text = "";
   my ($i, $c, $chn_list, $chn_ptr, $chip_list, $chip_ptr, $list);

   printf "\n----- SCAN MODULE CONTENTS -----\n";
   printf "NAME   = %s\n", $$mref{NAME};
   printf "INDEX  = %s\n", $$mref{INDEX};
   printf "LOC    = %s\n", $$mref{LOC};
   printf "MODE   = 0x%02X = %u\n", $$mref{MODE}, $$mref{MODE};
   printf "FLAG   = 0x%02X = %u\n", $$mref{FLAG}, $$mref{FLAG};
   printf "RACK   = %03u\n", $$mref{RACK};
   printf "SLOT   = %02u\n", $$mref{SLOT};
   printf "TYPE   = '%s'\n", $$mref{TYPE};
   printf "DESC   = \"%s\"\n", $$mref{DESC};
   printf "SELCHN = %u\n", $$mref{SELCHN};
   printf "NCHN   = %u\n", $$mref{NCHN};

   #
   #  Process each scan chain on the module
   #
   $chn_list = $$mref{CHN_REFS}; # List of chain definition pointers
   for ($c = 0; $c <= $#$chn_list; $c++)
   {  $chn_ptr   = $$chn_list[$c];
      $chip_list = $$chn_ptr[$CHN_IND_CREFS];

      printf "-- CHAIN %u OF MODULE %s --\n", $c, $$mref{NAME};
      printf "MODE   = 0x%02X = %u\n", $$chn_ptr[$CHN_IND_MODE], $$chn_ptr[$CHN_IND_MODE];
      printf "FLAG   = 0x%02X = %u\n", $$chn_ptr[$CHN_IND_FLAG], $$chn_ptr[$CHN_IND_FLAG];
      printf "IRLEN  = %u\n", $$chn_ptr[$CHN_IND_IRLEN];
      printf "DRLEN  = %u\n", $$chn_ptr[$CHN_IND_DRLEN];
      printf "NCHIP  = %u\n", $$chn_ptr[$CHN_IND_NCHIP];

      for ($i = 0; $i <= $#$chip_list; $i++)
      {  $chip_ptr = $$chip_list[$i];  # Pointer to chip definition
         $list     = $$chip_ptr[$CMP_IND_LIST];
         $text     = $$chip_ptr[$CMP_IND_PHY] . "/" . $$chip_ptr[$CMP_IND_LOG]; 
         printf "CHIP[%02u] : %-5s %-12s %-12s %-2u %-2u 0x%04X %-2u $list\n", $i, 
           $$chip_ptr[$CMP_IND_BOARD], $text, $$chip_ptr[$CMP_IND_DESC],
           $$chip_ptr[$CMP_IND_ILEN],  $$chip_ptr[$CMP_IND_IPOS], $$chip_ptr[$CMP_IND_FLAG],
           $$chip_ptr[$CMP_IND_CURR];
         if ( $flag && (ref($list) eq 'ARRAY') )
         {  foreach $jref (@$list)
            {  jtag_tdr_dump($jref);
            }
            print "\n";
         }
      }
   }
}


#-------------------------------------------------------------------------------

sub scan_module_init_jtdrs
{  return("Invalid number of parameters specified.") if ( $#_ != 0 );

   print "\nscan_module_init_jtdrs( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $mod_ptr = $_[0];                      # Reference to module structure
   return("Invalid module reference parameter.") if ( ref($mod_ptr) ne 'HASH' );
   my ($chn_list, $chn_ptr, $c, $chip_ptr, $desc, $name, $emess, $i, $ref, $def_name);

   #
   #  Process each chip on each scan chain
   #
   $chn_list = $$mod_ptr{CHN_REFS}; # List of chain definition pointers
   for ($c = 0; $c <= $#$chn_list; $c++)
   {  $chn_ptr   = $$chn_list[$c];
      $chip_list = $$chn_ptr[$CHN_IND_CREFS];

      foreach $chip_ptr ( @$chip_list )
      {
         my @list = ();                         # Array for JTAG TDR pointers
         $def_name = $$chip_ptr[$CMP_IND_DESC]; # Chip description/type ("shub01")
         if ( $def_name =~ /^shub2/i )
         {  $desc = "shub2_01";
         }
         elsif ( $def_name =~ /^shub/i )
         {  $desc = "shub_01";
         }
         elsif ( $def_name =~ /^pic/i )
         {  $desc = "pic_01";
         }
         elsif ( $def_name =~ /^tio/i )
         {  $desc = "tio_01";
         }
         elsif ( $def_name =~ /^nl4r/i )
         {  $desc = "nl4rtr_01";
         }
         elsif ( $def_name =~ /^lbi/i )
         {  $desc = "lbi_01";
         }
         else
         {  $desc = $def_name;
         }
         $name = $desc . "_tdr_defined";        # Name of definition variable

         if ( !defined($$name) )                # ASIC type TDRs not defined
         {  $ref = jtag_tdr_create_bypass_inst($$chip_ptr[$CMP_IND_ILEN]);
            $i   = 0;
            $list[$i] = $ref;                
            $i++;
         } 
         else # TDR definitions for the ASIC type are included
         {
            #  Process each valid OPCODE name for the chip description
            $name = $desc . "_opnames";         # Array of IR operation names
            for ($i = 0; $i <= $#$name; $i++)
            {
               #  Create the JTAG TDR structure for this instance
               $ref = jtag_tdr_create_inst($desc, $$name[$i]);
               if ( !jtag_tdr_is_valid($ref) )
               {  $emess  = "Unable to extract defintion for JTAG OPCODE.";
                  $emess .= "\nChip description: $desc";
                  $emess .= "\nJTAG OPCODE name: $$name[$i]";
                  return($emess);
               }
   
               #  Store the reference to the JTAG TDR structure in the list
               $list[$i] = $ref;                
            }
         }
   
         #
         #  Store the array of pointers in the chip definition and set current
         #
         if ( $i > 0 )
         {  $$chip_ptr[$CMP_IND_LIST] = \@list;
            $$chip_ptr[$CMP_IND_CURR] = 0;
         }
      }
   }

   return("");
}

#-------------------------------------------------------------------------------

sub scan_module_target
{
   printf "\nscan_module_target( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);
   my $mref    = $_[0];  # Reference to module structure
   my $chn_ind = ( ($#_ > 0) ? $_[1] : $$mref{SELCHN} );
   my $flag    = ( ($#_ > 1) ? $_[2] : $FALSE );

   #
   #  Save the currently selected chain
   #
   $$mref{SELCHN} = $chn_ind;

   #
   #  Target the specific module
   #
   jtag_target($$mref{LOC}, $$mref{INDEX}, $chn_ind, $flag);
}


#-------------------------------------------------------------------------------

sub scan_module_set_chip_opcode
{
   printf "\nscan_module_set_chip( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   return if ( $#_ != 4 );
   my $mref   = $_[0];  # Reference to module structure
   my $bname  = $_[1];  # Name of board in the module
   my $cname  = $_[2];  # Name of chip on the board
   my $ntype  = $_[3];  # Name type: logical or physical
   my $opname = $_[4];  # Name of instruction opcode to select
   my ($ind, $chip_ref, $jtdr_list, $jtdr_ref, $href);

   #  Get the description structure for the specified chip
   $chip_ref = scan_module_chip_ptr($mref, $bname, $cname, $ntype);
   if ( !defined($chip_ref) )
   {  $emess = "Unable to find specified chip in module configuration.";
      $emess .= "\nTarget: $$mref{NAME}:$bname:$cname";
      return;
   }
#printf "\nchip_ref  = $chip_ref = ( @$chip_ref )\n";

   #  Search thru the valid JTAG TDRs for the specified chip
   $jtdr_list = $$chip_ref[$CMP_IND_LIST]; # List of JTAG TDRs for the chip
#printf "\njtdr_list = $jtdr_list = ( @$jtdr_list )\n";
   for ($ind = 0; $ind <= $#$jtdr_list; $ind++)
   {  $jtdr_ref = $$jtdr_list[$ind];         # Structure for this JTAG TDR
#printf "\n%02u : jtdr_ref = $jtdr_ref = ( @$jtdr_ref )\n", $ind;
      $href     = $$jtdr_ref[$TDR_REF_DEF];  # Pointer to definition (HASH)
#printf "     OPNAME=%s\n", $$href{OPNAME};
      if ( $$href{OPNAME} eq $opname )
      {  $$chip_ref[$CMP_IND_CURR] = $ind;   # Set current JTAG TDR index
         return($jtdr_ref);                  # Return this JTAG TDR reference
      }
   }

   #  The desired JTAG TDR was not found for this chip
   $emess  = "Unable to find JTAG OPNAME for specified chip.";
   $emess .= "\nTarget: $$mref{NAME}:$bname:$cname";
   $emess .= "\nOPNAME: $opname";
   return;
}


#-------------------------------------------------------------------------------

sub scan_module_ir_len
{  
   print "\nscan_module_ir_len( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $mref = $_[0];              # Reference to scan module definition (HASH)
   if ( $#_ > 0 )
   {  $$mref{SELCHN} = $_[1];
   }
   my $chn_ind   = $$mref{SELCHN};
   my $chn_list  = $$mref{CHN_REFS}; # List of chain definition pointers
   my $chn_ptr   = $$chn_list[$chn_ind];
   return($$chn_ptr[$CHN_IND_IRLEN]);
}

#-------------------------------------------------------------------------------

sub scan_module_dr_len
{  
   print "\nscan_module_dr_len( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $mref = $_[0];              # Reference to scan module definition (HASH)
   if ( $#_ > 0 )
   {  $$mref{SELCHN} = $_[1];
   }
   my $chn_ind   = $$mref{SELCHN};
   my $chn_list  = $$mref{CHN_REFS}; # List of chain definition pointers
   my $chn_ptr   = $$chn_list[$chn_ind];
   return($$chn_ptr[$CHN_IND_DRLEN]);
}


#-------------------------------------------------------------------------------

sub scan_module_bypass_all
{  
   print "\nscan_module_bypass_all( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $mref = $_[0];              # Reference to scan module definition (HASH)
   if ( $#_ > 0 )
   {  $$mref{SELCHN} = $_[1];
   }
   my $chn_ind   = $$mref{SELCHN};
   my $chn_list  = $$mref{CHN_REFS}; # List of chain definition pointers
   my $chn_ptr   = $$chn_list[$chn_ind];
   my $chip_list = $$chn_ptr[$CHN_IND_CREFS];

   my ($chip_ref, $jtdr_list, $i, $jtdr_ref, $href, $found);
   foreach $chip_ref (@$chip_list)
   {  $jtdr_list = $$chip_ref[$CMP_IND_LIST];   # List of all JTAG TDR pointers
      for ($i = 0, $found = $FALSE; ( ($i <= $#$jtdr_list) && !$found ); $i++)
      {  $jtdr_ref = $$jtdr_list[$i];
         $href     = $$jtdr_ref[$TDR_REF_DEF];  # Pointer to definition (HASH)
         if ( $$href{OPNAME} eq 'BYPASS' )
         {  $$chip_ref[$CMP_IND_CURR] = $i;     # Current JTAG TDR pointer
            $found = $TRUE;
         }
      }
   }
}

#-------------------------------------------------------------------------------

sub scan_module_build_ir
{  
   print "\nscan_module_build_ir( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $mref = $_[0];              # Reference to scan module definition structure (HASH)
   if ( $#_ > 0 )
   {  $$mref{SELCHN} = $_[1];
   }
   my $chn_ind   = $$mref{SELCHN};
   my $chn_list  = $$mref{CHN_REFS}; # List of chain definition pointers
   my $chn_ptr   = $$chn_list[$chn_ind];
   my $chip_list = $$chn_ptr[$CHN_IND_CREFS];
   my @idata     = ();               # Array to hold chip's IR opcode
   my ($vref, $chip_ref, $jlist, $jtdr_ref, $pos, $len, $href);

   #  Initialize IR buffer and counters
   $vref = vec_create($$chn_ptr[$CHN_IND_IRLEN]);
   $pos = 0;
   $len = 0;

   #
   #  Process each chip in the scan chain
   #
   foreach $chip_ref (@$chip_list)
   {  
#print "\nprocessing chip: pos=$pos, len=$len, chip_ref=$chip_ref=(@$chip_ref)\n";
      $jlist    = $$chip_ref[$CMP_IND_LIST];            # List of all JTAG TDR pointers
      $jtdr_ref = $$jlist[ $$chip_ref[$CMP_IND_CURR] ]; # Current JTAG TDR pointer (ARRAY)
      $href     = $$jtdr_ref[ $TDR_REF_DEF ];           # Pointer to definition (HASH)
      $idata[0] = $$href{OPCODE};                       # Instruction OPCODE for current
#printf "opcode=0x%08lX\n", $idata[0];

      # Copy instruction OPCODE into vector
      vec_cpy_bits(0, \@idata, $pos, $vref, $$chip_ref[$CMP_IND_ILEN], 'right');

      $pos += $$chip_ref[$CMP_IND_ILEN];                # Update IR position
      $len += $$href{LENGTH};                           # Update DR length
   }

#print "build_ir done: pos=$pos, len=$len\n";
#vec_dump_hex("IR TDI DATA", $vref);

   $$chn_ptr[$CHN_IND_DRLEN] = $len;   # Save current DR chain length
   return($vref);          # Return array with IR TDI data
}


#-------------------------------------------------------------------------------

sub scan_module_build_dr
{  return(0) if ( $#_ != 0 );

   print "\nscan_module_build_dr( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $mref = $_[0];              # Reference to scan module definition structure (HASH)
   if ( $#_ > 0 )
   {  $$mref{SELCHN} = $_[1];
   }
   my $chn_ind   = $$mref{SELCHN};
   my $chn_list  = $$mref{CHN_REFS}; # List of chain definition pointers
   my $chn_ptr   = $$chn_list[$chn_ind];
   my $chip_list = $$chn_ptr[$CHN_IND_CREFS];
   my ($vref, $chip_ref, $jlist, $jtdr_ref, $pos, $len, $href);

   #  Initialize DR buffer and counters
   $vref = vec_create($$chn_ptr[$CHN_IND_DRLEN]);
   $pos = 0;

   #
   #  Process each chip in the scan chain
   #
   foreach $chip_ref (@$chip_list)
   {  
#print "\nprocessing chip: pos=$pos, chip_ref=$chip_ref=(@$chip_ref)\n";
      $jlist    = $$chip_ref[$CMP_IND_LIST];            # List of all JTAG TDR pointers
      $jtdr_ref = $$jlist[ $$chip_ref[$CMP_IND_CURR] ]; # Current JTAG TDR pointer (ARRAY)
      $href     = $$jtdr_ref[ $TDR_REF_DEF ];           # Pointer to definition (HASH)
      $dref     = $$jtdr_ref[ $TDR_REF_VAL ];           # Pointer to data (ARRAY)
#printf "dref=$dref=(@$dref)\n";

      # Copy chip TDR data into vector
      vec_cpy_bits(0, $dref, $pos, $vref, $$href{LENGTH}, 'right');

      $pos += $$href{LENGTH};                           # Update DR chain length
   }

#print "build_dr done: pos=$pos\n";
#vec_dump_hex("DR TDI DATA", $vref);

   if ( $pos != $$chn_ptr[$CHN_IND_DRLEN] )
   {  $$chn_ptr[$CHN_IND_DRLEN] = $len;  # Save current DR chain length
      # Need to set status for new IR data?
   }
   return($vref);            # Return array with DR TDI data
}


#-------------------------------------------------------------------------------

sub scan_module_parse_dr
{  return if ( $#_ != 1 );

   print "\nscan_module_parse_dr( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $mref = $_[0];              # Reference to scan module definition structure (HASH)
   my $vref = $_[1];              # Reference to vector data to parse
   if ( $#_ > 1 )
   {  $$mref{SELCHN} = $_[2];
   }
   my $chn_ind   = $$mref{SELCHN};
   my $chn_list  = $$mref{CHN_REFS}; # List of chain definition pointers
   my $chn_ptr   = $$chn_list[$chn_ind];
   my $chip_list = $$chn_ptr[$CHN_IND_CREFS];
   my ($chip_ref, $jlist, $jtdr_ref, $pos, $len, $href);
#vec_dump_hex("DR TDO DATA", $vref);

   #
   #  Process each chip in the scan chain
   #
   $pos = 0;
   foreach $chip_ref (@$chip_list)
   {  
#print "\nprocessing chip: pos=$pos, chip_ref=$chip_ref=(@$chip_ref)\n";
      $jlist    = $$chip_ref[$CMP_IND_LIST];            # List of all JTAG TDR pointers
      $jtdr_ref = $$jlist[ $$chip_ref[$CMP_IND_CURR] ]; # Current JTAG TDR pointer (ARRAY)
      $href     = $$jtdr_ref[ $TDR_REF_DEF ];           # Pointer to definition (HASH)
      $dref     = $$jtdr_ref[ $TDR_REF_VAL ];           # Pointer to data (ARRAY)
#printf "dref=$dref=(@$dref)  [0]=0x%08lX\n", $$dref[0];

      # Extract chip TDR data from vector
      vec_cpy_bits($pos, $vref, 0, $dref, $$href{LENGTH}, 'right');
#printf "dref=$dref=(@$dref)  [0]=0x%08lX\n", $$dref[0];

      $pos += $$href{LENGTH};                           # Update DR chain length
   }

#print "parse_dr done: pos=$pos\n";
   return;
}


#-------------------------------------------------------------------------------

sub scan_chip_get_jtdr_ptr
{  
   print "\nscan_chip_get_jtdr_ptr( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);
   my ($cref, $ind);

   if ( $#_ == 1 )
   {  $cref = $_[0];  # Reference to chip definition structure (ARRAY)
      return(0) if ( ref($cref) ne 'ARRAY' );
      $ind  = $_[1];  # Index of JTAG TDR pointer in the list
   }
   elsif ( $#_ == 0 )
   {  $cref = $_[0];  # Reference to chip definition structure (ARRAY)
      return(0) if ( ref($cref) ne 'ARRAY' );
      $ind  = $$cref[$CMP_IND_CURR]; # Index of JTAG TDR pointer in the list
   }
   else
   {  return(0);
   }
   my $aref = $$cref[$CMP_IND_LIST];
   return(0) if ( ($ind < 0) || ($ind > $#$aref) );
   return( $$aref[$ind] );
}


#-------------------------------------------------------------------------------

sub scan_chip_bypass_index
{  return(-1) if ( $#_ != 0 );

   print "\nscan_chip_bypass_index( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $chip_ref = $_[0];  # Reference to chip definition structure (ARRAY)
   my ($jtdr_list, $i, $jtdr_ref, $href); 
   $jtdr_list = $$chip_ref[$CMP_IND_LIST];           # List of all JTAG TDR pointers
   for ($i = 0; $i <= $#$jtdr_list; $i++)            # Process each JTAG TDR
   {  $jtdr_ref = $$jtdr_list[$i];                   # Reference of this JTAG TDR
      $href     = $$jtdr_ref[ $TDR_REF_DEF ];        # Pointer to definition (HASH)
      return($i) if ( $$href{OPNAME} eq 'BYPASS' );
   }
   return(-1);
}


#-------------------------------------------------------------------------------
#  NOTE: Not sure if the following two subroutines should live in this
#        file or not.  The use the scan configuration information, but require
#        "jtag_subs".
#-------------------------------------------------------------------------------

sub scan_module_write_ir
{
   printf "\nscan_module_write_ir( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $mref  = $_[0];  # Reference to module structure
   my $bname = $_[1];  # Board name on module/brick (i.e. "NODE" or "MOATB")
   my $cname = $_[2];  # Logical chip name on board (i.e. "SHUB0" or "TIO0")
   my $oname = $_[3];  # JTAG OPCODE/TDR name (i.e. "CONFIG" or "LLP4_NI0")
   my ($jtdr_ref, $chn_ind, $chip_ind, $ilen, $idata, @odata);

   #---------------------------------------
   #  Find the specified chip on the module
   #---------------------------------------
   ($chn_ind, $chip_ind) = scan_module_find_chip($mref, $bname, $cname, 'logical');
   if ( $chn_ind < 0 )
   {  $emess = "Unable to find specified chip in module configuration.";
      $emess .= "\nTarget: $$mref{NAME}:$bname:$cname";
      &$comm_err_fun($emess);
      return;
   }

   #---------------------------------------
   #  Target the specified module and select BYPASS
   #---------------------------------------
   scan_module_target($mref, $chn_ind);
   scan_module_bypass_all($mref);     # Select BYPASS opcode for all chips

   #---------------------------------------
   #  Get pointer for specified OPCODE/TDR
   #---------------------------------------
   $jtdr_ref = scan_module_set_chip_opcode($mref, $bname, $cname, 'logical', $oname);
   if ( !defined($jtdr_ref) )         # Was JTAG OPCODE/TDR supported on this chip?
   {  &$comm_err_fun($emess);
      return($FALSE);
   }

   #---------------------------------------
   #  Build the IR TDI data and load into the hardware
   #---------------------------------------
   $idata = scan_module_build_ir($mref);
   $ilen  = scan_module_ir_len($mref);
   @odata = jtag_scan_ir($ilen, $idata);
   if ( $debug_flag & $DEBUG_GENERAL )
   {  print "calling: jtag_scan_ir($ilen, $idata)\n";
      vec_dump_hex("IR TDI DATA", $idata);
      vec_dump_hex("IR TDO DATA", \@odata);
   }

   return($TRUE);
}


#-------------------------------------------------------------------------------

sub scan_module_write_tdr
{
   printf "\nscan_module_write_tdr( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $mref  = $_[0];  # Reference to module structure
   my $bname = $_[1];  # Board name on module/brick (i.e. "NODE" or "MOATB")
   my $cname = $_[2];  # Logical chip name on board (i.e. "SHUB0" or "TIO0")
   my $oname = $_[3];  # JTAG OPCODE/TDR name (i.e. "CONFIG" or "LLP4_NI0")
   my $bgnd  = $_[4];  # Background value for JTAG TDR: $JTDR_BGND_USE_*
   my $href  = ( ($#_ > 4) ? $_[5] : 0 );  # Reference to HASH structure with override values
   my $ir_load  = ( ($#_ > 5) ? $_[6] : $TRUE );
   my ($jtdr_ref, $idata, @odata, @fields, $field, $val, $dref, @data);
   my ($chn_ind, $chip_ind, $ilen, $dlen);

   #---------------------------------------
   #  Find the specified chip on the module
   #---------------------------------------
   ($chn_ind, $chip_ind) = scan_module_find_chip($mref, $bname, $cname, 'logical');
   if ( $chn_ind < 0 )
   {  $emess = "Unable to find specified chip in module configuration.";
      $emess .= "\nTarget: $$mref{NAME}:$bname:$cname";
      &$comm_err_fun($emess);
      return;
   }

   #---------------------------------------
   #  Target the specified module and select BYPASS
   #---------------------------------------
   scan_module_target($mref, $chn_ind);
   scan_module_bypass_all($mref);     # Select BYPASS opcode for all chips

   #---------------------------------------
   #  Get pointer for specified OPCODE/TDR
   #---------------------------------------
   $jtdr_ref = scan_module_set_chip_opcode($mref, $bname, $cname, 'logical', $oname);
   if ( !defined($jtdr_ref) )         # Was JTAG OPCODE/TDR supported on this chip?
   {  &$comm_err_fun($emess);
      return($FALSE);
   }

   #---------------------------------------
   #  Verify field names (if specified)
   #---------------------------------------
   @fields = ();
   if ( ref($href) eq 'HASH' )
   {  @fields = keys(%$href);
      foreach $field ( @fields )
      {  if ( jtag_tdr_is_valid_field($jtdr_ref, $field) == $FALSE )
         {  $emess = "Invalid field name provided for JTAG TDR data structure.";
            $emess .= "\nTarget chip: $$mref{NAME}:$bname:$cname";
            $emess .= "\nOPNAME/TDR:  $oname";
            $emess .= "\nField name:  $field";
            $ecode = -1;
            &$comm_err_fun($emess);
            return($FALSE);
         }
      }
   }

   #---------------------------------------
   #  Build the IR TDI data and load into the hardware
   #---------------------------------------
   if ( $ir_load )
   {  $idata = scan_module_build_ir($mref);
      $ilen  = scan_module_ir_len($mref);
      @odata = jtag_scan_ir($ilen, $idata);
      if ( $debug_flag & $DEBUG_GENERAL )
      {  print "calling: jtag_scan_ir($ilen, $idata)\n";
         vec_dump_hex("IR TDI DATA", $idata);
         vec_dump_hex("IR TDO DATA", \@odata);
      }
   }

   #---------------------------------------
   #  Read the actual TDR value from the hardware
   #---------------------------------------
   if ( $bgnd == $JTDR_BGND_USE_HARDWARE )
   {  jtag_tdr_reset($jtdr_ref);
      $idata = scan_module_build_dr($mref);
      $dlen  = scan_module_dr_len($mref);
      @odata = jtag_scan_dr($dlen, $idata, $JTAG_STATE_PDR);
      if ( $debug_flag & $DEBUG_DETAILS )
      {
printf "\nJTAG_TDR: Reading from hardware - $bname:$cname:$oname\n";
         print "calling: jtag_scan_dr($dlen, $idata, JTAG_STATE_PDR)\n";
         vec_dump_hex("DR TDI DATA", $idata);
         vec_dump_hex("DR TDO DATA", \@odata);
      }
      scan_module_parse_dr($mref, \@odata);
   }
   elsif ( $bgnd == $JTDR_BGND_USE_DEFAULT )
   {  jtag_tdr_reset($jtdr_ref);
      if ( $debug_flag & $DEBUG_DETAILS )
      {
printf "\nJTAG_TDR: Resetting TDR structure to default values\n";
      }
   }

   if ( ($#fields > -1) && ($debug_flag & $DEBUG_DETAILS) )
   {  printf "\nJTAG_TDR - Unmodified contents - $bname:$cname:$oname\n";
      jtag_tdr_dump_contents($jtdr_ref);
   }

   #---------------------------------------
   #  Change the specified fields in the TDR
   #---------------------------------------
   foreach $field ( @fields )
   {  $val  = $$href{$field};
      if ( ref($val) eq 'ARRAY' )
      {  $dref = $val;
      }
      else
      {  @data = ( $val );
         $dref = \@data;
      }
      jtag_tdr_set_field_value($jtdr_ref, $field, $dref);
   }

   if ( $debug_flag & $DEBUG_GENERAL )
   {  printf "\nJTAG_TDR: Writing to hardware - $bname:$cname:$oname\n";
      jtag_tdr_dump_contents($jtdr_ref);
   }

   #---------------------------------------
   #  Write the TDR structure to the hardware
   #---------------------------------------
   $idata = scan_module_build_dr($mref);
   $dlen  = scan_module_dr_len($mref);
   @odata = jtag_scan_dr($dlen, $idata);
   if ( $debug_flag & $DEBUG_GENERAL )
   {  print "calling: jtag_scan_dr($dlen, $idata)\n";
      vec_dump_hex("DR TDI DATA", $idata);
      vec_dump_hex("DR TDO DATA", \@odata);
   }

   return($TRUE);
}


#-------------------------------------------------------------------------------


1;
