#-------------------------------------------------------------------------------
#
#  FILE: shub2_subs.pm
#
#  DESCRIPTION:  This file contains PERL constants and subroutines that are
#     specific to the SN2 SHUB2 ASIC.
#
#  MODIFICATION HISTORY:
#
#     WHEN      WHO             WHAT
#     --------  --------------  ----------------------------------------------
#     05/11/04  Karen Beighley  Original source code (ported from SHUB).
#     07/19/04  Karen Beighley  Added MMRs and subroutine for packet generator.
#
#-------------------------------------------------------------------------------

$USE_SHUB2_SUBS_PM = 1;    # Define to allow test for module inclusion


#-------------------------------------------------------------------------------
#  DEPENDENCIES
#-------------------------------------------------------------------------------

if ( !defined($USE_JTAG_SUBS_PM) )
{  $emess = "Required module has not been included in PERL script.";
   $emess .= "\nThe \"use jtag_subs;\" statement must appear before using \"shub2_subs\".";
   dsp_error_and_exit($emess);
   exit -1;
}


#-------------------------------------------------------------------------------
#  GLOBAL VARIABLES
#-------------------------------------------------------------------------------

$SHUB2_IR_LEN = 7;

$SHUB2_MD_CHIPLET = 0;  # chiplet value for address <27:25> of PIO_CTRL field
$SHUB2_LB_CHIPLET = 1;  # chiplet value for address <27:25> of PIO_CTRL field
$SHUB2_PI_CHIPLET = 2;  # chiplet value for address <27:25> of PIO_CTRL field
$SHUB2_XN_CHIPLET = 5;  # chiplet value for address <27:25> of PIO_CTRL field


$CFG_MODE_NONE           = 0x00;
$CFG_MODE_RESET_ICNTLR   = 0x01;
$CFG_MODE_INIT_DLY_CTRL1 = 0x02;

#-------------------------------------------------------------------------------
#  PACKET GENERATOR SPECIFICS - The order of the address array cannot change!
#-------------------------------------------------------------------------------

@shub2_pktgen_addr =
( 0x0000000010060f00, # SH2_NODE_ID (NODE_ID field @ bit 32)
  0x0000000010063000, # SH2_DIAG_MSG_CNTRL
  0x0000000010063980, # SH2_DIAG_MSG_HDR0
  0x0000000010063a00, # SH2_DIAG_MSG_HDR1
  0x0000000010063080, # SH2_DIAG_MSG_DATA0L
  0x0000000010063100, # SH2_DIAG_MSG_DATA0U
  0x0000000010063180, # SH2_DIAG_MSG_DATA1L
  0x0000000010063200, # SH2_DIAG_MSG_DATA1U
  0x0000000010063280, # SH2_DIAG_MSG_DATA2L
  0x0000000010063300, # SH2_DIAG_MSG_DATA2U
  0x0000000010063380, # SH2_DIAG_MSG_DATA3L
  0x0000000010063400, # SH2_DIAG_MSG_DATA3U
  0x0000000010063480, # SH2_DIAG_MSG_DATA4L
  0x0000000010063500, # SH2_DIAG_MSG_DATA4U
  0x0000000010063580, # SH2_DIAG_MSG_DATA5L
  0x0000000010063600, # SH2_DIAG_MSG_DATA5U
  0x0000000010063680, # SH2_DIAG_MSG_DATA6L
  0x0000000010063700, # SH2_DIAG_MSG_DATA6U
  0x0000000010063780, # SH2_DIAG_MSG_DATA7L
  0x0000000010063800, # SH2_DIAG_MSG_DATA7U
  0x0000000010063880, # SH2_DIAG_MSG_DATA8L
  0x0000000010063900  # SH2_DIAG_MSG_DATA8U
);

#-------------------------------------------------------------------------------
#  ERROR REGISTER SPECIFICS - The order of the address array cannot change!
#-------------------------------------------------------------------------------

@shub2_ereg_addr =
(  0x0000000010050700, # SH2_LB_FIRST_ERROR
   0x0000000010050280, # SH2_LB_ERROR_DETAIL_1
   0x0000000010050300, # SH2_LB_ERROR_DETAIL_2
   0x0000000010050380, # SH2_LB_ERROR_DETAIL_3
   0x0000000010050400, # SH2_LB_ERROR_DETAIL_4
   0x0000000010050480, # SH2_LB_ERROR_DETAIL_5
   0x0000000010050680, # SH2_LB_ERROR_SUMMARY
   0x0000000010050600, # SH2_LB_ERROR_OVERFLOW
   0x0000000010050580, # SH2_LB_ERROR_MASK
   0x0000000010050688, # SH2_LB_ERROR_SUMMARY_ALIAS
   0x0000000010050608, # SH2_LB_ERROR_OVERFLOW_ALIAS
   0x0000000010050500, # SH2_LB_ERROR_DETAIL_6
);

#-------------------------------------------------------------------------------
#  INTERRUPT ENABLE REGISTER  - The order of the address array cannot change!
#-------------------------------------------------------------------------------

@shub2_int_enable_addr =
(  0x0000000010000500, # SH2_LOCAL_INT0_ENABLE
   0x0000000010000600, # SH2_LOCAL_INT1_ENABLE
   0x0000000010000700, # SH2_LOCAL_INT2_ENABLE
   0x0000000010000800, # SH2_LOCAL_INT3_ENABLE
   0x0000000010000900, # SH2_LOCAL_INT4_ENABLE
   0x0000000010000a00  # SH2_LOCAL_INT5_ENABLE
);



################################################################################
################################################################################

sub shub2_reset_sequence
{
   printf "\nshub2_reset_sequence( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $mref     = $_[0];  # Reference to module structure
   my $bname    = $_[1];  # Board name on module/brick (i.e. "NODE" or "MOATB")
   my $cname    = $_[2];  # Logical chip name on board (i.e. "SHUB0" or "TIO0")
   my $mode     = $_[3];  # Configuration mode/control flags
   my $tb_name  = (( $#_ > 3 ) ? $_[4] : "shub2_jtag" );
   my $mname    = $$mref{NAME};

   my $tp_data      = ( ($#_ > 0) ? $_[1] : 0 );
   my $lbi_hard_rst_reg = "$tb_name.lbi_hard_reset_reg";
   my $hard_rst_reg = "$tb_name.hard_reset_reg";
   my $soft_rst_reg = "$tb_name.soft_reset_reg";
   my $clk_rst_reg  = "$tb_name.clk_reset_reg";

   print "\n";
   print "------------------------------------------------------------------\n";
   print "--  SHUB2 RESET SEQUENCE FOR $mname:$bname:$cname\n";
   print "------------------------------------------------------------------\n";

   $time_stamp = jtag_simacc_time();
   printf "\n------------------------------------------------------------\n";
   printf "-> SIMULATOR TIMESTAMP : %llu\n", $time_stamp;
   printf "------------------------------------------------------------\n"; 
   print "\n->Asserting the HARD RESET signal on the simulator...\n  entity=$hard_rst_reg\n";
   jtag_simacc_set($JTAG_SIMACC_VEC, $hard_rst_reg, "1'b0");

   print "\n->Asserting the SOFT RESET signal on the simulator...\n  entity=$soft_rst_reg\n";
   jtag_simacc_set($JTAG_SIMACC_VEC, $soft_rst_reg, "1'b0");

   print "\n->Asserting the CLK RESET signal on the simulator...\n  entity=$clk_rst_reg\n";
   jtag_simacc_set($JTAG_SIMACC_VEC, $clk_rst_reg, "1'b0");


   if ( $prompt_flag )
   {  printf "\npress RETURN to continue...\n";
      $text = <STDIN>;
   }
   $time_stamp = jtag_simacc_time();
   printf "\n------------------------------------------------------------\n";
   printf "-> SIMULATOR TIMESTAMP : %llu\n", $time_stamp;
   printf "------------------------------------------------------------\n";

   print "\n->Resetting the TAP controller via TMS/TCK (SOFT reset)...\n";
   jtag_reset($JTAG_RESET_SOFT, $JTAG_STATE_RTI);

   my $oname = "IDCODE";
   my $mode  = $JTDR_BGND_USE_HARDWARE;
   printf "\n-> Reading DEVICE_IDENTIFICATION TDR\n";
   scan_module_write_tdr($mref, $bname, $cname, $oname, $mode);


   $time_stamp = jtag_simacc_time();
   printf "\n------------------------------------------------------------\n";
   printf "-> SIMULATOR TIMESTAMP : %llu\n", $time_stamp;
   printf "------------------------------------------------------------\n";


   printf "\n-> Loading IR: TDR_DEFAULT\n";
   scan_module_write_ir($mref, $bname, $cname, 'TDR_DEFAULT');

   if ( $prompt_flag )
   {  printf "\npress RETURN to continue...\n";
      $text = <STDIN>;
   }

   $time_stamp = jtag_simacc_time();
   printf "\n------------------------------------------------------------\n";
   printf "-> SIMULATOR TIMESTAMP : %llu\n", $time_stamp;
   printf "------------------------------------------------------------\n";


   shub2_chip_load_pre_config($mref, $bname, $cname);

   if ( $prompt_flag )
   {  printf "\npress RETURN to continue...\n";
      $text = <STDIN>;
   }


   $time_stamp = jtag_simacc_time();
   printf "\n------------------------------------------------------------\n";
   printf "-> SIMULATOR TIMESTAMP : %llu\n", $time_stamp;
   printf "------------------------------------------------------------\n";

   shub2_chip_load_clkcntl($mref, $bname, $cname, $tp_data);

   if ( $prompt_flag )
   {  printf "\npress RETURN to continue...\n";
      $text = <STDIN>;
   }


   $time_stamp = jtag_simacc_time();
   printf "\n------------------------------------------------------------\n";
   printf "-> SIMULATOR TIMESTAMP : %llu\n", $time_stamp;
   printf "------------------------------------------------------------\n";

   printf "\n-> Loading IR: JCLK_RST\n";
   scan_module_write_ir($mref, $bname, $cname, 'JCLK_RST');

   if ( $prompt_flag )
   {  printf "\npress RETURN to continue...\n";
      $text = <STDIN>;
   }

   printf "\n-> Loading IR: BYPASS to clear JCLK_RST instruction\n";
   scan_module_write_ir($mref, $bname, $cname, 'BYPASS');

   if ( $prompt_flag )
   {  printf "\npress RETURN to continue...\n";
      $text = <STDIN>;
   }


   print "\n->Deasserting the CLK RESET signal on the simulator...\n  entity=$clk_rst_reg\n";
   jtag_simacc_set($JTAG_SIMACC_VEC, $clk_rst_reg, "1'b1");

   if ( $prompt_flag )
   {  printf "\npress RETURN to continue...\n";
      $text = <STDIN>;
   }


   $time_stamp = jtag_simacc_time();
   printf "\n------------------------------------------------------------\n";
   printf "-> SIMULATOR TIMESTAMP : %llu\n", $time_stamp;
   printf "------------------------------------------------------------\n";


   shub2_chip_load_misc_config($mref, $bname, $cname);

   if ( $prompt_flag )
   {  printf "\npress RETURN to continue...\n";
      $text = <STDIN>;
   }


   shub2_chip_load_icntlr($mref, $bname, $cname, ($mode & $CFG_MODE_RESET_ICNTLR));

   if ( $prompt_flag )
   {  printf "\npress RETURN to continue...\n";
      $text = <STDIN>;
   }


   $time_stamp = jtag_simacc_time();
   printf "\n------------------------------------------------------------\n";
   printf "-> SIMULATOR TIMESTAMP : %llu\n", $time_stamp;
   printf "------------------------------------------------------------\n";


   shub2_chip_load_llp_config($mref, $bname, $cname, $mode);

   if ( $prompt_flag )
   {  printf "\npress RETURN to continue...\n";
      $text = <STDIN>;
   }


   $time_stamp = jtag_simacc_time();
   printf "\n------------------------------------------------------------\n";
   printf "-> SIMULATOR TIMESTAMP : %llu\n", $time_stamp;
   printf "------------------------------------------------------------\n";

   print "\n->Desserting the SOFT RESET signal on the simulator...\n  entity=$soft_rst_reg\n";
   jtag_simacc_set($JTAG_SIMACC_VEC, $soft_rst_reg, "1'b1");

   print "\n->Deasserting the HARD RESET signal on the simulator...\n  entity=$hard_rst_reg\n";
   jtag_simacc_set($JTAG_SIMACC_VEC, $hard_rst_reg, "1'b1");


   # Set the ENABLE_LINK bit for NI0 and NI1
   shub2_set_link_enable($mref, $bname, $cname, 0, $TRUE);
   shub2_set_link_enable($mref, $bname, $cname, 1, $TRUE);

   $time_stamp = jtag_simacc_time();
   printf "\n------------------------------------------------------------\n";
   printf "-> SIMULATOR TIMESTAMP : %llu\n", $time_stamp;
   printf "------------------------------------------------------------\n";
}


#-------------------------------------------------------------------------------

sub shub2_chip_load_pre_config
{
   printf "\nshub2_chip_load_pre_config( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $mref    = $_[0];  # Reference to module structure
   my $bname   = $_[1];  # Board name on module/brick (i.e. "NODE" or "MOATB")
   my $cname   = $_[2];  # Logical chip name on board (i.e. "SHUB0" or "TIO0")
   my ($oname, $mode, %hdata);

   $oname = "CONFIG";
   $mode  = $JTDR_BGND_USE_DEFAULT;
   %hdata = ();
   $hdata{ LB_DBUG_CONFIG_OE       } = 1;
   $hdata{ FSB_DEBUG_CONFIG_OE     } = 1;
   $hdata{ DQSGATE_EN              } = 1;
   $hdata{ FMAC_TXRX_CAPTURE_PHASE } = 1;
   $hdata{ JTAG_MCI_OVERRIDE       } = 1;
   $hdata{ JTAG_MCI_TARGET         } = 0x0021;
   $hdata{ JTAG_MCI_RESET_DELAY    } = 0x2A;
   $hdata{ JTAG_MCI_WRTACT_LATENCY } = 0;
   $hdata{ J_SEL_FSB_1X_BY2        } = 1;
   $hdata{ J_SEL_OSCFSB_CORE       } = 1;
   $hdata{ II_CLK_SEL              } = 1;
   printf "\n-> Loading DR: $oname\n";
   scan_module_write_tdr($mref, $bname, $cname, $oname, $mode, \%hdata);

   $oname = "FSB_POWER_ON_CFG";
   $mode  = $JTDR_BGND_USE_DEFAULT;
   %hdata = ();
   printf "\n-> Loading DR: $oname\n";
   scan_module_write_tdr($mref, $bname, $cname, $oname, $mode, \%hdata);

   $oname = "PROCESSOR_TYPE";
   $mode  = $JTDR_BGND_USE_DEFAULT;
   %hdata = ();
   printf "\n-> Loading DR: $oname\n";
   scan_module_write_tdr($mref, $bname, $cname, $oname, $mode, \%hdata);

   $oname = "JNR_CONFIG";
   $mode  = $JTDR_BGND_USE_DEFAULT;
   %hdata = ();
   printf "\n-> Loading DR: $oname\n";
   scan_module_write_tdr($mref, $bname, $cname, $oname, $mode, \%hdata);
}

#-------------------------------------------------------------------------------

sub shub2_chip_load_misc_config
{
   printf "\nshub2_chip_load_misc_config( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $mref  = $_[0];  # Reference to module structure
   my $bname = $_[1];  # Board name on module/brick (i.e. "NODE" or "MOATB")
   my $cname = $_[2];  # Logical chip name on board (i.e. "SHUB0" or "TIO0")
   my $rack  = $$mref{RACK};
   my $slot  = $$mref{SLOT};
   my $slab  = 0;
   my ($oname, $mode, $data, %hdata);

   $oname = "ADDRESS_MODE";
   $mode  = $JTDR_BGND_USE_DEFAULT;
   %hdata = ();
   $hdata{ CODE  } = 3;
   printf "\n-> Loading DR: $oname\n";
   scan_module_write_tdr($mref, $bname, $cname, $oname, $mode, \%hdata);

   #  Build data for RACK_SLOT_DATA TDR
   $data = (($rack << 26) | (($slot & 0x3f) << 20) | (($slab & 0xf) << 16) |
            ((0 & 0xff) << 8) | 0x6c);
   my @tmp = ( (($data >> 32) & 0xFFFFFFFF), ($data & 0xFFFFFFFF) );

   $oname = "RACK_SLOT_DATA";
   $mode  = $JTDR_BGND_USE_DEFAULT;
   %hdata = ();
   $hdata{ DATA  } = \@tmp;
   printf "\n-> Loading DR: $oname\n";
   scan_module_write_tdr($mref, $bname, $cname, $oname, $mode, \%hdata);
}


#-------------------------------------------------------------------------------

sub shub2_chip_load_icntlr
{
   printf "\nshub2_chip_load_icntlr( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $mref  = $_[0];  # Reference to module structure
   my $bname = $_[1];  # Board name on module/brick (i.e. "NODE" or "MOATB")
   my $cname = $_[2];  # Logical chip name on board (i.e. "SHUB0" or "TIO0")
   my $flag  = $_[3];  # Flag: TRUE=toggle RESET for each port

   my $oname = "ICNTLR";
   my $mode  = $JTDR_BGND_USE_DEFAULT;
   my %hdata = ();

   if ( $flag )
   {  $hdata{ NI0_RESET } = 1;
      $hdata{ NI1_RESET } = 1;
      printf "\n-> Loading DR: $oname with NIx_RESET asserted\n";
      scan_module_write_tdr($mref, $bname, $cname, $oname, $mode, \%hdata);
   }
 
   $hdata{ NI0_RESET } = 0;
   $hdata{ NI1_RESET } = 0;
   printf "\n-> Loading DR: $oname with NIx_RESET deasserted\n";
   scan_module_write_tdr($mref, $bname, $cname, $oname, $mode, \%hdata);
}


#-------------------------------------------------------------------------------

sub shub2_chip_load_clkcntl
{
   printf "\nshub2_chip_load_clkcntl( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);
   my $mref  = $_[0];  # Reference to module structure
   my $bname = $_[1];  # Board name on module/brick (i.e. "NODE" or "MOATB")
   my $cname = $_[2];  # Logical chip name on board (i.e. "SHUB0" or "TIO0")
   my $tpval = $_[3];  # Test point monitor value (for JSEL_TP)

   my $oname = "CLKCNTL";
   my $mode  = $JTDR_BGND_USE_DEFAULT;
   my %hdata = ();
   $hdata{ TP } = $tpval; 
   printf "\n-> Loading DR: $oname\n";
   scan_module_write_tdr($mref, $bname, $cname, $oname, $mode, \%hdata);
}


#-------------------------------------------------------------------------------

sub shub2_chip_load_llp_config
{
   printf "\nshub2_chip_load_llp_config( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $mref     = $_[0];  # Reference to module structure
   my $bname    = $_[1];  # Board name on module/brick (i.e. "NODE" or "MOATB")
   my $cname    = $_[2];  # Logical chip name on board (i.e. "SHUB0" or "TIO0")
   my $mode     = $_[3];  # Configuration mode/control flags
   my ($oname, $mode, %hdata, @data);

   #----------------------------------------
   #  Configure NI0 LLP
   #----------------------------------------
   $oname = "LLP4_NI0";
   $mode  = $JTDR_BGND_USE_DEFAULT;
   %hdata = ();
   $hdata{ BIT_MODE32         } = 1;
   $hdata{ ENABLE_AC          } = 0;
   $hdata{ ENABLE_TUNING      } = 0; # 1?
   $hdata{ ENABLE_QUAD        } = 0;
   $hdata{ ENABLE_LINK        } = 0;
   printf "\n-> Loading DR: $oname\n";
   scan_module_write_tdr($mref, $bname, $cname, $oname, $mode, \%hdata);

   #----------------------------------------
   #  Configure NI1 LLP
   #----------------------------------------
   $oname = "LLP4_NI1";
   $mode  = $JTDR_BGND_USE_DEFAULT;
   %hdata = ();
   $hdata{ BIT_MODE32         } = 1;
   $hdata{ ENABLE_AC          } = 0;
   $hdata{ ENABLE_TUNING      } = 0; # 1?
   $hdata{ ENABLE_QUAD        } = 0;
   $hdata{ ENABLE_LINK        } = 0;
   printf "\n-> Loading DR: $oname\n";
   scan_module_write_tdr($mref, $bname, $cname, $oname, $mode, \%hdata);

   #----------------------------------------
   #  Configure NI0 LLP_TEST : enable/disable test NULLs
   #----------------------------------------
   $oname = "LLP4TEST_NI0";
   $mode  = $JTDR_BGND_USE_DEFAULT;
   %hdata = ();
#   @data  = ( 0x00, 0x0A5FFFFF );
   @data  = ( 0x00, 0x0F012345 );
   $hdata{ TEST_PATTERN   }  = \@data;
   $hdata{ SEND_TEST_MODE }  = 0;
   $hdata{ NOISE_MODE }      = 0; # 2?
   $hdata{ LFSR_MODE }       = 0;
   printf "\n-> Loading DR: $oname\n";
   scan_module_write_tdr($mref, $bname, $cname, $oname, $mode, \%hdata);

   #----------------------------------------
   #  Configure NI1 LLP_TEST : enable/disable test NULLs
   #----------------------------------------
   $oname = "LLP4TEST_NI1";
   $mode  = $JTDR_BGND_USE_DEFAULT;
   %hdata = ();
#   @data  = ( 0x00, 0x0A5FFFFF );
   @data  = ( 0x00, 0x05A22222 );
   $hdata{ TEST_PATTERN   }  = \@data;
   $hdata{ SEND_TEST_MODE }  = 0;
   $hdata{ NOISE_MODE }      = 0; # 2?
   $hdata{ LFSR_MODE }       = 0;
   printf "\n-> Loading DR: $oname\n";
   scan_module_write_tdr($mref, $bname, $cname, $oname, $mode, \%hdata);

   #----------------------------------------
   #  Configure NI0 DELAY CONTROLLER for NL4
   #----------------------------------------
   $oname = "DLY_CTRL0_NI0";
   $mode  = $JTDR_BGND_USE_DEFAULT;
   %hdata = ();
   $hdata{ LLP_TUNING_CMPLT_EN } = 1;
   $hdata{ CLK_ADAPT_HOLD_MODE } = 0;
   $hdata{ SAMP_TIM_SEL        } = 0;
   printf "\n-> Loading DR: $oname\n";
   scan_module_write_tdr($mref, $bname, $cname, $oname, $mode, \%hdata);

   #----------------------------------------
   #  Configure NI1 DELAY CONTROLLER for NL4
   #----------------------------------------
   $oname = "DLY_CTRL0_NI1";
   $mode  = $JTDR_BGND_USE_DEFAULT;
   %hdata = ();
   $hdata{ LLP_TUNING_CMPLT_EN } = 1;
   $hdata{ CLK_ADAPT_HOLD_MODE } = 0;
   $hdata{ SAMP_TIM_SEL        } = 0;
   printf "\n-> Loading DR: $oname\n";
   scan_module_write_tdr($mref, $bname, $cname, $oname, $mode, \%hdata);

   #----------------------------------------
   #  Configure DELAY CONTROLLER 1 ?
   #----------------------------------------
   if ( $mode & $CFG_MODE_INIT_DLY_CTRL1 )
   {  $oname = "DLY_CTRL1_NI0";
      $mode  = $JTDR_BGND_USE_DEFAULT;
      %hdata = ();
      printf "\n-> Loading DR: $oname\n";
      scan_module_write_tdr($mref, $bname, $cname, $oname, $mode, \%hdata);

      $oname = "DLY_CTRL1_NI1";
      $mode  = $JTDR_BGND_USE_DEFAULT;
      %hdata = ();
      printf "\n-> Loading DR: $oname\n";
      scan_module_write_tdr($mref, $bname, $cname, $oname, $mode, \%hdata);
   }

   #----------------------------------------
   #  Load the WARM_FENCE TDR
   #----------------------------------------
   $oname = "WARM_FENCE";
   $mode  = $JTDR_BGND_USE_DEFAULT;
   %hdata = ();
   $hdata{ VALUE } = $shub2_fence_val;
   printf "\n-> Loading DR: $oname\n";
   scan_module_write_tdr($mref, $bname, $cname, $oname, $mode, \%hdata);
}


#-------------------------------------------------------------------------------

sub shub2_set_link_enable
{
   printf "\nshub2_set_link_enable( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $mref     = $_[0];  # Reference to module structure
   my $bname    = $_[1];  # Board name on module/brick (i.e. "NODE" or "MOATB")
   my $cname    = $_[2];  # Logical chip name on board (i.e. "SHUB0" or "TIO0")
   my $port     = $_[3];  # Port index: 0=NI0, 1=NI1
   my $value    = $_[4];  # Port enable value: 0=disable, 1=enable

   #----------------------------------------
   #  Configure NI1 LLP
   #----------------------------------------
   $oname = "LLP4_NI" . $port;
   $mode  = $JTDR_BGND_USE_CURRENT;
   %hdata = ();
   $hdata{ ENABLE_LINK        } = $value;
   printf "\n-> Loading DR: $oname\n";
   scan_module_write_tdr($mref, $bname, $cname, $oname, $mode, \%hdata);
}


#------------------------------------------------------------------------------
#
#    SUBROUTINE: shub2_pktgen_get_ids()
#
#    USAGE: @vals = shub2_pktgen_get_ids($op, $addr)
#           @vals = shub2_pktgen_get_ids($op, $addr, $src_id)
#           @vals = shub2_pktgen_get_ids($op, $addr, $src_id, $dst_id)
#           @vals = shub2_pktgen_get_ids($op, $addr, $src_id, $dst_id, $dst_cid)
#
#    DESCRIPTION:  This subroutine will determine the source and
#       destination IDs required for a packet generator request header.  The
#       source node ID, source chiplet ID, destination node ID and destination
#       chiplet ID are returned.  If no parameters are given, the chip's 
#       NODE_ID MMR is read, and the NODE_ID is extracted.  If the destination
#       node ID is not provided, then the source ID is used.  If the destination
#       chiplet ID is not provided, then it is determined from the operation
#       type ('mem' or 'mmr') and address.
#
#    RETURN VALUE: Array containing source and destination IDs.
#
#------------------------------------------------------------------------------

sub shub2_pktgen_get_ids
{
   my $num = $#_ + 1;

   printf "\nshub2_pktgen_get_ids(@_) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( ($num < 2) || ($num > 5) )
   {  $emess = "Invalid parameters passed to subroutine shub2_pktgen_get_ids().";
      $ecode = -1;
      return;
   }
   my $op   = $_[0];
   my $addr = $_[1];
   my ($tmp, $src_id, $src_chiplet, $dst_id, $dst_chiplet);

   #--------------------------------------- 
   #  Determine the NODE IDs
   #--------------------------------------- 
   if ( $num == 2 ) # No source or destination: read local ID
   {  $tmp = mmr_read($shub2_pktgen_addr[$PKTGEN_IND_ID]);
      return if ( $ecode < 0 );
      $src_id = ($tmp >> 32) & 0x7FF;
      $dst_id = $src_id;
   }
   elsif ( $num == 3 ) # Source ID provided: use for destination also
   {  $src_id = $_[2];
      $dst_id = $src_id;
   }
   else # Both source and destination ID provided
   {  $src_id = $_[2];
      $dst_id = $_[3];
   }

   #--------------------------------------- 
   #  Determine the CHIPLET IDs
   #--------------------------------------- 
   $src_chiplet = $SHUB2_LB_CHIPLET;
   if ( $num == 5 )
   {  $dst_chiplet = $_[4];
   }
   elsif ( $op eq 'mem' )
   {  $dst_chiplet = $SHUB2_MD_CHIPLET;
   }
   else
   {  $dst_chiplet = $SHUB2_LB_CHIPLET;
   }
   return($src_id, $src_chiplet, $dst_id, $dst_chiplet);
}


#-------------------------------------------------------------------------------

1;

