#
#/**************************************************************************
# *                                                                        *
# *               Copyright (C) 1989-1999 Silicon Graphics, Inc.           *
# *                                                                        *
# *  These coded instructions, statements, and computer programs  contain  *
# *  unpublished  proprietary  information of Silicon Graphics, Inc., and  *
# *  are protected by Federal copyright law.  They  may  not be disclosed  *
# *  to  third  parties  or copied or duplicated in any form, in whole or  *
# *  in part, without the prior written consent of Silicon Graphics, Inc.  *
# *                                                                        *
# **************************************************************************/
#
# @(#)$Id: smdk_subs.pm,v 1.2 2002/11/18 20:30:18 beighley Exp $


#-------------------------------------------------------------------------------
#
#  FILE: smdk_subs.pm
#
#  DESCRIPTION:  This file contains PERL subroutines to boot and interact
#     with the diagnostic kernel "sMDK".
#
#  NOTES: The "common_defs.pm" file, "sysco_sub.pm" file and "SysAccess"
#     module are required with this file.
#
#  SUBROUTINES:  The following subroutines are contained in this file:
#     (expected parameter usage displayed; some have optional parameters)
#     $text  = smdk_boot($command)
#     $text  = smdk_run($command)
#     $text  = smdk_cmd($command)
#     $text  = smdk_cmd($command, $flush)
#     $text  = smdk_dscr_read($diag_num, $scrn_name, $page_option)
#     %dhash = smdk_dscr_get_entries($text)
#     %dhash = smdk_dscr_get_entries($text, \@labels)
#     $val   = smdk_dscr_get_val($text, $label)
#     %dhash = smdk_dscr_read_get_entries($diag_num, $scrn_name, $page_option)
#     %dhash = smdk_dscr_read_get_entries($diag_num, $scrn_name, $page_option, \@labels)
#     $val   = smdk_dscr_read_get_val($diag_num, $scrn_name, $page_option, $label)
#
#
#  MODIFICATION HISTORY:
#
#     WHEN      WHO             WHAT
#     --------  --------------  -----------------------------------------
#     10/29/01  Karen Beighley  Original source code.
#     10/25/02  Karen Beighley  Added use of command error handler.
#     11/18/02  Karen Beighley  Added "flush" parameter: smdk_cmd().
#
#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
#
#    SUBROUTINE: smdk_boot()
#
#    USAGE: $text = smdk_boot($command)
#
#    DESCRIPTION: Issue a command to boot the diagnostic kernel sMDK.
#       The function will wait for the boot to complete, and confirm that
#       the "smdk>" prompt was reached.  If the system is at the "smdk>"
#       prompt, a "reset" will be issued and sMDK will then be booted.
#
#    RETURN VALUE: A string containing the response text from the command.
#
#-------------------------------------------------------------------------------

sub smdk_boot
{ 
   my $num = $#_ + 1;
   print "\nsmdk_boot(@_) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   $ecode = 0;
   $emess = "";

   if ( $num != 1 )
   {  $emess = "Invalid number of parameters passed to subroutine smdk_boot().";
      &$comm_err_fun($emess);
      return;
   }
   my $cmd_text = $_[0];

   my $rv = pcf_smdk_boot($cmd_text, $ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess, $rv);
   }
   return($rv);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: smdk_run()
#
#    USAGE: $text = smdk_run($command)
#
#    DESCRIPTION: Issue a diagnostic run to the diagnostic kernel sMDK.
#       The function will check the sMDK response for a load error.
#
#    RETURN VALUE: A string containing the response text from the command.
#
#-------------------------------------------------------------------------------

sub smdk_run
{ 
   my $num = $#_ + 1;
   print "\nsmdk_run(@_) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   $ecode = 0;
   $emess = "";

   if ( $num != 1 )
   {  $emess = "Invalid number of parameters passed to subroutine smdk_run().";
      &$comm_err_fun($emess);
      return;
   }
   my $cmd_text = $_[0];

   my $rv = pcf_smdk_run($cmd_text, $ecode, $emess);
   &$comm_err_fun($emess, $rv)   if ( $ecode < 0 );
   return($rv);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: smdk_cmd()
#
#    USAGE: $text = smdk_cmd($command)
#           $text = smdk_cmd($command, $flush)
#
#    DESCRIPTION: Issue a user-specified sMDK command.  The command will
#       be issued.  The optional "flush" parameter determines if
#       the response socket is cleared before the command is issued.  The
#       possible values for "flush" are $TRUE or $FALSE, and the default
#       is $TRUE (if not specified).  If $TRUE is specified for "flush", then
#       any waiting response data will be discarded before the comand is issued.
#
#    RETURN VALUE: A string containing the response text from the command.
#
#-------------------------------------------------------------------------------

sub smdk_cmd
{ 
   my $num = $#_ + 1;
   print "\nsmdk_cmd(@_) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   $ecode = 0;
   $emess = "";

   my $fflag = $TRUE;
   if ( $num == 2 )
   {  $fflag = $_[1];
   }
   elsif ( $num != 1 )
   {  $emess = "Invalid number of parameters passed to subroutine smdk_cmd().";
      &$comm_err_fun($emess);
      return;
   }
   my $cmd_text = $_[0];

   my $rv = pcf_smdk_cmd($cmd_text, $fflag, $ecode, $emess);
   &$comm_err_fun($emess, $rv)   if ( $ecode < 0 );
   return($rv);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: smdk_dscr_read()
#
#    USAGE: $text = smdk_dscr_read($diag_num, $scrn_name, $page_option)
#
#    DESCRIPTION: Issue a command to read a diagnostic screen from sMDK.
#       The "page_option" parameter is a string that contains the page number
#       and any additional options for the command (i.e. "5", "5 option").
#
#    RETURN VALUE: A string containing the response text from the command.
#
#-------------------------------------------------------------------------------

sub smdk_dscr_read
{
   my $num = $#_ + 1;
   print "\nsmdk_dscr_read(@_) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   $ecode = 0;
   $emess = "";

   if ( $num != 3 )
   {  $emess = "Invalid number of parameters passed to subroutine smdk_dscr_read().";
      &$comm_err_fun($emess);
      return;
   }
   my $cmd_text = "dscr $_[0] $_[1] $_[2]\n";

   my $rv = pcf_smdk_dscr_read($cmd_text, $ecode, $emess);
   &$comm_err_fun($emess, $rv)   if ( $ecode < 0 );
   return($rv);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: smdk_dscr_get_entries()
#
#    USAGE: %dhash = smdk_dscr_get_entries($text)
#           %dhash = smdk_dscr_get_entries($text, \@labels)
#
#    DESCRIPTION: Parse the text of a standard diagnostic screen and store
#       the label/value pairs in a HASH structure.  If the "labels" parameter
#       is given, only entries for those labels found in the text are
#       stored.  If the "labels" parameter is not given, then the HASH table
#       contains entries for all the labels in the text.
#
#    RETURN VALUE: A hash table containing label and value pairs.
#
#-------------------------------------------------------------------------------

sub smdk_dscr_get_entries
{
   my $num = $#_ + 1;
   print "\nsmdk_dscr_get_entries(@_) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   $ecode = 0;
   $emess = "";

   my $av_ptr = 0;
   if ( $num == 2 )
   {  $av_ptr = $_[1];
      if ( ref($av_ptr) ne 'ARRAY' )
      {  $emess = "Invalid parameter passed to subroutine smdk_dscr_get_entries().";
         $emess .= "\nSecond parameter must be an ARRAY reference.";
         &$comm_err_fun($emess);
         return;
      }
   }
   elsif ( $num != 1 )
   {  $emess = "Invalid number of parameters for subroutine smdk_dscr_get_entries().";
      &$comm_err_fun($emess);
      return;
   }
   my $scr_text = $_[0];

   my $rv = pcf_smdk_dscr_get_entries($scr_text, $av_ptr, $ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess);
      return;
   }
   return(%$rv);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: smdk_dscr_get_val()
#
#    USAGE: $val = smdk_dscr_get_val($text, $label)
#
#    DESCRIPTION: Parse the text of a standard diagnostic screen and store
#       the label/value pairs in a HASH structure.  If the "labels" parameter
#       is given, only entries for those labels found in the text are
#       stored.  If the "labels" parameter is not given, then the HASH table
#       contains entries for all the labels in the text.
#
#    RETURN VALUE: A scalar containing the value corresponding to the label.
#       If the label was not found in the text, an empty string is returned.
#
#-------------------------------------------------------------------------------

sub smdk_dscr_get_val
{
   my $num = $#_ + 1;
   print "\nsmdk_dscr_get_val(@_) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   $ecode = 0;
   $emess = "";

   if ( $num != 2 )
   {  $emess = "Invalid number of parameters for subroutine smdk_dscr_get_val().";
      &$comm_err_fun($emess);
      return;
   }
   my $scr_text = $_[0];
   my $label    = $_[1];

   my $rv = pcf_smdk_dscr_get_val($scr_text, $label, $ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess);
      return;
   }
   return($rv);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: smdk_dscr_read_get_entries()
#
#    USAGE: %dhash = smdk_dscr_read_get_entries($diag_num, $scrn_name, $page_option)
#           %dhash = smdk_dscr_read_get_entries($diag_num, $scrn_name, $page_option, \@labels)
#
#    DESCRIPTION:  Combination of subroutines "smdk_dscr_read()" and 
#       "smdk_dscr_get_entries()".  The screen is read and then parsed.
#
#    RETURN VALUE: A hash table containing label and value pairs.
#
#-------------------------------------------------------------------------------

sub smdk_dscr_read_get_entries
{
   my $num = $#_ + 1;
   print "\nsmdk_dscr_read_get_entries(@_) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   $ecode = 0;
   $emess = "";

   my $av_ptr = 0;
   if ( $num == 4 )
   {  $av_ptr = $_[3];
      if ( ref($av_ptr) ne 'ARRAY' )
      {  $emess = "Invalid parameter passed to subroutine smdk_dscr_read_get_entries().";
         $emess .= "\nLabels parameter must be an ARRAY reference.";
         &$comm_err_fun($emess);
         return;
      }
   }
   elsif ( $num != 3 )
   {  $emess = "Invalid number of parameters for subroutine smdk_dscr_read_get_entries().";
      &$comm_err_fun($emess);
      return;
   }
   my $cmd_text = "dscr $_[0] $_[1] $_[2]\n";

   my $rv = pcf_smdk_dscr_read_get_entries($cmd_text, $av_ptr, $ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess);
      return;
   }
   return(%$rv);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: smdk_dscr_read_get_val()
#
#    USAGE: $val = smdk_dscr_read_get_val($diag_num, $scrn_name, $page_option, $label)
#
#    DESCRIPTION:  Combination of subroutines "smdk_dscr_read()" and 
#       "smdk_dscr_get_val()".  The screen is read and then parsed.
#
#    RETURN VALUE: A scalar containing the value corresponding to the label.
#       If the label was not found in the text, an empty string is returned.
#
#-------------------------------------------------------------------------------

sub smdk_dscr_read_get_val
{
   my $num = $#_ + 1;
   print "\nsmdk_dscr_read_get_val(@_) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   $ecode = 0;
   $emess = "";

   if ( $num != 4 )
   {  $emess = "Invalid number of parameters for subroutine smdk_dscr_read_get_val().";
      &$comm_err_fun($emess);
      return;
   }
   my $cmd_text = "dscr $_[0] $_[1] $_[2]\n";
   my $label    = $_[3];

   my $rv = pcf_smdk_dscr_read_get_val($cmd_text, $label, $ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess);
      return;
   }
   return($rv);
}

#-------------------------------------------------------------------------------

1;
