#
#/**************************************************************************
# *                                                                        *
# *               Copyright (C) 2002-2005 Silicon Graphics, Inc.           *
# *                                                                        *
# *  These coded instructions, statements, and computer programs  contain  *
# *  unpublished  proprietary  information of Silicon Graphics, Inc., and  *
# *  are protected by Federal copyright law.  They  may  not be disclosed  *
# *  to  third  parties  or copied or duplicated in any form, in whole or  *
# *  in part, without the prior written consent of Silicon Graphics, Inc.  *
# *                                                                        *
# **************************************************************************/
#
# @(#)$Id: sys_ctrl_funcs.pm,v 1.33 2007/03/20 20:00:39 meyer Exp $
#
#  FILE: sys_ctrl_funcs.pm
#
#  DESCRIPTION:  This file contains PERL routines to extract system infomation
#                using system control infrastructure.
#
#  MODIFICATION HISTORY:
#
#     WHEN      WHO             WHAT
#     --------  --------------  -----------------------------------------
#     06/06/05  Gary Meyer      Original source code submitted for release.
#     06/07/05  Gary Meyer	Added shub version specific L1 commands in get_system_env().
#     06/14/05  Gary Meyer	More Changes for Shub2.
#     09/16/05  Gary Meyer	Updated check_leds() for 191010 chassis cpu numbering.
#     09/23/05  Gary Meyer	Updated parse_scan_config for 191010.
#     10/03/05  Gary Meyer	Fixed parsing problems in check_leds() and check_cpu_string().
#     10/24/05  Gary Meyer	Fixed another parsing problem in check_leds() for blade systems.
#     02/07/06  Gary Meyer	Updated to fix problem with IO Topology parsing when the "-ii"
#				argument is supplied to errdmp.  Added "cpu fsb" and "blade" commands.
#				Now parsing the "serial all" output to determine the blade and processor type.
#     02/21/06  Gary Meyer	Updates for FRU summary.  Also for changes in serial all output when
#			  	parsing cpu types.
#     03/21/06  Gary Meyer	Update to correct bug seen with IP71 blade type.
#     04/25/06  Gary Meyer	Updates to ensure correct cpu config with IP71 & IP75 blade types.
#     05/05/06  Gary Meyer	Now only running "cfg p" on systems with NL4 routers.
#     05/19/06  Gary Meyer	Updated for 1955 chassis type.  Removed more L1 commands if Shub2.
#     06/23/06  Gary Meyer	Updated so chassis type info is provided for Shub2 and TIO nodes.
#     07/24/06  Gary Meyer	Print chassis type with verbose messages.
#     09/14/06  Gary Meyer	Improved formatting of leds output.
#     09/20/06  Gary Meyer	Fixes for problems with ioprt and pci commands with Tollhouse systems.
#     10/13/06  Gary Meyer	Added check for total cpu count - generate warning if none are found.
#     11/01/06  Gary Meyer	Added "fru status" command for Tollhouse systems.
#     11/10/06  Gary Meyer	Corrected problem with incorrect error message when missing serial info.
#     11/16/06  Gary Meyer      Removed sysco_close_connection($l2_host) from the end of initial_l1_checks().
#     11/21/06  Gary Meyer      Added war disable/enable.  pwr vrm parsing too.
#     11/28/06  Gary Meyer      Updated to recognize IP79 & IP81 blade types.
#     01/10/07  Gary Meyer      Updated to generate the correct processor cache size.
#     01/11/07  Gary Meyer      Updated to recognize Montvale processor types.
#     03/14/07  Gary Meyer      Fixed problem with Compute Blade Summary table formatting.
#     03/19/07  Gary Meyer      Moving the l1 link, leds, and pwr vrm commands to occur at the beginning of the script.
#
#
#-------------------------------------------------------------------------------

################################################################################
#
#  Parse super-scope generated @module_names array.
#
#  OPUS bricks will look like:
#  brick=001c01 board=NODE chip=SHUB0
#  brick=001c01 board=RISER chip=PIC1
#
# C brick:
# CHIP[00] : NODE1 M7H6/SHUB1   shub_01      7  0  0x003F 0  ARRAY(0x883da98)
# CHIP[01] : NODE0 E2H6/SHUB0   shub_01      7  7  0x003F 0  ARRAY(0x8844818)
#
# I brick:
# CHIP[00] : PX    C6H9B/PIC1   pic_02       8  0  0x001F 0  ARRAY(0x8847868)
# CHIP[01] : PX    J1H9B/PIC0   pic_02       8  8  0x001F 0  ARRAY(0x884c83c)
# CHIP[02] : PX    P6H9B/PIC2   pic_02       8  16 0x001F 0  ARRAY(0x8851624)
#
# NL3R brick:
# CHIP[00] : R     F9B5/RTR     rtr_00       5  0  0x0000 0  ARRAY(0x841d348)
#
# NL4R brick:
# CHIP[00] : R     F8B5/RTR0    nl4rtr_01    7  0  0x0027 0  ARRAY(0x846e3c0)
#
# Vortix brick: (NL4R)
# target[0]=001c01  target[1]=R_TOP  target[2]=RTR
# target[0]=001c01  target[1]=R_BOT  target[2]=RTR
# target[0]=001c01  target[1]=NODE0  target[2]=SHUB0
# target[0]=001c01  target[1]=NODE1  target[2]=SHUB0
# target[0]=001c01  target[1]=NODE2  target[2]=SHUB0
# target[0]=001c01  target[1]=NODE3  target[2]=SHUB0
#
# 191010 chassis:
# ->Chip: 001c02:BLADE2:SHUB0
# target[0]=001c02  target[1]=BLADE2  target[2]=SHUB0
# ->Chip: 001c02:BLADE3:SHUB0
# target[0]=001c02  target[1]=BLADE3  target[2]=SHUB0
# ->Chip: 001c02:RTR_U:RTR1
# target[0]=001c02  target[1]=RTR_U  target[2]=RTR1
# ->Chip: 001c02:RTR_U:RTR0
# target[0]=001c02  target[1]=RTR_U  target[2]=RTR0
# ->Chip: 001c02:RTR_L:RTR1
# target[0]=001c02  target[1]=RTR_L  target[2]=RTR1
# ->Chip: 001c02:RTR_L:RTR0
# target[0]=001c02  target[1]=RTR_L  target[2]=RTR0
#
# 1955 chassis:
# ->Module name: 001c01
#  chip list: ( 001c01:BLADE0:TIO0 001c01:BLADE1:SHUB0 001c01:RTR_L:RTR0 001c01:RTR_R:RTR0 )
# ->Chip: 001c01:BLADE0:TIO0
# target[0]=001c01  target[1]=BLADE0  target[2]=TIO0
# ->Chip: 001c01:BLADE1:SHUB0
# target[0]=001c01  target[1]=BLADE1  target[2]=SHUB0
# ->Chip: 001c01:RTR_L:RTR0
# target[0]=001c01  target[1]=RTR_L  target[2]=RTR0
# ->Chip: 001c01:RTR_R:RTR0
# target[0]=001c01  target[1]=RTR_R  target[2]=RTR0


################################################################################

sub parse_scan_config {

  print "\nIn parse_scan_config function\n" if ($verbose_mode == 1);
  my $c_brick_count = 0;
  my $th_iru_count = 0;
  my $opus_brick_count = 0;
  my $vortix_brick_count = 0;
  my $mc_brick_count = 0;
  my $r_brick_count = 0;
  my $r4_brick_count = 0;
  my $io_brick_count = 0;
  my $tio_brick_count = 0;
  my $g_brick_count = 0;
  my $brick_count = 0;
  my $pic_count = 0;


  # handle the case where a user config is entered, and we only want to dump node 0
  if ( ($node_select == 0) && ($multi_node == 0) ) {
    $target_plane = 0;
  }

  # extract valid c and i-bricks from the configuration
  foreach $new_brick (@module_names)  {         #  This array is defined in superscope libraries
    my $rtr_type = 0;		# single router or Vortix

    $mref = $modules{ $new_brick };
    @chip_names = scan_module_chips($mref, 0xFFFF, 'logical');
    if ($verbose_mode == 1)  {
      printf "\n->Module name: $new_brick\n";
      printf "      chip list: ( @chip_names )\n";
    }

    if ($new_brick =~ /c/)  {
      @targets = split (/c/, $new_brick);
      if (($user_target == 0)  || (($user_target == 1) && ($targets[0] eq $target_rack) && ($targets[1] eq $target_slot)))  {
        $l1_c_brick_list[$c_brick_count]->{RACK} = $targets[0];
        $l1_c_brick_list[$c_brick_count]->{SLOT} = $targets[1];
        $c_brick_count++;
        $l1_mc_brick_list[$mc_brick_count]->{RACK} = $targets[0];
        $l1_mc_brick_list[$mc_brick_count]->{SLOT} = $targets[1];
        $mc_brick_count++;
	my $found_blade = 0;
        my $rtr_count = 0;
	foreach $chip_elem (@chip_names) {
          my @target = split(":", $chip_elem);
	  if (($chip_elem =~ /BLADE/) && ($found_blade == 0)) {
	    $th_iru_count++;
	    $found_blade = 1;
	  }
	  if (($found_blade == 1) && ($target[1] =~ /RTR_/)) {
	    $rtr_count++;
	  }
	}
	$rtr_type = ($rtr_count == 0) ? 0 : ($rtr_count == 2) ? 1 : 2;		#type 1 is 1955, type 2 is 191010
      }
      else  {
        printf "  Skipping C Brick - Does not match user designated Rack and Slot values.\n"  if ($verbose_mode == 1);
        next;
      }
    }
    elsif ($new_brick =~ /m/)  {
      @targets = split (/m/, $new_brick);
      if (($user_target == 0)  || (($user_target == 1) && ($targets[0] eq $target_rack) && ($targets[1] eq $target_slot)))  {
        $l1_mc_brick_list[$mc_brick_count]->{RACK} = $targets[0];
        $l1_mc_brick_list[$mc_brick_count]->{SLOT} = $targets[1];
        $mc_brick_count++;
      }
      else  {
        printf "  Skipping M Brick - Does not match user designated Rack and Slot values.\n"  if ($verbose_mode == 1);
        next;
      }
    }
    elsif ($new_brick =~ /g/)  {
      @targets = split (/g/, $new_brick);
      if (($user_target == 0)  || (($user_target == 1) && ($targets[0] eq $target_rack) && ($targets[1] eq $target_slot)))  {
        $l1_mc_brick_list[$mc_brick_count]->{RACK} = $targets[0];
        $l1_mc_brick_list[$mc_brick_count]->{SLOT} = $targets[1];
        $mc_brick_count++;
        $g_brick_count++;
      }
      else  {
        printf "  Skipping Graphics Brick - Does not match user designated Rack and Slot values.\n"  if ($verbose_mode == 1);
        next;
      }
    }
    elsif ($new_brick =~ /i/)  {                #  could be PIC or TIO
      @targets = split (/i/, $new_brick);
      if (($user_target == 0)  || (($user_target == 1) && ($targets[0] eq $target_rack) && ($targets[1] eq $target_slot)))  {
        @target = split(":", $chip_names[0]);
        if ( $target[2] =~ /^TIO/i ) {
          $l1_tio_brick_list[$tio_brick_count]->{RACK} = $targets[0];
          $l1_tio_brick_list[$tio_brick_count]->{SLOT} = $targets[1];
          $tio_brick_count++;
        }
        else  {
          $l1_io_brick_list[$io_brick_count]->{RACK} = $targets[0];
          $l1_io_brick_list[$io_brick_count]->{SLOT} = $targets[1];
          $io_brick_count++;
        }
      }
      else  {
        printf "  Skipping I Brick - Does not match user designated Rack and Slot values.\n"  if ($verbose_mode == 1);
        next;
      }
    }
    elsif ($new_brick =~ /t/)  {
      @targets = split (/t/, $new_brick);
      if (($user_target == 0)  || (($user_target == 1) && ($targets[0] eq $target_rack) && ($targets[1] eq $target_slot)))  {
        $l1_io_brick_list[$tio_brick_count]->{RACK} = $targets[0];
        $l1_io_brick_list[$tio_brick_count]->{SLOT} = $targets[1];
        $tio_brick_count++;
      }
      else  {
        printf "  Skipping T Brick - Does not match user designated Rack and Slot values.\n"  if ($verbose_mode == 1);
        next;
      }
    }
    elsif ($new_brick =~ /p/) {
      @targets = split (/p/, $new_brick);
      if (($user_target == 0)  || (($user_target == 1) && ($targets[0] eq $target_rack) && ($targets[1] eq $target_slot)))  {
        $l1_io_brick_list[$io_brick_count]->{RACK} = $targets[0];
        $l1_io_brick_list[$io_brick_count]->{SLOT} = $targets[1];
        $io_brick_count++;
      }
      else  {
        printf "  Skipping P Brick - Does not match user designated Rack and Slot values.\n"  if ($verbose_mode == 1);
        next;
      }
    }
    elsif ($new_brick =~ /r/) {
      @targets = split (/r/, $new_brick);
      if (($user_target == 0)  || (($user_target == 1) && ($targets[0] eq $target_rack) && ($targets[1] eq $target_slot)))  {
        $l1_r_brick_list[$r_brick_count]->{RACK} = $targets[0];
        $l1_r_brick_list[$r_brick_count]->{SLOT} = $targets[1];
        $r_brick_count++;
      }
      else  {
        printf "  Skipping R Brick - Does not match user designated Rack and Slot values.\n"  if ($verbose_mode == 1);
        next;
      }
    }

    $l1_brick_list[$brick_count]->{RACK} = $targets[0];
    $l1_brick_list[$brick_count]->{SLOT} = $targets[1];

    my $rack_exists = 0;
    foreach $rack_entry (@rack_list)  {
      if ($targets[0] == $rack_entry)  {
        $rack_exists = 1;
        if (( ! (defined $rack_entry->{CBRICK})) && (($new_brick =~ /c/) || ($new_brick =~ /m/)) )  {
          $rack_entry->{CBRICK} = 1;
          last;
        }
      }
    }
    if ($rack_exists == 0)  {
      push (@rack_list, $targets[0]);
      $rack_list[$#rack_list]->{RACK} = $targets[0];
      if (($new_brick =~ /c/) || ($new_brick =~ /m/))  {
        $rack_list[$#rack_list]->{CBRICK} = 1;
      }
    }

    my $shub_count = 0;
    my $shub2_count = 0;

   #  Process each name with a format = <brick>:<board>:<chip>
    foreach $curr_name ( @chip_names )  {
      printf "\n->Chip: $curr_name\n" if ($verbose_mode == 1);
      @target = split(":", $curr_name);

      printf "target[0]=$target[0]  target[1]=$target[1]  target[2]=$target[2]\n" if ($verbose_mode == 1);
      if ( $target[2] =~ /^SHUB0/i )
      {
	#  Get pointer to chip structure/array
        my $chip_ptr = scan_module_chip_ptr($mref, $target[1], $target[2], "logical");
	
	#  Extract the chip type/description
	my $chip_type = $$chip_ptr[$CMP_IND_DESC];

	if (($chip_type =~ /^shub2/i) && (($target_plane == 0) || ($target_plane == 2)))  {
          $sh2_target_list[$sh2_node_count]->{MODULE} = "$target[0]";
          $sh2_target_list[$sh2_node_count]->{NODE} = "$target[1]";
          $sh2_target_list[$sh2_node_count]->{CHIP} = "$target[2]";
          $sh2_target_list[$sh2_node_count]->{CHASSIS} = $rtr_type;
          printf("ADD SH2 SHUB0: brick=$target[0] board=$target[1] chip=$target[2]  CHASSIS=$rtr_type\n")  if ($verbose_mode == 1);
          $sh2_node_count++;
          $shub2_count++;
        }
	elsif (($chip_type =~ /^shub/i) && (($target_plane == 0) || ($target_plane == 2)))  {
          $target_list[$sh1_node_count]->{MODULE} = "$target[0]";
          $target_list[$sh1_node_count]->{NODE} = "$target[1]";
          $target_list[$sh1_node_count]->{CHIP} = "$target[2]";
          printf("ADD SH1 SHUB0: brick=$target[0] board=$target[1] chip=$target[2]\n")  if ($verbose_mode == 1);
          $sh1_node_count++;
          $shub_count++;
        }
      }
      elsif ( $target[2] =~ /^SHUB1/i )
      {
	#  Get pointer to chip structure/array
        my $chip_ptr = scan_module_chip_ptr($mref, $target[1], $target[2], "logical");

	#  Extract the chip type/description
	my $chip_type = $$chip_ptr[$CMP_IND_DESC];

	if (($chip_type =~ /^shub2/i ) && ($target_plane > 0)) {
          $sh2_target_list[$sh2_node_count]->{MODULE} = "$target[0]";
          $sh2_target_list[$sh2_node_count]->{NODE} = "$target[1]";
          $sh2_target_list[$sh2_node_count]->{CHIP} = "$target[2]";
          $sh2_target_list[$sh2_node_count]->{CHASSIS} = $rtr_type;
          printf("ADD SH2 SHUB1: brick=$target[0] board=$target[1] chip=$target[2]  CHASSIS=$rtr_type\n")  if ($verbose_mode == 1);
          $sh2_node_count++;
          $shub2_count++;
        }
	elsif (($chip_type =~ /^shub/i) && ($target_plane > 0)) {
          $target_list[$sh1_node_count]->{MODULE} = "$target[0]";
          $target_list[$sh1_node_count]->{NODE} = "$target[1]";
          $target_list[$sh1_node_count]->{CHIP} = "$target[2]";
          $sh1_node_count++;
          $shub_count++;
          printf("ADD SH1 SHUB1: brick=$target[0] board=$target[1] chip=$target[2]\n")  if ($verbose_mode == 1);
        }
      }
      elsif ( $target[2] =~ /^SHUB/i )
      {

        #  Get pointer to chip structure/array
        my $chip_ptr = scan_module_chip_ptr($mref, $target[1], $target[2], "logical");

        #  Extract the chip type/description
        my $chip_type = $$chip_ptr[$CMP_IND_DESC];

        if ($chip_type =~ /^shub2/i)  {
          $sh2_target_list[$sh2_node_count]->{MODULE} = "$target[0]";
          $sh2_target_list[$sh2_node_count]->{NODE} = "$target[1]";
          $sh2_target_list[$sh2_node_count]->{CHIP} = "$target[2]";
          $sh2_target_list[$sh2_node_count]->{CHASSIS} = $rtr_type;
          printf("ADD SH2 SHUB: brick=$target[0] board=$target[1] chip=$target[2]  CHASSIS=$rtr_type\n")  if ($verbose_mode == 1);
          $sh2_node_count++;
          $shub2_count++;
        }
        elsif ($chip_type =~ /^shub/i)  {
          $target_list[$sh1_node_count]->{MODULE} = "$target[0]";
          $target_list[$sh1_node_count]->{NODE} = "$target[1]";
          $target_list[$sh1_node_count]->{CHIP} = "$target[2]";
          $sh1_node_count++;
          $shub_count++;
          printf("ADD SH1 SHUB: brick=$target[0] board=$target[1] chip=$target[2]\n")  if ($verbose_mode == 1);
        }
      }
      elsif ( $target[2] =~ /^PIC/i )
      {
        $io_target_list[$pic_count]->{MODULE} = "$target[0]";
        $io_target_list[$pic_count]->{NODE} = "$target[1]";
        $io_target_list[$pic_count]->{CHIP} = "$target[2]";
        $pic_count++;
        printf("ADD PIC: brick=$target[0] board=$target[1] chip=$target[2]\n")  if ($verbose_mode == 1);
      }

      elsif ( $target[2] =~ /^RTR/i )
      {
          $mref = $modules{ $target[0] };
          $dref = scan_module_chip_details($mref, $target[1], $target[2], 'logical');
          if ( $$dref[$DETAIL_IND_TYPE] =~ /^nl4r/i )
          {
             $nl4r_target_list[$nl4r_node_count]->{MODULE} = "$target[0]";
             $nl4r_target_list[$nl4r_node_count]->{NODE}   = "$target[1]";
             $nl4r_target_list[$nl4r_node_count]->{CHIP}   = "$target[2]";
             $nl4r_target_list[$nl4r_node_count]->{CHASSIS}= $rtr_type;
             $nl4r_node_count++;
             $nl4r_present = 1;
             printf("ADD NL4R: brick=$target[0] board=$target[1] chip=$target[2]  CHASSIS=$rtr_type\n")  if ($verbose_mode == 1);
          }
      }

      elsif ( $target[2] =~ /^TIO/i )
      {
        $tio_target_list[$tio_node_count]->{MODULE} = "$target[0]";
        $tio_target_list[$tio_node_count]->{NODE} = "$target[1]";
        $tio_target_list[$tio_node_count]->{CHIP} = "$target[2]";
        $tio_target_list[$tio_node_count]->{CHASSIS} = $rtr_type;
        if (($tio_node_count +1) != $tio_brick_count) {
          $l1_io_brick_list[$tio_node_count]->{RACK} = $targets[0];
          $l1_io_brick_list[$tio_node_count]->{SLOT} = $targets[1];
        }
        $tio_node_count++;
        $tio_present = 1;
        printf("ADD TIO: brick=$target[0] board=$target[1] chip=$target[2]  CHASSIS=$rtr_type\n")  if ($verbose_mode == 1);
      }

#  Determine if this is a 2U opus (ip57) or 3U Altix (ip41) node board.
#  At some point, may need to revisit this to grab the {TYPE} hash key
#  from the module list provided by superscope

      if (($target[1] =~ /^NODE\b/i) && ($shub2_count == 0))
      {
        $l1_c_brick_list[$c_brick_count-1]->{OPUS} = 1;
        $l1_mc_brick_list[$mc_brick_count-1]->{OPUS} = 1;
        $target_list[$sh1_node_count-1]->{OPUS} = 1;
        $opus_brick_count++;
        if ($dump_ii_regs == 1)  {
          printf("\n This is an OPUS brick!!\n")  if ($verbose_mode == 1);
          push(@io_c_brick_list, $target[0]);
        }
      }

    }                       #  Go to the next chip

    if ($shub_count > 2)  {
        $l1_c_brick_list[$c_brick_count-1]->{VORTIX} = 1;
        $l1_mc_brick_list[$mc_brick_count-1]->{VORTIX} = 1;
        for (my $v_shub = 1; $v_shub < ($shub_count + 1); $v_shub++)  {
            $target_list[$sh1_node_count- $v_shub]->{VORTIX} = 1;
        }
        $vortix_brick_count++;
        printf("\n This is a VORTIX brick!!\n")  if ($verbose_mode == 1);
    }
    $brick_count++;     #  Go to the next brick
  }

  print "\n\n";

  my $altix_brick_count = $c_brick_count - $opus_brick_count - $vortix_brick_count - $th_iru_count;
  $rack_count = $#rack_list + 1;

  $config_summ = sprintf "\n\n=========================================================\n";
  $config_summ .= sprintf "\nCurrent system configuration summary:\n";
  if ($altix_brick_count > 0 )  {
    $config_summ .= sprintf "    Altix C bricks: %4d\n", $altix_brick_count;
  }
  if ($vortix_brick_count > 0 )  {
    $config_summ .= sprintf "     Vortix bricks: %4d\n", $vortix_brick_count;
  }
  if ($opus_brick_count > 0 )  {
    $config_summ .= sprintf "       Opus bricks: %4d\n", $opus_brick_count;
  }
  if ($th_iru_count > 0 )  {
    $config_summ .= sprintf "    Tollhouse IRUs: %4d\n", $th_iru_count;
  }
  if ($sh1_node_count > 0 )  {
    $config_summ .= sprintf "     Shub 1x chips: %4d\n", $sh1_node_count;
  }
  if ($sh2_node_count > 0 )  {
    $config_summ .= sprintf "     Shub 2x chips: %4d\n", $sh2_node_count;
  }
  if ($io_brick_count > 0 )  {
    $config_summ .= sprintf "         IO bricks: %4d\n", $io_brick_count;
  }
  if ($tio_brick_count > 0 )  {
    $config_summ .= sprintf "        TIO bricks: %4d\n", $tio_brick_count;
  }
  if ($g_brick_count > 0 )  {
    $config_summ .= sprintf "   Graphics bricks: %4d\n", $g_brick_count;
  }
  if ($tio_node_count > 0 )  {
    $config_summ .= sprintf "         TIO chips: %4d\n", $tio_node_count;
  }
  if ($r_brick_count > 0 )  {
    if ($nl4r_present == 1) {
        $config_summ .= sprintf "      NL4 R bricks: %4d\n", $r_brick_count;
    }
    else  {
        $config_summ .= sprintf "      NL3 R bricks: %4d\n", $r_brick_count;
    }
  }
  if ($nl4r_node_count > 0 )  {
    $config_summ .= sprintf "        NL4R chips: %4d\n", $nl4r_node_count;
  }
  $config_summ .=   sprintf "      System racks: %4d\n", $rack_count;
  $config_summ .= sprintf "\n\n------------------------------------------------------------------\n\n";

  print "$config_summ";

  if (($altix_brick_count == 0) && ($vortix_brick_count == 0) && ($opus_brick_count > 0))  {
    $l1_war_enable = 0;
  }
}


#-------------------------------------------------------------------------------
#
#  get_system_env()  subroutine
#
#  outputs the following environmental settings:
#       debug switch settings
#       port status
#       ni cfg settings
#       voltage margins
#       brick config
#
#-------------------------------------------------------------------------------

sub get_system_env()
{
   my $brick_count = 0;

        #  Get the PWR, MEM_DTR, and Processor SN's
   if ($shub_version == 1)  {
     $l1_output_string .= "\n->Recording brick serial information...\n";
     $l1_output_string .= send_l1_cmd('serial all', 'all_b');

     $l1_output_string .= "\n->Recording L1 fsb config...\n";
     $l1_output_string .= send_l1_cmd('shcfg fsb', 'c');
   }

   $l1_output_string .= "\n->Recording processor config...\n";
   $l1_output_string .= send_l1_cmd('cpu', 'mc');

   if ($shub_version == 1)  {
     $l1_output_string .= "\n->Recording nicfg ni0...\n";
     $l1_output_string .= send_l1_cmd('nicfg ni0', 'mc');

     $l1_output_string .= "\n->Recording nicfg ni1...\n";
     $l1_output_string .= send_l1_cmd('nicfg ni1', 'mc');
   }

   if (($nl4r_present) && ($#l1_r_brick_list >= 0)) {
       $l1_output_string .= "\n->Recording NL4R scramble mode settings...\n";
       $l1_output_string .= send_l1_cmd('nlcfg scramble', 'r');
   }

   if ($shub_version == 2)  {
       $l1_output_string .= "\n->Recording Blade config...\n";
       $l1_output_string .= send_l1_cmd('blade', 'c');
	
       $l1_output_string .= "\n->Recording Processor config...\n";
       $l1_output_string .= send_l1_cmd('cpu fsb', 'c');
	
       $l1_output_string .= "\n->Recording dimm status...\n";
       $l1_output_string .= send_l1_cmd('serial dimm status', 'all_c');
	
       $l1_output_string .= "\n->Recording fru status...\n";
       $l1_output_string .= send_l1_cmd('fru status', 'all');
   }

   $l1_output_string .= "\n->Recording Port status...\n";
   $l1_output_string .= send_l1_cmd('prt', 'all');

   if ($shub_version == 1)  {
     $l1_output_string .= "\n->Recording IO Port status...\n";
     $l1_output_string .= send_l1_cmd('ioprt', 'all');
   }

   $l1_output_string .= "\n->Recording PIC config info ...\n";
   $l1_output_string .= send_l1_cmd('pci', 'io');

   $l1_output_string .= "\n->Recording voltage margins...\n";
   $l1_output_string .= send_l1_cmd('mgn', 'all');

   $l1_output_string .= "\n->Recording L1 environmental info...\n";
   $l1_output_string .= send_l1_cmd('env', 'all_b');

   $l1_output_string .= "\n->Recording L1 log info...\n";
   $l1_output_string .= send_l1_cmd('log', 'all_b');

   $l1_output_string .= "\n->Recording debug switch settings...\n";
   $l1_output_string .= send_l1_cmd('debug', 'mc');

   if ($shub_version == 1)  {
       $l1_output_string .= "\n->Recording PromIce jumper settings...\n";
       $l1_output_string .= send_l1_cmd('promice', 'c');
   }

   $l1_output_string .= "\n->Recording L1 Flash status...\n";
   $l1_output_string .= send_l1_cmd('flash status', 'all');

   $l1_output_string .= "\n->Recording L2 firmware version...\n";
   $l1_output_string .= send_l1_cmd('version', '1');

   if ((($l1_war_enable == 1) && ($shub_version == 1)) || ($l1_war_status == 1))  {
       $l1_output_string .= "\n->Reenabling L1 war...\n";
       $l1_output_string .= send_l1_cmd('war e', 'c');
       if ($l1_war_status == 1) {
	  $l1_war_status = 2;
       }
   }
}


#-------------------------------------------------------------------------------
#
# function: send_l1_cmd
#
# called by the get_system_env() function to send the supplied string
# as a L1 command.
#
#  $com_mode can be:
#       1 - send a single command to all bricks - works best for L2 commands with limited output
#     all - send the L1 command on a per rack basis to all bricks in the config.
#   all_b - send the L1 command on a brick-by-brick basis to all bricks in the config.
#   all_c - send the L1 command on a brick-by-brick basis to all C bricks in the config.
#      mc - send the L1 command on a per rack basis to all m and c bricks in the config.
#       c - send the L1 command on a per rack basis to all c bricks in the config.
#      io - send the L1 command on a brick-by-brick basis to Ix & Px bricks only
#       r - send the L1 command on a brick-by-brick basis to all r bricks in the config.
#
#-------------------------------------------------------------------------------

sub send_l1_cmd  {

  my $cmd_to_send = "$_[0]";
  my $com_mode    = "$_[1]";
  my @sub_brick_list = ();
  my $cfg_brick;
  my $return_string = "";
  my $text = "";
  my $l1_target_slot = "*";

  if ($user_target == 1)  {
    $l1_target_slot = "$user_slot";
  }


  print "\nIn send_l1_cmd function - command is  $cmd_to_send  - mode is $com_mode\n" if ($verbose_mode == 1);

  if ($com_mode eq '1' )  {
    my $cmd  = sprintf("%c%s", $CHAR_CTRL_T, $cmd_to_send);
    $text = sysco_raw_cmd($cmd);
    if ( !defined($text) )
    {  $return_string = sprintf "\nSYSTEM STOPPED RESPONDING OR ENCOUNTERED ERROR - ignore remainder of sequence\n";
       $timeout_counter++;
       return;
    }
    else  {
      $return_string = "    response text >>>>$text\n<<<<\n";
    }
  }

  elsif ( $com_mode eq 'all' ) {
    foreach $cfg_rack (@rack_list)  {
      my $cmd  = sprintf("%cr %d s %s %s", $CHAR_CTRL_T, $cfg_rack->{RACK}, $l1_target_slot, $cmd_to_send);
      $text = sysco_raw_cmd($cmd);
      if ( !defined($text) )
      {  $return_string = sprintf "\nSYSTEM STOPPED RESPONDING OR ENCOUNTERED ERROR - ignore remainder of sequence\n";
         $timeout_counter++;
         return;
      }
      else  {
        $return_string .= "    response text >>>>$text\n<<<<\n";
      }
    }
  }
  elsif (( $com_mode eq 'c' ) || ( $com_mode eq 'mc' ))   {
    foreach $cfg_rack (@rack_list)  {
      if (defined $cfg_rack->{CBRICK})  {
        my $cmd  = sprintf("%cr %d s %s %s", $CHAR_CTRL_T, $cfg_rack->{RACK}, $l1_target_slot, $cmd_to_send);
        $text = sysco_raw_cmd($cmd);
        if ( !defined($text) )
        {  $return_string = sprintf "\nSYSTEM STOPPED RESPONDING OR ENCOUNTERED ERROR - ignore remainder of sequence\n";
           $timeout_counter++;
           return;
        }
        else  {
          $return_string .= "    response text >>>>$text\n<<<<\n";
        }
      }
    }
  }
  else  {
    if ( $com_mode eq 'io' )  {                 #  only the I & P bricks
      @sub_brick_list = @l1_io_brick_list;
    }
    elsif ( $com_mode eq 'r' )  {
      @sub_brick_list = @l1_r_brick_list;
    }
    elsif ( $com_mode eq 'all_b' )  {
      @sub_brick_list = @l1_brick_list;
    }
    elsif ( $com_mode eq 'all_c' )  {
      @sub_brick_list = @l1_c_brick_list;
    }

    foreach $cfg_brick (@sub_brick_list)  {
      my $cmd  = sprintf("%c%s.%s %s", $CHAR_CTRL_T, $cfg_brick->{RACK}, $cfg_brick->{SLOT}, $cmd_to_send);
      $text = sysco_raw_cmd($cmd);
      if ( !defined($text) )
      {  $return_string .= sprintf "\n$cmd:  SYSTEM STOPPED RESPONDING OR ENCOUNTERED ERROR - ignore remainder of sequence\n";
         $timeout_counter++;
         return;
      }
      else  {
        if ( $cmd_to_send eq 'log' )  {
          $return_string .= "$cmd_to_send response text >>>>" . check_log_string($text) . "\n\n";
        }
        else  {
          $return_string .= "$cmd_to_send response text >>>>$text\n<<<<\n";
        }
      }
    }
  }



  if ($cmd_to_send eq "cpu")  {
    check_cpu_string($return_string);
  }

  # if ($cmd_to_send eq "prt")  {
  #  check_prt_string($return_string);
  # }

  if ($cmd_to_send eq "link")  {
    check_link_string($return_string);
  }

  if ($cmd_to_send eq "leds")  {
    $new_return_string = check_leds_string($return_string);
    $return_string = $new_return_string;
  }

  if ($cmd_to_send eq "pwr vrm")  {
    $new_return_string = check_vrm_string($return_string);
    $return_string = $new_return_string;
  }

  if (($cmd_to_send eq 'cfg') && ( $com_mode eq 'io' ) && ( $dump_ii_regs == 1))  {
    check_iocfg_string($return_string);
  }

  if ($cmd_to_send eq "war")  {
    check_war_string($return_string);
  }

  return ($return_string);

}


#-------------------------------------------------------------------------------
#
#  this subroutine parses the string which contains the output
#  of the l1 "cpu" command.  The intent is to check for cpus
#  that aren't present or are disabled.
#
#-------------------------------------------------------------------------------

sub check_cpu_string  {

  my $cpu_string = "$_[0]";
  my $line_cnt   = 0;
  my $cpu_cnt    = 0;
  my $current_brick;
  my $prev_line;
  my $this_cpu_stat;

  my @str_lines = split(/\n/, $cpu_string);     # divide the output of the cpu cmd into lines

  while (defined ($str_lines[$line_cnt]))  {
    if (($str_lines[$line_cnt] =~ /CPU/) && ($str_lines[$line_cnt] =~ /Present/) && ($str_lines[$line_cnt] =~ /Enabled/)) {
      if ($verbose_mode == 1) {
        print "\nGot a match to CPU header line!!\n";
      }
      $prev_line = $line_cnt - 1;
      $current_brick = $str_lines[$prev_line];
      $current_brick =~ s/: //;
      $line_cnt +=2;
      my $valid_cpu_info = 1;
      while ($valid_cpu_info == 1)  {
        #print "\nIn the CPU loop\n";
	if ($str_lines[$line_cnt] !~ /^[0-9][A-D]/) {
	  $valid_cpu_info = 0;
          if ($verbose_mode == 1) {
            print "\nExiting CPU info for $current_brick at $line_cnt\n";
          }
	}
	else {
          $this_cpu_stat = 0;
          my @cpu_attrib = split(/\s+/, $str_lines[$line_cnt]);
          $cpu_attrib[0] =~ s/\s*//g;
          $cpu_attrib[1] =~ s/\s*//g;
          $cpu_attrib[2] =~ s/\s*//g;
          if ($verbose_mode == 1)  {
            print "$current_brick: cpu_attrib0: $cpu_attrib[0].\n";
            print "$current_brick: cpu_attrib1: $cpu_attrib[1].\n";
            print "$current_brick: cpu_attrib2: $cpu_attrib[2].\n";
          }
          if (($cpu_attrib[1] != 1) && ("$cpu_attrib[1]" !~ /[0-9]+/))  {
            last;
          }
          else  {
            if (($cpu_attrib[1] == 1) && ($cpu_attrib[2] == 0)) {
              $this_cpu_stat++;
              $l1_env_err = 1;
              $l1_env_err_msg .= "  ERROR!  CPU $cpu_attrib[0] Config error on $current_brick.\n";
              $l1_env_err_msg .= "  Expected CPU Enabled: 1     Actual CPU Enabled: $cpu_attrib[2].\n\n";
            }
  
            if ($this_cpu_stat == 0) {
              if ($verbose_mode == 1) {
                print "\n L1 CPU Status OK on Brick $current_brick, CPU $cpu_attrib[0]\n";
              }
            }
          }
          $line_cnt++;
        }
      }		# end of while loop
    }
    else  {
      if ($verbose_mode == 1) {
        print "\nNo Match found in string:  $str_lines[$line_cnt].\n";
      }
      $line_cnt++;
    }
  }

}

#-------------------------------------------------------------------------------
#
#  This subroutine checks the returned value of the io config topology L1 command.
#  The selected bricks are saved in the @io_c_brick_list array.
#
#
#-------------------------------------------------------------------------------


sub check_iocfg_string  {

  my $iocfg_string = "$_[0]";
  my $line_cnt     = 0;
  my $iocfg_cnt    = 0;
  my $new_c_brick;
  my $list_member;
  my $this_iocfg_stat;

  my @str_lines = split(/\n/, $iocfg_string);     # divide the output of the cfg cmd into lines

  while (defined ($str_lines[$line_cnt]))  {
    if ($str_lines[$line_cnt] =~ /\d\d\dc\d\d/) {               #  Is a C brick listed?
      my @cbrick_attrib = split(/\*/, $str_lines[$line_cnt]);	#  now uses a "*"
      if ($cbrick_attrib[1] !~ /\d\d\dc\d\d/) {
        @cbrick_attrib = split(/\+/, $str_lines[$line_cnt]);
      }
      if ($cbrick_attrib[1] !~ /\d\d\dc\d\d/) {
        @cbrick_attrib = split(/\-/, $str_lines[$line_cnt]);
      }
      $cbrick_attrib[1] =~ s/\s*//g;
      my $cbrick_name = $cbrick_attrib[1];
      $new_c_brick = 1;

      foreach $list_member (@io_c_brick_list) {         #  Check if C brick is already in the list
        if ($cbrick_name =~ /$list_member/) {
          $new_c_brick = 0;
          last;
        }
      }
      if ($new_c_brick == 1)  {
        push(@io_c_brick_list, $cbrick_name);
      }
    }
    $line_cnt++;
  }

  print "\n  The following C bricks will have their II registers dumped:\n";
  foreach $list_member (@io_c_brick_list) {
    print "\t$list_member\n";
  }
  print "\n";
}



#-------------------------------------------------------------------------------
#
#  this subroutine parses the string which contains the output
#  of the l1 "prt" command.  The intent is to check for network port
#  connections that are down, especially on router to router connections
#   which currently aren't being flagged in the prom.
#
#-------------------------------------------------------------------------------

sub check_prt_string  {

  my $prt_string = "$_[0]";
  my $line_cnt   = 0;
  my $prt_cnt    = 0;
  my $current_brick;
  my $prev_line;
  my $port_limit = 0;

  my @str_lines = split(/\n/, $prt_string);     # divide the output of the prt cmd into lines

  while (defined ($str_lines[$line_cnt]))  {
    if ($str_lines[$line_cnt] =~ /\d\d\d[r|c]\d\d:/) {
      # print "\nGot a match to !!\n";
      $current_brick = "$str_lines[$line_cnt]";
      $current_brick =~ s/://;
      $line_cnt +=3;
      if ($current_brick =~ /c/)  {
        $port_limit = 4;  }
      elsif ($current_brick =~ /r/)  {
        $port_limit = 8;
      }

      for ($prt_cnt = 0; $prt_cnt < $port_limit ; $prt_cnt++)  {
        if ($verbose_mode == 1) {
          print "\nIn the Port parsing loop\n";
        }
        my @prt_attrib = split(/\s+/, $str_lines[$line_cnt]);
        $prt_attrib[1] =~ s/\s*//g;
        $prt_attrib[2] =~ s/\s*//g;
        $prt_attrib[3] =~ s/\s*//g;
        # print "attrib1: $prt_attrib[1].\n";
        # print "attrib2: $prt_attrib[2].\n";
        # print "attrib3: $prt_attrib[3].\n";
        if (("$prt_attrib[2]" eq "0x0f") || ("$prt_attrib[2]" eq "0x02")) {
          if ($verbose_mode == 1) {
            print "\n L1 port Status OK on Brick $current_brick, Port $prt_attrib[1].\n";
          }
        }
        else  {
          $l1_env_err = 1;
          $l1_env_err_msg .= "  Error!  Incorrect port status on $current_brick,  Port $prt_attrib[1].\n";
          $l1_env_err_msg .= "  Expecting status of 0xf or 0x2     Actual port status value:  $prt_attrib[2]\n\n";
        }
        $line_cnt++;
      }
    }
    else  {
      if ($verbose_mode == 1) {
        print "\nNo Match found in string:  $str_lines[$line_cnt].\n";
      }
      $line_cnt++;
    }
  }

}


#-------------------------------------------------------------------------------
#
# parse the output of the link command, and capture any errors detected.
#
#-------------------------------------------------------------------------------

sub check_link_string  {

  my $link_string = "$_[0]";
  my $line_cnt   = 0;
  my $prt_cnt    = 0;
  my $current_brick;
  my $brick_acquired = 0;
  my $prev_cnt = 0;

  my @str_lines = split(/\n/, $link_string);    # divide the output of the link cmd into lines

  while (defined ($str_lines[$line_cnt]))  {
    if ($str_lines[$line_cnt] =~ /ERROR:/) {
      $prev_cnt = 1;
      $brick_acquired = 0;
      $current_brick = "Unknown";                       #  Initialize the value to something
      while ($brick_acquired == 0)  {
        if ($str_lines[$line_cnt - $prev_cnt] =~ /\d\d\d[a-z]\d\d/) {        #  Is a C, R, I, P, or other brick listed?
          $current_brick = "$str_lines[$line_cnt - $prev_cnt]";
          $current_brick =~ s/:*\s//;
          $brick_acquired = 1;
        }
        else  {
          $prev_cnt++;
          if (($line_cnt - $prev_cnt) < 0)  {
            print "\nALERT! Parse error occurred within check_link_string().";
            print "\n  Unable to determine location of C Brick with failing link status.\n\n";
            $brick_acquired = 1;
          }
        }
      }

      $l1_env_err = 1;
      $l1_env_err_msg .= "  Error!  Incorrect link status on $current_brick.\n";
      $l1_env_err_msg .= "  $str_lines[$line_cnt]\n\n";

      $l1_fru_summary .= sprintf "    Alert:  Incorrect link status detected on $current_brick.\n";
      $l1_fru_summary .= sprintf "      $str_lines[$line_cnt]\n\n";
      $l1_fru_report_ready++;
      $global_err_val |= $L1_LINK_ERR;
    }
    else  {
      if ($verbose_mode == 1) {
        print "\nNo Match found in string:  $str_lines[$line_cnt].\n";
      }
    }
      $line_cnt++;
  }                     #  End of while loop
}


#-------------------------------------------------------------------------------
#
#  This function parses the output of the l1 "leds" command for err indications.
#
#  Error indications are current as of Nov 12, 2003
#  LED messages are from /data/cf/sn2-diag-sw/sn2/sn2prom/latest/ledlist
#
#-------------------------------------------------------------------------------
sub check_leds_string  {

  my $leds_string = "$_[0]";
  my $line_cnt   = 0;
  my $prev_cnt    = 0;
  my $current_brick;
  my $brick_acquired = 0;
  my $led_entry = 0;
  my $last_cpu_line = 0;
  my $led_notify = "  <<<<<<<  Not toggling!";
  my $final_led_string = "";
  my @updated_led_string = ();

  my @str_lines = split(/\n/, $leds_string);

  while (defined ($str_lines[$line_cnt]))  {
    if (   ($str_lines[$line_cnt] =~ /fail/)            # 0x82, 0x83, 0xb0, 0xb5, 0xb6, 0xb8, 0xba, 0xbc, 0xc1
        || ($str_lines[$line_cnt] =~ /MCA/)             # 0x38, 0x39, 0x3a, 0x3e, 0x56
        || ($str_lines[$line_cnt] =~ /o memory/)        # 0x80, 0x84
        || ($str_lines[$line_cnt] =~ /timed out/)       # 0x81
        || ($str_lines[$line_cnt] =~ /Too many/)        # 0x8a
        || ($str_lines[$line_cnt] =~ /initialize/)      # 0x85
        || ($str_lines[$line_cnt] =~ /duplicate rack/)  # 0x86
        || ($str_lines[$line_cnt] =~ /invalid/)         # 0xb1
        || ($str_lines[$line_cnt] =~ /overflow/)        # 0xb7, 0xbd
        || ($str_lines[$line_cnt] =~ /[Bb]ad/)          # 0xb3, 0xb9, 0xbb
        || ($str_lines[$line_cnt] =~ /No PROM segment/) # 0xbe
        || ($str_lines[$line_cnt] =~ /Out of SAL/)      # 0xc0
        #  || ($str_lines[$line_cnt] =~ /SAL_PMI/)         # 0x57
        || ($str_lines[$line_cnt] =~ /Error/)           # 0xc2
        || ($str_lines[$line_cnt] =~ /Could not write/) # 0xc3
        || ($str_lines[$line_cnt] =~ /illegal/)   )     # 0xd8
    {
      # print "\nGot a match to !!\n";
      $prev_cnt = 1;
      $brick_acquired = 0;
      $current_brick = "Unknown";                       #  Initialize the value to something
      while ($brick_acquired == 0)  {
        if ($str_lines[$line_cnt - $prev_cnt] =~ /\d\d\dc\d\d/) {               #  Is a C brick listed?
          $current_brick = "$str_lines[$line_cnt - $prev_cnt]";
          $current_brick =~ s/:*\s//;
          $brick_acquired = 1;
        }
        else  {
          $prev_cnt++;
          if (($line_cnt - $prev_cnt) < 0)  {
            print "\nALERT! Parse error occurred within check_leds_string().";
            print "\n  Unable to determine location of C Brick with failing LED status.\n\n";
            $brick_acquired = 1;
          }
        }
      }

      $l1_env_err = 1;
      $l1_env_err_msg .= "  Error!  Brick led status indicates a error condition on $current_brick.\n";
      $l1_env_err_msg .= "  $str_lines[$line_cnt]\n\n";
    }
    else  {
      if ($verbose_mode == 1) {
        print "\nNo Match to a known error message found in string:  $str_lines[$line_cnt].\n";
      }
    }

        #  check for toggling LED values on each CPU that is present

    if ($str_lines[$line_cnt] !~ /CPU not present/)  {

      if (($str_lines[$line_cnt] =~ /CPU 0[A-D]/) || ($str_lines[$line_cnt] =~ /CPU  [A-D]/))  {
        if ($verbose_mode == 1) {
	  print "\ntoggle check 0: last_cpu_line: $last_cpu_line   led_entry: $led_entry\n";
        }
        if (($last_cpu_line != 0) && ($led_entry < 2))  {
          $no_toggle_count++;
          $l1_env_err = 1;
          $updated_led_string[$last_cpu_line] .= $led_notify;
        }
        if ($str_lines[$line_cnt] =~ /slave loop/) {
            $led_entry = 2;
        }
        else {
            $led_entry = 1;
        }
        $current_brick = "$str_lines[$line_cnt - 1]";
        $current_brick =~ s/:*\s//;
        $last_cpu_line = $line_cnt;             # record the value if we need to flag it later
      }
      elsif ($str_lines[$line_cnt] =~ /CPU [1-9][A-D]/)  {
        if ($verbose_mode == 1) {
	  print "\ntoggle check 1: last_cpu_line: $last_cpu_line   led_entry: $led_entry\n";
        }
        if (($last_cpu_line != 0) && ($led_entry < 2))  {
          $no_toggle_count++;
          $l1_env_err = 1;
          $updated_led_string[$last_cpu_line] .= $led_notify;
        }
        my @line_list = split(/:/, $str_lines[$line_cnt]);
        if ($str_lines[$line_cnt] =~ /slave loop/) {
            $led_entry = 2;
        }
        else {
            $led_entry = 1;
        }
        $last_cpu_line = $line_cnt;
      }
      elsif ($str_lines[$line_cnt] =~ /0x[\da-fA-F][\da-fA-F]:/) {
        if ($verbose_mode == 1) {
	  print "\ntoggle check 2: last_cpu_line: $last_cpu_line   led_entry: $led_entry\n";
        }
        $led_entry++;
      }
    }

    $updated_led_string[$line_cnt] = sprintf "%-45s", $str_lines[$line_cnt];
    $line_cnt++;
  }                     #  End of while loop

  if ($led_entry < 2)  {
    $no_toggle_count++;
    $l1_env_err = 1;
    $updated_led_string[$last_cpu_line] .= $led_notify;
  }

  foreach $str_line (@updated_led_string)  {
    $final_led_string .= "$str_line\n";
  }

  return($final_led_string);

}

#-------------------------------------------------------------------------------
#
#  This subroutine checks the L1 log output for ALERT, ERROR, WARNING, or PANIC
#  messages.
#
#-------------------------------------------------------------------------------

sub check_log_string  {

  my $log_string = "$_[0]";
  my $line_cnt   = 0;
  my $current_brick;
  my $prev_line;
  my $log_notify = "  <<<<<<<";
  my $final_log_string = "";


  my @str_lines = split(/\n/, $log_string);     # divide the output of the log cmd into lines

  $current_brick = $str_lines[1];
  $current_brick =~ s/: //;

  while (defined ($str_lines[$line_cnt]))  {
    if ($str_lines[$line_cnt] =~ /ALERT/)  {
      $l1_env_err = 1;
      $l1_env_err_msg .= "  ALERT message found in L1 log on $current_brick.\n";
      $l1_env_err_msg .= "  $str_lines[$line_cnt].\n\n";
      $str_lines[$line_cnt] .= $log_notify;
    }
    elsif ($str_lines[$line_cnt] =~ /ERROR/)  {
      $l1_env_err = 1;
      $l1_env_err_msg .= "  ERROR message found in L1 log on $current_brick.\n";
      $l1_env_err_msg .= "  $str_lines[$line_cnt].\n\n";
      $str_lines[$line_cnt] .= $log_notify;
    }
    elsif ($str_lines[$line_cnt] =~ /WARNING/)  {
      $l1_env_err = 1;
      $l1_env_err_msg .= "  WARNING message found in L1 log on $current_brick.\n";
      $l1_env_err_msg .= "  $str_lines[$line_cnt].\n\n";
      $str_lines[$line_cnt] .= $log_notify;
    }
    elsif ($str_lines[$line_cnt] =~ /PANIC/)  {
      $l1_env_err = 1;
      $l1_env_err_msg .= "  PANIC message found in L1 log on $current_brick.\n";
      $l1_env_err_msg .= "  $str_lines[$line_cnt].\n\n";
      $str_lines[$line_cnt] .= $log_notify;
    }
    elsif ($str_lines[$line_cnt] =~ /PANEL/)  {
      $l1_env_err = 1;
      $l1_env_err_msg .= "  PANEL switch event in L1 log on $current_brick.\n";
      $l1_env_err_msg .= "  $str_lines[$line_cnt].\n\n";
      $str_lines[$line_cnt] .= $log_notify;
    }
    $line_cnt++;
  }

  foreach $str_line (@str_lines)  {
    $final_log_string .= "$str_line\n";
  }

  return($final_log_string);

}

sub initial_l1_checks
{
  sysco_open_connection($l2_host, $DEFAULT_SYSCO_PORT, $l2_copy_num);
  sysco_set_timeout(25,0);      #  default is 10 seconds, but sometimes the link command takes longer

  $l1_output_string = "";
  $l1_output_string .= "\n->Recording system Serial Number ....\n";
  $l1_output_string .= send_l1_cmd('serial', '1');

  $l1_output_string .= "\n->Recording L1 system configuration summary....\n";
  $l1_output_string .= send_l1_cmd('cfg s', '1');

  $l1_output_string .= "\n->Recording L1 system power summary....\n";
  $l1_output_string .= send_l1_cmd('pwr s', '1');

  if ($nl4r_present == 1) {
    $l1_output_string .= "\n->Recording L1 port configuration details....\n";
    $l1_output_string .= send_l1_cmd('cfg p', 'all');
  }

  if ($fast_mode == 0)  {
    $l1_output_string .= "\n->Recording IO Topology info ...\n";
    $l1_output_string .= send_l1_cmd('cfg', 'io');

    $l1_output_string .= "\n->Recording system brick power status ....\n";
    $l1_output_string .= send_l1_cmd('pwr', 'all');

  }

  $l1_output_string .= "\n->Checking L1 war status ....\n";
  $l1_output_string .= send_l1_cmd('war', 'c');

  if (($l1_war_enable == 1) && ($shub_version == 1))  {
    sleep 2;
    $l1_output_string .= "\n->Checking L1 war status again....\n";
    $l1_output_string .= send_l1_cmd('war', 'c');
  }

  if ((($l1_war_enable == 1) && ($shub_version == 1)) || ($l1_war_status == 1))  {
    $l1_output_string .= "\n->Disabling L1 war ....\n";
    $l1_output_string .= send_l1_cmd('war d', 'c');
  }
	#  For Shub2, need to extract serial info first so we know
	#  the processor configuration.

  if ($shub_version == 2)  {
   $l1_output_string .= "\n->Recording brick serial information...\n";
   my $serial_info .= send_l1_cmd('serial all', 'all_b');
   parse_serial_info($serial_info);
   $l1_output_string .= $serial_info;
  }

  $l1_output_string .= "\n->Recording Link status...\n";
  $l1_output_string .= send_l1_cmd('link', '1');

  $l1_output_string .= "\n->Recording LED status...\n";
  my $leds_target = ($shub_version == 1) ? 'c' : 'all_c';
  $l1_output_string .= send_l1_cmd('leds', $leds_target);
  if ($no_toggle_count > 0)  {
    $l1_env_err_msg .= "  Warning!  Observed LED status on $no_toggle_count cpu(s) not toggling.\n\n";
    if ($no_toggle_count < 5) {
      $l1_fru_summary .= sprintf "\tAlert:  Observed LED status on $no_toggle_count cpu(s) not toggling.\n\n";
      $l1_fru_report_ready++;
    }
  }

  $l1_output_string .= "\n->Recording VRM Power status...\n";
  $l1_output_string .= send_l1_cmd('pwr vrm', 'all');

  print "$l1_output_string\n";

  $l1_output_string = "\n";		# clear the string variable for future use

  return;
}

sub do_L1_debug  {

    print "\n\n===========================================================================\n";
    $l2_fru_output = "\n\n\n->Recording FRU information captured by the L1...\n\n";
    my $tmp_junk = send_l1_cmd('fru capture', '1');
    $l2_fru_output .= send_l1_cmd('fru examine', '1');
    my @str_lines = split(/\n/, $l2_fru_output);        # divide the output into lines
    $filtered_l2_fru_output = "";
    foreach $line_out (@str_lines)  {           # filter out error messages from non-C bricks
      if ( $line_out =~ /ERROR:/)   {  next; }
      if ( $line_out =~ />>>>/)   {  next; }
      if ( $line_out =~ /<<<</)   {  next; }
      if ( $line_out =~ /ERRORS occurred while reading FRU data/)   {  next; }
      else  {  $filtered_l2_fru_output .= "\n$line_out";  }
    }

    $filtered_l2_fru_output .= "\n\n---------------------------------------------------------------------------\n";

    $filtered_l2_fru_output .= send_l1_cmd('fru deadlock v', '1');
    $filtered_l2_fru_output .= "\n\n===========================================================================\n";
    print "\n$filtered_l2_fru_output";
}

#
#  The following function parses the L1 serial info to determine the 
#  processor configuration on each blade.
#
#  The serial info on a 1932 looks like:
#
# 001c01:
# 
# Data                            Location      Value
# ------------------------------  ------------  --------
# Local System Serial Number      NVRAM         N3333333
# Reference System Serial Number  Attached L2   N3333333
# Local Brick Serial Number       EEPROM        NRZ758
# Reference Brick Serial Number   NVRAM         NRZ758
# 
# 
# EEPROM          Product Name   Serial        Part Number          Rev T/W
# --------------- -------------- ------------- -------------------- --- ------
# B0 BLADE ID     IP63           NRN187        030_2007_003         A   00
# B1 BLADE ID     IP73D          NSH271        030_2113_001         E   00
# L1 BOARD        2U_L1          NRZ758        030_1998_001         E   00
# POWER           1932BKPL_SHUB  NRV353        030_2033_001         D   00
# PS 1            DPS-500EBE     XPD0514005871 060-0178-003         S4
# PS 2            DPS-500EBE     XPD0514006013 060-0178-003         S4
# 
# EEPROM              Part Number  Serial Number    Stepping  Speed   Cache
# ----------          -----------  ---------------- -------- -------- -------
# B0 CPU 0A  *SAMPLE* 80549KC      00031d2b82164cde    3      1.4GHz  12.0MB ***
# B1 SOCKET  *SAMPLE* 80543KC      000301c45d7328f7    1      1.6GHz  6.0MB ***
# B1 SOCKET  *SAMPLE* 80543KC      00034834b06926e2    1      1.6GHz  6.0MB ***
# 
# EEPROM     JEDEC-SPD Info           Part Number        Rev  Speed  SGI      BC
# ---------- ------------------------ ------------------ ---- ------ -------- --
# B0 DIMM0 L CE000000000000005104E666 M3 93T6450FZ0-CCC   0F    5.0  NO DATA
# B0 DIMM1 L no hardware detected
#
# (dimm serial continues)
#

#  The serial info on a IP71 looks like:

#
# Data                            Location      Value
# ------------------------------  ------------  --------
# Local System Serial Number      NVRAM         T0000227
# Reference System Serial Number  Attached L2   T0000227
# Local Brick Serial Number       EEPROM        NXP409
# Reference Brick Serial Number   NVRAM         NXP409
# 
# 
# EEPROM          Product Name   Serial        Part Number          Rev T/W
# --------------- -------------- ------------- -------------------- --- ----
# NODE            IP71           NXP409        030_2107_001          H   00
# PS              DPS-550FB A    ZZT0547001067 060-0226-003         01
# 
# CPU EEPROM          Part Number Serial Number    Step  Speed/FSB    Cache
# ----------          ----------- ---------------- ---- ------------ -------
# *SAMPLE* CPU 0A     80549KE     0000738af3bb5d2c  4   1.60GHz/534   18.0MB  ***
# *SAMPLE* CPU 0C     80549KE     0000516b38011915  4   1.60GHz/534   18.0MB  ***
# 
# EEPROM     JEDEC-SPD Info           Part Number        Rev  Speed  SGI      BC
# ---------- ------------------------ ------------------ ---- ------ -------- --
# DIMM0 L    CE0000000000000050067602 M3 93T6450FZ0-CCC   0F    5.0  CSUMERR  --
# (dimm serial info continues)
#

sub parse_serial_info {


  my $serial_string = "$_[0]"; 
  @c_blade_list = ();
  my $line_cnt = 0;
  my $prev_cnt = 0;
  my $current_brick = "Unknown";                #  Initialize the value to something
  my $b_cnt = 0;				#  Blade count
  my $initial_cnt = 0;				#  First blade count on this brick
  my $final_cnt = 0;				#  Last blade count on this brick
  my $brick_acquired = 0;
  my $brick_sn = "";
  my $blade_match = 0;
  my $blade_flag = 0;
  
  my @str_lines = split(/\n/, $serial_string);

  print "\n  Serial string is : \n$serial_string\n\n" if ($verbose_mode == 1);

  while (defined ($str_lines[$line_cnt]))  {
    if ($str_lines[$line_cnt] =~ /^\d\d\dc\d\d/) {               #  Is a C brick listed?
      $current_brick = "$str_lines[$line_cnt]";
      $current_brick =~ s/:*\s//;
      $initial_cnt = $b_cnt;
      $brick_acquired = 1;
    }
    if ($str_lines[$line_cnt] =~ /Local Brick Serial Number/) {            #  Capture IRU SN
      my @line_list = split(/\s+/, $str_lines[$line_cnt]);
      $brick_sn = $line_list[5];
    }
    if (($str_lines[$line_cnt] =~ /^NODE\s+IP71/) && ($brick_acquired == 1)) {           #  Capture Blade location and Board Type
      my @line_list = split(/\s+/, $str_lines[$line_cnt]);
      printf ("\n  Blade type is: %s \n", $line_list[1]) if ($verbose_mode == 1);
      $c_blade_list[$b_cnt]->{INDEX} = $b_cnt;
      $c_blade_list[$b_cnt]->{IRU} = $current_brick;
      $c_blade_list[$b_cnt]->{IRU_SN} = $brick_sn;
      $c_blade_list[$b_cnt]->{PROC_CNT} = 0;
      $c_blade_list[$b_cnt]->{BLADE_LOC} = $current_brick;
      $c_blade_list[$b_cnt]->{BLADE_TYPE} = $line_list[1];
      $c_blade_list[$b_cnt]->{BLADE_SN} = $line_list[2];
      $c_blade_list[$b_cnt]->{BLADE_PN} = $line_list[3];
      $c_blade_list[$b_cnt]->{BLADE_REV} = $line_list[4];
      $b_cnt++;
      $blade_flag = 2;
    }
    if (($str_lines[$line_cnt] =~ /^B[0-9] BLADE ID/) && ($brick_acquired == 1)) {               #  Capture Blade location and Board Type
      my @line_list = split(/\s+/, $str_lines[$line_cnt]);
      printf ("\n  Blade location is: %s    Blade type is: %s \n", $line_list[0], $line_list[3]) if ($verbose_mode == 1);
      if ($line_list[3] =~ /^IP[678][1359]/)  {			# IP63, IP73, IP75, IP79, IP81
        $c_blade_list[$b_cnt]->{INDEX} = $b_cnt;
        $c_blade_list[$b_cnt]->{IRU} = $current_brick;
        $c_blade_list[$b_cnt]->{IRU_SN} = $brick_sn;
        $c_blade_list[$b_cnt]->{PROC_CNT} = 0;
        $c_blade_list[$b_cnt]->{BLADE_LOC} = $line_list[0];
        $c_blade_list[$b_cnt]->{BLADE_TYPE} = $line_list[3];
        $c_blade_list[$b_cnt]->{BLADE_SN} = $line_list[4];
        $c_blade_list[$b_cnt]->{BLADE_PN} = $line_list[5];
        $c_blade_list[$b_cnt]->{BLADE_REV} = $line_list[6];
        $b_cnt++;
  	$blade_flag = 1;
      }
    }
      #  At the end of the Blade list - set blade count to original
    if (((($str_lines[$line_cnt] =~ /Step/) && ($str_lines[$line_cnt] =~ /CPU/)) 
	 || ($str_lines[$line_cnt] =~ /Stepping/))
	 && ($brick_acquired == 1))  {      
      $final_cnt = $b_cnt;
      $b_cnt = $initial_cnt;
    }
	#  Capture proc info for Tollhouse IRU
    if ((($str_lines[$line_cnt] =~ /B[0-9] CPU/) || ($str_lines[$line_cnt] =~ /B[0-9] SOCKET/)) 
       && ($brick_acquired == 1)  && ($blade_flag == 1)
       && ($str_lines[$line_cnt] !~ /no hardware detected/) 
       && ($str_lines[$line_cnt] !~ /not present/) ) {    
      my @line_list = split(/\s+/, $str_lines[$line_cnt]);
      $blade_match = 0;
      while ($blade_match == 0) {
	# if ($line_list[0] =~ /$c_blade_list[$b_cnt]->{BLADE_LOC}/)
	if (($str_lines[$line_cnt] =~ /$c_blade_list[$b_cnt]->{BLADE_LOC} CPU/) ||
	    ($str_lines[$line_cnt] =~ /$c_blade_list[$b_cnt]->{BLADE_LOC} SOCKET/)) {
	  $blade_match = 1;
	}
	else {
	  $b_cnt++;
	  if ($b_cnt > $final_cnt) {           #  prevent endless loop if serial info changes
	    $blade_match = 2;
	    print "\n\nERROR!! Unable to find a blade match in the Blade CPU serial info.\n\n";
	  }
	}
      }
      $sh2_cpu_count++;
      $c_blade_list[$b_cnt]->{PROC_CNT}++;
      if ($c_blade_list[$b_cnt]->{PROC_CNT} == 1) {
	$c_blade_list[$b_cnt]->{PROC_TYPE} = get_sh2_proc_type($str_lines[$line_cnt]);
      }
      else {
	my $socket_1_cpu = get_sh2_proc_type($str_lines[$line_cnt]);
	if (($socket_1_cpu ne $c_blade_list[$b_cnt]->{PROC_TYPE}) || ($blade_match == 2)) {
	  $c_blade_list[$b_cnt]->{CFG_ERR} = 1;
	}
      }
    }

	#  Capture proc info for IP71 blade
    if ( ($str_lines[$line_cnt] =~ /CPU 0/) && ($brick_acquired == 1) && ($blade_flag == 2)
       && ($str_lines[$line_cnt] !~ /no hardware detected/) 
       && ($str_lines[$line_cnt] !~ /not present/) ) {    
      $sh2_cpu_count++;
      $c_blade_list[$b_cnt]->{PROC_CNT}++;
      if ($c_blade_list[$b_cnt]->{PROC_CNT} == 1) {
	$c_blade_list[$b_cnt]->{PROC_TYPE} = get_sh2_proc_type($str_lines[$line_cnt]);
      }
      else {
	my $socket_1_cpu = get_sh2_proc_type($str_lines[$line_cnt]);
	if ($socket_1_cpu ne $c_blade_list[$b_cnt]->{PROC_TYPE})  {
	  $c_blade_list[$b_cnt]->{CFG_ERR} = 1;
	}
      }
    }

    if ($str_lines[$line_cnt] =~ /JEDEC/) {     #  In the DIMM info area - done parsing this IRU
      $b_cnt = $final_cnt;
      $brick_acquired = 0;
    }  

    $line_cnt++;
  }                     #  End of while loop

  print "\n    Compute Blade Summary\n";
  foreach $cbld (@c_blade_list) {
    printf "\n  IRU: %s  Blade Loc.: %s  Type: %-9s  CPU Type: %-29s  CPU count: %d",
    $cbld->{IRU}, $cbld->{BLADE_LOC}, $cbld->{BLADE_TYPE}, $cbld->{PROC_TYPE}, $cbld->{PROC_CNT};
    if (defined $cbld->{CFG_ERR}) {
      print "  ** CFG_ERR **";
    }
  }
  print "\n\n    Total CPU count:  $sh2_cpu_count\n";

  if (($sh2_cpu_count == 0) && ($file_mode == 0))  {		#  $file_mode is set in errdmp
    print "\nALERT!!  Unable to read system serial information.\n";
    print "  If L2 password is enabled, disable the L2 password using \"pw clear\" and re-run errdmp.\n\n";
  }
  elsif (($sh2_cpu_count == 0) && ($file_mode == 1))  {
    print "\n  ALERT!!  Unable to read system serial information from the provided fru file. \n";
    print "  The integrity of the fru file may be in doubt. \n";
  }

}


sub get_sh2_proc_type {

  my $proc_str = "$_[0]";
  my $return_str = ($proc_str =~ /80549KC/) ? 'MONTECITO 400MHz/1.4GHz' :
                   ($proc_str =~ /80549KE/) ? 'MONTECITO 533MHz/1.6GHz' :
                   ($proc_str =~ /80549KF/) ? 'MONTECITO 667 MHz FSB' :
  		   ($proc_str =~ /80567KC/) ? 'MONTVALE 400MHz/1.4GHz' :
                   ($proc_str =~ /80567KE/) ? 'MONTVALE 533MHz/1.6GHz' :
                   ($proc_str =~ /80567KF/) ? 'MONTVALE 667 MHz FSB' :
                   ($proc_str =~ /80543KC/) ? 'MADISON 400MHz/1.6GHz' :
                   ($proc_str =~ /80543KE/) ? 'MADISON 533MHz/1.6GHz' :
                   ($proc_str =~ /80543KF/) ? 'MADISON 667MHz/1.6GHz' : 'UNKNOWN';

  if ($proc_str =~ /(\d+\.\d+MB)/) {
    my $cache_sz = $1;
    $cache_sz =~ s/\.0MB/MB/;
    $return_str .= "/" . $cache_sz;
  }
  return($return_str);
}

#---------------------------------------------------------------
#
#  Check the status of the L1 war for Shub2 Bug 951089
#
#---------------------------------------------------------------

sub check_war_string {

  my $war_string = "$_[0]";
  my @str_lines = split(/\n/, $war_string);    # divide the output of the war cmd into lines
  my $str_line = (); 

  foreach $str_line (@str_lines) {
    if ($str_line =~ /SHUB2 WAR enabled and active/) {
      $l1_war_status = 1;
      last;
    }
  }
}


#-------------------------------------------------------------------------------
#
#  This function parses the output of the l1 "pwr vrm" command for 
#  FAILED indications.
#
#-------------------------------------------------------------------------------
sub check_vrm_string  {

  my $vrm_string = "$_[0]";
  my $line_cnt   = 0;
  my $prev_cnt    = 0;
  my $current_brick;
  my $brick_acquired = 0;
  my $vrm_notify = "     <<<<<<< ";
  my $final_vrm_string = "";

  my @str_lines = split(/\n/, $vrm_string);

  while (defined $str_lines[$line_cnt]) {
    if ($str_lines[$line_cnt] =~ /FAILED/)     
    {
      $prev_cnt = 1;
      $brick_acquired = 0;
      $current_brick = "Unknown";                       #  Initialize the value to something
      while ($brick_acquired == 0)  {
        if ($str_lines[$line_cnt - $prev_cnt] =~ /\d\d\dc\d\d/) {               #  Is a C brick listed?
          $current_brick = "$str_lines[$line_cnt - $prev_cnt]";
          $current_brick =~ s/:*\s//;
          $brick_acquired = 1;
        }
        else  {
          $prev_cnt++;
          if (($line_cnt - $prev_cnt) < 0)  {
            print "\nALERT! Parse error occurred within check_vrm_string().";
            print "\n  Unable to determine location of C Brick with failing VRM status.\n\n";
            $brick_acquired = 1;
          }
        }
      }
 
      my @part_des = split(/    /, $str_lines[$line_cnt]);

      $l1_env_err = 1;
      $l1_env_err_msg .= sprintf "  Error!  Power VRM status indicates a error condition with %s in %s.\n",
        $part_des[0], $current_brick;
      $l1_env_err_msg .= "  $str_lines[$line_cnt]\n\n";
      $final_vrm_string .= sprintf "%-45s  %s\n", $str_lines[$line_cnt], $vrm_notify;
      $l1_fru_summary .= sprintf "\n    Alert:  Incorrect VRM status detected with %s on %s.\n", 
        $part_des[0], $current_brick;
      $l1_fru_summary .= sprintf "      $str_lines[$line_cnt]\n\n";
      $l1_fru_report_ready++;
    }
    else  {
      $final_vrm_string .= sprintf "%-45s\n", $str_lines[$line_cnt];
      if ($verbose_mode == 1) {
        print "\nNo Match to a known error message found in string:  $str_lines[$line_cnt].\n";
      }
    }
    $line_cnt++;
  }                     #  End of while loop

  $final_vrm_string .= "\n";
  return($final_vrm_string);

}

#-------------------------------------------------------------------------------


1;
