#
#/**************************************************************************
# *                                                                        *
# *               Copyright (C) 1989-1999 Silicon Graphics, Inc.           *
# *                                                                        *
# *  These coded instructions, statements, and computer programs  contain  *
# *  unpublished  proprietary  information of Silicon Graphics, Inc., and  *
# *  are protected by Federal copyright law.  They  may  not be disclosed  *
# *  to  third  parties  or copied or duplicated in any form, in whole or  *
# *  in part, without the prior written consent of Silicon Graphics, Inc.  *
# *                                                                        *
# **************************************************************************/
#
# @(#)$Id: sysco_subs.pm,v 1.9 2004/05/14 14:06:00 beighley Exp $


#-------------------------------------------------------------------------------
#
#  FILE: sysco_subs.pm
#
#  DESCRIPTION:  This file contains PERL subroutines to connect to a L2/L1
#     system controller and issue commands.
#
#  NOTES: The "common_defs.pm" file and "SysAccess" module are required when 
#     this file is used.
#
#  SUBROUTINES:  The following subroutines are contained in this file:
#     (expected parameter usage displayed; some have optional parameters)
#     sysco_open_connection($sysco_host)             
#     sysco_open_connection($sysco_host, $sysco_port)
#     sysco_close_connection()
#     sysco_set_timeout($num_sec, $num_usec);
#     sysco_set_wrap_flag($flag);
#     $text = sysco_cmd($command)
#     $text = sysco_cmd($command, \@bricks)
#     $text = sysco_raw_cmd($command)
#     $text = sysco_raw_cmd($command, $flush)
#     $num  = sysco_cfg(\@bricks)
#     $text = sysco_reset($TRUE|$FALSE)
#     $text = sysco_reset($TRUE|$FALSE, \@bricks)
#     $text = sysco_pwr($SYSCO_PWR_UP|$SYSCO_PWR_DOWN)
#     $text = sysco_pwr($SYSCO_PWR_UP|$SYSCO_PWR_DOWN, \@bricks)
#     $text = sysco_mgn($margin);
#     $text = sysco_mgn($margin, \@bricks)
#     $flag = sysco_is_pwr_on()
#     $flag = sysco_is_pwr_on(\@bricks)
#     $flag = sysco_is_pwr_off()
#     $flag = sysco_is_pwr_off(\@bricks)
#     $val  = sysco_get_ioexp($rack, $slot, $ioexp_addr)
#     $text = sysco_wait_for_cmd($WAIT_CMD_SYSCO|$WAIT_CMD_CONSOLE, $nsec)
#
#  MODIFICATION HISTORY:
#
#     WHEN      WHO             WHAT
#     --------  --------------  -----------------------------------------
#     10/29/01  Karen Beighley  Original source code.
#     03/22/02  Karen Beighley  Added subroutine: sysco_raw_cmd().
#     04/03/02  Karen Beighley  Added use of common error function pointer.
#     10/28/02  Karen Beighley  Added "flush" parameter: sysco_raw_cmd().
#                               Added subroutine: sysco_set_timeout()
#     03/07/03  Karen Beighley  Added subroutine: sysco_set_wrap_flag().
#     03/25/04  Karen Beighley  Added subroutine: sysco_get_ioexp().
#     05/14/04  Karen Beighley  Minor fix to subroutine: sysco_get_ioexp().
#
#-------------------------------------------------------------------------------

use SysAccess;    # JTAGAccess library built from C source code

#-------------------------------------------------------------------------------
#  CONSTANTS - The following values *cannot* change!
#-------------------------------------------------------------------------------

$SYSCO_PWR_UP=0;
$SYSCO_PWR_DOWN=1;

$SYSCO_MGN_DEFAULT=0;
$SYSCO_MGN_NORMAL=1;
$SYSCO_MGN_LOW=2;
$SYSCO_MGN_HIGH=3;

$WAIT_CMD_SYSCO=0;
$WAIT_CMD_CONSOLE=1;

$CHAR_CTRL_D=0x04;
$CHAR_CTRL_T=0x14;

$DEFAULT_SYSCO_PORT = 8001;
$SYSCO_COPY_INVALID = 0;

#-------------------------------------------------------------------------------
#  GLOBAL VARIABLES
#-------------------------------------------------------------------------------

$sysco_init = $FALSE;
$sysco_host = "localhost";
$sysco_port = $DEFAULT_SYSCO_PORT;
$sysco_parm = $SYSCO_COPY_INVALID;


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: sysco_open_connection()
#
#    USAGE: sysco_open_connection($sysco_host)             
#           sysco_open_connection($sysco_host, $sysco_port)
#           sysco_open_connection($sysco_host, $sysco_port, $sysco_parm)
#
#    DESCRIPTION: Open a connection to the specified system controller. 
#       The port is assumed to be 8001 if not passed as a parameter.
#       The option "sysco_parm" is used for the system controller access,
#       and contains the value associated with "--copy".
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub sysco_open_connection
{
   my $num = $#_ + 1;
   print "\nsysco_open_connection(@_) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   sysco_close_connection()  if ( $sysco_init == $TRUE );

   if ( $num == 3 )
   {  $sysco_port = defined($_[1]) ? $_[1] : $DEFAULT_SYSCO_PORT;
      $sysco_parm = defined($_[2]) ? $_[2] : $SYSCO_COPY_INVALID;
   }
   elsif ( $num == 2 )
   {  $sysco_port = defined($_[1]) ? $_[1] : $DEFAULT_SYSCO_PORT;
      $sysco_parm = $SYSCO_COPY_INVALID;
   }
   elsif ( $num != 1 )
   {  $emess = "Invalid parameters passed to subroutine sysco_open_connection().";
      &$comm_err_fun($emess);
      return;
   }
   $sysco_host = $_[0];
   $sysco_port += ($sysco_parm * 10);

   printf "\nOpening connection to system controller: host=$sysco_host, port=$sysco_port\n";

   pcf_open_connection($sysco_host, $sysco_port, $ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess);
      return;
   }

   printf "\tSuccessful connection and initialization with system controller.\n";

   $sysco_init = $TRUE;
   return;
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: sysco_close_connection()
#
#    USAGE: sysco_close_connection()
#
#    DESCRIPTION: Close a connection to the system controller (pre-opened). 
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub sysco_close_connection
{
   print "\nsysco_close_connection(@_)\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $sysco_init == $TRUE )
   {  pcf_close_connection();
      printf "\nClosing connection to system controller: host=$sysco_host, port=$sysco_port\n";
   }
   $sysco_init = $FALSE;
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: sysco_set_timeout()
#
#    USAGE: sysco_set_timeout($num_sec, $num_usec)
#
#    DESCRIPTION: Override the timeout internal for system controller commands.
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub sysco_set_timeout
{
   print "\nsysco_set_timeout(@_)\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $#_ != 1 )
   {  $emess = "Invalid number of parameters passed to subroutine sysco_set_timeout().";
      &$comm_err_fun($emess);
      return;
   }

   pcf_sysco_set_timeout($_[0], $_[1]);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: sysco_set_wrap_flag()
#
#    USAGE: sysco_set_wrap_flag($flag)
#
#    DESCRIPTION: Set the wrap flag for the read buffer: $TRUE means the input
#       read buffer will wrap around after the maximum characters are read.
#       $FALSE will cause an error to be returned when the read buffer is full.
#       The default value of the wrap flag is $TRUE.
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub sysco_set_wrap_flag
{
   print "\nsysco_set_wrap_flag(@_)\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $#_ != 0 )
   {  $emess = "Invalid number of parameters passed to subroutine sysco_set_wrap_flag().";
      &$comm_err_fun($emess);
      return;
   }

   pcf_sysco_set_wrap_flag($_[0]);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: sysco_cmd()
#
#    USAGE: $text = sysco_cmd($command)
#           $text = sysco_cmd($command, \@bricks)
#
#    DESCRIPTION: Issue a user-specified system controller command.  The
#       command will be issued to all bricks unless an array of brick
#       names is provided (this array is an optional parameter).
#       This subroutine will move to a system controller prompt before
#       issuing the specified command.
#
#    RETURN VALUE: A string containing the response text from the system
#       controller is returned.
#
#-------------------------------------------------------------------------------

sub sysco_cmd
{  
   my $num = $#_ + 1;
   print "\nsysco_cmd(@_) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   return($_[0]) if ( $sysco_init == $FALSE );

   $ecode = 0;
   $emess = "";

   my $av_ptr = 0;
   if ( $num == 2 )
   {  $av_ptr = $_[1];
      if ( ref($av_ptr) ne 'ARRAY' )
      {  $emess = "Invalid parameter passed to subroutine sysco_cmd().";
         $emess .= "\nSecond parameter must be an ARRAY reference.";
         &$comm_err_fun($emess);
         return;
      }
   }
   elsif ( $num != 1 )
   {  $emess = "Invalid number of parameters passed to subroutine sysco_cmd().";
      &$comm_err_fun($emess);
      return;
   }
   my $cmd_text = $_[0];

   my $rv = pcf_sysco_cmd($cmd_text, $av_ptr, $ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess, $rv);
      return;
   }
   return($rv);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: sysco_raw_cmd()
#
#    USAGE: $text = sysco_raw_cmd($command)
#           $text = sysco_raw_cmd($command, $flush)
#
#    DESCRIPTION: Issue a user-specified command.  This function does not
#       check the prompt state before issuing the command or the response data
#       (like "sysco_cmd()").  The optional "flush" parameter determines if
#       the response socket is cleared before the command is issued.  The
#       possible values for "flush" are $TRUE or $FALSE, and the default
#       is $TRUE (if not specified).  If $TRUE is specified for "flush", then
#       any waiting response data will be discarded before the comand is issued.
#
#    RETURN VALUE: A string containing the response text from the system
#       controller is returned.
#
#-------------------------------------------------------------------------------

sub sysco_raw_cmd
{  
   my $num = $#_ + 1;
   print "\nsysco_raw_cmd(@_) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   $ecode = 0;
   $emess = "";

   my $fflag = $TRUE;
   if ( $num == 2 )
   {  $fflag = $_[1];
   }
   elsif ( $num != 1 )
   {  $emess = "Invalid number of parameters passed to subroutine sysco_raw_cmd().";
      &$comm_err_fun($emess);
      return;
   }
   my $cmd_text = $_[0];

   my $rv = pcf_sysco_raw_cmd($cmd_text, $fflag, $ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess, $rv);
      return;
   }

   return($rv);
}

#-------------------------------------------------------------------------------
#
#    SUBROUTINE: sysco_cfg()
#
#    USAGE: $num = sysco_cfg(\@bricks)
#
#    DESCRIPTION: Read the system configuration from the system controller.
#
#    RETURN VALUE: The number of bricks detected by the system controller.
#       The list of brick names is placed in the "bricks" array.
#
#-------------------------------------------------------------------------------

sub sysco_cfg
{  
   my $num = $#_ + 1;
   print "\nsysco_cfg(@_) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   $ecode = 0;
   $emess = "";

   if ( $num != 1 )
   {  $emess = "Invalid number of parameters passed to subroutine sysco_cfg().";
      &$comm_err_fun($emess);
      return;
   }
   my $av_ptr = $_[0];
   if ( ref($av_ptr) ne 'ARRAY' )
   {  $emess = "Invalid parameter passed to subroutine sysco_cfg().";
      $emess .= "\nThe parameter must be an ARRAY reference.";
      &$comm_err_fun($emess);
      return;
   }

   my $rv = pcf_sysco_cfg($av_ptr, $ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess);
      return;
   }

   return($rv);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: sysco_reset()
#
#    USAGE: $text = sysco_reset($TRUE|$FALSE)
#           $text = sysco_reset($TRUE|$FALSE, \@bricks)
#
#    DESCRIPTION: Issue a "reset" to the system via the system controller. 
#       The first parameter is a flag:
#          $TRUE=wait for reset to complete (maintenance menu prompt)
#          $FALSE=issue reset and return; do not check progress
#       The command will be issued to all bricks unless an array of brick
#       names is provided (this array is an optional parameter).
#
#    RETURN VALUE: A string containing the response text from the system
#       controller is returned.
#
#-------------------------------------------------------------------------------

sub sysco_reset
{  
   my $num = $#_ + 1;
   print "\nsysco_reset(@_) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   $ecode = 0;
   $emess = "";

   my $av_ptr = 0;

   if ( $num == 2 )
   {  $av_ptr = $_[1];
      if ( ref($av_ptr) ne 'ARRAY' )
      {  $emess = "Invalid parameter passed to subroutine sysco_reset().";
         $emess .= "\nThe second parameter must be an ARRAY reference.";
         &$comm_err_fun($emess);
         return;
      }
   }
   elsif ( $num != 1 )
   {  $emess = "Invalid number of parameters passed to subroutine sysco_reset().";
      &$comm_err_fun($emess);
      return;
   }
   my $cmd_flag = $_[0];

   my $rv = pcf_sysco_reset($cmd_flag, $av_ptr, $ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess, $rv);
      return;
   }
   return($rv);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: sysco_pwr()
#
#    USAGE: $text = sysco_pwr($SYSCO_PWR_UP|$SYSCO_PWR_DOWN)
#           $text = sysco_pwr($SYSCO_PWR_UP|$SYSCO_PWR_DOWN, \@bricks)
#
#    DESCRIPTION: Issue a power command to the system controller. The
#       command will be issued to all bricks unless an array of brick
#       names is provided (this array is an optional parameter).
#
#    RETURN VALUE: A string containing the response text from the system
#       controller is returned.
#
#-------------------------------------------------------------------------------

sub sysco_pwr
{  
   my $num = $#_ + 1;
   print "\nsysco_pwr(@_) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   $ecode = 0;
   $emess = "";

   my $av_ptr = 0;
   if ( $num == 2 )
   {  $av_ptr = $_[1];
      if ( ref($av_ptr) ne 'ARRAY' )
      {  $emess = "Invalid parameter passed to subroutine sysco_pwr().";
         $emess .= "\nSecond parameter must be an ARRAY reference.";
         &$comm_err_fun($emess);
         return;
      }
   }
   elsif ( $num != 1 )
   {  $emess = "Invalid number of parameters passed to subroutine sysco_pwr().";
      &$comm_err_fun($emess);
      return;
   }
   my $cmd_opt = $_[0];
   if ( ($cmd_opt != $SYSCO_PWR_UP) && ($cmd_opt != $SYSCO_PWR_DOWN) )
   {  $emess = "Invalid power option parameter passed to subroutine sysco_pwr().";
      &$comm_err_fun($emess);
      return;
   }

   my $rv = pcf_sysco_pwr($cmd_opt, $av_ptr, $ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess, $rv);
      return;
   }
   return($rv);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: sysco_mgn()
#
#    USAGE: $text = sysco_mgn($margin);
#           $text = sysco_mgn($margin, \@bricks)
#
#           $margin is one of the following: $SYSCO_MGN_DEFAULT, 
#              $SYSCO_MGN_NORMAL, $SYSCO_MGN_LOW, or $SYSCO_MGN_HIGH
#
#    EXAMPLE: sysco_mgn($SYSCO_MGN_HIGH);
#
#    DESCRIPTION: Issue a margin command to the system controller. The
#       command will be issued to all bricks unless an array of brick
#       names is provided (this array is an optional parameter).
#
#    RETURN VALUE: A string containing the response text from the system
#       controller is returned.
#
#-------------------------------------------------------------------------------

sub sysco_mgn
{  
   my $num = $#_ + 1;
   print "\nsysco_mgn(@_) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   $ecode = 0;
   $emess = "";

   my $av_ptr = 0;
   if ( $num == 2 )
   {  $av_ptr = $_[1];
      if ( ref($av_ptr) ne 'ARRAY' )
      {  $emess = "Invalid parameter passed to subroutine sysco_mgn().";
         $emess .= "\nSecond parameter must be an ARRAY reference.";
         &$comm_err_fun($emess);
         return;
      }
   }
   elsif ( $num != 1 )
   {  $emess = "Invalid number of parameters passed to subroutine sysco_mgn().";
      &$comm_err_fun($emess);
      return;
   }
   my $cmd_opt = $_[0];
   if ( ($cmd_opt != $SYSCO_MGN_DEFAULT) && ($cmd_opt != $SYSCO_MGN_NORMAL) &&
        ($cmd_opt != $SYSCO_MGN_LOW) && ($cmd_opt != $SYSCO_MGN_HIGH) )
   {  $emess = "Invalid margin option parameter passed to subroutine sysco_mgn().";
      &$comm_err_fun($emess);
      return;
   }

   my $rv = pcf_sysco_mgn($cmd_opt, $av_ptr, $ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess, $rv);
      return;
   }
   return($rv);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: sysco_is_pwr_on()
#
#    USAGE: $flag = sysco_is_pwr_on()
#           $flag = sysco_is_pwr_on(\@bricks)
#
#    DESCRIPTION: Read the power status of the system to determine if the  
#      power is on.  All bricks in the system are checked, unless a specific
#      subset of bricks is targets with a parameter.
#
#    RETURN VALUE: $TRUE if the power is on for all targeted bricks.  
#                  $FALSE if the power is off for one or more targets bricks.
#
#-------------------------------------------------------------------------------

sub sysco_is_pwr_on
{  
   my $num = $#_ + 1;
   print "\nsysco_is_pwr_on(@_) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   $ecode = 0;
   $emess = "";

   my $av_ptr = 0;
   my $n_on   = 0;
   my $n_off  = 0;

   if ( $num == 1 )
   {  $av_ptr = $_[0];
      if ( ref($av_ptr) ne 'ARRAY' )
      {  $emess = "Invalid parameter passed to subroutine sysco_is_pwr_on().";
         $emess .= "\nThe parameter must be an ARRAY reference.";
         &$comm_err_fun($emess);
         return;
      }
   }
   elsif ( $num != 0 )
   {  $emess = "Invalid number of parameters passed to subroutine sysco_is_pwr_on().";
      &$comm_err_fun($emess);
      return;
   }

   my $rv = pcf_sysco_read_pwr($av_ptr, $n_on, $n_off, $ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess);
      return;
   }
   printf "done: on=$n_on, off=$n_off\n" if ($debug_flag & $DEBUG_GENERAL);

   return ($TRUE) if ( ($n_on > 0) && ($n_off == 0) );
   return($FALSE);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: sysco_is_pwr_off()
#
#    USAGE: $flag = sysco_is_pwr_off()
#           $flag = sysco_is_pwr_off(\@bricks)
#
#    DESCRIPTION: Read the power status of the system to determine if the  
#      power is off.  All bricks in the system are checked, unless a specific
#      subset of bricks is targets with a parameter.
#
#    RETURN VALUE: $TRUE if the power is off for all targeted bricks.  
#                  $FALSE if the power is on for one or more targets bricks.
#
#-------------------------------------------------------------------------------

sub sysco_is_pwr_off
{  
   my $num = $#_ + 1;
   print "\nsysco_is_pwr_off(@_) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   $ecode = 0;
   $emess = "";

   my $av_ptr = 0;
   my $n_on   = 0;
   my $n_off  = 0;

   if ( $num == 1 )
   {  $av_ptr = $_[0];
      if ( ref($av_ptr) ne 'ARRAY' )
      {  $emess = "Invalid parameter passed to subroutine sysco_is_pwr_off().";
         $emess .= "\nThe parameter must be an ARRAY reference.";
         &$comm_err_fun($emess);
         return;
      }
   }
   elsif ( $num != 0 )
   {  $emess = "Invalid number of parameters passed to subroutine sysco_is_pwr_off().";
      &$comm_err_fun($emess);
      return;
   }

   my $rv = pcf_sysco_read_pwr($av_ptr, $n_on, $n_off, $ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess);
      return;
   }
   printf "done: on=$n_on, off=$n_off\n" if ($debug_flag & $DEBUG_GENERAL);

   return ($TRUE) if ( ($n_off > 0) && ($n_on == 0) );
   return($FALSE);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: sysco_get_ioexp()
#
#    USAGE: $val = sysco_get_ioexp($rack, $slot, $ioexp_addr)
#
#    DESCRIPTION: The subroutine get the value of a specific I/O expander.
#
#    RETURN VALUE: Value of I/O expander, or undefined (if error).
#
#-------------------------------------------------------------------------------

sub sysco_get_ioexp
{
   my $num = $#_ + 1;
   print "\nsysco_get_ioexp(@_) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);
   
   $ecode = 0;
   $emess = "";

   if ( $num != 3 )
   {  $emess = "Invalid number of parameters passed to subroutine sysco_get_ioexp().";
      &$comm_err_fun($emess);
      return;
   }
   my $rack = $_[0]; # Rack number of brick
   my $slot = $_[1]; # Slot number of brick
   my $addr = $_[2]; # Address of I/O expander
   my ($cmd, $text, $i, $found, $val);

   $cmd  = "$rack.$slot tst ioexp get $addr";
   $text = sysco_cmd($cmd);

   my @parms = split(' ', $text);
   for ($i = 0, $found = $FALSE; (($i < $#parms) && !$found); $i++)
   {  $found = ( ($parms[$i] =~ /^Expander:/i) ? $TRUE : $FALSE );
   }
   return if ( !$found );

   $val = hex($parms[$i]);
   return($val);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: sysco_wait_for_cmd()
#
#    USAGE: $text = sysco_wait_for_cmd($WAIT_CMD_SYSCO|$WAIT_CMD_CONSOLE, $nsec)
#
#    DESCRIPTION: The subroutine will poll for response data from either
#      the system controller or the console.  The actual read function uses
#      the seconds specified as a timeout for data, and will loop until no 
#      data is available for a timeout interval (so the actual "wait" could
#      be some number of $nsec intervals.
#
#    RETURN VALUE: String containing response text from system.
#
#-------------------------------------------------------------------------------

sub sysco_wait_for_cmd
{
   my $num = $#_ + 1;
   print "\nsysco_wait_for_cmd(@_) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   $ecode = 0;
   $emess = "";

   if ( $num != 2 )
   {  $emess = "Invalid number of parameters passed to subroutine sysco_wait_for_cmd().";
      &$comm_err_fun($emess);
      return;
   }
   my $cmd_opt = $_[0];
   my $cmd_sec = $_[1];
   if ( ($cmd_opt != $WAIT_CMD_SYSCO) && ($cmd_opt != $WAIT_CMD_CONSOLE) )
   {  $emess = "Invalid option parameter passed to subroutine sysco_wait_for_cmd().";
      &$comm_err_fun($emess);
      return;
   }

   my $rv = pcf_sysco_wait_for_cmd($cmd_opt, $cmd_sec, $ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess, $rv);
      return;
   }

   return($rv);
}



#-------------------------------------------------------------------------------

1;
