#
#/**************************************************************************
# *                                                                        *
# *               Copyright (C) 1989-1999 Silicon Graphics, Inc.           *
# *                                                                        *
# *  These coded instructions, statements, and computer programs  contain  *
# *  unpublished  proprietary  information of Silicon Graphics, Inc., and  *
# *  are protected by Federal copyright law.  They  may  not be disclosed  *
# *  to  third  parties  or copied or duplicated in any form, in whole or  *
# *  in part, without the prior written consent of Silicon Graphics, Inc.  *
# *                                                                        *
# **************************************************************************/
#
# @(#)$Id: tdraccess.pm,v 1.12 2005/07/14 17:10:31 beighley Exp $
#-------------------------------------------------------------------------------
#
#  FILE: tdraccess.pm
#
#  DESCRIPTION:  This file contains PERL subroutines to create a JTAG access
#     window for a user to manipulate instruction and test data registers.
#
#  MODIFICATION HISTORY:
#
#     WHEN      WHO             WHAT
#     --------  --------------  -----------------------------------------
#     02/11/04  Karen Beighley  Added modification history to perl source.
#                               Added module scan chain index for resets.
#     03/04/04  Karen Beighley  Modified selection/display for instruction(s).
#     ...
#     07/13/05  Karen Beighley  Changed DR callback to use L1 command that
#                                  combines IR and DR scan operations.
#
#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
#  CONSTANTS - These functions should not change
#-------------------------------------------------------------------------------

my $dsp_cb_label = "Display individual register fields";
my $fld_font = '-*-helvetica-medium-r-normal--*-80-*-*-*-*-*-*';


################################################################################
#
#  The JTAG access data (JAD) structure contains the entries:
#  <frame_TDR> <frame_flds> <instruction_val> <display_val> <fld0 text> ... <fldN text>

$JAD_FRAME_TDR   = 0; # frame for TDR defintion
$JAD_FRAME_FLDS  = 1; # frame for field definition(s)
$JAD_VAL_OPNAME  = 2; # variable for instruction option menu
$JAD_VAL_DISPLAY = 3; # variable for field display checkbox
$JAD_TEXT_TDR    = 4; # text variable for TDR entry widget
$JAD_TEXT_FLD0   = 5; # text variable for field 0 entry widget

#  The JAD array is an array containing the JAD references for a JTAG TDR
#     access window.  There is one entry for each ASIC in the window.
#  ( <jad ref for ASIC 0>, ..., <jad ref for ASIC N> )
#
#  The JAD lookup is a hash table containing the reference to the JAD array
#     for each JTAG TDR access window created.
#     key   = module name
#     {key} = reference to JAD array
%jad_lookup  = ();

################################################################################
#
#  SUBROUTINES TO SUPPORT THE JTAG TDR ACCESS GUI
#
################################################################################

sub create_tdr_access_window
{  my $num = $#_ + 1;

   printf "\ncreate_tdr_access_window( @_ ) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 4 )
   {  printf "\nERROR - Invalid parameters passed to function!\n";
      return($FALSE);
   }
   my $MW        = $_[0]; # top-level window
   my $win_name  = $_[1]; # window name/label
   my $mod_ref   = $_[2]; # reference to module structure
   my $chn_ind   = $_[3]; # chain index on module

   if ( !scan_module_is_valid($mod_ref) )
   {  printf "\nERROR - Invalid parameters passed to function!\n";
      return($FALSE);
   }
   my ($chip_list, $chip_ref, $jad_list, $jad_ref, $i, $menubar, $casmenu, 
       $name, $frame, $label, $canvas, $bwidth, $btn, $cmd);
   $bwidth = 18;

   #  Create the menubar for the TDR access window
   $menubar  = $MW->Menu(-type => 'menubar');
   $MW->configure(-menu => $menubar);
   $casmenu = $menubar->cascade(-label => '~File', -tearoff => 0);
#   $casmenu->command(-label => 'Load Window Configuration ...');
#   $casmenu->command(-label => 'Save Window Configuration ...');
#   $casmenu->separator;
   $casmenu->command(-label => 'Close', -command => [ $MW => 'destroy' ]);

   $casmenu = $menubar->cascade(-label => '~Reset', -tearoff => 0);
   $casmenu->command(-label => 'Reset chip TAP(s) via TMS',  
      -command => [\&jar_cbf_reset, $mod_ref, $chn_ind, 'soft']);
   $casmenu->command(-label => 'Reset chip TAP(s) via TRST', 
      -command => [\&jar_cbf_reset, $mod_ref, $chn_ind, 'hard']);
   $casmenu->command(-label => 'Reset SIC and configure chain(s)',  
      -command => [\&jar_cbf_reset, $mod_ref, $chn_ind, 'sic' ]);

   $casmenu = $menubar->cascade(-label => '~Bypass', -tearoff => 0);
   $cmd = $casmenu->command(-label => 'All chips');  # need frame pointer for callback!
   $casmenu = $menubar->cascade(-label => '~Loop', -tearoff => 0, 
                 -state => 'disable');
   $casmenu->command(-label => 'IR Write',
      -command => [\&jar_cbf_loop, $mod_ref, $chn_ind, 'ir']);
   $casmenu->command(-label => 'DR Write/Discard',
      -command => [\&jar_cbf_loop, $mod_ref, $chn_ind, 'discard']);
   $casmenu->command(-label => 'DR Write/Display',
      -command => [\&jar_cbf_loop, $mod_ref, $chn_ind, 'display']);
   $casmenu->command(-label => 'DR Read/Recirculate',
      -command => [\&jar_cbf_loop, $mod_ref, $chn_ind, 'recirc' ]);

   #
   #  Create the frame to hold the TCK specifics
   #
   $frame  = $MW->Frame(-borderwidth => 10);
   $entry  = $frame->Entry(-width => 4);
   $btn    = $frame->Button(-text => 'Issue TCKs @ RTI', -width => $bwidth,
                -command => [\&jar_cbf_issue_tcks, $mod_ref, $chn_ind, $MW, $entry]);
   $label = $frame->Label(-text => 'Number of TCKs:', -anchor => 'e');
   $btn->pack(-side => 'right', -padx => 5);
   $entry->pack(-side => 'right', -fill => 'both');
   $label->pack(-side => 'right', -fill => 'both');
   $frame->pack(-side => 'top', -fill => 'both');

   #
   #  Create the header and main canvas
   #
   $frame  = $MW->Frame( );
   $label = $frame->Label(-width => 2);
   $label->pack(-side => 'left', -fill => 'both');
   $label = $frame->Label(-width => 22, -text => 'Chip Name',             -anchor => 'w');
   $label->pack(-side => 'left', -fill => 'both');
   $label = $frame->Label(-width => 24, -text => 'Instruction OPNAME',    -anchor => 'w');
   $label->pack(-side => 'left', -fill => 'both');
   $label = $frame->Label(-width => 20, -text => 'Value of Selected TDR', -anchor => 'w');
   $label->pack(-side => 'left', -fill => 'both');
   $frame->pack(-side => 'top', -fill => 'both');

   #
   #  Create the canvas/frame to hold the JTAG TDR specifics
   #
   # note: determine height by counting the number of ASICs and TDR fields (maximum per ASIC)
   $canvas = $MW->Scrolled('Canvas', -scrollbars => 'se',
                -scrollregion => ['0c', '0c', '70c', '200c'],
                -borderwidth => 2, -relief => 'sunken');
   $frame  = $canvas->Frame();

   #
   #  Create the JTAG access data structures for the module
   #
   $jad_list  = jad_list_create_inst($mod_ref, $chn_ind);
   dump_global_jad_lookup() if ($debug_flag & $DEBUG_DETAILS);

   #
   #  Create the frames/widgets for the JTAG TDR information
   #
   $chn_list  = $$mod_ref{CHN_REFS};       # List of chain definition pointers
   $chn_ptr   = $$chn_list[$chn_ind];      # Pointer to specific chain definition
   $chip_list = $$chn_ptr[$CHN_IND_CREFS]; # List of chip definition pointers
   for ($i = 0; $i <= $#$chip_list; $i++)
   {
      $chip_ref = $$chip_list[$i];           # Reference to chip definition structure
      $jad_ref  = $$jad_list[$i];            # Reference to JTAG access data structure

      #  Create the widgets for this register
      return($FALSE) if ( !create_tdr_access_widgets($frame, $mod_ref, $chip_ref, $jad_ref) );
   }

   #
   #  Pack the "top-level" JTAG TDR frame and canvas (display)
   #
   $cmd->configure(-command => [\&jar_cbf_bypass, $frame, $mod_ref, $chn_ind, 'all']);
   $frame->pack(-side => 'top', -fill => 'both', -expand => 'yes');
   $canvas->create('window', 4, 4, -anchor => 'nw', -window => $frame);
   $canvas->pack(-side => 'top', -fill => 'both', -expand => 'yes');


   #
   #  Create a frame to hold the action buttons
   #
   $frame  = $MW->Frame(-borderwidth => 10);
   $btn    = $frame->Button(-text => 'Write IR Only', -width => $bwidth,
                -command => [\&jar_cbf_write, $mod_ref, $chn_ind, $jad_list, 'ir']);
   $btn->pack(-side => 'left', -padx => 5);
   $btn    = $frame->Button(-text => 'DR Write/Discard', -width => $bwidth,
                -command => [\&jar_cbf_write, $mod_ref, $chn_ind, $jad_list, 'discard']);
   $btn->pack(-side => 'left', -padx => 5);
   $btn    = $frame->Button(-text => 'DR Write/Display', -width => $bwidth,
                -command => [\&jar_cbf_write, $mod_ref, $chn_ind, $jad_list, 'display']);
   $btn->pack(-side => 'left', -padx => 5);
   $btn    = $frame->Button(-text => 'DR Read/Recirculate', -width => $bwidth,
                -command => [\&jar_cbf_write, $mod_ref, $chn_ind, $jad_list, 'recirc']);
   $btn->pack(-side => 'left', -padx => 5);
   $label  = $frame->Label(-width => 8);
   $label->pack(-side => 'left', -padx => 5);
   $btn    = $frame->Button(-text => 'Close Window', -width => $bwidth, 
                -command => [ $MW => 'destroy' ]);
   $btn->pack(-side => 'left', -padx => 5);

   $frame->pack(-fill => 'both');

   return($TRUE);
}


#-------------------------------------------------------------------------------
#  JTAG TDR access widgets: create the frame and widgets for the TDR and fields
#-------------------------------------------------------------------------------

sub create_tdr_access_widgets
{  my $num = $#_ + 1;

   printf "\ncreate_tdr_access_widgets( @_ ) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 4 )
   {  printf "\nERROR - Invalid parameters passed to subroutine!\n";
      printf "        Subroutine: create_tdr_access_widgets()\n";
      return($FALSE);
   }
   my $frame    = $_[0];  # top-level frame
   my $mod_ref  = $_[1];  # pointer to module definition structure
   my $chip_ref = $_[2];  # pointer to chip definition structure
   my $jad_ref  = $_[3];  # pointer to JTAG access data structure

   if ( ref($$chip_ref[$CMP_IND_LIST]) ne 'ARRAY' )
   {  printf "\nERROR - Chip definition not initialized for module.\n";
      printf "        Logical name: $$chip_ref[$CMP_IND_BOARD]:$$chip_ref[$CMP_IND_LOG]\n";
      printf "        Subroutine: create_tdr_access_widgets()\n";
      return($FALSE);
   }
   my $jtdr_list = $$chip_ref[$CMP_IND_LIST]; # List of JTAG TDRs for the chip
   my ($sub_frame, $jtdr_ref, $i, $text, $menub, $entry, $label, $check);

   #  Create the frame for the MMR menu button and entry
   $sub_frame = $frame->Frame(-borderwidth => 10);
   $$jad_ref[$JAD_FRAME_TDR] = $sub_frame;  # Save frame reference for future use

   #  Add the menu button to the frame
   $text   = "$$chip_ref[$CMP_IND_BOARD]:$$chip_ref[$CMP_IND_LOG]";
   $menub = $sub_frame->Menubutton(-text => $text, -width => 20, -tearoff => 0,
                -relief => 'raised', -direction => 'below', -anchor => 'w');
   $text   = "Physical name: $$chip_ref[$CMP_IND_BOARD]:$$chip_ref[$CMP_IND_PHY]";
   $menub->command(-label => $text);
   $text   = "Chip type: $$chip_ref[$CMP_IND_DESC]";
   $menub->command(-label => $text);
   $text = "Instruction register length: $$chip_ref[$CMP_IND_ILEN]";
   $menub->command(-label => $text);

   #  Create the option menu for the instruction OPNAMEs
   my @names = ();
   for ($i = 0; $i <= $#$jtdr_list; $i++)
   {  $jtdr_ref = $$jtdr_list[$i];  # Reference to structure for this JTAG TDR
      $names[$i] = jtag_tdr_opname($jtdr_ref);
   }
   $jtdr_ref = $$jtdr_list[$$chip_ref[$CMP_IND_CURR]]; # Current JTAG TDR
   $$jad_ref[$JAD_VAL_OPNAME] = jtag_tdr_opname($jtdr_ref);
   $omenu = $sub_frame->Optionmenu(-options => \@names, -width => 20,
               -textvariable => \$$jad_ref[$JAD_VAL_OPNAME],
               -command => [\&jar_cbf_select_tdr, $frame, $chip_ref, $jad_ref]);

   #  Create the entry widget for the TDR value
   $entry = $sub_frame->Entry(-width => 50,
                -textvariable => \$$jad_ref[$JAD_TEXT_TDR]);
   $entry->bind( '<Return>' => [\&jar_cbf_entry_tdr, $chip_ref, $jad_ref] );
   $entry->bind( '<Leave>' => [\&jar_cbf_entry_tdr, $chip_ref, $jad_ref] );
   $entry->bind( '<Tab>' => [\&jar_cbf_entry_tdr, $chip_ref, $jad_ref] );
   $label = $sub_frame->Label(-text => 'show fields', -font => $fld_font);
   $check = $sub_frame->Checkbutton(-variable => \$$jad_ref[$JAD_VAL_DISPLAY],
      -command => [\&jar_cbf_dsp_checkbtn, $jad_ref]);

   #  Pack the widgets into the frame
   $menub->pack(-side => 'left');
   $omenu->pack(-side => 'left');
   $entry->pack(-side => 'left', -expand => 1);
   $label->pack(-side => 'left');
   $check->pack(-side => 'left');
   $sub_frame->pack(-side => 'top', -fill => 'both');

   #  Create the widgets for the TDR fields (use current JTAG TDR)
   $jtdr_ref = $$jtdr_list[$$chip_ref[$CMP_IND_CURR]]; # Current JTAG TDR
   create_tdr_field_widgets($frame, $jtdr_ref, $jad_ref);
   print "create: jad_ref=$jad_ref=( @$jad_ref )\n" if ($debug_flag & $DEBUG_DETAILS);

   return($TRUE);
}


#-------------------------------------------------------------------------------
#  JTAG TDR access widgets: create the frame and widgets for the fields 
#     of a specific JTAG TDR
#-------------------------------------------------------------------------------

# Assume that JTDR access data (JAD) array has been created and initialized
#    (specifically, the $JAD_FRAME_TDR entry must be valid)!

sub create_tdr_field_widgets
{  return($FALSE) if ( $#_ != 2 );

   print "\ncreate_tdr_field_widgets( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $frame    = $_[0];  #  Top-level frame
   my $jtdr_ref = $_[1];  #  Reference to the JTAG TDR structure
   my $jad_ref  = $_[2];  #  Reference to array with widget data

   my $dptr = $$jtdr_ref[$TDR_REF_DEF]; # pointer to definition structure
   my $fptr = $$dptr{FIELDS};           # pointer to field hash table
   my ($vref, $i, @names, $len, $row, $field, $aptr, $text, $label, $entry);

   #
   #  Initialize the entry variables for the TDR and each field
   #
   $vref = jtag_tdr_get_value($jtdr_ref);
   $$jad_ref[$JAD_TEXT_TDR] = vec_convert_num2str_hex(
      jtag_tdr_length($jtdr_ref), $vref, 'right');
   @names = jtag_tdr_get_fields($jtdr_ref, ORDER);
   for ($i = 0; $i <= $#names; $i++)
   {  $vref= jtag_tdr_get_field_value($jtdr_ref, $names[$i]);
      $len = jtag_tdr_field_parm($jtdr_ref, $names[$i], $TDR_FLD_LENGTH);
      $$jad_ref[$JAD_TEXT_FLD0+$i] = vec_convert_num2str_hex($len, $vref, 'right');
   }

   #
   #  Create the frame for the JTAG TDR field(s)
   #
   my $frame2 = $frame->Frame();
   $$jad_ref[$JAD_FRAME_FLDS] = $frame2;  # Store the frame pointer for future use

   #
   #  Create the widgets for each field in the JTAG TDR
   #
   for ($row = 0; $row <= $#names; $row++)
   {  $field = $names[$row];    # name of field from JTDR being processed
      $aptr  = $$fptr{$field};  # reference to field definition (ARRAY)

      if ( $$aptr[$TDR_FLD_LENGTH] == 1 )
      {  $text = sprintf("<%u>", $$aptr[$TDR_FLD_START]);
      }
      else
      {  $text = sprintf("<%u:%u>", 
            ($$aptr[$TDR_FLD_START] + $$aptr[$TDR_FLD_LENGTH] - 1),
            $$aptr[$TDR_FLD_START]);
      }
      $label = $frame2->Label(-width => 20, -anchor => 'w',  -font => $fld_font);
      Tk::grid( $label, -row => $row, -column => 0, -sticky => 'e');
      $label = $frame2->Label(-width => 20, -text => $field, -anchor => 'w', -font => $fld_font);
      Tk::grid( $label, -row => $row, -column => 1, -sticky => 'e');
      $label = $frame2->Label(-width => 8,  -text => $text,  -anchor => 'w', -font => $fld_font);
      Tk::grid( $label, -row => $row, -column => 2, -sticky => 'e');
      $entry = $frame2->Entry(-width => 26, -font   => $fld_font,
                   -textvariable => \$$jad_ref[$JAD_TEXT_FLD0+$row]);
      $entry->bind( '<Return>' => [\&jar_cbf_entry_field, $jtdr_ref, $jad_ref, $field] );
      $entry->bind( '<Leave>' => [\&jar_cbf_entry_field, $jtdr_ref, $jad_ref, $field] );
      $entry->bind( '<Tab>' => [\&jar_cbf_entry_field, $jtdr_ref, $jad_ref, $field] );
      Tk::grid( $entry, -row => $row, -column => 3, -sticky => 'e');
      $text  = sprintf("%s", $$aptr[$TDR_FLD_DESC]);
      $label = $frame2->Label(-text => $text, -anchor => 'w', -width => 50,
                -font   => $fld_font);
      Tk::grid( $label, -row => $row, -column => 4, -sticky => 'e');
      $frame2->gridRowconfigure(1,-weight => 1);
   }

   jar_cbf_dsp_checkbtn($jad_ref);
   #$frame2->pack(-after => $$jad_ref[$JAD_FRAME_TDR], -fill => 'both');
}


#-------------------------------------------------------------------------------
#  JTAG TDR access data: create a list of JAD data structures (one per ASIC)
#-------------------------------------------------------------------------------

sub jad_list_create_inst
{
   print "\njad_list_create_inst( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my @jad_list = ();  # Array to hold pointers (one per chip on module)
   return( \@jad_list ) if ( $#_ != 1 );
   my $mref    = $_[0]; # Reference to module definition structure
   my $chn_ind = $_[1]; # Chain number/index on module

   my ($chn_list, $chn_ptr, $chip_list, $i, $chip_ref, $jtdr_list, $jtdr_ref, $jad_ref);

   $chn_list  = $$mref{CHN_REFS};          # List of chain definition pointers
   $chn_ptr   = $$chn_list[$chn_ind];      # Pointer to specific chain definition
   $chip_list = $$chn_ptr[$CHN_IND_CREFS]; # List of chip definition pointers

   for ($i = 0; $i <= $#$chip_list; $i++)
   {  $chip_ref  = $$chip_list[$i];          # Reference for this specific chip
      $jtdr_list = $$chip_ref[$CMP_IND_LIST]; # List of JTAG TDRs for the chip
      $jtdr_ref  = $$jtdr_list[$$chip_ref[$CMP_IND_CURR]]; # Current JTAG TDR
      $jad_ref   = jad_create_inst($jtdr_ref); # Create JTAG TDR access data structure
      $jad_list[$i] = $jad_ref;
   }

   my $name = $$mref{NAME} . "-" . $chn_ind;
   $jad_lookup{ $name } = \@jad_list;   # Store the structure by name
   return( \@jad_list );
}


#-------------------------------------------------------------------------------

sub dump_global_jad_lookup
{  my ($name, $aref);
   print "\n";
   print "-------------------------------------------------------\n";
   print "- DUMPING JAD LOOKUP HASH TABLE:\n";
   print "-------------------------------------------------------\n";
   foreach $name ( keys %jad_lookup )
   {  print "\n$name = $jad_lookup{$name}\n";
      $aref = $jad_lookup{$name};
      print "array = ( @$aref )\n";
   }
}


#-------------------------------------------------------------------------------
#  JTAG TDR access data: create a JAD data structure for an ASIC
#     Initialize the entry fields with the specified TDR data
#-------------------------------------------------------------------------------

sub jad_create_inst
{  
   printf "\njad_create_inst( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my @jad_array = ( 0, 0, 0, 0, 0 );
   return(\@jad_array) if ( $#_ != 0 );

   my $jtdr_ref = $_[0];  #  Reference to the JTAG TDR structure
   my ($vref, $i, @names, $len);
   $vref = jtag_tdr_get_value($jtdr_ref);
   $jad_array[$JAD_TEXT_TDR] = vec_convert_num2str_hex(
      jtag_tdr_length($jtdr_ref), $vref, 'right');

   @names = jtag_tdr_get_fields($jtdr_ref, ORDER);
   for ($i = 0; $i <= $#names; $i++)
   {  $vref= jtag_tdr_get_field_value($jtdr_ref, $names[$i]);
      $len = jtag_tdr_field_parm($jtdr_ref, $names[$i], $TDR_FLD_LENGTH);
      $jad_array[$JAD_TEXT_FLD0+$i] = vec_convert_num2str_hex($len, $vref, 'right');
   }

   print "jad_array = ( @jad_array )\n" if ($debug_flag & $DEBUG_DETAILS);
   return(\@jad_array);
}


#-------------------------------------------------------------------------------

sub jad_update_text
{  
   printf "\njad_update_text( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);
   return if ( $#_ != 1 );
   my $jad_ref  = $_[0];  # Reference to JAD data structure
   my $jtdr_ref = $_[1];  #  Reference to the JTAG TDR structure
   my ($vref, $i, @names, $len);

#print "jad = $jad_ref = ( @$jad_ref )\n";
   $vref = jtag_tdr_get_value($jtdr_ref);
   $$jad_ref[$JAD_TEXT_TDR] = vec_convert_num2str_hex(
      jtag_tdr_length($jtdr_ref), $vref, 'right');

   @names = jtag_tdr_get_fields($jtdr_ref, ORDER);
   for ($i = 0; $i <= $#names; $i++)
   {  $vref= jtag_tdr_get_field_value($jtdr_ref, $names[$i]);
      $len = jtag_tdr_field_parm($jtdr_ref, $names[$i], $TDR_FLD_LENGTH);
      $$jad_ref[$JAD_TEXT_FLD0+$i] = vec_convert_num2str_hex($len, $vref, 'right');
   }
#print "jad = $jad_ref = ( @$jad_ref )\n";
}


#-------------------------------------------------------------------------------
#  JTAG TDR access: callback for TDR cascade menu
#-------------------------------------------------------------------------------

sub jar_cbf_select_tdr
{
   print "\njar_cbf_select_tdr( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $frame    = $_[0];  # Top-level frame for JTAG TDR access
   my $chip_ref = $_[1];  # Reference to chip structure
   my $jad_ref  = $_[2];  # Pointer to the JTAG access data 
   my $opname   = $$jad_ref[$JAD_VAL_OPNAME]; # Selected instruction opname (text)
   my ($jtdr_list, $jtdr_ref, $ind, $found);

   if ( $debug_flag & $DEBUG_GENERAL )
   {  printf "\nPARAMTERS:\n";
      printf "jad_ref = $jad_ref = ( @$jad_ref )\n";
      printf "opname  = $opname\n";
   }

   #  Search thru the valid JTAG TDRs for the specified OPNAME
   $jtdr_list = $$chip_ref[$CMP_IND_LIST]; # List of JTAG TDRs for the chip
   for ($ind=0, $found=$FALSE; (($ind <= $#$jtdr_list) && !$found); $ind++)
   {  $jtdr_ref = $$jtdr_list[$ind];         # Structure for this JTAG TDR
      $href     = $$jtdr_ref[$TDR_REF_DEF];  # Pointer to definition (HASH)
      $found    = ( $$href{OPNAME} eq $opname );
   }
   return if ( !$found );
   $ind--;
   return if ( $ind == $$chip_ref[$CMP_IND_CURR] );

   #  Destroy the widgets for the exisiting TDR fields
   my $fld_frame = $$jad_ref[$JAD_FRAME_FLDS];
   $fld_frame->destroy();

   #  Save the index as current and create widgets for TDR fields
   $$chip_ref[$CMP_IND_CURR] = $ind;
   create_tdr_field_widgets($frame, $jtdr_ref, $jad_ref);
}


#-------------------------------------------------------------------------------
#  JTAG TDR access: callback for reset command (in main menubar)
#-------------------------------------------------------------------------------

sub jar_cbf_reset
{
   print "\njar_cbf_reset( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $mref     = $_[0];  # Reference to module definition
   my $chn_ind  = $_[1];  # Chain number/index on module
   my $mode     = $_[2];  # Reset mode: soft, hard, sic

   #
   #  Target the specific module
   #
   scan_module_target($mref, $chn_ind);

   #
   #  Perform the specific reset type
   #
   if ( $mode eq 'soft' )
   {  jtag_reset($JTAG_RESET_SOFT);
   }
   elsif ( $mode eq 'hard' )
   {  jtag_reset($JTAG_RESET_HARD);
   }
   elsif ( $mode eq 'sic' )
   {  scancfg_hardware_init($$mref{NAME});
   }

}


#-------------------------------------------------------------------------------
#  JTAG TDR access: callback for loop command (in main menubar)
#-------------------------------------------------------------------------------

sub jar_cbf_loop
{
   print "\njar_cbf_loop( @_ )\n";
}


#-------------------------------------------------------------------------------
#  JTAG TDR access: issue a specific number of TCKs @ RTI
#-------------------------------------------------------------------------------

sub jar_cbf_issue_tcks
{
   print "\njar_cbf_issue_tcks( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $mref     = $_[0];  # Reference to module definition
   my $chn_ind  = $_[1];  # Chain number/index on module
   my $win      = $_[2];  # Pointer to window
   my $entry    = $_[3];  # Pointer to entry widget

   $val = $entry->get();  # Number of TCKS from entry
   if ( $val <= 0 )
   {  $win->messageBox(-icon => 'info', -type => 'OK',
            -message => 'Invalid number of TCKs specified.',
            -title => 'Issue TCKs @ RTI Error');
      return;
   }

   print "\n-> Issuing number of TCKs @ RTI = $val\n";

   #
   #  Target the specific module
   #
   scan_module_target($mref, $chn_ind);

   #
   #  Issue number of TCKs
   #
   while ( $val > 32 )
   {  jtag_state_to_rti(32);
      $val -= 32;
   }
   if ( $val > 0 )
   {  jtag_state_to_rti($val);
   }

}

#-------------------------------------------------------------------------------
#  JTAG TDR access: callback for bypass all command (in main menubar)
#-------------------------------------------------------------------------------

sub jar_cbf_bypass
{
   print "\njar_cbf_bypass( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $frame   = $_[0];  # top-level frame for module TDR widgets
   my $mref    = $_[1];  # Reference to module definition
   my $chn_ind = $_[2];  # Chain number/index on specified module
   my $mode    = $_[3];  # Bypass mode: all
   my ($chn_list, $chn_ptr, $chip_list, $name, $jad_list, $i, $chip_ptr, 
      $jad_ref, $jtdr_list, $jtdr_ref, $ind);

   #
   #  Extract the appropriate pointers
   #
   $chn_list  = $$mref{CHN_REFS};          # List of chain definition pointers
   $chn_ptr   = $$chn_list[$chn_ind];      # Pointer to specific chain definition
   $chip_list = $$chn_ptr[$CHN_IND_CREFS]; # List of chip definition pointers
   $name      = $$mref{NAME} . "-" . $chn_ind;
   $jad_list  = $jad_lookup{ $name };      # List of JTAG access data pointers

   #
   #  Process each chip on the module
   #
   for ($i = 0; $i <= $#$chip_list; $i++)
   {  $chip_ref  = $$chip_list[$i];          # Reference for this specific chip
      $jad_ref   = $$jad_list[$i];           # Reference for this specific chip
      $jtdr_list = $$chip_ref[$CMP_IND_LIST]; # List of JTAG TDRs for the chip
      $jtdr_ref  = $$jtdr_list[$$chip_ref[$CMP_IND_CURR]]; # Current JTAG TDR

      #  Get the index of the BYPASS TDR for the chip
      $ind = scan_chip_bypass_index($chip_ref);

      #  Update the data structure and display widgets
      jar_cbf_select_tdr($frame, $chip_ref, $ind, $jad_ref);
   }

}


#-------------------------------------------------------------------------------
#  JTAG TDR access: callback for write buttons (IR only, DR with parameters)
#-------------------------------------------------------------------------------

sub jar_cbf_write
{
   print "\njar_cbf_write( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $mref     = $_[0];  # Reference to module definition
   my $chn_ind  = $_[1];  # Chain number/index on specified module
   my $jad_list = $_[2];  # List of pointers (JTAG TDR access data)
   my $mode     = $_[3];  # Write mode: ir, discard, display, recirc

   my ($ilen, $ir_tdi, $dlen, $flag, $idata, @odata); # TDI and TDO data
   my ($chip_list, $i, $chip_ref, $jlist, $jtdr_ref);

   #
   #  Target the specific module
   #
   scan_module_target($mref, $chn_ind);

   #
   #  Extract IR length and IR TDI data
   #
   $ilen   = scan_module_ir_len($mref);
   $ir_tdi = scan_module_build_ir($mref);

   #
   #  Process IR scan only - load the instruction register(s)
   #
   if ( $mode eq 'ir' ) 
   {  
      @odata = jtag_scan_ir($ilen, $ir_tdi);

      if ( $debug_flag & $DEBUG_GENERAL )
      {  print "calling: jtag_scan_ir($ilen, $ir_tdi)\n";
         vec_dump_hex("IR TDI DATA", $ir_tdi);
         vec_dump_hex("IR TDO DATA", \@odata);

         $vec = vec_create($ilen);
         vec_cpy_bits(0, $ir_tdi, 0, $vec, $ilen);
         vec_flip_bits(0, (vec_compute_nlong($ilen) * $BITS_LONG), $vec);
         vec_dump_hex("FLIPPED IR TDI DATA", $vec);
      }

      return;
   }

   #
   #  Extract DR length and DR TDI data
   #
   $dlen  = scan_module_dr_len($mref);
   $idata = scan_module_build_dr($mref);
   $flag  = ( ( $mode eq 'recirc' ) ? $TRUE : $FALSE );

   @odata = jtag_dr_load($ilen, $ir_tdi, $dlen, $idata, $flag);

   if ( $debug_flag & $DEBUG_GENERAL )
   {  print "calling: jtag_dr_load($ilen, $ir_tdi, $dlen, $idata, $flag)\n";
      vec_dump_hex("DR TDI DATA", $idata);
      vec_dump_hex("DR TDO DATA", \@odata);

      $vec = vec_create($dlen);
      vec_cpy_bits(0, $idata, 0, $vec, $dlen);
      vec_flip_bits(0, (vec_compute_nlong($dlen) * $BITS_LONG), $vec);
      vec_dump_hex("FLIPPED DR TDI DATA", $vec);
   }

   #
   #  Update structures with DR TDO data?
   #
   if ( $mode ne 'discard' )
   {  
      #  Parse the TDO data and extract data for each ASIC
      scan_module_parse_dr($mref, \@odata);

      #  Update the display for each ASIC with the TDO data
      $chn_list  = $$mref{CHN_REFS};          # List of chain definition pointers
      $chn_ptr   = $$chn_list[$chn_ind];      # Pointer to specific chain definition
      $chip_list = $$chn_ptr[$CHN_IND_CREFS]; # List of chip definition pointers
      for ($i = 0; $i <= $#$chip_list; $i++)
      {  $chip_ref = $$chip_list[$i];  # Reference to chip definition structure
         $jlist    = $$chip_ref[$CMP_IND_LIST];            # List of all JTAG TDR pointers
         $jtdr_ref = $$jlist[ $$chip_ref[$CMP_IND_CURR] ]; # Current JTAG TDR pointer (ARRAY)
         jad_update_text($$jad_list[$i], $jtdr_ref);
      }
   }

}


#-------------------------------------------------------------------------------
#  JTAG TDR access: callback for field display checkbutton (per TDR)
#-------------------------------------------------------------------------------

sub jar_cbf_dsp_checkbtn
{  print "\njar_cbf_dsp_checkbtn(@_) : #parms=$#_\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my $jad_ref   = $_[0];                      # reference to JAD structure
   my $tdr_frame = $$jad_ref[$JAD_FRAME_TDR];  # frame containing MMR name/menu
   my $fld_frame = $$jad_ref[$JAD_FRAME_FLDS]; # frame containing fields

   #  Show or hide the frame containing the field definition(s)
   if ( $$jad_ref[$JAD_VAL_DISPLAY] == 0 )
   {  $fld_frame->pack("forget");
   }
   else
   {  $fld_frame->pack(-after => $tdr_frame, -fill => 'both');
   }
}


#-------------------------------------------------------------------------------
#  JTAG TDR access: callback for TDR entry (per TDR)
#-------------------------------------------------------------------------------

sub jar_cbf_entry_tdr
{  print "\njar_cbf_entry_tdr(@_) : #parms=$#_\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   # NOTE: confirm #parmeters = 3, and that $_[0] is a HASH
   my $href      = $_[0];  # hash pointer w/ Tk specifics
   my $chip_ref  = $_[1];  # reference to chip definition structure (ARRAY)
   my $jad_ref   = $_[2];  # pointer to JTDR access data (JAD) array

   #  Extract pointer to the JTAG TDR definition and field hash structures
   my $jtdr_ref = scan_chip_get_jtdr_ptr($chip_ref);
   my $dptr     = $$jtdr_ref[$TDR_REF_DEF]; # pointer to definition structure
   my $fptr     = $$dptr{FIELDS};           # pointer to field hash table

   #  Convert the value given for the register and store (structure & GUI entry)
   my ($vref, $len, $field, $aptr, $ind);
   $len = $$dptr{LENGTH};
   $vref = vec_convert_hex_str2num($len, $$jad_ref[$JAD_TEXT_TDR], 'right');
   jtag_tdr_set_value($jtdr_ref, $vref);
   $vref = jtag_tdr_get_value($jtdr_ref);
   $$jad_ref[$JAD_TEXT_TDR] = vec_convert_num2str_hex($len, $vref, 'right');

   #  Update the value of each field in the register
   foreach $field (keys %$fptr)
   {  $vref = jtag_tdr_get_field_value($jtdr_ref, $field); # field value
      $aptr = $$fptr{$field};         # pointer to field parameters (ARRAY)
      $ind  = $$aptr[$TDR_FLD_INDEX]; # index value from field parameters
      $$jad_ref[$JAD_TEXT_FLD0+$ind] = vec_convert_num2str_hex(
         $$aptr[$TDR_FLD_LENGTH], $vref, 'right');
   }
}


#-------------------------------------------------------------------------------
#  JTAG TDR access: callback for field entry (per TDR)
#-------------------------------------------------------------------------------

sub jar_cbf_entry_field
{  print "\njar_cbf_entry_field(@_) : #parms=$#_\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   # NOTE: confirm #parmeters = 3, and that $_[0] is a HASH
   my $href      = $_[0];  # hash pointer w/ Tk specifics
   my $jtdr_ref  = $_[1];  # reference to JTAG TDR instance (ARRAY)
   my $jad_ref   = $_[2];  # pointer to JTDR access data (JAD) array
   my $field     = $_[3];  # name of field affected

   #  Extract pointer to the JTAG TDR definition and field hash structures
   my $dptr     = $$jtdr_ref[$TDR_REF_DEF]; # pointer to definition structure
   my $fptr     = $$dptr{FIELDS};           # pointer to field hash table
   my $aptr     = $$fptr{$field};           # pointer to field definition
   my $ind      = $$aptr[$TDR_FLD_INDEX];   # index of field within TDR

   #  Convert the field text entry to a numeric value and store in JTAG TDR
   my $vref = vec_convert_hex_str2num($$aptr[$TDR_FLD_LENGTH],
             $$jad_ref[$JAD_TEXT_FLD0+$ind], 'right');
   jtag_tdr_set_field_value($jtdr_ref, $field, $vref);
   $vref = jtag_tdr_get_field_value($jtdr_ref, $field);
   $$jad_ref[$JAD_TEXT_FLD0+$ind] = vec_convert_num2str_hex(
         $$aptr[$TDR_FLD_LENGTH], $vref, 'right');

   #  Update the entry with the complete JTAG TDR value
   $vref = jtag_tdr_get_value($jtdr_ref);
   $$jad_ref[$JAD_TEXT_TDR] = vec_convert_num2str_hex($$dptr{LENGTH}, $vref, 
      'right');
}


#-------------------------------------------------------------------------------

1;
