# 
#/**************************************************************************
# *                                                                        *
# *               Copyright (C) 1989-1999 Silicon Graphics, Inc.           *
# *                                                                        *
# *  These coded instructions, statements, and computer programs  contain  *
# *  unpublished  proprietary  information of Silicon Graphics, Inc., and  *
# *  are protected by Federal copyright law.  They  may  not be disclosed  *
# *  to  third  parties  or copied or duplicated in any form, in whole or  *
# *  in part, without the prior written consent of Silicon Graphics, Inc.  *
# *                                                                        *
# **************************************************************************/
# 
# @(#)$Id: vstation_subs.pm,v 1.7 2003/05/06 18:01:49 beighley Exp $
#-------------------------------------------------------------------------------
#
#  FILE: vstation_subs.pm
#
#  DESCRIPTION:  This file contains PERL subroutines to interact with the
#     VStation emulator from Mentor Graphics (ikos).
#
#  NOTES: The "common_defs.pm" file is required when this file is used.
#
#  SUBROUTINES:  The following subroutines are contained in this file:
#     (expected parameter usage displayed; some have optional parameters)
#
#  -- The following support connect/disconnect to/from emulator command server
#  vstation_server_connect($vstation_host, $vstation_port)
#  vstation_server_disconnect()
#
#  -- The following will initialize, configure and shutdown the emulator
#  $flag = vstation_is_available() -- verify emulator is available
#  vstation_emulator_init() -- connect to emulator hardware
#  vstation_emulator_init($box_num) -- connect to emulator hardware
#  vstation_cd($dir_name) -- change current configuration/working directory 
#  vstation_configure($root_name, $probe_set) -- download design to emulator
#  vstation_configure($root_name, $probe_set, $speed)
#  vstation_emulator_shutdown() -- disconnect/quit from hardware (erase design)
#
#  -- The following functions will enable/disable external I/Os
#  vstation_io_disable() -- enable I/Os to the target system (out of tristate)
#  vstation_io_enable()  -- disable I/Os to the target system (tristate)
#
#  -- The following function issues a specific "vrun" command
#  $text = vstation_vrun_cmd($cmd) -- issue specific "vrun" command
#
#  -- The following functions will run and check trigger programs
#  vstation_dad_load_trig($obj_file) -- load the binary trigger program
#  vstation_dad_load_trig($obj_file, $interval)
#  vstation_dad_run() -- run the trigger machine (pre-loaded trigger file)
#  vstation_dad_load_trig_run($obj_file) -- load and run the binary trigger program
#  vstation_dad_load_trig_run($obj_file, $interval)
#  vstation_dad_stop() -- stop the trigger machine
#  @s_data = vstation_dad_status() -- get trigger machine status for all fields
#  $status = vstation_dad_status($field_num) -- get specific trigger field status
#  $flag = vstation_dad_wait($timeout) -- wait for data buffer ready ***NOT TESTED***
#  $flag = vstation_dad_wait($field_num, $value, $timeout) -- wait for specific field ***NOT TESTED***
#  vstation_dad_upload($dir_name) -- upload probe data to file
#  vstation_dad_upload($dir_name, $num_samp) -- upload probe data: specific # samples
#  vstation_dad_upload($dir_name, $num_samp, $trig_pos) -- upload probe data: specific # samples
#  vstation_replay_data($dir_name) -- replay raw probe data and translate to waveform
#  
#  -- The following functions will save and restore design state: registers and/or memories
#  vstation_design_save($dir_name) -- save registers and memories
#  vstation_design_save($dir_name, $what) -- save registers or memories
#  vstation_design_restore($dir_name)  -- restore registers and/or memories
#  vstation_design_restore($dir_name, $what) -- restore registers or memories
#
#  -- The following functions will save and restore state elements 
#  vstation_state_ctrl_get($sig_file) -- get state values and save to file
#  vstation_state_ctrl_get($sig_file, $fmt) -- get state values and save to file
#  vstation_state_ctrl_set($sig_file) -- set state values from file
#  vstation_state_ctrl_set($sig_file, $zero_flag) -- zero states not listed
#
#  -- The following will extract data from a file from emulator "state_get"
#  $val  = vstation_file_get_signal($sig_file, $sig_name)
#  @vals = vstation_file_get_signal_list($sig_file, \@sig_names)
#
#  -- The following functions will upload (read) and poke (write) memories
#  vstation_mem_reconfigure() -- downloads all memory content file(s)
#  vstation_mem_reconfigure($mem_name) -- downloads specific memory file
#  vstation_mem_clear() -- clears all user memories with 0s
#  vstation_mem_upload() -- upload all memories to default file
#  vstation_mem_upload($mem_name) -- upload memory/all to default file
#  vstation_mem_upload($mem_name, $out_file) -- upload memory/all to file
#  vstation_mem_upload($mem_name, $out_file, $beg_addr) -- upload location
#  vstation_mem_upload($mem_name, $out_file, $beg_addr, $end_addr) -- upload range
#  vstation_mem_poke($mem_name, $dat_file) -- write memory with data from file
#  vstation_mem_poke($mem_name, $addr, $data) -- write one location with data
#  vstation_mem_poke($mem_name, $beg_addr, $end_addr, $data) -- tile range with same data
#  vstation_mem_poke($mem_name, $beg_addr, $end_addr, \@data) -- write range with data
#
#  -- The following will extract memory data from a file from emulator "upload"
#  @vals = vstation_file_get_mem($mem_file) -- read entire file
#  @vals = vstation_file_get_mem($mem_file, $beg_addr, $end_addr) -- read section
#
#  -- The following will affect the behavior of the VStation command server
#  vstation_server_set_timeout($sec, $usec) -- set command response timeout values
#
#  -- The following will upload (read) and poke (write) jetson's system memory
#  --   The system memory is 72-bits wide and consists of 4*18 individual memories
#  --   These subroutines utilize the "vstation_mem_*" commands for the individual memories
#  jet_emu_mem_upload($base_out_name) -- upload DIMM memories to files
#  jet_emu_mem_upload($base_out_name, $beg_addr) -- upload specific location to files
#  jet_emu_mem_upload($base_out_name, $beg_addr, $end_addr) -- upload range to files
#  jet_emu_mem_poke($base_img_name) -- write DIMM memories with data from files
#
#  -- The following will extract system memory data or ECC from a complete data file
#  --   ***NOTE: THE FOLLOWING SUBROUTINES ARE NOT IMPLEMENTED IN THE SERVER***
#  @vals = jet_emu_file_get_data($mem_file) -- read entire file
#  @vals = jet_emu_file_get_data($mem_file, $beg_addr) -- read single location
#  @vals = jet_emu_file_get_data($mem_file, $beg_addr, $end_addr) -- read section
#  @vals = jet_emu_file_get_ecc($mem_file) -- read entire file
#  @vals = jet_emu_file_get_ecc($mem_file, $beg_addr) -- read single location
#  @vals = jet_emu_file_get_ecc($mem_file, $beg_addr, $end_addr) -- read section
#
#  MODIFICATION HISTORY:
#
#     WHEN      WHO             WHAT
#     --------  --------------  -----------------------------------------
#     03/31/03  Karen Beighley  Original source code for source control tree.
#     04/03/03  Karen Beighley  Added read file and timeout commands.
#     04/07/03  Karen Beighley  Modified format of "dad_status" subroutine.
#     05/06/03  Karen Beighley  Added subroutines for Jetson's system memory.
#
#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
#  CONSTANTS - These should not be changed!
#-------------------------------------------------------------------------------

$DEFAULT_VSTATION_PORT = 6543;

$DAD_STATUS_INDEX = 0;
$DAD_STATUS_LABEL = 1;
$DAD_STATUS_VALUE = 2;

#-------------------------------------------------------------------------------
#  GLOBAL VARIABLES
#-------------------------------------------------------------------------------

$vstation_serv = $FALSE;
$vstation_init = $FALSE;
$vstation_host = "localhost";
$vstation_port = $DEFAULT_VSTATION_PORT;


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: vstation_lib_msg_level()
#
#    USAGE: vstation_lib_msg_level($level)
#           vstation_lib_msg_level($level, $next_level)
#
#    DESCRIPTION: Set the debug message level within the VStationCtl library.
#       The "level" parameter is a bitmask using a combination of:
#       $DEBUG_FUNCTIONS, $DEBUG_GENERAL, $DEBUG_DETAILS, $DEBUG_VERBOSE
#       The optional "next_level" parameter will enable debug messages
#       in supporting libraries used by the VStationCtl module/library.
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub vstation_lib_msg_level
{
   my $num = $#_ + 1;
   print "\nvstation_lib_msg_level(@_) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);
   my $next_level = 0;

   if ( $num == 2 )
   {  $next_level = $_[2];
   }
   elsif ( $num != 1 )
   {  $emess = "Invalid parameters passed to subroutine vstation_lib_msg_level().";
      &$comm_err_fun($emess);
      return;
   }

   pcf_vstation_msg_level($_[0], $next_level);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: vstation_server_connect()
#
#    USAGE: vstation_server_connect($vstation_host)
#           vstation_server_connect($vstation_host, $vstation_port)
#
#    DESCRIPTION: Open a connection to the specified VStation emulator.
#       The port is assumed to be $DEFAULT_VSTATION_PORT if not provided.
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub vstation_server_connect
{
   my $num = $#_ + 1;
   print "\nvstation_server_connect( @_ ) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num == 2 )
   {  $vstation_host = $_[0];
      $vstation_port = $_[1];
   }
   elsif ( $num == 1 )
   {  $vstation_host = $_[0];
      $vstation_port = $DEFAULT_VSTATION_PORT;
   }
   else
   {  $ecode = -1;
      $emess = "Invalid parameters passed to subroutine vstation_server_connect().";
      &$comm_err_fun($emess);
      return;
   }

   vstation_server_disconnect()  if ( $vstation_serv == $TRUE );

   printf "\nOpening connection to VStation command server: host=$vstation_host, port=$vstation_port\n";

   pcf_vstation_server_connect($vstation_host, $vstation_port, $ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess);
      return;
   }

   $vstation_serv = $TRUE;
   printf "\tSuccessful connection and initialization with VStation command server.\n";
   return;
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: vstation_server_disconnect()
#
#    USAGE: vstation_server_disconnect()
#
#    DESCRIPTION: Close a connection to the VStation emulator (pre-opened).
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub vstation_server_disconnect
{
   print "\nvstation_server_disconnect( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $vstation_serv == $TRUE )
   {  pcf_vstation_server_disconnect();
      printf "\nClosing connection to VStation command server: host=$vstation_host, port=$vstation_port\n";
   }
   $vstation_serv = $FALSE;
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: vstation_is_available()
#
#    USAGE: $flag = vstation_is_available()
#
#    DESCRIPTION: The subroutine checks the availability of the VStation.
#
#    RETURN VALUE: $TRUE if the VStation is available; $FALSE if not available.
#
#-------------------------------------------------------------------------------

sub vstation_is_available
{
   my $num = $#_ + 1;
   print "\nvstation_is_available( @_ ) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 0 )
   {  $ecode = -1;
      $emess = "Invalid parameters passed to subroutine vstation_is_available().";
      &$comm_err_fun($emess);
      return($FALSE);
   }

   my $rv = pcf_vstation_is_available($ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess);
      return($FALSE);
   }
   return( $rv );
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: vstation_emulator_init()
#
#    USAGE: vstation_emulator_init()
#           vstation_emulator_init($box_num)
#
#    DESCRIPTION: This subroutine will connect the command server to the
#      VStation emulator hardware.  The emulator box will default to 1.
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub vstation_emulator_init
{
   my $num = $#_ + 1;
   print "\nvstation_emulator_init( @_ ) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   my $box_num = 1;

   if ( $num == 1 )
   {  $box_num = $_[0];
   }
   elsif ( $num != 0 )
   {  $ecode = -1;
      $emess = "Invalid parameters passed to subroutine vstation_emulator_init().";
      &$comm_err_fun($emess);
      return;
   }

   vstation_emulator_shutdown()  if ( $vstation_init == $TRUE );

   pcf_vstation_emulator_init($box_num, $ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess);
      return;
   }
   $vstation_init = $TRUE;
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: vstation_emulator_shutdown()
#
#    USAGE: vstation_emulator_shutdown()
#
#    DESCRIPTION: Close the connection to the VStation emulator.
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub vstation_emulator_shutdown
{
   print "\nvstation_emulator_shutdown( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $vstation_init == $TRUE )
   {  pcf_vstation_emulator_shutdown($ecode, $emess);
   }
   $vstation_init = $FALSE;
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: vstation_cd()
#
#    USAGE: vstation_cd($dir_name)
#
#    DESCRIPTION: The subroutine will change the current configuration
#      or working directory.
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub vstation_cd
{
   my $num = $#_ + 1;
   print "\nvstation_cd( @_ ) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 1 )
   {  $ecode = -1;
      $emess = "Invalid parameters passed to subroutine vstation_cd().";
      &$comm_err_fun($emess);
      return;
   }

   my $rv = pcf_vstation_cd($_[0], $ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess);
      return;
   }
   return( $rv );
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: vstation_configure()
#
#    USAGE: vstation_configure($root_name, $probe_set)
#           vstation_configure($root_name, $probe_set, $speed)
#
#    DESCRIPTION: This subroutine will download the design from the
#       current directory to the emulator.
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub vstation_configure
{
   my $num = $#_ + 1;
   print "\nvstation_configure( @_ ) : num=$num\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   my ($root_name, $probe_set, $speed);

   if ( $num == 3 )
   {  $root_name = $_[0];
      $probe_set = $_[1];
      $speed     = $_[2]; 
   }
   elsif ( $num == 2 )
   {  $root_name = $_[0];
      $probe_set = $_[1];
      $speed     = -1;
   }
   else
   {  $ecode = -1;
      $emess = "Invalid parameters passed to subroutine vstation_configure().";
      &$comm_err_fun($emess);
      return;
   }

   pcf_vstation_configure($root_name, $probe_set, $speed, $ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess);
      return;
   }
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: vstation_io_enable()
#
#    USAGE: vstation_io_enable()
#
#    DESCRIPTION: Enable I/O pods to external interface (out of tristate).
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub vstation_io_enable
{
   print "\nvstation_io_enable( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   pcf_vstation_io_enable($ecode, $emess);

   if ( $ecode < 0 )
   {  &$comm_err_fun($emess);
      return;
   }
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: vstation_io_disable()
#
#    USAGE: vstation_io_disable()
#
#    DESCRIPTION: Disable I/O pods to external interface (tristate).
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub vstation_io_disable
{
   print "\nvstation_io_disable( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   pcf_vstation_io_disable($ecode, $emess);

   if ( $ecode < 0 )
   {  &$comm_err_fun($emess);
      return;
   }
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: vstation_vrun_cmd()
#
#    USAGE: $text = vstation_vrun_cmd($cmd)
#
#    DESCRIPTION: The subroutine will issue a specific "vrun" command and 
#       return the response text.
#
#    RETURN VALUE: response text after command issued to "vrun"
#
#-------------------------------------------------------------------------------

sub vstation_vrun_cmd
{
   my $num = $#_ + 1;
   print "\nvstation_vrun_cmd( @_ ) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 1 )
   {  $ecode = -1;
      $emess = "Invalid parameters passed to subroutine vstation_vrun_cmd().";
      &$comm_err_fun($emess);
      return;
   }

   my $rv = pcf_vstation_vrun_cmd($_[0], $ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess);
      return;
   }
   return( $rv );
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: vstation_dad_load_trig()
#
#    USAGE: vstation_dad_load_trig($obj_file)
#           vstation_dad_load_trig($obj_file, $interval)
#
#    DESCRIPTION: The subroutine download a specific binary trigger program.
#       The optional "interval" parameter alters the snapshot interval for
#       fifo addresses.
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub vstation_dad_load_trig
{
   my $num = $#_ + 1;
   print "\nvstation_dad_load_trig( @_ ) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);
   my ($obj_file, $interval);

   if ( $num == 2 )
   {  $obj_file = $_[0];
      $interval = $_[1];
   }
   elsif ( $num == 1 )
   {  $obj_file = $_[0];
      $interval = -1;
   }
   else
   {  $ecode = -1;
      $emess = "Invalid parameters passed to subroutine vstation_dad_load_trig().";
      &$comm_err_fun($emess);
      return;
   }

   pcf_vstation_dad_load_trig($obj_file, $interval, $ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess);
      return;
   }
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: vstation_dad_run()
#
#    USAGE: vstation_dad_run()
#
#    DESCRIPTION: Run the trigger state machine; the trigger file must have
#       been previously loaded.
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub vstation_dad_run
{
   print "\nvstation_dad_run( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   pcf_vstation_dad_run($ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess);
      return;
   }
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: vstation_dad_load_trig_run()
#
#    USAGE: vstation_dad_load_trig_run($obj_file)
#           vstation_dad_load_trig_run($obj_file, $interval)
#
#    DESCRIPTION: The subroutine download a specific binary trigger program
#       and run the trigger state machine.
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub vstation_dad_load_trig_run
{
   my $num = $#_ + 1;
   print "\nvstation_dad_load_trig_run( @_ ) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);
   my ($obj_file, $interval);

   if ( $num == 2 )
   {  $obj_file = $_[0];
      $interval = $_[1];
   }
   elsif ( $num == 1 )
   {  $obj_file = $_[0];
      $interval = -1;
   }
   else
   {  $ecode = -1;
      $emess = "Invalid parameters passed to subroutine vstation_dad_load_trig_run().";
      &$comm_err_fun($emess);
      return;
   }

   pcf_vstation_dad_load_trig($obj_file, $interval, $ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess);
      return;
   }
   pcf_vstation_dad_run($ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess);
      return;
   }
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: vstation_dad_stop()
#
#    USAGE: vstation_dad_stop()
#
#    DESCRIPTION: Stop the trigger state machine (previously loaded and ran).
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub vstation_dad_stop
{
   print "\nvstation_dad_stop( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   pcf_vstation_dad_stop($ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess);
      return;
   }
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: vstation_dad_status()
#
#    USAGE: @s_data = vstation_dad_status()
#           $status = vstation_dad_status($field_num)
#
#    DESCRIPTION: This subroutine will read and extract the trigger machine
#      status; all fields will be returned if $field_num not provided
#
#    RETURN VALUE: An array of field status values or a specific status value.
#
#-------------------------------------------------------------------------------

sub vstation_dad_status
{
   my $num = $#_ + 1;
   print "\nvstation_dad_status( @_ ) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   my $field_num = -1;

   if ( $num == 1 )
   {  $field_num = $_[0];
   }
   elsif ( $num != 0 )
   {  $ecode = -1;
      $emess = "Invalid parameters passed to subroutine vstation_dad_status().";
      &$comm_err_fun($emess);
      return;
   }

   if ( $field_num == -1 )
   {  my $text = pcf_vstation_dad_status_all($ecode, $emess);
      if ( $ecode < 0 )
      {  &$comm_err_fun($emess);
         return;
      }
      $text =~ s/://g;    # Remove colon from label/name field
      my ($i, $j, $n, $elem, $name);
      my @list = split('\n', $text);
      my @rsp_data = ();
      for ($i = 0; $i <= $#list; $i++)
      {  $elem = $list[$i];
         my @tmp_data = split(' ', $elem);
         $n = $#tmp_data;
         if ( $n < 2 )
         {  $ecode = -1;
            $emess = "Unable to extract DAD status from response buffer.";
            $emess .= "\nInvalid format encountered while parsing text.";
            &$comm_err_fun($emess);
            return;
         }
         $name = $tmp_data[1];
         for ($j = 2; $j < $n; $j++)
         {  $name = $name . " " . $tmp_data[$j];
         }
         my @fld_data = ( $tmp_data[0], $name, $tmp_data[$n]);
         $rsp_data[$i] = \@fld_data;
      }
      return (@rsp_data);
   }
   else
   {  my $rv = pcf_vstation_dad_status_field($field_num, $ecode, $emess);
      if ( $ecode < 0 )
      {  &$comm_err_fun($emess);
         return;
      }
      return($rv);
   }
}



#-------------------------------------------------------------------------------
#
#    SUBROUTINE: vstation_dad_wait()
#
#    USAGE: $flag = vstation_dad_wait($timeout) 
#           $flag = vstation_dad_wait($field_num, $value, $timeout)
#
#    DESCRIPTION: The subroutine will wait on the trigger state machine
#      for either: the buffer ready bit (trigger occurred) or a specific
#      field value.  The "timeout" parameter is specified in seconds.
#
#    RETURN VALUE: $TRUE if condition occurred; $FALSE if timeout
#
#-------------------------------------------------------------------------------

sub vstation_dad_wait
{
   my $num = $#_ + 1;
   print "\nvstation_dad_wait( @_ ) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);
   my ($timeout, $fnum, $value);

   if ( $num == 3 )
   {  $fnum    = $_[0];
      $value   = $_[1];
      $timeout = $_[2];
   }
   elsif ( $num == 1 )
   {  $timeout = $_[0];
      $fnum    = -1;
      $value   = -1;
   }
   else
   {  $ecode = -1;
      $emess = "Invalid parameters passed to subroutine vstation_dad_wait().";
      &$comm_err_fun($emess);
      return;
   }

   my $rv = pcf_vstation_dad_wait($fnum, $value, $timeout, $ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess);
      return;
   }
   return($rv);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: vstation_dad_upload()
#
#    USAGE: vstation_dad_upload($dir_name)
#           vstation_dad_upload($dir_name, $num_samp)
#           vstation_dad_upload($dir_name, $num_samp, $trig_pos)
#
#    DESCRIPTION: The subroutine will upload probe data after capture.
#      The default number of samples processed is 2M.  If the number of
#      samples is specified, then the trigger position can be provided.
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub vstation_dad_upload
{
   my $num = $#_ + 1;
   print "\nvstation_dad_upload( @_ ) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);
   my ($dir_name, $num_samp, $trig_pos);

   if ( $num == 3 )
   {  $dir_name = $_[0];
      $num_samp = $_[1];
      $trig_pos = $_[2];
   }
   elsif ( $num == 2 )
   {  $dir_name = $_[0];
      $num_samp = $_[1];
      $trig_pos = -1;
   }
   elsif ( $num == 1 )
   {  $dir_name = $_[0];
      $num_samp = -1;
      $trig_pos = -1;
   }
   else
   {  $ecode = -1;
      $emess = "Invalid parameters passed to subroutine vstation_dad_upload().";
      &$comm_err_fun($emess);
      return;
   }

   pcf_vstation_dad_upload($dir_name, $num_samp, $trig_pos, $ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess);
      return;
   }
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: vstation_replay_data()
#
#    USAGE: vstation_replay_data($dir_name) 
#
#    DESCRIPTION: The subroutine perform a state replay based on existing
#      data in the specified .wave directory.  Creates file "waveform.vrc".
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub vstation_replay_data
{
   my $num = $#_ + 1;
   print "\nvstation_replay_data( @_ ) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 1 )
   {  $ecode = -1;
      $emess = "Invalid parameters passed to subroutine vstation_replay_data().";
      &$comm_err_fun($emess);
      return;
   }

   pcf_vstation_replay_data($_[0], $ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess);
      return;
   }
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: vstation_design_save()
#
#    USAGE: vstation_design_save($dir_name) 
#           vstation_design_save($dir_name, $what) 
#
#    DESCRIPTION: The subroutine will save the design state to the specified
#      directory.  The register state, memory state or both can be saved.
#      The state dump is not allowed when the emulator is incircuit with
#      external clocks running.  The "what" parameter can be:
#         'all' = save both user register state and memories
#         'mem' = save user memories only
#         'reg' = save user register state only
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub vstation_design_save
{
   my $num = $#_ + 1;
   print "\nvstation_design_save( @_ ) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   my ($dir_name, $what);

   if ( $num == 2 )
   {  $dir_name = $_[0];
      $what     = $_[1];
   }
   elsif ( $num == 1 )
   {  $dir_name = $_[0];
      $what     = 'all';
   }
   else
   {  $ecode = -1;
      $emess = "Invalid parameters passed to subroutine vstation_design_save().";
      &$comm_err_fun($emess);
      return;
   }
   if ( !(($what =~ /^mem/i) || ($what =~ /^reg/i) || ($what =~ /^all/i)) )
   {  $ecode = -1;
      $emess = "Invalid parameter passed to subroutine vstation_design_save().";
      $emess .= "\nUnknown parameter of what to save: $what";
      &$comm_err_fun($emess);
      return;
   }

   pcf_vstation_design_save($dir_name, $what, $ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess);
      return;
   }
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: vstation_design_restore()
#
#    USAGE: vstation_design_restore($dir_name) 
#           vstation_design_restore($dir_name, $what) 
#
#    DESCRIPTION: The subroutine will save the design state to the specified
#      directory.  The register state, memory state or both can be saved.
#      The state dump is not allowed when the emulator is incircuit with
#      external clocks running.  The "what" parameter can be:
#         'all' = save both user register state and memories
#         'mem' = save user memories only
#         'reg' = save user register state only
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub vstation_design_restore
{
   my $num = $#_ + 1;
   print "\nvstation_design_restore( @_ ) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   my ($dir_name, $what);

   if ( $num == 2 )
   {  $dir_name = $_[0];
      $what     = $_[1];
   }
   elsif ( $num == 1 )
   {  $dir_name = $_[0];
      $what     = 'all';
   }
   else
   {  $ecode = -1;
      $emess = "Invalid parameters passed to subroutine vstation_design_restore().";
      &$comm_err_fun($emess);
      return;
   }
   if ( !(($what =~ /^mem/i) || ($what =~ /^reg/i) || ($what =~ /^all/i)) )
   {  $ecode = -1;
      $emess = "Invalid parameter passed to subroutine vstation_design_restore().";
      $emess .= "\nUnknown parameter of what to restore: $what";
      &$comm_err_fun($emess);
      return;
   }

   pcf_vstation_design_restore($dir_name, $what, $ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess);
      return;
   }
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: vstation_state_ctrl_get()
#
#    USAGE: vstation_state_ctrl_get($sig_file) 
#           vstation_state_ctrl_get($sig_file, $fmt) 
#
#    DESCRIPTION: The subroutine will sample the state elements and write the
#      values to the specified file.  The default format is hexidecimal, and
#      can be changed by specifying a format character: 
#         'x' = hexidecimal (lower-case?)
#         'h' = hexidecimal (upper-case?)
#         'd' = decimal
#         'o' = octal
#         'b' = binary
#         'r' = raw binary dump
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub vstation_state_ctrl_get
{
   my $num = $#_ + 1;
   print "\nvstation_state_ctrl_get( @_ ) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   my ($sig_file, $fmt);

   if ( $num == 2 )
   {  $sig_file = $_[0];
      $fmt      = $_[1];
   }
   elsif ( $num == 1 )
   {  $sig_file = $_[0];
      $fmt      = 'h';
   }
   else
   {  $ecode = -1;
      $emess = "Invalid parameters passed to subroutine vstation_state_ctrl_get().";
      &$comm_err_fun($emess);
      return;
   }
   if ( !(($fmt =~ /^x/i) || ($fmt =~ /^h/i) || ($fmt =~ /^d/i) ||
          ($fmt =~ /^o/i) || ($fmt =~ /^b/i) || ($fmt =~ /^r/i)) )
   {  $ecode = -1;
      $emess = "Invalid parameter passed to subroutine vstation_state_ctrl_get().";
      $emess .= "\nUnknown file format parameter specified: $fmt";
      &$comm_err_fun($emess);
      return;
   }
   

   pcf_vstation_state_ctrl_get($sig_file, $fmt, $ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess);
      return;
   }
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: vstation_state_ctrl_set()
#
#    USAGE: vstation_state_ctrl_set($sig_file) 
#           vstation_state_ctrl_set($sig_file, $zero_flag) 
#
#    DESCRIPTION: The subroutine will set the state elements listed in the
#      the specified file.  The "zero_flag" can be $TRUE or $FALSE, where
#         $TRUE  = zero any state elements no listed in the file
#         $FALSE = do not change value of states not listed in file (default)
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub vstation_state_ctrl_set
{
   my $num = $#_ + 1;
   print "\nvstation_state_ctrl_set( @_ ) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   my ($sig_file, $flag);

   if ( $num == 2 )
   {  $sig_file = $_[0];
      $flag     = $_[1];
   }
   elsif ( $num == 1 )
   {  $sig_file = $_[0];
      $flag     = $FALSE;
   }
   else
   {  $ecode = -1;
      $emess = "Invalid parameters passed to subroutine vstation_state_ctrl_set().";
      &$comm_err_fun($emess);
      return;
   }

   pcf_vstation_state_ctrl_set($sig_file, $flag, $ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess);
      return;
   }
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: vstation_file_get_signal()
#
#    USAGE: $val = vstation_file_get_signal($sig_file, $sig_name) 
#
#    DESCRIPTION: The subroutine will extract a value from the state data
#       file (sig_file) for the specified signal (sig_name).
#
#    RETURN VALUE: string containing value from file
#
#-------------------------------------------------------------------------------

sub vstation_file_get_signal
{
   my $num = $#_ + 1;
   print "\nvstation_file_get_signal( @_ ) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 2 )
   {  $ecode = -1;
      $emess = "Invalid parameters passed to subroutine vstation_file_get_signal().";
      &$comm_err_fun($emess);
      return;
   }

   my $rv = pcf_vstation_file_get_signal($_[0], $_[1], $ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess);
      return;
   }
   return($rv);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: vstation_file_get_signal_list()
#
#    USAGE: @vals = vstation_file_get_signal($sig_file, \@sig_names) 
#
#    DESCRIPTION: The subroutine will extract a values from the state data
#       file (sig_file) for the list of specified signals (sig_name).
#
#    RETURN VALUE: array of strings containing values from file
#
#-------------------------------------------------------------------------------

sub vstation_file_get_signal_list
{
   my $num = $#_ + 1;
   print "\nvstation_file_get_signal_list( @_ ) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 2 )
   {  $ecode = -1;
      $emess = "Invalid parameters passed to subroutine vstation_file_get_signal_list().";
      &$comm_err_fun($emess);
      return;
   }
   if ( ref($_[1]) ne 'ARRAY' )
   {  $ecode = -1;
      $emess = "Invalid parameter passed to subroutine vstation_file_get_signal_list().";
      &$comm_err_fun($emess);
      return;
   }
   my @rv_data = ();
   my $rv = pcf_vstation_file_get_signal_list($_[0], $_[1], \@rv_data, $ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess);
      return;
   }
   return(@rv_data);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: vstation_mem_reconfigure()
#
#    USAGE: vstation_mem_reconfigure()
#           vstation_mem_reconfigure($mem_name)
#
#    DESCRIPTION: This subroutine will download memory with data from
#       contents files; zeros are put in memories without contents files.
#       The "mem_name" parameter can be used to specify a specific memory
#       to reconfigure (hierarchical pathname) or the word 'all'.
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub vstation_mem_reconfigure
{
   my $num = $#_ + 1;
   print "\nvstation_mem_reconfigure( @_ ) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   my $mem_name = "";

   if ( $num == 1 )
   {  $mem_name = $_[0];
   }
   elsif ( $num != 0 )
   {  $ecode = -1;
      $emess = "Invalid parameters passed to subroutine vstation_mem_reconfigure().";
      &$comm_err_fun($emess);
      return;
   }

   pcf_vstation_mem_reconfigure($mem_name, $ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess);
      return;
   }
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: vstation_mem_clear()
#
#    USAGE: vstation_mem_clear()
#
#    DESCRIPTION: The subroutine will clear all user memories with 0s.
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub vstation_mem_clear
{
   print "\nvstation_mem_clear( @_ )\n" if ($debug_flag & $DEBUG_FUNCTIONS);

   pcf_vstation_mem_clear($ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess);
      return;
   }
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: vstation_mem_upload()
#
#    USAGE: vstation_mem_upload($mem_name)
#           vstation_mem_upload($mem_name, $out_file)
#           vstation_mem_upload($mem_name, $out_file, $beg_addr)
#           vstation_mem_upload($mem_name, $out_file, $beg_addr, $end_addr)
#
#    DESCRIPTION: The subroutine will upload memory data from emulator to file.
#       The "mem_name" parameter can be used to specify a specific memory
#       to reconfigure (hierarchical pathname) or the word 'all'. The "out_file"
#       parameter is used as a pre-fix if 'all' is specified for "mem_name".
#       If "out_file" is not specified, the file is named using the memory name.
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub vstation_mem_upload
{
   my $num = $#_ + 1;
   print "\nvstation_mem_upload( @_ ) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);
   my ($mem_name, $out_file, $beg_addr, $end_addr);

   if ( $num == 4 )
   {  $mem_name = $_[0];
      $out_file = $_[1];
      $beg_addr = $_[2];
      $end_addr = $_[3];
   }
   elsif ( $num == 3 )
   {  $mem_name = $_[0];
      $out_file = $_[1];
      $beg_addr = $_[2];
      $end_addr = $_[2];
   }
   elsif ( $num == 2 )
   {  $mem_name = $_[0];
      $out_file = $_[1];
      $beg_addr = -1;
      $end_addr = -1;
   }
   elsif ( $num == 1 )
   {  $mem_name = $_[0];
      $out_file = "";
      $beg_addr = -1;
      $end_addr = -1;
   }
   elsif ( $num == 0 )
   {  $mem_name = 'all';
      $out_file = "";
      $beg_addr = -1;
      $end_addr = -1;
   }
   else
   {  $ecode = -1;
      $emess = "Invalid parameters passed to subroutine vstation_mem_upload().";
      &$comm_err_fun($emess);
      return;
   }

   pcf_vstation_mem_upload($mem_name, $out_file, $beg_addr, $end_addr, $ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess);
      return;
   }
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: vstation_mem_poke()
#
#    USAGE: vstation_mem_poke($mem_name, $dat_file)
#           vstation_mem_poke($mem_name, $addr, $data)
#           vstation_mem_poke($mem_name, $beg_addr, $end_addr, $data)
#           vstation_mem_poke($mem_name, $beg_addr, $end_addr, \@data)
#
#    DESCRIPTION: The subroutine will poke (write) one or a range of memory 
#       addresses with data specified in a file or in parameters.
#       The "mem_name" parameter specifies the memory to poke (hierarchical 
#       pathname).
#
#       To poke the memory with data specified in a file:
#          vstation_mem_poke($mem_name, $dat_file)
#
#       To poke one memory location with a specific value:
#           vstation_mem_poke($mem_name, $addr, $data)
#
#       To poke a range of addresses with the same value (tile memory):
#           vstation_mem_poke($mem_name, $beg_addr, $end_addr, $data)
#
#       To poke a range of addresses with specific values:
#           vstation_mem_poke($mem_name, $beg_addr, $end_addr, \@data)
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub vstation_mem_poke
{
   my $num = $#_ + 1;
   print "\nvstation_mem_poke( @_ ) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num == 2 )
   {  pcf_vstation_mem_poke_file($_[0], $_[1], $ecode, $emess);
   }
   elsif ( $num == 3 )
   {  pcf_vstation_mem_poke_addr($_[0], $_[1], $_[2], $ecode, $emess);
   }
   elsif ( $num == 4 )
   {  if ( ref($_[3]) eq 'ARRAY' )
      {  pcf_vstation_mem_poke_range($_[0], $_[1], $_[2], $_[3], $ecode, $emess);
      }
      else
      {  pcf_vstation_mem_poke_tile($_[0], $_[1], $_[2], $_[3], $ecode, $emess);
      }
   }
   else
   {  $ecode = -1;
      $emess = "Invalid parameters passed to subroutine vstation_mem_poke().";
      &$comm_err_fun($emess);
      return;
   }

   if ( $ecode < 0 )
   {  &$comm_err_fun($emess);
      return;
   }
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: vstation_file_get_mem()
#
#    USAGE: @vals = vstation_file_get_mem($mem_file)
#           @vals = vstation_file_get_mem($mem_file, $beg_addr, $end_addr)
#
#    DESCRIPTION: The subroutine extract data from a memory data file
#       created from an emultor memory upload.  The entire memory data is
#       returned if no address range is specified.
#
#    RETURN VALUE: an array of values
#
#-------------------------------------------------------------------------------

sub vstation_file_get_mem
{
   my $num = $#_ + 1;
   print "\nvstation_file_get_mem( @_ ) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);
   my ($mem_file, $beg_addr, $end_addr);

   if ( $num == 3 )
   {  $mem_file = $_[0];
      $beg_addr = $_[1];
      $end_addr = $_[2];
   }
   elsif ( $num == 2 )
   {  $mem_file = $_[0];
      $beg_addr = $_[1];
      $end_addr = $_[1];
   }
   elsif ( $num == 1 )
   {  $mem_file = $_[0];
      $beg_addr = -1;
      $end_addr = -1;
   }
   else
   {  $ecode = -1;
      $emess = "Invalid parameters passed to subroutine vstation_mem_upload().";
      &$comm_err_fun($emess);
      return;
   }

   my $rv = pcf_vstation_file_get_mem($mem_file, $beg_addr, $end_addr, $ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess);
      return;
   }
   return(@$rv);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: vstation_server_set_timeout()
#
#    USAGE: vstation_server_set_timeout($sec, $usec) 
#
#    DESCRIPTION: The subroutine will set the response timeout values for
#      the VStation command server.
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub vstation_server_set_timeout
{
   my $num = $#_ + 1;
   print "\nvstation_server_set_timeout( @_ ) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 2 )
   {  $ecode = -1;
      $emess = "Invalid parameters passed to subroutine vstation_server_set_timeout().";
      &$comm_err_fun($emess);
      return;
   }

   pcf_vstation_server_set_timeout($_[0], $_[1], $ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess);
      return;
   }
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: jet_emu_mem_upload()
#
#    USAGE: jet_emu_mem_upload($base_out_name)
#           jet_emu_mem_upload($base_out_name, $beg_addr)
#           jet_emu_mem_upload($base_out_name, $beg_addr, $end_addr)
#
#    DESCRIPTION: The subroutine will upload "system" memory from emulator to 
#       multiple files.  The "base_out_name" parameter is used as a pre-fix in
#       the output file names.  One file is created for each nibble memory.
#       There are 4 * 18 memories: (XL0, XR0, YL0, YR0) * (0, 1, ..., 17).
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub jet_emu_mem_upload
{
   my $num = $#_ + 1;
   print "\njet_emu_mem_upload( @_ ) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);
   my ($out_name, $beg_addr, $end_addr);

   if ( $num == 3 )
   {  $out_name = $_[0];
      $beg_addr = $_[1];
      $end_addr = $_[2];
   }
   elsif ( $num == 2 )
   {  $out_name = $_[0];
      $beg_addr = $_[1];
      $end_addr = $_[1];
   }
   elsif ( $num == 1 )
   {  $out_name = $_[0];
      $beg_addr = -1;
      $end_addr = -1;
   }
   else
   {  $ecode = -1;
      $emess = "Invalid parameters passed to subroutine jet_emu_mem_upload().";
      &$comm_err_fun($emess);
      return;
   }

   pcf_jet_emu_mem_upload($out_name, $beg_addr, $end_addr, $ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess);
      return;
   }
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: jet_emu_mem_poke()
#
#    USAGE: jet_emu_mem_poke($base_img_name) 
#
#    DESCRIPTION: The subroutine will poke (download) "system" memory in the 
#       emulator from multiple files.  The "base_img_name" parameter is used 
#       as a pre-fix in for each memory data file.  One file is required for 
#       each nibble memory (these files are created from the "nibble-izer").
#       There are 4 * 18 memories: (XL0, XR0, YL0, YR0) * (0, 1, ..., 17).
#
#    RETURN VALUE: none
#
#-------------------------------------------------------------------------------

sub jet_emu_mem_poke
{
   my $num = $#_ + 1;
   print "\njet_emu_mem_poke( @_ ) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);

   if ( $num != 1 )
   {  $ecode = -1;
      $emess = "Invalid parameters passed to subroutine jet_emu_mem_poke().";
      &$comm_err_fun($emess);
      return;
   }

   pcf_jet_emu_mem_poke($_[0], $ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess);
      return;
   }
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: jet_emu_file_get_data()
#
#    USAGE: @vals = jet_emu_file_get_data($mem_file)
#           @vals = jet_emu_file_get_data($mem_file, $beg_addr, $end_addr)
#
#    DESCRIPTION: The subroutine extract data from a memory data file
#       created from an emultor memory upload.  The entire memory data is
#       returned if no address range is specified.
#
#    RETURN VALUE: an array of values
#
#-------------------------------------------------------------------------------

sub jet_emu_file_get_data
{
   my $num = $#_ + 1;
   print "\njet_emu_file_get_data( @_ ) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);
   my ($mem_file, $beg_addr, $end_addr);

   if ( $num == 3 )
   {  $mem_file = $_[0];
      $beg_addr = $_[1];
      $end_addr = $_[2];
   }
   elsif ( $num == 2 )
   {  $mem_file = $_[0];
      $beg_addr = $_[1];
      $end_addr = $_[1];
   }
   elsif ( $num == 1 )
   {  $mem_file = $_[0];
      $beg_addr = -1;
      $end_addr = -1;
   }
   else
   {  $ecode = -1;
      $emess = "Invalid parameters passed to subroutine vstation_mem_upload().";
      &$comm_err_fun($emess);
      return;
   }

   my $rv = pcf_jet_emu_file_get_data($mem_file, $beg_addr, $end_addr, $ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess);
      return;
   }
   return(@$rv);
}


#-------------------------------------------------------------------------------
#
#    SUBROUTINE: jet_emu_file_get_ecc()
#
#    USAGE: @vals = jet_emu_file_get_ecc($mem_file)
#           @vals = jet_emu_file_get_ecc($mem_file, $beg_addr, $end_addr)
#
#    DESCRIPTION: The subroutine extract data from a memory data file
#       created from an emultor memory upload.  The entire memory data is
#       returned if no address range is specified.
#
#    RETURN VALUE: an array of values
#
#-------------------------------------------------------------------------------

sub jet_emu_file_get_ecc
{
   my $num = $#_ + 1;
   print "\njet_emu_file_get_ecc( @_ ) : num=$num\n"  if ($debug_flag & $DEBUG_FUNCTIONS);
   my ($mem_file, $beg_addr, $end_addr);

   if ( $num == 3 )
   {  $mem_file = $_[0];
      $beg_addr = $_[1];
      $end_addr = $_[2];
   }
   elsif ( $num == 2 )
   {  $mem_file = $_[0];
      $beg_addr = $_[1];
      $end_addr = $_[1];
   }
   elsif ( $num == 1 )
   {  $mem_file = $_[0];
      $beg_addr = -1;
      $end_addr = -1;
   }
   else
   {  $ecode = -1;
      $emess = "Invalid parameters passed to subroutine vstation_mem_upload().";
      &$comm_err_fun($emess);
      return;
   }

   my $rv = pcf_jet_emu_file_get_ecc($mem_file, $beg_addr, $end_addr, $ecode, $emess);
   if ( $ecode < 0 )
   {  &$comm_err_fun($emess);
      return;
   }
   return(@$rv);
}



#-------------------------------------------------------------------------------

1;

