#
#/**************************************************************************
# *                                                                        *
# *               Copyright (C) 2002-2005 Silicon Graphics, Inc.           *
# *                                                                        *
# *  These coded instructions, statements, and computer programs  contain  *
# *  unpublished  proprietary  information of Silicon Graphics, Inc., and  *
# *  are protected by Federal copyright law.  They  may  not be disclosed  *
# *  to  third  parties  or copied or duplicated in any form, in whole or  *
# *  in part, without the prior written consent of Silicon Graphics, Inc.  *
# *                                                                        *
# **************************************************************************/
#
# @(#)$Id: shub2_funcs.pm,v 1.108 2007/10/04 18:56:10 meyer Exp $
#
#  FILE: shub2_funcs.pm
#
#  DESCRIPTION:  This file contains PERL routines to dump and parse Shub2 registers.
#
#  MODIFICATION HISTORY:
#
#     WHEN      WHO             WHAT
#     --------  --------------  -----------------------------------------
#     06/06/05  Gary Meyer      Original source code submitted for release.
#     06/14/05  Gary Meyer      More changes for Shub2.
#     06/18/05  Gary Meyer      Still more changes for Shub2.
#     09/23/05  Gary Meyer      Fixed problem in capture_time_stamp() with printing 
#				Target Id addresses.
#     09/26/05  Gary Meyer	Error signature formatting update.
#     09/28/05  Gary Meyer	Added check of valid data before parsing mmr info.
#     10/24/05  Gary Meyer	Added LLP channel error parsing.  Fixed fast mode problems caused
#                               by not having the correct mmrs in the @SHUB2_SHORT_REG_LIST.
#     10/25/05  Gary Meyer	Update to prevent decoding psp mmrs on headless nodes.
#     10/27/05  Gary Meyer	Update for Shub2 LUT table dumping.
#     10/28/05  Gary Meyer	Update for additional arg needed for pi_deadlock_chk().
#     02/07/06  Gary Meyer	Updates for agent id parsing using AID_DECODE.  Added ECC Shadow mmrs.  
#				Added processor configuration checking.  Updated Processor FRU naming
#                               to match blade labels.
#     02/08/06  Gary Meyer	Now ignoring MD generated RED_BLACK_ERR_TIMEOUT.
#     02/21/06  Gary Meyer	Updates to handle the case where PI CE/UCE error detail is zero'd.
#				Added verbiage to indicate a HUNG BUS error.
#     02/23/06  Gary Meyer	Added code to handle error case of duplicate MMR data in all MMR's.
#     02/25/06  Gary Meyer	Corrected bug that would indicate the incorrect processor in the FRU
#				message for PI FSB HW failures.
#     03/13/06  Gary Meyer	Corrected bug that generated bogus processor config error.
#     03/17/06  Gary Meyer	Added polling of SH2_FSB_SYMMETRIC_AGENT_STATUS if incorrect status found.
#     03/28/06  Gary Meyer	Update to ensure proper FRU Summary ordering is observed when both write
#                               and read uce errors are detected in the PI.
#     03/31/06  Gary Meyer  	Changes to promote FRU score for Sustained snoop conflict on the FSB.
#     04/04/06  Gary Meyer  	Updates for Shadow ECC mmrs on headless nodes.
#     04/10/06  Gary Meyer      Change to promote FRU score for PI's with failing CAM reads.
#     04/11/06  Gary Meyer      Will now promote a Hung PI that's failing CAM reads or blocked request traffic.
#     04/25/06  Gary Meyer      Updates to ensure correct cpu config with IP71 & IP75 blade types.  
#				Generate a config error is fsb transactions are serialized.
#     05/06/06  Gary Meyer      Added check for FSB stall condition.
#     05/08/06  Gary Meyer      Added error signature for FSB stall condition.  Added Nack Hammer config check.
#     05/19/06  Gary Meyer      Updates for fru file parsing.  Added SH2_NODE_ID config checks.
#     05/27/06  Gary Meyer	Updates for exist status enumeration.  Fixed problem with MD error masks where
#				I wasn't taking into account unused bits 11,15,19, and 23, and the differences
#				in the bit mapping between SH2_MEM_ERROR_SUMMARY and SH2_MEM_FIRST_ERROR.
#     06/09/06  Gary Meyer	Updates to print rtrdmp command lines for numa corruption cases.
#     06/12/06  Gary Meyer	Fixed problem with Flush mode checking.  Producing a different config error
#				message when both flush parameters are set to 0.
#				Changed instances of $i in parse_sh2_mmr_list() to $reg_idx to prevent potential
#				endless loop due to conflicts with use of $i in other functions.
#				endless loop due to conflicts with use of $i in other functions.
#     06/13/06  Gary Meyer	Added fru summary details to MD detected errors in SH2_MEM_FIRST_ERROR.
#				Added WAR for BW# 953156.
#     06/14/06  Gary Meyer	Fix for case where multiple errors logged by the MD cause the suspect dimm location
#				to be attached to the wrong fru entry.  Added SH2_JUNK_BUS_TIME mmr.
#     06/19/06  Gary Meyer	Added check for ADVERT_BOOTED bit in SH2_SCRATCH1.
#     06/23/06  Gary Meyer	Added get_bp_connect() to indicate blade to router connection in FRU message.
#				Will flag fsb related config checks only on nodes having  > 0 processors.
#				Adding source node from SH2_PI_ERROR_DETAIL2 to FRU list if it's different
#				than the source node from the address found in SH2_PI_ERROR_DETAIL1.
#     06/26/06  Gary Meyer      Handle the case where a REQ_FORMAT or ADDR_ACCESS error was found in
#				SH2_PI_FIRST_ERROR, but later occurring FSB HW errors were also logged.
#     06/27/06  Gary Meyer      Updated so all ECC error detail fields get printed in the error summary.
#     06/30/06  Gary Meyer      Updated to better report the case of a unresponsive blade, and to improve
#				output generated with multiple error signatures per node.
#     07/03/06  Gary Meyer      Add SPEC_NASID assignment if the node wasn't booted properly, the nasid value is 0,
#				and a chiplet nomatch header has been captured by the XN.
#     07/05/06  Gary Meyer	Fixed formatting problem with error signature output.
#     07/11/06  Gary Meyer	Fixed some minor issues for handling failed jtag reads.
#     07/14/06  Gary Meyer	Fixed bug with generated the correct rtrdmp command line with PI Msg Len errors.
#     07/17/06  Gary Meyer	Now providing the COR & UNC values to decode_dir_state().
#     07/19/06  Gary Meyer	Added print statement when assigning SPEC_NASID.  Will re-assign the nasid value
#				of the node if both NI ports have the same destination node found in the chiplet
#				nomatch error detail.
#     07/24/06  Gary Meyer      Update to ensure all NASID values use 4 characters.
#     07/27/06  Gary Meyer      No longer derive source node for FRU from error detail2 with PI detected MSG_LEN error.
#				Display all fields in PI VC_STATUS mmrs.  Added more detail to fru reason for XN errors.
#     07/31/06  Gary Meyer      Only update $global_err_val with $SHUB_UCE_ERR for PI UCE's with syndrome != 0xFF.
#     08/01/06  Gary Meyer      Added SH2_CRB_MESSAGE_CONTROL mmr to the fast list.
#     08/02/06  Gary Meyer      Qualify the $fsb_stalled indication with PI CRB overuse in capture_sh2_pi_dbg_01().
#     08/10/06  Gary Meyer      Update to ensure valid PI shadow read ECC detail data doesn't get ignored.
#     08/23/06  Gary Meyer      Changed to have the PSP scratch mmrs occur after the Shadow ECC mmrs.  This is to
#				prevent FRU scoring problems if UCE's are only logged in the Shadow ECC mmrs.
#     08/31/06  Gary Meyer      Added XN QUEUE_CNTR mmrs.  Undid change from 08/10/06 to ensure all ECC detail regs
#				are qualified with a corresponding bit in the Shadow PI error summary or first error.
#     09/01/06  Gary Meyer      Now setting $shub_hw_err with any FIRST_ERROR occurrence to prevent fru scoring problems.
#     09/08/06  Gary Meyer      Will skip parse_sh2_mmr_list() if mmr accesses all fail in set_pio_timeout();
#     09/15/06  Gary Meyer      Added code to determine if a failing PIO was to local mmr space.  This is to prevent
#				skewing of FRU scoring and to prevent possible confusion regarding error detail.
#     09/19/06  Gary Meyer      Update in display_sh2_mem_cfg() to handle the case where a DEAD node causes output
#				list to be truncated.
#     09/22/06  Gary Meyer      Fix to chk_addr_mode() to handle the case where the read to the first Shub2 node fails.
#				This is to prevent using the default value for $nasid_index if that should occur.
#     10/18/06  Gary Meyer      Added check for global variables before dumping shub test points or PI CRB's.  
#				Initialize $SSCOPE_ERR before doing any jtag accesses.
#     10/28/06  Gary Meyer      Fixed for problem with fsb sw errors.  Fixed typo for FSB CE errors.
#     10/31/06  Gary Meyer      Added TSF & DIAG_MSG mmrs.
#     11/03/06  Gary Meyer      Will add Wedged Shub nodes to fru list.  Generate scandump message for these cases.
#     11/06/06  Gary Meyer      Recognize & report poisoned data with FSB_SHUB_UCE.  Prevent multiple proc errors on the same
#                               node from increasing the fru value unnecessarily.
#     11/06/06  Gary Meyer      Added DIMM_SUMM header for each node.
#     11/08/06  Gary Meyer      Added supporting functions for monitor_mode - BW 951089.
#     11/09/06  Gary Meyer      Add retry before moving on to the next shub node if Nack hammer disable/re-enable doesn't
#                               correct the hang.
#     11/09/06  Gary Meyer      Bug fix.  Wasn't capturing the flagged AMOPW_ERR registers.  Also, was parsing them wrong.
#     11/09/06  Gary Meyer      Added quick_crb_read() - will read all PI crb's while nack hammer is disabled to see if 
#                               that helps get the nack hammer state machine operational.
#     11/13/06  Gary Meyer      Added support for detection of address in header corruption signature.
#     11/16/06  Gary Meyer      Save rtrdmp parameters in $rtd_list for later calls to trace_numalink_route().
#     11/21/06  Gary Meyer      Fix for issue with decoding the SSYN field from AMOPW_ERR registers.  The actual syndrome
#				is in the upper 8 bits of the 16 bit SSYN field.  Print blade sn with hardware cfg info.
#				Added MD shadow ECC registers.  Parsing for these will have to come later.
#     11/21/06  Gary Meyer      First attempt to fix mysterious issue with parsing Shub2 ECC shadow mmrs.  Either the
#                               mmr_data is getting modified sometimes by the reg_set_value() function, or variables local
#                 		to the parse_sh2_mmr_list() function are getting modified somehow in a unexpected way. 
#     11/22/06  Gary Meyer      One more fix for issue with decoding the SSYN field from AMOPW_ERR registers.  The syndrome
#			     	bits are found in the lower or upper half of RSYN or SSYN depending on the failing address.
#     11/22/06  Gary Meyer      Temporary fix for issue with double meanings for ECC shadow mmrs.  Waiting for mrl to update
#				BW# 945281 before final fix can be done.
#     11/28/06  Gary Meyer      Updated to recognize IP79 & IP81 blade types.
#     11/29/06  Gary Meyer      Added check for correct prom version.
#     12/01/06  Gary Meyer      Added code to recognize signature for BW 945998.
#     12/04/06  Gary Meyer      Correcting issue with duplicate scratch mmrs for MD ECC shadow regs.  Changes to get directory
#				state from the proper blades on non-booted nodes.
#     12/16/06  Gary Meyer      Update to ensure valid PI shadow read ECC detail data doesn't get ignored.  If the PI_UCE bit
#                               is set in the shadow event occurred, or SHUB_FSB_UCE bit is set in shadow error summary, report it.
#				Globalizing some variables to allow address in header corruption flagging in file mode.
#     12/19/06  Gary Meyer      Changes to ensure correct fru scoring for Tollhouse IRU routers.
#     12/20/06  Gary Meyer      Improve formatting of backplane connection info in fru description.
#				Added L3 rtrdmp command line message if numalink corruption is found.
#     12/21/06  Gary Meyer      Fix for address in header corruption error reporting.
#     12/23/06  Gary Meyer      Update to handle MD directory UCE reporting issue in BW 959587.  Convert to use save_fru() .
#     01/02/07  Gary Meyer      Fix for address in header corruption error reporting - even_odd cacheline problem.
#     01/04/07  Gary Meyer      Corrected bug where DIMM location for Shadow ECC errors were not found in fru summary.
#				This was introduced with changes from 12/23/06.
#     01/11/07  Gary Meyer      Updated to recognize Montvale processor types.
#     01/22/07  Gary Meyer      Need to ignore FSB test point problems if there aren't any processors.
#     03-14-07  Gary Meyer      Update for physical dimm group remapping.
#     04-14-07  Gary Meyer      Need to ignore stale PI_ERROR_DETAIL1 & 2 if only data errors are found in PI_FIRST_ERROR or
#				PI_ERROR_SUMMARY or OVERFLOW.  Using $pi_dtl_vld to qualify.
#     04/16/07  Gary Meyer      Added code to recognize signature for BW 963416.
#     05/04/07  Gary Meyer      Update to ensure blade header gets printed even if Jtag reads fail.
#     05/15/07  Gary Meyer      Update to prevent bogus info from Shadow reg's from getting into error summary.
#     10/04/07  Gary Meyer      Added decode of LUT_ADDR from SH2_XNIILB_ERROR_DETAIL_3.
#
#-------------------------------------------------------------------------------

@SHUB2_REG_LIST = (
			"SH2_NODE_ID",			# MUST be FIRST in the list!
			"SH2_PROCESSOR_TYPE",
                        "SH2_RACK_SLOT_DATA",
			"SH2_ADDRESS_MODE",
                        "SH2_EVENT_OCCURRED",
                        "SH2_EVENT_OVERFLOW",
                        "SH2_FIRST_ERROR",

			"SH2_RTC",
			"SH2_GC_FIL_CTRL",
			"SH2_GC_SRC_CTRL",
                        "SH2_LB_PI_GCLK_SEL",

                        "SH2_CPU_PERM0",
                        "SH2_CPU_PERM1",
                        "SH2_CPU_PERM2",
                        "SH2_CPU_PERM3",
                        "SH2_CPU_PERM_OVR",
                        "SH2_EXT_IO_PERM0",
                        "SH2_EXT_IO_PERM1",
                        "SH2_EXT_IO_PERM2",
                        "SH2_EXT_IO_PERM3",
                        "SH2_EXT_IOI_ACCESS0",
                        "SH2_EXT_IOI_ACCESS1",
                        "SH2_EXT_IOI_ACCESS2",
                        "SH2_EXT_IOI_ACCESS3",
                        "SH2_NODE_PRESENT_TABLE",
                        "SH2_MD_DQLP_MMR_DIR_CONFIG",
                        "SH2_MD_DQLP_MMR_DIR_PRESVEC0",
                        "SH2_MD_DQLP_MMR_DIR_PRESVEC1",
                        "SH2_MD_DQLP_MMR_DIR_PRESVEC2",
                        "SH2_MD_DQLP_MMR_DIR_PRESVEC3",
                        "SH2_MD_DQRP_MMR_DIR_CONFIG",
                        "SH2_MD_DQRP_MMR_DIR_PRESVEC0",
                        "SH2_MD_DQRP_MMR_DIR_PRESVEC1",
                        "SH2_MD_DQRP_MMR_DIR_PRESVEC2",
                        "SH2_MD_DQRP_MMR_DIR_PRESVEC3",

			"SH2_PTC",
			"SH2_PTC_PARMS",
                        "SH2_SCRATCH0",
                        "SH2_SCRATCH1",
                        "SH2_SCRATCH2",
                        "SH2_SCRATCH3",
                        "SH2_SCRATCH4",
                        "SH2_SCRATCH5",
                        "SH2_SCRATCH6",
                        "SH2_SCRATCH7",
                        "SH2_SCRATCH8",
                        "SH2_SCRATCH9",
                        "SH2_SCRATCH10",
                        "SH2_SCRATCH11",
                        "SH2_SCRATCH12",
                        "SH2_SCRATCH13",
                        "SH2_SCRATCH14",
                        "SH2_SCRATCH15",
                        "SH2_SCRATCH16",
                        "SH2_SCRATCH17",
                        "SH2_SCRATCH18",
                        "SH2_SCRATCH19",
                        "SH2_SCRATCH20",
                        "SH2_SCRATCH21",
                        "SH2_SCRATCH22",
                        "SH2_SCRATCH23",
                        "SH2_SCRATCH24",
                        "SH2_XN_MD_IILB_CMP_EXP_DATA0",		# CPU A Stash
                        "SH2_XN_MD_IILB_CMP_EXP_DATA1",		# CPU B Stash
                        "SH2_XN_MD_IILB_CMP_ENABLE0",		# CPU C Stash
                        "SH2_XN_MD_IILB_CMP_ENABLE1",		# CPU D Stash
                        "SH2_XN_MD_NI0_CMP_EXP_DATA0",		# Locks
                        "SH2_XN_MD_NI0_CMP_EXP_DATA1",		# Flags
                        "SH2_XN_MD_NI0_CMP_ENABLE0",		# DIP Switches
                        "SH2_XN_MD_NI0_CMP_ENABLE1",		# Barrier Complete
                        "SH2_XN_MD_NI1_CMP_EXP_DATA0",		# Progress
                        "SH2_XN_MD_NI1_CMP_ENABLE0",		# Rendezvous
                        "SH2_XN_MD_SIC_CMP_HDR_ENABLE0",	# RZ Flags 0
                        "SH2_XN_MD_SIC_CMP_HDR_ENABLE1",	# RZ Flags 1
                        "SH2_XN_MD_NI1_CMP_EXP_DATA1",		# RZ Flags 2
                        "SH2_XN_MD_NI1_CMP_ENABLE1",		# RZ Flags 3

                        "SH2_XN_MD_SIC_CMP_EXP_HDR0",		# CPU A Target Identifier
                        "SH2_XN_MD_SIC_CMP_EXP_HDR1",		# CPU A Target Id Timestamp
                        "SH2_XN_NI0_MD_CMP_EXP_HDR1",		# CPU B Target Identifier
                        "SH2_XN_NI0_MD_CMP_EXP_HDR0",		# CPU B Target Id Timestamp
                        "SH2_XN_IILB_NI1_CMP_EXP_DATA0",	# CPU C Target Identifier
                        "SH2_XN_IILB_NI1_CMP_EXP_DATA1",	# CPU C Target Id Timestamp
                        "SH2_XN_IILB_NI0_CMP_EXP_DATA0",	# CPU D Target Identifier
                        "SH2_XN_IILB_NI0_CMP_EXP_DATA1",	# CPU D Target Id Timestamp

			"SH2_FIRMWARE_REMAP",
			"SH2_VGA_REMAP",
			"SH2_FSB_BINIT_CONTROL",
			"SH2_FSB_POWER_ON_CONFIG",
			"SH2_FSB_RESET_CONTROL",
			"SH2_FSB_SYSTEM_AGENT_CONFIG",
			"SH2_FSB_IO_READ_TRANS",
			"SH2_FSB_IO_READ_TRANS_ODD",
			"SH2_FSB_RESET_STATUS",
			"SH2_FSB_SYMMETRIC_AGENT_STATUS",
                        "SH2_CRB_MESSAGE_CONTROL",
                        "SH2_CRB_NACK_LIMIT",
                        "SH2_CRB_TIMEOUT_PRESCALE",
                        "SH2_CRB_TIMEOUT_SKID",

                        "SH2_EXTIO_INT0",
                        "SH2_EXTIO_INT0_CONFIG",
                        "SH2_EXTIO_INT0_ENABLE",
                        "SH2_EXTIO_INT1",
                        "SH2_EXTIO_INT1_CONFIG",
                        "SH2_EXTIO_INT1_ENABLE",
                        "SH2_EXTIO_INT2",
                        "SH2_EXTIO_INT2_CONFIG",
                        "SH2_EXTIO_INT2_ENABLE",
                        "SH2_EXTIO_INT3",
                        "SH2_EXTIO_INT3_CONFIG",
                        "SH2_EXTIO_INT3_ENABLE",
                        "SH2_INT_NODE_ID_CONFIG",
                        "SH2_IPI_INT",
                        "SH2_IPI_INT_ENABLE",
                        "SH2_IPI_ACCESS0",
                        "SH2_IPI_ACCESS1",
                        "SH2_IPI_ACCESS2",
                        "SH2_IPI_ACCESS3",
                        "SH2_LOCAL_INT0_CONFIG",
                        "SH2_LOCAL_INT0_ENABLE",
                        "SH2_LOCAL_INT1_CONFIG",
                        "SH2_LOCAL_INT1_ENABLE",
                        "SH2_LOCAL_INT2_CONFIG",
                        "SH2_LOCAL_INT2_ENABLE",
                        "SH2_LOCAL_INT3_CONFIG",
                        "SH2_LOCAL_INT3_ENABLE",
                        "SH2_LOCAL_INT4_CONFIG",
                        "SH2_LOCAL_INT4_ENABLE",
                        "SH2_LOCAL_INT5_CONFIG",
                        "SH2_LOCAL_INT5_ENABLE",
                        "SH2_LOCAL_INT6_CONFIG",
                        "SH2_LOCAL_INT6_ENABLE",
                        "SH2_LOCAL_INT7_CONFIG",
                        "SH2_LOCAL_INT7_ENABLE",
                        "SH2_LOCAL_INT8_CONFIG",
                        "SH2_LOCAL_INT8_ENABLE",
                        "SH2_LOCAL_INT9_CONFIG",
                        "SH2_LOCAL_INT9_ENABLE",
                        "SH2_PROFILE_INT_CONFIG",
                        "SH2_PROFILE_INT_ENABLE",
                        "SH2_RTC0_INT_CONFIG",
                        "SH2_RTC0_INT_ENABLE",
                        "SH2_RTC1_INT_CONFIG",
                        "SH2_RTC1_INT_ENABLE",
                        "SH2_RTC2_INT_CONFIG",
                        "SH2_RTC2_INT_ENABLE",
                        "SH2_RTC3_INT_CONFIG",
                        "SH2_RTC3_INT_ENABLE",
                        "SH2_BT0_INT_CONFIG",
                        "SH2_BT0_INT_ENABLE",
                        "SH2_BT1_INT_CONFIG",
                        "SH2_BT1_INT_ENABLE",
                        "SH2_BT2_INT_CONFIG",
                        "SH2_BT2_INT_ENABLE",
                        "SH2_BT3_INT_CONFIG",
                        "SH2_BT3_INT_ENABLE",
                        "SH2_NOTIFICATION_MEMORY_INT0_CONFIG",
                        "SH2_NOTIFICATION_MEMORY_INT0_ENABLE",
                        "SH2_NOTIFICATION_MEMORY_INT1_CONFIG",
                        "SH2_NOTIFICATION_MEMORY_INT1_ENABLE",
                        "SH2_NOTIFICATION_MEMORY_INT2_CONFIG",
                        "SH2_NOTIFICATION_MEMORY_INT2_ENABLE",
                        "SH2_NOTIFICATION_MEMORY_INT3_CONFIG",
                        "SH2_NOTIFICATION_MEMORY_INT3_ENABLE",

			"SH2_TSF_DIAG_MSG_CTL",
			"SH2_DIAG_MSG_CNTRL",
			"SH2_DIAG_MSG_HDR0",
			"SH2_DIAG_MSG_HDR1",
			"SH2_TSF_ARMED_STATE",
			"SH2_TSF_ARM_MASK",
			"SH2_TSF_DISARM_MASK",
			"SH2_TSF_COUNTER_PRESETS",
			"SH2_TSF_COUNTER_VALUE",
			"SH2_TSF_DECREMENT_CTL",
			"SH2_TSF_SOFTWARE_ARM",
			"SH2_TSF_SOFTWARE_DISARM",
			"SH2_TSF_ENABLE_CTL",
			"SH2_TSF_SOFTWARE_TRIGGERED",
			"SH2_TSF_TRIGGERED_STATE",
			"SH2_TSF_TRIGGER_MASK",

                        "SH2_DEBUG_SELECT",
                        "SH2_DEBUG_PORT_CONFIG",
                        "SH2_PERFORMANCE_COUNT0_CONTROL",
                        "SH2_PERFORMANCE_COUNT1_CONTROL",
                        "SH2_PERFORMANCE_COUNT2_CONTROL",
                        "SH2_PERFORMANCE_COUNT3_CONTROL",
                        "SH2_PERFORMANCE_COUNT4_CONTROL",
                        "SH2_PERFORMANCE_COUNT5_CONTROL",
                        "SH2_PERFORMANCE_COUNT6_CONTROL",
                        "SH2_PERFORMANCE_COUNT7_CONTROL",
                        "SH2_PERFORMANCE_COUNTER0",
                        "SH2_PERFORMANCE_COUNTER1",
                        "SH2_PERFORMANCE_COUNTER2",
                        "SH2_PERFORMANCE_COUNTER3",
                        "SH2_PERFORMANCE_COUNTER4",
                        "SH2_PERFORMANCE_COUNTER5",
                        "SH2_PERFORMANCE_COUNTER6",
                        "SH2_PERFORMANCE_COUNTER7",
                        "SH2_PROFILE_COUNTER",
                        "SH2_PROFILE_PEAK",

                        "SH2_MEM_CREDITS",		
                        "SH2_MEM_REDUCE_XN_TNUMS", 	
                        "SH2_XN_MD_RP_CRD_CTL", 	
                        "SH2_XN_MD_RQ_CRD_CTL", 	
                        "SH2_PI_MD_RP_CRD_CTL", 	
                        "SH2_PI_MD_RQ_CRD_CTL", 	
                        "SH2_PIO_RQ_CRD_CTL", 	
                        "SH2_MISC_MEM_CFG",
                        "SH2_MEM_RED_BLACK",
                        "SH2_X_DIMM_CFG",
                        "SH2_Y_DIMM_CFG",
                        "SH2_JNR_DIMM_CFG",
                        "SH2_X_REFRESH_CONTROL",
                        "SH2_Y_REFRESH_CONTROL",
                        "SH2_X_CFG",
                        "SH2_Y_CFG",
                       	"SH2_X_DQCT_CFG",
                        "SH2_Y_DQCT_CFG",
                        "SH2_X_ODT_CFG",
                        "SH2_Y_ODT_CFG",
                        "SH2_MD_DQLP_MMR_DIR_LOCVEC0",
                        "SH2_MD_DQLP_MMR_DIR_LOCVEC1",
                        "SH2_MD_DQLP_MMR_DIR_LOCVEC2",
                        "SH2_MD_DQLP_MMR_DIR_LOCVEC3",
                        "SH2_MD_DQLP_MMR_DIR_LOCVEC4",
                        "SH2_MD_DQLP_MMR_DIR_LOCVEC5",
                        "SH2_MD_DQLP_MMR_DIR_LOCVEC6",
                        "SH2_MD_DQLP_MMR_DIR_LOCVEC7",
                        "SH2_MD_DQLP_MMR_DIR_PRIVEC0",
                        "SH2_MD_DQLP_MMR_DIR_PRIVEC1",
                        "SH2_MD_DQLP_MMR_DIR_PRIVEC2",
                        "SH2_MD_DQLP_MMR_DIR_PRIVEC3",
                        "SH2_MD_DQLP_MMR_DIR_PRIVEC4",
                        "SH2_MD_DQLP_MMR_DIR_PRIVEC5",
                        "SH2_MD_DQLP_MMR_DIR_PRIVEC6",
                        "SH2_MD_DQLP_MMR_DIR_PRIVEC7",
                        "SH2_MD_DQRP_MMR_DIR_LOCVEC0",
                        "SH2_MD_DQRP_MMR_DIR_LOCVEC1",
                        "SH2_MD_DQRP_MMR_DIR_LOCVEC2",
                        "SH2_MD_DQRP_MMR_DIR_LOCVEC3",
                        "SH2_MD_DQRP_MMR_DIR_LOCVEC4",
                        "SH2_MD_DQRP_MMR_DIR_LOCVEC5",
                        "SH2_MD_DQRP_MMR_DIR_LOCVEC6",
                        "SH2_MD_DQRP_MMR_DIR_LOCVEC7",
                        "SH2_MD_DQRP_MMR_DIR_PRIVEC0",
                        "SH2_MD_DQRP_MMR_DIR_PRIVEC1",
                        "SH2_MD_DQRP_MMR_DIR_PRIVEC2",
                        "SH2_MD_DQRP_MMR_DIR_PRIVEC3",
                        "SH2_MD_DQRP_MMR_DIR_PRIVEC4",
                        "SH2_MD_DQRP_MMR_DIR_PRIVEC5",
                        "SH2_MD_DQRP_MMR_DIR_PRIVEC6",
                        "SH2_MD_DQRP_MMR_DIR_PRIVEC7",

                        "SH2_PI_FIRST_ERROR",
                        "SH2_PI_ERROR_SUMMARY",
                        "SH2_PI_ERROR_OVERFLOW",
                        "SH2_PI_ERROR_MASK",
                        "SH2_PI_AFI_ERROR_MASK",
                        "SH2_PI_CRBP_ERROR_MASK",
                        "SH2_PI_RDAT_ERROR_MASK",
                        "SH2_PI_WDAT_ERROR_MASK",
                        "SH2_PI_ERROR_DETAIL_1",
                        "SH2_PI_ERROR_DETAIL_2",
                        "SH2_PI_HW_TIME_STAMP",
                        "SH2_PI_CORRECTED_DETAIL_1",
                        "SH2_PI_CORRECTED_DETAIL_2",
                        "SH2_PI_CORRECTED_DETAIL_3",
                        "SH2_PI_CORRECTED_DETAIL_4",
                        "SH2_PI_CORRECTED_DETAIL_5",
                        "SH2_PI_CORRECTED_DETAIL_6",
                        "SH2_PI_COR_TIME_STAMP",
                        "SH2_PI_UNCORRECTED_DETAIL_1",
                        "SH2_PI_UNCORRECTED_DETAIL_2",
                        "SH2_PI_UNCORRECTED_DETAIL_3",
                        "SH2_PI_UNCORRECTED_DETAIL_4",
                        "SH2_PI_UNCORRECTED_DETAIL_5",
                        "SH2_PI_UNCORRECTED_DETAIL_6",
                        "SH2_PI_UNCOR_TIME_STAMP",

                        "SH2_PIO_WRITE_STATUS_00",
                        "SH2_PIO_WRITE_STATUS_01",
                        "SH2_PIO_WRITE_STATUS_10",
                        "SH2_PIO_WRITE_STATUS_11",
                        "SH2_PIO_WC_WRITE_STATUS_00",
                        "SH2_PIO_WC_WRITE_STATUS_01",
                        "SH2_PIO_WC_WRITE_STATUS_10",
                        "SH2_PIO_WC_WRITE_STATUS_11",
                        "SH2_MEMORY_WRITE_STATUS_00",
                        "SH2_MEMORY_WRITE_STATUS_01",
                        "SH2_MEMORY_WRITE_STATUS_10",
                        "SH2_MEMORY_WRITE_STATUS_11",
                        "SH2_MEMORY_WRITE_STATUS_NON_USER_00",
                        "SH2_MEMORY_WRITE_STATUS_NON_USER_01",
                        "SH2_MEMORY_WRITE_STATUS_NON_USER_10",
                        "SH2_MEMORY_WRITE_STATUS_NON_USER_11",
                        "SH2_PI_MD2PI_REPLY_VC_CONFIG",
                        "SH2_PI_MD2PI_REQUEST_VC_CONFIG",
                        "SH2_PI_MD2PI_REPLY_VC_STATUS",
                        "SH2_PI_MD2PI_REQUEST_VC_STATUS",
                        "SH2_PI_XN2PI_REPLY_VC_CONFIG",
                        "SH2_PI_XN2PI_REQUEST_VC_CONFIG",
                        "SH2_PI_XN2PI_REPLY_VC_STATUS",
                        "SH2_PI_XN2PI_REQUEST_VC_STATUS",	# This reg also has ICC status
                        "SH2_PI_PI2MD_REPLY_VC_STATUS",
                        "SH2_PI_PI2MD_REQUEST_VC_STATUS",
                        "SH2_PI_PI2XN_REPLY_VC_STATUS",
                        "SH2_PI_PI2XN_REQUEST_VC_STATUS",
                        "SH2_PI_CRB_STALL_SKID",
			"SH2_PI_WDAT_REPLAY_COUNT",
			"SH2_PI_WDAT_FSB_ACK_Q_CONFIG",
			"SH2_PI_WDAT_FSB_CMND_Q_CONFIG",
			"SH2_PI_RDAT_ACK_Q_CONFIG",
			"SH2_PI_RDAT_DAT_DFR_Q_CONFIG",

                        "SH2_LB_FIRST_ERROR",
                        "SH2_LB_ERROR_SUMMARY",
                        "SH2_LB_ERROR_OVERFLOW",
                        "SH2_LB_ERROR_MASK",
                        "SH2_LB_ERROR_DETAIL_1",
                        "SH2_LB_ERROR_DETAIL_2",
                        "SH2_LB_ERROR_DETAIL_3",
                        "SH2_LB_ERROR_DETAIL_4",
                        "SH2_LB_ERROR_DETAIL_5",
                        "SH2_LB_ERROR_DETAIL_6",
                        "SH2_LB_HW_TIME_STAMP",
                        "SH2_LB_LAST_CREDIT",
                        "SH2_LB_NACK_STATUS0",
                        "SH2_LB_NACK_STATUS1",
                        "SH2_LB_NACK_STATUS2",
                        "SH2_LB_NACK_STATUS3",
                        "SH2_PIO_NACK_RESET",
                        "SH2_CONVEYOR_BELT_TIME_OUT",
                        "SH2_CONVEYOR_BELT_CONFIG",
                        "SH2_PIO_TIME_OUT",
                        "SH2_JUNK_BUS_TIME",
                        "SH2_LB_CREDIT_STATUS",

                        "SH2_XN_FIRST_ERROR",
                        "SH2_XN_ERROR_SUMMARY",
                        "SH2_XN_ERROR_OVERFLOW",
                        "SH2_XN_ERROR_MASK",
                        "SH2_XN_HW_TIME_STAMP",
                        "SH2_XN_CORRECTED_DETAIL_1",
                        "SH2_XN_CORRECTED_DETAIL_2",
                        "SH2_XN_CORRECTED_DETAIL_3",
                        "SH2_XN_CORRECTED_DETAIL_4",
                        "SH2_XN_COR_TIME_STAMP",
                        "SH2_XN_UNCORRECTED_DETAIL_1",
                        "SH2_XN_UNCORRECTED_DETAIL_2",
                        "SH2_XN_UNCORRECTED_DETAIL_3",
                        "SH2_XN_UNCORRECTED_DETAIL_4",
                        "SH2_XN_UNCOR_TIME_STAMP",

                        "SH2_NI0_FIRST_ERROR_1",
                        "SH2_NI0_FIRST_ERROR_2",
                        "SH2_NI0_ERROR_SUMMARY_1",
                        "SH2_NI0_ERROR_SUMMARY_2",
                        "SH2_NI0_ERROR_OVERFLOW_1",
                        "SH2_NI0_ERROR_OVERFLOW_2",
                        "SH2_NI0_ERROR_MASK_1",
                        "SH2_NI0_ERROR_MASK_2",
                        "SH2_NI0_ERROR_DETAIL_1",
                        "SH2_NI0_ERROR_DETAIL_2",
                        "SH2_NI0_ERROR_DETAIL_3",

                        "SH2_NI1_FIRST_ERROR_1",
                        "SH2_NI1_FIRST_ERROR_2",
                        "SH2_NI1_ERROR_SUMMARY_1",
                        "SH2_NI1_ERROR_SUMMARY_2",
                        "SH2_NI1_ERROR_OVERFLOW_1",
                        "SH2_NI1_ERROR_OVERFLOW_2",
                        "SH2_NI1_ERROR_MASK_1",
                        "SH2_NI1_ERROR_MASK_2",
                        "SH2_NI1_ERROR_DETAIL_1",
                        "SH2_NI1_ERROR_DETAIL_2",
                        "SH2_NI1_ERROR_DETAIL_3",

                        "SH2_XNIILB_FIRST_ERROR",
                        "SH2_XNIILB_FIRST_ERROR_1",
                        "SH2_XNIILB_ERROR_SUMMARY",
                        "SH2_XNIILB_ERROR_SUMMARY_1",
                        "SH2_XNIILB_ERROR_OVERFLOW",
                        "SH2_XNIILB_ERROR_OVERFLOW_1",
                        "SH2_XNIILB_ERROR_MASK",
                        "SH2_XNIILB_ERROR_MASK_1",
                        "SH2_XNIILB_ERROR_DETAIL_1",
                        "SH2_XNIILB_ERROR_DETAIL_2",
                        "SH2_XNIILB_ERROR_DETAIL_3",

                        "SH2_XNPI_FIRST_ERROR",
                        "SH2_XNPI_ERROR_SUMMARY",
                        "SH2_XNPI_ERROR_OVERFLOW",
                        "SH2_XNPI_ERROR_MASK",
                        "SH2_XNPI_ERROR_DETAIL_1",

                        "SH2_XNMD_FIRST_ERROR",
                        "SH2_XNMD_FIRST_ERROR_1",
                       	"SH2_XNMD_ERROR_SUMMARY",
                        "SH2_XNMD_ERROR_SUMMARY_1",
                        "SH2_XNMD_ERROR_OVERFLOW",
                        "SH2_XNMD_ERROR_OVERFLOW_1",
                        "SH2_XNMD_ERROR_MASK",
                        "SH2_XNMD_ERROR_MASK_1",
                        "SH2_XNMD_ERROR_DETAIL_1",

                        "SH2_MEM_FIRST_ERROR",
                        "SH2_MEM_ERROR_SUMMARY",
                        "SH2_MEM_ERROR_OVERFLOW",
                        "SH2_MEM_ERROR_MASK",
                        "SH2_MISC_ERR_HDR_LOWER",
                        "SH2_MISC_ERR_HDR_UPPER",
                        "SH2_MISC_ADRS_ERR_HDR_LOWER_A",
                        "SH2_MD_HW_TIME_STAMP",
                        "SH2_DIR_COR_ERR_HDR_LOWER",
                        "SH2_DIR_COR_ERR_HDR_UPPER",
                        "SH2_DIR_COR_ADRS_ERR_HDR_LOWER_A",
                        "SH2_MD_COR_TIME_STAMP",
                        "SH2_DIR_UC_ERR_HDR_LOWER",
                        "SH2_DIR_UC_ERR_HDR_UPPER",
                        "SH2_DIR_UC_ADRS_ERR_HDR_LOWER_A",
                        "SH2_MD_UNCOR_TIME_STAMP",
                        "SH2_MMRBIST_BASE",
                        "SH2_MMRBIST_CTL",
                        "SH2_MMRBIST_ERR",
                        "SH2_MEM_CAPTURE_ADDR",
                        "SH2_MEM_CAPTURE_MASK",
                        "SH2_MEM_CAPTURE_HDR",
                        "SH2_MEM_CAPTURE_HDR_ADDR",
                        "SH2_MD_DQLP_MMR_XCERR1",
                        "SH2_MD_DQLP_MMR_XCERR2",
                        "SH2_MD_DQLP_MMR_XUERR1",
                        "SH2_MD_DQLP_MMR_XUERR2",
                        "SH2_MD_DQLP_MMR_XPERR0",
                        "SH2_MD_DQLP_MMR_XPERR1",
                        "SH2_MD_DQLP_MMR_YCERR1",
                        "SH2_MD_DQLP_MMR_YCERR2",
                        "SH2_MD_DQLP_MMR_YUERR1",
                        "SH2_MD_DQLP_MMR_YUERR2",
                        "SH2_MD_DQLP_MMR_YPERR0",
                        "SH2_MD_DQLP_MMR_YPERR1",
                        "SH2_MD_DQRP_MMR_XCERR1",
                        "SH2_MD_DQRP_MMR_XCERR2",
                        "SH2_MD_DQRP_MMR_XUERR1",
                        "SH2_MD_DQRP_MMR_XUERR2",
                        "SH2_MD_DQRP_MMR_XPERR0",
                        "SH2_MD_DQRP_MMR_XPERR1",
                        "SH2_MD_DQRP_MMR_YCERR1",
                        "SH2_MD_DQRP_MMR_YCERR2",
                        "SH2_MD_DQRP_MMR_YUERR1",
                        "SH2_MD_DQRP_MMR_YUERR2",
                        "SH2_MD_DQRP_MMR_YPERR0",
                        "SH2_MD_DQRP_MMR_YPERR1",
                        "SH2_MD_DQLS_MMR_XAMOPW_ERR",
                        "SH2_MD_DQRS_MMR_YAMOPW_ERR",



                        "SH2_BT_FIRST_ERROR",
                        "SH2_BT_ERROR_SUMMARY",
                        "SH2_BT_ERROR_OVERFLOW",
                        "SH2_BT_ERROR_MASK",
                        "SH2_BT_HW_TIME_STAMP",
                        "SH2_BT_XN_ERROR_DETAIL_1",
                        "SH2_BT_XN_ERROR_DETAIL_2",
                        "SH2_BT_ENG_ERROR_STAT",
			"SH2_BT_ENG_CSR_0",
			"SH2_BT_ENG_CSR_1",
			"SH2_BT_ENG_CSR_2",
			"SH2_BT_ENG_CSR_3",
                        "SH2_BT_IQ_STAT",
                        "SH2_BT_CRB_STAT_0",
                        "SH2_BT_CRB_SRC_ADDR_0",
                        "SH2_BT_CRB_DEST_ADDR_0",
                        "SH2_BT_CRB_STAT_1",
                        "SH2_BT_CRB_SRC_ADDR_1",
                        "SH2_BT_CRB_DEST_ADDR_1",
                        "SH2_BT_CRB_STAT_2",
                        "SH2_BT_CRB_SRC_ADDR_2",
                        "SH2_BT_CRB_DEST_ADDR_2",
                        "SH2_BT_CRB_STAT_3",
                        "SH2_BT_CRB_SRC_ADDR_3",
                        "SH2_BT_CRB_DEST_ADDR_3",
                        "SH2_BT_CRB_STAT_4",
                        "SH2_BT_CRB_SRC_ADDR_4",
                        "SH2_BT_CRB_DEST_ADDR_4",
                        "SH2_BT_CRB_STAT_5",
                        "SH2_BT_CRB_SRC_ADDR_5",
                        "SH2_BT_CRB_DEST_ADDR_5",
                        "SH2_BT_CRB_STAT_6",
                        "SH2_BT_CRB_SRC_ADDR_6",
                        "SH2_BT_CRB_DEST_ADDR_6",
                        "SH2_BT_CRB_STAT_7",
                        "SH2_BT_CRB_SRC_ADDR_7",
                        "SH2_BT_CRB_DEST_ADDR_7",
                        "SH2_BT_CRB_STAT_8",
                        "SH2_BT_CRB_SRC_ADDR_8",
                        "SH2_BT_CRB_DEST_ADDR_8",
                        "SH2_BT_CRB_STAT_9",
                        "SH2_BT_CRB_SRC_ADDR_9",
                        "SH2_BT_CRB_DEST_ADDR_9",
                        "SH2_BT_CRB_STAT_10",
                        "SH2_BT_CRB_SRC_ADDR_10",
                        "SH2_BT_CRB_DEST_ADDR_10",
                        "SH2_BT_CRB_STAT_11",
                        "SH2_BT_CRB_SRC_ADDR_11",
                        "SH2_BT_CRB_DEST_ADDR_11",
                        "SH2_BT_CRB_STAT_12",
                        "SH2_BT_CRB_SRC_ADDR_12",
                        "SH2_BT_CRB_DEST_ADDR_12",
                        "SH2_BT_CRB_STAT_13",
                        "SH2_BT_CRB_SRC_ADDR_13",
                        "SH2_BT_CRB_DEST_ADDR_13",
                        "SH2_BT_CRB_STAT_14",
                        "SH2_BT_CRB_SRC_ADDR_14",
                        "SH2_BT_CRB_DEST_ADDR_14",
                        "SH2_BT_CRB_STAT_15",
                        "SH2_BT_CRB_SRC_ADDR_15",
                        "SH2_BT_CRB_DEST_ADDR_15",
			"SH2_BT_ENG_SRC_ADDR_0",
			"SH2_BT_ENG_SRC_ADDR_1",
			"SH2_BT_ENG_SRC_ADDR_2",
			"SH2_BT_ENG_SRC_ADDR_3",
			"SH2_BT_ENG_DEST_ADDR_0",
			"SH2_BT_ENG_DEST_ADDR_1",
			"SH2_BT_ENG_DEST_ADDR_2",
			"SH2_BT_ENG_DEST_ADDR_3",
			"SH2_BT_ENG_NOTIF_ADDR_0",
			"SH2_BT_ENG_NOTIF_ADDR_1",
			"SH2_BT_ENG_NOTIF_ADDR_2",
			"SH2_BT_ENG_NOTIF_ADDR_3",


                        "SH2_NI0_LLP_STAT",
                        "SH2_NI0_LLP_RESET",
                        "SH2_NI0_LLP_RESET_EN",
                        "SH2_NI0_LLP_CHAN_MODE",
                        "SH2_NI0_LLP_CONFIG",
                        "SH2_NI0_LLP_TEST_CTL",
                        "SH2_NI0_LLP_CAPT_WD1",
                        "SH2_NI0_LLP_CAPT_WD2",
                        "SH2_NI0_LLP_CAPT_SBCB",
                        "SH2_NI0_LLP_ERR",
                        "SH2_NI0_LLP_ERR1",
                        "SH2_NI0_LLP_ERR2",
                        "SH2_NI0_LLP_PKT",
                        "SH2_NI1_LLP_STAT",
                        "SH2_NI1_LLP_RESET",
                        "SH2_NI1_LLP_RESET_EN",
                        "SH2_NI1_LLP_CHAN_MODE",
                        "SH2_NI1_LLP_CONFIG",
                        "SH2_NI1_LLP_TEST_CTL",
                        "SH2_NI1_LLP_CAPT_WD1",
                        "SH2_NI1_LLP_CAPT_WD2",
                        "SH2_NI1_LLP_CAPT_SBCB",
                        "SH2_NI1_LLP_ERR",
                        "SH2_NI1_LLP_ERR1",
                        "SH2_NI1_LLP_ERR2",
                        "SH2_NI1_LLP_PKT",

#       Shadow error registers for ECC errors from BW# 945281

			"SH2_XN_IILB_LB_CMP_EXP_DATA1", 	# SH2_EVENT_OCCURRED              
			"SH2_XN_PI_NI0_CMP_ENABLE0", 		# SH2_PI_FIRST_ERROR              
			"SH2_XN_PI_NI0_CMP_ENABLE1", 		# SH2_PI_ERROR_SUMMARY              

			"SH2_XN_IILB_LB_CMP_ENABLE0",		# SH2_PI_CORRECTED_DETAIL_1       
			"SH2_XN_IILB_LB_CMP_ENABLE1", 		# SH2_PI_CORRECTED_DETAIL_2       
			"SH2_XN_IILB_II_CMP_EXP_DATA0",		# SH2_PI_CORRECTED_DETAIL_3       
			"SH2_XN_IILB_II_CMP_EXP_DATA1",		# SH2_PI_CORRECTED_DETAIL_4       
			"SH2_XN_IILB_II_CMP_ENABLE0", 		# SH2_PI_CORRECTED_DETAIL_5       
			"SH2_XN_IILB_II_CMP_ENABLE1", 		# SH2_PI_CORRECTED_DETAIL_6       
			"SH2_XN_NI0_LLP_CMP_ENABLE0", 		# SH2_PI_COR_TIME_STAMP           

			"SH2_XN_NI0_LLP_CMP_ENABLE1", 		# SH2_PI_UNCORRECTED_DETAIL_1     
			"SH2_XN_NI0_MD_CMP_HDR_ENABLE0",	# SH2_PI_UNCORRECTED_DETAIL_2     
			"SH2_XN_NI0_MD_CMP_HDR_ENABLE1",	# SH2_PI_UNCORRECTED_DETAIL_3     
			"SH2_XN_NI1_LLP_CMP_EXP_DATA0",		# SH2_PI_UNCORRECTED_DETAIL_4     
			"SH2_XN_NI1_LLP_CMP_EXP_DATA1",		# SH2_PI_UNCORRECTED_DETAIL_5     
			"SH2_XN_NI1_LLP_CMP_ENABLE0", 		# SH2_PI_UNCORRECTED_DETAIL_6     
			"SH2_XN_NI1_LLP_CMP_ENABLE1",		# SH2_PI_UNCOR_TIME_STAMP         

			"SH2_XN_NI1_MD_CMP_EXP_HDR0", 		# SH2_XN_CORRECTED_DETAIL_1       
			"SH2_XN_NI1_MD_CMP_EXP_HDR1", 		# SH2_XN_CORRECTED_DETAIL_2       
			"SH2_XN_NI1_MD_CMP_HDR_ENABLE0",	# SH2_XN_CORRECTED_DETAIL_3       
			"SH2_XN_NI1_MD_CMP_HDR_ENABLE1",	# SH2_XN_CORRECTED_DETAIL_4       
			"SH2_XN_PI_IILB_CMP_EXP_DATA0",		# SH2_XN_COR_TIME_STAMP           

			"SH2_XN_PI_IILB_CMP_EXP_DATA1",		# SH2_XN_UNCORRECTED_DETAIL_1     
			"SH2_XN_PI_IILB_CMP_ENABLE0", 		# SH2_XN_UNCORRECTED_DETAIL_2     
			"SH2_XN_PI_IILB_CMP_ENABLE1",		# SH2_XN_UNCORRECTED_DETAIL_3     
			"SH2_XN_PI_NI0_CMP_EXP_DATA0", 		# SH2_XN_UNCORRECTED_DETAIL_4     
			"SH2_XN_PI_NI0_CMP_EXP_DATA1", 		# SH2_XN_UNCOR_TIME_STAMP         

			"SH2_XN_NI1_PI_CMP_ENABLE1",		# SH2_MISC_ERR_HDR_LOWER
			"SH2_XN_NI1_PI_CMP_ENABLE0",		# SH2_MISC_ERR_HDR_UPPER
			"SH2_XN_NI1_NI_CMP_EXP_DATA0",		# SH2_MISC_ADRS_ERR_HDR_LOWER_A
			"SH2_XN_NI1_IILB_CMP_ENABLE1",		# SH2_DIR_COR_ERR_HDR_LOWER
			"SH2_XN_NI1_PI_CMP_EXP_DATA0",		# SH2_DIR_COR_ERR_HDR_UPPER
			"SH2_XN_NI1_PI_CMP_EXP_DATA1",		# SH2_DIR_COR_ADRS_ERR_HDR_LOWER_A
			"SH2_XN_NI1_NI_CMP_EXP_DATA1",		# SH2_MD_COR_TIME_STAMP

			"SH2_XN_NI0_NI_CMP_EXP_DATA1",		# SH2_MD_DQLP_MMR_XCERR1
			"SH2_XN_NI0_NI_CMP_ENABLE0",		# SH2_MD_DQLP_MMR_XCERR2
			"SH2_XN_NI0_NI_CMP_ENABLE1",		# SH2_MD_DQLP_MMR_YCERR1
			"SH2_XN_NI1_NI_CMP_ENABLE0",		# SH2_MD_DQLP_MMR_YCERR2
			"SH2_XN_NI1_NI_CMP_ENABLE1",		# SH2_MD_DQLS_MMR_XAMOPW_ERR 
			"SH2_XN_PI_NI1_CMP_EXP_DATA0",		# SH2_MD_DQRP_MMR_XCERR1 
			"SH2_XN_PI_NI1_CMP_EXP_DATA1",		# SH2_MD_DQRP_MMR_XCERR2 
			"SH2_XN_NI1_IILB_CMP_EXP_DATA0",	# SH2_MD_DQRP_MMR_YCERR1 
			"SH2_XN_NI1_IILB_CMP_EXP_DATA1",	# SH2_MD_DQRP_MMR_YCERR2 
			"SH2_XN_NI1_IILB_CMP_ENABLE0",		# SH2_MD_DQRS_MMR_YAMOPW_ERR 

#	end of Shadow ECC mmrs

                        "SH2_XN_NI0_IILB_CMP_EXP_DATA0",	# CPU A PSP   (Core 0, Thread 0)
                        "SH2_XN_NI0_IILB_CMP_EXP_DATA1",	# CPU B PSP   (Core 0, Thread 1)
                        "SH2_XN_NI0_LLP_CMP_EXP_DATA0",		# CPU C PSP   (Core 1, Thread 0)
                        "SH2_XN_NI0_LLP_CMP_EXP_DATA1",		# CPU D PSP   (Core 1, Thread 1)

                        "SH2_XNPI_SIC_FLOW",
                        "SH2_XNPI_TO_NI0_PORT_FLOW",
                        "SH2_XNPI_TO_NI0_PORT_FLOW13",
                        "SH2_XNPI_TO_NI1_PORT_FLOW",
                        "SH2_XNPI_TO_NI1_PORT_FLOW13",
                        "SH2_XNPI_TO_IILB_PORT_FLOW",
                        "SH2_XNPI_STEERING_FIFO",
                        "SH2_XNPI_CREDIT_MGR_VC0",
                        "SH2_XNPI_CREDIT_MGR_VC2",

                        "SH2_XNMD_SIC_FLOW",
                        "SH2_XNMD_TO_NI0_PORT_FLOW",
                        "SH2_XNMD_TO_NI0_PORT_FLOW13",
                        "SH2_XNMD_TO_NI1_PORT_FLOW",
                        "SH2_XNMD_TO_NI1_PORT_FLOW13",
                        "SH2_XNMD_TO_IILB_PORT_FLOW",
                        "SH2_XNMD_STEERING_FIFO",
                        "SH2_XNMD_CREDIT_MGR_VC0",
                        "SH2_XNMD_CREDIT_MGR_VC2",

                        "SH2_XNII_INTRA_FLOW",
                        "SH2_XNLB_INTRA_FLOW",
                        "SH2_XNIILB_II_FIFO_CREDIT_CNT",
                        "SH2_XNIILB_LB_FIFO_CREDIT_CNT",
                        "SH2_XNIILB_FR_NI0_INTRA_FLOW_CREDIT",
                        "SH2_XNIILB_FR_NI1_INTRA_FLOW_CREDIT",
                        "SH2_XNIILB_FR_IILB_INTRA_FLOW_CREDIT",
                        "SH2_XNIILB_FR_MD_INTRA_FLOW_CREDIT",
                        "SH2_XNIILB_FR_PI_INTRA_FLOW_CREDIT",
                        "SH2_BT_XMIT_DYN_CREDIT_STAT",
                        "SH2_BT_XMIT_LAST_CREDIT_STAT",
                        "SH2_BT_NIQ_DYN_CREDIT_STAT",
                        "SH2_BT_NIQ_LAST_CREDIT_STAT",

                        "SH2_XNNI0_LLP_DEBIT_FLOW",
                        "SH2_XNNI0_LINK_0_FLOW",
                        "SH2_XNNI0_LINK_1_FLOW",
                        "SH2_XNNI0_LINK_2_FLOW",
                        "SH2_XNNI0_LINK_3_FLOW",
                        "SH2_XNNI0_0_INTRANI_FLOW",
                        "SH2_XNNI0_1_INTRANI_FLOW",
                        "SH2_XNNI0_2_INTRANI_FLOW",
                        "SH2_XNNI0_3_INTRANI_FLOW",
                        "SH2_XNNI0_VCSWITCH_FLOW",
                        "SH2_XNNI0_QUEUE_CNTR",
                        "SH2_XNNI0_TIMER_REG",
                        "SH2_XNNI0_IILB_TAILTO",
                        "SH2_XNNI0_NI_TAILTO",
                        "SH2_XNNI0_NI_TAILTO_SUPPORT",
                        "SH2_XNNI0_PI_TAILTO",
                        "SH2_XNNI0_MD_TAILTO",
                        "SH2_XNNI0_DEADLCK_FLOW",
                        "SH2_XNNI0_DEADLCK_FLOW_SUPPORT",

                        "SH2_XNNI1_LLP_DEBIT_FLOW",
                        "SH2_XNNI1_LINK_0_FLOW",
                        "SH2_XNNI1_LINK_1_FLOW",
                        "SH2_XNNI1_LINK_2_FLOW",
                        "SH2_XNNI1_LINK_3_FLOW",
                        "SH2_XNNI1_0_INTRANI_FLOW",
                        "SH2_XNNI1_1_INTRANI_FLOW",
                        "SH2_XNNI1_2_INTRANI_FLOW",
                        "SH2_XNNI1_3_INTRANI_FLOW",
                        "SH2_XNNI1_VCSWITCH_FLOW",
                        "SH2_XNNI1_QUEUE_CNTR",
                        "SH2_XNNI1_TIMER_REG",
                        "SH2_XNNI1_IILB_TAILTO",
                        "SH2_XNNI1_NI_TAILTO",
                        "SH2_XNNI1_NI_TAILTO_SUPPORT",
                        "SH2_XNNI1_PI_TAILTO",
                        "SH2_XNNI1_MD_TAILTO",
                        "SH2_XNNI1_DEADLCK_FLOW",
                        "SH2_XNNI1_DEADLCK_FLOW_SUPPORT",

                        "SH2_MCAST_INV_CONFIG",
                        "SH2_MCAST_MASK",
                        "SH2_MCAST_DEST_0A",
                        "SH2_MCAST_DEST_0B",
                        "SH2_MCAST_DEST_1A",
                        "SH2_MCAST_DEST_1B",
                        "SH2_MCAST_DEST_2A",
                        "SH2_MCAST_DEST_2B",
                        "SH2_MCAST_DEST_3A",
                        "SH2_MCAST_DEST_3B",
                        "SH2_MCAST_DEST_4A",
                        "SH2_MCAST_DEST_4B",
                        "SH2_MCAST_DEST_5A",
                        "SH2_MCAST_DEST_5B",
                        "SH2_MCAST_DEST_6A",
                        "SH2_MCAST_DEST_6B",
                        "SH2_MCAST_DEST_7A",
                        "SH2_MCAST_DEST_7B",
                        "SH2_MCAST_DEST_8A",
                        "SH2_MCAST_DEST_8B",
                        "SH2_MCAST_DEST_9A",
                        "SH2_MCAST_DEST_9B",
                        "SH2_MCAST_DEST_10A",
                        "SH2_MCAST_DEST_10B",
                        "SH2_MCAST_DEST_11A",
                        "SH2_MCAST_DEST_11B",
                        "SH2_MCAST_DEST_12A",
                        "SH2_MCAST_DEST_12B",
                        "SH2_MCAST_DEST_13A",
                        "SH2_MCAST_DEST_13B",
                        "SH2_MCAST_DEST_14A",
                        "SH2_MCAST_DEST_14B",
                        "SH2_MCAST_DEST_15A",
                        "SH2_MCAST_DEST_15B",
                        "SH2_MCAST_DEST_16A",
                        "SH2_MCAST_DEST_16B",
                        "SH2_MCAST_DEST_17A",
                        "SH2_MCAST_DEST_17B",
                        "SH2_MCAST_DEST_18A",
                        "SH2_MCAST_DEST_18B",
                        "SH2_MCAST_DEST_19A",
                        "SH2_MCAST_DEST_19B",
                        "SH2_MCAST_DEST_20A",
                        "SH2_MCAST_DEST_20B",
                        "SH2_MCAST_DEST_21A",
                        "SH2_MCAST_DEST_21B",
                        "SH2_MCAST_DEST_22A",
                        "SH2_MCAST_DEST_22B",
                        "SH2_MCAST_DEST_23A",
                        "SH2_MCAST_DEST_23B",
                        "SH2_MCAST_DEST_24A",
                        "SH2_MCAST_DEST_24B",
                        "SH2_MCAST_DEST_25A",
                        "SH2_MCAST_DEST_25B",
                        "SH2_MCAST_DEST_26A",
                        "SH2_MCAST_DEST_26B",
                        "SH2_MCAST_DEST_27A",
                        "SH2_MCAST_DEST_27B",
                        "SH2_MCAST_DEST_28A",
                        "SH2_MCAST_DEST_28B",
                        "SH2_MCAST_DEST_29A",
                        "SH2_MCAST_DEST_29B",
                        "SH2_MCAST_DEST_30A",
                        "SH2_MCAST_DEST_30B",
                        "SH2_MCAST_DEST_31A",
                        "SH2_MCAST_DEST_31B",
                        "SH2_MCAST_DEST_100A",
                        "SH2_MCAST_DEST_100B",
                        "SH2_MCAST_DEST_101A",
                        "SH2_MCAST_DEST_101B",
                        "SH2_MCAST_DEST_102A",
                        "SH2_MCAST_DEST_102B",
                        "SH2_MCAST_DEST_103A",
                        "SH2_MCAST_DEST_103B",
                        "SH2_MCAST_DEST_104A",
                        "SH2_MCAST_DEST_104B",
                        "SH2_MCAST_DEST_105A",
                        "SH2_MCAST_DEST_105B",
                        "SH2_MCAST_DEST_106A",
                        "SH2_MCAST_DEST_106B",
                        "SH2_MCAST_DEST_107A",
                        "SH2_MCAST_DEST_107B",
                        "SH2_MCAST_DEST_108A",
                        "SH2_MCAST_DEST_108B",
                        "SH2_MCAST_DEST_109A",
                        "SH2_MCAST_DEST_109B",
                        "SH2_MCAST_DEST_110A",
                        "SH2_MCAST_DEST_110B",
                        "SH2_MCAST_DEST_111A",
                        "SH2_MCAST_DEST_111B",
                        "SH2_MCAST_DEST_112A",
                        "SH2_MCAST_DEST_112B",
                        "SH2_MCAST_DEST_113A",
                        "SH2_MCAST_DEST_113B",
                        "SH2_MCAST_DEST_114A",
                        "SH2_MCAST_DEST_114B",
                        "SH2_MCAST_DEST_115A",
                        "SH2_MCAST_DEST_115B",
                        "SH2_MCAST_DEST_116A",
                        "SH2_MCAST_DEST_116B",
                        "SH2_MCAST_DEST_117A",
                        "SH2_MCAST_DEST_117B",
                        "SH2_MCAST_DEST_118A",
                        "SH2_MCAST_DEST_118B",
                        "SH2_MCAST_DEST_119A",
                        "SH2_MCAST_DEST_119B",
                        "SH2_MCAST_DEST_120A",
                        "SH2_MCAST_DEST_120B",
                        "SH2_MCAST_DEST_121A",
                        "SH2_MCAST_DEST_121B",
                        "SH2_MCAST_DEST_122A",
                        "SH2_MCAST_DEST_122B",
                        "SH2_MCAST_DEST_123A",
                        "SH2_MCAST_DEST_123B",
                        "SH2_MCAST_DEST_124A",
                        "SH2_MCAST_DEST_124B",
                        "SH2_MCAST_DEST_125A",
                        "SH2_MCAST_DEST_125B",
                        "SH2_MCAST_DEST_126A",
                        "SH2_MCAST_DEST_126B",
                        "SH2_MCAST_DEST_127A",
                        "SH2_MCAST_DEST_127B",
                        "SH2_MCAST_DEST_128A",
                        "SH2_MCAST_DEST_128B",
                        "SH2_MCAST_DEST_129A",
                        "SH2_MCAST_DEST_129B",
                        "SH2_MCAST_DEST_130A",
                        "SH2_MCAST_DEST_130B",
                        "SH2_MCAST_DEST_131A",
                        "SH2_MCAST_DEST_131B"

);

@SHUB2_SHORT_REG_LIST = (
                        "SH2_NODE_ID",                  # MUST be FIRST in the list!
			"SH2_PROCESSOR_TYPE",
                        "SH2_RACK_SLOT_DATA",
			"SH2_ADDRESS_MODE",
                        "SH2_EVENT_OCCURRED",
                        "SH2_EVENT_OVERFLOW",
                        "SH2_FIRST_ERROR",

                        "SH2_RTC",
                        "SH2_NODE_PRESENT_TABLE",
                        "SH2_MD_DQLP_MMR_DIR_CONFIG",
                        "SH2_MD_DQLP_MMR_DIR_PRESVEC0",
                        "SH2_MD_DQLP_MMR_DIR_PRESVEC1",
                        "SH2_MD_DQLP_MMR_DIR_PRESVEC2",
                        "SH2_MD_DQLP_MMR_DIR_PRESVEC3",
                        "SH2_MD_DQRP_MMR_DIR_CONFIG",
                        "SH2_MD_DQRP_MMR_DIR_PRESVEC0",
                        "SH2_MD_DQRP_MMR_DIR_PRESVEC1",
                        "SH2_MD_DQRP_MMR_DIR_PRESVEC2",
                        "SH2_MD_DQRP_MMR_DIR_PRESVEC3",

			"SH2_FSB_SYMMETRIC_AGENT_STATUS",
			"SH2_FSB_SYSTEM_AGENT_CONFIG",
                        "SH2_SCRATCH1",
		  	"SH2_CRB_MESSAGE_CONTROL",
                        "SH2_XN_MD_SIC_CMP_EXP_HDR0",           # CPU A Target Identifier
                        "SH2_XN_MD_SIC_CMP_EXP_HDR1",           # CPU A Target Id Timestamp
                        "SH2_XN_NI0_MD_CMP_EXP_HDR1",           # CPU B Target Identifier
                        "SH2_XN_NI0_MD_CMP_EXP_HDR0",           # CPU B Target Id Timestamp
                        "SH2_XN_IILB_NI1_CMP_EXP_DATA0",        # CPU C Target Identifier
                        "SH2_XN_IILB_NI1_CMP_EXP_DATA1",        # CPU C Target Id Timestamp
                        "SH2_XN_IILB_NI0_CMP_EXP_DATA0",        # CPU D Target Identifier
                        "SH2_XN_IILB_NI0_CMP_EXP_DATA1",        # CPU D Target Id Timestamp

                        "SH2_MEM_CREDITS",	
                        "SH2_X_DIMM_CFG",
                        "SH2_Y_DIMM_CFG",
                        "SH2_JNR_DIMM_CFG",

                        "SH2_PI_FIRST_ERROR",
                        "SH2_PI_ERROR_SUMMARY",
                        "SH2_PI_ERROR_OVERFLOW",
                        "SH2_PI_ERROR_MASK",
                        "SH2_PI_AFI_ERROR_MASK",
                        "SH2_PI_CRBP_ERROR_MASK",
                        "SH2_PI_RDAT_ERROR_MASK",
                        "SH2_PI_WDAT_ERROR_MASK",
                        "SH2_PI_ERROR_DETAIL_1",
                        "SH2_PI_ERROR_DETAIL_2",
                        "SH2_PI_HW_TIME_STAMP",
                        "SH2_PI_CORRECTED_DETAIL_1",
                        "SH2_PI_CORRECTED_DETAIL_2",
                        "SH2_PI_CORRECTED_DETAIL_3",
                        "SH2_PI_CORRECTED_DETAIL_4",
                        "SH2_PI_CORRECTED_DETAIL_5",
                        "SH2_PI_CORRECTED_DETAIL_6",
                        "SH2_PI_COR_TIME_STAMP",
                        "SH2_PI_UNCORRECTED_DETAIL_1",
                        "SH2_PI_UNCORRECTED_DETAIL_2",
                        "SH2_PI_UNCORRECTED_DETAIL_3",
                        "SH2_PI_UNCORRECTED_DETAIL_4",
                        "SH2_PI_UNCORRECTED_DETAIL_5",
                        "SH2_PI_UNCORRECTED_DETAIL_6",
                        "SH2_PI_UNCOR_TIME_STAMP",

                        "SH2_PIO_WRITE_STATUS_00",
                        "SH2_PIO_WRITE_STATUS_01",
                        "SH2_PIO_WRITE_STATUS_10",
                        "SH2_PIO_WRITE_STATUS_11",
                        "SH2_PIO_WC_WRITE_STATUS_00",
                        "SH2_PIO_WC_WRITE_STATUS_01",
                        "SH2_PIO_WC_WRITE_STATUS_10",
                        "SH2_PIO_WC_WRITE_STATUS_11",
                        "SH2_MEMORY_WRITE_STATUS_00",
                        "SH2_MEMORY_WRITE_STATUS_01",
                        "SH2_MEMORY_WRITE_STATUS_10",
                        "SH2_MEMORY_WRITE_STATUS_11",
                        "SH2_MEMORY_WRITE_STATUS_NON_USER_00",
                        "SH2_MEMORY_WRITE_STATUS_NON_USER_01",
                        "SH2_MEMORY_WRITE_STATUS_NON_USER_10",
                        "SH2_MEMORY_WRITE_STATUS_NON_USER_11",
                        "SH2_PI_MD2PI_REPLY_VC_CONFIG",
                        "SH2_PI_MD2PI_REQUEST_VC_CONFIG",
                        "SH2_PI_MD2PI_REPLY_VC_STATUS",
                        "SH2_PI_MD2PI_REQUEST_VC_STATUS",
                        "SH2_PI_XN2PI_REPLY_VC_CONFIG",
                        "SH2_PI_XN2PI_REQUEST_VC_CONFIG",
                        "SH2_PI_XN2PI_REPLY_VC_STATUS",
                        "SH2_PI_XN2PI_REQUEST_VC_STATUS",	# This reg also has ICC status
                        "SH2_PI_PI2MD_REPLY_VC_STATUS",
                        "SH2_PI_PI2MD_REQUEST_VC_STATUS",
                        "SH2_PI_PI2XN_REPLY_VC_STATUS",
                        "SH2_PI_PI2XN_REQUEST_VC_STATUS",
                        "SH2_PI_CRB_STALL_SKID",

                        "SH2_LB_FIRST_ERROR",
                        "SH2_LB_ERROR_SUMMARY",
                        "SH2_LB_ERROR_OVERFLOW",
                        "SH2_LB_ERROR_MASK",
                        "SH2_LB_ERROR_DETAIL_1",
                        "SH2_LB_ERROR_DETAIL_2",
                        "SH2_LB_ERROR_DETAIL_3",
                        "SH2_LB_ERROR_DETAIL_4",
                        "SH2_LB_ERROR_DETAIL_5",
                        "SH2_LB_ERROR_DETAIL_6",
                        "SH2_LB_HW_TIME_STAMP",

                        "SH2_XN_FIRST_ERROR",
                        "SH2_XN_ERROR_SUMMARY",
                        "SH2_XN_ERROR_OVERFLOW",
                        "SH2_XN_ERROR_MASK",
                        "SH2_XN_HW_TIME_STAMP",
                        "SH2_XN_CORRECTED_DETAIL_1",
                        "SH2_XN_CORRECTED_DETAIL_2",
                        "SH2_XN_CORRECTED_DETAIL_3",
                        "SH2_XN_CORRECTED_DETAIL_4",
                        "SH2_XN_COR_TIME_STAMP",
                        "SH2_XN_UNCORRECTED_DETAIL_1",
                        "SH2_XN_UNCORRECTED_DETAIL_2",
                        "SH2_XN_UNCORRECTED_DETAIL_3",
                        "SH2_XN_UNCORRECTED_DETAIL_4",
                        "SH2_XN_UNCOR_TIME_STAMP",

                        "SH2_NI0_FIRST_ERROR_1",
                        "SH2_NI0_FIRST_ERROR_2",
                        "SH2_NI0_ERROR_SUMMARY_1",
                        "SH2_NI0_ERROR_SUMMARY_2",
                        "SH2_NI0_ERROR_OVERFLOW_1",
                        "SH2_NI0_ERROR_OVERFLOW_2",
                        "SH2_NI0_ERROR_MASK_1",
                        "SH2_NI0_ERROR_MASK_2",
                        "SH2_NI0_ERROR_DETAIL_1",
                        "SH2_NI0_ERROR_DETAIL_2",
                        "SH2_NI0_ERROR_DETAIL_3",

                        "SH2_NI1_FIRST_ERROR_1",
                        "SH2_NI1_FIRST_ERROR_2",
                        "SH2_NI1_ERROR_SUMMARY_1",
                        "SH2_NI1_ERROR_SUMMARY_2",
                        "SH2_NI1_ERROR_OVERFLOW_1",
                        "SH2_NI1_ERROR_OVERFLOW_2",
                        "SH2_NI1_ERROR_MASK_1",
                        "SH2_NI1_ERROR_MASK_2",
                        "SH2_NI1_ERROR_DETAIL_1",
                        "SH2_NI1_ERROR_DETAIL_2",
                        "SH2_NI1_ERROR_DETAIL_3",

                        "SH2_XNIILB_FIRST_ERROR",
                        "SH2_XNIILB_FIRST_ERROR_1",
                        "SH2_XNIILB_ERROR_SUMMARY",
                        "SH2_XNIILB_ERROR_SUMMARY_1",
                        "SH2_XNIILB_ERROR_OVERFLOW",
                        "SH2_XNIILB_ERROR_OVERFLOW_1",
                        "SH2_XNIILB_ERROR_MASK",
                        "SH2_XNIILB_ERROR_MASK_1",
                        "SH2_XNIILB_ERROR_DETAIL_1",
                        "SH2_XNIILB_ERROR_DETAIL_2",
                        "SH2_XNIILB_ERROR_DETAIL_3",

                        "SH2_XNPI_FIRST_ERROR",
                        "SH2_XNPI_ERROR_SUMMARY",
                        "SH2_XNPI_ERROR_OVERFLOW",
                        "SH2_XNPI_ERROR_MASK",
                        "SH2_XNPI_ERROR_DETAIL_1",

                        "SH2_XNMD_FIRST_ERROR",
                        "SH2_XNMD_FIRST_ERROR_1",
                        "SH2_XNMD_ERROR_SUMMARY",
                        "SH2_XNMD_ERROR_SUMMARY_1",
                        "SH2_XNMD_ERROR_OVERFLOW",
                        "SH2_XNMD_ERROR_OVERFLOW_1",
                        "SH2_XNMD_ERROR_MASK",
                        "SH2_XNMD_ERROR_MASK_1",
                        "SH2_XNMD_ERROR_DETAIL_1",

                        "SH2_MEM_FIRST_ERROR",
                        "SH2_MEM_ERROR_SUMMARY",
                        "SH2_MEM_ERROR_OVERFLOW",
                        "SH2_MEM_ERROR_MASK",
                        "SH2_MISC_ERR_HDR_LOWER",
                        "SH2_MISC_ERR_HDR_UPPER",
                        "SH2_MISC_ADRS_ERR_HDR_LOWER_A",
                        "SH2_MD_HW_TIME_STAMP",
                        "SH2_DIR_COR_ERR_HDR_LOWER",
                        "SH2_DIR_COR_ERR_HDR_UPPER",
                        "SH2_DIR_COR_ADRS_ERR_HDR_LOWER_A",
                        "SH2_MD_COR_TIME_STAMP",
                        "SH2_DIR_UC_ERR_HDR_LOWER",
                        "SH2_DIR_UC_ERR_HDR_UPPER",
                        "SH2_DIR_UC_ADRS_ERR_HDR_LOWER_A",
                        "SH2_MD_UNCOR_TIME_STAMP",
                        "SH2_MMRBIST_ERR",
                        "SH2_MD_DQLP_MMR_XCERR1",
                        "SH2_MD_DQLP_MMR_XCERR2",
                        "SH2_MD_DQLP_MMR_XUERR1",
                        "SH2_MD_DQLP_MMR_XUERR2",
                        "SH2_MD_DQLP_MMR_XPERR0",
                        "SH2_MD_DQLP_MMR_XPERR1",
                        "SH2_MD_DQLP_MMR_YCERR1",
                        "SH2_MD_DQLP_MMR_YCERR2",
                        "SH2_MD_DQLP_MMR_YUERR1",
                        "SH2_MD_DQLP_MMR_YUERR2",
                        "SH2_MD_DQLP_MMR_YPERR0",
                        "SH2_MD_DQLP_MMR_YPERR1",
                        "SH2_MD_DQRP_MMR_XCERR1",
                        "SH2_MD_DQRP_MMR_XCERR2",
                        "SH2_MD_DQRP_MMR_XUERR1",
                        "SH2_MD_DQRP_MMR_XUERR2",
                        "SH2_MD_DQRP_MMR_XPERR0",
                        "SH2_MD_DQRP_MMR_XPERR1",
                        "SH2_MD_DQRP_MMR_YCERR1",
                        "SH2_MD_DQRP_MMR_YCERR2",
                        "SH2_MD_DQRP_MMR_YUERR1",
                        "SH2_MD_DQRP_MMR_YUERR2",
                        "SH2_MD_DQRP_MMR_YPERR0",
                        "SH2_MD_DQRP_MMR_YPERR1",
                        "SH2_MD_DQLS_MMR_XAMOPW_ERR",
                        "SH2_MD_DQRS_MMR_YAMOPW_ERR",

                        "SH2_BT_FIRST_ERROR",
                        "SH2_BT_ERROR_SUMMARY",
                        "SH2_BT_ERROR_OVERFLOW",
                        "SH2_BT_ERROR_MASK",
                        "SH2_BT_HW_TIME_STAMP",
                        "SH2_BT_XN_ERROR_DETAIL_1",
                        "SH2_BT_XN_ERROR_DETAIL_2",
                        "SH2_BT_ENG_ERROR_STAT",
			"SH2_BT_ENG_CSR_0",
			"SH2_BT_ENG_CSR_1",
			"SH2_BT_ENG_CSR_2",
			"SH2_BT_ENG_CSR_3",
                        "SH2_BT_IQ_STAT",
                        "SH2_BT_CRB_STAT_0",
                        "SH2_BT_CRB_SRC_ADDR_0",
                        "SH2_BT_CRB_DEST_ADDR_0",
                        "SH2_BT_CRB_STAT_1",
                        "SH2_BT_CRB_SRC_ADDR_1",
                        "SH2_BT_CRB_DEST_ADDR_1",
                        "SH2_BT_CRB_STAT_2",
                        "SH2_BT_CRB_SRC_ADDR_2",
                        "SH2_BT_CRB_DEST_ADDR_2",
                        "SH2_BT_CRB_STAT_3",
                        "SH2_BT_CRB_SRC_ADDR_3",
                        "SH2_BT_CRB_DEST_ADDR_3",
                        "SH2_BT_CRB_STAT_4",
                        "SH2_BT_CRB_SRC_ADDR_4",
                        "SH2_BT_CRB_DEST_ADDR_4",
                        "SH2_BT_CRB_STAT_5",
                        "SH2_BT_CRB_SRC_ADDR_5",
                        "SH2_BT_CRB_DEST_ADDR_5",
                        "SH2_BT_CRB_STAT_6",
                        "SH2_BT_CRB_SRC_ADDR_6",
                        "SH2_BT_CRB_DEST_ADDR_6",
                        "SH2_BT_CRB_STAT_7",
                        "SH2_BT_CRB_SRC_ADDR_7",
                        "SH2_BT_CRB_DEST_ADDR_7",
                        "SH2_BT_CRB_STAT_8",
                        "SH2_BT_CRB_SRC_ADDR_8",
                        "SH2_BT_CRB_DEST_ADDR_8",
                        "SH2_BT_CRB_STAT_9",
                        "SH2_BT_CRB_SRC_ADDR_9",
                        "SH2_BT_CRB_DEST_ADDR_9",
                        "SH2_BT_CRB_STAT_10",
                        "SH2_BT_CRB_SRC_ADDR_10",
                        "SH2_BT_CRB_DEST_ADDR_10",
                        "SH2_BT_CRB_STAT_11",
                        "SH2_BT_CRB_SRC_ADDR_11",
                        "SH2_BT_CRB_DEST_ADDR_11",
                        "SH2_BT_CRB_STAT_12",
                        "SH2_BT_CRB_SRC_ADDR_12",
                        "SH2_BT_CRB_DEST_ADDR_12",
                        "SH2_BT_CRB_STAT_13",
                        "SH2_BT_CRB_SRC_ADDR_13",
                        "SH2_BT_CRB_DEST_ADDR_13",
                        "SH2_BT_CRB_STAT_14",
                        "SH2_BT_CRB_SRC_ADDR_14",
                        "SH2_BT_CRB_DEST_ADDR_14",
                        "SH2_BT_CRB_STAT_15",
                        "SH2_BT_CRB_SRC_ADDR_15",
                        "SH2_BT_CRB_DEST_ADDR_15",

                        "SH2_NI0_LLP_STAT",
                        "SH2_NI0_LLP_RESET",
                        "SH2_NI0_LLP_RESET_EN",
                        "SH2_NI0_LLP_CHAN_MODE",
                        "SH2_NI0_LLP_CONFIG",
                        "SH2_NI0_LLP_TEST_CTL",
                        "SH2_NI0_LLP_CAPT_WD1",
                        "SH2_NI0_LLP_CAPT_WD2",
                        "SH2_NI0_LLP_CAPT_SBCB",
                        "SH2_NI0_LLP_ERR",
                        "SH2_NI0_LLP_ERR1",
                        "SH2_NI0_LLP_ERR2",
                        "SH2_NI0_LLP_PKT",
                        "SH2_NI1_LLP_STAT",
                        "SH2_NI1_LLP_RESET",
                        "SH2_NI1_LLP_RESET_EN",
                        "SH2_NI1_LLP_CHAN_MODE",
                        "SH2_NI1_LLP_CONFIG",
                        "SH2_NI1_LLP_TEST_CTL",
                        "SH2_NI1_LLP_CAPT_WD1",
                        "SH2_NI1_LLP_CAPT_WD2",
                        "SH2_NI1_LLP_CAPT_SBCB",
                        "SH2_NI1_LLP_ERR",
                        "SH2_NI1_LLP_ERR1",
                        "SH2_NI1_LLP_ERR2",
                        "SH2_NI1_LLP_PKT",

#       Shadow error registers for ECC errors from BW# 945281

			"SH2_XN_IILB_LB_CMP_EXP_DATA1", 	# SH2_EVENT_OCCURRED              
			"SH2_XN_PI_NI0_CMP_ENABLE0", 		# SH2_PI_FIRST_ERROR              
			"SH2_XN_PI_NI0_CMP_ENABLE1", 		# SH2_PI_ERROR_SUMMARY              

			"SH2_XN_IILB_LB_CMP_ENABLE0",		# SH2_PI_CORRECTED_DETAIL_1       
			"SH2_XN_IILB_LB_CMP_ENABLE1", 		# SH2_PI_CORRECTED_DETAIL_2       
			"SH2_XN_IILB_II_CMP_EXP_DATA0",		# SH2_PI_CORRECTED_DETAIL_3       
			"SH2_XN_IILB_II_CMP_EXP_DATA1",		# SH2_PI_CORRECTED_DETAIL_4       
			"SH2_XN_IILB_II_CMP_ENABLE0", 		# SH2_PI_CORRECTED_DETAIL_5       
			"SH2_XN_IILB_II_CMP_ENABLE1", 		# SH2_PI_CORRECTED_DETAIL_6       
			"SH2_XN_NI0_LLP_CMP_ENABLE0", 		# SH2_PI_COR_TIME_STAMP           

			"SH2_XN_NI0_LLP_CMP_ENABLE1", 		# SH2_PI_UNCORRECTED_DETAIL_1     
			"SH2_XN_NI0_MD_CMP_HDR_ENABLE0",	# SH2_PI_UNCORRECTED_DETAIL_2     
			"SH2_XN_NI0_MD_CMP_HDR_ENABLE1",	# SH2_PI_UNCORRECTED_DETAIL_3     
			"SH2_XN_NI1_LLP_CMP_EXP_DATA0",		# SH2_PI_UNCORRECTED_DETAIL_4     
			"SH2_XN_NI1_LLP_CMP_EXP_DATA1",		# SH2_PI_UNCORRECTED_DETAIL_5     
			"SH2_XN_NI1_LLP_CMP_ENABLE0", 		# SH2_PI_UNCORRECTED_DETAIL_6     
			"SH2_XN_NI1_LLP_CMP_ENABLE1",		# SH2_PI_UNCOR_TIME_STAMP         

			"SH2_XN_NI1_MD_CMP_EXP_HDR0", 		# SH2_XN_CORRECTED_DETAIL_1       
			"SH2_XN_NI1_MD_CMP_EXP_HDR1", 		# SH2_XN_CORRECTED_DETAIL_2       
			"SH2_XN_NI1_MD_CMP_HDR_ENABLE0",	# SH2_XN_CORRECTED_DETAIL_3       
			"SH2_XN_NI1_MD_CMP_HDR_ENABLE1",	# SH2_XN_CORRECTED_DETAIL_4       
			"SH2_XN_PI_IILB_CMP_EXP_DATA0",		# SH2_XN_COR_TIME_STAMP           

			"SH2_XN_PI_IILB_CMP_EXP_DATA1",		# SH2_XN_UNCORRECTED_DETAIL_1     
			"SH2_XN_PI_IILB_CMP_ENABLE0", 		# SH2_XN_UNCORRECTED_DETAIL_2     
			"SH2_XN_PI_IILB_CMP_ENABLE1",		# SH2_XN_UNCORRECTED_DETAIL_3     
			"SH2_XN_PI_NI0_CMP_EXP_DATA0", 		# SH2_XN_UNCORRECTED_DETAIL_4     
			"SH2_XN_PI_NI0_CMP_EXP_DATA1", 		# SH2_XN_UNCOR_TIME_STAMP         

                        "SH2_XN_NI1_PI_CMP_ENABLE1",            # SH2_MISC_ERR_HDR_LOWER
                        "SH2_XN_NI1_PI_CMP_ENABLE0",            # SH2_MISC_ERR_HDR_UPPER
                        "SH2_XN_NI1_NI_CMP_EXP_DATA0",          # SH2_MISC_ADRS_ERR_HDR_LOWER_A
                        "SH2_XN_NI1_IILB_CMP_ENABLE1",          # SH2_DIR_COR_ERR_HDR_LOWER
                        "SH2_XN_NI1_PI_CMP_EXP_DATA0",          # SH2_DIR_COR_ERR_HDR_UPPER
                        "SH2_XN_NI1_PI_CMP_EXP_DATA1",          # SH2_DIR_COR_ADRS_ERR_HDR_LOWER_A
                        "SH2_XN_NI1_NI_CMP_EXP_DATA1",          # SH2_MD_COR_TIME_STAMP

                        "SH2_XN_NI0_NI_CMP_EXP_DATA1",          # SH2_MD_DQLP_MMR_XCERR1
                        "SH2_XN_NI0_NI_CMP_ENABLE0",            # SH2_MD_DQLP_MMR_XCERR2
                        "SH2_XN_NI0_NI_CMP_ENABLE1",            # SH2_MD_DQLP_MMR_YCERR1
			"SH2_XN_NI1_NI_CMP_ENABLE0",		# SH2_MD_DQLP_MMR_YCERR2
			"SH2_XN_NI1_NI_CMP_ENABLE1",		# SH2_MD_DQLS_MMR_XAMOPW_ERR 
			"SH2_XN_PI_NI1_CMP_EXP_DATA0",		# SH2_MD_DQRP_MMR_XCERR1 
			"SH2_XN_PI_NI1_CMP_EXP_DATA1",		# SH2_MD_DQRP_MMR_XCERR2 
                        "SH2_XN_NI1_IILB_CMP_EXP_DATA0",        # SH2_MD_DQRP_MMR_YCERR1
                        "SH2_XN_NI1_IILB_CMP_EXP_DATA1",        # SH2_MD_DQRP_MMR_YCERR2
                        "SH2_XN_NI1_IILB_CMP_ENABLE0",          # SH2_MD_DQRS_MMR_YAMOPW_ERR

#	end of Shadow ECC mmrs

                        "SH2_XN_NI0_IILB_CMP_EXP_DATA0",	# CPU A PSP   (Core 0, Thread 0)
                        "SH2_XN_NI0_IILB_CMP_EXP_DATA1",	# CPU B PSP   (Core 0, Thread 1)
                        "SH2_XN_NI0_LLP_CMP_EXP_DATA0",		# CPU C PSP   (Core 1, Thread 0)
                        "SH2_XN_NI0_LLP_CMP_EXP_DATA1",		# CPU D PSP   (Core 1, Thread 1)

                        "SH2_XNPI_SIC_FLOW",
                        "SH2_XNPI_TO_NI0_PORT_FLOW",
                        "SH2_XNPI_TO_NI0_PORT_FLOW13",
                        "SH2_XNPI_TO_NI1_PORT_FLOW",
                        "SH2_XNPI_TO_NI1_PORT_FLOW13",
                        "SH2_XNPI_TO_IILB_PORT_FLOW",
                        "SH2_XNPI_STEERING_FIFO",
                        "SH2_XNPI_CREDIT_MGR_VC0",
                        "SH2_XNPI_CREDIT_MGR_VC2",

                        "SH2_XNMD_SIC_FLOW",
                        "SH2_XNMD_TO_NI0_PORT_FLOW",
                        "SH2_XNMD_TO_NI0_PORT_FLOW13",
                        "SH2_XNMD_TO_NI1_PORT_FLOW",
                        "SH2_XNMD_TO_NI1_PORT_FLOW13",
                        "SH2_XNMD_TO_IILB_PORT_FLOW",
                        "SH2_XNMD_STEERING_FIFO",
                        "SH2_XNMD_CREDIT_MGR_VC0",
                        "SH2_XNMD_CREDIT_MGR_VC2",

                        "SH2_XNII_INTRA_FLOW",
                        "SH2_XNLB_INTRA_FLOW",
                        "SH2_XNIILB_II_FIFO_CREDIT_CNT",
                        "SH2_XNIILB_LB_FIFO_CREDIT_CNT",
                        "SH2_XNIILB_FR_NI0_INTRA_FLOW_CREDIT",
                        "SH2_XNIILB_FR_NI1_INTRA_FLOW_CREDIT",
                        "SH2_XNIILB_FR_IILB_INTRA_FLOW_CREDIT",
                        "SH2_XNIILB_FR_MD_INTRA_FLOW_CREDIT",
                        "SH2_XNIILB_FR_PI_INTRA_FLOW_CREDIT",
                        "SH2_BT_XMIT_DYN_CREDIT_STAT",
                        "SH2_BT_XMIT_LAST_CREDIT_STAT",
                        "SH2_BT_NIQ_DYN_CREDIT_STAT",
                        "SH2_BT_NIQ_LAST_CREDIT_STAT",

                        "SH2_XNNI0_LLP_DEBIT_FLOW",
                        "SH2_XNNI0_LINK_0_FLOW",
                        "SH2_XNNI0_LINK_1_FLOW",
                        "SH2_XNNI0_LINK_2_FLOW",
                        "SH2_XNNI0_LINK_3_FLOW",
                        "SH2_XNNI0_0_INTRANI_FLOW",
                        "SH2_XNNI0_1_INTRANI_FLOW",
                        "SH2_XNNI0_2_INTRANI_FLOW",
                        "SH2_XNNI0_3_INTRANI_FLOW",
                        "SH2_XNNI0_VCSWITCH_FLOW",
                        "SH2_XNNI0_QUEUE_CNTR",
                        "SH2_XNNI0_TIMER_REG",
                        "SH2_XNNI0_IILB_TAILTO",
                        "SH2_XNNI0_NI_TAILTO",
                        "SH2_XNNI0_NI_TAILTO_SUPPORT",
                        "SH2_XNNI0_PI_TAILTO",
                        "SH2_XNNI0_MD_TAILTO",
                        "SH2_XNNI0_DEADLCK_FLOW",
                        "SH2_XNNI0_DEADLCK_FLOW_SUPPORT",

                        "SH2_XNNI1_LLP_DEBIT_FLOW",
                        "SH2_XNNI1_LINK_0_FLOW",
                        "SH2_XNNI1_LINK_1_FLOW",
                        "SH2_XNNI1_LINK_2_FLOW",
                        "SH2_XNNI1_LINK_3_FLOW",
                        "SH2_XNNI1_0_INTRANI_FLOW",
                        "SH2_XNNI1_1_INTRANI_FLOW",
                        "SH2_XNNI1_2_INTRANI_FLOW",
                        "SH2_XNNI1_3_INTRANI_FLOW",
                        "SH2_XNNI1_VCSWITCH_FLOW",
                        "SH2_XNNI1_QUEUE_CNTR",
                        "SH2_XNNI1_TIMER_REG",
                        "SH2_XNNI1_IILB_TAILTO",
                        "SH2_XNNI1_NI_TAILTO",
                        "SH2_XNNI1_NI_TAILTO_SUPPORT",
                        "SH2_XNNI1_PI_TAILTO",
                        "SH2_XNNI1_MD_TAILTO",
                        "SH2_XNNI1_DEADLCK_FLOW",
                        "SH2_XNNI1_DEADLCK_FLOW_SUPPORT"

);

@SHUB2_ABBREV_LIST = (
                        "SH2_NODE_ID"
);

#-------------------------------------------------------------------------------
#
#	Function to read & parse shub2 mmrs.
#
#-------------------------------------------------------------------------------

sub read_and_parse_shub2x {

    my @sh2_mmr_names = ($abbrev_mode == 1) ? @SHUB2_ABBREV_LIST : 
			($fast_mode == 1) ? @SHUB2_SHORT_REG_LIST : @SHUB2_REG_LIST;
    my @sh2_reg_list = ();

    my $i = 0;
    #--------------------------------------------
    #  Create the register structure for each MMR
    #--------------------------------------------
    for ($i = 0; $i <= $#sh2_mmr_names; $i++) {
        $name    = $sh2_mmr_names[$i];
        $reg_ptr = reg_create_mmr_inst("shub2_01", $name);
        if ( !reg_is_valid($reg_ptr) )
        {  $emess = "Unable to create register instance: $name.\n";
          dsp_error_and_exit($emess);
        }
        $sh2_reg_list[$i] = $reg_ptr;
    }

    $sh2_cnt = 0;

	#  continue in this while loop to read all Shub2 chips

    autoflush STDOUT 1; 		#  GEM  - debug

    while (defined $sh2_target_list[$sh2_cnt]) {

      $SSCOPE_ERR = 0;	
      mmr_target($SELECT, $sh2_target_list[$sh2_cnt]->{MODULE},
                      $sh2_target_list[$sh2_cnt]->{NODE},
                      $sh2_target_list[$sh2_cnt]->{CHIP});

      if ($SSCOPE_ERR == 1)  {
        printf "\n mmr_target() failed in read_and_parse_shub2x().  Aborting mmr read for %s of %s\n\n",
	  $sh2_target_list[$sh2_cnt]->{NODE}, $sh2_target_list[$sh2_cnt]->{MODULE};
        $SSCOPE_ERR = 0;
        $sh2_cnt++;
        next;
      }

      
      if ($show_clock == 1)  {
        chomp ($done_time = `date +'%T'`);
        printf "\n mmr_target\(\) completed at %s on %s of %s\n", $sh2_target_list[$sh2_cnt]->{NODE}, $sh2_target_list[$sh2_cnt]->{MODULE};
      }

      arm_pi_credits("shub2_01");

      if ($show_clock == 1)  {
        chomp ($done_time = `date +'%T'`);
        print " arm_pi_credits\(\) completed at $done_time\n";
      }

      set_pio_timeout(0xffff);	 		#  WAR for BW# 953156

      $xfr_complete = 0;
      $xfr_f_cnt = 0;
      while (($xfr_complete == 0) && ($xfr_f_cnt < 3) && ($sh2_target_list[$sh2_cnt]->{XFR_ERR_CNT} < 6))  {
        $SSCOPE_ERR = 0;
        reg_list_mmr_read(\@sh2_reg_list);		# actual mmr read occurs here
        if ($show_clock == 1)  {
          chomp ($done_time = `date +'%T'`);
          print " reg_list_mmr_read\(\) completed at $done_time\n";
	}

        if ($SSCOPE_ERR == 1)  {
          $sh2_target_list[$sh2_cnt]->{XFR_ERR_CNT}++;
          $xfr_f_cnt++;
          $SSCOPE_ERR = 0;
        }
        else {
          $xfr_complete = 1;
        }
      }

      if ($show_clock == 1)  {
        chomp ($done_time = `date +'%T'`);
        print " Begin parse_sh2_mmr_list\(\) at $done_time\n";
      }

      if ($xfr_complete == 1) {
        parse_sh2_mmr_list(\@sh2_reg_list);
      }
      else {		#  DEAD node
        $nasid_value = 0xDEAD;
	$sh2_target_list[$sh2_cnt]->{NASID} = $nasid_value;
        $sh2_node_hdr = sprintf "\n\nReading Shub2 Registers on %s of %s   NASID: 0x%04X   Rev: %d   (Shub2.%d)",
              $sh2_target_list[$sh2_cnt]->{NODE}, $sh2_target_list[$sh2_cnt]->{MODULE}, $nasid_value, 0, 0;
	my $scandmp_cmd = sprintf "scandump --l2 %s -chip %s:%s:%s", 
		$l2_host, $sh2_target_list[$sh2_cnt]->{MODULE}, $sh2_target_list[$sh2_cnt]->{NODE}, $sh2_target_list[$sh2_cnt]->{CHIP};
        my $err_msg = sprintf "\n   ERROR!  All MMR accesses attempted via jtag on this Blade have failed.\n ";
	$err_msg .= sprintf "        Suspect Shub has become wedged, and reply packets have backed up into the Shub LB.  \(BW# 957418\)\n";
        $err_msg .= sprintf "        Scandump this node for further fault isolation.   CMD:   %s\n\n", $scandmp_cmd;
	$sh2_node_hdr .= "$err_msg";
        printf "%s", $sh2_node_hdr;
        $sh2_target_list[$sh2_cnt]->{ERR_SUMM} = sprintf "\n\nError Registers found on Shub2 on %s of %s    NASID: 0x%04X   Rev: %d   (Shub2.%d)",
              $sh2_target_list[$sh2_cnt]->{NODE}, $sh2_target_list[$sh2_cnt]->{MODULE}, $nasid_value, 0, 0;
        $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "$err_msg";
        $any_sh2_error = 1;
        $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
	$global_err_val |= $SHUB_HW_ERR;
        $current_location = sprintf("%s, %s", $sh2_target_list[$sh2_cnt]->{MODULE}, $sh2_target_list[$sh2_cnt]->{NODE});

        my $fru_rsn = sprintf "All MMR reads failed due to suspected wedged Shub. \(BW# 957418\)  ";
	my $lcl_output = save_fru('SHUB2', $nasid_value, $current_location, $fru_rsn, 'WEDGED_SHUB');
	printf "$lcl_output";
      }

      if ($show_clock == 1)  {
        chomp ($done_time = `date +'%T'`);
        print "\n parse_sh2_mmr_list\(\) completed at $done_time\n";
      }

      $sh2_cnt++; 
    }

}


#-------------------------------------------------------------------------------
#
#
#-------------------------------------------------------------------------------

sub parse_shub2_node_id  {

	my $bad_mmr_data = $_[0];
	$nasid_value = ($bad_mmr_data == 0) ? reg_get_field_value($reg_ptr, 'NODE_ID') : 0xDEAD;
	$sh2_target_list[$sh2_cnt]->{REV} = reg_get_field_value($reg_ptr, 'REVISION');
	$prefix = sprintf ("0x%04X", $nasid_value);
	$sh2_target_list[$sh2_cnt]->{NASID} = $nasid_value;
        $sh2_node_hdr = sprintf "\n\nReading Shub2 Registers on %s of %s   NASID: 0x%04X   Rev: %d   (Shub2.%d)",
              $sh2_target_list[$sh2_cnt]->{NODE}, $sh2_target_list[$sh2_cnt]->{MODULE}, $nasid_value, $sh2_target_list[$sh2_cnt]->{REV},
	      $sh2_target_list[$sh2_cnt]->{REV} - 1;
        $sh2_target_list[$sh2_cnt]->{ERR_SUMM} = sprintf "\n\nError Registers found on Shub2 on %s of %s    NASID: 0x%04X   Rev: %d   (Shub2.%d)",
              $sh2_target_list[$sh2_cnt]->{NODE}, $sh2_target_list[$sh2_cnt]->{MODULE}, $nasid_value, $sh2_target_list[$sh2_cnt]->{REV},
              $sh2_target_list[$sh2_cnt]->{REV} - 1;
        $sh2_target_list[$sh2_cnt]->{CFG_SUMM} = sprintf "\n\nConfiguration Errors found on Shub2 on %s of %s    NASID: 0x%04X   Rev: %d   (Shub2.%d)",
              $sh2_target_list[$sh2_cnt]->{NODE}, $sh2_target_list[$sh2_cnt]->{MODULE}, $nasid_value, $sh2_target_list[$sh2_cnt]->{REV},
              $sh2_target_list[$sh2_cnt]->{REV} - 1;
        $sh2_target_list[$sh2_cnt]->{DIMM_SUMM} = sprintf "\n\nDIMM Problems found on %s of %s    NASID: 0x%04X    Rev: %d  (Shub 2.%d)\n\n",
              $sh2_target_list[$sh2_cnt]->{NODE}, $sh2_target_list[$sh2_cnt]->{MODULE}, $nasid_value, $sh2_target_list[$sh2_cnt]->{REV},
              $sh2_target_list[$sh2_cnt]->{REV} - 1;

  	if ($global_rev == 0)  {
    	  $global_rev = $sh2_target_list[$sh2_cnt]->{REV};
  	}
 	elsif ($global_rev != $sh2_target_list[$sh2_cnt]->{REV})  {
    	  printf ("The revision value on this brick doesn't match the machine rev of %d\n", $global_rev);
    	  $sh2_target_list[$sh2_cnt]->{CFG_SUMM} .= sprintf ("\nThe revision value on this brick doesn't match the machine rev of %d\n", $global_rev);
    	  $sh2_target_list[$sh2_cnt]->{CFG_ERR} = 1;
	  $any_cfg_error = 1;
    	  $global_rev_mismatch = 1;
  	}

        my $new_nasid = 1;
	if ($bad_mmr_data == 0) {
          foreach $lcl_nasid (@sh2_nasid_list) {
            if ($nasid_value == $lcl_nasid)  {
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf "\n   %s: ERROR!! NASID 0x%04X is already defined.\n", $prefix, $nasid_value;
              $any_sh2_error = 1;
              $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
	      $global_err_val |= $SHUB_HW_ERR;
              printf "\n\nALERT!! This Nasid value is already defined!!";
              $new_nasid = 0;
              last;
            }
          }
        }


        if ($new_nasid == 1)  {
            push(@sh2_nasid_list, $nasid_value);
        }
	$err_summ_msg = $sh2_target_list[$sh2_cnt]->{ERR_SUMM};
        my $node_reg_val = lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);

        my $lcl_shr_1k = reg_get_field_value($reg_ptr, 'SHARERS_1K');
	if (! defined $master_shr_1k) {
	  $master_shr_1k = $lcl_shr_1k;
	}
	else {			# check to ensure all nodes are setup the same
	  if ($master_shr_1k != $lcl_shr_1k) {
            $sh2_target_list[$sh2_cnt]->{CFG_SUMM} .= sprintf ("  The SHARERS_1K value on this node \( %d \) doesn't match the defined value of %d\n", $lcl_shr_1k, $master_shr_1k);
            $sh2_target_list[$sh2_cnt]->{CFG_ERR} = 1;
            $any_cfg_error = 1;
	  }
	}
        my $lcl_npb = reg_get_field_value($reg_ptr, 'NODES_PER_BIT');
	if (! defined $master_npb) {
	  $master_npb = $lcl_npb;
	}
	else {			# check to ensure all nodes are setup the same
	  if ($master_npb != $lcl_npb) {
            $sh2_target_list[$sh2_cnt]->{CFG_SUMM} .= sprintf ("  The NODES_PER_BIT value on this node \( %d \) doesn't match the defined value of %d\n", $lcl_npb, $master_npb);
            $sh2_target_list[$sh2_cnt]->{CFG_ERR} = 1;
            $any_cfg_error = 1;
	  }
	}
        my $lcl_shr_md = reg_get_field_value($reg_ptr, 'SHARING_MODE');
	if (! defined $master_shr_md) {
	  $master_shr_md = $lcl_shr_md;
	}
	else {			# check to ensure all nodes are setup the same
	  if ($master_shr_md != $lcl_shr_md) {
            $sh2_target_list[$sh2_cnt]->{CFG_SUMM} .= sprintf ("  The SHARING_MODE value on this node \( %d \) doesn't match the defined value of %d\n", $lcl_shr_md, $master_shr_1k);
            $sh2_target_list[$sh2_cnt]->{CFG_ERR} = 1;
            $any_cfg_error = 1;
	  }
	}

	return($node_reg_val);
}

#---------------------------------------------------------------------------------
#
#
#---------------------------------------------------------------------------------

sub print_proc_cfg {
    
    my $proc_type = $_[0];
    $sh2_target_list[$sh2_cnt]->{PROC_CFG} = get_proc_cfg($proc_type);
    $sh2_node_hdr .= sprintf "\n  Hardware Configuration: %s\n", $sh2_target_list[$sh2_cnt]->{PROC_CFG};
    printf "%s", $sh2_node_hdr;
    $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf "\n  Hardware Configuration: %s\n", $sh2_target_list[$sh2_cnt]->{PROC_CFG};
    $sh2_target_list[$sh2_cnt]->{CFG_SUMM} .= sprintf "\n  Hardware Configuration: %s\n", $sh2_target_list[$sh2_cnt]->{PROC_CFG};

}

#-------------------------------------------------------------------------------
#
#	Function to dump shub2 NODE_PRESENT table and display any non-zero entries.
#
#-------------------------------------------------------------------------------

sub dump_shub2_node_table  {

    my $tbl_ptr = reg_create_mmr_inst("shub2_01", "SH2_NODE_PRESENT_TABLE");
    if ( !reg_is_valid($tbl_ptr) )
    {  $emess = "Invalid MMR name specified in subroutine dump_shub2_node_table().";
       $emess .= "\nMMR name specified: SH2_NODE_PRESENT_TABLE";
       &$comm_err_fun($emess);
       return;
    }

    my @addr = ();
    my @data = ();
    for (my $jj = 0, $a = reg_addr($tbl_ptr); $jj < 128; $jj++, $a += 8)
    {  $addr[$jj] = $a;
    }

    my $xfr_complete = 0;
    my $xfr_f_cnt = 0;
    while (($xfr_complete == 0) && ($xfr_f_cnt < 3) && ($sh2_target_list[$sh2_cnt]->{XFR_ERR_CNT} < 6))  {
      $SSCOPE_ERR = 0;	
      @data = mmr_read_group(\@addr);
      if ($SSCOPE_ERR == 1)  {
        $xfr_f_cnt++;
        $sh2_target_list[$sh2_cnt]->{XFR_ERR_CNT}++;
        $SSCOPE_ERR = 0;
      }
      else {
        $xfr_complete = 1;
      }
    }

    my $mmr_display = "\n";
    my $e = 0;

    foreach $entry (@data)  {
      if ($entry > 0)  {
	my $lwr_node = $e * 64;	
	my $upr_node = $lwr_node + 63;	
	my $node_bound = sprintf "[%d:%d]", $upr_node, $lwr_node;
	$mmr_display .= sprintf "%s : SH2_NODE_PRESENT_TABLE # %3d   Addr: 0x%09llX : Nodes %11s  : Data = 0x%016llX\n",
		$prefix, $e, $addr[$e], $node_bound, $entry;
      }
      $e++;
    }
    $mmr_display .= "\n";
    return($mmr_display);
}


#-------------------------------------------------------------------------------
#
#	Function to update SH2_XN_PI_SIC_CMP_DATA0 so smdk knows the node is being dumped.
#	This is to prevent bogus NI link error threshold failures if a smdk diag is
#	being run while errdmp is running.
#
#	A value of 0xdeadbeef is written into the reg before modifying Shub testpoint
#	config mmrs, and the original value is written back into the reg after the
#	testpoint reading is complete.
#
#	May want to investigate making this a multi-chip operation.  GEM
#	Also, may need to do this as a seperate step inside read_shub2x_mmrs(), before
#	reading any mmrs.  Need to test first.
#
#-------------------------------------------------------------------------------


sub notify_smdk {

  $notify_reg = reg_create_mmr_inst("shub2_01", "SH2_XN_PI_SIC_CMP_DATA0");
  
  if ($_[0] == 1)  {		# save original value, and overwrite it
    my $xfr_complete = 0;
    my $xfr_f_cnt = 0;
    while (($xfr_complete == 0) && ($xfr_f_cnt < 3) && ($sh2_target_list[$sh2_cnt]->{XFR_ERR_CNT} < 6))  {
      $SSCOPE_ERR = 0;	
      reg_mmr_read($notify_reg);
      if ($SSCOPE_ERR == 1)  {
        $xfr_f_cnt++;
        $sh2_target_list[$sh2_cnt]->{XFR_ERR_CNT}++;
        $SSCOPE_ERR = 0;
        print "\n Error detected after reg_mmr_read() attempt in notify_smdk()\n";
      }
      else {
        $xfr_complete = 1;
      }
    }
    $sh2_target_list[$sh2_cnt]->{ORIG_SMDK_VAL} = reg_get_field_value($prom_war_reg, 'DATA0');
  }

  my $write_val = ($_[0] == 1) ? 0xdeadbeef : $sh2_target_list[$sh2_cnt]->{ORIG_SMDK_VAL};
  reg_set_field_value($notify_reg, 'DATA0', $write_val);

  my $xfr_complete = 0;
  my $xfr_f_cnt = 0;
  while (($xfr_complete == 0) && ($xfr_f_cnt < 3) && ($sh2_target_list[$sh2_cnt]->{XFR_ERR_CNT} < 6))  {
    $SSCOPE_ERR = 0;	
    reg_mmr_write($notify_reg);
    if ($SSCOPE_ERR == 1)  {
      $xfr_f_cnt++;
      $sh2_target_list[$sh2_cnt]->{XFR_ERR_CNT}++;
      $SSCOPE_ERR = 0;
      print "\n Error detected after reg_mmr_write() attempt in notify_smdk()\n";
    }
    else {
      $xfr_complete = 1;
    }
  }
}

#---------------------------------------------------------------------------------
#
#   Function to set parameters used for decoding and formatting the addresses
#   found in PI crbs, PI error detail regs, and PI detected memory errors.
#
#---------------------------------------------------------------------------------

sub chk_addr_mode {

  my $addr_mode = $_[0];

  if (! defined $first_addr_chk) {
    print "\n Capturing address mode of $addr_mode on the first node.\n"  if ($verbose_mode == 1);
    $sh2_addr_mode = $addr_mode;		# global variable
    $dimm_group_index = 36;			# global variable
    $first_addr_chk = 1;

    if ($sh2_addr_mode == 0)  {
      $nasid_index = 40;
      $nasid_mask =  0x03ff;
      $offset_mask = 0x3fffffffff;
    }
    elsif ($sh2_addr_mode == 1)  {
      $nasid_index = 39;
      $nasid_mask =  0x07ff;
      $offset_mask = 0x1fffffffff;
    }
    elsif ($sh2_addr_mode == 2)  {
      $nasid_index = 38;
      $nasid_mask =  0xfff;
      $offset_mask = 0x0fffffffff;
    }
    elsif ($sh2_addr_mode == 3)  {
      $nasid_index = 37;
      $nasid_mask =  0x1fff;
      $offset_mask = 0x07ffffffff;
    }
    elsif ($sh2_addr_mode == 4)  {
      $nasid_index = 36;
      $nasid_mask =  0x3fff;
      $offset_mask = 0x03ffffffff;
    }
  }
  else {
    if ($addr_mode != $sh2_addr_mode) {		# config error
          $sh2_target_list[$sh2_cnt]->{CFG_SUMM} .= sprintf "The address mode (%d) on NASID 0x%04X doesn't match the defined mode of %d\n", $addr_mode, $nasid_value, $sh2_addr_mode;
          $sh2_target_list[$sh2_cnt]->{CFG_ERR} = 1;
          $any_cfg_error = 1;
    }
  }
}

#---------------------------------------------------------------------------------
#
# * The following info describes how a resource ID for a given thread
# * is seen by the external agent. The external agent (Hub) uses
# * this value to assign CRBs and the like to the thread. On SHub2
# * this is determined by the TDR setting of SH2_PROCESSOR_TYPE.
# * SHub2 will use either bits [3:2] (type == 1) or bits [2:1]
# * (type == 0). The meaning of the bits also depends on the
# * format that the processor is using. On Montecito this is
# * configurable via A8#. On older cpus the format is fixed.
# * 
# * The two possible formats of AgentID as seen on the bus (and as
# * PAL_FIXED_ADDR) are:
# *
# * Legacy format (Madison and earlier, Montecito A8#=1) (TPPC):
# * AgentID[3]   - thread ID (if enabled)
# * AgentID[2:1] - package ID
# * AgendID[0]   - core ID (unused for legacy CPUs)
# *
# * New format (default on Montecito, A8#=0) (TCPP):
# * AgentID[3]   - thread ID (if enabled)
# * AgentID[2]   - core ID
# * AgentID[1:0] - package ID
# *
# * P - Packages  C - Cores/Package   T - Threads/Core
# *
# * Due to the way the bus is wired (for DP sku compatibility),
# * package IDs 0 and 2 are used.  Therefore the following AgentIDs
# * appear, with the given SH2_PROCESSOR_TYPE and A[8]# settings:
# *
# *               | SH2_PRO-    |       |         
# * CPU(s)        | CESSOR_TYPE | A[8]# | AgentIDs
# * --------------+-------------+-------+-----------------------------
# * 1 Legacy CPU  |      1      |  N/A  | 4b'0000
# * 2 Legacy CPUs |      1      |  N/A  | 4b'0000, 4b'0110
# * --------------+-------------+-------+-----------------------------
# * Montecito     |             |       |
# * 1 P, 1 C, 1 T |      1      |  0    | 4b'0000
# * 1 P, 2 C, 1 T |      1      |  0    | 4b'0000, 4b'0100
# * 1 P, 1 C, 2 T |      1      |  0    | 4b'0000, 4b,1000
# * 1 P, 2 C, 2 T |      1      |  0    | 4b'0000, 4b'1000, 4b'0100, 4b'1100
# * 2 P, 1 C, 1 T |      1      |  1    | 4b'0000, 4b'0110
# * 2 P, 2 C, 1 T |      0      |  0    | 4b'0000, 4b'0100, 4b'0011, 4b'0111
# * 2 P, 1 C, 2 T |      1      |  1    | 4b'0000, 4b'1000, 4b'0110, 4b'1110
# *
# * From the AgentIDs, SHub2 extracts resource IDs (which software
# * converts to cr.lid values (CPU IDs)) according to the following
# * table:
# *
# *               | SH2_PRO-    |        |    Resource    |      
# * CPU(s)        | CESSOR_TYPE | Format |       IDs      | CPU IDs
# * --------------+-------------+--------+----------------+-------------------
# * 1 Legacy CPU  |      1      |  TPPC  | 00             | (0)
# * 2 Legacy CPUs |      1      |  TPPC  | 00, 01         | (0), (2)
# * --------------+-------------+--------+----------------+-------------------
# * Montecito     |             |        |                |
# * 1 P, 1 C, 1 T |      1      |  TCPP  | 00             | (0)
# * 1 P, 2 C, 1 T |      1      |  TCPP  | 00, 01         | (0), (2)
# * 1 P, 1 C, 2 T |      1      |  TCPP  | 00, 10         | (0,1)
# * 1 P, 2 C, 2 T |      1      |  TCPP  | 00, 10, 01, 11 | (0,1), (2,3)
# * 2 P, 1 C, 1 T |      1      |  TPPC  | 00, 01         | (0), (2)
# * 2 P, 2 C, 1 T |      0      |  TCPP  | 00, 10, 01, 11 | (0), (1), (2), (3)
# * 2 P, 1 C, 2 T |      1      |  TPPC  | 00, 10, 01, 11 | (0,1), (2,3)
# *
# * Software converts a resource ID to a CPU ID by swapping the high
# * and low bits in the value.
#
#
#	Decode for errdmp's internal use of {AID_DECODE} :
#	(NOTE, if bit 0 is set in the decode, the TCPP method will be used 
#	 to translate the Agent ID bits)
#
#		0:  IP63, 1 Madison processor					
#		1:  IP63, 1 Montecito processor
#		2:  IP71/3/5, 1 or 2 Madison processors
#		3:  IP71/3/5, 1 Montecito or Montvale processor
#		4:  IP71/3/5, 2 Montecito or Montvale processors
#		5:  IP71/3/5, 2 Montecito or Montvale processors with SH2_PROCESSOR_TYPE=0
#
#
#---------------------------------------------------------------------------------

sub get_proc_cfg {

  my $mmr_proc_type = $_[0];	#  value from SH2_PROCESSOR_TYPE - 0= 1 or 2 Madison or 2 Montecito's,  1= 1 Montecito w/HT enabled
  my $bld = $sh2_target_list[$sh2_cnt]->{NODE};
  $bld =~ s/LADE//;
  my $brk = $sh2_target_list[$sh2_cnt]->{MODULE};
  my $cfg_string = "Config Error!  Unable to find this blade in the compute blade list.";
  my $blade_found = 0;
  my $part_name = "Processor";
 
  foreach $cbld (@c_blade_list) {
    if ( (($bld eq $cbld->{BLADE_LOC}) && ($brk eq $cbld->{IRU}))  # Tollhouse IRU case
       || (($brk eq $cbld->{IRU}) && ($cbld->{BLADE_TYPE} =~ /IP71/)) )  {
      $blade_found = 1;
      if ($cbld->{BLADE_TYPE} =~ /IP[78][1359]/) {
	$sh2_proc_type += 2;
      }
      if ($cbld->{PROC_CNT} == 2) {
        $part_name = "Processors";
 	if ($cbld->{PROC_TYPE} =~ /MONT/) {		# MONTECITO or MONTVALE
	  $sh2_proc_type += 2;
	}
      }
      if (($cbld->{PROC_CNT} == 1) && ($cbld->{PROC_TYPE} =~ /MONT/)) {     # MONTECITO or MONTVALE
	$sh2_proc_type += 1;
      }
      if (($sh2_proc_type == 4) && ($mmr_proc_type == 0)) {
        $sh2_target_list[$sh2_cnt]->{AID_DECODE} = 5;		# define how we will decode the Agent ID
	$sh2_proc_type += 1;
      }
      else {
          $sh2_target_list[$sh2_cnt]->{AID_DECODE} = $sh2_proc_type;
      }

      $cfg_string = sprintf "%s sn %s with %d %s %s  \(%d\)", $cbld->{BLADE_TYPE}, $cbld->{BLADE_SN}, 
           $cbld->{PROC_CNT}, $cbld->{PROC_TYPE}, $part_name, $sh2_target_list[$sh2_cnt]->{AID_DECODE};
      $processor_count = $cbld->{PROC_CNT};
      $blade_type = $cbld->{BLADE_TYPE};
      $blade_sn   = $cbld->{BLADE_SN};
      $processor_type = ($cbld->{PROC_TYPE} =~/MONTECITO/)? 'Montecito' :
			($cbld->{PROC_TYPE} =~/MONTVALE/)? 'Montvale' : 'Madison';
      last;
    }
  }
  
  if ($blade_found == 0)  {
    $sh2_target_list[$sh2_cnt]->{CFG_ERR} = 1;
  }
  return($cfg_string);
}

#---------------------------------------------------------------------------------
#
#
#---------------------------------------------------------------------------------

sub sh2_proc_chk {

  my $proc_type = $_[0];	#  value from SH2_PROCESSOR_TYPE - 0= 1 or 2 Madison or 2 Montecito's,  1= 1 Montecito w/HT enabled

  if (($proc_type == 0) && (($sh2_proc_type & 1) == 0)) {         # config error
    $sh2_target_list[$sh2_cnt]->{CFG_SUMM} .= sprintf ("The SH2_PROCESSOR_TYPE value \(%d\) on NASID 0x%04X doesn't match the defined type of %d\n", $proc_type, $nasid_value, $sh2_proc_type);
        $sh2_target_list[$sh2_cnt]->{CFG_ERR} = 1;
        $any_cfg_error = 1;
  }

}

#---------------------------------------------------------------------------------
#
#
#---------------------------------------------------------------------------------

sub capture_time_stamp {

    my $reg_ptr = $_[0];		# Register instance reference

    $any_sh2_error = 1;
    $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
    $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
    $time_list[$t_stamps]->{NASID} = $nasid_value;
    $time_list[$t_stamps]->{REG} = reg_name($reg_ptr);
    $time_list[$t_stamps]->{TIME} = reg_get_field_value($reg_ptr,'TIME');
    $time_list[$t_stamps]->{LOCATION} = $current_location;
    $t_stamps++;			# this variable is global from errdmp

    if (defined $sh2_target_list[$sh2_cnt]->{TARGET_0_VAL})  {
      $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf("\n    Target Identifier[0] Address: 0x%014llX ", $sh2_target_list[$sh2_cnt]->{TARGET_ID_0});          
      if (defined $sh2_target_list[$sh2_cnt]->{TARGET_ID_0_TIMESTAMP})  {
        $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf("    Target Id Time Stamp: 0x%016llX ", $sh2_target_list[$sh2_cnt]->{TARGET_ID_0_TIMESTAMP});
      }
      $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "\n";
    }

    if (defined $sh2_target_list[$sh2_cnt]->{TARGET_1_VAL})  {
      $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf("\n    Target Identifier[1] Address: 0x%014llX ", $sh2_target_list[$sh2_cnt]->{TARGET_ID_1});
      if (defined $sh2_target_list[$sh2_cnt]->{TARGET_ID_1_TIMESTAMP})  {
        $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf("    Target Id Time Stamp: 0x%016llX ", $sh2_target_list[$sh2_cnt]->{TARGET_ID_1_TIMESTAMP});
      }
      $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "\n";
    }

    if (defined $sh2_target_list[$sh2_cnt]->{TARGET_2_VAL})  {
      $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf("\n    Target Identifier[2] Address: 0x%014llX ", $sh2_target_list[$sh2_cnt]->{TARGET_ID_2});
      if (defined $sh2_target_list[$sh2_cnt]->{TARGET_ID_2_TIMESTAMP})  {
        $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf("    Target Id Time Stamp: 0x%016llX ", $sh2_target_list[$sh2_cnt]->{TARGET_ID_2_TIMESTAMP});
      }
      $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "\n";
    }

    if (defined $sh2_target_list[$sh2_cnt]->{TARGET_3_VAL})  {
      $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf("\n    Target Identifier[3] Address: 0x%014llX ", $sh2_target_list[$sh2_cnt]->{TARGET_ID_3});
      if (defined $sh2_target_list[$sh2_cnt]->{TARGET_ID_3_TIMESTAMP})  {
        $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf("    Target Id Time Stamp: 0x%016llX ", $sh2_target_list[$sh2_cnt]->{TARGET_ID_3_TIMESTAMP});
      }
      $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "\n";
    }
}

#-------------------------------------------------------------------------
#
# subroutine to convert a fsb address to a network address, taking into account
# the address mode selected in the shub
# 
# global variables are initialized in chk_addr_mode();
#
#-------------------------------------------------------------------------

sub fsb_to_net {

  my $fsb_address = $_[0];
  my $dir_address = 0;			#  address as seen at the directory
  my $dir_nasid = 0;			#  nasid value as seen at the directory

  my $fsb_dimm_group_index = $nasid_index - 4;
  my $dir_nasid = ($fsb_address >> $nasid_index) & $nasid_mask;
  $dir_address  = (($fsb_address >> $fsb_dimm_group_index) & 3) << $dimm_group_index;
  $dir_address += $fsb_address & ($offset_mask >> 2);
  $dir_address += (($fsb_address >> $nasid_index) & $nasid_mask) << 38;

  return($dir_address, $dir_nasid); 

}

#---------------------------------------------------------------------------------
#
#
#---------------------------------------------------------------------------------

sub decode_lb_stat {

  my $stat_val = $_[0];

  my $ret_string = "";

  if (($stat_val & 1) > 0) {
    $ret_string .= "Sequence Error   ";
  }
  if (($stat_val & 2) > 0) {
    $ret_string .= "Nack Overflow Error   ";
  }
  if (($stat_val & 4) > 0) {
    $ret_string .= "FProm Addressing Error   ";
  }
  if (($stat_val & 8) > 0) {
    $ret_string .= "Junk Bus Address Range Error   ";
  }
  if (($stat_val & 0x10) > 0) {
    $ret_string .= "Junk Bus Timeout Error   ";
  }

  return($ret_string);
}

#---------------------------------------------------------------------------------
#
#
#---------------------------------------------------------------------------------

sub display_sh2_mem_cfg
{
  print "\nMemory Configuration Summary:\n\n";
  print "                                               DIMM Group 0           DIMM Group 1           DIMM Group 2\n";
  print "    Location        NASID   Speed CAS_Lat    Size   Banks RV CS     Size   Banks RV CS     Size   Banks RV CS\n";
  print " ---------------    -----   -------------  -------- ----- -- --   -------- ----- -- --   -------- ----- -- --";

  my $m_cnt = 0;
  while (defined $sh2_target_list[$m_cnt]) {
    my $phys_loc = sprintf("%s of %s", $sh2_target_list[$m_cnt]->{NODE}, $sh2_target_list[$m_cnt]->{MODULE});
    printf "\n %15s  0x%04X   %s    %s",
        $phys_loc,
        $sh2_target_list[$m_cnt]->{NASID},
        $sh2_target_list[$m_cnt]->{DIMM_SPD},       # generated in get_sh2_mem_cfg() in md_dir_state.pm
        $sh2_target_list[$m_cnt]->{MEM_CFG};        #   "    "      "  "
    $m_cnt++;
  }
  print "\n\n";
}

#---------------------------------------------------------------------------------
#
#  This function re-reads the supplied credit MMR to check if
#  indicated Sic credit values are consistently low.
#
#---------------------------------------------------------------------------------

sub xn_sic_cred_chk {

  my $crd_reg = "$_[0]";
  my $read_cnt = $_[1];
  my $crd_type = "$_[2]";
  my $crd_thresh = $_[3];

  my $low_crd_cnt = 0;
  my $i = 0;

  for ($o_cnt = 0; $o_cnt < 10; $o_cnt++)  {	# repeat outer loop 10 times 
    my @crd_reg_list = ();
    $xnmd_reg_ptr = reg_create_mmr_inst("shub2_01", $crd_reg);
    for ($i = 0; $i < $read_cnt; $i++) {
      my $crd_reg_ptr = reg_create_mmr_inst("shub2_01", $crd_reg);
      if ( !reg_is_valid($crd_reg_ptr) )
      {  $emess = "Error in xn_sic_cred_chk().  Unable to create register instance: $crd_reg.\n";
         dsp_error_and_exit($emess);
      }
      $crd_reg_list[$i] = $crd_reg_ptr;
    }
  
    $xfr_complete = 0;
    $xfr_f_cnt = 0;
    while (($xfr_complete == 0) && ($xfr_f_cnt < 3) && ($sh2_target_list[$sh2_cnt]->{XFR_ERR_CNT} < 6))  {
      $SSCOPE_ERR = 0;	
      reg_list_mmr_read(\@crd_reg_list);              # actual mmr read occurs here
      if ($SSCOPE_ERR == 1)  {
        $sh2_target_list[$sh2_cnt]->{XFR_ERR_CNT}++;
        $xfr_f_cnt++;
        $SSCOPE_ERR = 0;
      }
      else {
        $xfr_complete = 1;
      }
    }  

    for ($i = 0; $i < $read_cnt; $i++) {
      my $crd_reg_ptr = $crd_reg_list[$i];
      if (reg_get_field_value($crd_reg_ptr, $crd_type) < $crd_thresh) {
        $low_crd_cnt++;
      }
    }
  }

  return($low_crd_cnt);
}

#---------------------------------------------------------------------------------
#
#
#---------------------------------------------------------------------------------

sub agent_check {
  my $agent_val = $_[0];
  my $bad_agent_status = 0;
  my $mask_val = 0x3;
  $agent_val = $agent_val & $mask_val;
  my $agent_status_msg = "";
  my $fru_msg = "";

  my $exp_agent_val = ($processor_count == 2) ? 3 : 
		      ($processor_count == 1) ? 1 : 0;
  
  if ($exp_agent_val != $agent_val) {
    $bad_agent_status = 1;
    $agent_status_msg = "\tAGENT STATUS: $agent_val     EXPECTED: $exp_agent_val\n";
    my $poll_count = 5;
    my $mmr_val_ok_cnt = ($file_mode == 1) ? 0 : poll_mmr_func('SH2_FSB_SYMMETRIC_AGENT_STATUS', $poll_count, $exp_agent_val, $mask_val);
    # if ($mmr_val_ok_cnt == 0) {
      # if (($exp_agent_val & $agent_val & 1) != 1)  {
        # $fru_msg .= "$processor_type CPU in SKT 0 isn't alive.  ";
      # }
      # if (($processor_count == 2) && (($exp_agent_val & $agent_val & 2) != 2)) {
        # $fru_msg .= "$processor_type CPU in SKT 1 isn't alive.  ";
      # }
      # $fru_list[$fru_count]->{DEVICE}   = "SHUB2";
      # $fru_list[$fru_count]->{NASID}    = $nasid_value;
      # $fru_list[$fru_count]->{LOCATION} = $current_location;
      # $fru_list[$fru_count]->{REASON}   = $fru_msg;
      # $sh2_output .= sprintf "\n%s: Added SHUB 2x NASID 0x%04X to the FRU list, entry %d.\n", $prefix, $sh2_target_list[$sh2_cnt]->{NASID}, $fru_count;
      # $fru_count++;
    # }
    if ($file_mode == 0) {
      $agent_status_msg .= sprintf "Re-read SH2_FSB_SYMMETRIC_AGENT_STATUS:  status was correct %d out of %d times.\n",
	$mmr_val_ok_cnt, ($poll_count * 10);
    }
  }

  print "AGENT STATUS:   $agent_val     EXPECTED:  $exp_agent_val\n" if ($verbose_mode == 1);
  return ($bad_agent_status, $agent_status_msg);
}

#---------------------------------------------------------------------------------
#
# This function re-reads the supplied register.  In each read instance,
# the value read is anded with the supplied mask value, and compared with
# the expected value.  A value returned is equal to the number of times
# the mask compared good.
#
#---------------------------------------------------------------------------------

sub poll_mmr_func {
  my $mmr_name = "$_[0]";
  my $read_cnt = $_[1];
  my $exp_val = $_[2];
  my $mask_val = $_[3];

  my $read_ok_cnt = 0;
  my $i = 0;

  for ($o_cnt = 0; $o_cnt < 10; $o_cnt++)  {    # repeat outer loop 10 times 
    my @poll_reg_list = (); 
    $poll_reg_ptr = reg_create_mmr_inst("shub2_01", $mmr_name); 
    for ($i = 0; $i < $read_cnt; $i++) { 
      my $crd_reg_ptr = reg_create_mmr_inst("shub2_01", $mmr_name); 
        if ( !reg_is_valid($poll_reg_ptr) ) {  
         $emess = "Error in poll_mmr_func().  Unable to create register instance: $mmr_name.\n";
         dsp_error_and_exit($emess);
      }
      $poll_reg_list[$i] = $poll_reg_ptr;
    }

    $xfr_complete = 0;
    $xfr_f_cnt = 0;
    while (($xfr_complete == 0) && ($xfr_f_cnt < 3) && ($sh2_target_list[$sh2_cnt]->{XFR_ERR_CNT} < 6))  {
      $SSCOPE_ERR = 0;	
      reg_list_mmr_read(\@poll_reg_list);              # actual mmr read occurs here
      if ($SSCOPE_ERR == 1)  {
        $sh2_target_list[$sh2_cnt]->{XFR_ERR_CNT}++;
        $xfr_f_cnt++;
        $SSCOPE_ERR = 0;
      }
      else {
        $xfr_complete = 1;
      }
    }

    for ($i = 0; $i < $read_cnt; $i++) {
      my $poll_reg_ptr = $poll_reg_list[$i];
      if ((reg_get_value($reg_ptr) & $mask_val) == $exp_val) {
        $read_ok_cnt++;
      }
    }
  }

  return($read_ok_cnt);
}

#---------------------------------------------------------------------------------
#
#  This function uses the supplied reg_list array to parse and display Shub2 MMR info.
#
#---------------------------------------------------------------------------------

sub parse_sh2_mmr_list  {

      my $sh2_reg_list_ref = $_[0];
      
      my @sh2_reg_list = ();
      @sh2_reg_list = @$sh2_reg_list_ref;
      
      $nasid_value = 0;
      $node_not_booted = 0;
      $remap_enabled = 0;
      $processor_count = 0;			# count of processors found per blade set in get_proc_cfg()
      $remap_offset = 0;
      $err_summ_msg = "";
      my $sh2_output = "";
      $blade_type = 'IP73';			# default
      $blade_sn = '??????';			# default
      $sh2_proc_type = 0;			# default
      my $shub_hw_err = 0;
      my $pi_dtl_vld = 0;
      my $pi_dtl_vld_mask = 0xFFE1FFFFF;	#  if any of these bits are set, we'll parse SH2_PI_ERROR_DETAIL_1 & 2
      my $pi_crb_err = 0;
      my $pi_to_err = 0;			# CRB Timeout in PI FIRST Error
      my $pi_to_err_s = 0;			# CRB Timeout in PI Error Summary
      my $pi_fsb_sw_err = 0;
      my $pi_fsb_hw_err = 0;
      my $pi_first_err_val = 0;
      my $pi_dtl1_node = 0;
      $pi_dtl1_addr = 0;
      my $pi_fsb_sw_fru_entry = 0;
      my $pi_fsb_hw_fru_entry = 0;
      my $pi_fsb_ce_fru_entry = 0;
      my $pi_fsb_uce_fru_entry = 0;
      my $pi_fsb_uce_poisoned = 0;
      my $pi_fsb_ce_err = 0;
      my $pi_shub_ce_err = 0;
      my $pi_fsb_uc_err = 0;
      my $pi_shub_uc_err = 0;
      my $pi_first_err_str = "";
      my $pi_fsb_proto_err = 0;
      my $pi_err_save_addr = 0;
      my $pi_pio_err = 0;
      $pi_xb_proto = 0;
      my $pi_xb_proto_addr = 0;
      my $bt_err = 0;
      my $bt_xn_err = 0;
      my $bt_crb_err = 0;
      my $lb_err = 0;
      my $bw_963416_attr = 0;
      my $md_data_err_uce = 0;
      my $md_data_err_ce = 0;
      my $md_err_s_mask = 0xF06FFFFFFFF;		# MEM_ERR_SUMMARY - Ignoring DIMM parity, RED_BLACK_ERR_TIMEOUT, & BIST_DONE
      my $md_err_fo_mask = 0x786FFFFFFFF;		# MEM_FIRST and OVERLOW - Ignoring DIMM parity, RED_BLACK_ERR_TIMEOUT, & BIST_DONE
      my $md_hw_err = 0;
      my $md_dqlp_uce = 0;				# for handling MD error reporting issue in BW# 959587
      my $md_dqrp_uce = 0;				# for handling MD error reporting issue in BW# 959587
      my $md_addr_lower = 0;				# for handling MD error reporting issue in BW# 959587
      my $md_dqlp_perr = 0;
      my $md_dqrp_perr = 0;
      my $get_misc_upper = 0;
      my $xn_err = 0;
      my $xn_pi_err = 0;
      my $xn_pi_lre = 0;
      my $xn_md_err = 0;
      my $xn_md_lre = 0;
      my $xn_iilb_err = 0;
      my $xn_iilb_dump_hdr = 0;
      my $xn_iilb_lre = 0;
      my $xn_md_ce = 0;
      my $xn_md_uc = 0;
      my $xn_ni0_err = 0;
      my $xn_ni0_dump_hdr = 0;
      my $xn_ni1_err = 0;
      my $xn_ni1_dump_hdr = 0;
      my $x_dimm_cfg = 0;
      my $ni0_link_up = 0;
      my $ni1_link_up = 0;
      my $valid_psp_state = 0;

      my $pi_numa_error = 0;
      my $pi_numa_err_src_node = 0;
      my $pi_numa_err_addr = 0;
      my $md_numa_error = 0;
      my $md_numa_err_src_node = 0;
      my $md_numa_err_src_chip = "";
      my $md_numa_err_addr = 0;

      my $shadow_sh_fsb_ce = 0;
      my $shadow_sh_fsb_uce = 0;
      my $shadow_fsb_sh_ce = 0;
      my $shadow_fsb_sh_uce = 0;
      my $shadow_pi_ce = 0;
      my $shadow_pi_uce = 0;
      my $shadow_pi_write_ce = 0;
      my $shadow_pi_write_uce = 0;
      my $shadow_pi_read_ce = 0;
      my $shadow_pi_read_uce = 0;
      my $shadow_xn_ce = 0;
      my $shadow_xn_uce = 0;
      my $first_pi_uce = 0;		#  1: read   2: write    3: read (shadow)   4: write (shadow)
      my $shadow_event_mask = 0x480;

      $current_location = sprintf("%s, %s", $sh2_target_list[$sh2_cnt]->{MODULE}, $sh2_target_list[$sh2_cnt]->{NODE});

	#  Error Signature variables
      my $low_md_rq_crd  = 0;		# BW 938724   MD_XN request credits stuck at < 2
      my $low_pi_rp_crd  = 0;		# BW 940379   PI_XN reply credits stuck at < 2
      my $low_pi_rq_crd  = 0;		# BW 
      my $md_cmd_is_prge = 0;		# BW 939024   MD PERR command is a PURGE
      my $md_cmd_is_xfer = 0;		# BW 939024   MD PERR command is a XFER
      my $xn_cmd_is_prge = 0;		# BW 939024   XN_IILB command is a PURGE
      # my $ptc_pwerr	 = 0;		# BW 939509   PI received a PWERR on a PTC write
      my $ptc_pwerr	 = 0;		# BW 947466   PI received a PWERR on a PTC write
      my $pi_wr_cmd_q_af = 0;		# BW 945998   PI RDAT's dat_dfr_q overflowing
      $fsb_stalled	 = 0;		# BW 951089   Hung CPUs due to Shub asserting BPRI (provided $pi_crb_skid == 0)
      $flush_cfg_err  = 0;		# flush parameters are causing infinite flush
      $pi_crb_skid    = 0;		# indication that PI crb resources are low


      my $reg_idx = 0;
      my $bad_read_cnt = 0;
      my $previous_data = 0;
      my $same_data_count = 0;

      $dimm_group_0_cfg = 0;              #  set to defaults
      $dimm_group_1_cfg = 1;
      $dimm_group_2_cfg = 2;
      $dimm_group_3_cfg = 3;

      while ($reg_idx < $#sh2_reg_list)  {	   # begin parsing MMR's here
  	my $notify = "";
	my $mmr_data_invalid = 0;

      	$reg_ptr = $sh2_reg_list[$reg_idx];
      	my $mmr_data = reg_get_value($reg_ptr);
      	my $mmr_addr = reg_addr($reg_ptr);
	my $mmr_name = reg_name($reg_ptr);

        if ($reg_idx == 0) {				# check for the case where all MMR data is the same
	  $previous_data = $mmr_data;
 	}
	else {
	  if ($mmr_data == $previous_data) {
	    $same_data_count++;
	    $mmr_data_invalid = ($same_data_count == $reg_idx) ? 1 : 0;
	  }
	  $previous_data = $mmr_data;
	}

	if ($mmr_data == $mmr_addr)  { 
	    printf "Suspect bad read:  Address: 0x%X      Value read: 0x%X\n", $mmr_addr, $mmr_data;			# GEM:  DEBUG
	    $bad_read_cnt++;
	    $mmr_data_invalid = ($bad_read_cnt > 10) ? 1 : 0;
   	}

	if (($mmr_data_invalid == 0) || ($reg_idx < 2)) {

	  if ( $mmr_name =~ /SH2_NODE_ID/ )  {	    # parse & check NODE_ID reg
	    $sh2_output .= parse_shub2_node_id($bad_read_cnt);
	  }
	  elsif (( $mmr_name =~ /SH2_RTC\b/ ) && ($sh2_cnt == 0))   {	 
	    $dump_time = $mmr_data << 8;			#  used in errdmp main script
            $sh2_output .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
	  }
	  elsif ( $mmr_name =~ /SH2_PROCESSOR_TYPE/ )  {	
	    print_proc_cfg(reg_get_field_value($reg_ptr, 'CODE'));
	    sh2_proc_chk(reg_get_field_value($reg_ptr, 'CODE'));
            $sh2_output .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
            $fru_location = sprintf("%s, %s \(%s\)", $sh2_target_list[$sh2_cnt]->{MODULE}, $sh2_target_list[$sh2_cnt]->{NODE}, $blade_sn);
	  }
	  elsif ( $mmr_name =~ /SH2_ADDRESS_MODE/ )  {	
	    chk_addr_mode(reg_get_field_value($reg_ptr, 'CODE'));
            $sh2_output .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
	  }
	  elsif (( $mmr_name =~ /SH2_FSB_SYMMETRIC_AGENT_STATUS/ ) && ($processor_count > 0))  {	
	    my $sym_ag_stat = 0;
	    my $sym_ag_err = "";
	    ($sym_ag_stat, $sym_ag_err) = agent_check($mmr_data);
	    if ($sym_ag_stat == 1)  {
	      $any_sh2_error = 1;
	      $notify = "  <<<<<<<";
	      $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
	      my $mmr_display = lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
              $sh2_output .= $mmr_display;
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "$mmr_display $sym_ag_err";
	      $global_err_val |= $SHUB_HW_ERR;
	    }
	    else {
              $sh2_output .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
	    }
	  }
	  elsif ( $mmr_name =~ /SH2_FIRMWARE_REMAP/ )  {	
	    $remap_enabled = reg_get_field_value($reg_ptr, 'ENABLE');
	    $remap_offset = reg_get_field_value($reg_ptr, 'OFFSET');
            $sh2_output .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
	  }
	  elsif (($mmr_name =~ /NODE_PRESENT/) && ($file_mode == 0)) {	    # dump complete NODE_PRESENT_TABLE
	    $sh2_output .= dump_shub2_node_table();
	  }
	  elsif (($mmr_name =~ /SH2_MEM_ERROR_MASK/) && ($sh2_cnt == 0))  {	    # save a copy
	    $sh2_md_mask = $mmr_data;
            $sh2_output .= lcl_reg_write_value($reg_ptr, 'valid', $prefix, 0, $notify);
	  }
	  elsif ( $mmr_name =~ /DIMM_CFG/ )  {	   
            $sh2_output .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
	    if ( $mmr_name =~ /X/ )  {	   
	      get_sh2_mem_cfg($mmr_data);
	      $x_dimm_cfg = $mmr_data;
	    }
	    else {
	      if ($mmr_data != $x_dimm_cfg) {
    	        $sh2_target_list[$sh2_cnt]->{CFG_SUMM} .= sprintf "The value in %s doesn't match the value in SH2_X_DIMM_CFG ", $mmr_name;
    	        $sh2_target_list[$sh2_cnt]->{CFG_ERR} = 1;
	        $any_cfg_error = 1;
	      }
	    }
	  }

	  elsif (($mmr_name =~ /SH2_FSB_SYSTEM_AGENT_CONFIG/) && ((reg_get_field_value($reg_ptr, 'SERIALIZE_FSB_EN')) == 1)
		&& ($processor_count > 0))  {
            $sh2_output .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
    	    $sh2_target_list[$sh2_cnt]->{CFG_ERR} = 1;
    	    $sh2_target_list[$sh2_cnt]->{CFG_SUMM} .= sprintf ("\n%s: FSB Transactions are serialized in %s \n", $prefix, $mmr_name);
	    $any_cfg_error = 1;
  	  }
	  elsif (($mmr_name =~ /SH2_CRB_MESSAGE_CONTROL/) && ($processor_count > 0))  {
            $sh2_output .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
	    if ((reg_get_field_value($reg_ptr, 'ENABLE_XB_PIPE_NACK_HAMMER')) == 0) { 
    	      $sh2_target_list[$sh2_cnt]->{CFG_ERR} = 1;
    	      $sh2_target_list[$sh2_cnt]->{CFG_SUMM} .= sprintf ("\n%s: XB_PIPE Nack Hammer is disabled in %s \n", $prefix, $mmr_name);
	      $any_cfg_error = 1;
  	    }
	    if ((reg_get_field_value($reg_ptr, 'ENABLE_FSB_NACK_HAMMER')) == 0) { 
    	      $sh2_target_list[$sh2_cnt]->{CFG_ERR} = 1;
    	      $sh2_target_list[$sh2_cnt]->{CFG_SUMM} .= sprintf ("\n%s: FSB Nack Hammer is disabled in %s \n", $prefix, $mmr_name);
	      $any_cfg_error = 1;
  	    }
	    if (((reg_get_field_value($reg_ptr, 'SYSTEM_FLUSH_DURATION')) == 0) && 
		((reg_get_field_value($reg_ptr, 'SYSTEM_FLUSH_INTERVAL')) == 0)) { 
    	      $sh2_target_list[$sh2_cnt]->{CFG_ERR} = 1;
    	      $sh2_target_list[$sh2_cnt]->{CFG_SUMM} .= sprintf ("\n%s: System Flush is disabled in %s \n", $prefix, $mmr_name);
	      $any_cfg_error = 1;
  	    }
	    if ( ((reg_get_field_value($reg_ptr, 'SYSTEM_FLUSH_DURATION')) > 0) && 
	        ((reg_get_field_value($reg_ptr, 'SYSTEM_FLUSH_DURATION')) == (reg_get_field_value($reg_ptr, 'SYSTEM_FLUSH_INTERVAL'))) ) { 
    	      $sh2_target_list[$sh2_cnt]->{CFG_ERR} = 1;
    	      $sh2_target_list[$sh2_cnt]->{CFG_SUMM} .= sprintf ("\n%s: FSB is parked due to System Flush settings in %s \n", $prefix, $mmr_name);
	      $any_cfg_error = 1;
	      $flush_cfg_err = 1;
  	    }
	  }

          elsif ($mmr_name =~ /SH2_SCRATCH1\b/ ) {
            $sh2_output .= lcl_reg_write_value($reg_ptr, 'valid', $prefix, 0, $notify);
            $sh2_target_list[$sh2_cnt]->{PROM_VER} = ($mmr_data >> 23) & 0xFF;
            $sh2_target_list[$sh2_cnt]->{PROM_REV} = ($mmr_data >> 31) & 0xFF;
            if (($sh2_target_list[$sh2_cnt]->{PROM_VER} != 0) || ($sh2_target_list[$sh2_cnt]->{PROM_REV} != 0)) {
              $valid_psp_state = 1;
	      if ($processor_count > 0)  {
	        $shadow_event_mask = 0x5a0;	# so we check for PI detected ECC errors found in shadow regs
	      }
              if ($global_prom_set == 0) {
		$global_prom_ver = $sh2_target_list[$sh2_cnt]->{PROM_VER};
		$global_prom_rev = $sh2_target_list[$sh2_cnt]->{PROM_REV};
		$global_prom_set = 1;
	      }
	      else {
		if ($sh2_target_list[$sh2_cnt]->{PROM_VER} != $global_prom_ver) {
	          $any_cfg_error = 1;
		  $sh2_target_list[$sh2_cnt]->{CFG_ERR} = 1;
		  $sh2_target_list[$sh2_cnt]->{CFG_SUMM} .= sprintf "\n%s:  Incorrect Prom Version found.  Actual: 0x%02X   Expected: 0x%02X\n",
		    $prefix, $sh2_target_list[$sh2_cnt]->{PROM_VER}, $global_prom_ver;
	        }
		if ($sh2_target_list[$sh2_cnt]->{PROM_REV} != $global_prom_rev) {
	          $any_cfg_error = 1;
		  $sh2_target_list[$sh2_cnt]->{CFG_ERR} = 1;
		  $sh2_target_list[$sh2_cnt]->{CFG_SUMM} .= sprintf "\n%s:  Incorrect Prom Revision found.  Actual: 0x%02X   Expected: 0x%02X\n",
		    $prefix, $sh2_target_list[$sh2_cnt]->{PROM_REV}, $global_prom_rev;
	        }
	      }
            }
	    if (($bw_955895_is_in == 0) 
             && ($sh2_target_list[$sh2_cnt]->{PROM_VER} > 0) 
             && ($sh2_target_list[$sh2_cnt]->{PROM_REV} > 0x1F)) {
	      $bw_955895_is_in = 1;
            }
	    if ((($mmr_data >> 40) & 1) == 0) {			#  check for ADVERT_BOOTED
	      $sh2_target_list[$sh2_cnt]->{CFG_ERR} = 1;
	      $sh2_target_list[$sh2_cnt]->{CFG_SUMM} .= sprintf ("\n%s: Bit 40 within %s indicates this node is NOT booted. \n", $prefix, $mmr_name);
	      $any_cfg_error = 1;
	      if ($nasid_value == 0) {
	        $node_not_booted = 1;
	      }
	    }
          }
  
	  elsif ( (($mmr_name =~ /SH2_FIRST_ERROR/) && ($mmr_data != 0))  
		  || (($mmr_name =~ /EVENT/) && (($mmr_data & 0xfffff) != 0)) )  {	    
	    $any_sh2_error = 1;
	    $notify = "  <<<<<<<";
	    $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
	    my $mmr_display = lcl_reg_write_value($reg_ptr, 'valid', $prefix, 0, $notify);
            $sh2_output .= $mmr_display;
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= $mmr_display;
	    if (($mmr_data & 0x71f)!= 0) {
	      $shub_hw_err = 1;
	      $global_err_val |= $SHUB_HW_ERR;
	    }
	  }
  
	  elsif (($mmr_name =~ /SH2_PI_FIRST_ERROR/) && ($mmr_data != 0))  {	    
	    $shub_hw_err = 1;
	    $any_sh2_error = 1;
	    $notify = "  <<<<<<<";
	    $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
	    my $mmr_display = lcl_reg_write_value($reg_ptr, 'valid', $prefix, 0, $notify);
            $sh2_output .= $mmr_display;
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= $mmr_display;
	    $pi_first_err_val = $mmr_data;
	    $pi_first_err_str = "HW Error";
	    $pi_dtl_vld = (($mmr_data & $pi_dtl_vld_mask) > 0) ? 1 : 0;
  
	    if ((reg_get_field_value($reg_ptr,'FSB_PROTO_ERR')) == 1) {
	      $pi_fsb_proto_err = 1;
	      $pi_fsb_proto_fru_entry = $fru_count;
	      $pi_first_err_str = "FSB Proto Error";
	    }
	    if ((reg_get_field_value($reg_ptr,'XB_PROTO_ERR')) == 1) {
	      $pi_first_err_str = "XB Pipe Protocol Error";
	      $pi_err_save_addr = 1;
	      $pi_xb_proto = 1;
	      $pi_xb_proto_fru_entry = $fru_count;
	    }
	    if ((reg_get_field_value($reg_ptr,'MEM_RP_ERR')) == 1) {
	      $pi_first_err_str = "Mem Reply Error";
	      $pi_err_save_addr = 1;
	    }
	    if ((reg_get_field_value($reg_ptr,'PIO_RP_ERR')) == 1) {
	      $pi_first_err_str = "PIO Reply Error";
	      $pi_pio_err = 1;
	    }
	    if ((reg_get_field_value($reg_ptr,'RRB_TO_ERR')) == 1) {
	      $pi_first_err_str = "RRB Timeout Error";
	      $pi_err_save_addr = 1;
	      $pi_to_err = 1;
	    }
	    if ((reg_get_field_value($reg_ptr,'WRB_TO_ERR')) == 1) {
	      $pi_first_err_str = "WRB Timeout Error";
	      $pi_err_save_addr = 1;
	      $pi_to_err = 1;
	    }
	    if ((reg_get_field_value($reg_ptr,'IRB_TO_ERR')) == 1) {
	      $pi_first_err_str = "IRB Timeout Error";
	      $pi_err_save_addr = 1;
	      $pi_to_err = 1;
	    }
	    if ((reg_get_field_value($reg_ptr,'PIO_TO_ERR')) == 1) {
	      $pi_first_err_str = "PIO Timeout Error";
	      $pi_pio_err = 1;
	      $pi_to_err = 1;
	    }
	    if ((reg_get_field_value($reg_ptr,'MSG_COLOR_ERR')) == 1) {
	      $pi_first_err_str = "Message Color Error";
	      $pi_err_save_addr = 1;
	    }
	    if ((reg_get_field_value($reg_ptr,'MSG_LENGTH')) == 1) {
	      $pi_first_err_str = "Message Length Error";
              $pi_first_err_str .= "\n\t\t\t\t\t\tSuspect Numa-Link packet corruption.";
              $pi_first_err_str .= "\n\t\t\t\t\t\tRun specified rtrdmp command to trace failing route.";
	      $pi_err_save_addr = 1;
      	      $pi_numa_error = 1;
	    }
	    if ((reg_get_field_value($reg_ptr,'NACK_OFLOW')) == 1) {
	      $pi_first_err_str = "Nack Overflow Error";
	      $pi_err_save_addr = 1;
	    }
	    if ((reg_get_field_value($reg_ptr,'FSB_SHUB_UCE')) == 1) {
	      $pi_first_err_str = "FSB to Shub Uncorrected Data Error";
	      $pi_fsb_uc_err = 1;
	      $pi_fsb_uce_fru_entry = $fru_count;
	      $first_pi_uce = 2;
	    }
	    if ((reg_get_field_value($reg_ptr,'FSB_SHUB_CE')) == 1) {
	      $pi_first_err_str = "FSB to Shub Corrected Data Error";
	      $pi_fsb_ce_err = 1;
	      $pi_fsb_ce_fru_entry = $fru_count;
	    }
	    if ((reg_get_field_value($reg_ptr,'SHUB_FSB_UCE')) == 1) {
	      $pi_first_err_str = "Shub to FSB Uncorrected Data Error";
	      $pi_shub_uc_err = 1;
	      $pi_uce_fru_entry = $fru_count;
	      $first_pi_uce = 1;
	    }
	    if ((reg_get_field_value($reg_ptr,'SHUB_FSB_CE')) == 1) {
	      $pi_first_err_str = "Shub to FSB Corrected Data Error";
	      $pi_shub_ce_err = 1;
	      $pi_ce_fru_entry = $fru_count;
	    }
	    if ((reg_get_field_value($reg_ptr,'IOQ_OVERRUN')) == 1) {
	      $pi_first_err_str = "FSB to Shub IOQ Overrun Error";
	      $pi_fsb_hw_err = 1;
	      $pi_fsb_hw_fru_entry = $fru_count;
	    }
	    if ((reg_get_field_value($reg_ptr,'HUNG_BUS')) == 1) {
	      $pi_first_err_str = "Hung Front-side Bus Error";
	    }
	    if ((reg_get_field_value($reg_ptr,'RSP_PARITY')) == 1) {
	      $pi_first_err_str = "FSB to Shub Response Parity Error";
	      $pi_fsb_hw_err = 1;
	      $pi_fsb_hw_fru_entry = $fru_count;
	    }
	    if ((reg_get_field_value($reg_ptr,'REQ_PARITY')) == 1) {
	      $pi_first_err_str = "FSB to Shub Request Parity Error";
	      $pi_fsb_hw_err = 1;
	      $pi_fsb_hw_fru_entry = $fru_count;
	    }
	    if ((reg_get_field_value($reg_ptr,'ADDR_PARITY')) == 1) {
	      $pi_first_err_str = "FSB to Shub Address Parity Error";
	      $pi_fsb_hw_err = 1;
	      $pi_fsb_hw_fru_entry = $fru_count;
	    }
	    if ((reg_get_field_value($reg_ptr,'ADDR_ACCESS')) == 1) {
	      $pi_first_err_str = "Address Access Error";
	      $pi_fsb_sw_err = 1;
	      $pi_fsb_sw_fru_entry = $fru_count;
	    }
	    if ((reg_get_field_value($reg_ptr,'REQ_FORMAT')) == 1) {
	      $pi_first_err_str = "Request Format Error";
	      $pi_fsb_sw_err = 1;
	      $pi_fsb_sw_fru_entry = $fru_count;
	    }
            my $fru_rsn = sprintf "Shub PI Reported %s.  ", $pi_first_err_str;
	    $sh2_output .= save_fru('SHUB2', $nasid_value, $current_location, $fru_rsn, 'PI_FIRST_ERR');
	  }
  
	  #  May also need to add 4 fsb_hw_err types into this block as well GEM
  
	  elsif ((($mmr_name =~ /SH2_PI_ERROR_SUMMARY/) || ($mmr_name =~ /SH2_PI_ERROR_OVERFLOW/)) && ($mmr_data != 0))  {
	    $any_sh2_error = 1;
	    $notify = "  <<<<<<<";
	    $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
	    my $mmr_display = lcl_reg_write_value($reg_ptr, 'valid', $prefix, 0, $notify);
            $sh2_output .= $mmr_display;
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= $mmr_display;
	    my $new_fru_err = 0;
	    my $err_sum_str = "";
	    if ($pi_dtl_vld == 0) {
	      $pi_dtl_vld = (($mmr_data & $pi_dtl_vld_mask) > 0) ? 1 : 0;
	    }
  
            if (((reg_get_field_value($reg_ptr,'FSB_SHUB_UCE')) == 1) && ($pi_fsb_uc_err == 0)) {
              $err_sum_str .= "FSB to Shub Uncorrected Data Error";
              $pi_fsb_uc_err = 1;
	      $new_fru_err = 1;
              $pi_fsb_uce_fru_entry = $fru_count;
            }
            if (((reg_get_field_value($reg_ptr,'FSB_SHUB_CE')) == 1) && ($pi_fsb_ce_err == 0)) {
              $err_sum_str .= "FSB to Shub Corrected Data Error";
              $pi_fsb_ce_err = 1;
	      $new_fru_err = 1;
              $pi_fsb_ce_fru_entry = $fru_count;
            }
            if (((reg_get_field_value($reg_ptr,'SHUB_FSB_UCE')) == 1) && ($pi_shub_uc_err == 0)) {
              $err_sum_str .= "Shub to FSB Uncorrected Data Error";
              $pi_shub_uc_err = 1;
	      $new_fru_err = 1;
              $pi_uce_fru_entry = $fru_count;
            }
            if (((reg_get_field_value($reg_ptr,'SHUB_FSB_CE')) == 1) && ($pi_shub_ce_err == 0)) {
              $err_sum_str .= "Shub to FSB Corrected Data Error";
              $pi_shub_ce_err = 1;
	      $new_fru_err = 1;
              $pi_ce_fru_entry = $fru_count;
            }
            if ((reg_get_field_value($reg_ptr,'RRB_TO_ERR')) == 1) {
              $pi_to_err_s = 1;
            }
            if ((reg_get_field_value($reg_ptr,'WRB_TO_ERR')) == 1) {
              $pi_to_err_s = 1;
            }
            if ((reg_get_field_value($reg_ptr,'IRB_TO_ERR')) == 1) {
              $pi_to_err_s = 1;
            }
            if ((reg_get_field_value($reg_ptr,'PIO_TO_ERR')) == 1) {
              $pi_to_err_s = 1;
            }

                # Handle the case where a REQ_FORMAT or ADDR_ACCESS error was found in
                # SH2_PI_FIRST_ERROR, but later occurring FSB HW errors were also logged.
                # Need to convert this error case from SW to HW.

            if ( ((reg_get_field_value($reg_ptr,'RSP_PARITY')) == 1)
              || ((reg_get_field_value($reg_ptr,'REQ_PARITY')) == 1)
              || ((reg_get_field_value($reg_ptr,'ADDR_PARITY')) == 1) ) {
              $pi_fsb_hw_err = 1;
              if ($pi_fsb_sw_err == 1) {
                $pi_fsb_hw_fru_entry = $pi_fsb_sw_fru_entry;
                $pi_fsb_sw_err = 0;
              }
            }
  
	    if ($new_fru_err == 1)  {
              my $fru_rsn = sprintf "Shub PI Reported %s.  ", $err_sum_str;
	      $sh2_output .= save_fru('SHUB2', $nasid_value, $current_location, $fru_rsn, 'PI_SUMMARY_ERR');
	    }
	  }
  
	  elsif (($mmr_name =~ /SH2_PI_ERROR_DETAIL/) && ($mmr_data != 0) && ($pi_dtl_vld == 1))  {
	    if ((($pi_first_err_val & 0xffc013fff) == 0) && ($mmr_name =~ /DETAIL_1/)) {
	      $pi_first_err_val = (($mmr_data & 1) == 1) ? 0x1 : 0x4000000;       # crb mode or fsb mode parsing
	    }
	    $global_err_val |= $SHUB_HW_ERR;
            $notify = "  <<<<<<<";
            my $mmr_display = lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
            $sh2_output .= $mmr_display;
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= $mmr_display;
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= decode_sh2_pi_err_detail($mmr_name, $mmr_data, $pi_first_err_val);
	    if (($mmr_name =~ /DETAIL_1/) && ($pi_err_save_addr == 1)) {
	      my ($net_addr, $net_nasid) = fsb_to_net($mmr_data & 0x3ffffffffff80); 
	      save_address($prefix, ($net_addr & 0x3fffffff80), $net_nasid, 'PI_DETAIL_1', $node_not_booted);
	    }
  
	    if (($mmr_name =~ /DETAIL_1/) && (($mmr_data & 1) == 1)) {
	      my ($net_addr, $net_nasid) = fsb_to_net($mmr_data & 0x3fffffffffff8); 
	      $pi_dtl1_node = $net_nasid;
	      $pi_dtl1_addr = $mmr_data & 0x3fffffffffff8;
	      if ((($mmr_data >> 56) & 0xff) == 0xfe) {
	        $pi_to_err = 1;
	      }
	      else {
	        $pi_crb_err = 1;
	      }
	      my $src_node_type = (($pi_err_save_addr == 1) && ($pi_numa_error == 0)) ? "Directory" : "Source";
	      my $lcl_mmr_space = (($pi_pio_err == 1) && (($net_addr & 0x3ffff00000000) == 0x200000000) ) ? 1 : 0;
	      if (($net_nasid != $nasid_value) && ($lcl_mmr_space == 0)) {
                my $fru_rsn  = sprintf "%s node indicated on NASID 0x%04X PI \(DETAIL1\).  ", $src_node_type, $nasid_value;
	        $sh2_output .= save_fru('SHUB2', $net_nasid, 'UNDEFINED', $fru_rsn, 'PI_REMOTE_ERR');
	      }
  
	      if ( (($net_addr & 0x1f000000f) == 0x170000000) && ($pi_pio_err == 1)) {
	         $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf "\n\tThe failing PIO address (0x%014llX) is a PTC access on NASID 0x%04X\n",
			  $net_addr, $net_nasid;
	        if ((($mmr_data >> 56) & 0xff) == 0x9f) {		# command type is PWERR
		  $ptc_pwerr = 1;
	        }
	      }
	      elsif (($net_nasid != $nasid_value) && ($lcl_mmr_space == 0)) {
	        $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf "\n\tThe failing address (0x%014llX) is on NASID 0x%04X\n",
			  $net_addr, $net_nasid;
	      }
	      elsif ($lcl_mmr_space == 1) {
	        $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf "\n\tThe failing address (0x%014llX) is in Local MMR Space on NASID 0x%04X\n",
			  $net_addr, $nasid_value;
	      }
	      if ($pi_numa_error == 1) {
      	        $pi_numa_err_src_node = $net_nasid;
      	        $pi_numa_err_addr     = $net_addr;
	      }
	    }
	       #  check if PI Ecode field = SH2_PI_CRB_ERR_CODE_PIO_RP
	    if (($mmr_name =~ /DETAIL_2/) && ($ptc_pwerr == 1) && ((($mmr_data >> 17) & 0x1f) == 3)) {
	      $ptc_pwerr++;			
	    }
	    if (($mmr_name =~ /DETAIL_2/) && ($pi_crb_err == 1) && ($pi_numa_error == 1)) {
	      my $dtl_chip = ($mmr_data >> 46) & 0x3;
	      $pi_dtl2_chip = ($dtl_chip == 0) ? "md" : ($dtl_chip == 2) ? "pi" : "lb";
	    }
	    if (($mmr_name =~ /DETAIL_2/) && ($pi_crb_err == 1) && ($pi_numa_error == 0)) {
	      $pi_dtl2_src_node = ($mmr_data >> 32) & 0x3fff;
	      if (($pi_dtl2_src_node != $nasid_value) && ($pi_dtl2_src_node != $pi_dtl1_node) && ($pi_to_err == 0)) {
                my $fru_rsn = sprintf "Source node indicated on NASID 0x%04X PI \(DETAIL2\).  ", $nasid_value;
	        $sh2_output .= save_fru('SHUB2', $pi_dtl2_src_node, 'UNDEFINED', $fru_rsn, 'PI_REMOTE_ERR');
	      }
	    }
	    if (($mmr_name =~ /DETAIL_2/) && ($pi_fsb_hw_err == 1)) {		# Agent ID will be in [24:21]
    	      my $proc_agent = (($sh2_target_list[$sh2_cnt]->{AID_DECODE} & 1) == 1) ? (($mmr_data >> 22) & 1) : (($mmr_data >> 23) & 1);
              my $proc_core =  (($sh2_target_list[$sh2_cnt]->{AID_DECODE} & 1) == 1) ? (($mmr_data >> 23) & 1) : 0 ;
    	      my $proc_thread = (($sh2_target_list[$sh2_cnt]->{AID_DECODE} & 1) == 1) ? (($mmr_data >> 24) & 1) : 0;
	      $bad_skt = ($processor_count == 2) ? $proc_agent : 0;
	      $fru_list[$pi_fsb_hw_fru_entry]->{REASON} .= sprintf "\n\t\t\t\t\t\tSuspect %s CPU in SKT %d, Shub or %s sn %s.", 
		$processor_type, $bad_skt, $blade_type, $blade_sn; 
	    }
	    if (($mmr_name =~ /DETAIL_2/) && ($pi_fsb_sw_err == 1)) {		# Agent ID will be in [24:21]
    	      my $proc_agent = (($sh2_target_list[$sh2_cnt]->{AID_DECODE} & 1) == 1) ? (($mmr_data >> 22) & 1) : (($mmr_data >> 23) & 1);
              my $proc_core =  (($sh2_target_list[$sh2_cnt]->{AID_DECODE} & 1) == 1) ? (($mmr_data >> 23) & 1) : 0 ;
    	      my $proc_thread = (($sh2_target_list[$sh2_cnt]->{AID_DECODE} & 1) == 1) ? (($mmr_data >> 24) & 1) : 0;
	      $bad_skt = ($processor_count == 2) ? $proc_agent : 0;
	      $fru_list[$pi_fsb_sw_fru_entry]->{REASON} .= sprintf "\n\t\t\t\t\t\tSuspect Software running on %s cpu in SKT %d, or the cpu is bad.", 
		$processor_type, $bad_skt, $blade_type, $blade_sn; 
	    }
	  }
  
	  elsif (($mmr_name =~ /SH2_PI_CORRECTED_DETAIL_1/) && ($pi_shub_ce_err == 1))  {
	    $global_err_val |= $SHUB_SBE_ERR;
            $notify = "  <<<<<<<";
            my $mmr_display = lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
            $sh2_output .= $mmr_display;
	    if ($mmr_data != 0)  {
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= $mmr_display;
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= decode_mem_err($mmr_name, $mmr_data, $prefix, $blade_type);
              $sh2_target_list[$sh2_cnt]->{DIMM_ERR} = 1;
              $any_dimm_error++;
	    }
	    else {
	      $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "\n Error detail in $mmr_name cleared - unable to get DIMM location.\n";
	      $pi_shub_ce_err = 0;
	    }
	  }
          elsif ((($mmr_name =~ /SH2_PI_CORRECTED_DETAIL_2/)||($mmr_name =~ /SH2_PI_CORRECTED_DETAIL_3/)) && ($pi_shub_ce_err == 1))  {
            $notify = "  <<<<<<<";
            my $mmr_display = lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
            $sh2_output .= $mmr_display;
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= $mmr_display;
 	    if ($mmr_name =~ /DETAIL_3/) {
    	      my $proc_agent = (($sh2_target_list[$sh2_cnt]->{AID_DECODE} & 1) == 1) ? (($mmr_data >> 6) & 1) : (($mmr_data >> 7) & 1);
              my $proc_core =  (($sh2_target_list[$sh2_cnt]->{AID_DECODE} & 1) == 1) ? (($mmr_data >> 7) & 1) : 0 ;
    	      my $proc_thread = (($sh2_target_list[$sh2_cnt]->{AID_DECODE} & 1) == 1) ? (($mmr_data >> 8) & 1) : 0;
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf "\n\t\t\t\tPROC:%d  CORE:%d  THREAD:%d\n\n",
                $proc_agent, $proc_core, $proc_thread;
	    }
  
	  }
  
          elsif (($mmr_name =~ /SH2_PI_UNCORRECTED_DETAIL_1/) && ($pi_shub_uc_err == 1))  {
            $notify = "  <<<<<<<";
            my $mmr_display = lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
            $sh2_output .= $mmr_display;
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= $mmr_display;
	    if ((reg_get_field_value($reg_ptr,'SYNDROME')) == 0xFF) {
	      my $byte_addr = (reg_get_field_value($reg_ptr,'ADDRESS')) << 3;
	      my ($net_addr, $net_nasid) = fsb_to_net($byte_addr & 0x3fffffffffff8); 
	      if ($net_nasid != $nasid_value) {
	        $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .=  sprintf "\n    Poisoned ECC received for Address: 0x%014llX on NASID 0x%04X\n\n", $byte_addr, $net_nasid;
                my $fru_rsn = sprintf "Source node indicated on NASID 0x%04X PI poisoned ECC error.  ", $nasid_value;
	        $sh2_output .= save_fru('SHUB2', $net_nasid, 'UNDEFINED', $fru_rsn, 'PI_SHUB_FSB_UCE_POIS');
	      }
	      else {
	        $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .=  sprintf "\n    Poisoned ECC received for Address: 0x%014llX\n\n", $byte_addr;
	      }
	    }
	    else {
	      if ($mmr_data != 0)  {
                $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= decode_mem_err($mmr_name, $mmr_data, $prefix, $blade_type);;
                $sh2_target_list[$sh2_cnt]->{DIMM_ERR} = 1;
                $any_dimm_error++;
	        $global_err_val |= $SHUB_UCE_ERR;
	      }
	      else {
	        $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "\n Error detail in $mmr_name cleared - unable to get DIMM location.\n";
	        $pi_shub_uc_err = 0;
	      }
	    }
          }
          elsif ((($mmr_name =~ /SH2_PI_UNCORRECTED_DETAIL_2/)||($mmr_name =~ /SH2_PI_UNCORRECTED_DETAIL_3/)) && ($pi_shub_uc_err == 1))  {
            $notify = "  <<<<<<<";
            my $mmr_display = lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
            $sh2_output .= $mmr_display;
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= $mmr_display;
 	    if ($mmr_name =~ /DETAIL_3/) {
    	      my $proc_agent = (($sh2_target_list[$sh2_cnt]->{AID_DECODE} & 1) == 1) ? (($mmr_data >> 6) & 1) : (($mmr_data >> 7) & 1);
              my $proc_core =  (($sh2_target_list[$sh2_cnt]->{AID_DECODE} & 1) == 1) ? (($mmr_data >> 7) & 1) : 0 ;
    	      my $proc_thread = (($sh2_target_list[$sh2_cnt]->{AID_DECODE} & 1) == 1) ? (($mmr_data >> 8) & 1) : 0;
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf "\n\t\t\t\tPROC:%d  CORE:%d  THREAD:%d\n\n",
                $proc_agent, $proc_core, $proc_thread;
	    }
          }
  
          elsif (($mmr_name =~ /SH2_PI_CORRECTED_DETAIL_4/) && ($pi_fsb_ce_err == 1))  {
	    $global_err_val |= $SHUB_SBE_ERR;
            $notify = "  <<<<<<<";
            my $mmr_display = lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
            $sh2_output .= $mmr_display;
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= $mmr_display;
	    if ($mmr_data != 0)  {
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= decode_proc_err($mmr_name, $mmr_data, $prefix);
	    }
	    else {
	      $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "\n Error detail in $mmr_name cleared - unable to get CPU location.\n";
	      $pi_fsb_ce_err = 0;
            }
          }
          elsif ((($mmr_name =~ /SH2_PI_CORRECTED_DETAIL_5/)||($mmr_name =~ /SH2_PI_CORRECTED_DETAIL_6/)) && ($pi_fsb_ce_err == 1))  {
            $notify = "  <<<<<<<";
            my $mmr_display = lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
            $sh2_output .= $mmr_display;
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= $mmr_display;
 	    if ($mmr_name =~ /DETAIL_6/) {
	      $bad_skt = ($processor_count == 2) ? (reg_get_field_value($reg_ptr,'CPU')) : 0;
	      $fru_list[$pi_fsb_ce_fru_entry]->{REASON} .= sprintf "\n\t\t\t\t\t\tSuspect %s cpu in SKT %d, Shub or %s sn %s .", 
		$processor_type, $bad_skt, $blade_type, $blade_sn; 
	    }
          }
  
          elsif (($mmr_name =~ /SH2_PI_UNCORRECTED_DETAIL_4/) && ($pi_fsb_uc_err == 1))  {
	    $global_err_val |= $SHUB_HW_ERR;
            $notify = "  <<<<<<<";
            my $mmr_display = lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
            $sh2_output .= $mmr_display;
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= $mmr_display;
	    if ($mmr_data != 0)  {
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= decode_proc_err($mmr_name, $mmr_data, $prefix);
	      $pi_fsb_uce_poisoned = ((reg_get_field_value($reg_ptr,'SYNDROME')) == 0x3) ? 1 : 0;
	    }
	    else {
	      $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "\n Error detail in $mmr_name cleared - unable to get CPU location.\n";
	      $pi_fsb_uc_err = 0;
	    }
          }
          elsif ((($mmr_name =~ /SH2_PI_UNCORRECTED_DETAIL_5/)||($mmr_name =~ /SH2_PI_UNCORRECTED_DETAIL_6/)) && ($pi_fsb_uc_err == 1))  {
            $notify = "  <<<<<<<";
            my $mmr_display = lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
            $sh2_output .= $mmr_display;
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= $mmr_display;
 	    if ($mmr_name =~ /DETAIL_6/) {
	      $bad_skt = ($processor_count == 2) ? (reg_get_field_value($reg_ptr,'CPU')) : 0;
	      my $cpu_string = ($first_pi_uce == 2) ? 'CPU in SKT' : 'cpu in SKT';
	      if ($pi_fsb_uce_poisoned == 1) {
	        $fru_list[$pi_fsb_uce_fru_entry]->{REASON} .= sprintf "\n\t\t\t\t\t\t%s %s %d wrote back previously poisoned data.", 
		  $processor_type, $cpu_string, $bad_skt, $blade_type, $blade_sn; 
	      }
	      else {
	        $fru_list[$pi_fsb_uce_fru_entry]->{REASON} .= sprintf "\n\t\t\t\t\t\tSuspect %s %s %d, Shub or %s  sn %s.", 
		  $processor_type, $cpu_string, $bad_skt, $blade_type, $blade_sn; 
	      }
	    }
          }
  
          elsif ((($mmr_name =~ /SH2_BT_FIRST_ERROR/)
                  || ($mmr_name =~ /SH2_BT_ERROR_SUMMARY/)
                  || ($mmr_name =~ /SH2_BT_ERROR_OVERFLOW/)) && ($mmr_data != 0))  {
	    $global_err_val |= $SHUB_HW_ERR;
            $any_sh2_error = 1;
	    $shub_hw_err = 1;
            $notify = "  <<<<<<<";
            $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
            my $mmr_display = lcl_reg_write_value($reg_ptr, 'valid', $prefix, 0, $notify);
            $sh2_output .= $mmr_display;
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= $mmr_display;
	    if ($bt_err == 0)  {
              $bt_err = 1;
              my $fru_rsn = sprintf "Shub BT Reported HW Error.  ";
	      $sh2_output .= save_fru('SHUB2', $nasid_value, $current_location, $fru_rsn, 'BT_HW_ERR');
	    }
	    if (($mmr_data & 0x9ff) > 0) {
	      $bt_xn_err = 1;
	    }
          }
          elsif (($mmr_name =~ /SH2_BT_XN_ERROR_DETAIL/) && ($bt_xn_err == 1)) {
            $notify = "  <<<<<<<";
            my $mmr_display = lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
            $sh2_output .= $mmr_display;
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= $mmr_display;
	    if (($mmr_name =~ /DETAIL_1/) && ((reg_get_field_value($reg_ptr,'VALID')) == 1)) {
	      my $bt_cmd = reg_get_field_value($reg_ptr,'COMMAND');
	      $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf "\n  Failing command: %s\n\n", $sn2net_command_list[$bt_cmd];
	      my $src_node = reg_get_field_value($reg_ptr,'SOURCE_NODE_ID');
	      if ($src_node != $nasid_value) {
                my $fru_rsn = sprintf "Source node indicated on NASID 0x%04X BT.  ", $nasid_value;
		$sh2_output .= save_fru('SHUB2', $src_node, 'UNDEFINED', $fru_rsn, 'BT_REMOTE_ERR');
	      }
	    }
	    if ($mmr_name =~ /DETAIL_2/) {
	      my $net_addr = $mmr_data & 0x3fffffff80;
	      my $net_nasid = ($mmr_data >> 38) & 0x3fff;
	      save_address($prefix, $net_addr, $net_nasid, 'BT_XN', $node_not_booted);
	    }
	  }
  
          elsif (($mmr_name =~ /SH2_BT_ENG_CSR/) && ((reg_get_field_value($reg_ptr,'ERROR')) == 1)) {
            $notify = "  <<<<<<<";
            my $mmr_display = lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
            $sh2_output .= $mmr_display;
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= $mmr_display;
	  }
          elsif (($mmr_name =~ /SH2_BT_ENG_ERROR_STAT/) && 
		   (  ((reg_get_field_value($reg_ptr,'ERROR0')) == 1)
		   || ((reg_get_field_value($reg_ptr,'ERROR1')) == 1)
		   || ((reg_get_field_value($reg_ptr,'ERROR2')) == 1)
		   || ((reg_get_field_value($reg_ptr,'ERROR3')) == 1)) ) {
            $notify = "  <<<<<<<";
            my $mmr_display = lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
            $sh2_output .= $mmr_display;
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= $mmr_display;
	  }
          elsif ( (($mmr_name =~ /SH2_BT_CRB_STAT/) && ((reg_get_field_value($reg_ptr,'CRB_VALID')) == 1)) &&
                   (  ((reg_get_field_value($reg_ptr,'CRB_ACCESS_ERROR')) == 1)
                   || ((reg_get_field_value($reg_ptr,'CRB_ERROR')) == 1)
                   || ((reg_get_field_value($reg_ptr,'CRB_TIMEOUT')) == 1)) ) {
            $notify = "  <<<<<<<";
            my $mmr_display = lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
            $sh2_output .= $mmr_display;
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= $mmr_display;
	    $bt_crb_err = ((reg_get_field_value($reg_ptr,'CRB_TYPE')) > 0x1b) ? 2 : 1;  #  1=BT Pull     2=BT Push
          }
          elsif ((($mmr_name =~ /SH2_BT_CRB_SRC/) && ($bt_crb_err == 1)) 
	      || (($mmr_name =~ /SH2_BT_CRB_DEST/) && ($bt_crb_err == 2)))  {
            $notify = "  <<<<<<<";
            my $mmr_display = lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
            $sh2_output .= $mmr_display;
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= $mmr_display;
	    my $net_addr = $mmr_data & 0x3fffffff80;
	    my $net_nasid = ($mmr_data >> 40) & 0x3fff;
	    save_address($prefix, $net_addr, $net_nasid, 'BT_CRB', $node_not_booted);
	    if ($net_nasid != $nasid_value) {
              my $fru_rsn = sprintf "Source node indicated on NASID 0x%04X BT CRB.  ", $nasid_value;
	      $sh2_output .= save_fru('SHUB2', $net_nasid, 'UNDEFINED', $fru_rsn, 'BT_CRB_ERR');
	    }
	    $bt_crb_err = 0;
	  }
  
          elsif ((($mmr_name =~ /SH2_LB_FIRST_ERROR/)
                  || ($mmr_name =~ /SH2_LB_ERROR_SUMMARY/)
                  || ($mmr_name =~ /SH2_LB_ERROR_OVERFLOW/)) && ($mmr_data != 0))  {
	    $global_err_val |= $SHUB_HW_ERR;
	    $shub_hw_err = 1;
            $any_sh2_error = 1;
            $notify = "  <<<<<<<";
            $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
            my $mmr_display = lcl_reg_write_value($reg_ptr, 'valid', $prefix, 0, $notify);
            $sh2_output .= $mmr_display;
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= $mmr_display;
	    if ($lb_err == 0)  {
              $lb_err = 1;
              my $fru_rsn = "Shub LB Reported HW Error.  ";
	      if ((reg_get_field_value($reg_ptr, 'JUNK_BUS_ERR')) == 1) {
		$bw_963416_attr = 1;
                $fru_rsn = "Shub LB Reported Junk Bus Error.  ";
	      }
	      $sh2_output .= save_fru('SHUB2', $nasid_value, $current_location, $fru_rsn, 'LB_HW_ERR');
	    }
          }
  
          elsif (($mmr_name =~ /SH2_LB_ERROR_DETAIL/) && ($lb_err == 1) && ($mmr_data != 0))  {
            $notify = "  <<<<<<<";
	    my $parse_mode = ($mmr_name =~ /5/) ? 'valid' : 'all';
            my $mmr_display = lcl_reg_write_value($reg_ptr, $parse_mode, $prefix, 0, $notify);
            $sh2_output .= $mmr_display;
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= $mmr_display;
	    if (($mmr_name =~ /DETAIL_1/) && ((reg_get_field_value($reg_ptr,'VALID')) == 1)) {
	      my $lb_cmd = reg_get_field_value($reg_ptr,'COMMAND');
	      $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf "\n  Failing command: %s\n\n", $sn2net_command_list[$lb_cmd];
	      my $src_node = reg_get_field_value($reg_ptr,'SOURCE_NODE_ID');
	      if ($src_node != $nasid_value) {
                my $fru_rsn = sprintf "Source node indicated on NASID 0x%04X LB.  ", $nasid_value;
	        $sh2_output .= save_fru('SHUB2', $src_node, 'UNDEFINED', $fru_rsn, 'LB_REMOTE_ERR');
	      }
	    }
	    if ($mmr_name =~ /DETAIL_2/) {
	      my $byte_addr = (reg_get_field_value($reg_ptr,'ADDRESS')) << 3;
	      $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf "\n    Actual Address: 0x%014llX\n\n", $byte_addr; 
	    }
	    if ($mmr_name =~ /DETAIL_6/) {
	      my $byte_addr = (reg_get_field_value($reg_ptr,'ADDRESS')) << 3;
	      my $lb_cmd = reg_get_field_value($reg_ptr,'CMD');
	      if ($lb_cmd == 0x19) {
		$bw_963416_attr++;
	      }

	      my $lb_mode = ((reg_get_field_value($reg_ptr,'MODE')) == 1) ? "NACK" : "NORMAL";
	      if ($lb_mode eq 'NORMAL') {
		$bw_963416_attr++;
	      }

	      my $lb_stat = decode_lb_stat(reg_get_field_value($reg_ptr,'STATUS'));
	      if ($lb_stat =~ /Sequence/) {
		$bw_963416_attr++;
	      }

	      $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf "\n    Actual Address: 0x%014llX    Command: %s   Mode: %s \n  Error Status:   %s\n\n",
		  $byte_addr, $sn2net_command_list[$lb_cmd], $lb_mode, $lb_stat;
	    }
 	  }
  
          elsif ((($mmr_name =~ /SH2_XN_FIRST_ERROR/)
                  || ($mmr_name =~ /SH2_XN_ERROR_SUMMARY/)
                  || ($mmr_name =~ /SH2_XN_ERROR_OVERFLOW/)) && ($mmr_data != 0))  {
            $any_sh2_error = 1;
	    $shub_hw_err = 1;
            $notify = "  <<<<<<<";
            $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
            my $mmr_display = lcl_reg_write_value($reg_ptr, 'valid', $prefix, 0, $notify);
            $sh2_output .= $mmr_display;
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= $mmr_display;
	    if (((reg_get_field_value($reg_ptr,'XNMD_SBE_ERROR')) == 1) && ($xn_md_ce == 0)) {
	      $global_err_val |= $SHUB_SBE_ERR;
              $sh2_target_list[$sh2_cnt]->{DIMM_ERR} = 1;
              $any_dimm_error++;
	      $xn_md_ce = 1;
	      $xn_ce_fru_entry = $fru_count;
              my $fru_rsn = "XN Reported SBE ECC Error.  ";
	      $sh2_output .= save_fru('SHUB2', $nasid_value, $current_location, $fru_rsn, 'XN_CE');
	    }
	    if (((reg_get_field_value($reg_ptr,'XNMD_UCE_ERROR')) == 1) && ($xn_md_uc == 0)) {
	      $global_err_val |= $SHUB_UCE_ERR;
              $sh2_target_list[$sh2_cnt]->{DIMM_ERR} = 1;
              $any_dimm_error++;
	      $xn_md_uc = 1;
	      $xn_uce_fru_entry = $fru_count;
              my $fru_rsn = "Shub XN Reported UCE ECC Error.  ";
	      $sh2_output .= save_fru('SHUB2', $nasid_value, $current_location, $fru_rsn, 'XN_UCE');
	    }
	    if (($xn_err == 0) && ($xn_md_ce == 0) && ($xn_md_uc == 0))  {
	      $global_err_val |= $SHUB_HW_ERR;
              $xn_err = 1;
              my $fru_rsn = "Shub XN Reported HW Error.  ";
	      $sh2_output .= save_fru('SHUB2', $nasid_value, $current_location, $fru_rsn, 'XN_HW');
	    }
          }
          elsif (($mmr_name =~ /SH2_XN_CORRECTED_DETAIL/) && ($xn_md_ce == 1)) {
            $notify = "  <<<<<<<";
	    my $disp_mode = ($mmr_name =~ /DETAIL_1/) ? "all" : "valid";
            my $mmr_display = lcl_reg_write_value($reg_ptr, $disp_mode, $prefix, 0, $notify);
            $sh2_output .= $mmr_display;
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= $mmr_display;
	    if ($mmr_name =~ /DETAIL_[134]/)  {
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .=  decode_mem_err($mmr_name, $mmr_data, $prefix, $blade_type);
	    }
	  }
          elsif (($mmr_name =~ /SH2_XN_UNCORRECTED_DETAIL/) && ($xn_md_uc == 1)) {
            $notify = "  <<<<<<<";
	    my $disp_mode = ($mmr_name =~ /DETAIL_1/) ? "all" : "valid";
            my $mmr_display = lcl_reg_write_value($reg_ptr, $disp_mode, $prefix, 0, $notify);
            $sh2_output .= $mmr_display;
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= $mmr_display;
	    if ($mmr_name =~ /DETAIL_[134]/)  {
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .=  decode_mem_err($mmr_name, $mmr_data, $prefix, $blade_type);
	    }
	  }
  
          elsif ((($mmr_name =~ /SH2_NI0_FIRST_ERROR/)
                  || ($mmr_name =~ /SH2_NI0_ERROR_SUMMARY/)
                  || ($mmr_name =~ /SH2_NI0_ERROR_OVERFLOW/)) && ($mmr_data != 0))  {
	    $global_err_val |= $SHUB_HW_ERR;
            $any_sh2_error = 1;
	    $shub_hw_err = 1;
            $notify = "  <<<<<<<";
            $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
            my $mmr_display = lcl_reg_write_value($reg_ptr, 'valid', $prefix, 0, $notify);
            $sh2_output .= $mmr_display;
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= $mmr_display;
	    if ($xn_ni0_err == 0)  {
              my $add_rtr_fru = 0;
	      my $rtr_loc = "";
	      my $rtr_err_type = "";
              $xn_ni0_err = 1;
	      my $fru_rsn = "";
	      if ((reg_get_field_value($reg_ptr,'RETRY_TIMEOUT_ERROR')) == 1) {
		my $bp_connection = get_bp_connect($sh2_target_list[$sh2_cnt]->{NODE}, $sh2_target_list[$sh2_cnt]->{CHASSIS}, 0);
                $fru_rsn = "Shub XN_NI0 LLP Retry Timeout Error.\n\t\t\t\t\t\t$bp_connection  ";
		$bp_connection =~ /(RTR_[LR]:[01])/;
		$rtr_loc = $1; 
		$add_rtr_fru = 1;
		$rtr_err_type = "Retry Timeout";
	      }
	      elsif ((reg_get_field_value($reg_ptr,'CHIPLET_NOMATCH')) == 1) {
                $fru_rsn = "Shub XN_NI0 detected Chiplet NoMatch Error.  ";
	      }
	      elsif (($mmr_name =~ /_1$/) && (($mmr_data & 0x3fffc3c0f00000ff) > 0)) {
		my $bp_connection = get_bp_connect($sh2_target_list[$sh2_cnt]->{NODE}, $sh2_target_list[$sh2_cnt]->{CHASSIS}, 0);
                $fru_rsn = "Shub XN_NI0 detected Packet Flow Error or Packet Corruption Error.\n\t\t\t\t\t\t$bp_connection ";
		$bp_connection =~ /(RTR_[LR]:[01])/;
		$rtr_loc = $1; 
		$add_rtr_fru = 1;
		$rtr_err_type = "Packet Flow";
	      }
	      elsif (($mmr_name =~ /_2$/) && (($mmr_data & 0xff) > 0)) {
		my $bp_connection = get_bp_connect($sh2_target_list[$sh2_cnt]->{NODE}, $sh2_target_list[$sh2_cnt]->{CHASSIS}, 0);
                $fru_rsn = "Shub XN_NI0 detected Packet Flow Error or Packet Corruption Error.\n\t\t\t\t\t\t$bp_connection ";
		$bp_connection =~ /(RTR_[LR]:[01])/;
		$rtr_loc = $1; 
		$add_rtr_fru = 1;
		$rtr_err_type = "Packet Flow";
	      }
	      else {
                $fru_rsn = "Shub XN_NI0 detected Internal Packet Flow Error or Packet Corruption Error.  ";
	      }
	      $sh2_output .= save_fru('SHUB2', $nasid_value, $current_location, $fru_rsn, 'XN_NI0_LINK_ERR');

	      if ($add_rtr_fru == 1) {
                my $fru_rtr_loc = "$sh2_target_list[$sh2_cnt]->{MODULE}, $rtr_loc";
                $fru_rsn = "Source router of $current_location Shub NI0 detected LLP $rtr_err_type Error";
	        $sh2_output .= save_fru('NL4R', 0xffffff, $fru_rtr_loc, $fru_rsn, 'XN_NI0_RTR_ERR');
		$global_err_val |= $NL4R_HW_ERR;
	      }
	    }
	    if ((reg_get_field_value($reg_ptr,'CHIPLET_NOMATCH')) == 1) {
	      $xn_ni0_dump_hdr = 1;
	    }
          }
          elsif (($mmr_name =~ /SH2_NI0_ERROR_DETAIL/) && ($xn_ni0_err == 1)) {
            $notify = "  <<<<<<<";
            $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
            my $mmr_display = lcl_reg_write_value($reg_ptr, 'valid', $prefix, 0, $notify);
            $sh2_output .= $mmr_display;
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= $mmr_display;
            if (($mmr_name =~ /DETAIL_1/) && ($xn_ni0_dump_hdr == 1)) {
              $sh2_target_list[$sh2_cnt]->{NI0_HDR0} = $mmr_data;
            }
            if (($mmr_name =~ /DETAIL_2/) && ($xn_ni0_dump_hdr == 1)) {
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= decode_shub2_sn2net_header($sh2_target_list[$sh2_cnt]->{NI0_HDR0}, $mmr_data);
              my $detail_nasid = (($mmr_data >> 16) & 0x7FF) | ((($mmr_data >> 29) & 0x7) << 11);
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf("\tThe above error detail indicates packet originated on NASID 0x%04X.\n", $detail_nasid);
	      if ($node_not_booted == 1)  {
                my $dst_nasid = (($sh2_target_list[$sh2_cnt]->{NI0_HDR0} >> 13) & 0x7FF)
			     | ((($sh2_target_list[$sh2_cnt]->{NI0_HDR0} >> 34) & 0x1) << 11)
			     | ((($sh2_target_list[$sh2_cnt]->{NI0_HDR0} >> 62) & 0x3) << 12);
		$sh2_target_list[$sh2_cnt]->{SPEC_NASID} = $dst_nasid;
		printf "\n%s:  Assigning the speculative nasid value on this node to 0x%04X.\n", $prefix, $dst_nasid;
	      }
              elsif ($detail_nasid != $nasid_value)  {
                my $fru_rsn = sprintf "Source node indicated in SHUB2 NASID 0x%04X XN_NI0 Chiplet NoMatch Header.  ", $nasid_value;
		$sh2_output .= save_fru('SHUB2', $detail_nasid, 'UNDEFINED', $fru_rsn, 'XN_NI0_CHIPLET_NOMATCH');
              }
            }
	  }
  
          elsif ((($mmr_name =~ /SH2_NI1_FIRST_ERROR/)
                  || ($mmr_name =~ /SH2_NI1_ERROR_SUMMARY/)
                  || ($mmr_name =~ /SH2_NI1_ERROR_OVERFLOW/)) && ($mmr_data != 0))  {
	    $global_err_val |= $SHUB_HW_ERR;
            $any_sh2_error = 1;
	    $shub_hw_err = 1;
            $notify = "  <<<<<<<";
            $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
            my $mmr_display = lcl_reg_write_value($reg_ptr, 'valid', $prefix, 0, $notify);
            $sh2_output .= $mmr_display;
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= $mmr_display;
	    if ($xn_ni1_err == 0)  {
              my $add_rtr_fru = 0;
	      my $rtr_loc = "";
	      my $rtr_err_type = "";
              $xn_ni1_err = 1;
	      my $fru_rsn = "";
	      if ((reg_get_field_value($reg_ptr,'RETRY_TIMEOUT_ERROR')) == 1) {
		my $bp_connection = get_bp_connect($sh2_target_list[$sh2_cnt]->{NODE}, $sh2_target_list[$sh2_cnt]->{CHASSIS}, 1);
                $fru_rsn = "Shub XN_NI1 LLP Retry Timeout Error.\n\t\t\t\t\t\t$bp_connection  ";
		$bp_connection =~ /(RTR_[LR]:[01])/;
		$rtr_loc = $1; 
		$add_rtr_fru = 1;
		$rtr_err_type = "Retry Timeout";
	      }
              elsif ((reg_get_field_value($reg_ptr,'CHIPLET_NOMATCH')) == 1) {
                $fru_rsn = "Shub XN_NI1 detected Chiplet NoMatch Error.  ";
              }
              elsif (($mmr_name =~ /_1$/) && (($mmr_data & 0x3fffc3c0f00000ff) > 0)) {
                my $bp_connection = get_bp_connect($sh2_target_list[$sh2_cnt]->{NODE}, $sh2_target_list[$sh2_cnt]->{CHASSIS}, 1);
                $fru_rsn = "Shub XN_NI1 detected Packet Flow Error or Packet Corruption Error.\n\t\t\t\t\t\t$bp_connection ";
		$bp_connection =~ /(RTR_[LR]:[01])/;
		$rtr_loc = $1; 
		$add_rtr_fru = 1;
		$rtr_err_type = "Packet Flow";
              }
              elsif (($mmr_name =~ /_2$/) && (($mmr_data & 0xff) > 0)) {
                my $bp_connection = get_bp_connect($sh2_target_list[$sh2_cnt]->{NODE}, $sh2_target_list[$sh2_cnt]->{CHASSIS}, 1);
                $fru_rsn = "Shub XN_NI1 detected Packet Flow Error or Packet Corruption Error.\n\t\t\t\t\t\t$bp_connection ";
		$bp_connection =~ /(RTR_[LR]:[01])/;
		$rtr_loc = $1; 
		$add_rtr_fru = 1;
		$rtr_err_type = "Packet Flow";
              }
              else {
                $fru_rsn = "Shub XN_NI1 detected Internal Packet Flow Error or Packet Corruption Error.  ";
              }
	      $sh2_output .= save_fru('SHUB2', $nasid_value, $current_location, $fru_rsn, 'XN_NI1_LINK_ERR');
	      if ($add_rtr_fru == 1) {
                my $fru_rtr_loc = "$sh2_target_list[$sh2_cnt]->{MODULE}, $rtr_loc";
                $fru_rsn = "Source router of $current_location Shub NI1 detected LLP $rtr_err_type Error";
	        $sh2_output .= save_fru('NL4R', 0xffffff, $fru_rtr_loc, $fru_rsn, 'XN_NI1_RTR_ERR');
		$global_err_val |= $NL4R_HW_ERR;
	      }
	    }
            if ((reg_get_field_value($reg_ptr,'CHIPLET_NOMATCH')) == 1) {
              $xn_ni1_dump_hdr = 1;
            }
          }
          elsif (($mmr_name =~ /SH2_NI1_ERROR_DETAIL/) && ($xn_ni1_err == 1)) {
            $notify = "  <<<<<<<";
            $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
            my $mmr_display = lcl_reg_write_value($reg_ptr, 'valid', $prefix, 0, $notify);
            $sh2_output .= $mmr_display;
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= $mmr_display;
            if (($mmr_name =~ /DETAIL_1/) && ($xn_ni1_dump_hdr == 1)) {
              $sh2_target_list[$sh2_cnt]->{NI1_HDR0} = $mmr_data;
            }
            if (($mmr_name =~ /DETAIL_2/) && ($xn_ni1_dump_hdr == 1)) {
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= decode_shub2_sn2net_header($sh2_target_list[$sh2_cnt]->{NI1_HDR0}, $mmr_data);
              my $detail_nasid = (($mmr_data >> 16) & 0x7FF) | ((($mmr_data >> 29) & 0x7) << 11);
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf("\tThe above error detail indicates packet originated on NASID 0x%04X.\n", $detail_nasid);
	      if ($node_not_booted == 1)  {
                my $dst_nasid = (($sh2_target_list[$sh2_cnt]->{NI1_HDR0} >> 13) & 0x7FF)
		             | ((($sh2_target_list[$sh2_cnt]->{NI1_HDR0} >> 34) & 0x1) << 11)
		             | ((($sh2_target_list[$sh2_cnt]->{NI1_HDR0} >> 62) & 0x3) << 12);
	  	if (! defined $sh2_target_list[$sh2_cnt]->{SPEC_NASID}) {
		  $sh2_target_list[$sh2_cnt]->{SPEC_NASID} = $dst_nasid;
		  printf "\n%s:  Assigning the speculative nasid value on this node to 0x%04X.\n", $prefix, $dst_nasid;
		}
		elsif ($sh2_target_list[$sh2_cnt]->{SPEC_NASID} == $dst_nasid) {
		  $sh2_target_list[$sh2_cnt]->{NASID} = $dst_nasid;
		  $nasid_value = $dst_nasid;
		  $prefix = sprintf ("0x%04X", $nasid_value);
		  printf "\n%s:  Re-assigning the nasid value on this node to 0x%04X.\n", $prefix, $nasid_value;
		}
	      }
              elsif ($detail_nasid != $nasid_value)  {
                my $fru_rsn = sprintf "Source node indicated in SHUB2 NASID 0x%04X XN_NI1 Chiplet NoMatch Header.  ", $nasid_value;
		$sh2_output .= save_fru('SHUB2', $detail_nasid, 'UNDEFINED', $fru_rsn, 'XN_NI1_CHIPLET_NOMATCH');
              }
            }
          }
  
          elsif ((($mmr_name =~ /SH2_XNIILB_FIRST_ERROR/)
                  || ($mmr_name =~ /SH2_XNIILB_ERROR_SUMMARY/)
                  || ($mmr_name =~ /SH2_XNIILB_ERROR_OVERFLOW/)) && ($mmr_data != 0))  {
	    $global_err_val |= $SHUB_HW_ERR;
            $any_sh2_error = 1;
	    $shub_hw_err = 1;
            $notify = "  <<<<<<<";
            $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
            my $mmr_display = lcl_reg_write_value($reg_ptr, 'valid', $prefix, 0, $notify);
            $sh2_output .= $mmr_display;
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= $mmr_display;
	    if ($xn_iilb_err == 0)  {
              $xn_iilb_err = 1;
	      my $xn_iilb_reason = "Shub XN_IILB detected Internal Packet Flow Error or Packet Corruption Error.  ";
              if ((reg_get_field_value($reg_ptr,'LUT_READ_ERROR')) == 1) {
                $xn_iilb_reason = "Shub XN_IILB detected Look-up Table Read Error.  ";
		$xn_iilb_lre = 1;
	      }
	      if ((reg_get_field_value($reg_ptr,'CHIPLET_NOMATCH')) == 1) {
                $xn_iilb_reason = "Shub XN_IILB detected Chiplet NoMatch Error.  ";
	      }
	      $sh2_output .= save_fru('SHUB2', $nasid_value, $current_location, $xn_iilb_reason, 'XNIILB_ERR');
	    }
  
	    if ((reg_get_field_value($reg_ptr,'CHIPLET_NOMATCH')) == 1) {
	      $xn_iilb_dump_hdr = 1;
	    }
          }
          elsif (($mmr_name =~ /SH2_XNIILB_ERROR_DETAIL/) && ($xn_iilb_err == 1)) {
            $notify = "  <<<<<<<";
            $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
            my $mmr_display = lcl_reg_write_value($reg_ptr, 'valid', $prefix, 0, $notify);
            $sh2_output .= $mmr_display;
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= $mmr_display;
            if (($mmr_name =~ /DETAIL_1/) && ($xn_iilb_dump_hdr == 1)) {
              $sh2_target_list[$sh2_cnt]->{IILB_HDR0} = $mmr_data;
	      if ((($mmr_data >> 24) & 0xff) == 0x8c) {
	        $xn_cmd_is_prge = 1;
	      }
            }
            if (($mmr_name =~ /DETAIL_2/) && ($xn_iilb_dump_hdr == 1)) {
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= decode_shub2_sn2net_header($sh2_target_list[$sh2_cnt]->{IILB_HDR0}, $mmr_data);
              my $detail_nasid = (($mmr_data >> 16) & 0x7FF) | ((($mmr_data >> 29) & 0x7) << 11);
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf("\tThe above error detail indicates packet originated on NASID 0x%04X.\n", $detail_nasid);
              if ($detail_nasid != $nasid_value)  {
                my $fru_rsn = sprintf("Source node indicated in SHUB2 NASID 0x%04X XN_IILB Chiplet NoMatch Header.  ", $nasid_value);
		$sh2_output .= save_fru('SHUB2', $detail_nasid, 'UNDEFINED', $fru_rsn, 'XN_IILB_HDR_ERR');
              }
	      else {
	        $xn_cmd_is_prge++;
	      }
            }
	    if (($mmr_name =~ /DETAIL_3/) && ($xn_iilb_lre == 1)) {
	      my $detail_info = decode_lut_addr($mmr_data);
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf("\n\tThe above error detail indicates packet was destined for %s .\n", $detail_info);
	    }
	  }
  
          elsif ((($mmr_name =~ /SH2_XNPI_FIRST_ERROR/)
                  || ($mmr_name =~ /SH2_XNPI_ERROR_SUMMARY/)
                  || ($mmr_name =~ /SH2_XNPI_ERROR_OVERFLOW/)) && ($mmr_data != 0))  {
	    $global_err_val |= $SHUB_HW_ERR;
            $any_sh2_error = 1;
	    $shub_hw_err = 1;
            $notify = "  <<<<<<<";
            $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
            my $mmr_display = lcl_reg_write_value($reg_ptr, 'valid', $prefix, 0, $notify);
            $sh2_output .= $mmr_display;
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= $mmr_display;
	    if ($xn_pi_err == 0)  {
              $xn_pi_err = 1;
              my $xnpi_reason = "Shub XN_PI detected Internal Packet Flow Error or Packet Corruption Error.  ";
              if ((reg_get_field_value($reg_ptr,'LUT_READ_ERROR')) == 1) {
                $xnpi_reason = "Shub XN_PI detected Look-up Table Read Error.  ";
		$xn_pi_lre = 1;
              }
	      $sh2_output .= save_fru('SHUB2', $nasid_value, $current_location, $xnpi_reason, 'XNPI_ERR');
	    }
          }
  
          elsif (($mmr_name =~ /SH2_XNPI_ERROR_DETAIL/) && ($xn_pi_err == 1)) {
            $notify = "  <<<<<<<";
            $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
            my $mmr_display = lcl_reg_write_value($reg_ptr, 'valid', $prefix, 0, $notify);
            $sh2_output .= $mmr_display;
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= $mmr_display;
	    if (($mmr_name =~ /DETAIL_3/) && ($xn_pi_lre == 1)) {
	      my $detail_info = decode_lut_addr($mmr_data);
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf("\n\tThe above error detail indicates packet was destined for %s .\n", $detail_info);
	    }
	  }
  
          elsif ((($mmr_name =~ /SH2_XNMD_FIRST_ERROR/)
                  || ($mmr_name =~ /SH2_XNMD_ERROR_SUMMARY/)
                  || ($mmr_name =~ /SH2_XNMD_ERROR_OVERFLOW/)) && ($mmr_data != 0))  {
            $any_sh2_error = 1;
	    $shub_hw_err = 1;
            $notify = "  <<<<<<<";
            $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
            my $mmr_display = lcl_reg_write_value($reg_ptr, 'valid', $prefix, 0, $notify);
            $sh2_output .= $mmr_display;
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= $mmr_display;
	    if (($xn_md_err == 0) && ($xn_md_ce == 0) && ($xn_md_uc == 0))  {
	      $global_err_val |= $SHUB_HW_ERR;
              $xn_md_err = 1;
	      my $xnmd_reason = "Shub XN_MD Reported Hardware Error.  ";
	      if ( ($mmr_name =~ /SH2_XNMD_FIRST_ERROR\b/) 
		|| ($mmr_name =~ /SH2_XNMD_ERROR_SUMMARY\b/) 
		|| ($mmr_name =~ /SH2_XNMD_ERROR_OVERFLOW\b/)) {
		if ((reg_get_field_value($reg_ptr,'LUT_READ_ERROR')) == 1) {
		  $xnmd_reason = "Shub XN_MD detected Look-up Table Read Error.  ";
		  $xn_md_lre = 1;
		}
		else {
		  $xnmd_reason = "Shub XN_MD detected Internal Packet Flow Error or Packet Corruption Error.  ";
		}
	      }
	      $sh2_output .= save_fru('SHUB2', $nasid_value, $current_location, $xnmd_reason, 'XNMD_ERR');
	    }
          }
  
          elsif (($mmr_name =~ /SH2_XNMD_ERROR_DETAIL/) && ($xn_md_err == 1)) {
            $notify = "  <<<<<<<";
            $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
            my $mmr_display = lcl_reg_write_value($reg_ptr, 'valid', $prefix, 0, $notify);
            $sh2_output .= $mmr_display;
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= $mmr_display;
	    if (($mmr_name =~ /DETAIL_3/) && ($xn_md_lre == 1)) {
	      my $detail_info = decode_lut_addr($mmr_data);
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf("\n\tThe above error detail indicates packet was destined for %s .\n", $detail_info);
	    }
	  }
  
	  elsif ((($mmr_name =~ /SH2_MEM_FIRST_ERROR/) 
		  || ($mmr_name =~ /SH2_MEM_ERROR_OVERFLOW/)) && (($mmr_data & $md_err_fo_mask) != 0))  {
            $any_sh2_error = 1;
	    $shub_hw_err = 1;
            $notify = "  <<<<<<<";
            $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
            my $mmr_display = lcl_reg_write_value($reg_ptr, 'valid', $prefix, 0, $notify);
            $sh2_output .= $mmr_display;
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= $mmr_display;
            if ($mmr_name =~ /FIRST/)  {
              my $md_reason = "";
              if (($mmr_data & 0x33330F0) > 0) {
		if (($mmr_data & 0x22220A0) > 0) {
                  $md_data_err_ce = 1;
	          $md_ce_fru_entry = $fru_count;
	          if ((reg_get_field_value($reg_ptr,'DQLP_DIR_COR')) == 1)      { $md_reason = "MD DQLP Detected a Corrected Directory Memory Error. "; }
	          elsif ((reg_get_field_value($reg_ptr,'DQRP_DIR_COR')) == 1)   { $md_reason = "MD DQRP Detected a Corrected Directory Memory Error. "; }
	          elsif ((reg_get_field_value($reg_ptr,'DQLP_INT_COR')) == 1)   { $md_reason = "MD DQLP Detected a Corrected AMO Data Error. "; }
	          elsif ((reg_get_field_value($reg_ptr,'DQLS_INT_COR')) == 1)   { 
		    $md_reason = "MD DQLS Detected a Corrected AMO Data Error. ";
	            $sh2_target_list[$sh2_cnt]->{DQLS_AMO_ERR} = 1;
	            $sh2_target_list[$sh2_cnt]->{DQLS_AMO_FRU} = $fru_count;
	          }
	          elsif ((reg_get_field_value($reg_ptr,'DQRP_INT_COR')) == 1)   { $md_reason = "MD DQRP Detected a Corrected AMO Data Error. "; }
	          elsif ((reg_get_field_value($reg_ptr,'DQRS_INT_COR')) == 1)   { 
		    $md_reason = "MD DQRS Detected a Corrected AMO Data Error. ";
	            $sh2_target_list[$sh2_cnt]->{DQRS_AMO_ERR} = 1;
	            $sh2_target_list[$sh2_cnt]->{DQRS_AMO_FRU} = $fru_count;
	          }
                  else  { $md_reason = "MD Reported Correctable ECC Error"; }
		  $global_err_val |= $SHUB_SBE_ERR;
		}
		else {
                  $md_data_err_uce = 1;
	          $md_uce_fru_entry = $fru_count;
	          if ((reg_get_field_value($reg_ptr,'DQLP_DIR_UC')) == 1)      { 
		    $md_reason = "MD DQLP Detected a Uncorrected Directory Memory Error. "; 
		    $md_dqlp_uce = 0x1;
                  }
	          elsif ((reg_get_field_value($reg_ptr,'DQRP_DIR_UC')) == 1)   { 
		    $md_reason = "MD DQRP Detected a Uncorrected Directory Memory Error. ";
		    $md_dqrp_uce = 0x1;
                  }
	          elsif ((reg_get_field_value($reg_ptr,'DQLP_INT_UC')) == 1)   { $md_reason = "MD DQLP Detected a Uncorrected AMO Data Error. "; }
	          elsif ((reg_get_field_value($reg_ptr,'DQLS_INT_UC')) == 1)   { 
		    $md_reason = "MD DQLS Detected a Uncorrected AMO Data Error. ";
	            $sh2_target_list[$sh2_cnt]->{DQLS_AMO_ERR} = 1;
	            $sh2_target_list[$sh2_cnt]->{DQLS_AMO_FRU} = $fru_count;
	          }
	          elsif ((reg_get_field_value($reg_ptr,'DQRP_INT_UC')) == 1)   { $md_reason = "MD DQRP Detected a Uncorrected AMO Data Error. "; }
	          elsif ((reg_get_field_value($reg_ptr,'DQRS_INT_UC')) == 1)   { 
		    $md_reason = "MD DQRS Detected a Uncorrected AMO Data Error. ";
	            $sh2_target_list[$sh2_cnt]->{DQRS_AMO_ERR} = 1;
	            $sh2_target_list[$sh2_cnt]->{DQRS_AMO_FRU} = $fru_count;
	          }
                  else  { $md_reason = "MD Reported Uncorrectable ECC Error"; }
		  $global_err_val |= $SHUB_UCE_ERR;
		}
              }
              if (($mmr_data & 0x786F444470F) > 0) {
                $md_hw_err = 1;
	        if ((reg_get_field_value($reg_ptr,'ILLEGAL_CMD')) == 1)       { $md_reason = "MD Received an Illegal Command. "; }
	        elsif ((reg_get_field_value($reg_ptr,'NONEXIST_ADDR')) == 1)  { $md_reason = "MD Received a request to Memory that doesn\'t exist.  "; }
	        elsif ((reg_get_field_value($reg_ptr,'DQLP_DIR_PERR')) == 1)  { 
		  $md_dqlp_perr = 0x1;
		  if ($md_dqlp_uce == 1)  {
		    $md_dqlp_uce_fru_entry = $fru_count;
		    $sh2_output .= save_fru('SHUB2', $nasid_value, $current_location, $md_reason, 'MD_DQLP_DIR_UC');
		  }
		  $md_reason = "MD Received a request that caused a directory protocol error.  ";
                }
	        elsif ((reg_get_field_value($reg_ptr,'DQRP_DIR_PERR')) == 1)  { 
		  $md_dqrp_perr = 0x1;
		  if ($md_dqrp_uce == 1)  {
		    $md_dqrp_uce_fru_entry = $fru_count;
		    $sh2_output .= save_fru('SHUB2', $nasid_value, $current_location, $md_reason, 'MD_DQRP_DIR_UC');
		  }
		  $md_reason = "MD Received a request that caused a directory protocol error.  ";
                }
	        elsif ((reg_get_field_value($reg_ptr,'ACX_INT_HW')) == 1)     { $md_reason = "MD Detected an Internal ACX RLM Hardware error.  "; }
	        elsif ((reg_get_field_value($reg_ptr,'ACY_INT_HW')) == 1)     { $md_reason = "MD Detected an Internal ACY RLM Hardware error.  "; }
	        elsif ((reg_get_field_value($reg_ptr,'DIR_ACC')) == 1)        { $md_reason = "MD Received a Processor Request to directory memory.  "; }
	        elsif ((reg_get_field_value($reg_ptr,'DQLP_INT_HW')) == 1)    { $md_reason = "MD Detected an Internal DQLP RLM Hardware error.  "; }
	        elsif ((reg_get_field_value($reg_ptr,'DQLS_INT_HW')) == 1)    { $md_reason = "MD Detected an Internal DQLS RLM Hardware error.  "; }
	        elsif ((reg_get_field_value($reg_ptr,'DQRP_INT_HW')) == 1)    { $md_reason = "MD Detected an Internal DQRP RLM Hardware error.  "; }
	        elsif ((reg_get_field_value($reg_ptr,'DQRS_INT_HW')) == 1)    { $md_reason = "MD Detected an Internal DQRS RLM Hardware error.  "; }
	        elsif ((reg_get_field_value($reg_ptr,'PI_REPLY_OVERFLOW')) == 1) { $md_reason = "MD Detected an PI to MD Reply Message Overflow error.  "; }
	        elsif ((reg_get_field_value($reg_ptr,'XN_REPLY_OVERFLOW')) == 1) { $md_reason = "MD Detected an XN to MD Reply Message Overflow error.  "; }
	        elsif ((reg_get_field_value($reg_ptr,'PI_REQUEST_OVERFLOW')) == 1) { $md_reason = "MD Detected an PI Request Message Overflow error.  "; }
	        elsif ((reg_get_field_value($reg_ptr,'XN_REQUEST_OVERFLOW')) == 1) { $md_reason = "MD Detected an XN Request Message Overflow error.  "; }
	        elsif ((reg_get_field_value($reg_ptr,'PI_PKT_SIZE')) == 1)    { $md_reason = "MD Detected a PI to MD Message Length error.  "; }
	        elsif ((reg_get_field_value($reg_ptr,'XN_PKT_SIZE')) == 1)    { $md_reason = "MD Detected a XN to MD Message Length error.  ";
                  $md_reason .= "\n\t\t\t\t\t\tSuspect Numa-Link packet corruption.";
                  $md_reason .= "\n\t\t\t\t\t\tRun specified rtrdmp command to trace failing route."; }
	        elsif ((reg_get_field_value($reg_ptr,'INCOMING_PI_REPLY_OVERFLOW')) == 1)    { $md_reason = "MD Detected a PI to MD Reply Message Overflow error overflow.  "; }
	        elsif ((reg_get_field_value($reg_ptr,'INCOMING_XN_REPLY_OVERFLOW')) == 1)    { $md_reason = "MD Detected a XN to MD Reply Message Overflow error overflow.  "; }
	        elsif ((reg_get_field_value($reg_ptr,'INCOMING_PI_REQUEST_OVERFLOW')) == 1)    { $md_reason = "MD Detected a PI to MD Request Message Overflow error overflow.  "; }
	        elsif ((reg_get_field_value($reg_ptr,'INCOMING_XN_REQUEST_OVERFLOW')) == 1)    { $md_reason = "MD Detected a XN to MD Request Message Overflow error overflow.  "; }
                else  { $md_reason = "MD Reported Hardware Error"; }
	        $global_err_val |= $SHUB_HW_ERR;
              }
	      $sh2_output .= save_fru('SHUB2', $nasid_value, $current_location, $md_reason, 'MD_ERR');
            }
	    if ((reg_get_field_value($reg_ptr,'DQLP_DIR_PERR')) == 1) {
	      $md_dqlp_perr = 1;
	    }
	    if ((reg_get_field_value($reg_ptr,'DQRP_DIR_PERR')) == 1) {
	      $md_dqrp_perr = 1;
	    }
	  }
  
	  elsif (($mmr_name =~ /SH2_MEM_ERROR_SUMMARY/) && (($mmr_data & $md_err_s_mask) != 0))  {
            $any_sh2_error = 1;
            $notify = "  <<<<<<<";
            $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
            my $mmr_display = lcl_reg_write_value($reg_ptr, 'valid', $prefix, 0, $notify);
            $sh2_output .= $mmr_display;
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= $mmr_display;
	    if ((($mmr_data & 0x22220A0) > 0) && ($md_data_err_ce == 0)) {
              $md_data_err_ce = 1;
	      $md_ce_fru_entry = $fru_count;
	      $global_err_val |= $SHUB_SBE_ERR;
              my $md_reason = "MD Reported Correctable ECC Error.";
	      $sh2_output .= save_fru('SHUB2', $nasid_value, $current_location, $md_reason, 'MD_DQLP_DIR_CE');
	    }
	    if ((($mmr_data & 0x1111050) > 0) && ($md_data_err_uce == 0)) {
              $md_data_err_uce = 1;
	      $md_uce_fru_entry = $fru_count;
	      $global_err_val |= $SHUB_UCE_ERR;
              my $md_reason = "MD Reported Uncorrectable ECC Error.";
	      $sh2_output .= save_fru('SHUB2', $nasid_value, $current_location, $md_reason, 'MD_DQLP_DATA_UC');
            }
            if ((($mmr_data & 0xF06F444470F) != 0) && ($md_hw_err == 0)){
              $md_hw_err = 1;
	      $global_err_val |= $SHUB_HW_ERR;
              my $md_reason = "MD Reported Hardware Error";
	      $sh2_output .= save_fru('SHUB2', $nasid_value, $current_location, $md_reason, 'MD_HW_ERR');
            }
            if (((reg_get_field_value($reg_ptr,'DQLS_INT_UC')) == 1) || ((reg_get_field_value($reg_ptr,'DQLS_INT_COR')) == 1)) {
	      $sh2_target_list[$sh2_cnt]->{DQLS_AMO_ERR} = 1;
	      if (! defined $sh2_target_list[$sh2_cnt]->{DQLS_AMO_FRU}) {
	        $sh2_target_list[$sh2_cnt]->{DQLS_AMO_FRU} = $fru_count - 1;
	      }
	    }
            if (((reg_get_field_value($reg_ptr,'DQRS_INT_UC')) == 1) || ((reg_get_field_value($reg_ptr,'DQRS_INT_COR')) == 1)) {
	      $sh2_target_list[$sh2_cnt]->{DQRS_AMO_ERR} = 1;
	      if (! defined $sh2_target_list[$sh2_cnt]->{DQRS_AMO_FRU}) {
	        $sh2_target_list[$sh2_cnt]->{DQRS_AMO_FRU} = $fru_count - 1;
	      }
	    }
	    if ((reg_get_field_value($reg_ptr,'DQLP_DIR_PERR')) == 1) {
	      $md_dqlp_perr = 1;
	    }
	    if ((reg_get_field_value($reg_ptr,'DQRP_DIR_PERR')) == 1) {
	      $md_dqrp_perr = 1;
	    }
	  }
  
	  elsif (($mmr_name =~ /SH2_MISC.*ERR_HDR_LOWER/) && ((reg_get_field_value($reg_ptr,'VALID')) == 1))  {
	    $get_misc_upper = 1;
	    $any_sh2_error = 1;
	    $notify = "  <<<<<<<";
	    $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
            $sh2_output .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
	    if ($mmr_name =~ /ADRS/) {
	      my $md_addr = (reg_get_field_value($reg_ptr,'ADDR')) << 3;
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf "\n\t\t\t\t\t\t    Actual Failing address: 0x%010llX\n", $md_addr;
      	      $md_numa_err_addr = $md_addr;
	      if (defined  $sh2_target_list[$sh2_cnt]->{DQLS_AMO_ERR})  {
                $sh2_target_list[$sh2_cnt]->{DQLS_AMO_ERR_ADDR} = $md_addr;
              }
              if (defined  $sh2_target_list[$sh2_cnt]->{DQRS_AMO_ERR})  {
                $sh2_target_list[$sh2_cnt]->{DQRS_AMO_ERR_ADDR} = $md_addr;
              }
              if (! defined  $sh2_target_list[$sh2_cnt]->{MD_BAD_ADDR})  {
	        save_address($prefix, ($md_addr & 0x3fffffff80), $nasid_value, 'MD_ADRS', $node_not_booted);
              }
	      $md_addr_lower = $mmr_data;
	    }
	    else {
	      my $src_node = (reg_get_field_value($reg_ptr,'SRC')) & 0x3fff;
      	      $md_numa_err_src_node = $src_node;
	      my $src_chip = ((reg_get_field_value($reg_ptr,'SRC')) >> 14) & 3;
	      $md_numa_err_src_chip = ($src_chip == 0) ? "md" : ($src_chip == 2) ? "pi" : "lb"; 
	      my $md_cmd = reg_get_field_value($reg_ptr,'CMD');
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf "\n  Source of failing message:  %s chiplet on NASID 0x%04X", $shub2_chiplet_list[$src_chip], $src_node;
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf "    Failing Command: %s\n", $sn2net_command_list[$md_cmd];
	    }
	  }
  
	  elsif (($mmr_name =~ /SH2_MISC_ERR_HDR_UPPER/) && ($get_misc_upper == 1))  {
	    $notify = "  <<<<<<<";
            $sh2_output .= lcl_reg_write_value($reg_ptr, 'valid', $prefix, 0, $notify);
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= lcl_reg_write_value($reg_ptr, 'valid', $prefix, 0, $notify);
	    my $save_src_fru = 0;
	    if ((reg_get_field_value($reg_ptr,'NONEXIST_ADDR')) == 1) {
	      $sh2_target_list[$sh2_cnt]->{MD_BAD_ADDR} = 1;		# so we don't read directory state on a bad address
	      $save_src_fru = 1;
	    }
	    elsif ((reg_get_field_value($reg_ptr,'XN_PKT_SIZE')) == 1) {
	      $md_numa_error = 1;		
	      $save_src_fru = 1;
	    }
	    elsif ((reg_get_field_value($reg_ptr,'DIR_ACC')) == 1) {
	      $save_src_fru = 1;
	    }
	    elsif ((reg_get_field_value($reg_ptr,'ILLEGAL_CMD')) == 1) {
	      $save_src_fru = 1;
	    }
	    elsif ((reg_get_field_value($reg_ptr,'DIR_PROTOCOL')) == 1) {
	      $save_src_fru = 1;
	    }
	    if (($save_src_fru == 1) && ($md_numa_err_src_node != $nasid_value))  {
              my $fru_rsn = "Source node indicated by NASID $prefix MD. ";
	      $sh2_output .= save_fru('SHUB2', $md_numa_err_src_node, 'UNDEFINED', $fru_rsn, 'MD_HW_ERR_REMOTE_NODE');
	    }
	  }
	  elsif (($mmr_name =~ /SH2_DIR_.*_ADRS_ERR_HDR_LOWER_A/) && ((reg_get_field_value($reg_ptr,'VALID')) == 1))  {
	    $any_sh2_error = 1;
	    $notify = "  <<<<<<<";
	    $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
            $sh2_output .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= decode_mem_err($mmr_name, $mmr_data, $prefix, $blade_type);;
            $sh2_target_list[$sh2_cnt]->{DIMM_ERR} = 1;
            $any_dimm_error++;
	  }
	  elsif (($mmr_name =~ /SH2_DIR_UC_ADRS_ERR_HDR_LOWER_A/) && ((reg_get_field_value($reg_ptr,'VALID')) == 0)
		 && ((($md_dqlp_perr == 1) && ($md_dqlp_uce == 1)) || (($md_dqrp_perr == 1) && ($md_dqrp_uce == 1))) )  {
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "\n Substituting protocol error address found in SH2_MISC_ADRS_ERR_HDR_LOWER_A\n";
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= decode_mem_err($mmr_name, $md_addr_lower, $prefix, $blade_type);;
            $sh2_target_list[$sh2_cnt]->{DIMM_ERR} = 1;
            $any_dimm_error++;
	  }
	  elsif (($mmr_name =~ /SH2_MD_DQLP_.*PERR/) && ($md_dqlp_perr == 1) && ($mmr_data != 0))  {
	    $notify = "  <<<<<<<";
            $sh2_output .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
	    if ($mmr_name =~ /PERR0/) {
	      my $src_node = (reg_get_field_value($reg_ptr,'SRC')) & 0x3fff;
	      my $src_chip = ((reg_get_field_value($reg_ptr,'SRC')) >> 14) & 3;
	      my $md_cmd = reg_get_field_value($reg_ptr,'CMD');
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf "\n  Source of failing message:  %s chiplet on NASID 0x%04X", $shub2_chiplet_list[$src_chip], $src_node;
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf "    Failing Command: %s\n", $sn2net_command_list[$md_cmd];
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf "  Directory State before PERR:  %s ", 
		decode_dir_state(reg_get_field_value($reg_ptr,'DIR'), reg_get_field_value($reg_ptr,'COR'), reg_get_field_value($reg_ptr,'UNC'));
	      if ($md_cmd == 0x8c) {
	        $md_cmd_is_prge = 1;
	      }
	      elsif ($md_cmd == 0xcc) {
	        $md_cmd_is_xfer = 1;
	      }
	    }
	    if ($mmr_name =~ /PERR1/) {
	      $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= decode_mybit($mmr_data);
	    }
	  }
	  elsif (($mmr_name =~ /SH2_MD_DQRP_.*PERR/) && ($md_dqrp_perr == 1) && ($mmr_data != 0))  {
	    $notify = "  <<<<<<<";
            $sh2_output .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
	    if ($mmr_name =~ /PERR0/) {
	      my $src_node = (reg_get_field_value($reg_ptr,'SRC')) & 0x3fff;
	      my $src_chip = ((reg_get_field_value($reg_ptr,'SRC')) >> 14) & 3;
	      my $md_cmd = reg_get_field_value($reg_ptr,'CMD');
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf "\n  Source of failing message:  %s chiplet on NASID 0x%04X", $shub2_chiplet_list[$src_chip], $src_node;
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf "    Failing Command: %s\n", $sn2net_command_list[$md_cmd];
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf "  Directory State before PERR:  %s ", 
		decode_dir_state(reg_get_field_value($reg_ptr,'DIR'), reg_get_field_value($reg_ptr,'COR'), reg_get_field_value($reg_ptr,'UNC'));
	      if ($md_cmd == 0x8c) {
	        $md_cmd_is_prge = 1;
	      }
	      elsif ($md_cmd == 0xcc) {
	        $md_cmd_is_xfer = 1;
	      }
	    }
	    if ($mmr_name =~ /PERR1/) {
	      $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= decode_mybit($mmr_data);
	    }
	  }
	  elsif (($mmr_name =~ /SH2_MD_DQLS_MMR_XAMOPW_ERR/) && ($sh2_target_list[$sh2_cnt]->{DQLS_AMO_ERR} == 1) && ($mmr_data != 0))  {
	    $notify = "  <<<<<<<";
            $sh2_output .= lcl_reg_write_value($reg_ptr, 'valid', $prefix, 0, $notify);
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
	    my $syn_shift_val = $sh2_target_list[$sh2_cnt]->{DQLS_AMO_ERR_ADDR} & 8;   # failure address determines where the syndrome is

            if (reg_get_field_value($reg_ptr,'SCOR') == 1)  {
              $sh2_target_list[$sh2_cnt]->{DQLS_AMO_SCOR} = 1;
              $sh2_target_list[$sh2_cnt]->{DQLS_AMO_SSYN} = ((reg_get_field_value($reg_ptr,'SSYN')) >> $syn_shift_val) & 0xff;
            }
            if (reg_get_field_value($reg_ptr,'SUNC') == 1)  {
              $sh2_target_list[$sh2_cnt]->{DQLS_AMO_SUNC} = 1;
              $sh2_target_list[$sh2_cnt]->{DQLS_AMO_SSYN} = ((reg_get_field_value($reg_ptr,'SSYN')) >> $syn_shift_val) & 0xff;
            }
            if (reg_get_field_value($reg_ptr,'RCOR') == 1)  {
              $sh2_target_list[$sh2_cnt]->{DQLS_AMO_RCOR} = 1;
              $sh2_target_list[$sh2_cnt]->{DQLS_AMO_RSYN} = ((reg_get_field_value($reg_ptr,'RSYN')) >> $syn_shift_val) & 0xff;
              $sh2_target_list[$sh2_cnt]->{DIMM_ERR} = 1;
              $any_dimm_error++;
            }
            if (reg_get_field_value($reg_ptr,'RUNC') == 1)  {
              $sh2_target_list[$sh2_cnt]->{DQLS_AMO_RUNC} = 1;
              $sh2_target_list[$sh2_cnt]->{DQLS_AMO_RSYN} = ((reg_get_field_value($reg_ptr,'RSYN')) >> $syn_shift_val) & 0xff;
              $sh2_target_list[$sh2_cnt]->{DIMM_ERR} = 1;
              $any_dimm_error++;
            }
            sh2_decode_amo_dimm_info('DQLS', $prefix, $blade_type);  
	  }
	  elsif (($mmr_name =~ /SH2_MD_DQRS_MMR_YAMOPW_ERR/) && ($sh2_target_list[$sh2_cnt]->{DQRS_AMO_ERR} == 1) && ($mmr_data != 0))  {

	    $notify = "  <<<<<<<";
            $sh2_output .= lcl_reg_write_value($reg_ptr, 'valid', $prefix, 0, $notify);
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
	    my $syn_shift_val = $sh2_target_list[$sh2_cnt]->{DQRS_AMO_ERR_ADDR} & 8;   # failure address determines where the syndrome is

            if (reg_get_field_value($reg_ptr,'SCOR') == 1)  {
              $sh2_target_list[$sh2_cnt]->{DQRS_AMO_SCOR} = 1;
              $sh2_target_list[$sh2_cnt]->{DQRS_AMO_SSYN} = ((reg_get_field_value($reg_ptr,'SSYN')) >> $syn_shift_val) & 0xff;
            }
            if (reg_get_field_value($reg_ptr,'SUNC') == 1)  {
              $sh2_target_list[$sh2_cnt]->{DQRS_AMO_SUNC} = 1;
              $sh2_target_list[$sh2_cnt]->{DQRS_AMO_SSYN} = ((reg_get_field_value($reg_ptr,'SSYN')) >> $syn_shift_val) & 0xff;
            }
            if (reg_get_field_value($reg_ptr,'RCOR') == 1)  {
              $sh2_target_list[$sh2_cnt]->{DQRS_AMO_RCOR} = 1;
              $sh2_target_list[$sh2_cnt]->{DQRS_AMO_RSYN} = ((reg_get_field_value($reg_ptr,'RSYN')) >> $syn_shift_val) & 0xff;
              $sh2_target_list[$sh2_cnt]->{DIMM_ERR} = 1;
              $any_dimm_error++;
            }
            if (reg_get_field_value($reg_ptr,'RUNC') == 1)  {
              $sh2_target_list[$sh2_cnt]->{DQRS_AMO_RUNC} = 1;
              $sh2_target_list[$sh2_cnt]->{DQRS_AMO_RSYN} = ((reg_get_field_value($reg_ptr,'RSYN')) >> $syn_shift_val) & 0xff;
              $sh2_target_list[$sh2_cnt]->{DIMM_ERR} = 1;
              $any_dimm_error++;
            }

            sh2_decode_amo_dimm_info('DQRS', $prefix, $blade_type);      #  GEM  need to coordinate FRU info
	  }
  
	  elsif ((($mmr_name =~ /SH2_XN_NI0_IILB_CMP_EXP_DATA/) || ($mmr_name =~ /SH2_XN_NI0_LLP_CMP_EXP_DATA/)) && ($valid_psp_state == 1) && (($mmr_data & $legit_psp_mask) == $legit_psp_mask))  {		# decode PSP scratch reg  - see BW# 935715
	    $any_sh2_error = 1;
	    $notify = "  <<<<<<<";
	    $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
            $sh2_output .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
            my $orig_fru_cnt = $fru_count;
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= decode_psp_mmr($mmr_name, $mmr_data, $mmr_addr, $prefix, $nasid_value, $current_location, $shub_hw_err);
	    if ($fru_count > $orig_fru_cnt) {
	      $shub_hw_err = 1;
	    }
	  }
  
	  elsif (($mmr_name =~ /STAMP/) && ((reg_get_field_value($reg_ptr,'VALID')) == 1))  {
	    $notify = "  <<<<<<<";
            $sh2_output .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
	    capture_time_stamp($reg_ptr);
  	  }
  
          elsif (($mmr_name =~ /SH2_XN_MD_SIC_CMP_EXP_HDR0/) && ((($mmr_data >> 56) & 0xff) == 0xaa))  {
			  # check for valid target id 0
            $sh2_target_list[$sh2_cnt]->{TARGET_ID_0} = $mmr_data & 0xffffffffffffff;
            $sh2_target_list[$sh2_cnt]->{TARGET_0_VAL} = 1;
            $sh2_output .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
            # print "\n\nTarget Identifier 0 is Valid!!!\n\n";
          }
  
          elsif ( ($mmr_name =~ /SH2_XN_MD_NI1_CMP_ENABLE1/)
            &&  ((($mmr_data >> 63) & 1) == 1)
            &&  ($sh2_target_list[$sh2_cnt]->{TARGET_0_VAL} == 1)  )  { #  check for valid target id time stamp
            $sh2_target_list[$sh2_cnt]->{TARGET_ID_0_TIMESTAMP} = $mmr_data & 0xffffffffffffff;
            $sh2_output .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
            # print "\n\nTarget Identifier 0 Timestamp is Valid!!!\n\n";
          }
  
          elsif (($mmr_name =~ /SH2_XN_NI0_MD_CMP_EXP_HDR1/) &&  ((($mmr_data >> 56) & 0xff) == 0xaa))  {  
			  # check for valid target id 1
            $sh2_target_list[$sh2_cnt]->{TARGET_ID_1} = $mmr_data & 0xffffffffffffff;
            $sh2_target_list[$sh2_cnt]->{TARGET_1_VAL} = 1;
            $sh2_output .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
            # print "\n\nTarget Identifier 1 is Valid!!!\n\n";
          }
  
          elsif (  ($mmr_name =~ /SH2_XN_NI0_MD_CMP_EXP_HDR0/)
            &&  ((($mmr_data >> 63) & 1) == 1)
            &&  ($sh2_target_list[$sh2_cnt]->{TARGET_1_VAL} == 1)  )  { #  check for valid target id time stamp
            $sh2_target_list[$sh2_cnt]->{TARGET_ID_1_TIMESTAMP} = $mmr_data & 0xffffffffffffff;
            $sh2_output .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
            # print "\n\nTarget Identifier 1 Timestamp is Valid!!!\n\n";
          }
  
          elsif (($mmr_name =~ /SH2_XN_IILB_NI1_CMP_EXP_DATA0/) && ((($mmr_data >> 56) & 0xff) == 0xaa))  {
		  	  # check for valid target id 2
            $sh2_target_list[$sh2_cnt]->{TARGET_ID_2} = $mmr_data & 0xffffffffffffff;
            $sh2_target_list[$sh2_cnt]->{TARGET_2_VAL} = 1;
            $sh2_output .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
            # print "\n\nTarget Identifier 2 is Valid!!!\n\n";
          }
  
          elsif ( ($mmr_name =~ /SH2_XN_IILB_NI1_CMP_EXP_DATA1/)
            &&  ((($mmr_data >> 63) & 1) == 1)
            &&  ($sh2_target_list[$sh2_cnt]->{TARGET_2_VAL} == 1)  )  { #  check for valid target id time stamp
            $sh2_target_list[$sh2_cnt]->{TARGET_ID_2_TIMESTAMP} = $mmr_data & 0xffffffffffffff;
            $sh2_output .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
            # print "\n\nTarget Identifier 2 Timestamp is Valid!!!\n\n";
          }
  
          elsif (($mmr_name =~ /SH2_XN_IILB_NI0_CMP_EXP_DATA0/) &&  ((($mmr_data >> 56) & 0xff) == 0xaa))  {  
			  # check for valid target id 3
            $sh2_target_list[$sh2_cnt]->{TARGET_ID_3} = $mmr_data & 0xffffffffffffff;
            $sh2_target_list[$sh2_cnt]->{TARGET_3_VAL} = 1;
            $sh2_output .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
            # print "\n\nTarget Identifier 3 is Valid!!!\n\n";
          }
  
          elsif (  ($mmr_name =~ /SH2_XN_IILB_NI0_CMP_EXP_DATA1/)
            &&  ((($mmr_data >> 63) & 1) == 1)
            &&  ($sh2_target_list[$sh2_cnt]->{TARGET_3_VAL} == 1)  )  { #  check for valid target id time stamp
            $sh2_target_list[$sh2_cnt]->{TARGET_ID_3_TIMESTAMP} = $mmr_data & 0xffffffffffffff;
            $sh2_output .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
            # print "\n\nTarget Identifier 3 Timestamp is Valid!!!\n\n";
          }
  
	  elsif ( $mmr_name =~ /SH2_PI_XN2PI_REQUEST_VC_STATUS/ )  {
	    my $icc_stat = reg_get_field_value($reg_ptr, 'ICC_STATUS');
	    my $icc_oflow = reg_get_field_value($reg_ptr, 'ICC_OVERFLOW');
	    my $icc_uflow = reg_get_field_value($reg_ptr, 'ICC_UNDERFLOW');
	    my $in_hdr = reg_get_field_value($reg_ptr, 'INPUT_HDR_CRD_STAT');
	    my $in_dat = reg_get_field_value($reg_ptr, 'INPUT_DAT_CRD_STAT');	#  apparently not working
	    my $in_que = reg_get_field_value($reg_ptr, 'INPUT_QUEUE_STAT');	#  apparently not working
	    if (($icc_stat < 5) || ($icc_oflow == 1) || ($icc_uflow == 1) || ($in_hdr < 2) ) {
	      if (($icc_oflow == 1) || ($icc_uflow == 1)) {
	        $global_err_val |= $SHUB_HW_ERR;
	      }  
	      else {
	        $global_err_val |= $SHUB_CRD_ERR;
	      }
	      $any_sh2_error = 1;
	      $notify = "  <<<<<<<";
	      $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
	    }
            $sh2_output .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
	  }
	  elsif ( $mmr_name =~ /SH2_PI_*_VC_STATUS/ )  {
            $sh2_output .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
	  }
  
	  elsif ($mmr_name =~ /CREDIT_MGR/)  {
	    if ( ((reg_get_field_value($reg_ptr, 'IILB_CNT_DYN')) < 2) 
	      || ((reg_get_field_value($reg_ptr, 'NI0_CNT_DYN')) < 2)
	      || ((reg_get_field_value($reg_ptr, 'NI1_CNT_DYN')) < 2))  {
	      $global_err_val |= $SHUB_CRD_ERR;
              $any_sh2_error = 1;
              $notify = "  <<<<<<<";
              $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= lcl_reg_write_value($reg_ptr, 'dyn', $prefix, 0, $notify);
	    }
            $sh2_output .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
	  }
	  elsif ($mmr_name =~ /SH2_XNMD_SIC_FLOW/)  {
	    if (((reg_get_field_value($reg_ptr, 'CREDIT_VC0_DYN')) < 2) || ((reg_get_field_value($reg_ptr, 'CREDIT_VC2_DYN')) < 2))  {
	      $global_err_val |= $SHUB_CRD_ERR;
              $any_sh2_error = 1;
              $notify = "  <<<<<<<";
              $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= lcl_reg_write_value($reg_ptr, 'dyn', $prefix, 0, $notify);
  
	      if ((reg_get_field_value($reg_ptr, 'CREDIT_VC0_DYN') < 2) && ($file_mode == 0)) {
	        my $chk_count = 30;
      	        my $low_crd_cnt = xn_sic_cred_chk($mmr_name, $chk_count, 'CREDIT_VC0_DYN', 2); 
	        if (($chk_count * 10) == $low_crd_cnt) {
      	          $low_md_rq_crd = 1;
	        }
	        else {
            	  $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf "\n    The VC0_DYN value in %s was less than 2 in %d out of %d read attempts.\n\n",
			  $mmr_name, $low_crd_cnt, ($chk_count * 10);
	        }
	      }
	    }
            $sh2_output .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
	  }
  
	  elsif ($mmr_name =~ /SH2_XNPI_SIC_FLOW/)  {
	    if (((reg_get_field_value($reg_ptr, 'CREDIT_VC0_DYN')) < 3) || ((reg_get_field_value($reg_ptr, 'CREDIT_VC2_DYN')) < 2))  {
	      $global_err_val |= $SHUB_CRD_ERR;
              $any_sh2_error = 1;
              $notify = "  <<<<<<<";
              $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= lcl_reg_write_value($reg_ptr, 'dyn', $prefix, 0, $notify);
              if ((reg_get_field_value($reg_ptr, 'CREDIT_VC2_DYN') < 2) && ($file_mode == 0)) {
                my $chk_count = 30;
                my $low_crd_cnt = xn_sic_cred_chk($mmr_name, $chk_count, 'CREDIT_VC2_DYN', 2);
                if (($chk_count * 10) == $low_crd_cnt) {
                  $low_pi_rp_crd = 1;
                }
                else {
                  $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf "\n    The SH2_XNPI_SIC_FLOW VC2_DYN value in %s was less than 2 in %d out of %d read attempts.\n\n",
                          $mmr_name, $low_crd_cnt, ($chk_count * 10);
                }
              }
              if ((reg_get_field_value($reg_ptr, 'CREDIT_VC0_DYN') < 3) && ($file_mode == 0)) {
                my $chk_count = 30;
                my $low_crd_cnt = xn_sic_cred_chk($mmr_name, $chk_count, 'CREDIT_VC0_DYN', 3);
                if (($chk_count * 10) == $low_crd_cnt) {
                  $low_pi_rq_crd = 1;
                }
                else {
                  $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf "\n    The SH2_XNPI_SIC_FLOW VC0_DYN value in %s was less than 2 in %d out of %d read attempts.\n\n",
                          $mmr_name, $low_crd_cnt, ($chk_count * 10);
                }
              }
	    }
	    if (reg_get_field_value($reg_ptr, 'CREDIT_VC2_DYN') == 0)  {
	      $sh2_target_list[$sh2_cnt]->{PI_DLOCK} = 1;
 	    }
            $sh2_output .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
	  }
  
	  elsif ($mmr_name =~ /SH2_XNII_INTRA_FLOW/)  {
	    if (((reg_get_field_value($reg_ptr, 'CREDIT_VC0_DYN')) < 3) || ((reg_get_field_value($reg_ptr, 'CREDIT_VC2_DYN')) < 3))  {
	      $global_err_val |= $SHUB_CRD_ERR;
              $any_sh2_error = 1;
              $notify = "  <<<<<<<";
              $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= lcl_reg_write_value($reg_ptr, 'dyn', $prefix, 0, $notify);
	    }
            $sh2_output .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
	  }
  
	  elsif ($mmr_name =~ /SH2_XNLB_INTRA_FLOW/)  {
	    if (((reg_get_field_value($reg_ptr, 'CREDIT_VC0_DYN')) < 3) || ((reg_get_field_value($reg_ptr, 'CREDIT_VC2_DYN')) < 2))  {
	      $global_err_val |= $SHUB_CRD_ERR;
              $any_sh2_error = 1;
              $notify = "  <<<<<<<";
              $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= lcl_reg_write_value($reg_ptr, 'dyn', $prefix, 0, $notify);
	    }
            $sh2_output .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
	  }
  
	  elsif (  ($mmr_name =~ /SH2_XNIILB_FR_NI0_INTRA_FLOW_CREDIT/)
	        || ($mmr_name =~ /SH2_XNIILB_FR_NI1_INTRA_FLOW_CREDIT/) 
	        || ($mmr_name =~ /SH2_XNIILB_FR_MD_INTRA_FLOW_CREDIT/) 
	        || ($mmr_name =~ /SH2_XNIILB_FR_PI_INTRA_FLOW_CREDIT/) 
	        || ($mmr_name =~ /SH2_XNIILB_FR_IILB_INTRA_FLOW_CREDIT/) )  {
	    if (((reg_get_field_value($reg_ptr, 'VC0_DYN')) < 3) || ((reg_get_field_value($reg_ptr, 'VC2_DYN')) < 3))  {
	      $global_err_val |= $SHUB_CRD_ERR;
              $any_sh2_error = 1;
              $notify = "  <<<<<<<";
              $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= lcl_reg_write_value($reg_ptr, 'dyn', $prefix, 0, $notify);
	    }
            $sh2_output .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
	  }
  
	  elsif ($mmr_name =~ /SH2_XNNI0_LINK_0_FLOW/)  {
	    if ((reg_get_field_value($reg_ptr, 'CREDIT_VC0_DYN') < 3) && ($ni0_link_up == 1))  {
	      $global_err_val |= $SHUB_CRD_ERR;
              $any_sh2_error = 1;
              $notify = "  <<<<<<<";
              $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= lcl_reg_write_value($reg_ptr, 'dyn', $prefix, 0, $notify);
	    }
            $sh2_output .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
	  }
  
	  elsif ($mmr_name =~ /SH2_XNNI0_LINK_1_FLOW/)  {
	    if ((reg_get_field_value($reg_ptr, 'CREDIT_VC1_DYN') < 3) && ($ni0_link_up == 1))  {
	      $global_err_val |= $SHUB_CRD_ERR;
              $any_sh2_error = 1;
              $notify = "  <<<<<<<";
              $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= lcl_reg_write_value($reg_ptr, 'dyn', $prefix, 0, $notify);
	    }
            $sh2_output .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
	  }
  
	  elsif ($mmr_name =~ /SH2_XNNI0_LINK_2_FLOW/)  {
	    if ((reg_get_field_value($reg_ptr, 'CREDIT_VC2_DYN') < 3) && ($ni0_link_up == 1))  {
	      $global_err_val |= $SHUB_CRD_ERR;
              $any_sh2_error = 1;
              $notify = "  <<<<<<<";
              $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= lcl_reg_write_value($reg_ptr, 'dyn', $prefix, 0, $notify);
	    }
            $sh2_output .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
	  }
  
	  elsif ($mmr_name =~ /SH2_XNNI0_LINK_3_FLOW/)  {
	    if ((reg_get_field_value($reg_ptr, 'CREDIT_VC3_DYN') < 3) && ($ni0_link_up == 1))  {
	      $global_err_val |= $SHUB_CRD_ERR;
              $any_sh2_error = 1;
              $notify = "  <<<<<<<";
              $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= lcl_reg_write_value($reg_ptr, 'dyn', $prefix, 0, $notify);
	    }
            $sh2_output .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
	  }
  
	  elsif ($mmr_name =~ /SH2_XNNI1_LINK_0_FLOW/)  {
	    if ((reg_get_field_value($reg_ptr, 'CREDIT_VC0_DYN') < 3) && ($ni1_link_up == 1))  {
	      $global_err_val |= $SHUB_CRD_ERR;
              $any_sh2_error = 1;
              $notify = "  <<<<<<<";
              $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= lcl_reg_write_value($reg_ptr, 'dyn', $prefix, 0, $notify);
	    }
            $sh2_output .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
	  }
  
	  elsif ($mmr_name =~ /SH2_XNNI1_LINK_1_FLOW/)  {
	    if ((reg_get_field_value($reg_ptr, 'CREDIT_VC1_DYN') < 3) && ($ni1_link_up == 1))  {
	      $global_err_val |= $SHUB_CRD_ERR;
              $any_sh2_error = 1;
              $notify = "  <<<<<<<";
              $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= lcl_reg_write_value($reg_ptr, 'dyn', $prefix, 0, $notify);
	    }
            $sh2_output .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
	  }
  
	  elsif ($mmr_name =~ /SH2_XNNI1_LINK_2_FLOW/)  {
	    if ((reg_get_field_value($reg_ptr, 'CREDIT_VC2_DYN') < 3) && ($ni1_link_up == 1))  {
	      $global_err_val |= $SHUB_CRD_ERR;
              $any_sh2_error = 1;
              $notify = "  <<<<<<<";
              $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= lcl_reg_write_value($reg_ptr, 'dyn', $prefix, 0, $notify);
	    }
            $sh2_output .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
	  }
  
	  elsif ($mmr_name =~ /SH2_XNNI1_LINK_3_FLOW/)  {
	    if ((reg_get_field_value($reg_ptr, 'CREDIT_VC3_DYN') < 3) && ($ni1_link_up == 1))  {
	      $global_err_val |= $SHUB_CRD_ERR;
              $any_sh2_error = 1;
              $notify = "  <<<<<<<";
              $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= lcl_reg_write_value($reg_ptr, 'dyn', $prefix, 0, $notify);
	    }
            $sh2_output .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
	  }
  
	  elsif ($mmr_name =~ /SH2_BT_XMIT_DYN_CREDIT_STAT/)  {
	    if ((reg_get_field_value($reg_ptr, 'XN_RSP_CREDIT') < 3) || (reg_get_field_value($reg_ptr, 'XN_REQ_CREDIT') < 3))  {
	      $global_err_val |= $SHUB_CRD_ERR;
              $any_sh2_error = 1;
              $notify = "  <<<<<<<";
              $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
	    }
            $sh2_output .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
	  }
  
	  elsif ($mmr_name =~ /SH2_BT_NIQ_DYN_CREDIT_STAT/)  {
	    if ((reg_get_field_value($reg_ptr, 'NIQ_RSP_CREDIT_CNT') < 3) || (reg_get_field_value($reg_ptr, 'NIQ_REQ_CREDIT_CNT') < 3))  {
	      $global_err_val |= $SHUB_CRD_ERR;
              $any_sh2_error = 1;
              $notify = "  <<<<<<<";
              $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
	    }
            $sh2_output .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
	  }
  
	  elsif (($mmr_name =~ /SH2_PIO_WRITE_STATUS/) 
	      && ((reg_get_field_value($reg_ptr, 'MULTI_WRITE_ERROR') == 1)
	      ||  (reg_get_field_value($reg_ptr, 'WRITE_DEADLOCK') == 1)
	      ||  (reg_get_field_value($reg_ptr, 'PENDING_WRITE_COUNT') > 0x10)
	      ||  (reg_get_field_value($reg_ptr, 'WRITE_ERROR') == 1)) )  {
	      $global_err_val |= $SHUB_HW_ERR;
              $any_sh2_error = 1;
              $notify = "  <<<<<<<";
              $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= lcl_reg_write_value($reg_ptr, 'valid', $prefix, 0, $notify);
              $sh2_output .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
	      my $error_addr = (reg_get_field_value($reg_ptr, 'WRITE_ERROR_ADDRESS')) << 3;
	      my ($net_addr, $net_nasid) = fsb_to_net($error_addr); 
	      if ((($net_addr >> 28) & 0xf) == 7)  {
	        $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf "\n\tThe failing PIO address (0x%014llX) is a PTC access on NASID 0x%04X\n",
			  $net_addr, $net_nasid;
	      }
	      else {
	        $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf "\n\tThe failing PIO address (0x%014llX) is on NASID 0x%04X\n",
			  $net_addr, $net_nasid;
	      }
  
	  }
          elsif (($mmr_name =~ /SH2_PIO_WC_WRITE_STATUS/)
              && ((reg_get_field_value($reg_ptr, 'MULTI_WRITE_ERROR') == 1) 
	      ||  (reg_get_field_value($reg_ptr, 'PENDING_WRITE_COUNT') > 0x10) 
	      ||  (reg_get_field_value($reg_ptr, 'WRITE_ERROR') == 1)) )  {
	      $global_err_val |= $SHUB_HW_ERR;
              $any_sh2_error = 1;
              $notify = "  <<<<<<<";
              $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= lcl_reg_write_value($reg_ptr, 'valid', $prefix, 0, $notify);
              $sh2_output .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
              my $error_addr = (reg_get_field_value($reg_ptr, 'WRITE_ERROR_ADDRESS')) << 3;
              my ($net_addr, $net_nasid) = fsb_to_net($error_addr);
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf "\n\tThe failing PIO address (0x%014llX) is on NASID 0x%04X\n",
                          $net_addr, $net_nasid;
          }
          elsif ($mmr_name =~ /LLP_STAT/) { 
	    if (reg_get_field_value($reg_ptr, 'LINK_RESET_STATE') == 0xe)  { 
	      if ($mmr_name =~ /NI0/) {
	        $ni0_link_up = 1;
	      }  
	      else  {
	        $ni1_link_up = 1;
	      }
	    }  
	    else {
	      $global_err_val |= $SHUB_HW_ERR;
              $any_sh2_error = 1;
              $notify = "  <<<<<<<";
              $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= lcl_reg_write_value($reg_ptr, 'valid', $prefix, 0, $notify);
	    } 
            $sh2_output .= lcl_reg_write_value($reg_ptr, 'all', $prefix, 0, $notify);
	  }
		#  Shadow MMR parsing starts here
          elsif (($mmr_name =~ /SH2_XN_IILB_LB_CMP_EXP_DATA1/) && (($mmr_data & $shadow_event_mask) != 0)) { 	

	    my $shadow_mmr = 'SH2_EVENT_OCCURRED';
	    $shadow_reg_ptr = reg_create_mmr_inst("shub2_01", $shadow_mmr);
	    reg_set_value($shadow_reg_ptr, $mmr_data);
	    my $shadow_msg = "\n         $mmr_name is shadow MMR for $shadow_mmr";
            $any_sh2_error = 1;
            $notify = "  <<<<<<<";
            $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
	    my $mmr_display = lcl_reg_write_value($shadow_reg_ptr, 'valid', $prefix, 0, $notify);
            $sh2_output .= "$shadow_msg $mmr_display";
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "$shadow_msg $mmr_display";

            if (((reg_get_field_value($shadow_reg_ptr,'PI_CE_INT')) == 1) && ($valid_psp_state == 1) && ($processor_count > 0)) {
              $global_err_val |= $SHUB_SBE_ERR;
              $shadow_pi_ce = 1;
            }
            if ((reg_get_field_value($shadow_reg_ptr,'XN_CE_INT')) == 1) {
              $shadow_xn_ce = 1;
	      $global_err_val |= $SHUB_SBE_ERR;
            }
            if (((reg_get_field_value($shadow_reg_ptr,'PI_UCE_INT')) == 1) && ($valid_psp_state == 1) && ($processor_count > 0)) {
              $shadow_pi_uce = 1;
	      $shub_hw_err = 1;
            }
            if ((reg_get_field_value($shadow_reg_ptr,'XN_UCE_INT')) == 1) {
              $shadow_xn_uce = 1;
	      $global_err_val |= $SHUB_UCE_ERR;
	      $shub_hw_err = 1;
            }

	  }
          elsif (($mmr_name =~ /SH2_XN_PI_NI0_CMP_ENABLE0/) && ($mmr_data != 0) 
	      && (($shadow_pi_ce == 1) || ($shadow_pi_uce == 1)) ) { 	
	    my $shadow_mmr = 'SH2_PI_FIRST_ERROR';
	    $shadow_reg_ptr = reg_create_mmr_inst("shub2_01", $shadow_mmr);
	    reg_set_value($shadow_reg_ptr, $mmr_data);
	    my $shadow_msg = "\n         $mmr_name is shadow MMR for $shadow_mmr";
            $notify = "  <<<<<<<";
            my $mmr_display = lcl_reg_write_value($shadow_reg_ptr, 'valid', $prefix, 0, $notify);
            $sh2_output .= "$shadow_msg $mmr_display";
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "$shadow_msg $mmr_display";
	    if ((reg_get_field_value($shadow_reg_ptr,'FSB_SHUB_UCE')) == 1) {	
	      $shadow_pi_write_uce = 1;
	      $first_pi_uce = 4;
	      $global_err_val |= $SHUB_HW_ERR;
	    }
	    if ((reg_get_field_value($shadow_reg_ptr,'SHUB_FSB_UCE')) == 1) {
	      $first_pi_uce = 3;
	      $shadow_pi_read_uce = 1;
	    }
	    if ((reg_get_field_value($shadow_reg_ptr,'SHUB_FSB_CE')) == 1) {
	      $shadow_pi_read_ce = 1;
	      $global_err_val |= $SHUB_SBE_ERR;
	    }
	  }
          elsif (($mmr_name =~ /SH2_XN_PI_NI0_CMP_ENABLE1/) && ($mmr_data != 0) 
	      && (($shadow_pi_ce == 1) || ($shadow_pi_uce == 1)) ) { 	
	    my $shadow_mmr = 'SH2_PI_ERROR_SUMMARY';
	    $shadow_reg_ptr = reg_create_mmr_inst("shub2_01", $shadow_mmr);
	    reg_set_value($shadow_reg_ptr, $mmr_data);
	    my $shadow_msg = "\n         $mmr_name is shadow MMR for $shadow_mmr";
            $notify = "  <<<<<<<";
            my $mmr_display = lcl_reg_write_value($shadow_reg_ptr, 'valid', $prefix, 0, $notify);
            $sh2_output .= "$shadow_msg $mmr_display";
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "$shadow_msg $mmr_display";
	    if ((reg_get_field_value($shadow_reg_ptr,'FSB_SHUB_UCE')) == 1) {		#  GEM - need to qualify PI shadow detail with these
	      $shadow_pi_write_uce = 1;
	      $global_err_val |= $SHUB_HW_ERR;
	    }
	    if ((reg_get_field_value($shadow_reg_ptr,'FSB_SHUB_CE')) == 1) {
	      $shadow_pi_write_ce = 1;
	      $global_err_val |= $SHUB_SBE_ERR;
	    }
	    if ((reg_get_field_value($shadow_reg_ptr,'SHUB_FSB_UCE')) == 1) {
	      $shadow_pi_read_uce = 1;
	    }
	    if ((reg_get_field_value($shadow_reg_ptr,'SHUB_FSB_CE')) == 1) {
	      $shadow_pi_read_ce = 1;
	      $global_err_val |= $SHUB_SBE_ERR;
	    }
	  }
	
		###############   PI CORRECTED SHADOW MMRS  #####################
          elsif (($mmr_name =~ /SH2_XN_IILB_LB_CMP_ENABLE0/) && ($mmr_data != 0) && ($shadow_pi_read_ce == 1) && ($shadow_pi_ce == 1)) { 	
	    my $shadow_mmr = 'SH2_PI_CORRECTED_DETAIL_1';
	    $shadow_reg_ptr = reg_create_mmr_inst("shub2_01", $shadow_mmr);
            my $reg_data = $mmr_data;
	    reg_set_value($shadow_reg_ptr, $reg_data);
	    my $shadow_msg = "\n         $mmr_name is shadow MMR for $shadow_mmr";
            $notify = "  <<<<<<<";
            my $mmr_display = lcl_reg_write_value($shadow_reg_ptr, 'all', $prefix, 0, $notify);
            $sh2_output .= "$shadow_msg $mmr_display";
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "$shadow_msg $mmr_display";
            $sh2_target_list[$sh2_cnt]->{DIMM_ERR} = 1;
            $any_dimm_error++;
	    $pi_shub_ce_err = 2;
	    $shadow_pi_read_ce = 1;
	    $pi_ce_fru_entry = $fru_count;
            my $fru_rsn = "PI Reported SHUB_FSB_CE.  ";
	    $sh2_output .= save_fru('SHUB2', $nasid_value, $current_location, $fru_rsn, 'SHADOW_SHUB_FSB_CE');
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= decode_mem_err($shadow_mmr, $mmr_data, $prefix, $blade_type);
	  }
          elsif (($mmr_name =~ /SH2_XN_IILB_LB_CMP_ENABLE1/) && ($shadow_pi_read_ce == 1) && ($pi_shub_ce_err == 2)) { 	
	    my $shadow_mmr = 'SH2_PI_CORRECTED_DETAIL_2';
	    $shadow_reg_ptr = reg_create_mmr_inst("shub2_01", $shadow_mmr);
	    reg_set_value($shadow_reg_ptr, $mmr_data);
	    my $shadow_msg = "\n         $mmr_name is shadow MMR for $shadow_mmr";
            $notify = "  <<<<<<<";
            my $mmr_display = lcl_reg_write_value($shadow_reg_ptr, 'all', $prefix, 0, $notify);
            $sh2_output .= "$shadow_msg $mmr_display";
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "$shadow_msg $mmr_display";
	  }
          elsif (($mmr_name =~ /SH2_XN_IILB_II_CMP_EXP_DATA0/) && ($shadow_pi_read_ce == 1) && ($pi_shub_ce_err == 2)) { 	
	    my $shadow_mmr = 'SH2_PI_CORRECTED_DETAIL_3';
	    $shadow_reg_ptr = reg_create_mmr_inst("shub2_01", $shadow_mmr);
            my $reg_data = $mmr_data;
	    reg_set_value($shadow_reg_ptr, $reg_data);
	    my $shadow_msg = "\n         $mmr_name is shadow MMR for $shadow_mmr";
            $notify = "  <<<<<<<";
            my $mmr_display = lcl_reg_write_value($shadow_reg_ptr, 'all', $prefix, 0, $notify);
            $sh2_output .= "$shadow_msg $mmr_display";
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "$shadow_msg $mmr_display";
    	    my $proc_agent = (($sh2_target_list[$sh2_cnt]->{AID_DECODE} & 1) == 1) ? (($mmr_data >> 6) & 1) : (($mmr_data >> 7) & 1);
            my $proc_core =  (($sh2_target_list[$sh2_cnt]->{AID_DECODE} & 1) == 1) ? (($mmr_data >> 7) & 1) : 0 ;
    	    my $proc_thread = (($sh2_target_list[$sh2_cnt]->{AID_DECODE} & 1) == 1) ? (($mmr_data >> 8) & 1) : 0;
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf "\n\t\t\t\tPROC:%d  CORE:%d  THREAD:%d\n\n",
              $proc_agent, $proc_core, $proc_thread;
	  }
          elsif (($mmr_name =~ /SH2_XN_IILB_II_CMP_EXP_DATA1/) && ($mmr_data != 0) && ($shadow_pi_write_ce == 1) && ($shadow_pi_ce == 1)) { 	
	    my $shadow_mmr = 'SH2_PI_CORRECTED_DETAIL_4';
	    $shadow_reg_ptr = reg_create_mmr_inst("shub2_01", $shadow_mmr);
            my $reg_data = $mmr_data;
	    reg_set_value($shadow_reg_ptr, $reg_data);
	    my $shadow_msg = "\n         $mmr_name is shadow MMR for $shadow_mmr";
            $notify = "  <<<<<<<";
            my $mmr_display = lcl_reg_write_value($shadow_reg_ptr, 'all', $prefix, 0, $notify);
            $sh2_output .= "$shadow_msg $mmr_display";
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "$shadow_msg $mmr_display";
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= decode_proc_err($shadow_mmr, $mmr_data, $prefix);
	    $pi_fsb_ce_err = 2;
	    $pi_fsb_ce_fru_entry = $fru_count;
            my $fru_rsn = "PI Reported FSB_SHUB_CE.  ";
	    $sh2_output .= save_fru('SHUB2', $nasid_value, $current_location, $fru_rsn, 'SHADOW_FSB_SHUB_CE');
	  }
          elsif (($mmr_name =~ /SH2_XN_IILB_II_CMP_ENABLE0/) && ($shadow_pi_write_ce == 1) && ($pi_fsb_ce_err == 2)) { 	
	    my $shadow_mmr = 'SH2_PI_CORRECTED_DETAIL_5';
	    $shadow_reg_ptr = reg_create_mmr_inst("shub2_01", $shadow_mmr);
	    reg_set_value($shadow_reg_ptr, $mmr_data);
	    my $shadow_msg = "\n         $mmr_name is shadow MMR for $shadow_mmr";
            $notify = "  <<<<<<<";
            my $mmr_display = lcl_reg_write_value($shadow_reg_ptr, 'all', $prefix, 0, $notify);
            $sh2_output .= "$shadow_msg $mmr_display";
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "$shadow_msg $mmr_display";
	  }
          elsif (($mmr_name =~ /SH2_XN_IILB_II_CMP_ENABLE1/) && ($shadow_pi_write_ce == 1) && ($pi_fsb_ce_err == 2)) { 	
	    my $shadow_mmr = 'SH2_PI_CORRECTED_DETAIL_6';
	    $shadow_reg_ptr = reg_create_mmr_inst("shub2_01", $shadow_mmr);
	    reg_set_value($shadow_reg_ptr, $mmr_data);
	    my $shadow_msg = "\n         $mmr_name is shadow MMR for $shadow_mmr";
            $notify = "  <<<<<<<";
            my $mmr_display = lcl_reg_write_value($shadow_reg_ptr, 'all', $prefix, 0, $notify);
            $sh2_output .= "$shadow_msg $mmr_display";
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "$shadow_msg $mmr_display";
	    my $bad_skt = ($processor_count == 2) ? (reg_get_field_value($shadow_reg_ptr,'CPU')) : 0;
	    $fru_list[$pi_fsb_ce_fru_entry]->{REASON} .= sprintf "\n\t\t\t\t\t\tSuspect %s cpu in SKT %d, Shub or %s sn %s .", 
		$processor_type, $bad_skt, $blade_type, $blade_sn; 
	  }
          elsif (($mmr_name =~ /SH2_XN_NI0_LLP_CMP_ENABLE0/) && ($shadow_pi_ce == 1) && ($mmr_data != 0)) { 	
	    my $shadow_mmr = 'SH2_PI_COR_TIME_STAMP';
	    $shadow_reg_ptr = reg_create_mmr_inst("shub2_01", $shadow_mmr);
	    reg_set_value($shadow_reg_ptr, $mmr_data);
	    my $shadow_msg = "\n         $mmr_name is shadow MMR for $shadow_mmr";
            $notify = "  <<<<<<<";
            my $mmr_display = lcl_reg_write_value($shadow_reg_ptr, 'valid', $prefix, 0, $notify);
            $sh2_output .= "$shadow_msg $mmr_display";
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "$shadow_msg";
	    capture_time_stamp($shadow_reg_ptr);
  	  }

		###############   PI UNCORRECTED SHADOW MMRS  #####################

          elsif (($mmr_name =~ /SH2_XN_NI0_LLP_CMP_ENABLE1/) && ($mmr_data != 0) && ($shadow_pi_read_uce == 1) && ($shadow_pi_uce == 1)) {
	    my $shadow_mmr = 'SH2_PI_UNCORRECTED_DETAIL_1';
	    $shadow_reg_ptr = reg_create_mmr_inst("shub2_01", $shadow_mmr);
            my $reg_data = $mmr_data;
	    reg_set_value($shadow_reg_ptr, $reg_data);
	    my $shadow_msg = "\n         $mmr_name is shadow MMR for $shadow_mmr";
            $notify = "  <<<<<<<";
            my $mmr_display = lcl_reg_write_value($shadow_reg_ptr, 'all', $prefix, 0, $notify);
            $sh2_output .= "$shadow_msg $mmr_display";
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "$shadow_msg $mmr_display";
	    $pi_shub_uc_err = 2;
	    # $shadow_pi_read_uce = 1;
            if ((reg_get_field_value($shadow_reg_ptr,'SYNDROME')) == 0xFF) {
              my $byte_addr = (reg_get_field_value($shadow_reg_ptr,'ADDRESS')) << 3;
              my ($net_addr, $net_nasid) = fsb_to_net($byte_addr & 0x3fffffffffff8);
              if ($net_nasid != $nasid_value) {
                $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .=  sprintf "\n    Poisoned ECC received for Address: 0x%014llX on NASID 0x%04X\n\n", $byte_addr, $net_nasid;
                my $fru_rsn = sprintf "Source node indicated on NASID 0x%04X PI poisoned ECC error.  ", $nasid_value;
	        $sh2_output .= save_fru('SHUB2', $net_nasid, 'UNDEFINED', $fru_rsn, 'SHADOW_SHUB_FSB_UCE_POIS');
              }
              else {
                $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .=  sprintf "\n    Poisoned ECC received for Address: 0x%014llX\n\n", $byte_addr;
              }
            }
            else {
              $sh2_target_list[$sh2_cnt]->{DIMM_ERR} = 1;
              $any_dimm_error++;
	      $global_err_val |= $SHUB_UCE_ERR;
	      $pi_uce_fru_entry = $fru_count;
              my $fru_rsn = sprintf "PI Reported SHUB_FSB_UCE.  ";
	      $sh2_output .= save_fru('SHUB2', $nasid_value, $current_location, $fru_rsn, 'SHADOW_SHUB_FSB_UCE');
              $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= decode_mem_err($shadow_mmr, $mmr_data, $prefix, $blade_type);
	    }
	  }
          elsif (($mmr_name =~ /SH2_XN_NI0_MD_CMP_HDR_ENABLE0/) && ($shadow_pi_read_uce == 1) && ($pi_shub_uc_err == 2)) { 	
	    my $shadow_mmr = 'SH2_PI_UNCORRECTED_DETAIL_2';
	    $shadow_reg_ptr = reg_create_mmr_inst("shub2_01", $shadow_mmr);
	    reg_set_value($shadow_reg_ptr, $mmr_data);
	    my $shadow_msg = "\n         $mmr_name is shadow MMR for $shadow_mmr";
            $notify = "  <<<<<<<";
            my $mmr_display = lcl_reg_write_value($shadow_reg_ptr, 'all', $prefix, 0, $notify);
            $sh2_output .= "$shadow_msg $mmr_display";
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "$shadow_msg $mmr_display";
	  }
          elsif (($mmr_name =~ /SH2_XN_NI0_MD_CMP_HDR_ENABLE1/) && ($shadow_pi_read_uce == 1) && ($pi_shub_uc_err == 2)) { 	
	    my $shadow_mmr = 'SH2_PI_UNCORRECTED_DETAIL_3';
	    $shadow_reg_ptr = reg_create_mmr_inst("shub2_01", $shadow_mmr);
            my $reg_data = $mmr_data;
	    reg_set_value($shadow_reg_ptr, $reg_data);
	    my $shadow_msg = "\n         $mmr_name is shadow MMR for $shadow_mmr";
            $notify = "  <<<<<<<";
            my $mmr_display = lcl_reg_write_value($shadow_reg_ptr, 'all', $prefix, 0, $notify);
            $sh2_output .= "$shadow_msg $mmr_display";
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "$shadow_msg $mmr_display";
    	    my $proc_agent = (($sh2_target_list[$sh2_cnt]->{AID_DECODE} & 1) == 1) ? (($mmr_data >> 6) & 1) : (($mmr_data >> 7) & 1);
            my $proc_core =  (($sh2_target_list[$sh2_cnt]->{AID_DECODE} & 1) == 1) ? (($mmr_data >> 7) & 1) : 0 ;
    	    my $proc_thread = (($sh2_target_list[$sh2_cnt]->{AID_DECODE} & 1) == 1) ? (($mmr_data >> 8) & 1) : 0;
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= sprintf "\n\t\t\t\tPROC:%d  CORE:%d  THREAD:%d\n\n",
              $proc_agent, $proc_core, $proc_thread;
	  }
          elsif (($mmr_name =~ /SH2_XN_NI1_LLP_CMP_EXP_DATA0/) && ($mmr_data != 0) && ($shadow_pi_write_uce == 1) && ($shadow_pi_uce == 1)) { 	
	    my $shadow_mmr = 'SH2_PI_UNCORRECTED_DETAIL_4';
	    $shadow_reg_ptr = reg_create_mmr_inst("shub2_01", $shadow_mmr);
            my $reg_data = $mmr_data;
	    reg_set_value($shadow_reg_ptr, $reg_data);
	    my $shadow_msg = "\n         $mmr_name is shadow MMR for $shadow_mmr";
            $notify = "  <<<<<<<";
            my $mmr_display = lcl_reg_write_value($shadow_reg_ptr, 'all', $prefix, 0, $notify);
            $sh2_output .= "$shadow_msg $mmr_display";
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "$shadow_msg $mmr_display";
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= decode_proc_err($shadow_mmr, $mmr_data, $prefix);
	    $pi_fsb_uc_err = 2;
	    $pi_fsb_uce_poisoned = ((reg_get_field_value($shadow_reg_ptr,'SYNDROME')) == 0x3) ? 1 : 0;
	    $pi_fsb_uce_fru_entry = $fru_count;
            my $fru_rsn = sprintf "PI Reported FSB_SHUB_UCE.  ";
	    $sh2_output .= save_fru('SHUB2', $nasid_value, $current_location, $fru_rsn, 'SHADOW_FSB_SHUB_UCE');
	  }
          elsif (($mmr_name =~ /SH2_XN_NI1_LLP_CMP_EXP_DATA1/) && ($shadow_pi_write_uce == 1) && ($pi_fsb_uc_err == 2)) { 	
	    my $shadow_mmr = 'SH2_PI_UNCORRECTED_DETAIL_5';
	    $shadow_reg_ptr = reg_create_mmr_inst("shub2_01", $shadow_mmr);
	    reg_set_value($shadow_reg_ptr, $mmr_data);
	    my $shadow_msg = "\n         $mmr_name is shadow MMR for $shadow_mmr";
            $notify = "  <<<<<<<";
            my $mmr_display = lcl_reg_write_value($shadow_reg_ptr, 'all', $prefix, 0, $notify);
            $sh2_output .= "$shadow_msg $mmr_display";
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "$shadow_msg $mmr_display";
	  }
          elsif (($mmr_name =~ /SH2_XN_NI1_LLP_CMP_ENABLE0/) && ($shadow_pi_write_uce == 1) && ($pi_fsb_uc_err == 2)) { 	
	    my $shadow_mmr = 'SH2_PI_UNCORRECTED_DETAIL_6';
	    $shadow_reg_ptr = reg_create_mmr_inst("shub2_01", $shadow_mmr);
	    reg_set_value($shadow_reg_ptr, $mmr_data);
	    my $shadow_msg = "\n         $mmr_name is shadow MMR for $shadow_mmr";
            $notify = "  <<<<<<<";
            my $mmr_display = lcl_reg_write_value($shadow_reg_ptr, 'all', $prefix, 0, $notify);
            $sh2_output .= "$shadow_msg $mmr_display";
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "$shadow_msg $mmr_display";
	    my $bad_skt = ($processor_count == 2) ? (reg_get_field_value($shadow_reg_ptr,'CPU')) : 0;
	    my $cpu_string = ($first_pi_uce == 4) ? 'CPU in SKT' : 'cpu in SKT';
	      if ($pi_fsb_uce_poisoned == 1) {
	        $fru_list[$pi_fsb_uce_fru_entry]->{REASON} .= sprintf "\n\t\t\t\t\t\t%s %s %d wrote back previously poisoned data.", 
		  $processor_type, $cpu_string, $bad_skt, $blade_type, $blade_sn; 
	      }
	      else {
	        $fru_list[$pi_fsb_uce_fru_entry]->{REASON} .= sprintf "\n\t\t\t\t\t\tSuspect %s %s %d, Shub or %s sn %s.", 
		    $processor_type, $cpu_string, $bad_skt, $blade_type, $blade_sn; 
	      }
	  }
          elsif (($mmr_name =~ /SH2_XN_NI1_LLP_CMP_ENABLE1/) && ($shadow_pi_uce == 1) && ($mmr_data != 0)) { 	
	    my $shadow_mmr = 'SH2_PI_UNCOR_TIME_STAMP';
	    $shadow_reg_ptr = reg_create_mmr_inst("shub2_01", $shadow_mmr);
	    reg_set_value($shadow_reg_ptr, $mmr_data);
	    my $shadow_msg = "\n         $mmr_name is shadow MMR for $shadow_mmr";
            $notify = "  <<<<<<<";
            my $mmr_display = lcl_reg_write_value($shadow_reg_ptr, 'valid', $prefix, 0, $notify);
            $sh2_output .= "$shadow_msg $mmr_display";
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "$shadow_msg";
	    capture_time_stamp($shadow_reg_ptr);
  	  }
		###############   XN CORRECTED SHADOW MMRS  #####################

          elsif (($mmr_name =~ /SH2_XN_NI1_MD_CMP_EXP_HDR0/) && ($mmr_data != 0) && ($shadow_xn_ce == 1)) { 	
	    my $shadow_mmr = 'SH2_XN_CORRECTED_DETAIL_1';
	    $shadow_reg_ptr = reg_create_mmr_inst("shub2_01", $shadow_mmr);
            my $reg_data = $mmr_data;
	    reg_set_value($shadow_reg_ptr, $reg_data);
	    my $shadow_msg = "\n         $mmr_name is shadow MMR for $shadow_mmr";
            $notify = "  <<<<<<<";
            my $mmr_display = lcl_reg_write_value($shadow_reg_ptr, 'all', $prefix, 0, $notify);
            $sh2_output .= "$shadow_msg $mmr_display";
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "$shadow_msg $mmr_display";
            $sh2_target_list[$sh2_cnt]->{DIMM_ERR} = 1;
            $any_dimm_error++;
	    $xn_md_ce = 2;
	    $xn_ce_fru_entry = $fru_count;
            my $fru_rsn = sprintf "XN Reported SBE ECC Error.  ";
	    $sh2_output .= save_fru('SHUB2', $nasid_value, $current_location, $fru_rsn, 'SHADOW_XN_CE');
	    $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= decode_mem_err($shadow_mmr, $mmr_data, $prefix, $blade_type);
	  }
          elsif (($mmr_name =~ /SH2_XN_NI1_MD_CMP_EXP_HDR1/) && ($shadow_xn_ce == 1) && ($xn_md_ce == 2)) { 	
	    my $shadow_mmr = 'SH2_XN_CORRECTED_DETAIL_2';
	    $shadow_reg_ptr = reg_create_mmr_inst("shub2_01", $shadow_mmr);
	    reg_set_value($shadow_reg_ptr, $mmr_data);
	    my $shadow_msg = "\n         $mmr_name is shadow MMR for $shadow_mmr";
            $notify = "  <<<<<<<";
            my $mmr_display = lcl_reg_write_value($shadow_reg_ptr, 'valid', $prefix, 0, $notify);
            $sh2_output .= "$shadow_msg $mmr_display";
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "$shadow_msg $mmr_display";
	  }
          elsif (($mmr_name =~ /SH2_XN_NI1_MD_CMP_HDR_ENABLE0/) && ($shadow_xn_ce == 1) && ($xn_md_ce == 2)) { 	
	    my $shadow_mmr = 'SH2_XN_CORRECTED_DETAIL_3';
	    $shadow_reg_ptr = reg_create_mmr_inst("shub2_01", $shadow_mmr);
            my $reg_data = $mmr_data;
	    reg_set_value($shadow_reg_ptr, $reg_data);
	    my $shadow_msg = "\n         $mmr_name is shadow MMR for $shadow_mmr";
            $notify = "  <<<<<<<";
            my $mmr_display = lcl_reg_write_value($shadow_reg_ptr, 'valid', $prefix, 0, $notify);
            $sh2_output .= "$shadow_msg $mmr_display";
	    $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= decode_mem_err($shadow_mmr, $mmr_data, $prefix, $blade_type);
	  }
          elsif (($mmr_name =~ /SH2_XN_NI1_MD_CMP_HDR_ENABLE1/) && ($shadow_xn_ce == 1) && ($xn_md_ce == 2)) { 	
	    my $shadow_mmr = 'SH2_XN_CORRECTED_DETAIL_4';
	    $shadow_reg_ptr = reg_create_mmr_inst("shub2_01", $shadow_mmr);
            my $reg_data = $mmr_data;
	    reg_set_value($shadow_reg_ptr, $reg_data);
	    my $shadow_msg = "\n         $mmr_name is shadow MMR for $shadow_mmr";
            $notify = "  <<<<<<<";
            my $mmr_display = lcl_reg_write_value($shadow_reg_ptr, 'valid', $prefix, 0, $notify);
            $sh2_output .= "$shadow_msg $mmr_display";
	    $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= decode_mem_err($shadow_mmr, $mmr_data, $prefix, $blade_type);
	  }
          elsif (($mmr_name =~ /SH2_XN_PI_IILB_CMP_EXP_DATA0/) && ($shadow_xn_ce == 1) && ($mmr_data != 0)) { 	
	    my $shadow_mmr = 'SH2_XN_COR_TIME_STAMP';
	    $shadow_reg_ptr = reg_create_mmr_inst("shub2_01", $shadow_mmr);
	    reg_set_value($shadow_reg_ptr, $mmr_data);
	    my $shadow_msg = "\n         $mmr_name is shadow MMR for $shadow_mmr";
            $notify = "  <<<<<<<";
            my $mmr_display = lcl_reg_write_value($shadow_reg_ptr, 'valid', $prefix, 0, $notify);
            $sh2_output .= "$shadow_msg $mmr_display";
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "$shadow_msg";
	    capture_time_stamp($shadow_reg_ptr);
  	  }

		###############   XN UNCORRECTED SHADOW MMRS  #####################

          elsif (($mmr_name =~ /SH2_XN_PI_IILB_CMP_EXP_DATA1/) && ($mmr_data != 0) && ($shadow_xn_uce == 1)) {
            my $shadow_mmr = 'SH2_XN_UNCORRECTED_DETAIL_1';
            $shadow_reg_ptr = reg_create_mmr_inst("shub2_01", $shadow_mmr);
            my $reg_data = $mmr_data;
            reg_set_value($shadow_reg_ptr, $reg_data);
	    my $shadow_msg = "\n         $mmr_name is shadow MMR for $shadow_mmr";
            $notify = "  <<<<<<<";
            my $mmr_display = lcl_reg_write_value($shadow_reg_ptr, 'all', $prefix, 0, $notify);
            $sh2_output .= "$shadow_msg $mmr_display";
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "$shadow_msg $mmr_display";
            $sh2_target_list[$sh2_cnt]->{DIMM_ERR} = 1;
            $any_dimm_error++;
            $xn_md_uc = 2;
            $xn_uce_fru_entry = $fru_count;
            my $fru_rsn = sprintf "XN Reported UCE ECC Error.  ";
	    $sh2_output .= save_fru('SHUB2', $nasid_value, $current_location, $fru_rsn, 'SHADOW_XN_UCE');
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .=  decode_mem_err($shadow_mmr, $mmr_data, $prefix, $blade_type);
          }
          elsif (($mmr_name =~ /SH2_XN_PI_IILB_CMP_ENABLE0/) && ($shadow_xn_uce == 1) && ($xn_md_uc == 2)) { 	
	    my $shadow_mmr = 'SH2_XN_UNCORRECTED_DETAIL_2';
	    $shadow_reg_ptr = reg_create_mmr_inst("shub2_01", $shadow_mmr);
	    reg_set_value($shadow_reg_ptr, $mmr_data);
	    my $shadow_msg = "\n         $mmr_name is shadow MMR for $shadow_mmr";
            $notify = "  <<<<<<<";
            my $mmr_display = lcl_reg_write_value($shadow_reg_ptr, 'valid', $prefix, 0, $notify);
            $sh2_output .= "$shadow_msg $mmr_display";
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "$shadow_msg $mmr_display";
	  }
          elsif (($mmr_name =~ /SH2_XN_PI_IILB_CMP_ENABLE1/) && ($shadow_xn_uce == 1) && ($xn_md_uc == 2)) { 	
	    my $shadow_mmr = 'SH2_XN_UNCORRECTED_DETAIL_3';
	    $shadow_reg_ptr = reg_create_mmr_inst("shub2_01", $shadow_mmr);
            my $reg_data = $mmr_data;
	    reg_set_value($shadow_reg_ptr, $reg_data);
	    my $shadow_msg = "\n         $mmr_name is shadow MMR for $shadow_mmr";
            $notify = "  <<<<<<<";
            my $mmr_display = lcl_reg_write_value($shadow_reg_ptr, 'valid', $prefix, 0, $notify);
            $sh2_output .= "$shadow_msg $mmr_display";
	    $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= decode_mem_err($shadow_mmr, $mmr_data, $prefix, $blade_type);
	  }
          elsif (($mmr_name =~ /SH2_XN_PI_NI0_CMP_EXP_DATA0/) && ($shadow_xn_uce == 1) && ($xn_md_uc == 2)) { 	
	    my $shadow_mmr = 'SH2_XN_UNCORRECTED_DETAIL_4';
	    $shadow_reg_ptr = reg_create_mmr_inst("shub2_01", $shadow_mmr);
            my $reg_data = $mmr_data;
	    reg_set_value($shadow_reg_ptr, $reg_data);
	    my $shadow_msg = "\n         $mmr_name is shadow MMR for $shadow_mmr";
            $notify = "  <<<<<<<";
            my $mmr_display = lcl_reg_write_value($shadow_reg_ptr, 'valid', $prefix, 0, $notify);
            $sh2_output .= "$shadow_msg $mmr_display";
	    $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= decode_mem_err($shadow_mmr, $mmr_data, $prefix, $blade_type);
	  }
          elsif (($mmr_name =~ /SH2_XN_PI_NI0_CMP_EXP_DATA1/) && ($shadow_xn_uce == 1) && ($mmr_data != 0)) { 	
	    my $shadow_mmr = 'SH2_XN_UNCOR_TIME_STAMP';
	    $shadow_reg_ptr = reg_create_mmr_inst("shub2_01", $shadow_mmr);
	    reg_set_value($shadow_reg_ptr, $mmr_data);
	    my $shadow_msg = "\n         $mmr_name is shadow MMR for $shadow_mmr";
            $notify = "  <<<<<<<";
            my $mmr_display = lcl_reg_write_value($shadow_reg_ptr, 'valid', $prefix, 0, $notify);
            $sh2_output .= "$shadow_msg $mmr_display";
            $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "$shadow_msg";
	    capture_time_stamp($shadow_reg_ptr);
  	  }
	   ############  End of ECC SHADOW MMR Parsing  ############

       ########  Fall through case - no errors or credits that could be reported in a error summary
          else {
            $sh2_output .= lcl_reg_write_value($reg_ptr, 'valid', $prefix, 0, $notify);
          }
  
        }
	  $reg_idx++;	#  continue to next MMR
      }			#  MMR Parsing ends here

      if ($bad_read_cnt > 0) {
	print "\n\nALERT:  Found $bad_read_cnt suspect reads out of $#sh2_reg_list mmrs.\n\n";	#  GEM  Debug
      }
      print "$sh2_output\n\n\n";          #  completed parsing for the current chip

      if ($sh2_target_list[$sh2_cnt]->{PI_DLOCK} == 1) {	# re-read if initial read saw 0 PI->XN reply credits
        if (pi_deadlock_chk("shub2_01", 1) == 1) {
          $pi_vc_deadlock_list[$dl_cnt]->{NASID} = $sh2_target_list[$sh2_cnt]->{NASID};
          $pi_vc_deadlock_list[$dl_cnt]->{LOCATION} = $current_location;
          $sh2_target_list[$sh2_cnt]->{PI_DLOCK_MSG} = sprintf "%s: Confirmed Reply VC Deadlocked PI.", $prefix;
          $dl_cnt++;
        }
        else  {
          $sh2_target_list[$sh2_cnt]->{PI_DLOCK} = 0;
        }
      }
      
      if (($same_data_count + 1) == $reg_idx) {
	$sh2_target_list[$sh2_cnt]->{ERR_SUMM} = "$err_summ_msg"; 
	$sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "\n\n     Duplicate Data found in all MMR's on this node.\n"; 
	$sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "     Error report for this node will not be generated due to corrupted data from JTAG.\n"; 
        $any_sh2_error = 1;
        $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
	$global_err_val |= $SHUB_HW_ERR;
      }
      
      if ($bad_read_cnt == $reg_idx) {
	$sh2_target_list[$sh2_cnt]->{ERR_SUMM} = "$err_summ_msg"; 
	$sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "\n\n     Invalid Data \(address values\) found in all MMR's on this node.\n"; 
	$sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "     Error report for this node will not be generated due to corrupted data from JTAG.\n"; 
        $any_sh2_error = 1;
        $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
	$global_err_val |= $SHUB_HW_ERR;
        my $fru_rsn = sprintf "Fatal unknown error causing JTAG reads to fail on this blade.  ";
	my $lcl_output = save_fru('SHUB2', $nasid_value, $current_location, $fru_rsn, 'FATAL_JTAG');
	printf "$lcl_output";
      }
      
      if ((($same_data_count + 1) == $reg_idx) || ($bad_read_cnt > 10)) {
	print "\n\nALERT:   Not continuing to extract Shub state due to corrupted data from JTAG.\n\n";	
      }
      elsif ($file_mode == 0) {			#  if we're using real HW
	my $cam_status = 0;
        if (($fast_mode == 0) && ($abbrev_mode == 0)) { 
          if ($show_pi_state == 1) {		#  global variable from errdmp main
	    notify_smdk(1);
            my $tp_chk_val = capture_sh2_pi_dbg_00();	
            if (($tp_chk_val & 2) == 2) {
	      $global_err_val |= $SHUB_HW_ERR;
              my $fru_rsn = sprintf "PI is Hung due to Sustained snoop conflict.  ";
	      my $lcl_output = save_fru('SHUB2', $nasid_value, $current_location, $fru_rsn, 'PI_SNOOP_CONFLICT');
	      printf "$lcl_output";
	    }
	    if ($flush_cfg_err == 0) {
              ($fsb_stalled, $pi_crb_skid, $pi_wr_cmd_q_af) = capture_sh2_pi_dbg_01();
	    }
            if (($fsb_stalled == 1) && ($processor_count > 0)) {
	      $global_err_val |= $SHUB_HW_ERR;
	      $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "\n    Shub PI testpoints indicate BPRI is continually asserted on the Front-side Bus.\n"; 
	      $sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "    This will prevent the processors on this node from issuing any requests.\n"; 
              $any_sh2_error = 1;
  
              $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
              my $fru_rsn = sprintf "Shub PI is continually asserting BPRI on the FSB.  ";
	      my $lcl_output = save_fru('SHUB2', $nasid_value, $current_location, $fru_rsn, 'PI_FSB_BPRI');
	      printf "$lcl_output";
	    }
	    notify_smdk(0);
          }
	  if ($show_crb_state == 1) {		#  global variable from errdmp main
            $cam_status = read_all_crb("shub2_01");
	  }
        }
        if ((($pi_to_err == 1) || ($pi_to_err_s == 1)) && ($fast_mode == 1))  {
	  $show_rrb_state = 1;
	  $show_wrb_state = 1;
	  $show_irb_state = 1;
	  $show_prb_state = 1;
          my $cam_status = read_all_crb("shub2_01");
	  $show_rrb_state = 0;
	  $show_wrb_state = 0;
	  $show_irb_state = 0;
	  $show_prb_state = 0;
        }
	if (($cam_status == 1) || ($low_pi_rq_crd == 1)) {
	  my $fru_msg = (($cam_status == 1) && ($low_pi_rq_crd == 1)) ? "Requests into PI are blocked and CRB Cam reads are failing." :
		        (($cam_status == 0) && ($low_pi_rq_crd == 1)) ? "Requests into PI are blocked." : "CRB Cam reads are failing.";
          my $fru_rsn = sprintf "PI is Hung - %s  ", $fru_msg;
	  my $lcl_output = save_fru('SHUB2', $nasid_value, $current_location, $fru_rsn, 'PI_FSB_XB_HUNG');
	  printf "$lcl_output";
          $any_sh2_error = 1;
          $sh2_target_list[$sh2_cnt]->{ERROR} = 1;
	  $global_err_val |= $SHUB_HW_ERR;
	}

        if (  ($low_md_rq_crd ==1) 
          || ($low_pi_rp_crd == 1) 
          || ($md_cmd_is_prge == 1) 
          || ($xn_cmd_is_prge == 2) 
          || ($ptc_pwerr > 0) 
          || ($md_cmd_is_xfer == 1) 
          || (($low_pi_rp_crd == 1) && ($pi_wr_cmd_q_af == 1))
          || (($fsb_stalled == 1) && ($pi_crb_skid == 0) && ($processor_count > 0)) ) {
	  my $sig_msg = "\($current_location\) : ";
          my $tab_val = "\n\t\t\t\t\t\t   ";
          my $first_one = 1;
	  if ($low_md_rq_crd == 1)  {   $sig_msg .= "BW 938724: MDXN request credits stuck.\n";  $first_one = 0; }
	  if (($low_pi_rp_crd == 1) && ($pi_wr_cmd_q_af == 0))  
				    {   if ($first_one == 0) {  $sig_msg .= $tab_val; }
				        $sig_msg .= "BW 940739: PIXN reply credits stuck.  ";
				        $first_one = 0; 
				    }
	  if (($low_pi_rp_crd == 1) && ($pi_wr_cmd_q_af == 1))  
				    {   if ($first_one == 0) {  $sig_msg .= $tab_val; }
				        $sig_msg .= "BW 945998: PI RDAT dat_dfr_q overflowing.  ";
				        $first_one = 0; 
				    }
	  if ($md_cmd_is_prge == 1) {   if ($first_one == 0) {  $sig_msg .= $tab_val; }
					$sig_msg .= "BW 939024: PI Purge Issue. \(MD PERR response\) ";
					$first_one = 0; 
				    }
	  if ($xn_cmd_is_prge == 2) {   if ($first_one == 0) {  $sig_msg .= $tab_val; }
					$sig_msg .= "BW 939024: PI Purge Issue. \(XN_IILB Chiplet Nomatch\) ";
					$first_one = 0; 
				    }
	  if ($ptc_pwerr > 0)       {   if ($first_one == 0) {  $sig_msg .= $tab_val; }
					$sig_msg .= "BW 947466: LB Times out PTC write.  ";
					$first_one = 0; 
				    }
	  if ($md_cmd_is_xfer == 1) {   if ($first_one == 0) {  $sig_msg .= $tab_val; }
				        $sig_msg .= "BW 939692: PI XFER Issue. \(MD PERR response\) ";
					$first_one = 0; 
				    }
	  if (($fsb_stalled == 1) && ($pi_crb_skid == 0) && ($processor_count > 0))    {   
					if ($first_one == 0) {  $sig_msg .= $tab_val; }
					$sig_msg .= "BW 951089: Hung CPUs due to Shub asserting BPRI. ";
					$first_one = 0; 
				    }
  
          $sh2_target_list[$sh2_cnt]->{ERR_SIG} = $sig_msg;

        }
        if ($dump_luts == 1)  {  dump_valid_lut('SHUB2', $nasid_value);  }
      }

      if ($bw_963416_attr == 4) {   
	my $sig_msg = "\($current_location\) : ";
        my $tab_val = "\n\t\t\t\t\t\t   ";
	if (defined $sh2_target_list[$sh2_cnt]->{ERR_SIG}) {  
	  $sig_msg .= $tab_val; 
	}
	$sig_msg .= "BW 963416: Junk Bus Sequence Error during boot.  ";
        $sh2_target_list[$sh2_cnt]->{ERR_SIG} = $sig_msg;
      }

      if ($pi_numa_error == 1) {		#  PI detected message length error
	my $ev_od = ((($pi_numa_err_addr >> 7) & 1) == 1) ? "Odd" : "Even"; 
	my $bool_ev_od = ((($pi_numa_err_addr >> 7) & 1) == 1) ? 0 : 1 ; 
	my $cl_ev_od = ((($pi_numa_err_addr >> 7) & 1) == 1) ? "-odd_cache_lines" : "-even_cache_lines" ; 
	my $rtd_msg = sprintf "\n  smdk command:  run rtrdmp -t R -n N0x%X,N0x%X      \t\t#   %s cacheline failed",
	    $pi_numa_err_src_node, $nasid_value, $ev_od;
        if ( $l2_host eq "" ) {
	  $l2_host = "[L2 host ip address]";
	}
	$rtd_msg .= sprintf "\n    L3 command:  rtrdmp --l2 %s -src_nasid 0x%X -dst_nasid 0x%X %s  ",
	    $l2_host, $pi_numa_err_src_node, $nasid_value, $cl_ev_od;
	$sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "\n\n Possible Numa-Link error corruption detected by the PI on this node.";
	$sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "\n Run the following command line within smdk to determine the routes used.";
	$sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "    $rtd_msg\n\n";

	$rtd_list[$rtd_cnt]->{RTD_MSG} = "$rtd_msg";
	$rtd_list[$rtd_cnt]->{SRC_NASID} = $pi_numa_err_src_node;
	$rtd_list[$rtd_cnt]->{DST_NASID} = $nasid_value;
	$rtd_list[$rtd_cnt]->{CHIPLET}   = "md";
	$rtd_list[$rtd_cnt]->{EVEN_ODD}  = $bool_ev_od;
	$rtd_list[$rtd_cnt]->{REQUEST}   = 0;	        	#  failing packet would most likely have been a response
	$rtd_cnt++;
      }

      if ($md_numa_error == 1) {		#  MD detected message length error
	my $ev_od = ((($md_numa_err_addr >> 7) & 1) == 1) ? "Odd" : "Even"; 
	my $bool_ev_od = ((($md_numa_err_addr >> 7) & 1) == 1) ? 0 : 1 ; 
	my $cl_ev_od = ((($md_numa_err_addr >> 7) & 1) == 1) ? "-odd_cache_lines" : "-even_cache_lines" ; 
	my $rtd_msg = sprintf "\n  smdk command:  run rtrdmp -t R -n N0x%X,N0x%X      \t\t#   %s cacheline failed",
	    $md_numa_err_src_node, $nasid_value, $ev_od;
        if ( $l2_host eq "" ) {
	  $l2_host = "[L2 host ip address]";
	}
	$rtd_msg .= sprintf "\n    L3 command:  rtrdmp --l2 %s -src_nasid 0x%X -dst_nasid 0x%X %s  ",
	    $l2_host, $md_numa_err_src_node, $nasid_value, $cl_ev_od;
	$sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "\n\n Possible Numa-Link error corruption detected by the MD on this node.";
	$sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "\n Run the following command line within smdk to determine the routes used.";
	$sh2_target_list[$sh2_cnt]->{ERR_SUMM} .= "    $rtd_msg\n\n";

	$rtd_list[$rtd_cnt]->{RTD_MSG} = "$rtd_msg";
        $rtd_list[$rtd_cnt]->{SRC_NASID} = $md_numa_err_src_node;
        $rtd_list[$rtd_cnt]->{DST_NASID} = $nasid_value;
        $rtd_list[$rtd_cnt]->{CHIPLET}   = $md_numa_err_src_chip;
        $rtd_list[$rtd_cnt]->{EVEN_ODD}  = $bool_ev_od;
        $rtd_list[$rtd_cnt]->{REQUEST}   = 1;	        	#  failing packet would most likely have been a request
	$rtd_cnt++;
      }

}

#----------------------------------------------------------------
#
# This function writes the supplied value to the SH2_PIO_TIME_OUT mmr.  This
# is being done to work around the bug descibed in :
#    953156: SH2:LB Conflict between jtag & processor mmr reads can crash system.
#
#----------------------------------------------------------------

sub set_pio_timeout {

  my $write_val = $_[0];
  my $to_reg = reg_create_mmr_inst('shub2_01', 'SH2_PIO_TIME_OUT');
  reg_set_field_value($to_reg, 'VALUE', $write_val);
 
  my $xfr_complete = 0;
  my $xfr_f_cnt = 0;
  while (($xfr_complete == 0) && ($xfr_f_cnt < 3) && ($sh2_target_list[$sh2_cnt]->{XFR_ERR_CNT} < 6))  {
    $SSCOPE_ERR = 0;	
    reg_mmr_write($to_reg);
    if ($SSCOPE_ERR == 1)  {
      $xfr_f_cnt++;
      $sh2_target_list[$sh2_cnt]->{XFR_ERR_CNT}++;
      $SSCOPE_ERR = 0;
      print "\n Error detected after reg_mmr_write() attempt in set_pio_timeout()\n";
    }
    else {
      $xfr_complete = 1;
    }
  }
}

#---------------------------------------------------------------------
#
#  Subroutine to determine the backplane connection between the supplied
#  blade location and NI value and the tollhouse chassis router.
#
#  usage:    get_bp_connect($sh2_target_list[$sh2_cnt]->{NODE}, $sh2_target_list[$sh2_cnt]->{CHASSIS}, 1);
#
#---------------------------------------------------------------------
sub get_bp_connect {

  my $blade_loc = "$_[0]";
  my $bp_type = $_[1];			# backplane type  1=1955    2=191010 
  my $ni_port = $_[2];			# 0=NI0    1=NI1

  my $blade_index = $blade_loc;
  $blade_index =~ s/BLADE//;
  my $blade_offset = ($bp_type == 1) ? 5 : 10;
  if ($ni_port == 1) {
     $blade_index += $blade_offset;
  }
  
  my @FIVE_LIST = (
 	"RTR_L:0, Port F",		# NI0
 	"RTR_L:0, Port E",
 	"RTR_L:0, Port D",
 	"RTR_L:0, Port C",
 	"RTR_L:0, Port B",
 	"RTR_R:0, Port B",		# NI1
 	"RTR_R:0, Port C",
 	"RTR_R:0, Port D",
 	"RTR_R:0, Port E",
 	"RTR_R:0, Port F"
  );
  
  my @TEN_LIST = (
 	"RTR_L:0, Port F",		# NI0
 	"RTR_L:0, Port E",
 	"RTR_L:0, Port D",
 	"RTR_L:0, Port C",
 	"RTR_L:0, Port B",
 	"RTR_L:1, Port F",
 	"RTR_L:1, Port E",
 	"RTR_L:1, Port D",
 	"RTR_L:1, Port C",
 	"RTR_L:1, Port B",
 	"RTR_R:1, Port B",		# NI1
 	"RTR_R:1, Port C",
 	"RTR_R:1, Port D",
 	"RTR_R:1, Port E",
 	"RTR_R:1, Port F",
 	"RTR_R:0, Port B",
 	"RTR_R:0, Port C",
 	"RTR_R:0, Port D",
 	"RTR_R:0, Port E",
 	"RTR_R:0, Port F"
  );
  
  my $conn_name = ($bp_type == 1) ? sprintf "Connected to $FIVE_LIST[$blade_index] via 1955 backplane.":
  		  ($bp_type == 2) ? sprintf "Connected to $TEN_LIST[$blade_index] via 191010 backplane.":
		  'Connected to external NI port. ';
	        
  return($conn_name);

}

#---------------------------------------------------------------------
#  Subroutine to monitor all shub2 nodes for BW# 951089
#
#
#---------------------------------------------------------------------

sub monitor_for_BW951089 {

  autoflush STDOUT 1; 		#  GEM  - debug
  my $loop_cnt = 0;
  my $err_cnt = 0;

  print "\n Beginning Shub2 PI monitor activity to detect PI Nack Hammer overflow events. \n";

  while ($monitor_mode == 1) {
    $sh2_cnt = 0;
    while (defined $sh2_target_list[$sh2_cnt]) {

      $SSCOPE_ERR = 0;
      mmr_target($SELECT, $sh2_target_list[$sh2_cnt]->{MODULE},
                      $sh2_target_list[$sh2_cnt]->{NODE},
                      $sh2_target_list[$sh2_cnt]->{CHIP});

      if ($SSCOPE_ERR == 1)  {
        printf "\n mmr_target() failed in monitor_for_BW951089().  Aborting Shub2 Monitor for %s of %s\n\n",
          $sh2_target_list[$sh2_cnt]->{NODE}, $sh2_target_list[$sh2_cnt]->{MODULE};
        $SSCOPE_ERR = 0;
        $sh2_cnt++;
        next;
      }

      my $fsb_stalled = 1;
      my $pi_crb_skid = 0;
      my $retry_cnt = 0;
      
      while (($retry_cnt < 9) && ($fsb_stalled == 1)) {

        ($fsb_stalled, $pi_crb_skid) = capture_sh2_pi_dbg_01();	
        if ($fsb_stalled == 1) {
          chomp ($fail_time = `date +'%T'`);
	  printf "\n ALERT!! Detected stalled FSB on %s of %s at %s - clearing nack hammer -  retry count: %d \n",  
	    $sh2_target_list[$sh2_cnt]->{NODE}, $sh2_target_list[$sh2_cnt]->{MODULE}, $fail_time, $retry_cnt;
	  clear_nack_hammer();
	  $err_cnt++;
	  $retry_cnt++;
        }
      }

      if ($retry_cnt > 0) {
        my $result_msg = ($fsb_stalled == 0) ? "Successfully" : "Unsuccessfully";
	printf "    %s cleared PI Nack Hammer overflow after %d attempts.\n", $result_msg, $retry_cnt;
	printf "    Total number of errors detected since monitor was started:  %d.\n", $err_cnt;
      }

      $sh2_cnt++;
    }
    if (($loop_cnt % 20) == 0) { 
      chomp ($done_time = `date +'%T'`);
      printf "\n  Completed loop # %d at %s  ", ($loop_cnt + 1), $done_time;
    }
    else {
      printf ".";
    }
    $loop_cnt++;
  }

  print "\n\n";
}


sub clear_nack_hammer {

  my $reg_ptr = reg_create_mmr_inst("shub2_01", "SH2_CRB_MESSAGE_CONTROL");
  
  my $xfr_complete = 0;
  my $xfr_f_cnt = 0;
  while (($xfr_complete == 0) && ($xfr_f_cnt < 3) && ($sh2_target_list[$sh2_cnt]->{XFR_ERR_CNT} < 6))  {
    $SSCOPE_ERR = 0;
    reg_mmr_read($reg_ptr);
    if ($SSCOPE_ERR == 1)  {
      $xfr_f_cnt++;
      $sh2_target_list[$sh2_cnt]->{XFR_ERR_CNT}++;
      $SSCOPE_ERR = 0;
      print "\n Error detected after reg_mmr_read() attempt in clear_nack_hammer()\n";
    }
    else {
      $xfr_complete = 1;
    }
  }

    #  first clear the nack hammers by writing a zero, then re-enable

  for (my $write_val == 0; $write_val < 2; $write_val++) {
    reg_set_field_value($reg_ptr, 'ENABLE_XB_PIPE_NACK_HAMMER', $write_val);
    reg_set_field_value($reg_ptr, 'ENABLE_FSB_NACK_HAMMER', $write_val);

    $xfr_complete = 0;
    $xfr_f_cnt = 0;
    while (($xfr_complete == 0) && ($xfr_f_cnt < 3) && ($sh2_target_list[$sh2_cnt]->{XFR_ERR_CNT} < 6))  {
      $SSCOPE_ERR = 0;
      reg_mmr_write($reg_ptr);
      if ($SSCOPE_ERR == 1)  {
        $xfr_f_cnt++;
        $sh2_target_list[$sh2_cnt]->{XFR_ERR_CNT}++;
        $SSCOPE_ERR = 0;
        print "\n Error detected after reg_mmr_read() attempt in clear_nack_hammer()\n";
      }
      else {
        $xfr_complete = 1;
      }
    }

    if ($write_val == 0) {
      my $cam_status = quick_crb_read();
    }
  }
  
}

sub quick_crb_read {

  print "Reading PI CRBs..\n";

  $cam_address_inst = reg_create_mmr_inst("shub2_01", "SH2_PI_CAM_ADDRESS_READ_DATA");
  $cam_lpra_inst = reg_create_mmr_inst("shub2_01", "SH2_PI_CAM_LPRA_READ_DATA");
  $cam_state_inst = reg_create_mmr_inst("shub2_01", "SH2_PI_CAM_STATE_READ_DATA");
  $cam_control_inst = reg_create_mmr_inst("shub2_01", "SH2_PI_CAM_CONTROL");

  for (my $qcrb_type = 0; $qcrb_type < 4; $qcrb_type++ ) {
    my $crb_limit = ($qcrb_type < 2) ? 32 : ($qcrb_type == 2) ? 64 : 8;
    for($crb=0; $crb < $crb_limit; $crb++) {
      my $crb_valid = read_cam($qcrb_type, $crb);   
    }
  }

}


sub decode_lut_addr {

  my $val_0 = $_[0];
  my $nsd_val = $val_0 & 0x3fff;
  my $chip_val = ($val_0 >> 14) & 0x3;
  my $ret_msg = "%s chiplet on NASID 0x%04X", $shub2_chiplet_list[$chip_val], $nsd_val;
  return($ret_msg);

}


#-------------------------------------------------------------------------------
1;
